//
//  unicodelib.h
//
//  Copyright (c) 2020 Yuji Hirose. All rights reserved.
//  MIT License
//

#pragma once

#include <cstdlib>
#include <cstring>
#include <string>
#include <unordered_map>
#include <vector>

#if !defined(__cplusplus) || __cplusplus < 201703L
#error "Requires complete C++17 support"
#endif

namespace unicode {

//-----------------------------------------------------------------------------
// General Category
//-----------------------------------------------------------------------------

enum class GeneralCategory {
  Lu,
  Uppercase_Letter = Lu,
  Ll,
  Lowercase_Letter = Ll,
  Lt,
  Titlecase_Letter = Lt,
  Lm,
  Modifier_Letter = Lm,
  Lo,
  Other_Letter = Lo,
  Mn,
  Nonspacing_Mark = Mn,
  Mc,
  Spacing_Mark = Mc,
  Me,
  Enclosing_Mark = Me,
  Nd,
  Decimal_Number = Nd,
  Nl,
  Letter_Number = Nl,
  No,
  Other_Number = No,
  Pc,
  Connector_Punctuation = Pc,
  Pd,
  Dash_Punctuation = Pd,
  Ps,
  Open_Punctuation = Ps,
  Pe,
  Close_Punctuation = Pe,
  Pi,
  Initial_Punctuation = Pi,
  Pf,
  Final_Punctuation = Pf,
  Po,
  Other_Punctuation = Po,
  Sm,
  Math_Symbol = Sm,
  Sc,
  Currency_Symbol = Sc,
  Sk,
  Modifier_Symbol = Sk,
  So,
  Other_Symbol = So,
  Zs,
  Space_Separator = Zs,
  Zl,
  Line_Separator = Zl,
  Zp,
  Paragraph_Separator = Zp,
  Cc,
  Control = Cc,
  Cf,
  Format = Cf,
  Cs,
  Surrogate = Cs,
  Co,
  Private_Use = Co,
  Cn,
  Unassigned = Cn,
};

GeneralCategory general_category(char32_t cp);

bool is_cased_letter_category(GeneralCategory gc);
bool is_letter_category(GeneralCategory gc);
bool is_mark_category(GeneralCategory gc);
bool is_number_category(GeneralCategory gc);
bool is_punctuation_category(GeneralCategory gc);
bool is_symbol_category(GeneralCategory gc);
bool is_separator_category(GeneralCategory gc);
bool is_other_category(GeneralCategory gc);

bool is_cased_letter(char32_t cp);
bool is_letter(char32_t cp);
bool is_mark(char32_t cp);
bool is_number(char32_t cp);
bool is_punctuation(char32_t cp);
bool is_symbol(char32_t cp);
bool is_separator(char32_t cp);
bool is_other(char32_t cp);

//-----------------------------------------------------------------------------
// Property
//-----------------------------------------------------------------------------

bool is_white_space(char32_t cp);
bool is_bidi_control(char32_t cp);
bool is_join_control(char32_t cp);
bool is_dash(char32_t cp);
bool is_hyphen(char32_t cp);
bool is_quotation_mark(char32_t cp);
bool is_terminal_punctuation(char32_t cp);
bool is_other_math(char32_t cp);
bool is_hex_digit(char32_t cp);
bool is_ascii_hex_digit(char32_t cp);
bool is_other_alphabetic(char32_t cp);
bool is_ideographic(char32_t cp);
bool is_diacritic(char32_t cp);
bool is_extender(char32_t cp);
bool is_other_lowercase(char32_t cp);
bool is_other_uppercase(char32_t cp);
bool is_noncharacter_code_point(char32_t cp);
bool is_other_grapheme_extend(char32_t cp);
bool is_ids_binary_operator(char32_t cp);
bool is_ids_trinary_operator(char32_t cp);
bool is_radical(char32_t cp);
bool is_unified_ideograph(char32_t cp);
bool is_other_default_ignorable_code_point(char32_t cp);
bool is_deprecated(char32_t cp);
bool is_soft_dotted(char32_t cp);
bool is_logical_order_exception(char32_t cp);
bool is_other_id_start(char32_t cp);
bool is_other_id_continue(char32_t cp);
bool is_sentence_terminal(char32_t cp);
bool is_variation_selector(char32_t cp);
bool is_pattern_white_space(char32_t cp);
bool is_pattern_syntax(char32_t cp);
bool is_prepended_concatenation_mark(char32_t cp);

//-----------------------------------------------------------------------------
// Derived Property
//-----------------------------------------------------------------------------

bool is_math(char32_t cp);
bool is_alphabetic(char32_t cp);
bool is_lowercase(char32_t cp);
bool is_uppercase(char32_t cp);
bool is_cased(char32_t cp);
bool is_case_ignorable(char32_t cp);
bool is_changes_when_lowercased(char32_t cp);
bool is_changes_when_uppercased(char32_t cp);
bool is_changes_when_titlecased(char32_t cp);
bool is_changes_when_casefolded(char32_t cp);
bool is_changes_when_casemapped(char32_t cp);
bool is_id_start(char32_t cp);
bool is_id_continue(char32_t cp);
bool is_xid_start(char32_t cp);
bool is_xid_continue(char32_t cp);
bool is_default_ignorable_code_point(char32_t cp);
bool is_grapheme_extend(char32_t cp);
bool is_grapheme_base(char32_t cp);
bool is_grapheme_link(char32_t cp);

//-----------------------------------------------------------------------------
// Case
//-----------------------------------------------------------------------------

char32_t simple_uppercase_mapping(char32_t cp);
char32_t simple_lowercase_mapping(char32_t cp);
char32_t simple_titlecase_mapping(char32_t cp);
char32_t simple_case_folding(char32_t cp);

std::u32string to_uppercase(const char32_t *s32, size_t l,
                            const char *lang = nullptr);
std::u32string to_lowercase(const char32_t *s32, size_t l,
                            const char *lang = nullptr);
std::u32string to_titlecase(const char32_t *s32, size_t l,
                            const char *lang = nullptr);
std::u32string to_case_fold(
    const char32_t *s32, size_t l,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false);

bool is_uppercase(const char32_t *s32, size_t l);
bool is_lowercase(const char32_t *s32, size_t l);
bool is_titlecase(const char32_t *s32, size_t l);
bool is_case_fold(const char32_t *s32, size_t l);

bool caseless_match(
    const char32_t *s1, size_t l1, const char32_t *s2, size_t l2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false);

bool canonical_caseless_match(
    const char32_t *s1, size_t l1, const char32_t *s2, size_t l2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false);

bool compatibility_caseless_match(
    const char32_t *s1, size_t l1, const char32_t *s2, size_t l2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false);

//-----------------------------------------------------------------------------
// Text Segmentation
//-----------------------------------------------------------------------------

bool is_graphic_character(char32_t cp);
bool is_base_character(char32_t cp);
bool is_combining_character(char32_t cp);

size_t combining_character_sequence_length(const char32_t *s32, size_t l);
size_t combining_character_sequence_count(const char32_t *s32, size_t l);

size_t extended_combining_character_sequence_length(const char32_t *s32,
                                                    size_t l);
size_t extended_combining_character_sequence_count(const char32_t *s32,
                                                   size_t l);

// The following functions support 'extended' grapheme cluster, not 'legacy'
// graphme cluster.
bool is_grapheme_boundary(const char32_t *s32, size_t l, size_t i);
size_t grapheme_length(const char32_t *s32, size_t l);
size_t grapheme_count(const char32_t *s32, size_t l);

bool is_word_boundary(const char32_t *s32, size_t l, size_t i);

bool is_sentence_boundary(const char32_t *s32, size_t l, size_t i);

//-----------------------------------------------------------------------------
// Block
//-----------------------------------------------------------------------------

// This is generated from 'Blocks.txt' in Unicode database.
// `python scripts/gen_property_values.py < UCD/Blocks.txt`
enum class Block {
  Unassigned,
  BasicLatin,
  Latin1Supplement,
  LatinExtendedA,
  LatinExtendedB,
  IPAExtensions,
  SpacingModifierLetters,
  CombiningDiacriticalMarks,
  GreekAndCoptic,
  Cyrillic,
  CyrillicSupplement,
  Armenian,
  Hebrew,
  Arabic,
  Syriac,
  ArabicSupplement,
  Thaana,
  NKo,
  Samaritan,
  Mandaic,
  SyriacSupplement,
  ArabicExtendedA,
  Devanagari,
  Bengali,
  Gurmukhi,
  Gujarati,
  Oriya,
  Tamil,
  Telugu,
  Kannada,
  Malayalam,
  Sinhala,
  Thai,
  Lao,
  Tibetan,
  Myanmar,
  Georgian,
  HangulJamo,
  Ethiopic,
  EthiopicSupplement,
  Cherokee,
  UnifiedCanadianAboriginalSyllabics,
  Ogham,
  Runic,
  Tagalog,
  Hanunoo,
  Buhid,
  Tagbanwa,
  Khmer,
  Mongolian,
  UnifiedCanadianAboriginalSyllabicsExtended,
  Limbu,
  TaiLe,
  NewTaiLue,
  KhmerSymbols,
  Buginese,
  TaiTham,
  CombiningDiacriticalMarksExtended,
  Balinese,
  Sundanese,
  Batak,
  Lepcha,
  OlChiki,
  CyrillicExtendedC,
  GeorgianExtended,
  SundaneseSupplement,
  VedicExtensions,
  PhoneticExtensions,
  PhoneticExtensionsSupplement,
  CombiningDiacriticalMarksSupplement,
  LatinExtendedAdditional,
  GreekExtended,
  GeneralPunctuation,
  SuperscriptsAndSubscripts,
  CurrencySymbols,
  CombiningDiacriticalMarksForSymbols,
  LetterlikeSymbols,
  NumberForms,
  Arrows,
  MathematicalOperators,
  MiscellaneousTechnical,
  ControlPictures,
  OpticalCharacterRecognition,
  EnclosedAlphanumerics,
  BoxDrawing,
  BlockElements,
  GeometricShapes,
  MiscellaneousSymbols,
  Dingbats,
  MiscellaneousMathematicalSymbolsA,
  SupplementalArrowsA,
  BraillePatterns,
  SupplementalArrowsB,
  MiscellaneousMathematicalSymbolsB,
  SupplementalMathematicalOperators,
  MiscellaneousSymbolsAndArrows,
  Glagolitic,
  LatinExtendedC,
  Coptic,
  GeorgianSupplement,
  Tifinagh,
  EthiopicExtended,
  CyrillicExtendedA,
  SupplementalPunctuation,
  CJKRadicalsSupplement,
  KangxiRadicals,
  IdeographicDescriptionCharacters,
  CJKSymbolsAndPunctuation,
  Hiragana,
  Katakana,
  Bopomofo,
  HangulCompatibilityJamo,
  Kanbun,
  BopomofoExtended,
  CJKStrokes,
  KatakanaPhoneticExtensions,
  EnclosedCJKLettersAndMonths,
  CJKCompatibility,
  CJKUnifiedIdeographsExtensionA,
  YijingHexagramSymbols,
  CJKUnifiedIdeographs,
  YiSyllables,
  YiRadicals,
  Lisu,
  Vai,
  CyrillicExtendedB,
  Bamum,
  ModifierToneLetters,
  LatinExtendedD,
  SylotiNagri,
  CommonIndicNumberForms,
  PhagsPa,
  Saurashtra,
  DevanagariExtended,
  KayahLi,
  Rejang,
  HangulJamoExtendedA,
  Javanese,
  MyanmarExtendedB,
  Cham,
  MyanmarExtendedA,
  TaiViet,
  MeeteiMayekExtensions,
  EthiopicExtendedA,
  LatinExtendedE,
  CherokeeSupplement,
  MeeteiMayek,
  HangulSyllables,
  HangulJamoExtendedB,
  HighSurrogates,
  HighPrivateUseSurrogates,
  LowSurrogates,
  PrivateUseArea,
  CJKCompatibilityIdeographs,
  AlphabeticPresentationForms,
  ArabicPresentationFormsA,
  VariationSelectors,
  VerticalForms,
  CombiningHalfMarks,
  CJKCompatibilityForms,
  SmallFormVariants,
  ArabicPresentationFormsB,
  HalfwidthAndFullwidthForms,
  Specials,
  LinearBSyllabary,
  LinearBIdeograms,
  AegeanNumbers,
  AncientGreekNumbers,
  AncientSymbols,
  PhaistosDisc,
  Lycian,
  Carian,
  CopticEpactNumbers,
  OldItalic,
  Gothic,
  OldPermic,
  Ugaritic,
  OldPersian,
  Deseret,
  Shavian,
  Osmanya,
  Osage,
  Elbasan,
  CaucasianAlbanian,
  LinearA,
  CypriotSyllabary,
  ImperialAramaic,
  Palmyrene,
  Nabataean,
  Hatran,
  Phoenician,
  Lydian,
  MeroiticHieroglyphs,
  MeroiticCursive,
  Kharoshthi,
  OldSouthArabian,
  OldNorthArabian,
  Manichaean,
  Avestan,
  InscriptionalParthian,
  InscriptionalPahlavi,
  PsalterPahlavi,
  OldTurkic,
  OldHungarian,
  HanifiRohingya,
  RumiNumeralSymbols,
  Yezidi,
  OldSogdian,
  Sogdian,
  Chorasmian,
  Elymaic,
  Brahmi,
  Kaithi,
  SoraSompeng,
  Chakma,
  Mahajani,
  Sharada,
  SinhalaArchaicNumbers,
  Khojki,
  Multani,
  Khudawadi,
  Grantha,
  Newa,
  Tirhuta,
  Siddham,
  Modi,
  MongolianSupplement,
  Takri,
  Ahom,
  Dogra,
  WarangCiti,
  DivesAkuru,
  Nandinagari,
  ZanabazarSquare,
  Soyombo,
  PauCinHau,
  Bhaiksuki,
  Marchen,
  MasaramGondi,
  GunjalaGondi,
  Makasar,
  LisuSupplement,
  TamilSupplement,
  Cuneiform,
  CuneiformNumbersAndPunctuation,
  EarlyDynasticCuneiform,
  EgyptianHieroglyphs,
  EgyptianHieroglyphFormatControls,
  AnatolianHieroglyphs,
  BamumSupplement,
  Mro,
  BassaVah,
  PahawhHmong,
  Medefaidrin,
  Miao,
  IdeographicSymbolsAndPunctuation,
  Tangut,
  TangutComponents,
  KhitanSmallScript,
  TangutSupplement,
  KanaSupplement,
  KanaExtendedA,
  SmallKanaExtension,
  Nushu,
  Duployan,
  ShorthandFormatControls,
  ByzantineMusicalSymbols,
  MusicalSymbols,
  AncientGreekMusicalNotation,
  MayanNumerals,
  TaiXuanJingSymbols,
  CountingRodNumerals,
  MathematicalAlphanumericSymbols,
  SuttonSignWriting,
  GlagoliticSupplement,
  NyiakengPuachueHmong,
  Wancho,
  MendeKikakui,
  Adlam,
  IndicSiyaqNumbers,
  OttomanSiyaqNumbers,
  ArabicMathematicalAlphabeticSymbols,
  MahjongTiles,
  DominoTiles,
  PlayingCards,
  EnclosedAlphanumericSupplement,
  EnclosedIdeographicSupplement,
  MiscellaneousSymbolsAndPictographs,
  Emoticons,
  OrnamentalDingbats,
  TransportAndMapSymbols,
  AlchemicalSymbols,
  GeometricShapesExtended,
  SupplementalArrowsC,
  SupplementalSymbolsAndPictographs,
  ChessSymbols,
  SymbolsAndPictographsExtendedA,
  SymbolsForLegacyComputing,
  CJKUnifiedIdeographsExtensionB,
  CJKUnifiedIdeographsExtensionC,
  CJKUnifiedIdeographsExtensionD,
  CJKUnifiedIdeographsExtensionE,
  CJKUnifiedIdeographsExtensionF,
  CJKCompatibilityIdeographsSupplement,
  CJKUnifiedIdeographsExtensionG,
  Tags,
  VariationSelectorsSupplement,
  SupplementaryPrivateUseAreaA,
  SupplementaryPrivateUseAreaB,
};

Block block(char32_t cp);

//-----------------------------------------------------------------------------
// Script
//-----------------------------------------------------------------------------

// This is generated from 'Scripts.txt' in Unicode database.
// `python scripts/gen_property_values.py < UCD/Scripts.txt`
enum class Script {
  Unassigned,
  Common,
  Latin,
  Greek,
  Cyrillic,
  Armenian,
  Hebrew,
  Arabic,
  Syriac,
  Thaana,
  Devanagari,
  Bengali,
  Gurmukhi,
  Gujarati,
  Oriya,
  Tamil,
  Telugu,
  Kannada,
  Malayalam,
  Sinhala,
  Thai,
  Lao,
  Tibetan,
  Myanmar,
  Georgian,
  Hangul,
  Ethiopic,
  Cherokee,
  Canadian_Aboriginal,
  Ogham,
  Runic,
  Khmer,
  Mongolian,
  Hiragana,
  Katakana,
  Bopomofo,
  Han,
  Yi,
  Old_Italic,
  Gothic,
  Deseret,
  Inherited,
  Tagalog,
  Hanunoo,
  Buhid,
  Tagbanwa,
  Limbu,
  Tai_Le,
  Linear_B,
  Ugaritic,
  Shavian,
  Osmanya,
  Cypriot,
  Braille,
  Buginese,
  Coptic,
  New_Tai_Lue,
  Glagolitic,
  Tifinagh,
  Syloti_Nagri,
  Old_Persian,
  Kharoshthi,
  Balinese,
  Cuneiform,
  Phoenician,
  Phags_Pa,
  Nko,
  Sundanese,
  Lepcha,
  Ol_Chiki,
  Vai,
  Saurashtra,
  Kayah_Li,
  Rejang,
  Lycian,
  Carian,
  Lydian,
  Cham,
  Tai_Tham,
  Tai_Viet,
  Avestan,
  Egyptian_Hieroglyphs,
  Samaritan,
  Lisu,
  Bamum,
  Javanese,
  Meetei_Mayek,
  Imperial_Aramaic,
  Old_South_Arabian,
  Inscriptional_Parthian,
  Inscriptional_Pahlavi,
  Old_Turkic,
  Kaithi,
  Batak,
  Brahmi,
  Mandaic,
  Chakma,
  Meroitic_Cursive,
  Meroitic_Hieroglyphs,
  Miao,
  Sharada,
  Sora_Sompeng,
  Takri,
  Caucasian_Albanian,
  Bassa_Vah,
  Duployan,
  Elbasan,
  Grantha,
  Pahawh_Hmong,
  Khojki,
  Linear_A,
  Mahajani,
  Manichaean,
  Mende_Kikakui,
  Modi,
  Mro,
  Old_North_Arabian,
  Nabataean,
  Palmyrene,
  Pau_Cin_Hau,
  Old_Permic,
  Psalter_Pahlavi,
  Siddham,
  Khudawadi,
  Tirhuta,
  Warang_Citi,
  Ahom,
  Anatolian_Hieroglyphs,
  Hatran,
  Multani,
  Old_Hungarian,
  SignWriting,
  Adlam,
  Bhaiksuki,
  Marchen,
  Newa,
  Osage,
  Tangut,
  Masaram_Gondi,
  Nushu,
  Soyombo,
  Zanabazar_Square,
  Dogra,
  Gunjala_Gondi,
  Makasar,
  Medefaidrin,
  Hanifi_Rohingya,
  Sogdian,
  Old_Sogdian,
  Elymaic,
  Nandinagari,
  Nyiakeng_Puachue_Hmong,
  Wancho,
  Chorasmian,
  Dives_Akuru,
  Khitan_Small_Script,
  Yezidi,
};

Script script(char32_t cp);
bool is_script(Script sc, char32_t cp);

//-----------------------------------------------------------------------------
// Normalization
//-----------------------------------------------------------------------------

std::u32string to_nfc(const char32_t *s32, size_t l);
std::u32string to_nfd(const char32_t *s32, size_t l);
std::u32string to_nfkc(const char32_t *s32, size_t l);
std::u32string to_nfkd(const char32_t *s32, size_t l);

//-----------------------------------------------------------------------------
// Inline Wrapper functions
//-----------------------------------------------------------------------------

inline std::u32string to_uppercase(const std::u32string_view s32,
                                   const char *lang = nullptr) {
  return to_uppercase(s32.data(), s32.length(), lang);
}

inline std::u32string to_uppercase(const char32_t *s32,
                                   const char *lang = nullptr) {
  return to_uppercase(s32, std::char_traits<char32_t>::length(s32), lang);
}

inline std::u32string to_lowercase(const std::u32string_view s32,
                                   const char *lang = nullptr) {
  return to_lowercase(s32.data(), s32.length(), lang);
}

inline std::u32string to_lowercase(const char32_t *s32,
                                   const char *lang = nullptr) {
  return to_lowercase(s32, std::char_traits<char32_t>::length(s32), lang);
}

inline std::u32string to_titlecase(const std::u32string_view s32,
                                   const char *lang = nullptr) {
  return to_titlecase(s32.data(), s32.length(), lang);
}

inline std::u32string to_titlecase(const char32_t *s32,
                                   const char *lang = nullptr) {
  return to_titlecase(s32, std::char_traits<char32_t>::length(s32), lang);
}

inline std::u32string to_case_fold(
    const std::u32string_view s32,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false) {
  return to_case_fold(s32.data(), s32.length(),
                      special_case_for_uppercase_I_and_dotted_uppercase_I);
}

inline std::u32string to_case_fold(
    const char32_t *s32,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false) {
  return to_case_fold(s32, std::char_traits<char32_t>::length(s32),
                      special_case_for_uppercase_I_and_dotted_uppercase_I);
}

inline bool is_uppercase(const std::u32string_view s32) {
  return is_uppercase(s32.data(), s32.length());
}

inline bool is_uppercase(const char32_t *s32) {
  return is_uppercase(s32, std::char_traits<char32_t>::length(s32));
}

inline bool is_lowercase(const std::u32string_view s32) {
  return is_lowercase(s32.data(), s32.length());
}

inline bool is_lowercase(const char32_t *s32) {
  return is_lowercase(s32, std::char_traits<char32_t>::length(s32));
}

inline bool is_titlecase(const std::u32string_view s32) {
  return is_titlecase(s32.data(), s32.length());
}

inline bool is_titlecase(const char32_t *s32) {
  return is_titlecase(s32, std::char_traits<char32_t>::length(s32));
}

inline bool is_case_fold(const std::u32string_view s32) {
  return is_case_fold(s32.data(), s32.length());
}

inline bool is_case_fold(const char32_t *s32) {
  return is_case_fold(s32, std::char_traits<char32_t>::length(s32));
}

inline bool caseless_match(
    const std::u32string_view s1, const std::u32string_view s2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false) {
  return caseless_match(s1.data(), s1.length(), s2.data(), s2.length(),
                        special_case_for_uppercase_I_and_dotted_uppercase_I);
}

inline bool caseless_match(
    const char32_t *s1, const char32_t *s2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false) {
  return caseless_match(s1, std::char_traits<char32_t>::length(s1), s2,
                        std::char_traits<char32_t>::length(s2),
                        special_case_for_uppercase_I_and_dotted_uppercase_I);
}

inline bool canonical_caseless_match(
    const std::u32string_view s1, const std::u32string_view s2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false) {
  return canonical_caseless_match(
      s1.data(), s1.length(), s2.data(), s2.length(),
      special_case_for_uppercase_I_and_dotted_uppercase_I);
}

inline bool canonical_caseless_match(
    const char32_t *s1, const char32_t *s2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false) {
  return canonical_caseless_match(
      s1, std::char_traits<char32_t>::length(s1), s2,
      std::char_traits<char32_t>::length(s2),
      special_case_for_uppercase_I_and_dotted_uppercase_I);
}

inline bool compatibility_caseless_match(
    const std::u32string_view s1, const std::u32string_view s2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false) {
  return compatibility_caseless_match(
      s1.data(), s1.length(), s2.data(), s2.length(),
      special_case_for_uppercase_I_and_dotted_uppercase_I);
}

inline bool compatibility_caseless_match(
    const char32_t *s1, const char32_t *s2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I = false) {
  return compatibility_caseless_match(
      s1, std::char_traits<char32_t>::length(s1), s2,
      std::char_traits<char32_t>::length(s2),
      special_case_for_uppercase_I_and_dotted_uppercase_I);
}

inline std::u32string to_nfc(const std::u32string_view s32) {
  return to_nfc(s32.data(), s32.length());
}

inline std::u32string to_nfc(const char32_t *s32) {
  return to_nfc(s32, std::char_traits<char32_t>::length(s32));
}

inline std::u32string to_nfd(const std::u32string_view s32) {
  return to_nfd(s32.data(), s32.length());
}

inline std::u32string to_nfd(const char32_t *s32) {
  return to_nfd(s32, std::char_traits<char32_t>::length(s32));
}

inline std::u32string to_nfkc(const std::u32string_view s32) {
  return to_nfkc(s32.data(), s32.length());
}

inline std::u32string to_nfkc(const char32_t *s32) {
  return to_nfkc(s32, std::char_traits<char32_t>::length(s32));
}

inline std::u32string to_nfkd(const std::u32string_view s32) {
  return to_nfkd(s32.data(), s32.length());
}

inline std::u32string to_nfkd(const char32_t *s32) {
  return to_nfkd(s32, std::char_traits<char32_t>::length(s32));
}

inline size_t grapheme_count(const std::u32string_view s32) {
  return grapheme_count(s32.data(), s32.length());
}

inline size_t grapheme_count(const char32_t *s32) {
  return grapheme_count(s32, std::char_traits<char32_t>::length(s32));
}

inline size_t grapheme_length(const std::u32string_view s32) {
  return grapheme_length(s32.data(), s32.length());
}

inline size_t grapheme_length(const char32_t *s32) {
  return grapheme_length(s32, std::char_traits<char32_t>::length(s32));
}

// ----------------------------------------------------------------------------

/*
 * Implementation
 */

enum class CaseMappingType {
  Upper = 0,
  Lower,
  Title,
};

enum class SpecialCasingContext {
  Unassigned,
  Final_Sigma,
  Not_Final_Sigma,
  After_Soft_Dotted,
  More_Above,
  Before_Dot,
  Not_Before_Dot,
  After_I,
};

struct SpecialCasing {
  const char32_t *lower;
  const char32_t *title;
  const char32_t *upper;
  const char *language;
  SpecialCasingContext context;

  const char32_t *case_mapping_codes(CaseMappingType type) const {
    switch (type) {
      case CaseMappingType::Upper:
        return upper;
      case CaseMappingType::Lower:
        return lower;
      case CaseMappingType::Title:
        return title;
    }
    // NOTREACHED
    return nullptr;
  }
};

struct CaseFolding {
  const char32_t C;
  const char32_t S;
  const char32_t *F;
  const char32_t T;
};

struct NormalizationProperties {
  int combining_class;
  const char *compat_format;
  const char32_t *codes;
};

// This is generated from 'GraphemeBreakProperty.txt' in Unicode database.
// `python scripts/gen_property_values.py <
// UCD/auxiliary/GraphemeBreakProperty.txt`
enum class GraphemeBreak {
  Unassigned,
  Prepend,
  CR,
  LF,
  Control,
  Extend,
  Regional_Indicator,
  SpacingMark,
  L,
  V,
  T,
  LV,
  LVT,
  ZWJ,
};

// This is generated from 'WordBreakProperty.txt' in Unicode database.
// `python scripts/gen_property_values.py < UCD/auxiliary/WordBreakProperty.txt`
enum class WordBreak {
  Unassigned,
  Double_Quote,
  Single_Quote,
  Hebrew_Letter,
  CR,
  LF,
  Newline,
  Extend,
  Regional_Indicator,
  Format,
  Katakana,
  ALetter,
  MidLetter,
  MidNum,
  MidNumLet,
  Numeric,
  ExtendNumLet,
  ZWJ,
  WSegSpace,
};

// This is generated from 'SentenceBreakProperty.txt' in Unicode database.
// `python scripts/gen_property_values.py <
// UCD/auxiliary/SentenceBreakProperty.txt`
enum class SentenceBreak {
  Unassigned,
  CR,
  LF,
  Extend,
  Sep,
  Format,
  Sp,
  Lower,
  Upper,
  OLetter,
  Numeric,
  ATerm,
  STerm,
  Close,
  SContinue,
};

// This is generated from 'emoji-data.txt' in Unicode database.
// `python scripts/gen_property_values.py < UCD/emoji/emoji-data.txt`
enum class Emoji {
  Unassigned,
  Emoji,
  Emoji_Presentation,
  Emoji_Modifier,
  Emoji_Modifier_Base,
  Emoji_Component,
  Extended_Pictographic,
};

//-----------------------------------------------------------------------------
// Unicode Data
//-----------------------------------------------------------------------------

// This is generated from Unicode database.
// `python scripts/gen_tables.py UCD`
namespace _general_category_properties {
using T = GeneralCategory;
const auto D = GeneralCategory::Cn;
static const size_t _block_size = 128;
static const GeneralCategory _0[] = { T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Zs,T::Po,T::Po,T::Po,T::Sc,T::Po,T::Po,T::Po,T::Ps,T::Pe,T::Po,T::Sm,T::Po,T::Pd,T::Po,T::Po,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Po,T::Po,T::Sm,T::Sm,T::Sm,T::Po,T::Po,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ps,T::Po,T::Pe,T::Sk,T::Pc,T::Sk,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ps,T::Sm,T::Pe,T::Sm,T::Cc, };
static const GeneralCategory _1[] = { T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Cc,T::Zs,T::Po,T::Sc,T::Sc,T::Sc,T::Sc,T::So,T::Po,T::Sk,T::So,T::Lo,T::Pi,T::Sm,T::Cf,T::So,T::Sk,T::So,T::Sm,T::No,T::No,T::Sk,T::Ll,T::Po,T::Po,T::Sk,T::No,T::Lo,T::Pf,T::No,T::No,T::No,T::Po,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Sm,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _2[] = { T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll, };
static const GeneralCategory _3[] = { T::Ll,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll,T::Ll,T::Lo,T::Lu,T::Ll,T::Ll,T::Ll,T::Lo,T::Lo,T::Lo,T::Lo,T::Lu,T::Lt,T::Ll,T::Lu,T::Lt,T::Ll,T::Lu,T::Lt,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Lu,T::Lt,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll, };
static const GeneralCategory _4[] = { T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _5[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lo,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Sk,T::Sk,T::Sk,T::Sk,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Lm,T::Sk,T::Lm,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk, };
static const GeneralCategory _6[] = { T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lu,T::Ll,T::Lu,T::Ll,T::Lm,T::Sk,T::Lu,T::Ll,D,D,T::Lm,T::Ll,T::Ll,T::Ll,T::Po,T::Lu, };
static const GeneralCategory _7[] = { D,D,D,D,T::Sk,T::Sk,T::Lu,T::Po,T::Lu,T::Lu,T::Lu,D,T::Lu,D,T::Lu,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Ll,T::Sm,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu, };
static const GeneralCategory _8[] = { T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll, };
static const GeneralCategory _9[] = { T::Lu,T::Ll,T::So,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Me,T::Me,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll, };
static const GeneralCategory _10[] = { T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,D,T::Lm,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _11[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Po,T::Pd,D,D,T::So,T::So,T::Sc,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Pd,T::Mn,T::Po,T::Mn,T::Mn,T::Po,T::Mn,T::Mn,T::Po,T::Mn,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _12[] = { T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Sm,T::Sm,T::Sm,T::Po,T::Po,T::Sc,T::Po,T::Po,T::So,T::So,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Cf,D,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lm,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Po,T::Po,T::Po,T::Po,T::Lo,T::Lo,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _13[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Po,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Cf,T::So,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lm,T::Lm,T::Mn,T::Mn,T::So,T::Mn,T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Lo,T::Lo,T::Lo,T::So,T::So,T::Lo, };
static const GeneralCategory _14[] = { T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,D,T::Cf,T::Lo,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _15[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lm,T::Lm,T::So,T::Po,T::Po,T::Po,T::Lm,D,D,T::Mn,T::Sc,T::Sc, };
static const GeneralCategory _16[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Lm,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lm,T::Mn,T::Mn,T::Mn,T::Lm,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,D,D,T::Po,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _17[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Cf,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn, };
static const GeneralCategory _18[] = { T::Mn,T::Mn,T::Mn,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mc,T::Mn,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Mc,T::Mc,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Po,T::Po,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Po,T::Lm,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _19[] = { T::Lo,T::Mn,T::Mc,T::Mc,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Mn,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,D,D,T::Mc,T::Mc,D,D,T::Mc,T::Mc,T::Mn,T::Lo,D,D,D,D,D,D,D,D,T::Mc,D,D,D,D,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Lo,T::Lo,T::Sc,T::Sc,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::Sc,T::Lo,T::Po,T::Mn,D, };
static const GeneralCategory _20[] = { D,T::Mn,T::Mn,T::Mc,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,D,D,T::Mn,D,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,D,D,D,D,T::Mn,T::Mn,D,D,T::Mn,T::Mn,T::Mn,D,D,D,T::Mn,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Mn,T::Po,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _21[] = { D,T::Mn,T::Mn,T::Mc,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Mn,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Mn,T::Mn,T::Mc,D,T::Mc,T::Mc,T::Mn,D,D,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Mn,T::Mn,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Po,T::Sc,D,D,D,D,D,D,D,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn, };
static const GeneralCategory _22[] = { D,T::Mn,T::Mc,T::Mc,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Mn,T::Lo,T::Mc,T::Mn,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,D,D,T::Mc,T::Mc,D,D,T::Mc,T::Mc,T::Mn,D,D,D,D,D,D,D,T::Mn,T::Mn,T::Mc,D,D,D,D,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::So,T::Lo,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,D, };
static const GeneralCategory _23[] = { D,D,T::Mn,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,T::Lo,T::Lo,D,T::Lo,D,T::Lo,T::Lo,D,D,D,T::Lo,T::Lo,D,D,D,T::Lo,T::Lo,T::Lo,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,T::Mc,T::Mc,T::Mn,T::Mc,T::Mc,D,D,D,T::Mc,T::Mc,T::Mc,D,T::Mc,T::Mc,T::Mc,T::Mn,D,D,T::Lo,D,D,D,D,D,D,T::Mc,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::Sc,T::So,D,D,D,D,D, };
static const GeneralCategory _24[] = { T::Mn,T::Mc,T::Mc,T::Mc,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,T::Lo,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,D,T::Mn,T::Mn,T::Mn,D,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,D,D,D,T::Mn,T::Mn,D,T::Lo,T::Lo,T::Lo,D,D,D,D,D,T::Lo,T::Lo,T::Mn,T::Mn,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,D,T::Po,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So, };
static const GeneralCategory _25[] = { T::Lo,T::Mn,T::Mc,T::Mc,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Mn,T::Lo,T::Mc,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,D,T::Mn,T::Mc,T::Mc,D,T::Mc,T::Mc,T::Mn,T::Mn,D,D,D,D,D,D,D,T::Mc,T::Mc,D,D,D,D,D,D,D,T::Lo,D,T::Lo,T::Lo,T::Mn,T::Mn,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _26[] = { T::Mn,T::Mn,T::Mc,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Mc,T::Mc,T::Mc,D,T::Mc,T::Mc,T::Mc,T::Mn,T::Lo,T::So,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Mc,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _27[] = { D,T::Mn,T::Mc,T::Mc,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,T::Mn,D,D,D,D,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,D,T::Mn,D,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,T::Mc,T::Mc,T::Po,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _28[] = { D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,T::Sc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lm,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _29[] = { D,T::Lo,T::Lo,D,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lm,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _30[] = { T::Lo,T::So,T::So,T::So,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::So,T::Po,T::So,T::So,T::So,T::Mn,T::Mn,T::So,T::So,T::So,T::So,T::So,T::So,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::Mn,T::So,T::Mn,T::So,T::Mn,T::Ps,T::Pe,T::Ps,T::Pe,T::Mc,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc, };
static const GeneralCategory _31[] = { T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Mn,T::So,T::So,T::So,T::So,T::So,T::So,D,T::So,T::So,T::Po,T::Po,T::Po,T::Po,T::Po,T::So,T::So,T::So,T::So,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _32[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Lo,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Lo,T::Mc,T::Mc,T::Mc,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _33[] = { T::Lo,T::Lo,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Lo,T::Mc,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Mc,T::Mc,T::Mc,T::Mn,T::So,T::So,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,T::Lu,D,D,D,D,D,T::Lu,D,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Po,T::Lm,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _36[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _37[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _38[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Mn,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D, };
static const GeneralCategory _39[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,D, };
static const GeneralCategory _40[] = { T::Pd,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _44[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::So,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _45[] = { T::Zs,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Ps,T::Pe,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Po,T::Po,T::Po,T::Nl,T::Nl,T::Nl,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D, };
static const GeneralCategory _46[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Po,T::Po,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,D,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _47[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Lm,T::Po,T::Po,T::Po,T::Sc,T::Lo,T::Mn,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D, };
static const GeneralCategory _48[] = { T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Pd,T::Po,T::Po,T::Po,T::Po,T::Mn,T::Mn,T::Mn,T::Cf,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lm,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D, };
static const GeneralCategory _49[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Lo,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _50[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mc,T::Mc,T::Mc,D,D,D,D,T::Mc,T::Mc,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,D,D,D,D,T::So,D,D,D,T::Po,T::Po,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _51[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::No,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _52[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,D,D,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mn,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Mn,T::Mc,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,T::Mn, };
static const GeneralCategory _53[] = { T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Lm,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Me,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _54[] = { T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Mc,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D, };
static const GeneralCategory _55[] = { T::Mn,T::Mn,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mc,T::Mn,T::Mn,T::Mc,T::Mc,T::Mc,T::Mn,T::Mc,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,D,D,D,D,D,D,D,D,T::Po,T::Po,T::Po,T::Po, };
static const GeneralCategory _56[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,D,D,D,T::Po,T::Po,T::Po,T::Po,T::Po,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,T::Lo,T::Lo,T::Lo,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Po,T::Po, };
static const GeneralCategory _57[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,D,D,D,D,D,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,D,T::Lu,T::Lu,T::Lu,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,T::Mn,T::Mn,T::Mn,T::Po,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Lo,T::Lo,T::Mc,T::Mn,T::Mn,T::Lo,D,D,D,D,D, };
static const GeneralCategory _58[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _59[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn, };
static const GeneralCategory _60[] = { T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll, };
static const GeneralCategory _61[] = { T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll, };
static const GeneralCategory _62[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,T::Lu,D,T::Lu,D,T::Lu,D,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,D, };
static const GeneralCategory _63[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Lt,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lt,T::Sk,T::Ll,T::Sk,T::Sk,T::Sk,T::Ll,T::Ll,T::Ll,D,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lt,T::Sk,T::Sk,T::Sk,T::Ll,T::Ll,T::Ll,T::Ll,D,D,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,D,T::Sk,T::Sk,T::Sk,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Sk,T::Sk,T::Sk,D,D,T::Ll,T::Ll,T::Ll,D,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lt,T::Sk,T::Sk,D, };
static const GeneralCategory _64[] = { T::Zs,T::Zs,T::Zs,T::Zs,T::Zs,T::Zs,T::Zs,T::Zs,T::Zs,T::Zs,T::Zs,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Pd,T::Pd,T::Pd,T::Pd,T::Pd,T::Pd,T::Po,T::Po,T::Pi,T::Pf,T::Ps,T::Pi,T::Pi,T::Pf,T::Ps,T::Pi,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Zl,T::Zp,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Zs,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Pi,T::Pf,T::Po,T::Po,T::Po,T::Po,T::Pc,T::Pc,T::Po,T::Po,T::Po,T::Sm,T::Ps,T::Pe,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Sm,T::Po,T::Pc,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Zs,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,D,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::No,T::Lm,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::Sm,T::Sm,T::Sm,T::Ps,T::Pe,T::Lm, };
static const GeneralCategory _65[] = { T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::Sm,T::Sm,T::Sm,T::Ps,T::Pe,D,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,D,D,D,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,T::Sc,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Me,T::Me,T::Me,T::Me,T::Mn,T::Me,T::Me,T::Me,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _66[] = { T::So,T::So,T::Lu,T::So,T::So,T::So,T::So,T::Lu,T::So,T::So,T::Ll,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Ll,T::So,T::Lu,T::So,T::So,T::Sm,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::So,T::So,T::So,T::So,T::So,T::So,T::Lu,T::So,T::Lu,T::So,T::Lu,T::So,T::Lu,T::Lu,T::Lu,T::Lu,T::So,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Lo,T::Lo,T::Lo,T::Lo,T::Ll,T::So,T::So,T::Ll,T::Ll,T::Lu,T::Lu,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::So,T::Sm,T::So,T::So,T::Ll,T::So,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl, };
static const GeneralCategory _67[] = { T::Nl,T::Nl,T::Nl,T::Lu,T::Ll,T::Nl,T::Nl,T::Nl,T::Nl,T::No,T::So,T::So,D,D,D,D,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::So,T::So,T::So,T::So,T::So,T::Sm,T::Sm,T::So,T::So,T::So,T::So,T::Sm,T::So,T::So,T::Sm,T::So,T::So,T::Sm,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::Sm,T::So,T::So,T::Sm,T::So,T::Sm,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm, };
static const GeneralCategory _70[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Ps,T::Pe,T::Ps,T::Pe,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::Sm,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Ps,T::Pe,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::So,T::So,T::So, };
static const GeneralCategory _71[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _72[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No, };
static const GeneralCategory _73[] = { T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No, };
static const GeneralCategory _75[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm, };
static const GeneralCategory _76[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _78[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No, };
static const GeneralCategory _79[] = { T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Ps,T::Pe,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm, };
static const GeneralCategory _83[] = { T::Sm,T::Sm,T::Sm,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Ps,T::Pe,T::Ps,T::Pe,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Ps,T::Pe,T::Sm,T::Sm, };
static const GeneralCategory _86[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::So,T::So,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::Sm,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _87[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _88[] = { T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lm,T::Lm,T::Lu,T::Lu, };
static const GeneralCategory _89[] = { T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::So,T::So,T::So,T::So,T::So,T::So,T::Lu,T::Ll,T::Lu,T::Ll,T::Mn,T::Mn,T::Mn,T::Lu,T::Ll,D,D,D,D,D,T::Po,T::Po,T::Po,T::Po,T::No,T::Po,T::Po, };
static const GeneralCategory _90[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,T::Ll,D,D,D,D,D,T::Ll,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,T::Lm,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Mn, };
static const GeneralCategory _91[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn, };
static const GeneralCategory _92[] = { T::Po,T::Po,T::Pi,T::Pf,T::Pi,T::Pf,T::Po,T::Po,T::Po,T::Pi,T::Pf,T::Po,T::Pi,T::Pf,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Pd,T::Po,T::Po,T::Pd,T::Po,T::Pi,T::Pf,T::Po,T::Po,T::Pi,T::Pf,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Po,T::Po,T::Po,T::Po,T::Po,T::Lm,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Pd,T::Pd,T::Po,T::Po,T::Po,T::Po,T::Pd,T::Po,T::Ps,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::So,T::So,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _93[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _95[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D, };
static const GeneralCategory _96[] = { T::Zs,T::Po,T::Po,T::Po,T::So,T::Lm,T::Lo,T::Nl,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::So,T::So,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Pd,T::Ps,T::Pe,T::Pe,T::So,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Pd,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::So,T::So,T::Nl,T::Nl,T::Nl,T::Lm,T::Lo,T::Po,T::So,T::So,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _97[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Mn,T::Mn,T::Sk,T::Sk,T::Lm,T::Lm,T::Lo,T::Pd,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Po,T::Lm,T::Lm,T::Lm,T::Lo, };
static const GeneralCategory _98[] = { D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _99[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::So,T::So,T::No,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _100[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _101[] = { T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _155[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _319[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D, };
static const GeneralCategory _320[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lm,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _329[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Po,T::Po, };
static const GeneralCategory _332[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lm,T::Po,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lo,T::Mn,T::Me,T::Me,T::Me,T::Po,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Lm, };
static const GeneralCategory _333[] = { T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lm,T::Lm,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D, };
static const GeneralCategory _334[] = { T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Sk,T::Sk,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Ll, };
static const GeneralCategory _335[] = { T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lm,T::Sk,T::Sk,T::Lu,T::Ll,T::Lu,T::Ll,T::Lo,T::Lu,T::Ll,T::Lu,T::Ll,T::Ll,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,T::Lu,T::Ll,D,D,T::Lu,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Lu,T::Ll,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lu,T::Ll,T::Lo,T::Lm,T::Lm,T::Ll,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _336[] = { T::Lo,T::Lo,T::Mn,T::Lo,T::Lo,T::Lo,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mn,T::Mn,T::Mc,T::So,T::So,T::So,T::So,T::Mn,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::So,T::Sc,T::So,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D, };
static const GeneralCategory _337[] = { T::Mc,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,D,D,D,D,D,D,D,D,T::Po,T::Po,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Po,T::Po,T::Po,T::Lo,T::Po,T::Lo,T::Lo,T::Mn, };
static const GeneralCategory _338[] = { T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,D,D,D,D,D,D,D,D,D,D,D,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D, };
static const GeneralCategory _339[] = { T::Mn,T::Mn,T::Mn,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mc,T::Mc,T::Mc,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,D,T::Lm,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Lm,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D, };
static const GeneralCategory _340[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mc,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,T::Po,T::Po,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lm,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::So,T::So,T::So,T::Lo,T::Mc,T::Mn,T::Mc,T::Lo,T::Lo, };
static const GeneralCategory _341[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Lo,T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Lo,T::Mn,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lm,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mn,T::Mn,T::Mc,T::Mc,T::Po,T::Po,T::Lo,T::Lm,T::Lm,T::Mc,T::Mn,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _342[] = { D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Sk,T::Lm,T::Lm,T::Lm,T::Lm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lm,T::Sk,T::Sk,D,D,D,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _343[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mn,T::Mc,T::Mc,T::Mn,T::Mc,T::Mc,T::Po,T::Mc,T::Mn,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D, };
static const GeneralCategory _431[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D, };
static const GeneralCategory _500[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _501[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _502[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,D,D,D,D,D,D,D,D,D,D,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,D,D,D,D,T::Lo,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Sm,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _503[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _506[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Pe,T::Ps,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _507[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Sc,T::So,D,D, };
static const GeneralCategory _508[] = { T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Ps,T::Pe,T::Po,D,D,D,D,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Pd,T::Pd,T::Pc,T::Pc,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Po,T::Po,T::Ps,T::Pe,T::Po,T::Po,T::Po,T::Po,T::Pc,T::Pc,T::Pc,T::Po,T::Po,T::Po,D,T::Po,T::Po,T::Po,T::Po,T::Pd,T::Ps,T::Pe,T::Ps,T::Pe,T::Ps,T::Pe,T::Po,T::Po,T::Po,T::Sm,T::Pd,T::Sm,T::Sm,T::Sm,D,T::Po,T::Sc,T::Po,T::Po,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _509[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Cf, };
static const GeneralCategory _510[] = { D,T::Po,T::Po,T::Po,T::Sc,T::Po,T::Po,T::Po,T::Ps,T::Pe,T::Po,T::Sm,T::Po,T::Pd,T::Po,T::Po,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Po,T::Po,T::Sm,T::Sm,T::Sm,T::Po,T::Po,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ps,T::Po,T::Pe,T::Sk,T::Pc,T::Sk,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ps,T::Sm,T::Pe,T::Sm,T::Ps,T::Pe,T::Po,T::Ps,T::Pe,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lm,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _511[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lm,T::Lm,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,D,D,D,T::Sc,T::Sc,T::Sm,T::Sk,T::So,T::Sc,T::Sc,D,T::So,T::Sm,T::Sm,T::Sm,T::Sm,T::So,T::So,D,D,D,D,D,D,D,D,D,D,T::Cf,T::Cf,T::Cf,T::So,T::So,D,D, };
static const GeneralCategory _512[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _513[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D, };
static const GeneralCategory _514[] = { T::Po,T::Po,T::Po,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::No,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _515[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::No,T::No,T::So,T::So,T::So,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Mn,D,D, };
static const GeneralCategory _517[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Mn,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D, };
static const GeneralCategory _518[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Nl,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Nl,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,D, };
static const GeneralCategory _519[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Po,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _520[] = { T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _521[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,D,D,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,D,D,D, };
static const GeneralCategory _522[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _526[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _528[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,D,D,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Po,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::So,T::So,T::No,T::No,T::No,T::No,T::No,T::No,T::No, };
static const GeneralCategory _529[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No, };
static const GeneralCategory _530[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _531[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,T::No,T::No,T::Lo,T::Lo,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No, };
static const GeneralCategory _532[] = { T::Lo,T::Mn,T::Mn,T::Mn,D,T::Mn,T::Mn,D,D,D,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Mn,T::Mn,T::Mn,D,D,D,D,T::Mn,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::No,T::No,T::Po, };
static const GeneralCategory _533[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::No,T::No,T::No,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::So,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _534[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No, };
static const GeneralCategory _535[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _536[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _537[] = { T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,D,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No, };
static const GeneralCategory _538[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _540[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D, };
static const GeneralCategory _541[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Mn,T::Mn,T::Pd,D,D,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _542[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::Lo,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::No,T::No,T::No,T::No,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _543[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _544[] = { T::Mc,T::Mn,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Mn, };
static const GeneralCategory _545[] = { T::Mn,T::Mn,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Po,T::Po,T::Cf,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,T::Cf,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D, };
static const GeneralCategory _546[] = { T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Po,T::Po,T::Po,T::Po,T::Lo,T::Mc,T::Mc,T::Lo,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Po,T::Po,T::Lo,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _547[] = { T::Mn,T::Mn,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Lo,T::Lo,T::Lo,T::Lo,T::Po,T::Po,T::Po,T::Po,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Mc,T::Mn,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Lo,T::Po,T::Lo,T::Po,T::Po,T::Po,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _548[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mc,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _549[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Po,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D, };
static const GeneralCategory _550[] = { T::Mn,T::Mn,T::Mc,T::Mc,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Mn,T::Mn,T::Lo,T::Mc,T::Mc,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,D,D,T::Mc,T::Mc,D,D,T::Mc,T::Mc,T::Mc,D,D,T::Lo,D,D,D,D,D,D,T::Mc,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _552[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Po,T::Po,T::Po,T::Po,T::Po,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Po,T::Po,D,T::Po,T::Mn,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _553[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Lo,T::Lo,T::Po,T::Lo,D,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _555[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,D,D,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _556[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mc,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Lo,D,D,D,D,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _557[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mc,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Lo,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _558[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::No,T::No,T::Po,T::Po,T::Po,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _560[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _561[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo, };
static const GeneralCategory _562[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,D,T::Mc,T::Mc,D,D,T::Mn,T::Mn,T::Mc,T::Mn,T::Lo,T::Mc,T::Lo,T::Mc,T::Mn,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _563[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,D,D,T::Mn,T::Mn,T::Mc,T::Mc,T::Mc,T::Mc,T::Mn,T::Lo,T::Po,T::Lo,T::Mc,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _564[] = { T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Po,T::Mn,D,D,D,D,D,D,D,D,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _565[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Lo,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D, };
static const GeneralCategory _568[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Lo,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,T::Po,T::Po,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _569[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Mc,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,T::Mc,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _570[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,T::Mn,D,T::Mn,T::Mn,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lo,T::Mn,D,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _571[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,D,T::Mn,T::Mn,D,T::Mc,T::Mc,T::Mn,T::Mc,T::Mn,T::Lo,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _573[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mc,T::Mc,T::Po,T::Po,D,D,D,D,D,D,D, };
static const GeneralCategory _575[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sc,T::Sc,T::Sc,T::Sc,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Po, };
static const GeneralCategory _583[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _584[] = { T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,T::Nl,D,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _586[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _616[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _652[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _724[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _725[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _726[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Po,T::Po,T::Po,T::Po,T::Po,T::So,T::So,T::So,T::So,T::Lm,T::Lm,T::Lm,T::Lm,T::Po,T::So,D,D,D,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _727[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _732[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _733[] = { T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _734[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,T::Mn,T::Lo,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc, };
static const GeneralCategory _735[] = { T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,D,D,D,D,D,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lm,T::Lm,T::Po,T::Lm,T::Mn,D,D,D,D,D,D,D,D,D,D,D,T::Mc,T::Mc,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _783[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D, };
static const GeneralCategory _793[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _794[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _866[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _869[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D, };
static const GeneralCategory _888[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D, };
static const GeneralCategory _889[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::So,T::Mn,T::Mn,T::Po,T::Cf,T::Cf,T::Cf,T::Cf,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _929[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _930[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Mc,T::Mc,T::Mn,T::Mn,T::Mn,T::So,T::So,T::So,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Mc,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn, };
static const GeneralCategory _931[] = { T::Mn,T::Mn,T::Mn,T::So,T::So,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Mn,T::Mn,T::Mn,T::Mn,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _932[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Mn,T::Mn,T::Mn,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _933[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _934[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D, };
static const GeneralCategory _936[] = { T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu, };
static const GeneralCategory _937[] = { T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,D,T::Lu,T::Lu,D,D,T::Lu,D,D,T::Lu,T::Lu,D,D,T::Lu,T::Lu,T::Lu,T::Lu,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,D,T::Ll,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _938[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,D,T::Lu,T::Lu,T::Lu,T::Lu,D,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,D,T::Lu,T::Lu,T::Lu,T::Lu,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,T::Lu,D,D,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,D,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu, };
static const GeneralCategory _939[] = { T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _940[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu, };
static const GeneralCategory _941[] = { T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,D,D,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _942[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll, };
static const GeneralCategory _943[] = { T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Sm,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Lu,T::Ll,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd, };
static const GeneralCategory _948[] = { T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::So,T::So,T::So,T::So,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Mn,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _949[] = { T::So,T::So,T::So,T::So,T::Mn,T::So,T::So,T::Po,T::Po,T::Po,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _960[] = { T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,T::Mn,T::Mn,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _962[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,T::Lm,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,T::Lo,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _965[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Mn,T::Mn,T::Mn,T::Mn,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,T::Sc, };
static const GeneralCategory _977[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _978[] = { T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Lu,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Ll,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Lm,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,T::Po,T::Po,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _984[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No, };
static const GeneralCategory _985[] = { T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::No,T::No,T::No,T::Sc,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _986[] = { D,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _988[] = { T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,D,D,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,D,T::Lo,D,D,D,D,D,D,T::Lo,D,D,D,D,T::Lo,D,T::Lo,D,T::Lo,D,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,D,T::Lo,D,D,T::Lo,D,T::Lo,D,T::Lo,D,T::Lo,D,T::Lo,D,T::Lo,T::Lo,D,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,D, };
static const GeneralCategory _989[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Sm,T::Sm,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _992[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _993[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _994[] = { T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::No,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _995[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _996[] = { T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _999[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::Sk,T::Sk,T::Sk,T::Sk,T::Sk, };
static const GeneralCategory _1005[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D, };
static const GeneralCategory _1006[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _1007[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _1008[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _1009[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _1010[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _1011[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So, };
static const GeneralCategory _1012[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,T::So,T::So,T::So,T::So,T::So,D,D,D,T::So,T::So,T::So,D,D,D,D,D, };
static const GeneralCategory _1013[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _1015[] = { T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,T::So,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,T::Nd,D,D,D,D,D,D, };
static const GeneralCategory _1357[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _1390[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _1392[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _1437[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo, };
static const GeneralCategory _1495[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _1524[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _1574[] = { T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _7168[] = { D,T::Cf,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf,T::Cf, };
static const GeneralCategory _7171[] = { T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,T::Mn,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GeneralCategory _8191[] = { T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,D,D, };
static const GeneralCategory _8703[] = { T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,D,D, };
static const GeneralCategory *_blocks[] = {
 _0,_1,_2,_3,_4,_5,_6,_7,
 _8,_9,_10,_11,_12,_13,_14,_15,
 _16,_17,_18,_19,_20,_21,_22,_23,
 _24,_25,_26,_27,_28,_29,_30,_31,
 _32,_33,0,0,_36,_37,_38,_39,
 _40,0,0,0,_44,_45,_46,_47,
 _48,_49,_50,_51,_52,_53,_54,_55,
 _56,_57,_58,_59,_60,_61,_62,_63,
 _64,_65,_66,_67,0,0,_70,_71,
 _72,_73,0,_75,_76,0,_78,_79,
 0,0,0,_83,0,0,_86,_87,
 _88,_89,_90,_91,_92,_93,0,_95,
 _96,_97,_98,_99,_100,_101,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_155,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_319,
 _320,0,0,0,0,0,0,0,
 0,_329,0,0,_332,_333,_334,_335,
 _336,_337,_338,_339,_340,_341,_342,_343,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_431,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_500,_501,_502,_503,
 0,0,_506,_507,_508,_509,_510,_511,
 _512,_513,_514,_515,0,_517,_518,_519,
 _520,_521,_522,0,0,0,_526,0,
 _528,_529,_530,_531,_532,_533,_534,_535,
 _536,_537,_538,0,_540,_541,_542,_543,
 _544,_545,_546,_547,_548,_549,_550,0,
 _552,_553,0,_555,_556,_557,_558,0,
 _560,_561,_562,_563,_564,_565,0,0,
 _568,_569,_570,_571,0,_573,0,_575,
 0,0,0,0,0,0,0,_583,
 _584,0,_586,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _616,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_652,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_724,_725,_726,_727,
 0,0,0,0,_732,_733,_734,_735,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_783,
 0,0,0,0,0,0,0,0,
 0,_793,_794,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_866,0,0,_869,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _888,_889,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_929,_930,_931,_932,_933,_934,0,
 _936,_937,_938,_939,_940,_941,_942,_943,
 0,0,0,0,_948,_949,0,0,
 0,0,0,0,0,0,0,0,
 _960,0,_962,0,0,_965,0,0,
 0,0,0,0,0,0,0,0,
 0,_977,_978,0,0,0,0,0,
 _984,_985,_986,0,_988,_989,0,0,
 _992,_993,_994,_995,_996,0,0,_999,
 0,0,0,0,0,_1005,_1006,_1007,
 _1008,_1009,_1010,_1011,_1012,_1013,0,_1015,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1357,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1390,0,
 _1392,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1437,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_1495,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_1524,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1574,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _7168,0,0,_7171,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_8191,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_8703,
};
static const GeneralCategory _block_values[] = {
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::Lo,T::Lo,D,D,D,D,
 D,T::Lo,T::Lo,T::Lo,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,T::Sm,T::Sm,D,D,
 D,D,T::So,D,D,T::So,D,D,
 T::So,T::So,T::Sm,D,T::Sm,T::Sm,D,D,
 D,D,D,D,D,D,T::So,D,
 D,D,D,D,D,D,T::So,T::So,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,
 D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,D,T::Lo,T::Lo,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,
 T::Cs,T::Cs,T::Cs,T::Cs,T::Cs,T::Cs,T::Cs,T::Cs,
 T::Cs,T::Cs,T::Cs,T::Cs,T::Cs,T::Cs,T::Cs,T::Cs,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Lo,T::Lo,D,D,D,D,
 T::Lo,T::Lo,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,T::Lo,T::Lo,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,
 D,T::Lo,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Lo,T::Lo,D,T::Lo,T::Lo,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::So,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::So,T::So,T::So,T::So,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Lo,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,T::So,D,
 T::So,T::So,T::So,T::So,T::So,D,D,D,
 D,D,D,D,D,D,T::So,D,
 D,D,D,D,D,D,D,D,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,
 D,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Lo,T::Lo,T::Lo,T::Lo,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,
 T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,T::Lo,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::Mn,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,D,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,
 T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,T::Co,D,
};
inline GeneralCategory get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}
namespace _properties {
static const size_t _block_size = 128;
static const uint64_t _0[] = { 0,0,0,0,0,0,0,0,0,1073741825,1073741825,1073741825,1073741825,1073741825,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1073741825,2415919168,2147483680,2147483648,2147483648,2147483648,2147483648,2147483680,2147483648,2147483648,2147483648,2147483648,2147483712,2147483672,2415919168,2147483648,768,768,768,768,768,768,768,768,768,768,2147483712,2147483712,2147483648,2147483648,2147483648,2415919168,2147483648,768,768,768,768,768,768,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2147483648,2147483648,2147483648,2147487872,0,2147487744,768,768,768,768,768,768,0,0,16777216,16777216,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2147483648,2147483648,2147483648,2147483648,0, };
static const uint64_t _1[] = { 0,0,0,0,0,1073741825,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,4096,2147483648,16384,2147483680,2147483648,16,2147483648,4096,2147483648,2147483648,0,0,4096,0,2147483648,134230016,4096,0,16384,2147483680,0,0,0,2147483648,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2147483648,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2147483648,0,0,0,0,0,0,0,0, };
static const uint64_t _2[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16777216,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8388608,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _4[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16777216,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16777216,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _5[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16777216,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,20480,20480,16797696,20480,20480,20480,20480,20480,20480,4096,4096,4096,4096,4096,4096,4096,20480,20480,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,12288,12288,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,20480,20480,20480,20480,20480,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096, };
static const uint64_t _6[] = { 4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,21504,4096,4096,4096,4096,4096,4096,4096,4096,4096,4194304,4096,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,0,4096,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,0,0,0,0,20480,0,0,0,64,0, };
static const uint64_t _7[] = { 0,0,0,0,4096,4096,0,134217792,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,128,128,0,0,128,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,128,0,16777216,128,128,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _8[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16777216,0,16777216,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _9[] = { 0,0,0,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _10[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _11[] = { 0,0,0,0,0,0,0,0,0,268435520,24,0,0,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,8,5120,0,5120,5120,64,5120,1024,0,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _12[] = { 4294967296,4294967296,4294967296,4294967296,4294967296,4294967296,0,0,0,0,0,0,64,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,64,2,0,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8192,0,0,0,0,0,0,0,0,0,0,5120,5120,5120,5120,5120,5120,5120,5120,1024,1024,1024,1024,5120,4096,1024,1024,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,0,0,8388608,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _13[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,0,1024,1024,1024,1024,1024,1024,1024,4294967296,0,4096,4096,1024,1024,1024,1024,4096,4096,1024,1024,0,4096,4096,4096,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _14[] = { 268435520,268435520,268435520,64,64,64,64,64,64,64,64,0,64,0,0,4294967296,0,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _15[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,0,64,268435520,8192,0,0,0,0,0, };
static const uint64_t _16[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,4096,4096,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,1024,1024,1024,0,1024,1024,1024,1024,0,0,0,64,64,64,64,64,64,64,268435520,64,268435520,64,64,64,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _17[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,0,4294967296,5120,5120,5120,5120,5120,5120,5120,4096,4096,4096,4096,4096,4096,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,5120,1024, };
static const uint64_t _18[] = { 1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,4096,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,1024,1024,0,4096,4096,4096,4096,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,1024,1024,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _19[] = { 0,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,132096,1024,1024,1024,1024,1024,1024,0,0,1024,1024,0,0,1024,1024,4096,0,0,0,0,0,0,0,0,0,132096,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _20[] = { 0,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,1024,1024,1024,1024,1024,0,0,0,0,1024,1024,0,0,1024,1024,4096,0,0,0,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,1024,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _21[] = { 0,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,1024,1024,1024,1024,1024,1024,1024,1024,0,1024,1024,1024,0,1024,1024,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,4096,4096,4096, };
static const uint64_t _22[] = { 0,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,132096,1024,1024,1024,1024,1024,1024,0,0,1024,1024,0,0,1024,1024,4096,0,0,0,0,0,0,0,12288,1024,132096,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _23[] = { 0,0,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,132096,1024,1024,1024,1024,0,0,0,1024,1024,1024,0,1024,1024,1024,4096,0,0,0,0,0,0,0,0,0,132096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _24[] = { 1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,0,1024,1024,1024,0,1024,1024,1024,4096,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _25[] = { 0,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,1024,1024,1024,1024,132096,1024,1024,0,1024,1024,1024,0,1024,1024,1024,4096,0,0,0,0,0,0,0,132096,132096,0,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _26[] = { 1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,0,132096,1024,1024,1024,1024,1024,1024,0,1024,1024,1024,0,1024,1024,1024,4096,0,0,0,0,0,0,0,0,0,132096,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _27[] = { 0,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,0,0,0,132096,1024,1024,1024,1024,1024,0,1024,0,1024,1024,1024,1024,1024,1024,1024,132096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _28[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,0,0,1024,1024,1024,1024,1024,1024,1024,0,0,0,0,0,33554432,33554432,33554432,33554432,33554432,0,8192,4096,4096,4096,4096,4096,4096,1024,4096,0,0,0,0,0,0,0,0,0,0,0,64,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _29[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,0,0,1024,1024,1024,1024,1024,1024,4096,1024,1024,0,0,0,33554432,33554432,33554432,33554432,33554432,0,8192,0,4096,4096,4096,4096,4096,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _30[] = { 0,0,0,0,0,0,0,0,64,0,0,0,0,64,64,64,64,64,64,0,0,0,0,0,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,4096,0,4096,0,0,0,0,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,8389632,1024,8389632,1024,1024,1024,1024,1024,1024, };
static const uint64_t _31[] = { 1024,1024,4096,4096,4096,0,4096,4096,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _32[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,1024,4096,4096,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,268435520,268435520,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,0,0,0,0,1024,1024,1024,0,1024,5120,5120,0,0,1024,1024,5120,5120,5120,5120,5120,0,0,0,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _33[] = { 0,0,1024,1024,1024,1024,1024,5120,5120,5120,5120,5120,5120,5120,0,5120,0,0,0,0,0,0,0,0,0,0,5120,5120,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _34[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4194304,4194304,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _38[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,0,64,268435520,64,64,64,64,268435520,268435520,134217728,134217728,134217728,134217728,134217728,134217728,134217728,134217728,134217728,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _40[] = { 8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _44[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _45[] = { 1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,64,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _46[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,0,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _47[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8388608,8388608,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4194304,4194304,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,64,64,64,0,0,0,64,0,0,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _48[] = { 0,0,64,268435520,64,64,24,0,64,268435520,8192,536870912,536870912,536870912,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8192,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _49[] = { 0,0,0,0,0,67109888,67109888,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _50[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,4096,0,0,0,0,0,0,0,0,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _51[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,33554432,33554432,33554432,0,0,33554432,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,134217728,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _52[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,4096,4096,4096,4096,4096,4096,0,0,4096, };
static const uint64_t _53[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8192,268435520,268435520,268435520,268435520,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _54[] = { 1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,132096,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,268435520,0,64,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _55[] = { 1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _56[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,13312,4096,0,0,0,268435520,268435520,64,64,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,12288,4096,4096,268435520,268435520, };
static const uint64_t _57[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,4096,0,0,0,0,0,0,4096,0,0,4096,4096,4096,0,0,0,0,0,0, };
static const uint64_t _58[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,20480,16797696,20480,20480,20480,20480,20480,20480,20480,20480,0,0,0,0,0,0,0,0,0,0,0,0,0,16384,0,0,0,0,0,0,0, };
static const uint64_t _59[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16777216,0,0,0,0,16384,16384,16384,16384,16384,16384,16384,16384,16384,16793600,16384,16384,16384,16793600,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,4096,4096,4096,0,0,0,4096,4096,4096, };
static const uint64_t _60[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16777216,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _61[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16777216,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _63[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,0, };
static const uint64_t _64[] = { 1,1,1,1,1,1,1,1,1,1,1,0,131076,4,1073741826,1073741826,2147483672,2147483672,2147483656,2147483656,2147483656,2147483656,2147483776,2147483648,2147483680,2147483680,2147483680,2147483680,2147483680,2147483680,2147483680,2147483680,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,1073741825,1073741825,2,2,2,2,2,1,2147483648,2147483648,2147483776,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483680,2147483680,2147483648,2415919168,2415919168,2147483648,0,128,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2415919168,2415919168,2415919168,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483656,0,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,1,0,128,128,128,128,4194304,2,2,2,2,8388608,8388608,8388608,8388608,8388608,8388608,0,16793600,0,0,0,0,0,0,0,0,0,8,0,128,128,16384, };
static const uint64_t _65[] = { 0,0,0,0,0,0,0,0,0,0,0,8,0,128,128,0,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,128,128,128,128,128,128,128,128,128,128,128,128,0,0,0,0,128,0,0,0,128,128,0,0,0,0,128,128,128,128,128,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _66[] = { 0,0,128,0,0,0,0,128,0,0,128,128,128,128,128,128,128,128,128,128,0,128,0,0,67108864,128,128,128,128,128,0,0,0,0,0,0,128,0,0,0,128,128,0,0,128,128,67108864,128,128,128,0,128,128,128,128,128,128,0,0,0,128,128,128,128,0,0,0,0,0,128,128,128,16777344,16777344,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384,16384, };
static const uint64_t _67[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483648,2147483776,2147483776,2147483648,2147483776,2147483776,2147483648,2147483776,2147483648,2147483776,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483776,2147483776,2147483648,2147483776,2147483648,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483648,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648, };
static const uint64_t _68[] = { 2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483656,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648, };
static const uint64_t _70[] = { 2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2155872256,2155872256,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648, };
static const uint64_t _71[] = { 2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483648,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648, };
static const uint64_t _72[] = { 2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _73[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,17408,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _75[] = { 2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483648,2147483648,2147483776,2147483776,2147483648,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483648,2147483776,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648, };
static const uint64_t _76[] = { 2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483648,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648, };
static const uint64_t _78[] = { 2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _79[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648, };
static const uint64_t _83[] = { 2147483648,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483776,2147483776,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483776,2147483776,2147483648,2147483648, };
static const uint64_t _88[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16793600,16384,0,0, };
static const uint64_t _89[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _91[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024, };
static const uint64_t _92[] = { 2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483672,2147483648,2147483648,2147483656,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2415919168,2147487744,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483656,2147483656,2415919168,2147483648,2147483648,2147483648,2147483656,2147483712,2147483680,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483712,2147483648,2147483712,2147483712,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648, };
static const uint64_t _93[] = { 1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,0,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _95[] = { 1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,1048576,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,262144,262144,524288,524288,262144,262144,262144,262144,262144,262144,262144,262144,0,0,0,0, };
static const uint64_t _96[] = { 1,2147483712,2415919168,2147483648,0,8192,2048,2048,2147483648,2147483648,2147483648,2147483648,2147483680,2147483680,2147483680,2147483680,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483648,2147483656,2147483680,2147483680,2147483680,2147483648,2048,2048,2048,2048,2048,2048,2048,2048,2048,4096,4096,4096,4096,135168,135168,2147483656,8192,8192,8192,8192,8192,0,0,2048,2048,2048,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _97[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,67112960,67112960,8192,8192,0,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16,12288,8192,8192,0, };
static const uint64_t _98[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4194304,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _155[] = { 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _319[] = { 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,0,0,0, };
static const uint64_t _320[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8192,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _329[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,268435520, };
static const uint64_t _332[] = { 0,0,0,0,0,0,0,0,0,0,0,0,8192,64,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,0,4096, };
static const uint64_t _333[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,20480,20480,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,0,268435520,64,64,64,268435520,0,0,0,0,0,0,0,0, };
static const uint64_t _334[] = { 4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,16384,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _335[] = { 0,0,0,0,0,0,0,0,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,20480,20480,0,0,0,0,0,0, };
static const uint64_t _336[] = { 0,0,1024,0,0,0,0,0,0,0,0,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,268435520,0,0,0,0,0,0,0,0, };
static const uint64_t _337[] = { 1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,1024,0,0,0,0,0,0,0,0,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,1024, };
static const uint64_t _338[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,4096,4096,4096,4096,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _339[] = { 1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,0,0,0,0,0,0,64,268435520,268435520,0,0,0,0,0,8192,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5120,8192,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _340[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,1024,0,0,0,0,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8192,0,0,0,0,0,0,0,0,0,0,5120,5120,5120,0,0, };
static const uint64_t _341[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,0,1024,1024,1024,33554432,33554432,1024,1024,33554432,0,33554432,33554432,0,1024,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8192,0,64,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,268435520,268435520,0,8192,8192,1024,4096,0,0,0,0,0,0,0,0,0, };
static const uint64_t _342[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,20480,20480,20480,20480,0,0,0,0,0,0,0,0,0,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _343[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,268435520,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _500[] = { 2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2099200,2099200,2048,2099200,2048,2099200,2099200,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2099200,2048,2099200,2048,2099200,2099200,2048,2048,2099200,2099200,2099200,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,0,0,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048, };
static const uint64_t _501[] = { 2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _502[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,5120,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _506[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2147483648,2147483648,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _507[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,65536,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _508[] = { 536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,8,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,32,32,32,32,2147483648,2147483648,0,0,0,0,0,0,0,0,0,64,64,268435520,0,64,64,268435520,268435520,8,0,0,0,0,0,0,0,0,128,0,152,0,0,0,0,128,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _510[] = { 0,268435520,32,0,0,0,0,32,0,0,0,0,64,24,268435520,0,256,256,256,256,256,256,256,256,256,256,64,64,0,0,0,268435520,0,256,256,256,256,256,256,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,128,0,4224,0,4096,256,256,256,256,256,256,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,32,32,64,16,0,0,0,0,0,0,0,0,0,0,12288,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _511[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,135168,135168,4194304,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,0,0,0,0,0,0,0,0,0,0,0,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,0,0,0,0,0,65536,65536, };
static const uint64_t _517[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _518[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,0,0,0,0,0, };
static const uint64_t _519[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _528[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _530[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _532[] = { 0,1024,1024,1024,0,1024,1024,0,0,0,0,0,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _533[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,0,0,0,0,0,0,0,0,0,64,64,64,64,64,64,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _534[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,64,64,64,64,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _535[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,64,64,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _538[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,5120,5120,5120,5120,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _541[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _542[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,268435520,268435520,268435520,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _544[] = { 1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,268435520,268435520,64,64,64,64,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _545[] = { 0,0,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,0,0,4294967296,268435520,268435520,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,4294967296,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _546[] = { 1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,268435520,268435520,268435520,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _547[] = { 1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,0,0,0,0,268435520,268435520,0,0,0,4096,4096,4096,268435520,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _548[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,1024,268435520,268435520,64,268435520,268435520,0,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _549[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _550[] = { 1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,0,132096,1024,1024,1024,1024,1024,1024,0,0,1024,1024,0,0,1024,1024,4096,0,0,0,0,0,0,0,0,0,132096,0,0,0,0,0,8192,0,0,0,0,1024,1024,0,0,4096,4096,4096,4096,4096,4096,4096,0,0,0,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _552[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,1024,1024,1024,4096,0,0,0,0,268435520,268435520,64,0,0,0,0,0,0,0,0,0,0,0,0,64,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _553[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,132096,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,132096,1024,1024,1024,1024,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _555[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,132096,1024,1024,1024,1024,1024,1024,0,0,1024,1024,1024,1024,1024,1024,1024,4096,4096,0,268435520,268435520,64,64,8192,8192,8192,268435520,268435520,268435520,268435520,268435520,268435520,268435520,268435520,268435520,268435520,268435520,268435520,268435520,268435520,268435520,0,0,0,0,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _556[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,1024,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _557[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _558[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _560[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _562[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,132096,1024,1024,1024,1024,1024,0,1024,1024,0,0,1024,1024,4096,4096,0,1024,0,1024,4096,268435520,0,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _563[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,0,0,1024,1024,1024,1024,1024,1024,4096,0,0,0,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _564[] = { 0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,1024,1024,1024,1024,1024,0,1024,1024,1024,1024,0,0,0,268435520,268435520,0,0,0,4096,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _565[] = { 0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,8192,4096,0,268435520,268435520,0,0,0,0,64,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _568[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,0,1024,1024,1024,1024,1024,1024,1024,4096,0,268435520,268435520,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _569[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _570[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,1024,0,0,0,1024,0,1024,1024,0,1024,1024,1024,4096,1024,4096,4096,0,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _571[] = { 0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,1024,0,1024,1024,0,1024,1024,1024,1024,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _573[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,1024,1024,1024,268435520,268435520,0,0,0,0,0,0,0, };
static const uint64_t _584[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,64,64,64,64,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _724[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,268435520,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _725[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096,268435520,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _726[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,268435520,268435520,64,0,0,0,0,0,0,0,0,8192,8192,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _733[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,64,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _734[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024, };
static const uint64_t _735[] = { 1024,1024,1024,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,5120,5120,5120,5120,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8192,8192,0,8192,2048,0,0,0,0,0,0,0,0,0,0,0,5120,5120,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _783[] = { 2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,0,0,0,0,0,0,0,0, };
static const uint64_t _793[] = { 2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _794[] = { 2048,2048,2048,2048,2048,2048,2048,2048,2048,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _866[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048, };
static const uint64_t _869[] = { 2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,0,0,0,0, };
static const uint64_t _889[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1024,268435520,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _930[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,0,4096,4096,4096,0,0,0,4096,135168,135168,135168,135168,135168,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096, };
static const uint64_t _931[] = { 4096,4096,4096,0,0,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _936[] = { 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128, };
static const uint64_t _937[] = { 128,128,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,0,0,128,0,0,128,128,0,0,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,0,128,0,128,16777344,16777344,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128, };
static const uint64_t _938[] = { 128,128,128,128,128,128,0,128,128,128,128,0,0,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,0,128,128,128,128,128,0,128,0,0,0,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128, };
static const uint64_t _939[] = { 128,128,128,128,128,128,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128, };
static const uint64_t _940[] = { 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128, };
static const uint64_t _941[] = { 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,16777344,16777344,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128, };
static const uint64_t _942[] = { 128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128, };
static const uint64_t _943[] = { 128,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,0,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128, };
static const uint64_t _949[] = { 0,0,0,0,0,0,0,64,268435520,64,64,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _960[] = { 1024,1024,1024,1024,1024,1024,1024,0,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,1024,0,0,1024,1024,1024,1024,1024,1024,1024,0,1024,1024,0,1024,1024,1024,1024,1024,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _962[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,0,8192,8192,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _965[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _977[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,4096,4096,4096,4096,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _978[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,12288,12288,12288,1024,4096,4096,4096,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _988[] = { 128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,128,128,0,128,0,0,128,0,128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,0,128,0,128,0,0,0,0,0,0,128,0,0,0,0,128,0,128,0,128,0,128,128,128,0,128,128,0,128,0,0,128,0,128,0,128,0,128,0,128,0,128,128,0,128,0,0,128,128,128,128,0,128,128,128,128,128,128,128,0,128,128,128,128,0,128,128,128,128,0,128,0, };
static const uint64_t _989[] = { 128,128,128,128,128,128,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,0,0,0,0,128,128,128,0,128,128,128,128,128,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _994[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,0,0,0,0,0,0,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,0,0,0,0,0,0,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,33792, };
static const uint64_t _995[] = { 33792,33792,33792,33792,33792,33792,33792,33792,33792,33792,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592,8589934592, };
static const uint64_t _1023[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _1357[] = { 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _1390[] = { 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,0,0,0,0,0,0,0,0,0,0,0,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200, };
static const uint64_t _1392[] = { 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,0,0,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200, };
static const uint64_t _1437[] = { 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200, };
static const uint64_t _1495[] = { 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _1524[] = { 2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,2048,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _1535[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _1574[] = { 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint64_t _2047[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _2559[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _3071[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _3583[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _4095[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _4607[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _5119[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _5631[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _6143[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _6655[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _7167[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _7168[] = { 4194304,8388608,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint64_t _7171[] = { 536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,536870912,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304, };
static const uint64_t _7679[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _8191[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t _8703[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,65536,65536, };
static const uint64_t *_blocks[] = {
 _0,_1,_2,0,_4,_5,_6,_7,
 _8,_9,_10,_11,_12,_13,_14,_15,
 _16,_17,_18,_19,_20,_21,_22,_23,
 _24,_25,_26,_27,_28,_29,_30,_31,
 _32,_33,_34,0,0,0,_38,0,
 _40,0,0,0,_44,_45,_46,_47,
 _48,_49,_50,_51,_52,_53,_54,_55,
 _56,_57,_58,_59,_60,_61,0,_63,
 _64,_65,_66,_67,_68,0,_70,_71,
 _72,_73,0,_75,_76,0,_78,_79,
 0,0,0,_83,0,0,0,0,
 _88,_89,0,_91,_92,_93,0,_95,
 _96,_97,_98,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_155,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_319,
 _320,0,0,0,0,0,0,0,
 0,_329,0,0,_332,_333,_334,_335,
 _336,_337,_338,_339,_340,_341,_342,_343,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_500,_501,_502,0,
 0,0,_506,_507,_508,0,_510,_511,
 0,0,0,0,0,_517,_518,_519,
 0,0,0,0,0,0,0,0,
 _528,0,_530,0,_532,_533,_534,_535,
 0,0,_538,0,0,_541,_542,0,
 _544,_545,_546,_547,_548,_549,_550,0,
 _552,_553,0,_555,_556,_557,_558,0,
 _560,0,_562,_563,_564,_565,0,0,
 _568,_569,_570,_571,0,_573,0,0,
 0,0,0,0,0,0,0,0,
 _584,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_724,_725,_726,0,
 0,0,0,0,0,_733,_734,_735,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_783,
 0,0,0,0,0,0,0,0,
 0,_793,_794,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_866,0,0,_869,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_889,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_930,_931,0,0,0,0,
 _936,_937,_938,_939,_940,_941,_942,_943,
 0,0,0,0,0,_949,0,0,
 0,0,0,0,0,0,0,0,
 _960,0,_962,0,0,_965,0,0,
 0,0,0,0,0,0,0,0,
 0,_977,_978,0,0,0,0,0,
 0,0,0,0,_988,_989,0,0,
 0,0,_994,_995,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_1023,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1357,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1390,0,
 _1392,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1437,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_1495,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_1524,0,0,0,
 0,0,0,0,0,0,0,_1535,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1574,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_2047,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_2559,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_3071,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_3583,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_4095,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_4607,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_5119,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_5631,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_6143,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_6655,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_7167,
 _7168,0,0,_7171,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_7679,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_8191,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_8703,
};
static const uint64_t _block_values[] = {
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,2147483648,0,0,
 0,0,2147483648,0,0,2147483648,0,0,
 2147483648,2147483648,2147483648,0,2147483648,2147483648,2147483648,2147483648,
 0,0,0,0,0,0,1048576,0,
 0,0,0,0,0,0,0,0,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,0,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,2048,2048,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 2048,2048,2048,2048,2048,2048,2048,2048,
 2048,2048,2048,2048,2048,2048,2048,2048,
 2048,2048,2048,2048,2048,2048,2048,2048,
 2048,2048,2048,2048,2048,2048,2048,2048,
 2048,2048,2048,2048,2048,2048,2048,2048,
 2048,2048,2048,2048,2048,2048,2048,0,
 2048,2048,2048,2048,2048,2048,2048,2048,
 2048,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,2048,2048,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,0,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,0,2099200,
 0,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,0,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 2048,2048,2048,2048,0,0,0,0,
 0,0,0,0,0,0,0,0,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,2099200,2099200,
 2099200,2099200,2099200,2099200,2099200,2099200,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,4194304,536870912,0,4194304,4194304,4194304,4194304,
 4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,
 4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,
 4194304,4194304,4194304,4194304,4194304,4194304,4194304,4194304,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
inline uint64_t get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}
namespace _derived_core_properties {
static const size_t _block_size = 128;
static const uint32_t _0[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131104,131072,131072,131072,131073,131072,131072,131104,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131104,131072,131073,131073,131073,131072,131072,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,131072,131072,131072,131105,151552,131104,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,131072,131073,131072,131073,0, };
static const uint32_t _1[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131104,131072,161814,131072,131073,32800,131072,131104,131072,131073,131072,131072,131104,163734,131072,151584,131104,131072,161814,131072,131072,131072,131072,131072,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,131073,163418,163418,163418,163418,163418,163418,163418,163734,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,131073,163222,163222,163222,163222,163222,163222,163222,163222, };
static const uint32_t _2[] = { 163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,161814,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163734,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163418,163222,163418,163222,163418,163222,163734, };
static const uint32_t _3[] = { 163222,163418,163418,163222,163418,163222,163418,163418,163222,163418,163418,163418,163222,161814,163418,163418,163418,163418,163222,163418,163418,163222,163418,163418,163418,163222,163222,161814,163418,163418,163222,163418,163418,163222,163418,163222,163418,163222,163418,163418,163222,163418,161814,161814,163418,163222,163418,163418,163222,163418,163418,163418,163222,163418,163222,163418,163418,163222,161814,161794,163418,163222,161814,163222,161794,161794,161794,161794,163674,163538,163222,163674,163538,163222,163674,163538,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163222,163674,163538,163222,163418,163222,163418,163418,163418,163222,163418,163222,163418,163222,163418,163222, };
static const uint32_t _4[] = { 163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,161814,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,161814,161814,161814,161814,161814,161814,163418,163418,163222,163418,163418,163222,163222,163418,163222,163418,163418,163418,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163222,163222,163222,163222,163222,161814,163222,163222,161814,163222,161814,163222,163222,161814,161814,161814,163222,163222,161814,163222,161814,163222,163222,161814,163222,163222,163222,163222,163222,161814,161814,163222,161814,163222,163222,161814,161814,163222,161814,161814,161814,161814,161814,161814,161814,163222,161814,161814, };
static const uint32_t _5[] = { 163222,161814,163222,163222,161814,161814,161814,163222,163222,163222,163222,163222,163222,161814,161814,161814,161814,161814,163222,161814,161794,161814,161814,161814,161814,161814,161814,161814,161814,163222,163222,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161846,161846,161846,161846,161846,161846,161846,161846,161846,161826,161826,161826,161826,161826,161826,161826,161846,161846,131104,131104,131104,131104,161826,161826,161826,161826,161826,161826,161826,161826,161826,161826,161826,161826,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,161846,161846,161846,161846,161846,131104,131104,131104,131104,131104,131104,131104,161826,131104,161826,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104, };
static const uint32_t _6[] = { 86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,87990,86048,86048,86048,86048,86048,86048,86048,86048,86048,118816,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,163418,163222,163418,163222,161826,131104,163418,163222,0,0,137270,163222,163222,163222,131072,163418, };
static const uint32_t _7[] = { 0,0,0,0,131104,131104,163418,151584,163418,163418,163418,0,163418,0,163418,163418,163222,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,0,163418,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163734,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163418,163735,163735,161819,161818,161818,163735,163734,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163735,163735,163222,163222,163419,163735,131073,163418,163222,163418,163418,163222,161814,163418,163418,163418, };
static const uint32_t _8[] = { 163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222, };
static const uint32_t _9[] = { 163418,163222,131072,86048,86048,86048,86048,86048,65568,65568,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222, };
static const uint32_t _10[] = { 163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,0,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,0,0,161826,131072,131072,131072,131072,131072,131104,161814,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222, };
static const uint32_t _11[] = { 163222,163222,163222,163222,163222,163222,163222,163734,161814,131072,131072,0,0,131072,131072,131072,0,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,131072,86050,131072,86050,86050,131072,86050,86050,131072,86050,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,161794,161794,161794,161794,131072,131104,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _12[] = { 32,32,32,32,32,32,131073,131073,131073,131072,131072,131072,131072,131072,131072,131072,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,131072,32800,0,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161826,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86048,86050,86050,86050,86050,86050,86050,86050,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,161794,161794,86050,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _13[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,161794,86050,86050,86050,86050,86050,86050,86050,32,131072,86048,86048,86050,86050,86050,86050,161826,161826,86050,86050,131072,86048,86048,86048,86050,161794,161794,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,161794,161794,161794,131072,131072,161794, };
static const uint32_t _14[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,32,161794,86050,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _15[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,86048,86048,86048,86048,86048,86048,86048,86048,161826,161826,131072,131072,131072,131072,161826,0,0,86048,131072,131072, };
static const uint32_t _16[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86048,86048,161826,86050,86050,86050,86050,86050,86050,86050,86050,86050,161826,86050,86050,86050,161826,86050,86050,86050,86050,86048,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,86048,86048,0,0,131072,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _17[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,86048,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86048,86048,32,86050,86050,86050,86050,86050,86050,86050,86048,86048,86048,86048,86048,86048,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050, };
static const uint32_t _18[] = { 86050,86050,86050,151554,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,151554,86048,161794,151554,151554,151554,86050,86050,86050,86050,86050,86050,86050,86050,151554,151554,151554,151554,348192,151554,151554,161794,86048,86048,86048,86048,86050,86050,86050,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,131072,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,161826,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _19[] = { 161794,86050,151554,151554,0,161794,161794,161794,161794,161794,161794,161794,161794,0,0,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,0,0,0,161794,161794,161794,161794,0,0,86048,161794,86018,151554,151554,86050,86050,86050,86050,0,0,151554,151554,0,0,151554,151554,348192,161794,0,0,0,0,0,0,0,0,86018,0,0,0,0,161794,161794,0,161794,161794,161794,86050,86050,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,161794,161794,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,161794,131072,86048,0, };
static const uint32_t _20[] = { 0,86050,86050,151554,0,161794,161794,161794,161794,161794,161794,0,0,0,0,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,0,161794,161794,0,161794,161794,0,0,86048,0,151554,151554,151554,86050,86050,0,0,0,0,86050,86050,0,0,86050,86050,348192,0,0,0,86050,0,0,0,0,0,0,0,161794,161794,161794,161794,0,161794,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,86050,86050,161794,161794,161794,86050,131072,0,0,0,0,0,0,0,0,0, };
static const uint32_t _21[] = { 0,86050,86050,151554,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,0,161794,161794,161794,161794,161794,0,0,86048,161794,151554,151554,151554,86050,86050,86050,86050,86050,0,86050,86050,151554,0,151554,151554,348192,0,0,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,86050,86050,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,0,0,0,0,0,0,0,161794,86050,86050,86050,86048,86048,86048, };
static const uint32_t _22[] = { 0,86050,151554,151554,0,161794,161794,161794,161794,161794,161794,161794,161794,0,0,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,0,161794,161794,161794,161794,161794,0,0,86048,161794,86018,86050,151554,86050,86050,86050,86050,0,0,151554,151554,0,0,151554,151554,348192,0,0,0,0,0,0,0,86048,86050,86018,0,0,0,0,161794,161794,0,161794,161794,161794,86050,86050,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,161794,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0, };
static const uint32_t _23[] = { 0,0,86050,161794,0,161794,161794,161794,161794,161794,161794,0,0,0,161794,161794,161794,0,161794,161794,161794,161794,0,0,0,161794,161794,0,161794,0,161794,161794,0,0,0,161794,161794,0,0,0,161794,161794,161794,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,86018,151554,86050,151554,151554,0,0,0,151554,151554,151554,0,151554,151554,151554,348192,0,0,161794,0,0,0,0,0,0,86018,0,0,0,0,0,0,0,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0, };
static const uint32_t _24[] = { 86050,151554,151554,151554,86048,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,161794,86050,86050,86050,151554,151554,151554,151554,0,86050,86050,86050,0,86050,86050,86050,348192,0,0,0,0,0,0,0,86050,86050,0,161794,161794,161794,0,0,0,0,0,161794,161794,86050,86050,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _25[] = { 161794,86050,151554,151554,131072,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,0,0,86048,161794,151554,86050,151554,151554,86018,151554,151554,0,86050,151554,151554,0,151554,151554,86050,348192,0,0,0,0,0,0,0,86018,86018,0,0,0,0,0,0,0,161794,0,161794,161794,86050,86050,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _26[] = { 86050,86050,151554,151554,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,348192,348192,161794,86018,151554,151554,86050,86050,86050,86050,0,151554,151554,151554,0,151554,151554,151554,348192,161794,131072,0,0,0,0,161794,161794,161794,86018,131072,131072,131072,131072,131072,131072,131072,161794,161794,161794,86050,86050,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,161794,161794,161794,161794,161794,161794, };
static const uint32_t _27[] = { 0,86050,151554,151554,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,0,0,161794,161794,161794,161794,161794,161794,161794,0,0,0,348192,0,0,0,0,86018,151554,151554,86050,86050,86050,0,86050,0,151554,151554,151554,151554,151554,151554,151554,86018,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,151554,151554,131072,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _28[] = { 0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,161794,153602,86050,86050,86050,86050,86050,86050,348194,0,0,0,0,131072,161794,161794,161794,161794,161794,161794,161826,86048,86048,86048,86048,86048,86048,86050,86048,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _29[] = { 0,161794,161794,0,161794,0,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,161794,153602,86050,86050,86050,86050,86050,86050,348192,86050,86050,161794,0,0,161794,161794,161794,161794,161794,0,161826,0,86048,86048,86048,86048,86048,86050,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _30[] = { 161794,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,86048,86048,131072,131072,131072,131072,131072,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,86048,131072,86048,131072,86048,131072,131072,131072,131072,151552,151552,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,151554, };
static const uint32_t _31[] = { 86050,86050,86048,86048,348192,131072,86048,86048,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,0,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,0,131072,131072,131072,131072,131072,131072,131072,131072,86048,131072,131072,131072,131072,131072,131072,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _32[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,86050,86050,86050,86050,151554,86050,86050,86050,86050,86050,86048,151554,348192,348192,151554,151554,86050,86050,161794,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,131072,131072,161794,161794,161794,161794,161794,161794,151554,151554,86050,86050,161794,161794,161794,161794,86050,86050,86050,161794,151554,151554,151554,161794,161794,151554,151554,151554,151554,151554,151554,151554,161794,161794,161794,86050,86050,86050,86050,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _33[] = { 161794,161794,86050,151554,151554,86050,86050,151554,151554,151554,151554,151554,151554,86050,161794,151554,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,151554,151554,151554,86050,131072,131072,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,0,163418,0,0,0,0,0,163418,0,0,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,162966,131072,161826,162966,162966,162966, };
static const uint32_t _34[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,194562,194562,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _36[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,0,161794,0,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _37[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,0,161794,0,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _38[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,86048,86048,86048,131072,131072,131072,131072,131072,131072,131072,131072,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0, };
static const uint32_t _39[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,162906,0,0,163734,163734,163734,163734,163734,163734,0,0, };
static const uint32_t _40[] = { 131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _44[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _45[] = { 131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0, };
static const uint32_t _46[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,86050,86050,348192,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,348192,131072,131072,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,0,86050,86050,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _47[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,118816,118816,151554,86050,86050,86050,86050,86050,86050,86050,151554,151554,151554,151554,151554,151554,151554,151554,86050,151554,151554,86048,86048,86048,86048,86048,86048,86048,86048,86048,348192,86048,131072,131072,131072,161826,131072,131072,131072,131072,161794,86048,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0, };
static const uint32_t _48[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,118816,118816,118816,32800,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161826,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0, };
static const uint32_t _49[] = { 161794,161794,161794,161794,161794,96290,96290,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,161794,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _50[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,86050,86050,86050,151554,151554,151554,151554,86050,86050,151554,151554,151554,0,0,0,0,151554,151554,86050,151554,151554,151554,151554,151554,151554,86048,86048,86048,0,0,0,0,131072,0,0,0,131072,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _51[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _52[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,151554,151554,86050,0,0,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,86050,151554,86050,86050,86050,86050,86050,86050,86050,0,348192,151554,86050,151554,151554,86050,86050,86050,86050,86050,86050,86050,86050,151554,151554,151554,151554,151554,151554,86050,86050,86048,86048,86048,86048,86048,86048,86048,86048,0,0,86048, };
static const uint32_t _53[] = { 151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,161826,131072,131072,131072,131072,131072,131072,0,0,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,65568,86050,86050,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _54[] = { 86050,86050,86050,86050,151554,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,86018,86050,86050,86050,86050,86050,151554,86050,151554,151554,151554,151554,151554,86050,151554,413696,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,86048,86048,86048,86048,86048,86048,86048,86048,86048,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0, };
static const uint32_t _55[] = { 86050,86050,151554,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,86050,86050,86050,86050,151554,151554,86050,86050,413696,348192,86050,86050,161794,161794,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,151554,86050,86050,151554,151554,151554,86050,151554,86050,86050,86050,413696,413696,0,0,0,0,0,0,0,0,131072,131072,131072,131072, };
static const uint32_t _56[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,151554,151554,151554,151554,151554,151554,86050,86050,86050,86050,86050,86050,86050,86050,151554,151554,86050,86048,0,0,0,131072,131072,131072,131072,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,161794,161794,161794,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161826,161826,161826,161826,161826,161826,131072,131072, };
static const uint32_t _57[] = { 163734,163734,163734,163734,163734,163734,163734,163734,163734,0,0,0,0,0,0,0,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,0,0,163418,163418,163418,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,86048,86048,86048,131072,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,151552,86048,86048,86048,86048,86048,86048,86048,161794,161794,161794,161794,86048,161794,161794,161794,161794,161794,161794,86048,161794,161794,151552,86048,86048,161794,0,0,0,0,0, };
static const uint32_t _58[] = { 161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161846,163222,161814,161814,161814,163222,161814,161814, };
static const uint32_t _59[] = { 161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,163222,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86048,86048,86048,86048,86048,0,86048,86048,86048,86048,86048, };
static const uint32_t _60[] = { 163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222, };
static const uint32_t _61[] = { 163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163222,163222,163222,163222,163734,163734,161814,161814,163418,161814,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222, };
static const uint32_t _62[] = { 163222,163222,163222,163222,163222,163222,163222,163222,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,0,0,163418,163418,163418,163418,163418,163418,0,0,163222,163222,163222,163222,163222,163222,163222,163222,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,163222,163222,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,0,0,163418,163418,163418,163418,163418,163418,0,0,163222,163222,163222,163222,163222,163222,163222,163222,0,163418,0,163418,0,163418,0,163418,163222,163222,163222,163222,163222,163222,163222,163222,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,0,0, };
static const uint32_t _63[] = { 163734,163734,163734,163734,163734,163734,163734,163734,163538,163538,163538,163538,163538,163538,163538,163538,163734,163734,163734,163734,163734,163734,163734,163734,163538,163538,163538,163538,163538,163538,163538,163538,163734,163734,163734,163734,163734,163734,163734,163734,163538,163538,163538,163538,163538,163538,163538,163538,163222,163222,163734,163734,163734,0,163222,163734,163418,163418,163418,163418,163538,131104,163222,131104,131104,131104,163734,163734,163734,0,163222,163734,163418,163418,163418,163418,163538,131104,131104,131104,163222,163222,163222,163222,0,0,163222,163222,163418,163418,163418,163418,0,131104,131104,131104,163222,163222,163222,163222,163222,163222,163222,163222,163418,163418,163418,163418,163418,131104,131104,131104,0,0,163734,163734,163734,0,163222,163734,163418,163418,163418,163418,163538,131104,131104,0, };
static const uint32_t _64[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,32800,98336,32800,32800,32800,131072,131072,131072,131072,131072,131072,131073,131072,131104,131104,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131104,131072,131072,131104,0,0,32800,32800,32800,32800,32800,131072,131072,131072,131073,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,151552,151553,131072,131072,131072,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131072,151552,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,32800,32801,32801,32801,32801,32768,32800,32800,32800,32800,32800,32800,32800,32800,32800,32800,131072,161846,0,0,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131073,161846, };
static const uint32_t _65[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131073,0,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,161846,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,86049,86049,86049,86049,86049,86049,86049,86049,86049,86049,86049,86049,86049,65568,65568,65568,65568,86049,65568,65568,65568,86049,86049,86048,86048,86048,86048,86049,86049,86049,86049,86049,86048,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _66[] = { 131072,131072,161819,131072,131072,131072,131072,161819,131072,131072,161815,161819,161819,161819,161815,161815,161819,161819,161819,161815,131072,161819,131072,131072,161793,161819,161819,161819,161819,161819,131072,131072,131072,131072,131072,131072,161819,131072,163418,131072,161819,131073,163418,163418,161819,161819,161792,161815,161819,161819,163418,161819,161815,161795,161795,161795,161795,161814,131072,131072,161815,161815,161819,161819,131073,131073,131073,131073,131073,161819,161815,161815,161815,161815,131072,131073,131072,131072,163222,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222, };
static const uint32_t _67[] = { 161794,161794,161794,163418,163222,161794,161794,161794,161794,131072,131072,131072,0,0,0,0,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131072,131073,131073,131073,131073,131073,131073,131072,131073,131073,131072,131072,131072,131072,131073,131073,131072,131072,131072,131072,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131072,131073,131072,131072,131072,131072,131072,131072,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073, };
static const uint32_t _70[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131072,131072,131072, };
static const uint32_t _71[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131072,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131073,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _72[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _73[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132698,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,132502,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _75[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131072,131072,131072,131072,131073,131073,131073,131073,131073,131073,131072,131072,131072,131072,131073,131073,131072,131072,131073,131073,131072,131072,131072,131073,131073,131073,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131072,131073,131072,131072,131073,131073,131073,131073,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131073,131073,131073,131073, };
static const uint32_t _76[] = { 131072,131072,131072,131072,131072,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131072,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _79[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073, };
static const uint32_t _86[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131073,131072,131072,131073,131073,131073,131073,131073,131073,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _87[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _88[] = { 163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,0,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,0,163418,163222,163418,163418,163418,163222,163222,163418,163222,163418,163222,163418,163222,163418,163418,163418,163418,161814,163418,163222,161814,163418,163222,161814,161814,161814,161814,161814,161846,161846,163418,163418, };
static const uint32_t _89[] = { 163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,161814,131072,131072,131072,131072,131072,131072,163418,163222,163418,163222,86048,86048,86048,163418,163222,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _90[] = { 163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,0,163222,0,0,0,0,0,163222,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,161826,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,348192, };
static const uint32_t _91[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050, };
static const uint32_t _92[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131106,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _93[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _95[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0, };
static const uint32_t _96[] = { 131072,131072,131072,131072,131072,161826,161794,161794,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,86048,86048,86048,86016,86016,131072,161826,161826,161826,161826,161826,131072,131072,161794,161794,161794,161826,161794,131072,131072,131072,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _97[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,86048,86048,137248,137248,161826,161826,161794,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,161826,161826,161826,161794, };
static const uint32_t _98[] = { 0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,194562,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _99[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _100[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _155[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _319[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0, };
static const uint32_t _320[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161826,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _329[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161826,161826,161826,161826,161826,161826,131072,131072, };
static const uint32_t _332[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161826,131072,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,161794,86048,65568,65568,65568,131072,86050,86050,86050,86050,86050,86050,86050,86050,86048,86048,131072,161826, };
static const uint32_t _333[] = { 163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,161846,161846,86050,86050,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,86048,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0, };
static const uint32_t _334[] = { 131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,161826,161826,161826,161826,161826,161826,161826,161826,161826,131104,131104,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,161814,161814,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,161846,161814,161814,161814,161814,161814,161814,161814,161814,163418,163222,163418,163222,163418,163418,163222, };
static const uint32_t _335[] = { 163418,163222,163418,163222,163418,163222,163418,163222,161826,131104,131104,163418,163222,163418,161814,161794,163418,163222,163418,163222,163222,161814,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163418,163418,163418,163418,161814,163418,163418,163418,163418,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,163418,163222,0,0,163418,163222,163418,163418,163418,163418,163222,163418,163222,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,163418,163222,161794,161846,161846,161814,161794,161794,161794,161794,161794, };
static const uint32_t _336[] = { 161794,161794,86050,161794,161794,161794,348192,161794,161794,161794,161794,86050,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,86050,86050,151554,131072,131072,131072,131072,348192,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,131072,0,0,0,0,0,0,0,0, };
static const uint32_t _337[] = { 151554,151554,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,348192,86050,0,0,0,0,0,0,0,0,131072,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,161794,161794,161794,161794,161794,161794,131072,131072,131072,161794,131072,161794,161794,86050, };
static const uint32_t _338[] = { 151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,86048,86048,86048,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,151554,413696,0,0,0,0,0,0,0,0,0,0,0,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0, };
static const uint32_t _339[] = { 86050,86050,86050,151554,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,151554,151554,86050,86050,86050,86050,151554,151554,86050,86050,151554,151554,413696,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,161826,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,131072,131072,161794,161794,161794,161794,161794,86050,161826,161794,161794,161794,161794,161794,161794,161794,161794,161794,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,161794,161794,161794,161794,161794,0, };
static const uint32_t _340[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,86050,151554,151554,86050,86050,151554,151554,86050,86050,0,0,0,0,0,0,0,0,0,161794,161794,161794,86050,161794,161794,161794,161794,161794,161794,161794,161794,86050,151554,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,131072,131072,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161826,161794,161794,161794,161794,161794,161794,131072,131072,131072,161794,151554,86050,151554,161794,161794, };
static const uint32_t _341[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,161794,86050,86050,86050,161794,161794,86050,86050,161794,161794,161794,161794,161794,86050,86048,161794,86048,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161826,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,86050,86050,151554,151554,131072,131072,161794,161826,161826,151554,348192,0,0,0,0,0,0,0,0,0, };
static const uint32_t _342[] = { 0,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,161814,163222,161814,161814,161814,161814,161814,161814,161814,131104,161846,161846,161846,161846,161814,161814,161814,161814,161814,161814,161814,161814,161814,161826,131104,131104,0,0,0,0,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734, };
static const uint32_t _343[] = { 163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,163734,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,86050,151554,151554,86050,151554,151554,131072,151552,348192,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0, };
static const uint32_t _431[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0, };
static const uint32_t _500[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _501[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _502[] = { 163734,163734,163734,163734,163734,163734,163734,0,0,0,0,0,0,0,0,0,0,0,0,163734,163734,163734,163734,163734,0,0,0,0,0,161794,86050,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131073,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,0,161794,0,161794,161794,0,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _503[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,131104,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _504[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,137218,137218,137218,137218,137218,137218,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _506[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _507[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,137218,137218,131072,131072,0,0, };
static const uint32_t _508[] = { 118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,131072,131072,131072,131104,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,131072,131072,131072,151552,151552,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,151552,151552,151552,131072,131072,131104,0,131072,131104,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131073,131073,131073,131073,131073,131073,0,131073,131072,131072,131072,0,0,0,0,137218,161794,137218,161794,137218,0,137218,161794,137218,161794,137218,161794,137218,161794,137218,161794, };
static const uint32_t _509[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,32800, };
static const uint32_t _510[] = { 0,131072,131072,131072,131072,131072,131072,131104,131072,131072,131072,131073,131072,131072,131104,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131104,131072,131073,131073,131073,131072,131072,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,131072,131073,131072,131105,151552,131104,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,131072,131073,131072,131073,131072,131072,131072,131072,131072,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161826,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _511[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,88098,88098,194562,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,0,0,0,131072,131072,131073,131104,131072,131072,131072,0,131072,131073,131073,131073,131073,131072,131072,0,32768,32768,32768,32768,32768,32768,32768,32768,32768,32,32,32,131072,131072,0,0, };
static const uint32_t _512[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _513[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0, };
static const uint32_t _514[] = { 131072,131072,131072,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _515[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,86048,0,0, };
static const uint32_t _517[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,86048,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0, };
static const uint32_t _518[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,0,0,0,0,0, };
static const uint32_t _519[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,131072,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _520[] = { 163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _521[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,0,0,0,0,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,0,0,0,0, };
static const uint32_t _522[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _526[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _528[] = { 161794,161794,161794,161794,161794,161794,0,0,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,0,0,0,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _529[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,0,0,0,0,0,131072,131072,131072,131072,131072, };
static const uint32_t _530[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,131072,131072,131072,0,0,0,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _531[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,131072,131072,161794,161794,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _532[] = { 161794,86050,86050,86050,0,86050,86050,0,0,0,0,0,86050,86050,86050,86050,161794,161794,161794,161794,0,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,86048,86048,86048,0,0,0,0,348192,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072, };
static const uint32_t _533[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,86048,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0, };
static const uint32_t _534[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,131072,131072,131072,131072,131072,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,131072,131072,131072,131072,131072,131072,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _535[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _536[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _537[] = { 163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,0,0,0,0,0,0,0,0,0,0,0,0,0,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072, };
static const uint32_t _538[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,0,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _540[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0, };
static const uint32_t _541[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,86050,86050,131072,0,0,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _542[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,161794,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _543[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0, };
static const uint32_t _544[] = { 151554,86050,151554,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,348192,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,348192, };
static const uint32_t _545[] = { 86048,86048,151554,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,151554,86050,86050,86050,86050,151554,151554,348192,86048,131072,131072,32,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,32,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0, };
static const uint32_t _546[] = { 86050,86050,86050,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,151554,86050,86050,86050,86050,86050,86050,348192,348192,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,161794,151554,151554,161794,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,131072,131072,161794,0,0,0,0,0,0,0,0,0, };
static const uint32_t _547[] = { 86050,86050,151554,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,151554,86050,86050,86050,86050,86050,86050,86050,86050,86050,151554,413696,161794,161794,161794,161794,131072,131072,131072,131072,86048,86048,86048,86048,131072,151554,86050,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,161794,131072,161794,131072,131072,131072,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _548[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,151554,86050,86050,86050,151554,151554,86050,413696,86048,86050,131072,131072,131072,131072,131072,131072,86050,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _549[] = { 161794,161794,161794,161794,161794,161794,161794,0,161794,0,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,131072,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,151554,151554,151554,86050,86050,86050,86050,86050,86050,86048,348192,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0, };
static const uint32_t _550[] = { 86050,86050,151554,151554,0,161794,161794,161794,161794,161794,161794,161794,161794,0,0,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,0,161794,161794,161794,161794,161794,0,86048,86048,161794,86018,151554,86050,151554,151554,151554,151554,0,0,151554,151554,0,0,151554,151554,413696,0,0,161794,0,0,0,0,0,0,86018,0,0,0,0,0,161794,161794,161794,161794,161794,151554,151554,0,0,86048,86048,86048,86048,86048,86048,86048,0,0,0,86048,86048,86048,86048,86048,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _552[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,151554,86050,86050,86050,86050,86050,86050,86050,86050,151554,151554,348192,86050,86050,151554,86048,161794,161794,161794,161794,131072,131072,131072,131072,131072,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,0,131072,86048,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _553[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86018,151554,151554,86050,86050,86050,86050,86050,86050,151554,86050,151554,151554,86018,151554,86050,86050,151554,348192,86048,161794,161794,131072,161794,0,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _555[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86018,151554,151554,86050,86050,86050,86050,0,0,151554,151554,151554,151554,86050,86050,151554,348192,86048,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,161794,161794,161794,161794,86050,86050,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _556[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,151554,86050,86050,86050,86050,86050,86050,86050,86050,151554,151554,86050,151554,348192,86050,131072,131072,131072,161794,0,0,0,0,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _557[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,151554,86050,151554,151554,86050,86050,86050,86050,86050,86050,413696,86048,161794,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _558[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,86050,86050,86050,151554,151554,86050,86050,86050,86050,151554,86050,86050,86050,86050,348192,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _560[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,151554,86050,86050,86050,86050,86050,86050,86050,86050,86050,151554,348192,86048,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _561[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,161794, };
static const uint32_t _562[] = { 161794,161794,161794,161794,161794,161794,161794,0,0,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86018,151554,151554,151554,151554,151554,0,151554,151554,0,0,86050,86050,413696,348192,161794,151554,161794,151554,86048,131072,131072,131072,0,0,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _563[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,151554,86050,86050,86050,86050,0,0,86050,86050,151554,151554,151554,151554,348192,161794,131072,161794,151554,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _564[] = { 161794,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,348192,86050,86050,86050,86050,151554,161794,86050,86050,86050,86050,131072,131072,131072,131072,131072,131072,131072,131072,348192,0,0,0,0,0,0,0,0,161794,86050,86050,86050,86050,86050,86050,151554,151554,86050,86050,86050,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _565[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,151554,86048,348192,131072,131072,131072,161794,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0, };
static const uint32_t _568[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,86050,86050,86050,86050,86050,86050,86050,0,86050,86050,86050,86050,86050,86050,151554,348192,161794,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,131072,131072,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _569[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,0,151554,86050,86050,86050,86050,86050,86050,86050,151554,86050,86050,151554,86050,86050,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _570[] = { 161794,161794,161794,161794,161794,161794,161794,0,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,86050,86050,86050,86050,0,0,0,86050,0,86050,86050,0,86050,86050,86050,86048,86050,348192,348192,161794,86050,0,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,0,161794,161794,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _571[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,151554,151554,151554,151554,151554,0,86050,86050,0,151554,151554,86050,151554,348192,161794,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _573[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86050,86050,151554,151554,131072,131072,0,0,0,0,0,0,0, };
static const uint32_t _575[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,131072, };
static const uint32_t _583[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _584[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _586[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _616[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,32,32,32,32,32,32,32,32,32,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _652[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _724[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _725[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,86048,86048,86048,86048,86048,131072,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _726[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,86048,86048,86048,86048,86048,86048,131072,131072,131072,131072,131072,131072,131072,131072,131072,161826,161826,161826,161826,131072,131072,0,0,0,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,131072,131072,131072,131072,131072,131072,131072,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,161794,161794,161794, };
static const uint32_t _727[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _732[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222, };
static const uint32_t _733[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _734[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,86050,161794,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554,151554, };
static const uint32_t _735[] = { 151554,151554,151554,151554,151554,151554,151554,151554,0,0,0,0,0,0,0,86050,86050,86050,86050,161826,161826,161826,161826,161826,161826,161826,161826,161826,161826,161826,161826,161826,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161826,161826,131072,161826,86048,0,0,0,0,0,0,0,0,0,0,0,151554,151554,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _783[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0, };
static const uint32_t _793[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _794[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _866[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _869[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0, };
static const uint32_t _888[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0, };
static const uint32_t _889[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,131072,86048,86050,131072,32800,32800,32800,32800,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _929[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _930[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,86016,151552,86048,86048,86048,131072,131072,131072,151552,86016,86016,86016,86016,86016,32800,32800,32800,32800,32800,32800,32800,32800,86048,86048,86048,86048,86048, };
static const uint32_t _931[] = { 86048,86048,86048,131072,131072,86048,86048,86048,86048,86048,86048,86048,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,86048,86048,86048,86048,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _932[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,86048,86048,86048,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _933[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _934[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0, };
static const uint32_t _936[] = { 161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161815,161815,161815,161815,161815,161815,161815,0,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819, };
static const uint32_t _937[] = { 161819,161819,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161819,0,161819,161819,0,0,161819,0,0,161819,161819,0,0,161819,161819,161819,161819,0,161819,161819,161819,161819,161819,161819,161819,161819,161815,161815,161815,161815,0,161815,0,161815,161815,161815,161815,161815,161815,161815,0,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815, };
static const uint32_t _938[] = { 161815,161815,161815,161815,161819,161819,0,161819,161819,161819,161819,0,0,161819,161819,161819,161819,161819,161819,161819,161819,0,161819,161819,161819,161819,161819,161819,161819,0,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161819,161819,0,161819,161819,161819,161819,0,161819,161819,161819,161819,161819,0,161819,0,0,0,161819,161819,161819,161819,161819,161819,161819,0,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819, };
static const uint32_t _939[] = { 161819,161819,161819,161819,161819,161819,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815, };
static const uint32_t _940[] = { 161815,161815,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819, };
static const uint32_t _941[] = { 161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,0,0,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,131073,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,131073,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,131073,161815,161815,161815,161815, };
static const uint32_t _942[] = { 161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,131073,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,131073,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,131073,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,131073,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815, };
static const uint32_t _943[] = { 161815,161815,161815,161815,161815,161815,161815,161815,161815,131073,161815,161815,161815,161815,161815,161815,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,161819,131073,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,161815,131073,161815,161815,161815,161815,161815,161815,161819,161815,0,0,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553,151553, };
static const uint32_t _948[] = { 86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,131072,131072,131072,131072,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,131072,131072,131072,131072,131072,131072,131072,131072,86048,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _949[] = { 131072,131072,131072,131072,86048,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,86048,86048,86048,86048,86048,0,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,86048,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _960[] = { 86050,86050,86050,86050,86050,86050,86050,0,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,86050,0,0,86050,86050,86050,86050,86050,86050,86050,0,86050,86050,0,86050,86050,86050,86050,86050,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _962[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,86048,86048,86048,86048,86048,86048,86048,161826,161826,161826,161826,161826,161826,161826,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,161794,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _965[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,86048,86048,86048,86048,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,131072, };
static const uint32_t _977[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,86048,86048,86048,86048,86048,86048,86048,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _978[] = { 163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163418,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,163222,86048,86048,86048,86050,86048,86048,86048,161826,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _984[] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _985[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _986[] = { 0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _988[] = { 161795,161795,161795,161795,0,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,0,161795,161795,0,161795,0,0,161795,0,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,0,161795,161795,161795,161795,0,161795,0,161795,0,0,0,0,0,0,161795,0,0,0,0,161795,0,161795,0,161795,0,161795,161795,161795,0,161795,161795,0,161795,0,0,161795,0,161795,0,161795,0,161795,0,161795,0,161795,161795,0,161795,0,0,161795,161795,161795,161795,0,161795,161795,161795,161795,161795,161795,161795,0,161795,161795,161795,161795,0,161795,161795,161795,161795,0,161795,0, };
static const uint32_t _989[] = { 161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,0,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,0,0,0,0,0,161795,161795,161795,0,161795,161795,161795,161795,161795,0,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,161795,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131073,131073,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _992[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _993[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _994[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131072,131072,131072,131072,131072,131072,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131072,131072,131072,131072,131072,131072,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131098, };
static const uint32_t _995[] = { 131098,131098,131098,131098,131098,131098,131098,131098,131098,131098,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _996[] = { 131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _999[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131104,131104,131104,131104,131104, };
static const uint32_t _1005[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0, };
static const uint32_t _1006[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _1007[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _1008[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _1009[] = { 131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _1010[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,131072,131072,131072,131072,131072,131072, };
static const uint32_t _1011[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072, };
static const uint32_t _1012[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,131072,131072,131072,131072,131072,0,0,0,131072,131072,131072,0,0,0,0,0, };
static const uint32_t _1013[] = { 131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _1015[] = { 131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,131072,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,151552,151552,151552,151552,151552,151552,151552,151552,151552,151552,0,0,0,0,0,0, };
static const uint32_t _1357[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _1390[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _1392[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _1437[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794, };
static const uint32_t _1495[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _1524[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _1574[] = { 161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,161794,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, };
static const uint32_t _7168[] = { 32768,32800,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336,98336, };
static const uint32_t _7171[] = { 118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,118816,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768,32768, };
static const uint32_t *_blocks[] = {
 _0,_1,_2,_3,_4,_5,_6,_7,
 _8,_9,_10,_11,_12,_13,_14,_15,
 _16,_17,_18,_19,_20,_21,_22,_23,
 _24,_25,_26,_27,_28,_29,_30,_31,
 _32,_33,_34,0,_36,_37,_38,_39,
 _40,0,0,0,_44,_45,_46,_47,
 _48,_49,_50,_51,_52,_53,_54,_55,
 _56,_57,_58,_59,_60,_61,_62,_63,
 _64,_65,_66,_67,0,0,_70,_71,
 _72,_73,0,_75,_76,0,0,_79,
 0,0,0,0,0,0,_86,_87,
 _88,_89,_90,_91,_92,_93,0,_95,
 _96,_97,_98,_99,_100,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_155,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_319,
 _320,0,0,0,0,0,0,0,
 0,_329,0,0,_332,_333,_334,_335,
 _336,_337,_338,_339,_340,_341,_342,_343,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_431,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_500,_501,_502,_503,
 _504,0,_506,_507,_508,_509,_510,_511,
 _512,_513,_514,_515,0,_517,_518,_519,
 _520,_521,_522,0,0,0,_526,0,
 _528,_529,_530,_531,_532,_533,_534,_535,
 _536,_537,_538,0,_540,_541,_542,_543,
 _544,_545,_546,_547,_548,_549,_550,0,
 _552,_553,0,_555,_556,_557,_558,0,
 _560,_561,_562,_563,_564,_565,0,0,
 _568,_569,_570,_571,0,_573,0,_575,
 0,0,0,0,0,0,0,_583,
 _584,0,_586,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _616,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_652,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_724,_725,_726,_727,
 0,0,0,0,_732,_733,_734,_735,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_783,
 0,0,0,0,0,0,0,0,
 0,_793,_794,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_866,0,0,_869,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _888,_889,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_929,_930,_931,_932,_933,_934,0,
 _936,_937,_938,_939,_940,_941,_942,_943,
 0,0,0,0,_948,_949,0,0,
 0,0,0,0,0,0,0,0,
 _960,0,_962,0,0,_965,0,0,
 0,0,0,0,0,0,0,0,
 0,_977,_978,0,0,0,0,0,
 _984,_985,_986,0,_988,_989,0,0,
 _992,_993,_994,_995,_996,0,0,_999,
 0,0,0,0,0,_1005,_1006,_1007,
 _1008,_1009,_1010,_1011,_1012,_1013,0,_1015,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1357,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1390,0,
 _1392,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1437,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_1495,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_1524,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1574,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _7168,0,0,_7171,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
static const uint32_t _block_values[] = {
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,161794,0,0,0,0,
 0,161794,161794,161794,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,131073,131073,0,0,
 0,0,131072,0,0,131072,131072,0,
 131072,131072,131073,131073,131073,131073,0,0,
 0,0,0,0,0,0,131072,0,
 0,0,0,0,0,131072,131072,131072,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,0,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,0,
 0,161794,161794,161794,161794,161794,161794,161794,
 161794,0,161794,161794,0,0,0,0,
 0,0,0,0,0,0,0,0,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,161794,161794,0,0,0,0,
 0,161794,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,161794,161794,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 161794,161794,161794,161794,161794,161794,161794,0,
 0,161794,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 161794,161794,161794,161794,161794,161794,161794,161794,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 161794,161794,161794,161794,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 161794,161794,161794,161794,0,0,0,0,
 0,0,0,0,0,0,0,0,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,0,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 161794,161794,0,161794,161794,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 131072,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 131072,131072,131072,131072,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 161794,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,131072,0,
 131072,131072,131072,131072,131072,0,0,0,
 0,0,0,0,0,0,131072,0,
 0,0,0,0,0,0,0,0,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,0,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,0,161794,
 0,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,0,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 161794,161794,161794,161794,0,0,0,0,
 0,0,0,0,0,0,0,0,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,161794,161794,
 161794,161794,161794,161794,161794,161794,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,32768,118816,0,32768,32768,32768,32768,
 32768,32768,32768,32768,32768,32768,32768,32768,
 32768,32768,32768,32768,32768,32768,32768,32768,
 32768,32768,32768,32768,32768,32768,32768,32768,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
inline uint32_t get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}
const std::unordered_map<char32_t, const char32_t*> _simple_case_mappings = {
{ 0x00000041, U"\U00000041\U00000061\U00000041" },
{ 0x00000042, U"\U00000042\U00000062\U00000042" },
{ 0x00000043, U"\U00000043\U00000063\U00000043" },
{ 0x00000044, U"\U00000044\U00000064\U00000044" },
{ 0x00000045, U"\U00000045\U00000065\U00000045" },
{ 0x00000046, U"\U00000046\U00000066\U00000046" },
{ 0x00000047, U"\U00000047\U00000067\U00000047" },
{ 0x00000048, U"\U00000048\U00000068\U00000048" },
{ 0x00000049, U"\U00000049\U00000069\U00000049" },
{ 0x0000004A, U"\U0000004A\U0000006A\U0000004A" },
{ 0x0000004B, U"\U0000004B\U0000006B\U0000004B" },
{ 0x0000004C, U"\U0000004C\U0000006C\U0000004C" },
{ 0x0000004D, U"\U0000004D\U0000006D\U0000004D" },
{ 0x0000004E, U"\U0000004E\U0000006E\U0000004E" },
{ 0x0000004F, U"\U0000004F\U0000006F\U0000004F" },
{ 0x00000050, U"\U00000050\U00000070\U00000050" },
{ 0x00000051, U"\U00000051\U00000071\U00000051" },
{ 0x00000052, U"\U00000052\U00000072\U00000052" },
{ 0x00000053, U"\U00000053\U00000073\U00000053" },
{ 0x00000054, U"\U00000054\U00000074\U00000054" },
{ 0x00000055, U"\U00000055\U00000075\U00000055" },
{ 0x00000056, U"\U00000056\U00000076\U00000056" },
{ 0x00000057, U"\U00000057\U00000077\U00000057" },
{ 0x00000058, U"\U00000058\U00000078\U00000058" },
{ 0x00000059, U"\U00000059\U00000079\U00000059" },
{ 0x0000005A, U"\U0000005A\U0000007A\U0000005A" },
{ 0x00000061, U"\U00000041\U00000061\U00000041" },
{ 0x00000062, U"\U00000042\U00000062\U00000042" },
{ 0x00000063, U"\U00000043\U00000063\U00000043" },
{ 0x00000064, U"\U00000044\U00000064\U00000044" },
{ 0x00000065, U"\U00000045\U00000065\U00000045" },
{ 0x00000066, U"\U00000046\U00000066\U00000046" },
{ 0x00000067, U"\U00000047\U00000067\U00000047" },
{ 0x00000068, U"\U00000048\U00000068\U00000048" },
{ 0x00000069, U"\U00000049\U00000069\U00000049" },
{ 0x0000006A, U"\U0000004A\U0000006A\U0000004A" },
{ 0x0000006B, U"\U0000004B\U0000006B\U0000004B" },
{ 0x0000006C, U"\U0000004C\U0000006C\U0000004C" },
{ 0x0000006D, U"\U0000004D\U0000006D\U0000004D" },
{ 0x0000006E, U"\U0000004E\U0000006E\U0000004E" },
{ 0x0000006F, U"\U0000004F\U0000006F\U0000004F" },
{ 0x00000070, U"\U00000050\U00000070\U00000050" },
{ 0x00000071, U"\U00000051\U00000071\U00000051" },
{ 0x00000072, U"\U00000052\U00000072\U00000052" },
{ 0x00000073, U"\U00000053\U00000073\U00000053" },
{ 0x00000074, U"\U00000054\U00000074\U00000054" },
{ 0x00000075, U"\U00000055\U00000075\U00000055" },
{ 0x00000076, U"\U00000056\U00000076\U00000056" },
{ 0x00000077, U"\U00000057\U00000077\U00000057" },
{ 0x00000078, U"\U00000058\U00000078\U00000058" },
{ 0x00000079, U"\U00000059\U00000079\U00000059" },
{ 0x0000007A, U"\U0000005A\U0000007A\U0000005A" },
{ 0x000000B5, U"\U0000039C\U000000B5\U0000039C" },
{ 0x000000C0, U"\U000000C0\U000000E0\U000000C0" },
{ 0x000000C1, U"\U000000C1\U000000E1\U000000C1" },
{ 0x000000C2, U"\U000000C2\U000000E2\U000000C2" },
{ 0x000000C3, U"\U000000C3\U000000E3\U000000C3" },
{ 0x000000C4, U"\U000000C4\U000000E4\U000000C4" },
{ 0x000000C5, U"\U000000C5\U000000E5\U000000C5" },
{ 0x000000C6, U"\U000000C6\U000000E6\U000000C6" },
{ 0x000000C7, U"\U000000C7\U000000E7\U000000C7" },
{ 0x000000C8, U"\U000000C8\U000000E8\U000000C8" },
{ 0x000000C9, U"\U000000C9\U000000E9\U000000C9" },
{ 0x000000CA, U"\U000000CA\U000000EA\U000000CA" },
{ 0x000000CB, U"\U000000CB\U000000EB\U000000CB" },
{ 0x000000CC, U"\U000000CC\U000000EC\U000000CC" },
{ 0x000000CD, U"\U000000CD\U000000ED\U000000CD" },
{ 0x000000CE, U"\U000000CE\U000000EE\U000000CE" },
{ 0x000000CF, U"\U000000CF\U000000EF\U000000CF" },
{ 0x000000D0, U"\U000000D0\U000000F0\U000000D0" },
{ 0x000000D1, U"\U000000D1\U000000F1\U000000D1" },
{ 0x000000D2, U"\U000000D2\U000000F2\U000000D2" },
{ 0x000000D3, U"\U000000D3\U000000F3\U000000D3" },
{ 0x000000D4, U"\U000000D4\U000000F4\U000000D4" },
{ 0x000000D5, U"\U000000D5\U000000F5\U000000D5" },
{ 0x000000D6, U"\U000000D6\U000000F6\U000000D6" },
{ 0x000000D8, U"\U000000D8\U000000F8\U000000D8" },
{ 0x000000D9, U"\U000000D9\U000000F9\U000000D9" },
{ 0x000000DA, U"\U000000DA\U000000FA\U000000DA" },
{ 0x000000DB, U"\U000000DB\U000000FB\U000000DB" },
{ 0x000000DC, U"\U000000DC\U000000FC\U000000DC" },
{ 0x000000DD, U"\U000000DD\U000000FD\U000000DD" },
{ 0x000000DE, U"\U000000DE\U000000FE\U000000DE" },
{ 0x000000E0, U"\U000000C0\U000000E0\U000000C0" },
{ 0x000000E1, U"\U000000C1\U000000E1\U000000C1" },
{ 0x000000E2, U"\U000000C2\U000000E2\U000000C2" },
{ 0x000000E3, U"\U000000C3\U000000E3\U000000C3" },
{ 0x000000E4, U"\U000000C4\U000000E4\U000000C4" },
{ 0x000000E5, U"\U000000C5\U000000E5\U000000C5" },
{ 0x000000E6, U"\U000000C6\U000000E6\U000000C6" },
{ 0x000000E7, U"\U000000C7\U000000E7\U000000C7" },
{ 0x000000E8, U"\U000000C8\U000000E8\U000000C8" },
{ 0x000000E9, U"\U000000C9\U000000E9\U000000C9" },
{ 0x000000EA, U"\U000000CA\U000000EA\U000000CA" },
{ 0x000000EB, U"\U000000CB\U000000EB\U000000CB" },
{ 0x000000EC, U"\U000000CC\U000000EC\U000000CC" },
{ 0x000000ED, U"\U000000CD\U000000ED\U000000CD" },
{ 0x000000EE, U"\U000000CE\U000000EE\U000000CE" },
{ 0x000000EF, U"\U000000CF\U000000EF\U000000CF" },
{ 0x000000F0, U"\U000000D0\U000000F0\U000000D0" },
{ 0x000000F1, U"\U000000D1\U000000F1\U000000D1" },
{ 0x000000F2, U"\U000000D2\U000000F2\U000000D2" },
{ 0x000000F3, U"\U000000D3\U000000F3\U000000D3" },
{ 0x000000F4, U"\U000000D4\U000000F4\U000000D4" },
{ 0x000000F5, U"\U000000D5\U000000F5\U000000D5" },
{ 0x000000F6, U"\U000000D6\U000000F6\U000000D6" },
{ 0x000000F8, U"\U000000D8\U000000F8\U000000D8" },
{ 0x000000F9, U"\U000000D9\U000000F9\U000000D9" },
{ 0x000000FA, U"\U000000DA\U000000FA\U000000DA" },
{ 0x000000FB, U"\U000000DB\U000000FB\U000000DB" },
{ 0x000000FC, U"\U000000DC\U000000FC\U000000DC" },
{ 0x000000FD, U"\U000000DD\U000000FD\U000000DD" },
{ 0x000000FE, U"\U000000DE\U000000FE\U000000DE" },
{ 0x000000FF, U"\U00000178\U000000FF\U00000178" },
{ 0x00000100, U"\U00000100\U00000101\U00000100" },
{ 0x00000101, U"\U00000100\U00000101\U00000100" },
{ 0x00000102, U"\U00000102\U00000103\U00000102" },
{ 0x00000103, U"\U00000102\U00000103\U00000102" },
{ 0x00000104, U"\U00000104\U00000105\U00000104" },
{ 0x00000105, U"\U00000104\U00000105\U00000104" },
{ 0x00000106, U"\U00000106\U00000107\U00000106" },
{ 0x00000107, U"\U00000106\U00000107\U00000106" },
{ 0x00000108, U"\U00000108\U00000109\U00000108" },
{ 0x00000109, U"\U00000108\U00000109\U00000108" },
{ 0x0000010A, U"\U0000010A\U0000010B\U0000010A" },
{ 0x0000010B, U"\U0000010A\U0000010B\U0000010A" },
{ 0x0000010C, U"\U0000010C\U0000010D\U0000010C" },
{ 0x0000010D, U"\U0000010C\U0000010D\U0000010C" },
{ 0x0000010E, U"\U0000010E\U0000010F\U0000010E" },
{ 0x0000010F, U"\U0000010E\U0000010F\U0000010E" },
{ 0x00000110, U"\U00000110\U00000111\U00000110" },
{ 0x00000111, U"\U00000110\U00000111\U00000110" },
{ 0x00000112, U"\U00000112\U00000113\U00000112" },
{ 0x00000113, U"\U00000112\U00000113\U00000112" },
{ 0x00000114, U"\U00000114\U00000115\U00000114" },
{ 0x00000115, U"\U00000114\U00000115\U00000114" },
{ 0x00000116, U"\U00000116\U00000117\U00000116" },
{ 0x00000117, U"\U00000116\U00000117\U00000116" },
{ 0x00000118, U"\U00000118\U00000119\U00000118" },
{ 0x00000119, U"\U00000118\U00000119\U00000118" },
{ 0x0000011A, U"\U0000011A\U0000011B\U0000011A" },
{ 0x0000011B, U"\U0000011A\U0000011B\U0000011A" },
{ 0x0000011C, U"\U0000011C\U0000011D\U0000011C" },
{ 0x0000011D, U"\U0000011C\U0000011D\U0000011C" },
{ 0x0000011E, U"\U0000011E\U0000011F\U0000011E" },
{ 0x0000011F, U"\U0000011E\U0000011F\U0000011E" },
{ 0x00000120, U"\U00000120\U00000121\U00000120" },
{ 0x00000121, U"\U00000120\U00000121\U00000120" },
{ 0x00000122, U"\U00000122\U00000123\U00000122" },
{ 0x00000123, U"\U00000122\U00000123\U00000122" },
{ 0x00000124, U"\U00000124\U00000125\U00000124" },
{ 0x00000125, U"\U00000124\U00000125\U00000124" },
{ 0x00000126, U"\U00000126\U00000127\U00000126" },
{ 0x00000127, U"\U00000126\U00000127\U00000126" },
{ 0x00000128, U"\U00000128\U00000129\U00000128" },
{ 0x00000129, U"\U00000128\U00000129\U00000128" },
{ 0x0000012A, U"\U0000012A\U0000012B\U0000012A" },
{ 0x0000012B, U"\U0000012A\U0000012B\U0000012A" },
{ 0x0000012C, U"\U0000012C\U0000012D\U0000012C" },
{ 0x0000012D, U"\U0000012C\U0000012D\U0000012C" },
{ 0x0000012E, U"\U0000012E\U0000012F\U0000012E" },
{ 0x0000012F, U"\U0000012E\U0000012F\U0000012E" },
{ 0x00000130, U"\U00000130\U00000069\U00000130" },
{ 0x00000131, U"\U00000049\U00000131\U00000049" },
{ 0x00000132, U"\U00000132\U00000133\U00000132" },
{ 0x00000133, U"\U00000132\U00000133\U00000132" },
{ 0x00000134, U"\U00000134\U00000135\U00000134" },
{ 0x00000135, U"\U00000134\U00000135\U00000134" },
{ 0x00000136, U"\U00000136\U00000137\U00000136" },
{ 0x00000137, U"\U00000136\U00000137\U00000136" },
{ 0x00000139, U"\U00000139\U0000013A\U00000139" },
{ 0x0000013A, U"\U00000139\U0000013A\U00000139" },
{ 0x0000013B, U"\U0000013B\U0000013C\U0000013B" },
{ 0x0000013C, U"\U0000013B\U0000013C\U0000013B" },
{ 0x0000013D, U"\U0000013D\U0000013E\U0000013D" },
{ 0x0000013E, U"\U0000013D\U0000013E\U0000013D" },
{ 0x0000013F, U"\U0000013F\U00000140\U0000013F" },
{ 0x00000140, U"\U0000013F\U00000140\U0000013F" },
{ 0x00000141, U"\U00000141\U00000142\U00000141" },
{ 0x00000142, U"\U00000141\U00000142\U00000141" },
{ 0x00000143, U"\U00000143\U00000144\U00000143" },
{ 0x00000144, U"\U00000143\U00000144\U00000143" },
{ 0x00000145, U"\U00000145\U00000146\U00000145" },
{ 0x00000146, U"\U00000145\U00000146\U00000145" },
{ 0x00000147, U"\U00000147\U00000148\U00000147" },
{ 0x00000148, U"\U00000147\U00000148\U00000147" },
{ 0x0000014A, U"\U0000014A\U0000014B\U0000014A" },
{ 0x0000014B, U"\U0000014A\U0000014B\U0000014A" },
{ 0x0000014C, U"\U0000014C\U0000014D\U0000014C" },
{ 0x0000014D, U"\U0000014C\U0000014D\U0000014C" },
{ 0x0000014E, U"\U0000014E\U0000014F\U0000014E" },
{ 0x0000014F, U"\U0000014E\U0000014F\U0000014E" },
{ 0x00000150, U"\U00000150\U00000151\U00000150" },
{ 0x00000151, U"\U00000150\U00000151\U00000150" },
{ 0x00000152, U"\U00000152\U00000153\U00000152" },
{ 0x00000153, U"\U00000152\U00000153\U00000152" },
{ 0x00000154, U"\U00000154\U00000155\U00000154" },
{ 0x00000155, U"\U00000154\U00000155\U00000154" },
{ 0x00000156, U"\U00000156\U00000157\U00000156" },
{ 0x00000157, U"\U00000156\U00000157\U00000156" },
{ 0x00000158, U"\U00000158\U00000159\U00000158" },
{ 0x00000159, U"\U00000158\U00000159\U00000158" },
{ 0x0000015A, U"\U0000015A\U0000015B\U0000015A" },
{ 0x0000015B, U"\U0000015A\U0000015B\U0000015A" },
{ 0x0000015C, U"\U0000015C\U0000015D\U0000015C" },
{ 0x0000015D, U"\U0000015C\U0000015D\U0000015C" },
{ 0x0000015E, U"\U0000015E\U0000015F\U0000015E" },
{ 0x0000015F, U"\U0000015E\U0000015F\U0000015E" },
{ 0x00000160, U"\U00000160\U00000161\U00000160" },
{ 0x00000161, U"\U00000160\U00000161\U00000160" },
{ 0x00000162, U"\U00000162\U00000163\U00000162" },
{ 0x00000163, U"\U00000162\U00000163\U00000162" },
{ 0x00000164, U"\U00000164\U00000165\U00000164" },
{ 0x00000165, U"\U00000164\U00000165\U00000164" },
{ 0x00000166, U"\U00000166\U00000167\U00000166" },
{ 0x00000167, U"\U00000166\U00000167\U00000166" },
{ 0x00000168, U"\U00000168\U00000169\U00000168" },
{ 0x00000169, U"\U00000168\U00000169\U00000168" },
{ 0x0000016A, U"\U0000016A\U0000016B\U0000016A" },
{ 0x0000016B, U"\U0000016A\U0000016B\U0000016A" },
{ 0x0000016C, U"\U0000016C\U0000016D\U0000016C" },
{ 0x0000016D, U"\U0000016C\U0000016D\U0000016C" },
{ 0x0000016E, U"\U0000016E\U0000016F\U0000016E" },
{ 0x0000016F, U"\U0000016E\U0000016F\U0000016E" },
{ 0x00000170, U"\U00000170\U00000171\U00000170" },
{ 0x00000171, U"\U00000170\U00000171\U00000170" },
{ 0x00000172, U"\U00000172\U00000173\U00000172" },
{ 0x00000173, U"\U00000172\U00000173\U00000172" },
{ 0x00000174, U"\U00000174\U00000175\U00000174" },
{ 0x00000175, U"\U00000174\U00000175\U00000174" },
{ 0x00000176, U"\U00000176\U00000177\U00000176" },
{ 0x00000177, U"\U00000176\U00000177\U00000176" },
{ 0x00000178, U"\U00000178\U000000FF\U00000178" },
{ 0x00000179, U"\U00000179\U0000017A\U00000179" },
{ 0x0000017A, U"\U00000179\U0000017A\U00000179" },
{ 0x0000017B, U"\U0000017B\U0000017C\U0000017B" },
{ 0x0000017C, U"\U0000017B\U0000017C\U0000017B" },
{ 0x0000017D, U"\U0000017D\U0000017E\U0000017D" },
{ 0x0000017E, U"\U0000017D\U0000017E\U0000017D" },
{ 0x0000017F, U"\U00000053\U0000017F\U00000053" },
{ 0x00000180, U"\U00000243\U00000180\U00000243" },
{ 0x00000181, U"\U00000181\U00000253\U00000181" },
{ 0x00000182, U"\U00000182\U00000183\U00000182" },
{ 0x00000183, U"\U00000182\U00000183\U00000182" },
{ 0x00000184, U"\U00000184\U00000185\U00000184" },
{ 0x00000185, U"\U00000184\U00000185\U00000184" },
{ 0x00000186, U"\U00000186\U00000254\U00000186" },
{ 0x00000187, U"\U00000187\U00000188\U00000187" },
{ 0x00000188, U"\U00000187\U00000188\U00000187" },
{ 0x00000189, U"\U00000189\U00000256\U00000189" },
{ 0x0000018A, U"\U0000018A\U00000257\U0000018A" },
{ 0x0000018B, U"\U0000018B\U0000018C\U0000018B" },
{ 0x0000018C, U"\U0000018B\U0000018C\U0000018B" },
{ 0x0000018E, U"\U0000018E\U000001DD\U0000018E" },
{ 0x0000018F, U"\U0000018F\U00000259\U0000018F" },
{ 0x00000190, U"\U00000190\U0000025B\U00000190" },
{ 0x00000191, U"\U00000191\U00000192\U00000191" },
{ 0x00000192, U"\U00000191\U00000192\U00000191" },
{ 0x00000193, U"\U00000193\U00000260\U00000193" },
{ 0x00000194, U"\U00000194\U00000263\U00000194" },
{ 0x00000195, U"\U000001F6\U00000195\U000001F6" },
{ 0x00000196, U"\U00000196\U00000269\U00000196" },
{ 0x00000197, U"\U00000197\U00000268\U00000197" },
{ 0x00000198, U"\U00000198\U00000199\U00000198" },
{ 0x00000199, U"\U00000198\U00000199\U00000198" },
{ 0x0000019A, U"\U0000023D\U0000019A\U0000023D" },
{ 0x0000019C, U"\U0000019C\U0000026F\U0000019C" },
{ 0x0000019D, U"\U0000019D\U00000272\U0000019D" },
{ 0x0000019E, U"\U00000220\U0000019E\U00000220" },
{ 0x0000019F, U"\U0000019F\U00000275\U0000019F" },
{ 0x000001A0, U"\U000001A0\U000001A1\U000001A0" },
{ 0x000001A1, U"\U000001A0\U000001A1\U000001A0" },
{ 0x000001A2, U"\U000001A2\U000001A3\U000001A2" },
{ 0x000001A3, U"\U000001A2\U000001A3\U000001A2" },
{ 0x000001A4, U"\U000001A4\U000001A5\U000001A4" },
{ 0x000001A5, U"\U000001A4\U000001A5\U000001A4" },
{ 0x000001A6, U"\U000001A6\U00000280\U000001A6" },
{ 0x000001A7, U"\U000001A7\U000001A8\U000001A7" },
{ 0x000001A8, U"\U000001A7\U000001A8\U000001A7" },
{ 0x000001A9, U"\U000001A9\U00000283\U000001A9" },
{ 0x000001AC, U"\U000001AC\U000001AD\U000001AC" },
{ 0x000001AD, U"\U000001AC\U000001AD\U000001AC" },
{ 0x000001AE, U"\U000001AE\U00000288\U000001AE" },
{ 0x000001AF, U"\U000001AF\U000001B0\U000001AF" },
{ 0x000001B0, U"\U000001AF\U000001B0\U000001AF" },
{ 0x000001B1, U"\U000001B1\U0000028A\U000001B1" },
{ 0x000001B2, U"\U000001B2\U0000028B\U000001B2" },
{ 0x000001B3, U"\U000001B3\U000001B4\U000001B3" },
{ 0x000001B4, U"\U000001B3\U000001B4\U000001B3" },
{ 0x000001B5, U"\U000001B5\U000001B6\U000001B5" },
{ 0x000001B6, U"\U000001B5\U000001B6\U000001B5" },
{ 0x000001B7, U"\U000001B7\U00000292\U000001B7" },
{ 0x000001B8, U"\U000001B8\U000001B9\U000001B8" },
{ 0x000001B9, U"\U000001B8\U000001B9\U000001B8" },
{ 0x000001BC, U"\U000001BC\U000001BD\U000001BC" },
{ 0x000001BD, U"\U000001BC\U000001BD\U000001BC" },
{ 0x000001BF, U"\U000001F7\U000001BF\U000001F7" },
{ 0x000001C4, U"\U000001C4\U000001C6\U000001C5" },
{ 0x000001C5, U"\U000001C4\U000001C6\U000001C5" },
{ 0x000001C6, U"\U000001C4\U000001C6\U000001C5" },
{ 0x000001C7, U"\U000001C7\U000001C9\U000001C8" },
{ 0x000001C8, U"\U000001C7\U000001C9\U000001C8" },
{ 0x000001C9, U"\U000001C7\U000001C9\U000001C8" },
{ 0x000001CA, U"\U000001CA\U000001CC\U000001CB" },
{ 0x000001CB, U"\U000001CA\U000001CC\U000001CB" },
{ 0x000001CC, U"\U000001CA\U000001CC\U000001CB" },
{ 0x000001CD, U"\U000001CD\U000001CE\U000001CD" },
{ 0x000001CE, U"\U000001CD\U000001CE\U000001CD" },
{ 0x000001CF, U"\U000001CF\U000001D0\U000001CF" },
{ 0x000001D0, U"\U000001CF\U000001D0\U000001CF" },
{ 0x000001D1, U"\U000001D1\U000001D2\U000001D1" },
{ 0x000001D2, U"\U000001D1\U000001D2\U000001D1" },
{ 0x000001D3, U"\U000001D3\U000001D4\U000001D3" },
{ 0x000001D4, U"\U000001D3\U000001D4\U000001D3" },
{ 0x000001D5, U"\U000001D5\U000001D6\U000001D5" },
{ 0x000001D6, U"\U000001D5\U000001D6\U000001D5" },
{ 0x000001D7, U"\U000001D7\U000001D8\U000001D7" },
{ 0x000001D8, U"\U000001D7\U000001D8\U000001D7" },
{ 0x000001D9, U"\U000001D9\U000001DA\U000001D9" },
{ 0x000001DA, U"\U000001D9\U000001DA\U000001D9" },
{ 0x000001DB, U"\U000001DB\U000001DC\U000001DB" },
{ 0x000001DC, U"\U000001DB\U000001DC\U000001DB" },
{ 0x000001DD, U"\U0000018E\U000001DD\U0000018E" },
{ 0x000001DE, U"\U000001DE\U000001DF\U000001DE" },
{ 0x000001DF, U"\U000001DE\U000001DF\U000001DE" },
{ 0x000001E0, U"\U000001E0\U000001E1\U000001E0" },
{ 0x000001E1, U"\U000001E0\U000001E1\U000001E0" },
{ 0x000001E2, U"\U000001E2\U000001E3\U000001E2" },
{ 0x000001E3, U"\U000001E2\U000001E3\U000001E2" },
{ 0x000001E4, U"\U000001E4\U000001E5\U000001E4" },
{ 0x000001E5, U"\U000001E4\U000001E5\U000001E4" },
{ 0x000001E6, U"\U000001E6\U000001E7\U000001E6" },
{ 0x000001E7, U"\U000001E6\U000001E7\U000001E6" },
{ 0x000001E8, U"\U000001E8\U000001E9\U000001E8" },
{ 0x000001E9, U"\U000001E8\U000001E9\U000001E8" },
{ 0x000001EA, U"\U000001EA\U000001EB\U000001EA" },
{ 0x000001EB, U"\U000001EA\U000001EB\U000001EA" },
{ 0x000001EC, U"\U000001EC\U000001ED\U000001EC" },
{ 0x000001ED, U"\U000001EC\U000001ED\U000001EC" },
{ 0x000001EE, U"\U000001EE\U000001EF\U000001EE" },
{ 0x000001EF, U"\U000001EE\U000001EF\U000001EE" },
{ 0x000001F1, U"\U000001F1\U000001F3\U000001F2" },
{ 0x000001F2, U"\U000001F1\U000001F3\U000001F2" },
{ 0x000001F3, U"\U000001F1\U000001F3\U000001F2" },
{ 0x000001F4, U"\U000001F4\U000001F5\U000001F4" },
{ 0x000001F5, U"\U000001F4\U000001F5\U000001F4" },
{ 0x000001F6, U"\U000001F6\U00000195\U000001F6" },
{ 0x000001F7, U"\U000001F7\U000001BF\U000001F7" },
{ 0x000001F8, U"\U000001F8\U000001F9\U000001F8" },
{ 0x000001F9, U"\U000001F8\U000001F9\U000001F8" },
{ 0x000001FA, U"\U000001FA\U000001FB\U000001FA" },
{ 0x000001FB, U"\U000001FA\U000001FB\U000001FA" },
{ 0x000001FC, U"\U000001FC\U000001FD\U000001FC" },
{ 0x000001FD, U"\U000001FC\U000001FD\U000001FC" },
{ 0x000001FE, U"\U000001FE\U000001FF\U000001FE" },
{ 0x000001FF, U"\U000001FE\U000001FF\U000001FE" },
{ 0x00000200, U"\U00000200\U00000201\U00000200" },
{ 0x00000201, U"\U00000200\U00000201\U00000200" },
{ 0x00000202, U"\U00000202\U00000203\U00000202" },
{ 0x00000203, U"\U00000202\U00000203\U00000202" },
{ 0x00000204, U"\U00000204\U00000205\U00000204" },
{ 0x00000205, U"\U00000204\U00000205\U00000204" },
{ 0x00000206, U"\U00000206\U00000207\U00000206" },
{ 0x00000207, U"\U00000206\U00000207\U00000206" },
{ 0x00000208, U"\U00000208\U00000209\U00000208" },
{ 0x00000209, U"\U00000208\U00000209\U00000208" },
{ 0x0000020A, U"\U0000020A\U0000020B\U0000020A" },
{ 0x0000020B, U"\U0000020A\U0000020B\U0000020A" },
{ 0x0000020C, U"\U0000020C\U0000020D\U0000020C" },
{ 0x0000020D, U"\U0000020C\U0000020D\U0000020C" },
{ 0x0000020E, U"\U0000020E\U0000020F\U0000020E" },
{ 0x0000020F, U"\U0000020E\U0000020F\U0000020E" },
{ 0x00000210, U"\U00000210\U00000211\U00000210" },
{ 0x00000211, U"\U00000210\U00000211\U00000210" },
{ 0x00000212, U"\U00000212\U00000213\U00000212" },
{ 0x00000213, U"\U00000212\U00000213\U00000212" },
{ 0x00000214, U"\U00000214\U00000215\U00000214" },
{ 0x00000215, U"\U00000214\U00000215\U00000214" },
{ 0x00000216, U"\U00000216\U00000217\U00000216" },
{ 0x00000217, U"\U00000216\U00000217\U00000216" },
{ 0x00000218, U"\U00000218\U00000219\U00000218" },
{ 0x00000219, U"\U00000218\U00000219\U00000218" },
{ 0x0000021A, U"\U0000021A\U0000021B\U0000021A" },
{ 0x0000021B, U"\U0000021A\U0000021B\U0000021A" },
{ 0x0000021C, U"\U0000021C\U0000021D\U0000021C" },
{ 0x0000021D, U"\U0000021C\U0000021D\U0000021C" },
{ 0x0000021E, U"\U0000021E\U0000021F\U0000021E" },
{ 0x0000021F, U"\U0000021E\U0000021F\U0000021E" },
{ 0x00000220, U"\U00000220\U0000019E\U00000220" },
{ 0x00000222, U"\U00000222\U00000223\U00000222" },
{ 0x00000223, U"\U00000222\U00000223\U00000222" },
{ 0x00000224, U"\U00000224\U00000225\U00000224" },
{ 0x00000225, U"\U00000224\U00000225\U00000224" },
{ 0x00000226, U"\U00000226\U00000227\U00000226" },
{ 0x00000227, U"\U00000226\U00000227\U00000226" },
{ 0x00000228, U"\U00000228\U00000229\U00000228" },
{ 0x00000229, U"\U00000228\U00000229\U00000228" },
{ 0x0000022A, U"\U0000022A\U0000022B\U0000022A" },
{ 0x0000022B, U"\U0000022A\U0000022B\U0000022A" },
{ 0x0000022C, U"\U0000022C\U0000022D\U0000022C" },
{ 0x0000022D, U"\U0000022C\U0000022D\U0000022C" },
{ 0x0000022E, U"\U0000022E\U0000022F\U0000022E" },
{ 0x0000022F, U"\U0000022E\U0000022F\U0000022E" },
{ 0x00000230, U"\U00000230\U00000231\U00000230" },
{ 0x00000231, U"\U00000230\U00000231\U00000230" },
{ 0x00000232, U"\U00000232\U00000233\U00000232" },
{ 0x00000233, U"\U00000232\U00000233\U00000232" },
{ 0x0000023A, U"\U0000023A\U00002C65\U0000023A" },
{ 0x0000023B, U"\U0000023B\U0000023C\U0000023B" },
{ 0x0000023C, U"\U0000023B\U0000023C\U0000023B" },
{ 0x0000023D, U"\U0000023D\U0000019A\U0000023D" },
{ 0x0000023E, U"\U0000023E\U00002C66\U0000023E" },
{ 0x0000023F, U"\U00002C7E\U0000023F\U00002C7E" },
{ 0x00000240, U"\U00002C7F\U00000240\U00002C7F" },
{ 0x00000241, U"\U00000241\U00000242\U00000241" },
{ 0x00000242, U"\U00000241\U00000242\U00000241" },
{ 0x00000243, U"\U00000243\U00000180\U00000243" },
{ 0x00000244, U"\U00000244\U00000289\U00000244" },
{ 0x00000245, U"\U00000245\U0000028C\U00000245" },
{ 0x00000246, U"\U00000246\U00000247\U00000246" },
{ 0x00000247, U"\U00000246\U00000247\U00000246" },
{ 0x00000248, U"\U00000248\U00000249\U00000248" },
{ 0x00000249, U"\U00000248\U00000249\U00000248" },
{ 0x0000024A, U"\U0000024A\U0000024B\U0000024A" },
{ 0x0000024B, U"\U0000024A\U0000024B\U0000024A" },
{ 0x0000024C, U"\U0000024C\U0000024D\U0000024C" },
{ 0x0000024D, U"\U0000024C\U0000024D\U0000024C" },
{ 0x0000024E, U"\U0000024E\U0000024F\U0000024E" },
{ 0x0000024F, U"\U0000024E\U0000024F\U0000024E" },
{ 0x00000250, U"\U00002C6F\U00000250\U00002C6F" },
{ 0x00000251, U"\U00002C6D\U00000251\U00002C6D" },
{ 0x00000252, U"\U00002C70\U00000252\U00002C70" },
{ 0x00000253, U"\U00000181\U00000253\U00000181" },
{ 0x00000254, U"\U00000186\U00000254\U00000186" },
{ 0x00000256, U"\U00000189\U00000256\U00000189" },
{ 0x00000257, U"\U0000018A\U00000257\U0000018A" },
{ 0x00000259, U"\U0000018F\U00000259\U0000018F" },
{ 0x0000025B, U"\U00000190\U0000025B\U00000190" },
{ 0x0000025C, U"\U0000A7AB\U0000025C\U0000A7AB" },
{ 0x00000260, U"\U00000193\U00000260\U00000193" },
{ 0x00000261, U"\U0000A7AC\U00000261\U0000A7AC" },
{ 0x00000263, U"\U00000194\U00000263\U00000194" },
{ 0x00000265, U"\U0000A78D\U00000265\U0000A78D" },
{ 0x00000266, U"\U0000A7AA\U00000266\U0000A7AA" },
{ 0x00000268, U"\U00000197\U00000268\U00000197" },
{ 0x00000269, U"\U00000196\U00000269\U00000196" },
{ 0x0000026A, U"\U0000A7AE\U0000026A\U0000A7AE" },
{ 0x0000026B, U"\U00002C62\U0000026B\U00002C62" },
{ 0x0000026C, U"\U0000A7AD\U0000026C\U0000A7AD" },
{ 0x0000026F, U"\U0000019C\U0000026F\U0000019C" },
{ 0x00000271, U"\U00002C6E\U00000271\U00002C6E" },
{ 0x00000272, U"\U0000019D\U00000272\U0000019D" },
{ 0x00000275, U"\U0000019F\U00000275\U0000019F" },
{ 0x0000027D, U"\U00002C64\U0000027D\U00002C64" },
{ 0x00000280, U"\U000001A6\U00000280\U000001A6" },
{ 0x00000282, U"\U0000A7C5\U00000282\U0000A7C5" },
{ 0x00000283, U"\U000001A9\U00000283\U000001A9" },
{ 0x00000287, U"\U0000A7B1\U00000287\U0000A7B1" },
{ 0x00000288, U"\U000001AE\U00000288\U000001AE" },
{ 0x00000289, U"\U00000244\U00000289\U00000244" },
{ 0x0000028A, U"\U000001B1\U0000028A\U000001B1" },
{ 0x0000028B, U"\U000001B2\U0000028B\U000001B2" },
{ 0x0000028C, U"\U00000245\U0000028C\U00000245" },
{ 0x00000292, U"\U000001B7\U00000292\U000001B7" },
{ 0x0000029D, U"\U0000A7B2\U0000029D\U0000A7B2" },
{ 0x0000029E, U"\U0000A7B0\U0000029E\U0000A7B0" },
{ 0x00000345, U"\U00000399\U00000345\U00000399" },
{ 0x00000370, U"\U00000370\U00000371\U00000370" },
{ 0x00000371, U"\U00000370\U00000371\U00000370" },
{ 0x00000372, U"\U00000372\U00000373\U00000372" },
{ 0x00000373, U"\U00000372\U00000373\U00000372" },
{ 0x00000376, U"\U00000376\U00000377\U00000376" },
{ 0x00000377, U"\U00000376\U00000377\U00000376" },
{ 0x0000037B, U"\U000003FD\U0000037B\U000003FD" },
{ 0x0000037C, U"\U000003FE\U0000037C\U000003FE" },
{ 0x0000037D, U"\U000003FF\U0000037D\U000003FF" },
{ 0x0000037F, U"\U0000037F\U000003F3\U0000037F" },
{ 0x00000386, U"\U00000386\U000003AC\U00000386" },
{ 0x00000388, U"\U00000388\U000003AD\U00000388" },
{ 0x00000389, U"\U00000389\U000003AE\U00000389" },
{ 0x0000038A, U"\U0000038A\U000003AF\U0000038A" },
{ 0x0000038C, U"\U0000038C\U000003CC\U0000038C" },
{ 0x0000038E, U"\U0000038E\U000003CD\U0000038E" },
{ 0x0000038F, U"\U0000038F\U000003CE\U0000038F" },
{ 0x00000391, U"\U00000391\U000003B1\U00000391" },
{ 0x00000392, U"\U00000392\U000003B2\U00000392" },
{ 0x00000393, U"\U00000393\U000003B3\U00000393" },
{ 0x00000394, U"\U00000394\U000003B4\U00000394" },
{ 0x00000395, U"\U00000395\U000003B5\U00000395" },
{ 0x00000396, U"\U00000396\U000003B6\U00000396" },
{ 0x00000397, U"\U00000397\U000003B7\U00000397" },
{ 0x00000398, U"\U00000398\U000003B8\U00000398" },
{ 0x00000399, U"\U00000399\U000003B9\U00000399" },
{ 0x0000039A, U"\U0000039A\U000003BA\U0000039A" },
{ 0x0000039B, U"\U0000039B\U000003BB\U0000039B" },
{ 0x0000039C, U"\U0000039C\U000003BC\U0000039C" },
{ 0x0000039D, U"\U0000039D\U000003BD\U0000039D" },
{ 0x0000039E, U"\U0000039E\U000003BE\U0000039E" },
{ 0x0000039F, U"\U0000039F\U000003BF\U0000039F" },
{ 0x000003A0, U"\U000003A0\U000003C0\U000003A0" },
{ 0x000003A1, U"\U000003A1\U000003C1\U000003A1" },
{ 0x000003A3, U"\U000003A3\U000003C3\U000003A3" },
{ 0x000003A4, U"\U000003A4\U000003C4\U000003A4" },
{ 0x000003A5, U"\U000003A5\U000003C5\U000003A5" },
{ 0x000003A6, U"\U000003A6\U000003C6\U000003A6" },
{ 0x000003A7, U"\U000003A7\U000003C7\U000003A7" },
{ 0x000003A8, U"\U000003A8\U000003C8\U000003A8" },
{ 0x000003A9, U"\U000003A9\U000003C9\U000003A9" },
{ 0x000003AA, U"\U000003AA\U000003CA\U000003AA" },
{ 0x000003AB, U"\U000003AB\U000003CB\U000003AB" },
{ 0x000003AC, U"\U00000386\U000003AC\U00000386" },
{ 0x000003AD, U"\U00000388\U000003AD\U00000388" },
{ 0x000003AE, U"\U00000389\U000003AE\U00000389" },
{ 0x000003AF, U"\U0000038A\U000003AF\U0000038A" },
{ 0x000003B1, U"\U00000391\U000003B1\U00000391" },
{ 0x000003B2, U"\U00000392\U000003B2\U00000392" },
{ 0x000003B3, U"\U00000393\U000003B3\U00000393" },
{ 0x000003B4, U"\U00000394\U000003B4\U00000394" },
{ 0x000003B5, U"\U00000395\U000003B5\U00000395" },
{ 0x000003B6, U"\U00000396\U000003B6\U00000396" },
{ 0x000003B7, U"\U00000397\U000003B7\U00000397" },
{ 0x000003B8, U"\U00000398\U000003B8\U00000398" },
{ 0x000003B9, U"\U00000399\U000003B9\U00000399" },
{ 0x000003BA, U"\U0000039A\U000003BA\U0000039A" },
{ 0x000003BB, U"\U0000039B\U000003BB\U0000039B" },
{ 0x000003BC, U"\U0000039C\U000003BC\U0000039C" },
{ 0x000003BD, U"\U0000039D\U000003BD\U0000039D" },
{ 0x000003BE, U"\U0000039E\U000003BE\U0000039E" },
{ 0x000003BF, U"\U0000039F\U000003BF\U0000039F" },
{ 0x000003C0, U"\U000003A0\U000003C0\U000003A0" },
{ 0x000003C1, U"\U000003A1\U000003C1\U000003A1" },
{ 0x000003C2, U"\U000003A3\U000003C2\U000003A3" },
{ 0x000003C3, U"\U000003A3\U000003C3\U000003A3" },
{ 0x000003C4, U"\U000003A4\U000003C4\U000003A4" },
{ 0x000003C5, U"\U000003A5\U000003C5\U000003A5" },
{ 0x000003C6, U"\U000003A6\U000003C6\U000003A6" },
{ 0x000003C7, U"\U000003A7\U000003C7\U000003A7" },
{ 0x000003C8, U"\U000003A8\U000003C8\U000003A8" },
{ 0x000003C9, U"\U000003A9\U000003C9\U000003A9" },
{ 0x000003CA, U"\U000003AA\U000003CA\U000003AA" },
{ 0x000003CB, U"\U000003AB\U000003CB\U000003AB" },
{ 0x000003CC, U"\U0000038C\U000003CC\U0000038C" },
{ 0x000003CD, U"\U0000038E\U000003CD\U0000038E" },
{ 0x000003CE, U"\U0000038F\U000003CE\U0000038F" },
{ 0x000003CF, U"\U000003CF\U000003D7\U000003CF" },
{ 0x000003D0, U"\U00000392\U000003D0\U00000392" },
{ 0x000003D1, U"\U00000398\U000003D1\U00000398" },
{ 0x000003D5, U"\U000003A6\U000003D5\U000003A6" },
{ 0x000003D6, U"\U000003A0\U000003D6\U000003A0" },
{ 0x000003D7, U"\U000003CF\U000003D7\U000003CF" },
{ 0x000003D8, U"\U000003D8\U000003D9\U000003D8" },
{ 0x000003D9, U"\U000003D8\U000003D9\U000003D8" },
{ 0x000003DA, U"\U000003DA\U000003DB\U000003DA" },
{ 0x000003DB, U"\U000003DA\U000003DB\U000003DA" },
{ 0x000003DC, U"\U000003DC\U000003DD\U000003DC" },
{ 0x000003DD, U"\U000003DC\U000003DD\U000003DC" },
{ 0x000003DE, U"\U000003DE\U000003DF\U000003DE" },
{ 0x000003DF, U"\U000003DE\U000003DF\U000003DE" },
{ 0x000003E0, U"\U000003E0\U000003E1\U000003E0" },
{ 0x000003E1, U"\U000003E0\U000003E1\U000003E0" },
{ 0x000003E2, U"\U000003E2\U000003E3\U000003E2" },
{ 0x000003E3, U"\U000003E2\U000003E3\U000003E2" },
{ 0x000003E4, U"\U000003E4\U000003E5\U000003E4" },
{ 0x000003E5, U"\U000003E4\U000003E5\U000003E4" },
{ 0x000003E6, U"\U000003E6\U000003E7\U000003E6" },
{ 0x000003E7, U"\U000003E6\U000003E7\U000003E6" },
{ 0x000003E8, U"\U000003E8\U000003E9\U000003E8" },
{ 0x000003E9, U"\U000003E8\U000003E9\U000003E8" },
{ 0x000003EA, U"\U000003EA\U000003EB\U000003EA" },
{ 0x000003EB, U"\U000003EA\U000003EB\U000003EA" },
{ 0x000003EC, U"\U000003EC\U000003ED\U000003EC" },
{ 0x000003ED, U"\U000003EC\U000003ED\U000003EC" },
{ 0x000003EE, U"\U000003EE\U000003EF\U000003EE" },
{ 0x000003EF, U"\U000003EE\U000003EF\U000003EE" },
{ 0x000003F0, U"\U0000039A\U000003F0\U0000039A" },
{ 0x000003F1, U"\U000003A1\U000003F1\U000003A1" },
{ 0x000003F2, U"\U000003F9\U000003F2\U000003F9" },
{ 0x000003F3, U"\U0000037F\U000003F3\U0000037F" },
{ 0x000003F4, U"\U000003F4\U000003B8\U000003F4" },
{ 0x000003F5, U"\U00000395\U000003F5\U00000395" },
{ 0x000003F7, U"\U000003F7\U000003F8\U000003F7" },
{ 0x000003F8, U"\U000003F7\U000003F8\U000003F7" },
{ 0x000003F9, U"\U000003F9\U000003F2\U000003F9" },
{ 0x000003FA, U"\U000003FA\U000003FB\U000003FA" },
{ 0x000003FB, U"\U000003FA\U000003FB\U000003FA" },
{ 0x000003FD, U"\U000003FD\U0000037B\U000003FD" },
{ 0x000003FE, U"\U000003FE\U0000037C\U000003FE" },
{ 0x000003FF, U"\U000003FF\U0000037D\U000003FF" },
{ 0x00000400, U"\U00000400\U00000450\U00000400" },
{ 0x00000401, U"\U00000401\U00000451\U00000401" },
{ 0x00000402, U"\U00000402\U00000452\U00000402" },
{ 0x00000403, U"\U00000403\U00000453\U00000403" },
{ 0x00000404, U"\U00000404\U00000454\U00000404" },
{ 0x00000405, U"\U00000405\U00000455\U00000405" },
{ 0x00000406, U"\U00000406\U00000456\U00000406" },
{ 0x00000407, U"\U00000407\U00000457\U00000407" },
{ 0x00000408, U"\U00000408\U00000458\U00000408" },
{ 0x00000409, U"\U00000409\U00000459\U00000409" },
{ 0x0000040A, U"\U0000040A\U0000045A\U0000040A" },
{ 0x0000040B, U"\U0000040B\U0000045B\U0000040B" },
{ 0x0000040C, U"\U0000040C\U0000045C\U0000040C" },
{ 0x0000040D, U"\U0000040D\U0000045D\U0000040D" },
{ 0x0000040E, U"\U0000040E\U0000045E\U0000040E" },
{ 0x0000040F, U"\U0000040F\U0000045F\U0000040F" },
{ 0x00000410, U"\U00000410\U00000430\U00000410" },
{ 0x00000411, U"\U00000411\U00000431\U00000411" },
{ 0x00000412, U"\U00000412\U00000432\U00000412" },
{ 0x00000413, U"\U00000413\U00000433\U00000413" },
{ 0x00000414, U"\U00000414\U00000434\U00000414" },
{ 0x00000415, U"\U00000415\U00000435\U00000415" },
{ 0x00000416, U"\U00000416\U00000436\U00000416" },
{ 0x00000417, U"\U00000417\U00000437\U00000417" },
{ 0x00000418, U"\U00000418\U00000438\U00000418" },
{ 0x00000419, U"\U00000419\U00000439\U00000419" },
{ 0x0000041A, U"\U0000041A\U0000043A\U0000041A" },
{ 0x0000041B, U"\U0000041B\U0000043B\U0000041B" },
{ 0x0000041C, U"\U0000041C\U0000043C\U0000041C" },
{ 0x0000041D, U"\U0000041D\U0000043D\U0000041D" },
{ 0x0000041E, U"\U0000041E\U0000043E\U0000041E" },
{ 0x0000041F, U"\U0000041F\U0000043F\U0000041F" },
{ 0x00000420, U"\U00000420\U00000440\U00000420" },
{ 0x00000421, U"\U00000421\U00000441\U00000421" },
{ 0x00000422, U"\U00000422\U00000442\U00000422" },
{ 0x00000423, U"\U00000423\U00000443\U00000423" },
{ 0x00000424, U"\U00000424\U00000444\U00000424" },
{ 0x00000425, U"\U00000425\U00000445\U00000425" },
{ 0x00000426, U"\U00000426\U00000446\U00000426" },
{ 0x00000427, U"\U00000427\U00000447\U00000427" },
{ 0x00000428, U"\U00000428\U00000448\U00000428" },
{ 0x00000429, U"\U00000429\U00000449\U00000429" },
{ 0x0000042A, U"\U0000042A\U0000044A\U0000042A" },
{ 0x0000042B, U"\U0000042B\U0000044B\U0000042B" },
{ 0x0000042C, U"\U0000042C\U0000044C\U0000042C" },
{ 0x0000042D, U"\U0000042D\U0000044D\U0000042D" },
{ 0x0000042E, U"\U0000042E\U0000044E\U0000042E" },
{ 0x0000042F, U"\U0000042F\U0000044F\U0000042F" },
{ 0x00000430, U"\U00000410\U00000430\U00000410" },
{ 0x00000431, U"\U00000411\U00000431\U00000411" },
{ 0x00000432, U"\U00000412\U00000432\U00000412" },
{ 0x00000433, U"\U00000413\U00000433\U00000413" },
{ 0x00000434, U"\U00000414\U00000434\U00000414" },
{ 0x00000435, U"\U00000415\U00000435\U00000415" },
{ 0x00000436, U"\U00000416\U00000436\U00000416" },
{ 0x00000437, U"\U00000417\U00000437\U00000417" },
{ 0x00000438, U"\U00000418\U00000438\U00000418" },
{ 0x00000439, U"\U00000419\U00000439\U00000419" },
{ 0x0000043A, U"\U0000041A\U0000043A\U0000041A" },
{ 0x0000043B, U"\U0000041B\U0000043B\U0000041B" },
{ 0x0000043C, U"\U0000041C\U0000043C\U0000041C" },
{ 0x0000043D, U"\U0000041D\U0000043D\U0000041D" },
{ 0x0000043E, U"\U0000041E\U0000043E\U0000041E" },
{ 0x0000043F, U"\U0000041F\U0000043F\U0000041F" },
{ 0x00000440, U"\U00000420\U00000440\U00000420" },
{ 0x00000441, U"\U00000421\U00000441\U00000421" },
{ 0x00000442, U"\U00000422\U00000442\U00000422" },
{ 0x00000443, U"\U00000423\U00000443\U00000423" },
{ 0x00000444, U"\U00000424\U00000444\U00000424" },
{ 0x00000445, U"\U00000425\U00000445\U00000425" },
{ 0x00000446, U"\U00000426\U00000446\U00000426" },
{ 0x00000447, U"\U00000427\U00000447\U00000427" },
{ 0x00000448, U"\U00000428\U00000448\U00000428" },
{ 0x00000449, U"\U00000429\U00000449\U00000429" },
{ 0x0000044A, U"\U0000042A\U0000044A\U0000042A" },
{ 0x0000044B, U"\U0000042B\U0000044B\U0000042B" },
{ 0x0000044C, U"\U0000042C\U0000044C\U0000042C" },
{ 0x0000044D, U"\U0000042D\U0000044D\U0000042D" },
{ 0x0000044E, U"\U0000042E\U0000044E\U0000042E" },
{ 0x0000044F, U"\U0000042F\U0000044F\U0000042F" },
{ 0x00000450, U"\U00000400\U00000450\U00000400" },
{ 0x00000451, U"\U00000401\U00000451\U00000401" },
{ 0x00000452, U"\U00000402\U00000452\U00000402" },
{ 0x00000453, U"\U00000403\U00000453\U00000403" },
{ 0x00000454, U"\U00000404\U00000454\U00000404" },
{ 0x00000455, U"\U00000405\U00000455\U00000405" },
{ 0x00000456, U"\U00000406\U00000456\U00000406" },
{ 0x00000457, U"\U00000407\U00000457\U00000407" },
{ 0x00000458, U"\U00000408\U00000458\U00000408" },
{ 0x00000459, U"\U00000409\U00000459\U00000409" },
{ 0x0000045A, U"\U0000040A\U0000045A\U0000040A" },
{ 0x0000045B, U"\U0000040B\U0000045B\U0000040B" },
{ 0x0000045C, U"\U0000040C\U0000045C\U0000040C" },
{ 0x0000045D, U"\U0000040D\U0000045D\U0000040D" },
{ 0x0000045E, U"\U0000040E\U0000045E\U0000040E" },
{ 0x0000045F, U"\U0000040F\U0000045F\U0000040F" },
{ 0x00000460, U"\U00000460\U00000461\U00000460" },
{ 0x00000461, U"\U00000460\U00000461\U00000460" },
{ 0x00000462, U"\U00000462\U00000463\U00000462" },
{ 0x00000463, U"\U00000462\U00000463\U00000462" },
{ 0x00000464, U"\U00000464\U00000465\U00000464" },
{ 0x00000465, U"\U00000464\U00000465\U00000464" },
{ 0x00000466, U"\U00000466\U00000467\U00000466" },
{ 0x00000467, U"\U00000466\U00000467\U00000466" },
{ 0x00000468, U"\U00000468\U00000469\U00000468" },
{ 0x00000469, U"\U00000468\U00000469\U00000468" },
{ 0x0000046A, U"\U0000046A\U0000046B\U0000046A" },
{ 0x0000046B, U"\U0000046A\U0000046B\U0000046A" },
{ 0x0000046C, U"\U0000046C\U0000046D\U0000046C" },
{ 0x0000046D, U"\U0000046C\U0000046D\U0000046C" },
{ 0x0000046E, U"\U0000046E\U0000046F\U0000046E" },
{ 0x0000046F, U"\U0000046E\U0000046F\U0000046E" },
{ 0x00000470, U"\U00000470\U00000471\U00000470" },
{ 0x00000471, U"\U00000470\U00000471\U00000470" },
{ 0x00000472, U"\U00000472\U00000473\U00000472" },
{ 0x00000473, U"\U00000472\U00000473\U00000472" },
{ 0x00000474, U"\U00000474\U00000475\U00000474" },
{ 0x00000475, U"\U00000474\U00000475\U00000474" },
{ 0x00000476, U"\U00000476\U00000477\U00000476" },
{ 0x00000477, U"\U00000476\U00000477\U00000476" },
{ 0x00000478, U"\U00000478\U00000479\U00000478" },
{ 0x00000479, U"\U00000478\U00000479\U00000478" },
{ 0x0000047A, U"\U0000047A\U0000047B\U0000047A" },
{ 0x0000047B, U"\U0000047A\U0000047B\U0000047A" },
{ 0x0000047C, U"\U0000047C\U0000047D\U0000047C" },
{ 0x0000047D, U"\U0000047C\U0000047D\U0000047C" },
{ 0x0000047E, U"\U0000047E\U0000047F\U0000047E" },
{ 0x0000047F, U"\U0000047E\U0000047F\U0000047E" },
{ 0x00000480, U"\U00000480\U00000481\U00000480" },
{ 0x00000481, U"\U00000480\U00000481\U00000480" },
{ 0x0000048A, U"\U0000048A\U0000048B\U0000048A" },
{ 0x0000048B, U"\U0000048A\U0000048B\U0000048A" },
{ 0x0000048C, U"\U0000048C\U0000048D\U0000048C" },
{ 0x0000048D, U"\U0000048C\U0000048D\U0000048C" },
{ 0x0000048E, U"\U0000048E\U0000048F\U0000048E" },
{ 0x0000048F, U"\U0000048E\U0000048F\U0000048E" },
{ 0x00000490, U"\U00000490\U00000491\U00000490" },
{ 0x00000491, U"\U00000490\U00000491\U00000490" },
{ 0x00000492, U"\U00000492\U00000493\U00000492" },
{ 0x00000493, U"\U00000492\U00000493\U00000492" },
{ 0x00000494, U"\U00000494\U00000495\U00000494" },
{ 0x00000495, U"\U00000494\U00000495\U00000494" },
{ 0x00000496, U"\U00000496\U00000497\U00000496" },
{ 0x00000497, U"\U00000496\U00000497\U00000496" },
{ 0x00000498, U"\U00000498\U00000499\U00000498" },
{ 0x00000499, U"\U00000498\U00000499\U00000498" },
{ 0x0000049A, U"\U0000049A\U0000049B\U0000049A" },
{ 0x0000049B, U"\U0000049A\U0000049B\U0000049A" },
{ 0x0000049C, U"\U0000049C\U0000049D\U0000049C" },
{ 0x0000049D, U"\U0000049C\U0000049D\U0000049C" },
{ 0x0000049E, U"\U0000049E\U0000049F\U0000049E" },
{ 0x0000049F, U"\U0000049E\U0000049F\U0000049E" },
{ 0x000004A0, U"\U000004A0\U000004A1\U000004A0" },
{ 0x000004A1, U"\U000004A0\U000004A1\U000004A0" },
{ 0x000004A2, U"\U000004A2\U000004A3\U000004A2" },
{ 0x000004A3, U"\U000004A2\U000004A3\U000004A2" },
{ 0x000004A4, U"\U000004A4\U000004A5\U000004A4" },
{ 0x000004A5, U"\U000004A4\U000004A5\U000004A4" },
{ 0x000004A6, U"\U000004A6\U000004A7\U000004A6" },
{ 0x000004A7, U"\U000004A6\U000004A7\U000004A6" },
{ 0x000004A8, U"\U000004A8\U000004A9\U000004A8" },
{ 0x000004A9, U"\U000004A8\U000004A9\U000004A8" },
{ 0x000004AA, U"\U000004AA\U000004AB\U000004AA" },
{ 0x000004AB, U"\U000004AA\U000004AB\U000004AA" },
{ 0x000004AC, U"\U000004AC\U000004AD\U000004AC" },
{ 0x000004AD, U"\U000004AC\U000004AD\U000004AC" },
{ 0x000004AE, U"\U000004AE\U000004AF\U000004AE" },
{ 0x000004AF, U"\U000004AE\U000004AF\U000004AE" },
{ 0x000004B0, U"\U000004B0\U000004B1\U000004B0" },
{ 0x000004B1, U"\U000004B0\U000004B1\U000004B0" },
{ 0x000004B2, U"\U000004B2\U000004B3\U000004B2" },
{ 0x000004B3, U"\U000004B2\U000004B3\U000004B2" },
{ 0x000004B4, U"\U000004B4\U000004B5\U000004B4" },
{ 0x000004B5, U"\U000004B4\U000004B5\U000004B4" },
{ 0x000004B6, U"\U000004B6\U000004B7\U000004B6" },
{ 0x000004B7, U"\U000004B6\U000004B7\U000004B6" },
{ 0x000004B8, U"\U000004B8\U000004B9\U000004B8" },
{ 0x000004B9, U"\U000004B8\U000004B9\U000004B8" },
{ 0x000004BA, U"\U000004BA\U000004BB\U000004BA" },
{ 0x000004BB, U"\U000004BA\U000004BB\U000004BA" },
{ 0x000004BC, U"\U000004BC\U000004BD\U000004BC" },
{ 0x000004BD, U"\U000004BC\U000004BD\U000004BC" },
{ 0x000004BE, U"\U000004BE\U000004BF\U000004BE" },
{ 0x000004BF, U"\U000004BE\U000004BF\U000004BE" },
{ 0x000004C0, U"\U000004C0\U000004CF\U000004C0" },
{ 0x000004C1, U"\U000004C1\U000004C2\U000004C1" },
{ 0x000004C2, U"\U000004C1\U000004C2\U000004C1" },
{ 0x000004C3, U"\U000004C3\U000004C4\U000004C3" },
{ 0x000004C4, U"\U000004C3\U000004C4\U000004C3" },
{ 0x000004C5, U"\U000004C5\U000004C6\U000004C5" },
{ 0x000004C6, U"\U000004C5\U000004C6\U000004C5" },
{ 0x000004C7, U"\U000004C7\U000004C8\U000004C7" },
{ 0x000004C8, U"\U000004C7\U000004C8\U000004C7" },
{ 0x000004C9, U"\U000004C9\U000004CA\U000004C9" },
{ 0x000004CA, U"\U000004C9\U000004CA\U000004C9" },
{ 0x000004CB, U"\U000004CB\U000004CC\U000004CB" },
{ 0x000004CC, U"\U000004CB\U000004CC\U000004CB" },
{ 0x000004CD, U"\U000004CD\U000004CE\U000004CD" },
{ 0x000004CE, U"\U000004CD\U000004CE\U000004CD" },
{ 0x000004CF, U"\U000004C0\U000004CF\U000004C0" },
{ 0x000004D0, U"\U000004D0\U000004D1\U000004D0" },
{ 0x000004D1, U"\U000004D0\U000004D1\U000004D0" },
{ 0x000004D2, U"\U000004D2\U000004D3\U000004D2" },
{ 0x000004D3, U"\U000004D2\U000004D3\U000004D2" },
{ 0x000004D4, U"\U000004D4\U000004D5\U000004D4" },
{ 0x000004D5, U"\U000004D4\U000004D5\U000004D4" },
{ 0x000004D6, U"\U000004D6\U000004D7\U000004D6" },
{ 0x000004D7, U"\U000004D6\U000004D7\U000004D6" },
{ 0x000004D8, U"\U000004D8\U000004D9\U000004D8" },
{ 0x000004D9, U"\U000004D8\U000004D9\U000004D8" },
{ 0x000004DA, U"\U000004DA\U000004DB\U000004DA" },
{ 0x000004DB, U"\U000004DA\U000004DB\U000004DA" },
{ 0x000004DC, U"\U000004DC\U000004DD\U000004DC" },
{ 0x000004DD, U"\U000004DC\U000004DD\U000004DC" },
{ 0x000004DE, U"\U000004DE\U000004DF\U000004DE" },
{ 0x000004DF, U"\U000004DE\U000004DF\U000004DE" },
{ 0x000004E0, U"\U000004E0\U000004E1\U000004E0" },
{ 0x000004E1, U"\U000004E0\U000004E1\U000004E0" },
{ 0x000004E2, U"\U000004E2\U000004E3\U000004E2" },
{ 0x000004E3, U"\U000004E2\U000004E3\U000004E2" },
{ 0x000004E4, U"\U000004E4\U000004E5\U000004E4" },
{ 0x000004E5, U"\U000004E4\U000004E5\U000004E4" },
{ 0x000004E6, U"\U000004E6\U000004E7\U000004E6" },
{ 0x000004E7, U"\U000004E6\U000004E7\U000004E6" },
{ 0x000004E8, U"\U000004E8\U000004E9\U000004E8" },
{ 0x000004E9, U"\U000004E8\U000004E9\U000004E8" },
{ 0x000004EA, U"\U000004EA\U000004EB\U000004EA" },
{ 0x000004EB, U"\U000004EA\U000004EB\U000004EA" },
{ 0x000004EC, U"\U000004EC\U000004ED\U000004EC" },
{ 0x000004ED, U"\U000004EC\U000004ED\U000004EC" },
{ 0x000004EE, U"\U000004EE\U000004EF\U000004EE" },
{ 0x000004EF, U"\U000004EE\U000004EF\U000004EE" },
{ 0x000004F0, U"\U000004F0\U000004F1\U000004F0" },
{ 0x000004F1, U"\U000004F0\U000004F1\U000004F0" },
{ 0x000004F2, U"\U000004F2\U000004F3\U000004F2" },
{ 0x000004F3, U"\U000004F2\U000004F3\U000004F2" },
{ 0x000004F4, U"\U000004F4\U000004F5\U000004F4" },
{ 0x000004F5, U"\U000004F4\U000004F5\U000004F4" },
{ 0x000004F6, U"\U000004F6\U000004F7\U000004F6" },
{ 0x000004F7, U"\U000004F6\U000004F7\U000004F6" },
{ 0x000004F8, U"\U000004F8\U000004F9\U000004F8" },
{ 0x000004F9, U"\U000004F8\U000004F9\U000004F8" },
{ 0x000004FA, U"\U000004FA\U000004FB\U000004FA" },
{ 0x000004FB, U"\U000004FA\U000004FB\U000004FA" },
{ 0x000004FC, U"\U000004FC\U000004FD\U000004FC" },
{ 0x000004FD, U"\U000004FC\U000004FD\U000004FC" },
{ 0x000004FE, U"\U000004FE\U000004FF\U000004FE" },
{ 0x000004FF, U"\U000004FE\U000004FF\U000004FE" },
{ 0x00000500, U"\U00000500\U00000501\U00000500" },
{ 0x00000501, U"\U00000500\U00000501\U00000500" },
{ 0x00000502, U"\U00000502\U00000503\U00000502" },
{ 0x00000503, U"\U00000502\U00000503\U00000502" },
{ 0x00000504, U"\U00000504\U00000505\U00000504" },
{ 0x00000505, U"\U00000504\U00000505\U00000504" },
{ 0x00000506, U"\U00000506\U00000507\U00000506" },
{ 0x00000507, U"\U00000506\U00000507\U00000506" },
{ 0x00000508, U"\U00000508\U00000509\U00000508" },
{ 0x00000509, U"\U00000508\U00000509\U00000508" },
{ 0x0000050A, U"\U0000050A\U0000050B\U0000050A" },
{ 0x0000050B, U"\U0000050A\U0000050B\U0000050A" },
{ 0x0000050C, U"\U0000050C\U0000050D\U0000050C" },
{ 0x0000050D, U"\U0000050C\U0000050D\U0000050C" },
{ 0x0000050E, U"\U0000050E\U0000050F\U0000050E" },
{ 0x0000050F, U"\U0000050E\U0000050F\U0000050E" },
{ 0x00000510, U"\U00000510\U00000511\U00000510" },
{ 0x00000511, U"\U00000510\U00000511\U00000510" },
{ 0x00000512, U"\U00000512\U00000513\U00000512" },
{ 0x00000513, U"\U00000512\U00000513\U00000512" },
{ 0x00000514, U"\U00000514\U00000515\U00000514" },
{ 0x00000515, U"\U00000514\U00000515\U00000514" },
{ 0x00000516, U"\U00000516\U00000517\U00000516" },
{ 0x00000517, U"\U00000516\U00000517\U00000516" },
{ 0x00000518, U"\U00000518\U00000519\U00000518" },
{ 0x00000519, U"\U00000518\U00000519\U00000518" },
{ 0x0000051A, U"\U0000051A\U0000051B\U0000051A" },
{ 0x0000051B, U"\U0000051A\U0000051B\U0000051A" },
{ 0x0000051C, U"\U0000051C\U0000051D\U0000051C" },
{ 0x0000051D, U"\U0000051C\U0000051D\U0000051C" },
{ 0x0000051E, U"\U0000051E\U0000051F\U0000051E" },
{ 0x0000051F, U"\U0000051E\U0000051F\U0000051E" },
{ 0x00000520, U"\U00000520\U00000521\U00000520" },
{ 0x00000521, U"\U00000520\U00000521\U00000520" },
{ 0x00000522, U"\U00000522\U00000523\U00000522" },
{ 0x00000523, U"\U00000522\U00000523\U00000522" },
{ 0x00000524, U"\U00000524\U00000525\U00000524" },
{ 0x00000525, U"\U00000524\U00000525\U00000524" },
{ 0x00000526, U"\U00000526\U00000527\U00000526" },
{ 0x00000527, U"\U00000526\U00000527\U00000526" },
{ 0x00000528, U"\U00000528\U00000529\U00000528" },
{ 0x00000529, U"\U00000528\U00000529\U00000528" },
{ 0x0000052A, U"\U0000052A\U0000052B\U0000052A" },
{ 0x0000052B, U"\U0000052A\U0000052B\U0000052A" },
{ 0x0000052C, U"\U0000052C\U0000052D\U0000052C" },
{ 0x0000052D, U"\U0000052C\U0000052D\U0000052C" },
{ 0x0000052E, U"\U0000052E\U0000052F\U0000052E" },
{ 0x0000052F, U"\U0000052E\U0000052F\U0000052E" },
{ 0x00000531, U"\U00000531\U00000561\U00000531" },
{ 0x00000532, U"\U00000532\U00000562\U00000532" },
{ 0x00000533, U"\U00000533\U00000563\U00000533" },
{ 0x00000534, U"\U00000534\U00000564\U00000534" },
{ 0x00000535, U"\U00000535\U00000565\U00000535" },
{ 0x00000536, U"\U00000536\U00000566\U00000536" },
{ 0x00000537, U"\U00000537\U00000567\U00000537" },
{ 0x00000538, U"\U00000538\U00000568\U00000538" },
{ 0x00000539, U"\U00000539\U00000569\U00000539" },
{ 0x0000053A, U"\U0000053A\U0000056A\U0000053A" },
{ 0x0000053B, U"\U0000053B\U0000056B\U0000053B" },
{ 0x0000053C, U"\U0000053C\U0000056C\U0000053C" },
{ 0x0000053D, U"\U0000053D\U0000056D\U0000053D" },
{ 0x0000053E, U"\U0000053E\U0000056E\U0000053E" },
{ 0x0000053F, U"\U0000053F\U0000056F\U0000053F" },
{ 0x00000540, U"\U00000540\U00000570\U00000540" },
{ 0x00000541, U"\U00000541\U00000571\U00000541" },
{ 0x00000542, U"\U00000542\U00000572\U00000542" },
{ 0x00000543, U"\U00000543\U00000573\U00000543" },
{ 0x00000544, U"\U00000544\U00000574\U00000544" },
{ 0x00000545, U"\U00000545\U00000575\U00000545" },
{ 0x00000546, U"\U00000546\U00000576\U00000546" },
{ 0x00000547, U"\U00000547\U00000577\U00000547" },
{ 0x00000548, U"\U00000548\U00000578\U00000548" },
{ 0x00000549, U"\U00000549\U00000579\U00000549" },
{ 0x0000054A, U"\U0000054A\U0000057A\U0000054A" },
{ 0x0000054B, U"\U0000054B\U0000057B\U0000054B" },
{ 0x0000054C, U"\U0000054C\U0000057C\U0000054C" },
{ 0x0000054D, U"\U0000054D\U0000057D\U0000054D" },
{ 0x0000054E, U"\U0000054E\U0000057E\U0000054E" },
{ 0x0000054F, U"\U0000054F\U0000057F\U0000054F" },
{ 0x00000550, U"\U00000550\U00000580\U00000550" },
{ 0x00000551, U"\U00000551\U00000581\U00000551" },
{ 0x00000552, U"\U00000552\U00000582\U00000552" },
{ 0x00000553, U"\U00000553\U00000583\U00000553" },
{ 0x00000554, U"\U00000554\U00000584\U00000554" },
{ 0x00000555, U"\U00000555\U00000585\U00000555" },
{ 0x00000556, U"\U00000556\U00000586\U00000556" },
{ 0x00000561, U"\U00000531\U00000561\U00000531" },
{ 0x00000562, U"\U00000532\U00000562\U00000532" },
{ 0x00000563, U"\U00000533\U00000563\U00000533" },
{ 0x00000564, U"\U00000534\U00000564\U00000534" },
{ 0x00000565, U"\U00000535\U00000565\U00000535" },
{ 0x00000566, U"\U00000536\U00000566\U00000536" },
{ 0x00000567, U"\U00000537\U00000567\U00000537" },
{ 0x00000568, U"\U00000538\U00000568\U00000538" },
{ 0x00000569, U"\U00000539\U00000569\U00000539" },
{ 0x0000056A, U"\U0000053A\U0000056A\U0000053A" },
{ 0x0000056B, U"\U0000053B\U0000056B\U0000053B" },
{ 0x0000056C, U"\U0000053C\U0000056C\U0000053C" },
{ 0x0000056D, U"\U0000053D\U0000056D\U0000053D" },
{ 0x0000056E, U"\U0000053E\U0000056E\U0000053E" },
{ 0x0000056F, U"\U0000053F\U0000056F\U0000053F" },
{ 0x00000570, U"\U00000540\U00000570\U00000540" },
{ 0x00000571, U"\U00000541\U00000571\U00000541" },
{ 0x00000572, U"\U00000542\U00000572\U00000542" },
{ 0x00000573, U"\U00000543\U00000573\U00000543" },
{ 0x00000574, U"\U00000544\U00000574\U00000544" },
{ 0x00000575, U"\U00000545\U00000575\U00000545" },
{ 0x00000576, U"\U00000546\U00000576\U00000546" },
{ 0x00000577, U"\U00000547\U00000577\U00000547" },
{ 0x00000578, U"\U00000548\U00000578\U00000548" },
{ 0x00000579, U"\U00000549\U00000579\U00000549" },
{ 0x0000057A, U"\U0000054A\U0000057A\U0000054A" },
{ 0x0000057B, U"\U0000054B\U0000057B\U0000054B" },
{ 0x0000057C, U"\U0000054C\U0000057C\U0000054C" },
{ 0x0000057D, U"\U0000054D\U0000057D\U0000054D" },
{ 0x0000057E, U"\U0000054E\U0000057E\U0000054E" },
{ 0x0000057F, U"\U0000054F\U0000057F\U0000054F" },
{ 0x00000580, U"\U00000550\U00000580\U00000550" },
{ 0x00000581, U"\U00000551\U00000581\U00000551" },
{ 0x00000582, U"\U00000552\U00000582\U00000552" },
{ 0x00000583, U"\U00000553\U00000583\U00000553" },
{ 0x00000584, U"\U00000554\U00000584\U00000554" },
{ 0x00000585, U"\U00000555\U00000585\U00000555" },
{ 0x00000586, U"\U00000556\U00000586\U00000556" },
{ 0x000010A0, U"\U000010A0\U00002D00\U000010A0" },
{ 0x000010A1, U"\U000010A1\U00002D01\U000010A1" },
{ 0x000010A2, U"\U000010A2\U00002D02\U000010A2" },
{ 0x000010A3, U"\U000010A3\U00002D03\U000010A3" },
{ 0x000010A4, U"\U000010A4\U00002D04\U000010A4" },
{ 0x000010A5, U"\U000010A5\U00002D05\U000010A5" },
{ 0x000010A6, U"\U000010A6\U00002D06\U000010A6" },
{ 0x000010A7, U"\U000010A7\U00002D07\U000010A7" },
{ 0x000010A8, U"\U000010A8\U00002D08\U000010A8" },
{ 0x000010A9, U"\U000010A9\U00002D09\U000010A9" },
{ 0x000010AA, U"\U000010AA\U00002D0A\U000010AA" },
{ 0x000010AB, U"\U000010AB\U00002D0B\U000010AB" },
{ 0x000010AC, U"\U000010AC\U00002D0C\U000010AC" },
{ 0x000010AD, U"\U000010AD\U00002D0D\U000010AD" },
{ 0x000010AE, U"\U000010AE\U00002D0E\U000010AE" },
{ 0x000010AF, U"\U000010AF\U00002D0F\U000010AF" },
{ 0x000010B0, U"\U000010B0\U00002D10\U000010B0" },
{ 0x000010B1, U"\U000010B1\U00002D11\U000010B1" },
{ 0x000010B2, U"\U000010B2\U00002D12\U000010B2" },
{ 0x000010B3, U"\U000010B3\U00002D13\U000010B3" },
{ 0x000010B4, U"\U000010B4\U00002D14\U000010B4" },
{ 0x000010B5, U"\U000010B5\U00002D15\U000010B5" },
{ 0x000010B6, U"\U000010B6\U00002D16\U000010B6" },
{ 0x000010B7, U"\U000010B7\U00002D17\U000010B7" },
{ 0x000010B8, U"\U000010B8\U00002D18\U000010B8" },
{ 0x000010B9, U"\U000010B9\U00002D19\U000010B9" },
{ 0x000010BA, U"\U000010BA\U00002D1A\U000010BA" },
{ 0x000010BB, U"\U000010BB\U00002D1B\U000010BB" },
{ 0x000010BC, U"\U000010BC\U00002D1C\U000010BC" },
{ 0x000010BD, U"\U000010BD\U00002D1D\U000010BD" },
{ 0x000010BE, U"\U000010BE\U00002D1E\U000010BE" },
{ 0x000010BF, U"\U000010BF\U00002D1F\U000010BF" },
{ 0x000010C0, U"\U000010C0\U00002D20\U000010C0" },
{ 0x000010C1, U"\U000010C1\U00002D21\U000010C1" },
{ 0x000010C2, U"\U000010C2\U00002D22\U000010C2" },
{ 0x000010C3, U"\U000010C3\U00002D23\U000010C3" },
{ 0x000010C4, U"\U000010C4\U00002D24\U000010C4" },
{ 0x000010C5, U"\U000010C5\U00002D25\U000010C5" },
{ 0x000010C7, U"\U000010C7\U00002D27\U000010C7" },
{ 0x000010CD, U"\U000010CD\U00002D2D\U000010CD" },
{ 0x000010D0, U"\U00001C90\U000010D0\U000010D0" },
{ 0x000010D1, U"\U00001C91\U000010D1\U000010D1" },
{ 0x000010D2, U"\U00001C92\U000010D2\U000010D2" },
{ 0x000010D3, U"\U00001C93\U000010D3\U000010D3" },
{ 0x000010D4, U"\U00001C94\U000010D4\U000010D4" },
{ 0x000010D5, U"\U00001C95\U000010D5\U000010D5" },
{ 0x000010D6, U"\U00001C96\U000010D6\U000010D6" },
{ 0x000010D7, U"\U00001C97\U000010D7\U000010D7" },
{ 0x000010D8, U"\U00001C98\U000010D8\U000010D8" },
{ 0x000010D9, U"\U00001C99\U000010D9\U000010D9" },
{ 0x000010DA, U"\U00001C9A\U000010DA\U000010DA" },
{ 0x000010DB, U"\U00001C9B\U000010DB\U000010DB" },
{ 0x000010DC, U"\U00001C9C\U000010DC\U000010DC" },
{ 0x000010DD, U"\U00001C9D\U000010DD\U000010DD" },
{ 0x000010DE, U"\U00001C9E\U000010DE\U000010DE" },
{ 0x000010DF, U"\U00001C9F\U000010DF\U000010DF" },
{ 0x000010E0, U"\U00001CA0\U000010E0\U000010E0" },
{ 0x000010E1, U"\U00001CA1\U000010E1\U000010E1" },
{ 0x000010E2, U"\U00001CA2\U000010E2\U000010E2" },
{ 0x000010E3, U"\U00001CA3\U000010E3\U000010E3" },
{ 0x000010E4, U"\U00001CA4\U000010E4\U000010E4" },
{ 0x000010E5, U"\U00001CA5\U000010E5\U000010E5" },
{ 0x000010E6, U"\U00001CA6\U000010E6\U000010E6" },
{ 0x000010E7, U"\U00001CA7\U000010E7\U000010E7" },
{ 0x000010E8, U"\U00001CA8\U000010E8\U000010E8" },
{ 0x000010E9, U"\U00001CA9\U000010E9\U000010E9" },
{ 0x000010EA, U"\U00001CAA\U000010EA\U000010EA" },
{ 0x000010EB, U"\U00001CAB\U000010EB\U000010EB" },
{ 0x000010EC, U"\U00001CAC\U000010EC\U000010EC" },
{ 0x000010ED, U"\U00001CAD\U000010ED\U000010ED" },
{ 0x000010EE, U"\U00001CAE\U000010EE\U000010EE" },
{ 0x000010EF, U"\U00001CAF\U000010EF\U000010EF" },
{ 0x000010F0, U"\U00001CB0\U000010F0\U000010F0" },
{ 0x000010F1, U"\U00001CB1\U000010F1\U000010F1" },
{ 0x000010F2, U"\U00001CB2\U000010F2\U000010F2" },
{ 0x000010F3, U"\U00001CB3\U000010F3\U000010F3" },
{ 0x000010F4, U"\U00001CB4\U000010F4\U000010F4" },
{ 0x000010F5, U"\U00001CB5\U000010F5\U000010F5" },
{ 0x000010F6, U"\U00001CB6\U000010F6\U000010F6" },
{ 0x000010F7, U"\U00001CB7\U000010F7\U000010F7" },
{ 0x000010F8, U"\U00001CB8\U000010F8\U000010F8" },
{ 0x000010F9, U"\U00001CB9\U000010F9\U000010F9" },
{ 0x000010FA, U"\U00001CBA\U000010FA\U000010FA" },
{ 0x000010FD, U"\U00001CBD\U000010FD\U000010FD" },
{ 0x000010FE, U"\U00001CBE\U000010FE\U000010FE" },
{ 0x000010FF, U"\U00001CBF\U000010FF\U000010FF" },
{ 0x000013A0, U"\U000013A0\U0000AB70\U000013A0" },
{ 0x000013A1, U"\U000013A1\U0000AB71\U000013A1" },
{ 0x000013A2, U"\U000013A2\U0000AB72\U000013A2" },
{ 0x000013A3, U"\U000013A3\U0000AB73\U000013A3" },
{ 0x000013A4, U"\U000013A4\U0000AB74\U000013A4" },
{ 0x000013A5, U"\U000013A5\U0000AB75\U000013A5" },
{ 0x000013A6, U"\U000013A6\U0000AB76\U000013A6" },
{ 0x000013A7, U"\U000013A7\U0000AB77\U000013A7" },
{ 0x000013A8, U"\U000013A8\U0000AB78\U000013A8" },
{ 0x000013A9, U"\U000013A9\U0000AB79\U000013A9" },
{ 0x000013AA, U"\U000013AA\U0000AB7A\U000013AA" },
{ 0x000013AB, U"\U000013AB\U0000AB7B\U000013AB" },
{ 0x000013AC, U"\U000013AC\U0000AB7C\U000013AC" },
{ 0x000013AD, U"\U000013AD\U0000AB7D\U000013AD" },
{ 0x000013AE, U"\U000013AE\U0000AB7E\U000013AE" },
{ 0x000013AF, U"\U000013AF\U0000AB7F\U000013AF" },
{ 0x000013B0, U"\U000013B0\U0000AB80\U000013B0" },
{ 0x000013B1, U"\U000013B1\U0000AB81\U000013B1" },
{ 0x000013B2, U"\U000013B2\U0000AB82\U000013B2" },
{ 0x000013B3, U"\U000013B3\U0000AB83\U000013B3" },
{ 0x000013B4, U"\U000013B4\U0000AB84\U000013B4" },
{ 0x000013B5, U"\U000013B5\U0000AB85\U000013B5" },
{ 0x000013B6, U"\U000013B6\U0000AB86\U000013B6" },
{ 0x000013B7, U"\U000013B7\U0000AB87\U000013B7" },
{ 0x000013B8, U"\U000013B8\U0000AB88\U000013B8" },
{ 0x000013B9, U"\U000013B9\U0000AB89\U000013B9" },
{ 0x000013BA, U"\U000013BA\U0000AB8A\U000013BA" },
{ 0x000013BB, U"\U000013BB\U0000AB8B\U000013BB" },
{ 0x000013BC, U"\U000013BC\U0000AB8C\U000013BC" },
{ 0x000013BD, U"\U000013BD\U0000AB8D\U000013BD" },
{ 0x000013BE, U"\U000013BE\U0000AB8E\U000013BE" },
{ 0x000013BF, U"\U000013BF\U0000AB8F\U000013BF" },
{ 0x000013C0, U"\U000013C0\U0000AB90\U000013C0" },
{ 0x000013C1, U"\U000013C1\U0000AB91\U000013C1" },
{ 0x000013C2, U"\U000013C2\U0000AB92\U000013C2" },
{ 0x000013C3, U"\U000013C3\U0000AB93\U000013C3" },
{ 0x000013C4, U"\U000013C4\U0000AB94\U000013C4" },
{ 0x000013C5, U"\U000013C5\U0000AB95\U000013C5" },
{ 0x000013C6, U"\U000013C6\U0000AB96\U000013C6" },
{ 0x000013C7, U"\U000013C7\U0000AB97\U000013C7" },
{ 0x000013C8, U"\U000013C8\U0000AB98\U000013C8" },
{ 0x000013C9, U"\U000013C9\U0000AB99\U000013C9" },
{ 0x000013CA, U"\U000013CA\U0000AB9A\U000013CA" },
{ 0x000013CB, U"\U000013CB\U0000AB9B\U000013CB" },
{ 0x000013CC, U"\U000013CC\U0000AB9C\U000013CC" },
{ 0x000013CD, U"\U000013CD\U0000AB9D\U000013CD" },
{ 0x000013CE, U"\U000013CE\U0000AB9E\U000013CE" },
{ 0x000013CF, U"\U000013CF\U0000AB9F\U000013CF" },
{ 0x000013D0, U"\U000013D0\U0000ABA0\U000013D0" },
{ 0x000013D1, U"\U000013D1\U0000ABA1\U000013D1" },
{ 0x000013D2, U"\U000013D2\U0000ABA2\U000013D2" },
{ 0x000013D3, U"\U000013D3\U0000ABA3\U000013D3" },
{ 0x000013D4, U"\U000013D4\U0000ABA4\U000013D4" },
{ 0x000013D5, U"\U000013D5\U0000ABA5\U000013D5" },
{ 0x000013D6, U"\U000013D6\U0000ABA6\U000013D6" },
{ 0x000013D7, U"\U000013D7\U0000ABA7\U000013D7" },
{ 0x000013D8, U"\U000013D8\U0000ABA8\U000013D8" },
{ 0x000013D9, U"\U000013D9\U0000ABA9\U000013D9" },
{ 0x000013DA, U"\U000013DA\U0000ABAA\U000013DA" },
{ 0x000013DB, U"\U000013DB\U0000ABAB\U000013DB" },
{ 0x000013DC, U"\U000013DC\U0000ABAC\U000013DC" },
{ 0x000013DD, U"\U000013DD\U0000ABAD\U000013DD" },
{ 0x000013DE, U"\U000013DE\U0000ABAE\U000013DE" },
{ 0x000013DF, U"\U000013DF\U0000ABAF\U000013DF" },
{ 0x000013E0, U"\U000013E0\U0000ABB0\U000013E0" },
{ 0x000013E1, U"\U000013E1\U0000ABB1\U000013E1" },
{ 0x000013E2, U"\U000013E2\U0000ABB2\U000013E2" },
{ 0x000013E3, U"\U000013E3\U0000ABB3\U000013E3" },
{ 0x000013E4, U"\U000013E4\U0000ABB4\U000013E4" },
{ 0x000013E5, U"\U000013E5\U0000ABB5\U000013E5" },
{ 0x000013E6, U"\U000013E6\U0000ABB6\U000013E6" },
{ 0x000013E7, U"\U000013E7\U0000ABB7\U000013E7" },
{ 0x000013E8, U"\U000013E8\U0000ABB8\U000013E8" },
{ 0x000013E9, U"\U000013E9\U0000ABB9\U000013E9" },
{ 0x000013EA, U"\U000013EA\U0000ABBA\U000013EA" },
{ 0x000013EB, U"\U000013EB\U0000ABBB\U000013EB" },
{ 0x000013EC, U"\U000013EC\U0000ABBC\U000013EC" },
{ 0x000013ED, U"\U000013ED\U0000ABBD\U000013ED" },
{ 0x000013EE, U"\U000013EE\U0000ABBE\U000013EE" },
{ 0x000013EF, U"\U000013EF\U0000ABBF\U000013EF" },
{ 0x000013F0, U"\U000013F0\U000013F8\U000013F0" },
{ 0x000013F1, U"\U000013F1\U000013F9\U000013F1" },
{ 0x000013F2, U"\U000013F2\U000013FA\U000013F2" },
{ 0x000013F3, U"\U000013F3\U000013FB\U000013F3" },
{ 0x000013F4, U"\U000013F4\U000013FC\U000013F4" },
{ 0x000013F5, U"\U000013F5\U000013FD\U000013F5" },
{ 0x000013F8, U"\U000013F0\U000013F8\U000013F0" },
{ 0x000013F9, U"\U000013F1\U000013F9\U000013F1" },
{ 0x000013FA, U"\U000013F2\U000013FA\U000013F2" },
{ 0x000013FB, U"\U000013F3\U000013FB\U000013F3" },
{ 0x000013FC, U"\U000013F4\U000013FC\U000013F4" },
{ 0x000013FD, U"\U000013F5\U000013FD\U000013F5" },
{ 0x00001C80, U"\U00000412\U00001C80\U00000412" },
{ 0x00001C81, U"\U00000414\U00001C81\U00000414" },
{ 0x00001C82, U"\U0000041E\U00001C82\U0000041E" },
{ 0x00001C83, U"\U00000421\U00001C83\U00000421" },
{ 0x00001C84, U"\U00000422\U00001C84\U00000422" },
{ 0x00001C85, U"\U00000422\U00001C85\U00000422" },
{ 0x00001C86, U"\U0000042A\U00001C86\U0000042A" },
{ 0x00001C87, U"\U00000462\U00001C87\U00000462" },
{ 0x00001C88, U"\U0000A64A\U00001C88\U0000A64A" },
{ 0x00001C90, U"\U00001C90\U000010D0\U00001C90" },
{ 0x00001C91, U"\U00001C91\U000010D1\U00001C91" },
{ 0x00001C92, U"\U00001C92\U000010D2\U00001C92" },
{ 0x00001C93, U"\U00001C93\U000010D3\U00001C93" },
{ 0x00001C94, U"\U00001C94\U000010D4\U00001C94" },
{ 0x00001C95, U"\U00001C95\U000010D5\U00001C95" },
{ 0x00001C96, U"\U00001C96\U000010D6\U00001C96" },
{ 0x00001C97, U"\U00001C97\U000010D7\U00001C97" },
{ 0x00001C98, U"\U00001C98\U000010D8\U00001C98" },
{ 0x00001C99, U"\U00001C99\U000010D9\U00001C99" },
{ 0x00001C9A, U"\U00001C9A\U000010DA\U00001C9A" },
{ 0x00001C9B, U"\U00001C9B\U000010DB\U00001C9B" },
{ 0x00001C9C, U"\U00001C9C\U000010DC\U00001C9C" },
{ 0x00001C9D, U"\U00001C9D\U000010DD\U00001C9D" },
{ 0x00001C9E, U"\U00001C9E\U000010DE\U00001C9E" },
{ 0x00001C9F, U"\U00001C9F\U000010DF\U00001C9F" },
{ 0x00001CA0, U"\U00001CA0\U000010E0\U00001CA0" },
{ 0x00001CA1, U"\U00001CA1\U000010E1\U00001CA1" },
{ 0x00001CA2, U"\U00001CA2\U000010E2\U00001CA2" },
{ 0x00001CA3, U"\U00001CA3\U000010E3\U00001CA3" },
{ 0x00001CA4, U"\U00001CA4\U000010E4\U00001CA4" },
{ 0x00001CA5, U"\U00001CA5\U000010E5\U00001CA5" },
{ 0x00001CA6, U"\U00001CA6\U000010E6\U00001CA6" },
{ 0x00001CA7, U"\U00001CA7\U000010E7\U00001CA7" },
{ 0x00001CA8, U"\U00001CA8\U000010E8\U00001CA8" },
{ 0x00001CA9, U"\U00001CA9\U000010E9\U00001CA9" },
{ 0x00001CAA, U"\U00001CAA\U000010EA\U00001CAA" },
{ 0x00001CAB, U"\U00001CAB\U000010EB\U00001CAB" },
{ 0x00001CAC, U"\U00001CAC\U000010EC\U00001CAC" },
{ 0x00001CAD, U"\U00001CAD\U000010ED\U00001CAD" },
{ 0x00001CAE, U"\U00001CAE\U000010EE\U00001CAE" },
{ 0x00001CAF, U"\U00001CAF\U000010EF\U00001CAF" },
{ 0x00001CB0, U"\U00001CB0\U000010F0\U00001CB0" },
{ 0x00001CB1, U"\U00001CB1\U000010F1\U00001CB1" },
{ 0x00001CB2, U"\U00001CB2\U000010F2\U00001CB2" },
{ 0x00001CB3, U"\U00001CB3\U000010F3\U00001CB3" },
{ 0x00001CB4, U"\U00001CB4\U000010F4\U00001CB4" },
{ 0x00001CB5, U"\U00001CB5\U000010F5\U00001CB5" },
{ 0x00001CB6, U"\U00001CB6\U000010F6\U00001CB6" },
{ 0x00001CB7, U"\U00001CB7\U000010F7\U00001CB7" },
{ 0x00001CB8, U"\U00001CB8\U000010F8\U00001CB8" },
{ 0x00001CB9, U"\U00001CB9\U000010F9\U00001CB9" },
{ 0x00001CBA, U"\U00001CBA\U000010FA\U00001CBA" },
{ 0x00001CBD, U"\U00001CBD\U000010FD\U00001CBD" },
{ 0x00001CBE, U"\U00001CBE\U000010FE\U00001CBE" },
{ 0x00001CBF, U"\U00001CBF\U000010FF\U00001CBF" },
{ 0x00001D79, U"\U0000A77D\U00001D79\U0000A77D" },
{ 0x00001D7D, U"\U00002C63\U00001D7D\U00002C63" },
{ 0x00001D8E, U"\U0000A7C6\U00001D8E\U0000A7C6" },
{ 0x00001E00, U"\U00001E00\U00001E01\U00001E00" },
{ 0x00001E01, U"\U00001E00\U00001E01\U00001E00" },
{ 0x00001E02, U"\U00001E02\U00001E03\U00001E02" },
{ 0x00001E03, U"\U00001E02\U00001E03\U00001E02" },
{ 0x00001E04, U"\U00001E04\U00001E05\U00001E04" },
{ 0x00001E05, U"\U00001E04\U00001E05\U00001E04" },
{ 0x00001E06, U"\U00001E06\U00001E07\U00001E06" },
{ 0x00001E07, U"\U00001E06\U00001E07\U00001E06" },
{ 0x00001E08, U"\U00001E08\U00001E09\U00001E08" },
{ 0x00001E09, U"\U00001E08\U00001E09\U00001E08" },
{ 0x00001E0A, U"\U00001E0A\U00001E0B\U00001E0A" },
{ 0x00001E0B, U"\U00001E0A\U00001E0B\U00001E0A" },
{ 0x00001E0C, U"\U00001E0C\U00001E0D\U00001E0C" },
{ 0x00001E0D, U"\U00001E0C\U00001E0D\U00001E0C" },
{ 0x00001E0E, U"\U00001E0E\U00001E0F\U00001E0E" },
{ 0x00001E0F, U"\U00001E0E\U00001E0F\U00001E0E" },
{ 0x00001E10, U"\U00001E10\U00001E11\U00001E10" },
{ 0x00001E11, U"\U00001E10\U00001E11\U00001E10" },
{ 0x00001E12, U"\U00001E12\U00001E13\U00001E12" },
{ 0x00001E13, U"\U00001E12\U00001E13\U00001E12" },
{ 0x00001E14, U"\U00001E14\U00001E15\U00001E14" },
{ 0x00001E15, U"\U00001E14\U00001E15\U00001E14" },
{ 0x00001E16, U"\U00001E16\U00001E17\U00001E16" },
{ 0x00001E17, U"\U00001E16\U00001E17\U00001E16" },
{ 0x00001E18, U"\U00001E18\U00001E19\U00001E18" },
{ 0x00001E19, U"\U00001E18\U00001E19\U00001E18" },
{ 0x00001E1A, U"\U00001E1A\U00001E1B\U00001E1A" },
{ 0x00001E1B, U"\U00001E1A\U00001E1B\U00001E1A" },
{ 0x00001E1C, U"\U00001E1C\U00001E1D\U00001E1C" },
{ 0x00001E1D, U"\U00001E1C\U00001E1D\U00001E1C" },
{ 0x00001E1E, U"\U00001E1E\U00001E1F\U00001E1E" },
{ 0x00001E1F, U"\U00001E1E\U00001E1F\U00001E1E" },
{ 0x00001E20, U"\U00001E20\U00001E21\U00001E20" },
{ 0x00001E21, U"\U00001E20\U00001E21\U00001E20" },
{ 0x00001E22, U"\U00001E22\U00001E23\U00001E22" },
{ 0x00001E23, U"\U00001E22\U00001E23\U00001E22" },
{ 0x00001E24, U"\U00001E24\U00001E25\U00001E24" },
{ 0x00001E25, U"\U00001E24\U00001E25\U00001E24" },
{ 0x00001E26, U"\U00001E26\U00001E27\U00001E26" },
{ 0x00001E27, U"\U00001E26\U00001E27\U00001E26" },
{ 0x00001E28, U"\U00001E28\U00001E29\U00001E28" },
{ 0x00001E29, U"\U00001E28\U00001E29\U00001E28" },
{ 0x00001E2A, U"\U00001E2A\U00001E2B\U00001E2A" },
{ 0x00001E2B, U"\U00001E2A\U00001E2B\U00001E2A" },
{ 0x00001E2C, U"\U00001E2C\U00001E2D\U00001E2C" },
{ 0x00001E2D, U"\U00001E2C\U00001E2D\U00001E2C" },
{ 0x00001E2E, U"\U00001E2E\U00001E2F\U00001E2E" },
{ 0x00001E2F, U"\U00001E2E\U00001E2F\U00001E2E" },
{ 0x00001E30, U"\U00001E30\U00001E31\U00001E30" },
{ 0x00001E31, U"\U00001E30\U00001E31\U00001E30" },
{ 0x00001E32, U"\U00001E32\U00001E33\U00001E32" },
{ 0x00001E33, U"\U00001E32\U00001E33\U00001E32" },
{ 0x00001E34, U"\U00001E34\U00001E35\U00001E34" },
{ 0x00001E35, U"\U00001E34\U00001E35\U00001E34" },
{ 0x00001E36, U"\U00001E36\U00001E37\U00001E36" },
{ 0x00001E37, U"\U00001E36\U00001E37\U00001E36" },
{ 0x00001E38, U"\U00001E38\U00001E39\U00001E38" },
{ 0x00001E39, U"\U00001E38\U00001E39\U00001E38" },
{ 0x00001E3A, U"\U00001E3A\U00001E3B\U00001E3A" },
{ 0x00001E3B, U"\U00001E3A\U00001E3B\U00001E3A" },
{ 0x00001E3C, U"\U00001E3C\U00001E3D\U00001E3C" },
{ 0x00001E3D, U"\U00001E3C\U00001E3D\U00001E3C" },
{ 0x00001E3E, U"\U00001E3E\U00001E3F\U00001E3E" },
{ 0x00001E3F, U"\U00001E3E\U00001E3F\U00001E3E" },
{ 0x00001E40, U"\U00001E40\U00001E41\U00001E40" },
{ 0x00001E41, U"\U00001E40\U00001E41\U00001E40" },
{ 0x00001E42, U"\U00001E42\U00001E43\U00001E42" },
{ 0x00001E43, U"\U00001E42\U00001E43\U00001E42" },
{ 0x00001E44, U"\U00001E44\U00001E45\U00001E44" },
{ 0x00001E45, U"\U00001E44\U00001E45\U00001E44" },
{ 0x00001E46, U"\U00001E46\U00001E47\U00001E46" },
{ 0x00001E47, U"\U00001E46\U00001E47\U00001E46" },
{ 0x00001E48, U"\U00001E48\U00001E49\U00001E48" },
{ 0x00001E49, U"\U00001E48\U00001E49\U00001E48" },
{ 0x00001E4A, U"\U00001E4A\U00001E4B\U00001E4A" },
{ 0x00001E4B, U"\U00001E4A\U00001E4B\U00001E4A" },
{ 0x00001E4C, U"\U00001E4C\U00001E4D\U00001E4C" },
{ 0x00001E4D, U"\U00001E4C\U00001E4D\U00001E4C" },
{ 0x00001E4E, U"\U00001E4E\U00001E4F\U00001E4E" },
{ 0x00001E4F, U"\U00001E4E\U00001E4F\U00001E4E" },
{ 0x00001E50, U"\U00001E50\U00001E51\U00001E50" },
{ 0x00001E51, U"\U00001E50\U00001E51\U00001E50" },
{ 0x00001E52, U"\U00001E52\U00001E53\U00001E52" },
{ 0x00001E53, U"\U00001E52\U00001E53\U00001E52" },
{ 0x00001E54, U"\U00001E54\U00001E55\U00001E54" },
{ 0x00001E55, U"\U00001E54\U00001E55\U00001E54" },
{ 0x00001E56, U"\U00001E56\U00001E57\U00001E56" },
{ 0x00001E57, U"\U00001E56\U00001E57\U00001E56" },
{ 0x00001E58, U"\U00001E58\U00001E59\U00001E58" },
{ 0x00001E59, U"\U00001E58\U00001E59\U00001E58" },
{ 0x00001E5A, U"\U00001E5A\U00001E5B\U00001E5A" },
{ 0x00001E5B, U"\U00001E5A\U00001E5B\U00001E5A" },
{ 0x00001E5C, U"\U00001E5C\U00001E5D\U00001E5C" },
{ 0x00001E5D, U"\U00001E5C\U00001E5D\U00001E5C" },
{ 0x00001E5E, U"\U00001E5E\U00001E5F\U00001E5E" },
{ 0x00001E5F, U"\U00001E5E\U00001E5F\U00001E5E" },
{ 0x00001E60, U"\U00001E60\U00001E61\U00001E60" },
{ 0x00001E61, U"\U00001E60\U00001E61\U00001E60" },
{ 0x00001E62, U"\U00001E62\U00001E63\U00001E62" },
{ 0x00001E63, U"\U00001E62\U00001E63\U00001E62" },
{ 0x00001E64, U"\U00001E64\U00001E65\U00001E64" },
{ 0x00001E65, U"\U00001E64\U00001E65\U00001E64" },
{ 0x00001E66, U"\U00001E66\U00001E67\U00001E66" },
{ 0x00001E67, U"\U00001E66\U00001E67\U00001E66" },
{ 0x00001E68, U"\U00001E68\U00001E69\U00001E68" },
{ 0x00001E69, U"\U00001E68\U00001E69\U00001E68" },
{ 0x00001E6A, U"\U00001E6A\U00001E6B\U00001E6A" },
{ 0x00001E6B, U"\U00001E6A\U00001E6B\U00001E6A" },
{ 0x00001E6C, U"\U00001E6C\U00001E6D\U00001E6C" },
{ 0x00001E6D, U"\U00001E6C\U00001E6D\U00001E6C" },
{ 0x00001E6E, U"\U00001E6E\U00001E6F\U00001E6E" },
{ 0x00001E6F, U"\U00001E6E\U00001E6F\U00001E6E" },
{ 0x00001E70, U"\U00001E70\U00001E71\U00001E70" },
{ 0x00001E71, U"\U00001E70\U00001E71\U00001E70" },
{ 0x00001E72, U"\U00001E72\U00001E73\U00001E72" },
{ 0x00001E73, U"\U00001E72\U00001E73\U00001E72" },
{ 0x00001E74, U"\U00001E74\U00001E75\U00001E74" },
{ 0x00001E75, U"\U00001E74\U00001E75\U00001E74" },
{ 0x00001E76, U"\U00001E76\U00001E77\U00001E76" },
{ 0x00001E77, U"\U00001E76\U00001E77\U00001E76" },
{ 0x00001E78, U"\U00001E78\U00001E79\U00001E78" },
{ 0x00001E79, U"\U00001E78\U00001E79\U00001E78" },
{ 0x00001E7A, U"\U00001E7A\U00001E7B\U00001E7A" },
{ 0x00001E7B, U"\U00001E7A\U00001E7B\U00001E7A" },
{ 0x00001E7C, U"\U00001E7C\U00001E7D\U00001E7C" },
{ 0x00001E7D, U"\U00001E7C\U00001E7D\U00001E7C" },
{ 0x00001E7E, U"\U00001E7E\U00001E7F\U00001E7E" },
{ 0x00001E7F, U"\U00001E7E\U00001E7F\U00001E7E" },
{ 0x00001E80, U"\U00001E80\U00001E81\U00001E80" },
{ 0x00001E81, U"\U00001E80\U00001E81\U00001E80" },
{ 0x00001E82, U"\U00001E82\U00001E83\U00001E82" },
{ 0x00001E83, U"\U00001E82\U00001E83\U00001E82" },
{ 0x00001E84, U"\U00001E84\U00001E85\U00001E84" },
{ 0x00001E85, U"\U00001E84\U00001E85\U00001E84" },
{ 0x00001E86, U"\U00001E86\U00001E87\U00001E86" },
{ 0x00001E87, U"\U00001E86\U00001E87\U00001E86" },
{ 0x00001E88, U"\U00001E88\U00001E89\U00001E88" },
{ 0x00001E89, U"\U00001E88\U00001E89\U00001E88" },
{ 0x00001E8A, U"\U00001E8A\U00001E8B\U00001E8A" },
{ 0x00001E8B, U"\U00001E8A\U00001E8B\U00001E8A" },
{ 0x00001E8C, U"\U00001E8C\U00001E8D\U00001E8C" },
{ 0x00001E8D, U"\U00001E8C\U00001E8D\U00001E8C" },
{ 0x00001E8E, U"\U00001E8E\U00001E8F\U00001E8E" },
{ 0x00001E8F, U"\U00001E8E\U00001E8F\U00001E8E" },
{ 0x00001E90, U"\U00001E90\U00001E91\U00001E90" },
{ 0x00001E91, U"\U00001E90\U00001E91\U00001E90" },
{ 0x00001E92, U"\U00001E92\U00001E93\U00001E92" },
{ 0x00001E93, U"\U00001E92\U00001E93\U00001E92" },
{ 0x00001E94, U"\U00001E94\U00001E95\U00001E94" },
{ 0x00001E95, U"\U00001E94\U00001E95\U00001E94" },
{ 0x00001E9B, U"\U00001E60\U00001E9B\U00001E60" },
{ 0x00001E9E, U"\U00001E9E\U000000DF\U00001E9E" },
{ 0x00001EA0, U"\U00001EA0\U00001EA1\U00001EA0" },
{ 0x00001EA1, U"\U00001EA0\U00001EA1\U00001EA0" },
{ 0x00001EA2, U"\U00001EA2\U00001EA3\U00001EA2" },
{ 0x00001EA3, U"\U00001EA2\U00001EA3\U00001EA2" },
{ 0x00001EA4, U"\U00001EA4\U00001EA5\U00001EA4" },
{ 0x00001EA5, U"\U00001EA4\U00001EA5\U00001EA4" },
{ 0x00001EA6, U"\U00001EA6\U00001EA7\U00001EA6" },
{ 0x00001EA7, U"\U00001EA6\U00001EA7\U00001EA6" },
{ 0x00001EA8, U"\U00001EA8\U00001EA9\U00001EA8" },
{ 0x00001EA9, U"\U00001EA8\U00001EA9\U00001EA8" },
{ 0x00001EAA, U"\U00001EAA\U00001EAB\U00001EAA" },
{ 0x00001EAB, U"\U00001EAA\U00001EAB\U00001EAA" },
{ 0x00001EAC, U"\U00001EAC\U00001EAD\U00001EAC" },
{ 0x00001EAD, U"\U00001EAC\U00001EAD\U00001EAC" },
{ 0x00001EAE, U"\U00001EAE\U00001EAF\U00001EAE" },
{ 0x00001EAF, U"\U00001EAE\U00001EAF\U00001EAE" },
{ 0x00001EB0, U"\U00001EB0\U00001EB1\U00001EB0" },
{ 0x00001EB1, U"\U00001EB0\U00001EB1\U00001EB0" },
{ 0x00001EB2, U"\U00001EB2\U00001EB3\U00001EB2" },
{ 0x00001EB3, U"\U00001EB2\U00001EB3\U00001EB2" },
{ 0x00001EB4, U"\U00001EB4\U00001EB5\U00001EB4" },
{ 0x00001EB5, U"\U00001EB4\U00001EB5\U00001EB4" },
{ 0x00001EB6, U"\U00001EB6\U00001EB7\U00001EB6" },
{ 0x00001EB7, U"\U00001EB6\U00001EB7\U00001EB6" },
{ 0x00001EB8, U"\U00001EB8\U00001EB9\U00001EB8" },
{ 0x00001EB9, U"\U00001EB8\U00001EB9\U00001EB8" },
{ 0x00001EBA, U"\U00001EBA\U00001EBB\U00001EBA" },
{ 0x00001EBB, U"\U00001EBA\U00001EBB\U00001EBA" },
{ 0x00001EBC, U"\U00001EBC\U00001EBD\U00001EBC" },
{ 0x00001EBD, U"\U00001EBC\U00001EBD\U00001EBC" },
{ 0x00001EBE, U"\U00001EBE\U00001EBF\U00001EBE" },
{ 0x00001EBF, U"\U00001EBE\U00001EBF\U00001EBE" },
{ 0x00001EC0, U"\U00001EC0\U00001EC1\U00001EC0" },
{ 0x00001EC1, U"\U00001EC0\U00001EC1\U00001EC0" },
{ 0x00001EC2, U"\U00001EC2\U00001EC3\U00001EC2" },
{ 0x00001EC3, U"\U00001EC2\U00001EC3\U00001EC2" },
{ 0x00001EC4, U"\U00001EC4\U00001EC5\U00001EC4" },
{ 0x00001EC5, U"\U00001EC4\U00001EC5\U00001EC4" },
{ 0x00001EC6, U"\U00001EC6\U00001EC7\U00001EC6" },
{ 0x00001EC7, U"\U00001EC6\U00001EC7\U00001EC6" },
{ 0x00001EC8, U"\U00001EC8\U00001EC9\U00001EC8" },
{ 0x00001EC9, U"\U00001EC8\U00001EC9\U00001EC8" },
{ 0x00001ECA, U"\U00001ECA\U00001ECB\U00001ECA" },
{ 0x00001ECB, U"\U00001ECA\U00001ECB\U00001ECA" },
{ 0x00001ECC, U"\U00001ECC\U00001ECD\U00001ECC" },
{ 0x00001ECD, U"\U00001ECC\U00001ECD\U00001ECC" },
{ 0x00001ECE, U"\U00001ECE\U00001ECF\U00001ECE" },
{ 0x00001ECF, U"\U00001ECE\U00001ECF\U00001ECE" },
{ 0x00001ED0, U"\U00001ED0\U00001ED1\U00001ED0" },
{ 0x00001ED1, U"\U00001ED0\U00001ED1\U00001ED0" },
{ 0x00001ED2, U"\U00001ED2\U00001ED3\U00001ED2" },
{ 0x00001ED3, U"\U00001ED2\U00001ED3\U00001ED2" },
{ 0x00001ED4, U"\U00001ED4\U00001ED5\U00001ED4" },
{ 0x00001ED5, U"\U00001ED4\U00001ED5\U00001ED4" },
{ 0x00001ED6, U"\U00001ED6\U00001ED7\U00001ED6" },
{ 0x00001ED7, U"\U00001ED6\U00001ED7\U00001ED6" },
{ 0x00001ED8, U"\U00001ED8\U00001ED9\U00001ED8" },
{ 0x00001ED9, U"\U00001ED8\U00001ED9\U00001ED8" },
{ 0x00001EDA, U"\U00001EDA\U00001EDB\U00001EDA" },
{ 0x00001EDB, U"\U00001EDA\U00001EDB\U00001EDA" },
{ 0x00001EDC, U"\U00001EDC\U00001EDD\U00001EDC" },
{ 0x00001EDD, U"\U00001EDC\U00001EDD\U00001EDC" },
{ 0x00001EDE, U"\U00001EDE\U00001EDF\U00001EDE" },
{ 0x00001EDF, U"\U00001EDE\U00001EDF\U00001EDE" },
{ 0x00001EE0, U"\U00001EE0\U00001EE1\U00001EE0" },
{ 0x00001EE1, U"\U00001EE0\U00001EE1\U00001EE0" },
{ 0x00001EE2, U"\U00001EE2\U00001EE3\U00001EE2" },
{ 0x00001EE3, U"\U00001EE2\U00001EE3\U00001EE2" },
{ 0x00001EE4, U"\U00001EE4\U00001EE5\U00001EE4" },
{ 0x00001EE5, U"\U00001EE4\U00001EE5\U00001EE4" },
{ 0x00001EE6, U"\U00001EE6\U00001EE7\U00001EE6" },
{ 0x00001EE7, U"\U00001EE6\U00001EE7\U00001EE6" },
{ 0x00001EE8, U"\U00001EE8\U00001EE9\U00001EE8" },
{ 0x00001EE9, U"\U00001EE8\U00001EE9\U00001EE8" },
{ 0x00001EEA, U"\U00001EEA\U00001EEB\U00001EEA" },
{ 0x00001EEB, U"\U00001EEA\U00001EEB\U00001EEA" },
{ 0x00001EEC, U"\U00001EEC\U00001EED\U00001EEC" },
{ 0x00001EED, U"\U00001EEC\U00001EED\U00001EEC" },
{ 0x00001EEE, U"\U00001EEE\U00001EEF\U00001EEE" },
{ 0x00001EEF, U"\U00001EEE\U00001EEF\U00001EEE" },
{ 0x00001EF0, U"\U00001EF0\U00001EF1\U00001EF0" },
{ 0x00001EF1, U"\U00001EF0\U00001EF1\U00001EF0" },
{ 0x00001EF2, U"\U00001EF2\U00001EF3\U00001EF2" },
{ 0x00001EF3, U"\U00001EF2\U00001EF3\U00001EF2" },
{ 0x00001EF4, U"\U00001EF4\U00001EF5\U00001EF4" },
{ 0x00001EF5, U"\U00001EF4\U00001EF5\U00001EF4" },
{ 0x00001EF6, U"\U00001EF6\U00001EF7\U00001EF6" },
{ 0x00001EF7, U"\U00001EF6\U00001EF7\U00001EF6" },
{ 0x00001EF8, U"\U00001EF8\U00001EF9\U00001EF8" },
{ 0x00001EF9, U"\U00001EF8\U00001EF9\U00001EF8" },
{ 0x00001EFA, U"\U00001EFA\U00001EFB\U00001EFA" },
{ 0x00001EFB, U"\U00001EFA\U00001EFB\U00001EFA" },
{ 0x00001EFC, U"\U00001EFC\U00001EFD\U00001EFC" },
{ 0x00001EFD, U"\U00001EFC\U00001EFD\U00001EFC" },
{ 0x00001EFE, U"\U00001EFE\U00001EFF\U00001EFE" },
{ 0x00001EFF, U"\U00001EFE\U00001EFF\U00001EFE" },
{ 0x00001F00, U"\U00001F08\U00001F00\U00001F08" },
{ 0x00001F01, U"\U00001F09\U00001F01\U00001F09" },
{ 0x00001F02, U"\U00001F0A\U00001F02\U00001F0A" },
{ 0x00001F03, U"\U00001F0B\U00001F03\U00001F0B" },
{ 0x00001F04, U"\U00001F0C\U00001F04\U00001F0C" },
{ 0x00001F05, U"\U00001F0D\U00001F05\U00001F0D" },
{ 0x00001F06, U"\U00001F0E\U00001F06\U00001F0E" },
{ 0x00001F07, U"\U00001F0F\U00001F07\U00001F0F" },
{ 0x00001F08, U"\U00001F08\U00001F00\U00001F08" },
{ 0x00001F09, U"\U00001F09\U00001F01\U00001F09" },
{ 0x00001F0A, U"\U00001F0A\U00001F02\U00001F0A" },
{ 0x00001F0B, U"\U00001F0B\U00001F03\U00001F0B" },
{ 0x00001F0C, U"\U00001F0C\U00001F04\U00001F0C" },
{ 0x00001F0D, U"\U00001F0D\U00001F05\U00001F0D" },
{ 0x00001F0E, U"\U00001F0E\U00001F06\U00001F0E" },
{ 0x00001F0F, U"\U00001F0F\U00001F07\U00001F0F" },
{ 0x00001F10, U"\U00001F18\U00001F10\U00001F18" },
{ 0x00001F11, U"\U00001F19\U00001F11\U00001F19" },
{ 0x00001F12, U"\U00001F1A\U00001F12\U00001F1A" },
{ 0x00001F13, U"\U00001F1B\U00001F13\U00001F1B" },
{ 0x00001F14, U"\U00001F1C\U00001F14\U00001F1C" },
{ 0x00001F15, U"\U00001F1D\U00001F15\U00001F1D" },
{ 0x00001F18, U"\U00001F18\U00001F10\U00001F18" },
{ 0x00001F19, U"\U00001F19\U00001F11\U00001F19" },
{ 0x00001F1A, U"\U00001F1A\U00001F12\U00001F1A" },
{ 0x00001F1B, U"\U00001F1B\U00001F13\U00001F1B" },
{ 0x00001F1C, U"\U00001F1C\U00001F14\U00001F1C" },
{ 0x00001F1D, U"\U00001F1D\U00001F15\U00001F1D" },
{ 0x00001F20, U"\U00001F28\U00001F20\U00001F28" },
{ 0x00001F21, U"\U00001F29\U00001F21\U00001F29" },
{ 0x00001F22, U"\U00001F2A\U00001F22\U00001F2A" },
{ 0x00001F23, U"\U00001F2B\U00001F23\U00001F2B" },
{ 0x00001F24, U"\U00001F2C\U00001F24\U00001F2C" },
{ 0x00001F25, U"\U00001F2D\U00001F25\U00001F2D" },
{ 0x00001F26, U"\U00001F2E\U00001F26\U00001F2E" },
{ 0x00001F27, U"\U00001F2F\U00001F27\U00001F2F" },
{ 0x00001F28, U"\U00001F28\U00001F20\U00001F28" },
{ 0x00001F29, U"\U00001F29\U00001F21\U00001F29" },
{ 0x00001F2A, U"\U00001F2A\U00001F22\U00001F2A" },
{ 0x00001F2B, U"\U00001F2B\U00001F23\U00001F2B" },
{ 0x00001F2C, U"\U00001F2C\U00001F24\U00001F2C" },
{ 0x00001F2D, U"\U00001F2D\U00001F25\U00001F2D" },
{ 0x00001F2E, U"\U00001F2E\U00001F26\U00001F2E" },
{ 0x00001F2F, U"\U00001F2F\U00001F27\U00001F2F" },
{ 0x00001F30, U"\U00001F38\U00001F30\U00001F38" },
{ 0x00001F31, U"\U00001F39\U00001F31\U00001F39" },
{ 0x00001F32, U"\U00001F3A\U00001F32\U00001F3A" },
{ 0x00001F33, U"\U00001F3B\U00001F33\U00001F3B" },
{ 0x00001F34, U"\U00001F3C\U00001F34\U00001F3C" },
{ 0x00001F35, U"\U00001F3D\U00001F35\U00001F3D" },
{ 0x00001F36, U"\U00001F3E\U00001F36\U00001F3E" },
{ 0x00001F37, U"\U00001F3F\U00001F37\U00001F3F" },
{ 0x00001F38, U"\U00001F38\U00001F30\U00001F38" },
{ 0x00001F39, U"\U00001F39\U00001F31\U00001F39" },
{ 0x00001F3A, U"\U00001F3A\U00001F32\U00001F3A" },
{ 0x00001F3B, U"\U00001F3B\U00001F33\U00001F3B" },
{ 0x00001F3C, U"\U00001F3C\U00001F34\U00001F3C" },
{ 0x00001F3D, U"\U00001F3D\U00001F35\U00001F3D" },
{ 0x00001F3E, U"\U00001F3E\U00001F36\U00001F3E" },
{ 0x00001F3F, U"\U00001F3F\U00001F37\U00001F3F" },
{ 0x00001F40, U"\U00001F48\U00001F40\U00001F48" },
{ 0x00001F41, U"\U00001F49\U00001F41\U00001F49" },
{ 0x00001F42, U"\U00001F4A\U00001F42\U00001F4A" },
{ 0x00001F43, U"\U00001F4B\U00001F43\U00001F4B" },
{ 0x00001F44, U"\U00001F4C\U00001F44\U00001F4C" },
{ 0x00001F45, U"\U00001F4D\U00001F45\U00001F4D" },
{ 0x00001F48, U"\U00001F48\U00001F40\U00001F48" },
{ 0x00001F49, U"\U00001F49\U00001F41\U00001F49" },
{ 0x00001F4A, U"\U00001F4A\U00001F42\U00001F4A" },
{ 0x00001F4B, U"\U00001F4B\U00001F43\U00001F4B" },
{ 0x00001F4C, U"\U00001F4C\U00001F44\U00001F4C" },
{ 0x00001F4D, U"\U00001F4D\U00001F45\U00001F4D" },
{ 0x00001F51, U"\U00001F59\U00001F51\U00001F59" },
{ 0x00001F53, U"\U00001F5B\U00001F53\U00001F5B" },
{ 0x00001F55, U"\U00001F5D\U00001F55\U00001F5D" },
{ 0x00001F57, U"\U00001F5F\U00001F57\U00001F5F" },
{ 0x00001F59, U"\U00001F59\U00001F51\U00001F59" },
{ 0x00001F5B, U"\U00001F5B\U00001F53\U00001F5B" },
{ 0x00001F5D, U"\U00001F5D\U00001F55\U00001F5D" },
{ 0x00001F5F, U"\U00001F5F\U00001F57\U00001F5F" },
{ 0x00001F60, U"\U00001F68\U00001F60\U00001F68" },
{ 0x00001F61, U"\U00001F69\U00001F61\U00001F69" },
{ 0x00001F62, U"\U00001F6A\U00001F62\U00001F6A" },
{ 0x00001F63, U"\U00001F6B\U00001F63\U00001F6B" },
{ 0x00001F64, U"\U00001F6C\U00001F64\U00001F6C" },
{ 0x00001F65, U"\U00001F6D\U00001F65\U00001F6D" },
{ 0x00001F66, U"\U00001F6E\U00001F66\U00001F6E" },
{ 0x00001F67, U"\U00001F6F\U00001F67\U00001F6F" },
{ 0x00001F68, U"\U00001F68\U00001F60\U00001F68" },
{ 0x00001F69, U"\U00001F69\U00001F61\U00001F69" },
{ 0x00001F6A, U"\U00001F6A\U00001F62\U00001F6A" },
{ 0x00001F6B, U"\U00001F6B\U00001F63\U00001F6B" },
{ 0x00001F6C, U"\U00001F6C\U00001F64\U00001F6C" },
{ 0x00001F6D, U"\U00001F6D\U00001F65\U00001F6D" },
{ 0x00001F6E, U"\U00001F6E\U00001F66\U00001F6E" },
{ 0x00001F6F, U"\U00001F6F\U00001F67\U00001F6F" },
{ 0x00001F70, U"\U00001FBA\U00001F70\U00001FBA" },
{ 0x00001F71, U"\U00001FBB\U00001F71\U00001FBB" },
{ 0x00001F72, U"\U00001FC8\U00001F72\U00001FC8" },
{ 0x00001F73, U"\U00001FC9\U00001F73\U00001FC9" },
{ 0x00001F74, U"\U00001FCA\U00001F74\U00001FCA" },
{ 0x00001F75, U"\U00001FCB\U00001F75\U00001FCB" },
{ 0x00001F76, U"\U00001FDA\U00001F76\U00001FDA" },
{ 0x00001F77, U"\U00001FDB\U00001F77\U00001FDB" },
{ 0x00001F78, U"\U00001FF8\U00001F78\U00001FF8" },
{ 0x00001F79, U"\U00001FF9\U00001F79\U00001FF9" },
{ 0x00001F7A, U"\U00001FEA\U00001F7A\U00001FEA" },
{ 0x00001F7B, U"\U00001FEB\U00001F7B\U00001FEB" },
{ 0x00001F7C, U"\U00001FFA\U00001F7C\U00001FFA" },
{ 0x00001F7D, U"\U00001FFB\U00001F7D\U00001FFB" },
{ 0x00001F80, U"\U00001F88\U00001F80\U00001F88" },
{ 0x00001F81, U"\U00001F89\U00001F81\U00001F89" },
{ 0x00001F82, U"\U00001F8A\U00001F82\U00001F8A" },
{ 0x00001F83, U"\U00001F8B\U00001F83\U00001F8B" },
{ 0x00001F84, U"\U00001F8C\U00001F84\U00001F8C" },
{ 0x00001F85, U"\U00001F8D\U00001F85\U00001F8D" },
{ 0x00001F86, U"\U00001F8E\U00001F86\U00001F8E" },
{ 0x00001F87, U"\U00001F8F\U00001F87\U00001F8F" },
{ 0x00001F88, U"\U00001F88\U00001F80\U00001F88" },
{ 0x00001F89, U"\U00001F89\U00001F81\U00001F89" },
{ 0x00001F8A, U"\U00001F8A\U00001F82\U00001F8A" },
{ 0x00001F8B, U"\U00001F8B\U00001F83\U00001F8B" },
{ 0x00001F8C, U"\U00001F8C\U00001F84\U00001F8C" },
{ 0x00001F8D, U"\U00001F8D\U00001F85\U00001F8D" },
{ 0x00001F8E, U"\U00001F8E\U00001F86\U00001F8E" },
{ 0x00001F8F, U"\U00001F8F\U00001F87\U00001F8F" },
{ 0x00001F90, U"\U00001F98\U00001F90\U00001F98" },
{ 0x00001F91, U"\U00001F99\U00001F91\U00001F99" },
{ 0x00001F92, U"\U00001F9A\U00001F92\U00001F9A" },
{ 0x00001F93, U"\U00001F9B\U00001F93\U00001F9B" },
{ 0x00001F94, U"\U00001F9C\U00001F94\U00001F9C" },
{ 0x00001F95, U"\U00001F9D\U00001F95\U00001F9D" },
{ 0x00001F96, U"\U00001F9E\U00001F96\U00001F9E" },
{ 0x00001F97, U"\U00001F9F\U00001F97\U00001F9F" },
{ 0x00001F98, U"\U00001F98\U00001F90\U00001F98" },
{ 0x00001F99, U"\U00001F99\U00001F91\U00001F99" },
{ 0x00001F9A, U"\U00001F9A\U00001F92\U00001F9A" },
{ 0x00001F9B, U"\U00001F9B\U00001F93\U00001F9B" },
{ 0x00001F9C, U"\U00001F9C\U00001F94\U00001F9C" },
{ 0x00001F9D, U"\U00001F9D\U00001F95\U00001F9D" },
{ 0x00001F9E, U"\U00001F9E\U00001F96\U00001F9E" },
{ 0x00001F9F, U"\U00001F9F\U00001F97\U00001F9F" },
{ 0x00001FA0, U"\U00001FA8\U00001FA0\U00001FA8" },
{ 0x00001FA1, U"\U00001FA9\U00001FA1\U00001FA9" },
{ 0x00001FA2, U"\U00001FAA\U00001FA2\U00001FAA" },
{ 0x00001FA3, U"\U00001FAB\U00001FA3\U00001FAB" },
{ 0x00001FA4, U"\U00001FAC\U00001FA4\U00001FAC" },
{ 0x00001FA5, U"\U00001FAD\U00001FA5\U00001FAD" },
{ 0x00001FA6, U"\U00001FAE\U00001FA6\U00001FAE" },
{ 0x00001FA7, U"\U00001FAF\U00001FA7\U00001FAF" },
{ 0x00001FA8, U"\U00001FA8\U00001FA0\U00001FA8" },
{ 0x00001FA9, U"\U00001FA9\U00001FA1\U00001FA9" },
{ 0x00001FAA, U"\U00001FAA\U00001FA2\U00001FAA" },
{ 0x00001FAB, U"\U00001FAB\U00001FA3\U00001FAB" },
{ 0x00001FAC, U"\U00001FAC\U00001FA4\U00001FAC" },
{ 0x00001FAD, U"\U00001FAD\U00001FA5\U00001FAD" },
{ 0x00001FAE, U"\U00001FAE\U00001FA6\U00001FAE" },
{ 0x00001FAF, U"\U00001FAF\U00001FA7\U00001FAF" },
{ 0x00001FB0, U"\U00001FB8\U00001FB0\U00001FB8" },
{ 0x00001FB1, U"\U00001FB9\U00001FB1\U00001FB9" },
{ 0x00001FB3, U"\U00001FBC\U00001FB3\U00001FBC" },
{ 0x00001FB8, U"\U00001FB8\U00001FB0\U00001FB8" },
{ 0x00001FB9, U"\U00001FB9\U00001FB1\U00001FB9" },
{ 0x00001FBA, U"\U00001FBA\U00001F70\U00001FBA" },
{ 0x00001FBB, U"\U00001FBB\U00001F71\U00001FBB" },
{ 0x00001FBC, U"\U00001FBC\U00001FB3\U00001FBC" },
{ 0x00001FBE, U"\U00000399\U00001FBE\U00000399" },
{ 0x00001FC3, U"\U00001FCC\U00001FC3\U00001FCC" },
{ 0x00001FC8, U"\U00001FC8\U00001F72\U00001FC8" },
{ 0x00001FC9, U"\U00001FC9\U00001F73\U00001FC9" },
{ 0x00001FCA, U"\U00001FCA\U00001F74\U00001FCA" },
{ 0x00001FCB, U"\U00001FCB\U00001F75\U00001FCB" },
{ 0x00001FCC, U"\U00001FCC\U00001FC3\U00001FCC" },
{ 0x00001FD0, U"\U00001FD8\U00001FD0\U00001FD8" },
{ 0x00001FD1, U"\U00001FD9\U00001FD1\U00001FD9" },
{ 0x00001FD8, U"\U00001FD8\U00001FD0\U00001FD8" },
{ 0x00001FD9, U"\U00001FD9\U00001FD1\U00001FD9" },
{ 0x00001FDA, U"\U00001FDA\U00001F76\U00001FDA" },
{ 0x00001FDB, U"\U00001FDB\U00001F77\U00001FDB" },
{ 0x00001FE0, U"\U00001FE8\U00001FE0\U00001FE8" },
{ 0x00001FE1, U"\U00001FE9\U00001FE1\U00001FE9" },
{ 0x00001FE5, U"\U00001FEC\U00001FE5\U00001FEC" },
{ 0x00001FE8, U"\U00001FE8\U00001FE0\U00001FE8" },
{ 0x00001FE9, U"\U00001FE9\U00001FE1\U00001FE9" },
{ 0x00001FEA, U"\U00001FEA\U00001F7A\U00001FEA" },
{ 0x00001FEB, U"\U00001FEB\U00001F7B\U00001FEB" },
{ 0x00001FEC, U"\U00001FEC\U00001FE5\U00001FEC" },
{ 0x00001FF3, U"\U00001FFC\U00001FF3\U00001FFC" },
{ 0x00001FF8, U"\U00001FF8\U00001F78\U00001FF8" },
{ 0x00001FF9, U"\U00001FF9\U00001F79\U00001FF9" },
{ 0x00001FFA, U"\U00001FFA\U00001F7C\U00001FFA" },
{ 0x00001FFB, U"\U00001FFB\U00001F7D\U00001FFB" },
{ 0x00001FFC, U"\U00001FFC\U00001FF3\U00001FFC" },
{ 0x00002126, U"\U00002126\U000003C9\U00002126" },
{ 0x0000212A, U"\U0000212A\U0000006B\U0000212A" },
{ 0x0000212B, U"\U0000212B\U000000E5\U0000212B" },
{ 0x00002132, U"\U00002132\U0000214E\U00002132" },
{ 0x0000214E, U"\U00002132\U0000214E\U00002132" },
{ 0x00002160, U"\U00002160\U00002170\U00002160" },
{ 0x00002161, U"\U00002161\U00002171\U00002161" },
{ 0x00002162, U"\U00002162\U00002172\U00002162" },
{ 0x00002163, U"\U00002163\U00002173\U00002163" },
{ 0x00002164, U"\U00002164\U00002174\U00002164" },
{ 0x00002165, U"\U00002165\U00002175\U00002165" },
{ 0x00002166, U"\U00002166\U00002176\U00002166" },
{ 0x00002167, U"\U00002167\U00002177\U00002167" },
{ 0x00002168, U"\U00002168\U00002178\U00002168" },
{ 0x00002169, U"\U00002169\U00002179\U00002169" },
{ 0x0000216A, U"\U0000216A\U0000217A\U0000216A" },
{ 0x0000216B, U"\U0000216B\U0000217B\U0000216B" },
{ 0x0000216C, U"\U0000216C\U0000217C\U0000216C" },
{ 0x0000216D, U"\U0000216D\U0000217D\U0000216D" },
{ 0x0000216E, U"\U0000216E\U0000217E\U0000216E" },
{ 0x0000216F, U"\U0000216F\U0000217F\U0000216F" },
{ 0x00002170, U"\U00002160\U00002170\U00002160" },
{ 0x00002171, U"\U00002161\U00002171\U00002161" },
{ 0x00002172, U"\U00002162\U00002172\U00002162" },
{ 0x00002173, U"\U00002163\U00002173\U00002163" },
{ 0x00002174, U"\U00002164\U00002174\U00002164" },
{ 0x00002175, U"\U00002165\U00002175\U00002165" },
{ 0x00002176, U"\U00002166\U00002176\U00002166" },
{ 0x00002177, U"\U00002167\U00002177\U00002167" },
{ 0x00002178, U"\U00002168\U00002178\U00002168" },
{ 0x00002179, U"\U00002169\U00002179\U00002169" },
{ 0x0000217A, U"\U0000216A\U0000217A\U0000216A" },
{ 0x0000217B, U"\U0000216B\U0000217B\U0000216B" },
{ 0x0000217C, U"\U0000216C\U0000217C\U0000216C" },
{ 0x0000217D, U"\U0000216D\U0000217D\U0000216D" },
{ 0x0000217E, U"\U0000216E\U0000217E\U0000216E" },
{ 0x0000217F, U"\U0000216F\U0000217F\U0000216F" },
{ 0x00002183, U"\U00002183\U00002184\U00002183" },
{ 0x00002184, U"\U00002183\U00002184\U00002183" },
{ 0x000024B6, U"\U000024B6\U000024D0\U000024B6" },
{ 0x000024B7, U"\U000024B7\U000024D1\U000024B7" },
{ 0x000024B8, U"\U000024B8\U000024D2\U000024B8" },
{ 0x000024B9, U"\U000024B9\U000024D3\U000024B9" },
{ 0x000024BA, U"\U000024BA\U000024D4\U000024BA" },
{ 0x000024BB, U"\U000024BB\U000024D5\U000024BB" },
{ 0x000024BC, U"\U000024BC\U000024D6\U000024BC" },
{ 0x000024BD, U"\U000024BD\U000024D7\U000024BD" },
{ 0x000024BE, U"\U000024BE\U000024D8\U000024BE" },
{ 0x000024BF, U"\U000024BF\U000024D9\U000024BF" },
{ 0x000024C0, U"\U000024C0\U000024DA\U000024C0" },
{ 0x000024C1, U"\U000024C1\U000024DB\U000024C1" },
{ 0x000024C2, U"\U000024C2\U000024DC\U000024C2" },
{ 0x000024C3, U"\U000024C3\U000024DD\U000024C3" },
{ 0x000024C4, U"\U000024C4\U000024DE\U000024C4" },
{ 0x000024C5, U"\U000024C5\U000024DF\U000024C5" },
{ 0x000024C6, U"\U000024C6\U000024E0\U000024C6" },
{ 0x000024C7, U"\U000024C7\U000024E1\U000024C7" },
{ 0x000024C8, U"\U000024C8\U000024E2\U000024C8" },
{ 0x000024C9, U"\U000024C9\U000024E3\U000024C9" },
{ 0x000024CA, U"\U000024CA\U000024E4\U000024CA" },
{ 0x000024CB, U"\U000024CB\U000024E5\U000024CB" },
{ 0x000024CC, U"\U000024CC\U000024E6\U000024CC" },
{ 0x000024CD, U"\U000024CD\U000024E7\U000024CD" },
{ 0x000024CE, U"\U000024CE\U000024E8\U000024CE" },
{ 0x000024CF, U"\U000024CF\U000024E9\U000024CF" },
{ 0x000024D0, U"\U000024B6\U000024D0\U000024B6" },
{ 0x000024D1, U"\U000024B7\U000024D1\U000024B7" },
{ 0x000024D2, U"\U000024B8\U000024D2\U000024B8" },
{ 0x000024D3, U"\U000024B9\U000024D3\U000024B9" },
{ 0x000024D4, U"\U000024BA\U000024D4\U000024BA" },
{ 0x000024D5, U"\U000024BB\U000024D5\U000024BB" },
{ 0x000024D6, U"\U000024BC\U000024D6\U000024BC" },
{ 0x000024D7, U"\U000024BD\U000024D7\U000024BD" },
{ 0x000024D8, U"\U000024BE\U000024D8\U000024BE" },
{ 0x000024D9, U"\U000024BF\U000024D9\U000024BF" },
{ 0x000024DA, U"\U000024C0\U000024DA\U000024C0" },
{ 0x000024DB, U"\U000024C1\U000024DB\U000024C1" },
{ 0x000024DC, U"\U000024C2\U000024DC\U000024C2" },
{ 0x000024DD, U"\U000024C3\U000024DD\U000024C3" },
{ 0x000024DE, U"\U000024C4\U000024DE\U000024C4" },
{ 0x000024DF, U"\U000024C5\U000024DF\U000024C5" },
{ 0x000024E0, U"\U000024C6\U000024E0\U000024C6" },
{ 0x000024E1, U"\U000024C7\U000024E1\U000024C7" },
{ 0x000024E2, U"\U000024C8\U000024E2\U000024C8" },
{ 0x000024E3, U"\U000024C9\U000024E3\U000024C9" },
{ 0x000024E4, U"\U000024CA\U000024E4\U000024CA" },
{ 0x000024E5, U"\U000024CB\U000024E5\U000024CB" },
{ 0x000024E6, U"\U000024CC\U000024E6\U000024CC" },
{ 0x000024E7, U"\U000024CD\U000024E7\U000024CD" },
{ 0x000024E8, U"\U000024CE\U000024E8\U000024CE" },
{ 0x000024E9, U"\U000024CF\U000024E9\U000024CF" },
{ 0x00002C00, U"\U00002C00\U00002C30\U00002C00" },
{ 0x00002C01, U"\U00002C01\U00002C31\U00002C01" },
{ 0x00002C02, U"\U00002C02\U00002C32\U00002C02" },
{ 0x00002C03, U"\U00002C03\U00002C33\U00002C03" },
{ 0x00002C04, U"\U00002C04\U00002C34\U00002C04" },
{ 0x00002C05, U"\U00002C05\U00002C35\U00002C05" },
{ 0x00002C06, U"\U00002C06\U00002C36\U00002C06" },
{ 0x00002C07, U"\U00002C07\U00002C37\U00002C07" },
{ 0x00002C08, U"\U00002C08\U00002C38\U00002C08" },
{ 0x00002C09, U"\U00002C09\U00002C39\U00002C09" },
{ 0x00002C0A, U"\U00002C0A\U00002C3A\U00002C0A" },
{ 0x00002C0B, U"\U00002C0B\U00002C3B\U00002C0B" },
{ 0x00002C0C, U"\U00002C0C\U00002C3C\U00002C0C" },
{ 0x00002C0D, U"\U00002C0D\U00002C3D\U00002C0D" },
{ 0x00002C0E, U"\U00002C0E\U00002C3E\U00002C0E" },
{ 0x00002C0F, U"\U00002C0F\U00002C3F\U00002C0F" },
{ 0x00002C10, U"\U00002C10\U00002C40\U00002C10" },
{ 0x00002C11, U"\U00002C11\U00002C41\U00002C11" },
{ 0x00002C12, U"\U00002C12\U00002C42\U00002C12" },
{ 0x00002C13, U"\U00002C13\U00002C43\U00002C13" },
{ 0x00002C14, U"\U00002C14\U00002C44\U00002C14" },
{ 0x00002C15, U"\U00002C15\U00002C45\U00002C15" },
{ 0x00002C16, U"\U00002C16\U00002C46\U00002C16" },
{ 0x00002C17, U"\U00002C17\U00002C47\U00002C17" },
{ 0x00002C18, U"\U00002C18\U00002C48\U00002C18" },
{ 0x00002C19, U"\U00002C19\U00002C49\U00002C19" },
{ 0x00002C1A, U"\U00002C1A\U00002C4A\U00002C1A" },
{ 0x00002C1B, U"\U00002C1B\U00002C4B\U00002C1B" },
{ 0x00002C1C, U"\U00002C1C\U00002C4C\U00002C1C" },
{ 0x00002C1D, U"\U00002C1D\U00002C4D\U00002C1D" },
{ 0x00002C1E, U"\U00002C1E\U00002C4E\U00002C1E" },
{ 0x00002C1F, U"\U00002C1F\U00002C4F\U00002C1F" },
{ 0x00002C20, U"\U00002C20\U00002C50\U00002C20" },
{ 0x00002C21, U"\U00002C21\U00002C51\U00002C21" },
{ 0x00002C22, U"\U00002C22\U00002C52\U00002C22" },
{ 0x00002C23, U"\U00002C23\U00002C53\U00002C23" },
{ 0x00002C24, U"\U00002C24\U00002C54\U00002C24" },
{ 0x00002C25, U"\U00002C25\U00002C55\U00002C25" },
{ 0x00002C26, U"\U00002C26\U00002C56\U00002C26" },
{ 0x00002C27, U"\U00002C27\U00002C57\U00002C27" },
{ 0x00002C28, U"\U00002C28\U00002C58\U00002C28" },
{ 0x00002C29, U"\U00002C29\U00002C59\U00002C29" },
{ 0x00002C2A, U"\U00002C2A\U00002C5A\U00002C2A" },
{ 0x00002C2B, U"\U00002C2B\U00002C5B\U00002C2B" },
{ 0x00002C2C, U"\U00002C2C\U00002C5C\U00002C2C" },
{ 0x00002C2D, U"\U00002C2D\U00002C5D\U00002C2D" },
{ 0x00002C2E, U"\U00002C2E\U00002C5E\U00002C2E" },
{ 0x00002C30, U"\U00002C00\U00002C30\U00002C00" },
{ 0x00002C31, U"\U00002C01\U00002C31\U00002C01" },
{ 0x00002C32, U"\U00002C02\U00002C32\U00002C02" },
{ 0x00002C33, U"\U00002C03\U00002C33\U00002C03" },
{ 0x00002C34, U"\U00002C04\U00002C34\U00002C04" },
{ 0x00002C35, U"\U00002C05\U00002C35\U00002C05" },
{ 0x00002C36, U"\U00002C06\U00002C36\U00002C06" },
{ 0x00002C37, U"\U00002C07\U00002C37\U00002C07" },
{ 0x00002C38, U"\U00002C08\U00002C38\U00002C08" },
{ 0x00002C39, U"\U00002C09\U00002C39\U00002C09" },
{ 0x00002C3A, U"\U00002C0A\U00002C3A\U00002C0A" },
{ 0x00002C3B, U"\U00002C0B\U00002C3B\U00002C0B" },
{ 0x00002C3C, U"\U00002C0C\U00002C3C\U00002C0C" },
{ 0x00002C3D, U"\U00002C0D\U00002C3D\U00002C0D" },
{ 0x00002C3E, U"\U00002C0E\U00002C3E\U00002C0E" },
{ 0x00002C3F, U"\U00002C0F\U00002C3F\U00002C0F" },
{ 0x00002C40, U"\U00002C10\U00002C40\U00002C10" },
{ 0x00002C41, U"\U00002C11\U00002C41\U00002C11" },
{ 0x00002C42, U"\U00002C12\U00002C42\U00002C12" },
{ 0x00002C43, U"\U00002C13\U00002C43\U00002C13" },
{ 0x00002C44, U"\U00002C14\U00002C44\U00002C14" },
{ 0x00002C45, U"\U00002C15\U00002C45\U00002C15" },
{ 0x00002C46, U"\U00002C16\U00002C46\U00002C16" },
{ 0x00002C47, U"\U00002C17\U00002C47\U00002C17" },
{ 0x00002C48, U"\U00002C18\U00002C48\U00002C18" },
{ 0x00002C49, U"\U00002C19\U00002C49\U00002C19" },
{ 0x00002C4A, U"\U00002C1A\U00002C4A\U00002C1A" },
{ 0x00002C4B, U"\U00002C1B\U00002C4B\U00002C1B" },
{ 0x00002C4C, U"\U00002C1C\U00002C4C\U00002C1C" },
{ 0x00002C4D, U"\U00002C1D\U00002C4D\U00002C1D" },
{ 0x00002C4E, U"\U00002C1E\U00002C4E\U00002C1E" },
{ 0x00002C4F, U"\U00002C1F\U00002C4F\U00002C1F" },
{ 0x00002C50, U"\U00002C20\U00002C50\U00002C20" },
{ 0x00002C51, U"\U00002C21\U00002C51\U00002C21" },
{ 0x00002C52, U"\U00002C22\U00002C52\U00002C22" },
{ 0x00002C53, U"\U00002C23\U00002C53\U00002C23" },
{ 0x00002C54, U"\U00002C24\U00002C54\U00002C24" },
{ 0x00002C55, U"\U00002C25\U00002C55\U00002C25" },
{ 0x00002C56, U"\U00002C26\U00002C56\U00002C26" },
{ 0x00002C57, U"\U00002C27\U00002C57\U00002C27" },
{ 0x00002C58, U"\U00002C28\U00002C58\U00002C28" },
{ 0x00002C59, U"\U00002C29\U00002C59\U00002C29" },
{ 0x00002C5A, U"\U00002C2A\U00002C5A\U00002C2A" },
{ 0x00002C5B, U"\U00002C2B\U00002C5B\U00002C2B" },
{ 0x00002C5C, U"\U00002C2C\U00002C5C\U00002C2C" },
{ 0x00002C5D, U"\U00002C2D\U00002C5D\U00002C2D" },
{ 0x00002C5E, U"\U00002C2E\U00002C5E\U00002C2E" },
{ 0x00002C60, U"\U00002C60\U00002C61\U00002C60" },
{ 0x00002C61, U"\U00002C60\U00002C61\U00002C60" },
{ 0x00002C62, U"\U00002C62\U0000026B\U00002C62" },
{ 0x00002C63, U"\U00002C63\U00001D7D\U00002C63" },
{ 0x00002C64, U"\U00002C64\U0000027D\U00002C64" },
{ 0x00002C65, U"\U0000023A\U00002C65\U0000023A" },
{ 0x00002C66, U"\U0000023E\U00002C66\U0000023E" },
{ 0x00002C67, U"\U00002C67\U00002C68\U00002C67" },
{ 0x00002C68, U"\U00002C67\U00002C68\U00002C67" },
{ 0x00002C69, U"\U00002C69\U00002C6A\U00002C69" },
{ 0x00002C6A, U"\U00002C69\U00002C6A\U00002C69" },
{ 0x00002C6B, U"\U00002C6B\U00002C6C\U00002C6B" },
{ 0x00002C6C, U"\U00002C6B\U00002C6C\U00002C6B" },
{ 0x00002C6D, U"\U00002C6D\U00000251\U00002C6D" },
{ 0x00002C6E, U"\U00002C6E\U00000271\U00002C6E" },
{ 0x00002C6F, U"\U00002C6F\U00000250\U00002C6F" },
{ 0x00002C70, U"\U00002C70\U00000252\U00002C70" },
{ 0x00002C72, U"\U00002C72\U00002C73\U00002C72" },
{ 0x00002C73, U"\U00002C72\U00002C73\U00002C72" },
{ 0x00002C75, U"\U00002C75\U00002C76\U00002C75" },
{ 0x00002C76, U"\U00002C75\U00002C76\U00002C75" },
{ 0x00002C7E, U"\U00002C7E\U0000023F\U00002C7E" },
{ 0x00002C7F, U"\U00002C7F\U00000240\U00002C7F" },
{ 0x00002C80, U"\U00002C80\U00002C81\U00002C80" },
{ 0x00002C81, U"\U00002C80\U00002C81\U00002C80" },
{ 0x00002C82, U"\U00002C82\U00002C83\U00002C82" },
{ 0x00002C83, U"\U00002C82\U00002C83\U00002C82" },
{ 0x00002C84, U"\U00002C84\U00002C85\U00002C84" },
{ 0x00002C85, U"\U00002C84\U00002C85\U00002C84" },
{ 0x00002C86, U"\U00002C86\U00002C87\U00002C86" },
{ 0x00002C87, U"\U00002C86\U00002C87\U00002C86" },
{ 0x00002C88, U"\U00002C88\U00002C89\U00002C88" },
{ 0x00002C89, U"\U00002C88\U00002C89\U00002C88" },
{ 0x00002C8A, U"\U00002C8A\U00002C8B\U00002C8A" },
{ 0x00002C8B, U"\U00002C8A\U00002C8B\U00002C8A" },
{ 0x00002C8C, U"\U00002C8C\U00002C8D\U00002C8C" },
{ 0x00002C8D, U"\U00002C8C\U00002C8D\U00002C8C" },
{ 0x00002C8E, U"\U00002C8E\U00002C8F\U00002C8E" },
{ 0x00002C8F, U"\U00002C8E\U00002C8F\U00002C8E" },
{ 0x00002C90, U"\U00002C90\U00002C91\U00002C90" },
{ 0x00002C91, U"\U00002C90\U00002C91\U00002C90" },
{ 0x00002C92, U"\U00002C92\U00002C93\U00002C92" },
{ 0x00002C93, U"\U00002C92\U00002C93\U00002C92" },
{ 0x00002C94, U"\U00002C94\U00002C95\U00002C94" },
{ 0x00002C95, U"\U00002C94\U00002C95\U00002C94" },
{ 0x00002C96, U"\U00002C96\U00002C97\U00002C96" },
{ 0x00002C97, U"\U00002C96\U00002C97\U00002C96" },
{ 0x00002C98, U"\U00002C98\U00002C99\U00002C98" },
{ 0x00002C99, U"\U00002C98\U00002C99\U00002C98" },
{ 0x00002C9A, U"\U00002C9A\U00002C9B\U00002C9A" },
{ 0x00002C9B, U"\U00002C9A\U00002C9B\U00002C9A" },
{ 0x00002C9C, U"\U00002C9C\U00002C9D\U00002C9C" },
{ 0x00002C9D, U"\U00002C9C\U00002C9D\U00002C9C" },
{ 0x00002C9E, U"\U00002C9E\U00002C9F\U00002C9E" },
{ 0x00002C9F, U"\U00002C9E\U00002C9F\U00002C9E" },
{ 0x00002CA0, U"\U00002CA0\U00002CA1\U00002CA0" },
{ 0x00002CA1, U"\U00002CA0\U00002CA1\U00002CA0" },
{ 0x00002CA2, U"\U00002CA2\U00002CA3\U00002CA2" },
{ 0x00002CA3, U"\U00002CA2\U00002CA3\U00002CA2" },
{ 0x00002CA4, U"\U00002CA4\U00002CA5\U00002CA4" },
{ 0x00002CA5, U"\U00002CA4\U00002CA5\U00002CA4" },
{ 0x00002CA6, U"\U00002CA6\U00002CA7\U00002CA6" },
{ 0x00002CA7, U"\U00002CA6\U00002CA7\U00002CA6" },
{ 0x00002CA8, U"\U00002CA8\U00002CA9\U00002CA8" },
{ 0x00002CA9, U"\U00002CA8\U00002CA9\U00002CA8" },
{ 0x00002CAA, U"\U00002CAA\U00002CAB\U00002CAA" },
{ 0x00002CAB, U"\U00002CAA\U00002CAB\U00002CAA" },
{ 0x00002CAC, U"\U00002CAC\U00002CAD\U00002CAC" },
{ 0x00002CAD, U"\U00002CAC\U00002CAD\U00002CAC" },
{ 0x00002CAE, U"\U00002CAE\U00002CAF\U00002CAE" },
{ 0x00002CAF, U"\U00002CAE\U00002CAF\U00002CAE" },
{ 0x00002CB0, U"\U00002CB0\U00002CB1\U00002CB0" },
{ 0x00002CB1, U"\U00002CB0\U00002CB1\U00002CB0" },
{ 0x00002CB2, U"\U00002CB2\U00002CB3\U00002CB2" },
{ 0x00002CB3, U"\U00002CB2\U00002CB3\U00002CB2" },
{ 0x00002CB4, U"\U00002CB4\U00002CB5\U00002CB4" },
{ 0x00002CB5, U"\U00002CB4\U00002CB5\U00002CB4" },
{ 0x00002CB6, U"\U00002CB6\U00002CB7\U00002CB6" },
{ 0x00002CB7, U"\U00002CB6\U00002CB7\U00002CB6" },
{ 0x00002CB8, U"\U00002CB8\U00002CB9\U00002CB8" },
{ 0x00002CB9, U"\U00002CB8\U00002CB9\U00002CB8" },
{ 0x00002CBA, U"\U00002CBA\U00002CBB\U00002CBA" },
{ 0x00002CBB, U"\U00002CBA\U00002CBB\U00002CBA" },
{ 0x00002CBC, U"\U00002CBC\U00002CBD\U00002CBC" },
{ 0x00002CBD, U"\U00002CBC\U00002CBD\U00002CBC" },
{ 0x00002CBE, U"\U00002CBE\U00002CBF\U00002CBE" },
{ 0x00002CBF, U"\U00002CBE\U00002CBF\U00002CBE" },
{ 0x00002CC0, U"\U00002CC0\U00002CC1\U00002CC0" },
{ 0x00002CC1, U"\U00002CC0\U00002CC1\U00002CC0" },
{ 0x00002CC2, U"\U00002CC2\U00002CC3\U00002CC2" },
{ 0x00002CC3, U"\U00002CC2\U00002CC3\U00002CC2" },
{ 0x00002CC4, U"\U00002CC4\U00002CC5\U00002CC4" },
{ 0x00002CC5, U"\U00002CC4\U00002CC5\U00002CC4" },
{ 0x00002CC6, U"\U00002CC6\U00002CC7\U00002CC6" },
{ 0x00002CC7, U"\U00002CC6\U00002CC7\U00002CC6" },
{ 0x00002CC8, U"\U00002CC8\U00002CC9\U00002CC8" },
{ 0x00002CC9, U"\U00002CC8\U00002CC9\U00002CC8" },
{ 0x00002CCA, U"\U00002CCA\U00002CCB\U00002CCA" },
{ 0x00002CCB, U"\U00002CCA\U00002CCB\U00002CCA" },
{ 0x00002CCC, U"\U00002CCC\U00002CCD\U00002CCC" },
{ 0x00002CCD, U"\U00002CCC\U00002CCD\U00002CCC" },
{ 0x00002CCE, U"\U00002CCE\U00002CCF\U00002CCE" },
{ 0x00002CCF, U"\U00002CCE\U00002CCF\U00002CCE" },
{ 0x00002CD0, U"\U00002CD0\U00002CD1\U00002CD0" },
{ 0x00002CD1, U"\U00002CD0\U00002CD1\U00002CD0" },
{ 0x00002CD2, U"\U00002CD2\U00002CD3\U00002CD2" },
{ 0x00002CD3, U"\U00002CD2\U00002CD3\U00002CD2" },
{ 0x00002CD4, U"\U00002CD4\U00002CD5\U00002CD4" },
{ 0x00002CD5, U"\U00002CD4\U00002CD5\U00002CD4" },
{ 0x00002CD6, U"\U00002CD6\U00002CD7\U00002CD6" },
{ 0x00002CD7, U"\U00002CD6\U00002CD7\U00002CD6" },
{ 0x00002CD8, U"\U00002CD8\U00002CD9\U00002CD8" },
{ 0x00002CD9, U"\U00002CD8\U00002CD9\U00002CD8" },
{ 0x00002CDA, U"\U00002CDA\U00002CDB\U00002CDA" },
{ 0x00002CDB, U"\U00002CDA\U00002CDB\U00002CDA" },
{ 0x00002CDC, U"\U00002CDC\U00002CDD\U00002CDC" },
{ 0x00002CDD, U"\U00002CDC\U00002CDD\U00002CDC" },
{ 0x00002CDE, U"\U00002CDE\U00002CDF\U00002CDE" },
{ 0x00002CDF, U"\U00002CDE\U00002CDF\U00002CDE" },
{ 0x00002CE0, U"\U00002CE0\U00002CE1\U00002CE0" },
{ 0x00002CE1, U"\U00002CE0\U00002CE1\U00002CE0" },
{ 0x00002CE2, U"\U00002CE2\U00002CE3\U00002CE2" },
{ 0x00002CE3, U"\U00002CE2\U00002CE3\U00002CE2" },
{ 0x00002CEB, U"\U00002CEB\U00002CEC\U00002CEB" },
{ 0x00002CEC, U"\U00002CEB\U00002CEC\U00002CEB" },
{ 0x00002CED, U"\U00002CED\U00002CEE\U00002CED" },
{ 0x00002CEE, U"\U00002CED\U00002CEE\U00002CED" },
{ 0x00002CF2, U"\U00002CF2\U00002CF3\U00002CF2" },
{ 0x00002CF3, U"\U00002CF2\U00002CF3\U00002CF2" },
{ 0x00002D00, U"\U000010A0\U00002D00\U000010A0" },
{ 0x00002D01, U"\U000010A1\U00002D01\U000010A1" },
{ 0x00002D02, U"\U000010A2\U00002D02\U000010A2" },
{ 0x00002D03, U"\U000010A3\U00002D03\U000010A3" },
{ 0x00002D04, U"\U000010A4\U00002D04\U000010A4" },
{ 0x00002D05, U"\U000010A5\U00002D05\U000010A5" },
{ 0x00002D06, U"\U000010A6\U00002D06\U000010A6" },
{ 0x00002D07, U"\U000010A7\U00002D07\U000010A7" },
{ 0x00002D08, U"\U000010A8\U00002D08\U000010A8" },
{ 0x00002D09, U"\U000010A9\U00002D09\U000010A9" },
{ 0x00002D0A, U"\U000010AA\U00002D0A\U000010AA" },
{ 0x00002D0B, U"\U000010AB\U00002D0B\U000010AB" },
{ 0x00002D0C, U"\U000010AC\U00002D0C\U000010AC" },
{ 0x00002D0D, U"\U000010AD\U00002D0D\U000010AD" },
{ 0x00002D0E, U"\U000010AE\U00002D0E\U000010AE" },
{ 0x00002D0F, U"\U000010AF\U00002D0F\U000010AF" },
{ 0x00002D10, U"\U000010B0\U00002D10\U000010B0" },
{ 0x00002D11, U"\U000010B1\U00002D11\U000010B1" },
{ 0x00002D12, U"\U000010B2\U00002D12\U000010B2" },
{ 0x00002D13, U"\U000010B3\U00002D13\U000010B3" },
{ 0x00002D14, U"\U000010B4\U00002D14\U000010B4" },
{ 0x00002D15, U"\U000010B5\U00002D15\U000010B5" },
{ 0x00002D16, U"\U000010B6\U00002D16\U000010B6" },
{ 0x00002D17, U"\U000010B7\U00002D17\U000010B7" },
{ 0x00002D18, U"\U000010B8\U00002D18\U000010B8" },
{ 0x00002D19, U"\U000010B9\U00002D19\U000010B9" },
{ 0x00002D1A, U"\U000010BA\U00002D1A\U000010BA" },
{ 0x00002D1B, U"\U000010BB\U00002D1B\U000010BB" },
{ 0x00002D1C, U"\U000010BC\U00002D1C\U000010BC" },
{ 0x00002D1D, U"\U000010BD\U00002D1D\U000010BD" },
{ 0x00002D1E, U"\U000010BE\U00002D1E\U000010BE" },
{ 0x00002D1F, U"\U000010BF\U00002D1F\U000010BF" },
{ 0x00002D20, U"\U000010C0\U00002D20\U000010C0" },
{ 0x00002D21, U"\U000010C1\U00002D21\U000010C1" },
{ 0x00002D22, U"\U000010C2\U00002D22\U000010C2" },
{ 0x00002D23, U"\U000010C3\U00002D23\U000010C3" },
{ 0x00002D24, U"\U000010C4\U00002D24\U000010C4" },
{ 0x00002D25, U"\U000010C5\U00002D25\U000010C5" },
{ 0x00002D27, U"\U000010C7\U00002D27\U000010C7" },
{ 0x00002D2D, U"\U000010CD\U00002D2D\U000010CD" },
{ 0x0000A640, U"\U0000A640\U0000A641\U0000A640" },
{ 0x0000A641, U"\U0000A640\U0000A641\U0000A640" },
{ 0x0000A642, U"\U0000A642\U0000A643\U0000A642" },
{ 0x0000A643, U"\U0000A642\U0000A643\U0000A642" },
{ 0x0000A644, U"\U0000A644\U0000A645\U0000A644" },
{ 0x0000A645, U"\U0000A644\U0000A645\U0000A644" },
{ 0x0000A646, U"\U0000A646\U0000A647\U0000A646" },
{ 0x0000A647, U"\U0000A646\U0000A647\U0000A646" },
{ 0x0000A648, U"\U0000A648\U0000A649\U0000A648" },
{ 0x0000A649, U"\U0000A648\U0000A649\U0000A648" },
{ 0x0000A64A, U"\U0000A64A\U0000A64B\U0000A64A" },
{ 0x0000A64B, U"\U0000A64A\U0000A64B\U0000A64A" },
{ 0x0000A64C, U"\U0000A64C\U0000A64D\U0000A64C" },
{ 0x0000A64D, U"\U0000A64C\U0000A64D\U0000A64C" },
{ 0x0000A64E, U"\U0000A64E\U0000A64F\U0000A64E" },
{ 0x0000A64F, U"\U0000A64E\U0000A64F\U0000A64E" },
{ 0x0000A650, U"\U0000A650\U0000A651\U0000A650" },
{ 0x0000A651, U"\U0000A650\U0000A651\U0000A650" },
{ 0x0000A652, U"\U0000A652\U0000A653\U0000A652" },
{ 0x0000A653, U"\U0000A652\U0000A653\U0000A652" },
{ 0x0000A654, U"\U0000A654\U0000A655\U0000A654" },
{ 0x0000A655, U"\U0000A654\U0000A655\U0000A654" },
{ 0x0000A656, U"\U0000A656\U0000A657\U0000A656" },
{ 0x0000A657, U"\U0000A656\U0000A657\U0000A656" },
{ 0x0000A658, U"\U0000A658\U0000A659\U0000A658" },
{ 0x0000A659, U"\U0000A658\U0000A659\U0000A658" },
{ 0x0000A65A, U"\U0000A65A\U0000A65B\U0000A65A" },
{ 0x0000A65B, U"\U0000A65A\U0000A65B\U0000A65A" },
{ 0x0000A65C, U"\U0000A65C\U0000A65D\U0000A65C" },
{ 0x0000A65D, U"\U0000A65C\U0000A65D\U0000A65C" },
{ 0x0000A65E, U"\U0000A65E\U0000A65F\U0000A65E" },
{ 0x0000A65F, U"\U0000A65E\U0000A65F\U0000A65E" },
{ 0x0000A660, U"\U0000A660\U0000A661\U0000A660" },
{ 0x0000A661, U"\U0000A660\U0000A661\U0000A660" },
{ 0x0000A662, U"\U0000A662\U0000A663\U0000A662" },
{ 0x0000A663, U"\U0000A662\U0000A663\U0000A662" },
{ 0x0000A664, U"\U0000A664\U0000A665\U0000A664" },
{ 0x0000A665, U"\U0000A664\U0000A665\U0000A664" },
{ 0x0000A666, U"\U0000A666\U0000A667\U0000A666" },
{ 0x0000A667, U"\U0000A666\U0000A667\U0000A666" },
{ 0x0000A668, U"\U0000A668\U0000A669\U0000A668" },
{ 0x0000A669, U"\U0000A668\U0000A669\U0000A668" },
{ 0x0000A66A, U"\U0000A66A\U0000A66B\U0000A66A" },
{ 0x0000A66B, U"\U0000A66A\U0000A66B\U0000A66A" },
{ 0x0000A66C, U"\U0000A66C\U0000A66D\U0000A66C" },
{ 0x0000A66D, U"\U0000A66C\U0000A66D\U0000A66C" },
{ 0x0000A680, U"\U0000A680\U0000A681\U0000A680" },
{ 0x0000A681, U"\U0000A680\U0000A681\U0000A680" },
{ 0x0000A682, U"\U0000A682\U0000A683\U0000A682" },
{ 0x0000A683, U"\U0000A682\U0000A683\U0000A682" },
{ 0x0000A684, U"\U0000A684\U0000A685\U0000A684" },
{ 0x0000A685, U"\U0000A684\U0000A685\U0000A684" },
{ 0x0000A686, U"\U0000A686\U0000A687\U0000A686" },
{ 0x0000A687, U"\U0000A686\U0000A687\U0000A686" },
{ 0x0000A688, U"\U0000A688\U0000A689\U0000A688" },
{ 0x0000A689, U"\U0000A688\U0000A689\U0000A688" },
{ 0x0000A68A, U"\U0000A68A\U0000A68B\U0000A68A" },
{ 0x0000A68B, U"\U0000A68A\U0000A68B\U0000A68A" },
{ 0x0000A68C, U"\U0000A68C\U0000A68D\U0000A68C" },
{ 0x0000A68D, U"\U0000A68C\U0000A68D\U0000A68C" },
{ 0x0000A68E, U"\U0000A68E\U0000A68F\U0000A68E" },
{ 0x0000A68F, U"\U0000A68E\U0000A68F\U0000A68E" },
{ 0x0000A690, U"\U0000A690\U0000A691\U0000A690" },
{ 0x0000A691, U"\U0000A690\U0000A691\U0000A690" },
{ 0x0000A692, U"\U0000A692\U0000A693\U0000A692" },
{ 0x0000A693, U"\U0000A692\U0000A693\U0000A692" },
{ 0x0000A694, U"\U0000A694\U0000A695\U0000A694" },
{ 0x0000A695, U"\U0000A694\U0000A695\U0000A694" },
{ 0x0000A696, U"\U0000A696\U0000A697\U0000A696" },
{ 0x0000A697, U"\U0000A696\U0000A697\U0000A696" },
{ 0x0000A698, U"\U0000A698\U0000A699\U0000A698" },
{ 0x0000A699, U"\U0000A698\U0000A699\U0000A698" },
{ 0x0000A69A, U"\U0000A69A\U0000A69B\U0000A69A" },
{ 0x0000A69B, U"\U0000A69A\U0000A69B\U0000A69A" },
{ 0x0000A722, U"\U0000A722\U0000A723\U0000A722" },
{ 0x0000A723, U"\U0000A722\U0000A723\U0000A722" },
{ 0x0000A724, U"\U0000A724\U0000A725\U0000A724" },
{ 0x0000A725, U"\U0000A724\U0000A725\U0000A724" },
{ 0x0000A726, U"\U0000A726\U0000A727\U0000A726" },
{ 0x0000A727, U"\U0000A726\U0000A727\U0000A726" },
{ 0x0000A728, U"\U0000A728\U0000A729\U0000A728" },
{ 0x0000A729, U"\U0000A728\U0000A729\U0000A728" },
{ 0x0000A72A, U"\U0000A72A\U0000A72B\U0000A72A" },
{ 0x0000A72B, U"\U0000A72A\U0000A72B\U0000A72A" },
{ 0x0000A72C, U"\U0000A72C\U0000A72D\U0000A72C" },
{ 0x0000A72D, U"\U0000A72C\U0000A72D\U0000A72C" },
{ 0x0000A72E, U"\U0000A72E\U0000A72F\U0000A72E" },
{ 0x0000A72F, U"\U0000A72E\U0000A72F\U0000A72E" },
{ 0x0000A732, U"\U0000A732\U0000A733\U0000A732" },
{ 0x0000A733, U"\U0000A732\U0000A733\U0000A732" },
{ 0x0000A734, U"\U0000A734\U0000A735\U0000A734" },
{ 0x0000A735, U"\U0000A734\U0000A735\U0000A734" },
{ 0x0000A736, U"\U0000A736\U0000A737\U0000A736" },
{ 0x0000A737, U"\U0000A736\U0000A737\U0000A736" },
{ 0x0000A738, U"\U0000A738\U0000A739\U0000A738" },
{ 0x0000A739, U"\U0000A738\U0000A739\U0000A738" },
{ 0x0000A73A, U"\U0000A73A\U0000A73B\U0000A73A" },
{ 0x0000A73B, U"\U0000A73A\U0000A73B\U0000A73A" },
{ 0x0000A73C, U"\U0000A73C\U0000A73D\U0000A73C" },
{ 0x0000A73D, U"\U0000A73C\U0000A73D\U0000A73C" },
{ 0x0000A73E, U"\U0000A73E\U0000A73F\U0000A73E" },
{ 0x0000A73F, U"\U0000A73E\U0000A73F\U0000A73E" },
{ 0x0000A740, U"\U0000A740\U0000A741\U0000A740" },
{ 0x0000A741, U"\U0000A740\U0000A741\U0000A740" },
{ 0x0000A742, U"\U0000A742\U0000A743\U0000A742" },
{ 0x0000A743, U"\U0000A742\U0000A743\U0000A742" },
{ 0x0000A744, U"\U0000A744\U0000A745\U0000A744" },
{ 0x0000A745, U"\U0000A744\U0000A745\U0000A744" },
{ 0x0000A746, U"\U0000A746\U0000A747\U0000A746" },
{ 0x0000A747, U"\U0000A746\U0000A747\U0000A746" },
{ 0x0000A748, U"\U0000A748\U0000A749\U0000A748" },
{ 0x0000A749, U"\U0000A748\U0000A749\U0000A748" },
{ 0x0000A74A, U"\U0000A74A\U0000A74B\U0000A74A" },
{ 0x0000A74B, U"\U0000A74A\U0000A74B\U0000A74A" },
{ 0x0000A74C, U"\U0000A74C\U0000A74D\U0000A74C" },
{ 0x0000A74D, U"\U0000A74C\U0000A74D\U0000A74C" },
{ 0x0000A74E, U"\U0000A74E\U0000A74F\U0000A74E" },
{ 0x0000A74F, U"\U0000A74E\U0000A74F\U0000A74E" },
{ 0x0000A750, U"\U0000A750\U0000A751\U0000A750" },
{ 0x0000A751, U"\U0000A750\U0000A751\U0000A750" },
{ 0x0000A752, U"\U0000A752\U0000A753\U0000A752" },
{ 0x0000A753, U"\U0000A752\U0000A753\U0000A752" },
{ 0x0000A754, U"\U0000A754\U0000A755\U0000A754" },
{ 0x0000A755, U"\U0000A754\U0000A755\U0000A754" },
{ 0x0000A756, U"\U0000A756\U0000A757\U0000A756" },
{ 0x0000A757, U"\U0000A756\U0000A757\U0000A756" },
{ 0x0000A758, U"\U0000A758\U0000A759\U0000A758" },
{ 0x0000A759, U"\U0000A758\U0000A759\U0000A758" },
{ 0x0000A75A, U"\U0000A75A\U0000A75B\U0000A75A" },
{ 0x0000A75B, U"\U0000A75A\U0000A75B\U0000A75A" },
{ 0x0000A75C, U"\U0000A75C\U0000A75D\U0000A75C" },
{ 0x0000A75D, U"\U0000A75C\U0000A75D\U0000A75C" },
{ 0x0000A75E, U"\U0000A75E\U0000A75F\U0000A75E" },
{ 0x0000A75F, U"\U0000A75E\U0000A75F\U0000A75E" },
{ 0x0000A760, U"\U0000A760\U0000A761\U0000A760" },
{ 0x0000A761, U"\U0000A760\U0000A761\U0000A760" },
{ 0x0000A762, U"\U0000A762\U0000A763\U0000A762" },
{ 0x0000A763, U"\U0000A762\U0000A763\U0000A762" },
{ 0x0000A764, U"\U0000A764\U0000A765\U0000A764" },
{ 0x0000A765, U"\U0000A764\U0000A765\U0000A764" },
{ 0x0000A766, U"\U0000A766\U0000A767\U0000A766" },
{ 0x0000A767, U"\U0000A766\U0000A767\U0000A766" },
{ 0x0000A768, U"\U0000A768\U0000A769\U0000A768" },
{ 0x0000A769, U"\U0000A768\U0000A769\U0000A768" },
{ 0x0000A76A, U"\U0000A76A\U0000A76B\U0000A76A" },
{ 0x0000A76B, U"\U0000A76A\U0000A76B\U0000A76A" },
{ 0x0000A76C, U"\U0000A76C\U0000A76D\U0000A76C" },
{ 0x0000A76D, U"\U0000A76C\U0000A76D\U0000A76C" },
{ 0x0000A76E, U"\U0000A76E\U0000A76F\U0000A76E" },
{ 0x0000A76F, U"\U0000A76E\U0000A76F\U0000A76E" },
{ 0x0000A779, U"\U0000A779\U0000A77A\U0000A779" },
{ 0x0000A77A, U"\U0000A779\U0000A77A\U0000A779" },
{ 0x0000A77B, U"\U0000A77B\U0000A77C\U0000A77B" },
{ 0x0000A77C, U"\U0000A77B\U0000A77C\U0000A77B" },
{ 0x0000A77D, U"\U0000A77D\U00001D79\U0000A77D" },
{ 0x0000A77E, U"\U0000A77E\U0000A77F\U0000A77E" },
{ 0x0000A77F, U"\U0000A77E\U0000A77F\U0000A77E" },
{ 0x0000A780, U"\U0000A780\U0000A781\U0000A780" },
{ 0x0000A781, U"\U0000A780\U0000A781\U0000A780" },
{ 0x0000A782, U"\U0000A782\U0000A783\U0000A782" },
{ 0x0000A783, U"\U0000A782\U0000A783\U0000A782" },
{ 0x0000A784, U"\U0000A784\U0000A785\U0000A784" },
{ 0x0000A785, U"\U0000A784\U0000A785\U0000A784" },
{ 0x0000A786, U"\U0000A786\U0000A787\U0000A786" },
{ 0x0000A787, U"\U0000A786\U0000A787\U0000A786" },
{ 0x0000A78B, U"\U0000A78B\U0000A78C\U0000A78B" },
{ 0x0000A78C, U"\U0000A78B\U0000A78C\U0000A78B" },
{ 0x0000A78D, U"\U0000A78D\U00000265\U0000A78D" },
{ 0x0000A790, U"\U0000A790\U0000A791\U0000A790" },
{ 0x0000A791, U"\U0000A790\U0000A791\U0000A790" },
{ 0x0000A792, U"\U0000A792\U0000A793\U0000A792" },
{ 0x0000A793, U"\U0000A792\U0000A793\U0000A792" },
{ 0x0000A794, U"\U0000A7C4\U0000A794\U0000A7C4" },
{ 0x0000A796, U"\U0000A796\U0000A797\U0000A796" },
{ 0x0000A797, U"\U0000A796\U0000A797\U0000A796" },
{ 0x0000A798, U"\U0000A798\U0000A799\U0000A798" },
{ 0x0000A799, U"\U0000A798\U0000A799\U0000A798" },
{ 0x0000A79A, U"\U0000A79A\U0000A79B\U0000A79A" },
{ 0x0000A79B, U"\U0000A79A\U0000A79B\U0000A79A" },
{ 0x0000A79C, U"\U0000A79C\U0000A79D\U0000A79C" },
{ 0x0000A79D, U"\U0000A79C\U0000A79D\U0000A79C" },
{ 0x0000A79E, U"\U0000A79E\U0000A79F\U0000A79E" },
{ 0x0000A79F, U"\U0000A79E\U0000A79F\U0000A79E" },
{ 0x0000A7A0, U"\U0000A7A0\U0000A7A1\U0000A7A0" },
{ 0x0000A7A1, U"\U0000A7A0\U0000A7A1\U0000A7A0" },
{ 0x0000A7A2, U"\U0000A7A2\U0000A7A3\U0000A7A2" },
{ 0x0000A7A3, U"\U0000A7A2\U0000A7A3\U0000A7A2" },
{ 0x0000A7A4, U"\U0000A7A4\U0000A7A5\U0000A7A4" },
{ 0x0000A7A5, U"\U0000A7A4\U0000A7A5\U0000A7A4" },
{ 0x0000A7A6, U"\U0000A7A6\U0000A7A7\U0000A7A6" },
{ 0x0000A7A7, U"\U0000A7A6\U0000A7A7\U0000A7A6" },
{ 0x0000A7A8, U"\U0000A7A8\U0000A7A9\U0000A7A8" },
{ 0x0000A7A9, U"\U0000A7A8\U0000A7A9\U0000A7A8" },
{ 0x0000A7AA, U"\U0000A7AA\U00000266\U0000A7AA" },
{ 0x0000A7AB, U"\U0000A7AB\U0000025C\U0000A7AB" },
{ 0x0000A7AC, U"\U0000A7AC\U00000261\U0000A7AC" },
{ 0x0000A7AD, U"\U0000A7AD\U0000026C\U0000A7AD" },
{ 0x0000A7AE, U"\U0000A7AE\U0000026A\U0000A7AE" },
{ 0x0000A7B0, U"\U0000A7B0\U0000029E\U0000A7B0" },
{ 0x0000A7B1, U"\U0000A7B1\U00000287\U0000A7B1" },
{ 0x0000A7B2, U"\U0000A7B2\U0000029D\U0000A7B2" },
{ 0x0000A7B3, U"\U0000A7B3\U0000AB53\U0000A7B3" },
{ 0x0000A7B4, U"\U0000A7B4\U0000A7B5\U0000A7B4" },
{ 0x0000A7B5, U"\U0000A7B4\U0000A7B5\U0000A7B4" },
{ 0x0000A7B6, U"\U0000A7B6\U0000A7B7\U0000A7B6" },
{ 0x0000A7B7, U"\U0000A7B6\U0000A7B7\U0000A7B6" },
{ 0x0000A7B8, U"\U0000A7B8\U0000A7B9\U0000A7B8" },
{ 0x0000A7B9, U"\U0000A7B8\U0000A7B9\U0000A7B8" },
{ 0x0000A7BA, U"\U0000A7BA\U0000A7BB\U0000A7BA" },
{ 0x0000A7BB, U"\U0000A7BA\U0000A7BB\U0000A7BA" },
{ 0x0000A7BC, U"\U0000A7BC\U0000A7BD\U0000A7BC" },
{ 0x0000A7BD, U"\U0000A7BC\U0000A7BD\U0000A7BC" },
{ 0x0000A7BE, U"\U0000A7BE\U0000A7BF\U0000A7BE" },
{ 0x0000A7BF, U"\U0000A7BE\U0000A7BF\U0000A7BE" },
{ 0x0000A7C2, U"\U0000A7C2\U0000A7C3\U0000A7C2" },
{ 0x0000A7C3, U"\U0000A7C2\U0000A7C3\U0000A7C2" },
{ 0x0000A7C4, U"\U0000A7C4\U0000A794\U0000A7C4" },
{ 0x0000A7C5, U"\U0000A7C5\U00000282\U0000A7C5" },
{ 0x0000A7C6, U"\U0000A7C6\U00001D8E\U0000A7C6" },
{ 0x0000A7C7, U"\U0000A7C7\U0000A7C8\U0000A7C7" },
{ 0x0000A7C8, U"\U0000A7C7\U0000A7C8\U0000A7C7" },
{ 0x0000A7C9, U"\U0000A7C9\U0000A7CA\U0000A7C9" },
{ 0x0000A7CA, U"\U0000A7C9\U0000A7CA\U0000A7C9" },
{ 0x0000A7F5, U"\U0000A7F5\U0000A7F6\U0000A7F5" },
{ 0x0000A7F6, U"\U0000A7F5\U0000A7F6\U0000A7F5" },
{ 0x0000AB53, U"\U0000A7B3\U0000AB53\U0000A7B3" },
{ 0x0000AB70, U"\U000013A0\U0000AB70\U000013A0" },
{ 0x0000AB71, U"\U000013A1\U0000AB71\U000013A1" },
{ 0x0000AB72, U"\U000013A2\U0000AB72\U000013A2" },
{ 0x0000AB73, U"\U000013A3\U0000AB73\U000013A3" },
{ 0x0000AB74, U"\U000013A4\U0000AB74\U000013A4" },
{ 0x0000AB75, U"\U000013A5\U0000AB75\U000013A5" },
{ 0x0000AB76, U"\U000013A6\U0000AB76\U000013A6" },
{ 0x0000AB77, U"\U000013A7\U0000AB77\U000013A7" },
{ 0x0000AB78, U"\U000013A8\U0000AB78\U000013A8" },
{ 0x0000AB79, U"\U000013A9\U0000AB79\U000013A9" },
{ 0x0000AB7A, U"\U000013AA\U0000AB7A\U000013AA" },
{ 0x0000AB7B, U"\U000013AB\U0000AB7B\U000013AB" },
{ 0x0000AB7C, U"\U000013AC\U0000AB7C\U000013AC" },
{ 0x0000AB7D, U"\U000013AD\U0000AB7D\U000013AD" },
{ 0x0000AB7E, U"\U000013AE\U0000AB7E\U000013AE" },
{ 0x0000AB7F, U"\U000013AF\U0000AB7F\U000013AF" },
{ 0x0000AB80, U"\U000013B0\U0000AB80\U000013B0" },
{ 0x0000AB81, U"\U000013B1\U0000AB81\U000013B1" },
{ 0x0000AB82, U"\U000013B2\U0000AB82\U000013B2" },
{ 0x0000AB83, U"\U000013B3\U0000AB83\U000013B3" },
{ 0x0000AB84, U"\U000013B4\U0000AB84\U000013B4" },
{ 0x0000AB85, U"\U000013B5\U0000AB85\U000013B5" },
{ 0x0000AB86, U"\U000013B6\U0000AB86\U000013B6" },
{ 0x0000AB87, U"\U000013B7\U0000AB87\U000013B7" },
{ 0x0000AB88, U"\U000013B8\U0000AB88\U000013B8" },
{ 0x0000AB89, U"\U000013B9\U0000AB89\U000013B9" },
{ 0x0000AB8A, U"\U000013BA\U0000AB8A\U000013BA" },
{ 0x0000AB8B, U"\U000013BB\U0000AB8B\U000013BB" },
{ 0x0000AB8C, U"\U000013BC\U0000AB8C\U000013BC" },
{ 0x0000AB8D, U"\U000013BD\U0000AB8D\U000013BD" },
{ 0x0000AB8E, U"\U000013BE\U0000AB8E\U000013BE" },
{ 0x0000AB8F, U"\U000013BF\U0000AB8F\U000013BF" },
{ 0x0000AB90, U"\U000013C0\U0000AB90\U000013C0" },
{ 0x0000AB91, U"\U000013C1\U0000AB91\U000013C1" },
{ 0x0000AB92, U"\U000013C2\U0000AB92\U000013C2" },
{ 0x0000AB93, U"\U000013C3\U0000AB93\U000013C3" },
{ 0x0000AB94, U"\U000013C4\U0000AB94\U000013C4" },
{ 0x0000AB95, U"\U000013C5\U0000AB95\U000013C5" },
{ 0x0000AB96, U"\U000013C6\U0000AB96\U000013C6" },
{ 0x0000AB97, U"\U000013C7\U0000AB97\U000013C7" },
{ 0x0000AB98, U"\U000013C8\U0000AB98\U000013C8" },
{ 0x0000AB99, U"\U000013C9\U0000AB99\U000013C9" },
{ 0x0000AB9A, U"\U000013CA\U0000AB9A\U000013CA" },
{ 0x0000AB9B, U"\U000013CB\U0000AB9B\U000013CB" },
{ 0x0000AB9C, U"\U000013CC\U0000AB9C\U000013CC" },
{ 0x0000AB9D, U"\U000013CD\U0000AB9D\U000013CD" },
{ 0x0000AB9E, U"\U000013CE\U0000AB9E\U000013CE" },
{ 0x0000AB9F, U"\U000013CF\U0000AB9F\U000013CF" },
{ 0x0000ABA0, U"\U000013D0\U0000ABA0\U000013D0" },
{ 0x0000ABA1, U"\U000013D1\U0000ABA1\U000013D1" },
{ 0x0000ABA2, U"\U000013D2\U0000ABA2\U000013D2" },
{ 0x0000ABA3, U"\U000013D3\U0000ABA3\U000013D3" },
{ 0x0000ABA4, U"\U000013D4\U0000ABA4\U000013D4" },
{ 0x0000ABA5, U"\U000013D5\U0000ABA5\U000013D5" },
{ 0x0000ABA6, U"\U000013D6\U0000ABA6\U000013D6" },
{ 0x0000ABA7, U"\U000013D7\U0000ABA7\U000013D7" },
{ 0x0000ABA8, U"\U000013D8\U0000ABA8\U000013D8" },
{ 0x0000ABA9, U"\U000013D9\U0000ABA9\U000013D9" },
{ 0x0000ABAA, U"\U000013DA\U0000ABAA\U000013DA" },
{ 0x0000ABAB, U"\U000013DB\U0000ABAB\U000013DB" },
{ 0x0000ABAC, U"\U000013DC\U0000ABAC\U000013DC" },
{ 0x0000ABAD, U"\U000013DD\U0000ABAD\U000013DD" },
{ 0x0000ABAE, U"\U000013DE\U0000ABAE\U000013DE" },
{ 0x0000ABAF, U"\U000013DF\U0000ABAF\U000013DF" },
{ 0x0000ABB0, U"\U000013E0\U0000ABB0\U000013E0" },
{ 0x0000ABB1, U"\U000013E1\U0000ABB1\U000013E1" },
{ 0x0000ABB2, U"\U000013E2\U0000ABB2\U000013E2" },
{ 0x0000ABB3, U"\U000013E3\U0000ABB3\U000013E3" },
{ 0x0000ABB4, U"\U000013E4\U0000ABB4\U000013E4" },
{ 0x0000ABB5, U"\U000013E5\U0000ABB5\U000013E5" },
{ 0x0000ABB6, U"\U000013E6\U0000ABB6\U000013E6" },
{ 0x0000ABB7, U"\U000013E7\U0000ABB7\U000013E7" },
{ 0x0000ABB8, U"\U000013E8\U0000ABB8\U000013E8" },
{ 0x0000ABB9, U"\U000013E9\U0000ABB9\U000013E9" },
{ 0x0000ABBA, U"\U000013EA\U0000ABBA\U000013EA" },
{ 0x0000ABBB, U"\U000013EB\U0000ABBB\U000013EB" },
{ 0x0000ABBC, U"\U000013EC\U0000ABBC\U000013EC" },
{ 0x0000ABBD, U"\U000013ED\U0000ABBD\U000013ED" },
{ 0x0000ABBE, U"\U000013EE\U0000ABBE\U000013EE" },
{ 0x0000ABBF, U"\U000013EF\U0000ABBF\U000013EF" },
{ 0x0000FF21, U"\U0000FF21\U0000FF41\U0000FF21" },
{ 0x0000FF22, U"\U0000FF22\U0000FF42\U0000FF22" },
{ 0x0000FF23, U"\U0000FF23\U0000FF43\U0000FF23" },
{ 0x0000FF24, U"\U0000FF24\U0000FF44\U0000FF24" },
{ 0x0000FF25, U"\U0000FF25\U0000FF45\U0000FF25" },
{ 0x0000FF26, U"\U0000FF26\U0000FF46\U0000FF26" },
{ 0x0000FF27, U"\U0000FF27\U0000FF47\U0000FF27" },
{ 0x0000FF28, U"\U0000FF28\U0000FF48\U0000FF28" },
{ 0x0000FF29, U"\U0000FF29\U0000FF49\U0000FF29" },
{ 0x0000FF2A, U"\U0000FF2A\U0000FF4A\U0000FF2A" },
{ 0x0000FF2B, U"\U0000FF2B\U0000FF4B\U0000FF2B" },
{ 0x0000FF2C, U"\U0000FF2C\U0000FF4C\U0000FF2C" },
{ 0x0000FF2D, U"\U0000FF2D\U0000FF4D\U0000FF2D" },
{ 0x0000FF2E, U"\U0000FF2E\U0000FF4E\U0000FF2E" },
{ 0x0000FF2F, U"\U0000FF2F\U0000FF4F\U0000FF2F" },
{ 0x0000FF30, U"\U0000FF30\U0000FF50\U0000FF30" },
{ 0x0000FF31, U"\U0000FF31\U0000FF51\U0000FF31" },
{ 0x0000FF32, U"\U0000FF32\U0000FF52\U0000FF32" },
{ 0x0000FF33, U"\U0000FF33\U0000FF53\U0000FF33" },
{ 0x0000FF34, U"\U0000FF34\U0000FF54\U0000FF34" },
{ 0x0000FF35, U"\U0000FF35\U0000FF55\U0000FF35" },
{ 0x0000FF36, U"\U0000FF36\U0000FF56\U0000FF36" },
{ 0x0000FF37, U"\U0000FF37\U0000FF57\U0000FF37" },
{ 0x0000FF38, U"\U0000FF38\U0000FF58\U0000FF38" },
{ 0x0000FF39, U"\U0000FF39\U0000FF59\U0000FF39" },
{ 0x0000FF3A, U"\U0000FF3A\U0000FF5A\U0000FF3A" },
{ 0x0000FF41, U"\U0000FF21\U0000FF41\U0000FF21" },
{ 0x0000FF42, U"\U0000FF22\U0000FF42\U0000FF22" },
{ 0x0000FF43, U"\U0000FF23\U0000FF43\U0000FF23" },
{ 0x0000FF44, U"\U0000FF24\U0000FF44\U0000FF24" },
{ 0x0000FF45, U"\U0000FF25\U0000FF45\U0000FF25" },
{ 0x0000FF46, U"\U0000FF26\U0000FF46\U0000FF26" },
{ 0x0000FF47, U"\U0000FF27\U0000FF47\U0000FF27" },
{ 0x0000FF48, U"\U0000FF28\U0000FF48\U0000FF28" },
{ 0x0000FF49, U"\U0000FF29\U0000FF49\U0000FF29" },
{ 0x0000FF4A, U"\U0000FF2A\U0000FF4A\U0000FF2A" },
{ 0x0000FF4B, U"\U0000FF2B\U0000FF4B\U0000FF2B" },
{ 0x0000FF4C, U"\U0000FF2C\U0000FF4C\U0000FF2C" },
{ 0x0000FF4D, U"\U0000FF2D\U0000FF4D\U0000FF2D" },
{ 0x0000FF4E, U"\U0000FF2E\U0000FF4E\U0000FF2E" },
{ 0x0000FF4F, U"\U0000FF2F\U0000FF4F\U0000FF2F" },
{ 0x0000FF50, U"\U0000FF30\U0000FF50\U0000FF30" },
{ 0x0000FF51, U"\U0000FF31\U0000FF51\U0000FF31" },
{ 0x0000FF52, U"\U0000FF32\U0000FF52\U0000FF32" },
{ 0x0000FF53, U"\U0000FF33\U0000FF53\U0000FF33" },
{ 0x0000FF54, U"\U0000FF34\U0000FF54\U0000FF34" },
{ 0x0000FF55, U"\U0000FF35\U0000FF55\U0000FF35" },
{ 0x0000FF56, U"\U0000FF36\U0000FF56\U0000FF36" },
{ 0x0000FF57, U"\U0000FF37\U0000FF57\U0000FF37" },
{ 0x0000FF58, U"\U0000FF38\U0000FF58\U0000FF38" },
{ 0x0000FF59, U"\U0000FF39\U0000FF59\U0000FF39" },
{ 0x0000FF5A, U"\U0000FF3A\U0000FF5A\U0000FF3A" },
{ 0x00010400, U"\U00010400\U00010428\U00010400" },
{ 0x00010401, U"\U00010401\U00010429\U00010401" },
{ 0x00010402, U"\U00010402\U0001042A\U00010402" },
{ 0x00010403, U"\U00010403\U0001042B\U00010403" },
{ 0x00010404, U"\U00010404\U0001042C\U00010404" },
{ 0x00010405, U"\U00010405\U0001042D\U00010405" },
{ 0x00010406, U"\U00010406\U0001042E\U00010406" },
{ 0x00010407, U"\U00010407\U0001042F\U00010407" },
{ 0x00010408, U"\U00010408\U00010430\U00010408" },
{ 0x00010409, U"\U00010409\U00010431\U00010409" },
{ 0x0001040A, U"\U0001040A\U00010432\U0001040A" },
{ 0x0001040B, U"\U0001040B\U00010433\U0001040B" },
{ 0x0001040C, U"\U0001040C\U00010434\U0001040C" },
{ 0x0001040D, U"\U0001040D\U00010435\U0001040D" },
{ 0x0001040E, U"\U0001040E\U00010436\U0001040E" },
{ 0x0001040F, U"\U0001040F\U00010437\U0001040F" },
{ 0x00010410, U"\U00010410\U00010438\U00010410" },
{ 0x00010411, U"\U00010411\U00010439\U00010411" },
{ 0x00010412, U"\U00010412\U0001043A\U00010412" },
{ 0x00010413, U"\U00010413\U0001043B\U00010413" },
{ 0x00010414, U"\U00010414\U0001043C\U00010414" },
{ 0x00010415, U"\U00010415\U0001043D\U00010415" },
{ 0x00010416, U"\U00010416\U0001043E\U00010416" },
{ 0x00010417, U"\U00010417\U0001043F\U00010417" },
{ 0x00010418, U"\U00010418\U00010440\U00010418" },
{ 0x00010419, U"\U00010419\U00010441\U00010419" },
{ 0x0001041A, U"\U0001041A\U00010442\U0001041A" },
{ 0x0001041B, U"\U0001041B\U00010443\U0001041B" },
{ 0x0001041C, U"\U0001041C\U00010444\U0001041C" },
{ 0x0001041D, U"\U0001041D\U00010445\U0001041D" },
{ 0x0001041E, U"\U0001041E\U00010446\U0001041E" },
{ 0x0001041F, U"\U0001041F\U00010447\U0001041F" },
{ 0x00010420, U"\U00010420\U00010448\U00010420" },
{ 0x00010421, U"\U00010421\U00010449\U00010421" },
{ 0x00010422, U"\U00010422\U0001044A\U00010422" },
{ 0x00010423, U"\U00010423\U0001044B\U00010423" },
{ 0x00010424, U"\U00010424\U0001044C\U00010424" },
{ 0x00010425, U"\U00010425\U0001044D\U00010425" },
{ 0x00010426, U"\U00010426\U0001044E\U00010426" },
{ 0x00010427, U"\U00010427\U0001044F\U00010427" },
{ 0x00010428, U"\U00010400\U00010428\U00010400" },
{ 0x00010429, U"\U00010401\U00010429\U00010401" },
{ 0x0001042A, U"\U00010402\U0001042A\U00010402" },
{ 0x0001042B, U"\U00010403\U0001042B\U00010403" },
{ 0x0001042C, U"\U00010404\U0001042C\U00010404" },
{ 0x0001042D, U"\U00010405\U0001042D\U00010405" },
{ 0x0001042E, U"\U00010406\U0001042E\U00010406" },
{ 0x0001042F, U"\U00010407\U0001042F\U00010407" },
{ 0x00010430, U"\U00010408\U00010430\U00010408" },
{ 0x00010431, U"\U00010409\U00010431\U00010409" },
{ 0x00010432, U"\U0001040A\U00010432\U0001040A" },
{ 0x00010433, U"\U0001040B\U00010433\U0001040B" },
{ 0x00010434, U"\U0001040C\U00010434\U0001040C" },
{ 0x00010435, U"\U0001040D\U00010435\U0001040D" },
{ 0x00010436, U"\U0001040E\U00010436\U0001040E" },
{ 0x00010437, U"\U0001040F\U00010437\U0001040F" },
{ 0x00010438, U"\U00010410\U00010438\U00010410" },
{ 0x00010439, U"\U00010411\U00010439\U00010411" },
{ 0x0001043A, U"\U00010412\U0001043A\U00010412" },
{ 0x0001043B, U"\U00010413\U0001043B\U00010413" },
{ 0x0001043C, U"\U00010414\U0001043C\U00010414" },
{ 0x0001043D, U"\U00010415\U0001043D\U00010415" },
{ 0x0001043E, U"\U00010416\U0001043E\U00010416" },
{ 0x0001043F, U"\U00010417\U0001043F\U00010417" },
{ 0x00010440, U"\U00010418\U00010440\U00010418" },
{ 0x00010441, U"\U00010419\U00010441\U00010419" },
{ 0x00010442, U"\U0001041A\U00010442\U0001041A" },
{ 0x00010443, U"\U0001041B\U00010443\U0001041B" },
{ 0x00010444, U"\U0001041C\U00010444\U0001041C" },
{ 0x00010445, U"\U0001041D\U00010445\U0001041D" },
{ 0x00010446, U"\U0001041E\U00010446\U0001041E" },
{ 0x00010447, U"\U0001041F\U00010447\U0001041F" },
{ 0x00010448, U"\U00010420\U00010448\U00010420" },
{ 0x00010449, U"\U00010421\U00010449\U00010421" },
{ 0x0001044A, U"\U00010422\U0001044A\U00010422" },
{ 0x0001044B, U"\U00010423\U0001044B\U00010423" },
{ 0x0001044C, U"\U00010424\U0001044C\U00010424" },
{ 0x0001044D, U"\U00010425\U0001044D\U00010425" },
{ 0x0001044E, U"\U00010426\U0001044E\U00010426" },
{ 0x0001044F, U"\U00010427\U0001044F\U00010427" },
{ 0x000104B0, U"\U000104B0\U000104D8\U000104B0" },
{ 0x000104B1, U"\U000104B1\U000104D9\U000104B1" },
{ 0x000104B2, U"\U000104B2\U000104DA\U000104B2" },
{ 0x000104B3, U"\U000104B3\U000104DB\U000104B3" },
{ 0x000104B4, U"\U000104B4\U000104DC\U000104B4" },
{ 0x000104B5, U"\U000104B5\U000104DD\U000104B5" },
{ 0x000104B6, U"\U000104B6\U000104DE\U000104B6" },
{ 0x000104B7, U"\U000104B7\U000104DF\U000104B7" },
{ 0x000104B8, U"\U000104B8\U000104E0\U000104B8" },
{ 0x000104B9, U"\U000104B9\U000104E1\U000104B9" },
{ 0x000104BA, U"\U000104BA\U000104E2\U000104BA" },
{ 0x000104BB, U"\U000104BB\U000104E3\U000104BB" },
{ 0x000104BC, U"\U000104BC\U000104E4\U000104BC" },
{ 0x000104BD, U"\U000104BD\U000104E5\U000104BD" },
{ 0x000104BE, U"\U000104BE\U000104E6\U000104BE" },
{ 0x000104BF, U"\U000104BF\U000104E7\U000104BF" },
{ 0x000104C0, U"\U000104C0\U000104E8\U000104C0" },
{ 0x000104C1, U"\U000104C1\U000104E9\U000104C1" },
{ 0x000104C2, U"\U000104C2\U000104EA\U000104C2" },
{ 0x000104C3, U"\U000104C3\U000104EB\U000104C3" },
{ 0x000104C4, U"\U000104C4\U000104EC\U000104C4" },
{ 0x000104C5, U"\U000104C5\U000104ED\U000104C5" },
{ 0x000104C6, U"\U000104C6\U000104EE\U000104C6" },
{ 0x000104C7, U"\U000104C7\U000104EF\U000104C7" },
{ 0x000104C8, U"\U000104C8\U000104F0\U000104C8" },
{ 0x000104C9, U"\U000104C9\U000104F1\U000104C9" },
{ 0x000104CA, U"\U000104CA\U000104F2\U000104CA" },
{ 0x000104CB, U"\U000104CB\U000104F3\U000104CB" },
{ 0x000104CC, U"\U000104CC\U000104F4\U000104CC" },
{ 0x000104CD, U"\U000104CD\U000104F5\U000104CD" },
{ 0x000104CE, U"\U000104CE\U000104F6\U000104CE" },
{ 0x000104CF, U"\U000104CF\U000104F7\U000104CF" },
{ 0x000104D0, U"\U000104D0\U000104F8\U000104D0" },
{ 0x000104D1, U"\U000104D1\U000104F9\U000104D1" },
{ 0x000104D2, U"\U000104D2\U000104FA\U000104D2" },
{ 0x000104D3, U"\U000104D3\U000104FB\U000104D3" },
{ 0x000104D8, U"\U000104B0\U000104D8\U000104B0" },
{ 0x000104D9, U"\U000104B1\U000104D9\U000104B1" },
{ 0x000104DA, U"\U000104B2\U000104DA\U000104B2" },
{ 0x000104DB, U"\U000104B3\U000104DB\U000104B3" },
{ 0x000104DC, U"\U000104B4\U000104DC\U000104B4" },
{ 0x000104DD, U"\U000104B5\U000104DD\U000104B5" },
{ 0x000104DE, U"\U000104B6\U000104DE\U000104B6" },
{ 0x000104DF, U"\U000104B7\U000104DF\U000104B7" },
{ 0x000104E0, U"\U000104B8\U000104E0\U000104B8" },
{ 0x000104E1, U"\U000104B9\U000104E1\U000104B9" },
{ 0x000104E2, U"\U000104BA\U000104E2\U000104BA" },
{ 0x000104E3, U"\U000104BB\U000104E3\U000104BB" },
{ 0x000104E4, U"\U000104BC\U000104E4\U000104BC" },
{ 0x000104E5, U"\U000104BD\U000104E5\U000104BD" },
{ 0x000104E6, U"\U000104BE\U000104E6\U000104BE" },
{ 0x000104E7, U"\U000104BF\U000104E7\U000104BF" },
{ 0x000104E8, U"\U000104C0\U000104E8\U000104C0" },
{ 0x000104E9, U"\U000104C1\U000104E9\U000104C1" },
{ 0x000104EA, U"\U000104C2\U000104EA\U000104C2" },
{ 0x000104EB, U"\U000104C3\U000104EB\U000104C3" },
{ 0x000104EC, U"\U000104C4\U000104EC\U000104C4" },
{ 0x000104ED, U"\U000104C5\U000104ED\U000104C5" },
{ 0x000104EE, U"\U000104C6\U000104EE\U000104C6" },
{ 0x000104EF, U"\U000104C7\U000104EF\U000104C7" },
{ 0x000104F0, U"\U000104C8\U000104F0\U000104C8" },
{ 0x000104F1, U"\U000104C9\U000104F1\U000104C9" },
{ 0x000104F2, U"\U000104CA\U000104F2\U000104CA" },
{ 0x000104F3, U"\U000104CB\U000104F3\U000104CB" },
{ 0x000104F4, U"\U000104CC\U000104F4\U000104CC" },
{ 0x000104F5, U"\U000104CD\U000104F5\U000104CD" },
{ 0x000104F6, U"\U000104CE\U000104F6\U000104CE" },
{ 0x000104F7, U"\U000104CF\U000104F7\U000104CF" },
{ 0x000104F8, U"\U000104D0\U000104F8\U000104D0" },
{ 0x000104F9, U"\U000104D1\U000104F9\U000104D1" },
{ 0x000104FA, U"\U000104D2\U000104FA\U000104D2" },
{ 0x000104FB, U"\U000104D3\U000104FB\U000104D3" },
{ 0x00010C80, U"\U00010C80\U00010CC0\U00010C80" },
{ 0x00010C81, U"\U00010C81\U00010CC1\U00010C81" },
{ 0x00010C82, U"\U00010C82\U00010CC2\U00010C82" },
{ 0x00010C83, U"\U00010C83\U00010CC3\U00010C83" },
{ 0x00010C84, U"\U00010C84\U00010CC4\U00010C84" },
{ 0x00010C85, U"\U00010C85\U00010CC5\U00010C85" },
{ 0x00010C86, U"\U00010C86\U00010CC6\U00010C86" },
{ 0x00010C87, U"\U00010C87\U00010CC7\U00010C87" },
{ 0x00010C88, U"\U00010C88\U00010CC8\U00010C88" },
{ 0x00010C89, U"\U00010C89\U00010CC9\U00010C89" },
{ 0x00010C8A, U"\U00010C8A\U00010CCA\U00010C8A" },
{ 0x00010C8B, U"\U00010C8B\U00010CCB\U00010C8B" },
{ 0x00010C8C, U"\U00010C8C\U00010CCC\U00010C8C" },
{ 0x00010C8D, U"\U00010C8D\U00010CCD\U00010C8D" },
{ 0x00010C8E, U"\U00010C8E\U00010CCE\U00010C8E" },
{ 0x00010C8F, U"\U00010C8F\U00010CCF\U00010C8F" },
{ 0x00010C90, U"\U00010C90\U00010CD0\U00010C90" },
{ 0x00010C91, U"\U00010C91\U00010CD1\U00010C91" },
{ 0x00010C92, U"\U00010C92\U00010CD2\U00010C92" },
{ 0x00010C93, U"\U00010C93\U00010CD3\U00010C93" },
{ 0x00010C94, U"\U00010C94\U00010CD4\U00010C94" },
{ 0x00010C95, U"\U00010C95\U00010CD5\U00010C95" },
{ 0x00010C96, U"\U00010C96\U00010CD6\U00010C96" },
{ 0x00010C97, U"\U00010C97\U00010CD7\U00010C97" },
{ 0x00010C98, U"\U00010C98\U00010CD8\U00010C98" },
{ 0x00010C99, U"\U00010C99\U00010CD9\U00010C99" },
{ 0x00010C9A, U"\U00010C9A\U00010CDA\U00010C9A" },
{ 0x00010C9B, U"\U00010C9B\U00010CDB\U00010C9B" },
{ 0x00010C9C, U"\U00010C9C\U00010CDC\U00010C9C" },
{ 0x00010C9D, U"\U00010C9D\U00010CDD\U00010C9D" },
{ 0x00010C9E, U"\U00010C9E\U00010CDE\U00010C9E" },
{ 0x00010C9F, U"\U00010C9F\U00010CDF\U00010C9F" },
{ 0x00010CA0, U"\U00010CA0\U00010CE0\U00010CA0" },
{ 0x00010CA1, U"\U00010CA1\U00010CE1\U00010CA1" },
{ 0x00010CA2, U"\U00010CA2\U00010CE2\U00010CA2" },
{ 0x00010CA3, U"\U00010CA3\U00010CE3\U00010CA3" },
{ 0x00010CA4, U"\U00010CA4\U00010CE4\U00010CA4" },
{ 0x00010CA5, U"\U00010CA5\U00010CE5\U00010CA5" },
{ 0x00010CA6, U"\U00010CA6\U00010CE6\U00010CA6" },
{ 0x00010CA7, U"\U00010CA7\U00010CE7\U00010CA7" },
{ 0x00010CA8, U"\U00010CA8\U00010CE8\U00010CA8" },
{ 0x00010CA9, U"\U00010CA9\U00010CE9\U00010CA9" },
{ 0x00010CAA, U"\U00010CAA\U00010CEA\U00010CAA" },
{ 0x00010CAB, U"\U00010CAB\U00010CEB\U00010CAB" },
{ 0x00010CAC, U"\U00010CAC\U00010CEC\U00010CAC" },
{ 0x00010CAD, U"\U00010CAD\U00010CED\U00010CAD" },
{ 0x00010CAE, U"\U00010CAE\U00010CEE\U00010CAE" },
{ 0x00010CAF, U"\U00010CAF\U00010CEF\U00010CAF" },
{ 0x00010CB0, U"\U00010CB0\U00010CF0\U00010CB0" },
{ 0x00010CB1, U"\U00010CB1\U00010CF1\U00010CB1" },
{ 0x00010CB2, U"\U00010CB2\U00010CF2\U00010CB2" },
{ 0x00010CC0, U"\U00010C80\U00010CC0\U00010C80" },
{ 0x00010CC1, U"\U00010C81\U00010CC1\U00010C81" },
{ 0x00010CC2, U"\U00010C82\U00010CC2\U00010C82" },
{ 0x00010CC3, U"\U00010C83\U00010CC3\U00010C83" },
{ 0x00010CC4, U"\U00010C84\U00010CC4\U00010C84" },
{ 0x00010CC5, U"\U00010C85\U00010CC5\U00010C85" },
{ 0x00010CC6, U"\U00010C86\U00010CC6\U00010C86" },
{ 0x00010CC7, U"\U00010C87\U00010CC7\U00010C87" },
{ 0x00010CC8, U"\U00010C88\U00010CC8\U00010C88" },
{ 0x00010CC9, U"\U00010C89\U00010CC9\U00010C89" },
{ 0x00010CCA, U"\U00010C8A\U00010CCA\U00010C8A" },
{ 0x00010CCB, U"\U00010C8B\U00010CCB\U00010C8B" },
{ 0x00010CCC, U"\U00010C8C\U00010CCC\U00010C8C" },
{ 0x00010CCD, U"\U00010C8D\U00010CCD\U00010C8D" },
{ 0x00010CCE, U"\U00010C8E\U00010CCE\U00010C8E" },
{ 0x00010CCF, U"\U00010C8F\U00010CCF\U00010C8F" },
{ 0x00010CD0, U"\U00010C90\U00010CD0\U00010C90" },
{ 0x00010CD1, U"\U00010C91\U00010CD1\U00010C91" },
{ 0x00010CD2, U"\U00010C92\U00010CD2\U00010C92" },
{ 0x00010CD3, U"\U00010C93\U00010CD3\U00010C93" },
{ 0x00010CD4, U"\U00010C94\U00010CD4\U00010C94" },
{ 0x00010CD5, U"\U00010C95\U00010CD5\U00010C95" },
{ 0x00010CD6, U"\U00010C96\U00010CD6\U00010C96" },
{ 0x00010CD7, U"\U00010C97\U00010CD7\U00010C97" },
{ 0x00010CD8, U"\U00010C98\U00010CD8\U00010C98" },
{ 0x00010CD9, U"\U00010C99\U00010CD9\U00010C99" },
{ 0x00010CDA, U"\U00010C9A\U00010CDA\U00010C9A" },
{ 0x00010CDB, U"\U00010C9B\U00010CDB\U00010C9B" },
{ 0x00010CDC, U"\U00010C9C\U00010CDC\U00010C9C" },
{ 0x00010CDD, U"\U00010C9D\U00010CDD\U00010C9D" },
{ 0x00010CDE, U"\U00010C9E\U00010CDE\U00010C9E" },
{ 0x00010CDF, U"\U00010C9F\U00010CDF\U00010C9F" },
{ 0x00010CE0, U"\U00010CA0\U00010CE0\U00010CA0" },
{ 0x00010CE1, U"\U00010CA1\U00010CE1\U00010CA1" },
{ 0x00010CE2, U"\U00010CA2\U00010CE2\U00010CA2" },
{ 0x00010CE3, U"\U00010CA3\U00010CE3\U00010CA3" },
{ 0x00010CE4, U"\U00010CA4\U00010CE4\U00010CA4" },
{ 0x00010CE5, U"\U00010CA5\U00010CE5\U00010CA5" },
{ 0x00010CE6, U"\U00010CA6\U00010CE6\U00010CA6" },
{ 0x00010CE7, U"\U00010CA7\U00010CE7\U00010CA7" },
{ 0x00010CE8, U"\U00010CA8\U00010CE8\U00010CA8" },
{ 0x00010CE9, U"\U00010CA9\U00010CE9\U00010CA9" },
{ 0x00010CEA, U"\U00010CAA\U00010CEA\U00010CAA" },
{ 0x00010CEB, U"\U00010CAB\U00010CEB\U00010CAB" },
{ 0x00010CEC, U"\U00010CAC\U00010CEC\U00010CAC" },
{ 0x00010CED, U"\U00010CAD\U00010CED\U00010CAD" },
{ 0x00010CEE, U"\U00010CAE\U00010CEE\U00010CAE" },
{ 0x00010CEF, U"\U00010CAF\U00010CEF\U00010CAF" },
{ 0x00010CF0, U"\U00010CB0\U00010CF0\U00010CB0" },
{ 0x00010CF1, U"\U00010CB1\U00010CF1\U00010CB1" },
{ 0x00010CF2, U"\U00010CB2\U00010CF2\U00010CB2" },
{ 0x000118A0, U"\U000118A0\U000118C0\U000118A0" },
{ 0x000118A1, U"\U000118A1\U000118C1\U000118A1" },
{ 0x000118A2, U"\U000118A2\U000118C2\U000118A2" },
{ 0x000118A3, U"\U000118A3\U000118C3\U000118A3" },
{ 0x000118A4, U"\U000118A4\U000118C4\U000118A4" },
{ 0x000118A5, U"\U000118A5\U000118C5\U000118A5" },
{ 0x000118A6, U"\U000118A6\U000118C6\U000118A6" },
{ 0x000118A7, U"\U000118A7\U000118C7\U000118A7" },
{ 0x000118A8, U"\U000118A8\U000118C8\U000118A8" },
{ 0x000118A9, U"\U000118A9\U000118C9\U000118A9" },
{ 0x000118AA, U"\U000118AA\U000118CA\U000118AA" },
{ 0x000118AB, U"\U000118AB\U000118CB\U000118AB" },
{ 0x000118AC, U"\U000118AC\U000118CC\U000118AC" },
{ 0x000118AD, U"\U000118AD\U000118CD\U000118AD" },
{ 0x000118AE, U"\U000118AE\U000118CE\U000118AE" },
{ 0x000118AF, U"\U000118AF\U000118CF\U000118AF" },
{ 0x000118B0, U"\U000118B0\U000118D0\U000118B0" },
{ 0x000118B1, U"\U000118B1\U000118D1\U000118B1" },
{ 0x000118B2, U"\U000118B2\U000118D2\U000118B2" },
{ 0x000118B3, U"\U000118B3\U000118D3\U000118B3" },
{ 0x000118B4, U"\U000118B4\U000118D4\U000118B4" },
{ 0x000118B5, U"\U000118B5\U000118D5\U000118B5" },
{ 0x000118B6, U"\U000118B6\U000118D6\U000118B6" },
{ 0x000118B7, U"\U000118B7\U000118D7\U000118B7" },
{ 0x000118B8, U"\U000118B8\U000118D8\U000118B8" },
{ 0x000118B9, U"\U000118B9\U000118D9\U000118B9" },
{ 0x000118BA, U"\U000118BA\U000118DA\U000118BA" },
{ 0x000118BB, U"\U000118BB\U000118DB\U000118BB" },
{ 0x000118BC, U"\U000118BC\U000118DC\U000118BC" },
{ 0x000118BD, U"\U000118BD\U000118DD\U000118BD" },
{ 0x000118BE, U"\U000118BE\U000118DE\U000118BE" },
{ 0x000118BF, U"\U000118BF\U000118DF\U000118BF" },
{ 0x000118C0, U"\U000118A0\U000118C0\U000118A0" },
{ 0x000118C1, U"\U000118A1\U000118C1\U000118A1" },
{ 0x000118C2, U"\U000118A2\U000118C2\U000118A2" },
{ 0x000118C3, U"\U000118A3\U000118C3\U000118A3" },
{ 0x000118C4, U"\U000118A4\U000118C4\U000118A4" },
{ 0x000118C5, U"\U000118A5\U000118C5\U000118A5" },
{ 0x000118C6, U"\U000118A6\U000118C6\U000118A6" },
{ 0x000118C7, U"\U000118A7\U000118C7\U000118A7" },
{ 0x000118C8, U"\U000118A8\U000118C8\U000118A8" },
{ 0x000118C9, U"\U000118A9\U000118C9\U000118A9" },
{ 0x000118CA, U"\U000118AA\U000118CA\U000118AA" },
{ 0x000118CB, U"\U000118AB\U000118CB\U000118AB" },
{ 0x000118CC, U"\U000118AC\U000118CC\U000118AC" },
{ 0x000118CD, U"\U000118AD\U000118CD\U000118AD" },
{ 0x000118CE, U"\U000118AE\U000118CE\U000118AE" },
{ 0x000118CF, U"\U000118AF\U000118CF\U000118AF" },
{ 0x000118D0, U"\U000118B0\U000118D0\U000118B0" },
{ 0x000118D1, U"\U000118B1\U000118D1\U000118B1" },
{ 0x000118D2, U"\U000118B2\U000118D2\U000118B2" },
{ 0x000118D3, U"\U000118B3\U000118D3\U000118B3" },
{ 0x000118D4, U"\U000118B4\U000118D4\U000118B4" },
{ 0x000118D5, U"\U000118B5\U000118D5\U000118B5" },
{ 0x000118D6, U"\U000118B6\U000118D6\U000118B6" },
{ 0x000118D7, U"\U000118B7\U000118D7\U000118B7" },
{ 0x000118D8, U"\U000118B8\U000118D8\U000118B8" },
{ 0x000118D9, U"\U000118B9\U000118D9\U000118B9" },
{ 0x000118DA, U"\U000118BA\U000118DA\U000118BA" },
{ 0x000118DB, U"\U000118BB\U000118DB\U000118BB" },
{ 0x000118DC, U"\U000118BC\U000118DC\U000118BC" },
{ 0x000118DD, U"\U000118BD\U000118DD\U000118BD" },
{ 0x000118DE, U"\U000118BE\U000118DE\U000118BE" },
{ 0x000118DF, U"\U000118BF\U000118DF\U000118BF" },
{ 0x00016E40, U"\U00016E40\U00016E60\U00016E40" },
{ 0x00016E41, U"\U00016E41\U00016E61\U00016E41" },
{ 0x00016E42, U"\U00016E42\U00016E62\U00016E42" },
{ 0x00016E43, U"\U00016E43\U00016E63\U00016E43" },
{ 0x00016E44, U"\U00016E44\U00016E64\U00016E44" },
{ 0x00016E45, U"\U00016E45\U00016E65\U00016E45" },
{ 0x00016E46, U"\U00016E46\U00016E66\U00016E46" },
{ 0x00016E47, U"\U00016E47\U00016E67\U00016E47" },
{ 0x00016E48, U"\U00016E48\U00016E68\U00016E48" },
{ 0x00016E49, U"\U00016E49\U00016E69\U00016E49" },
{ 0x00016E4A, U"\U00016E4A\U00016E6A\U00016E4A" },
{ 0x00016E4B, U"\U00016E4B\U00016E6B\U00016E4B" },
{ 0x00016E4C, U"\U00016E4C\U00016E6C\U00016E4C" },
{ 0x00016E4D, U"\U00016E4D\U00016E6D\U00016E4D" },
{ 0x00016E4E, U"\U00016E4E\U00016E6E\U00016E4E" },
{ 0x00016E4F, U"\U00016E4F\U00016E6F\U00016E4F" },
{ 0x00016E50, U"\U00016E50\U00016E70\U00016E50" },
{ 0x00016E51, U"\U00016E51\U00016E71\U00016E51" },
{ 0x00016E52, U"\U00016E52\U00016E72\U00016E52" },
{ 0x00016E53, U"\U00016E53\U00016E73\U00016E53" },
{ 0x00016E54, U"\U00016E54\U00016E74\U00016E54" },
{ 0x00016E55, U"\U00016E55\U00016E75\U00016E55" },
{ 0x00016E56, U"\U00016E56\U00016E76\U00016E56" },
{ 0x00016E57, U"\U00016E57\U00016E77\U00016E57" },
{ 0x00016E58, U"\U00016E58\U00016E78\U00016E58" },
{ 0x00016E59, U"\U00016E59\U00016E79\U00016E59" },
{ 0x00016E5A, U"\U00016E5A\U00016E7A\U00016E5A" },
{ 0x00016E5B, U"\U00016E5B\U00016E7B\U00016E5B" },
{ 0x00016E5C, U"\U00016E5C\U00016E7C\U00016E5C" },
{ 0x00016E5D, U"\U00016E5D\U00016E7D\U00016E5D" },
{ 0x00016E5E, U"\U00016E5E\U00016E7E\U00016E5E" },
{ 0x00016E5F, U"\U00016E5F\U00016E7F\U00016E5F" },
{ 0x00016E60, U"\U00016E40\U00016E60\U00016E40" },
{ 0x00016E61, U"\U00016E41\U00016E61\U00016E41" },
{ 0x00016E62, U"\U00016E42\U00016E62\U00016E42" },
{ 0x00016E63, U"\U00016E43\U00016E63\U00016E43" },
{ 0x00016E64, U"\U00016E44\U00016E64\U00016E44" },
{ 0x00016E65, U"\U00016E45\U00016E65\U00016E45" },
{ 0x00016E66, U"\U00016E46\U00016E66\U00016E46" },
{ 0x00016E67, U"\U00016E47\U00016E67\U00016E47" },
{ 0x00016E68, U"\U00016E48\U00016E68\U00016E48" },
{ 0x00016E69, U"\U00016E49\U00016E69\U00016E49" },
{ 0x00016E6A, U"\U00016E4A\U00016E6A\U00016E4A" },
{ 0x00016E6B, U"\U00016E4B\U00016E6B\U00016E4B" },
{ 0x00016E6C, U"\U00016E4C\U00016E6C\U00016E4C" },
{ 0x00016E6D, U"\U00016E4D\U00016E6D\U00016E4D" },
{ 0x00016E6E, U"\U00016E4E\U00016E6E\U00016E4E" },
{ 0x00016E6F, U"\U00016E4F\U00016E6F\U00016E4F" },
{ 0x00016E70, U"\U00016E50\U00016E70\U00016E50" },
{ 0x00016E71, U"\U00016E51\U00016E71\U00016E51" },
{ 0x00016E72, U"\U00016E52\U00016E72\U00016E52" },
{ 0x00016E73, U"\U00016E53\U00016E73\U00016E53" },
{ 0x00016E74, U"\U00016E54\U00016E74\U00016E54" },
{ 0x00016E75, U"\U00016E55\U00016E75\U00016E55" },
{ 0x00016E76, U"\U00016E56\U00016E76\U00016E56" },
{ 0x00016E77, U"\U00016E57\U00016E77\U00016E57" },
{ 0x00016E78, U"\U00016E58\U00016E78\U00016E58" },
{ 0x00016E79, U"\U00016E59\U00016E79\U00016E59" },
{ 0x00016E7A, U"\U00016E5A\U00016E7A\U00016E5A" },
{ 0x00016E7B, U"\U00016E5B\U00016E7B\U00016E5B" },
{ 0x00016E7C, U"\U00016E5C\U00016E7C\U00016E5C" },
{ 0x00016E7D, U"\U00016E5D\U00016E7D\U00016E5D" },
{ 0x00016E7E, U"\U00016E5E\U00016E7E\U00016E5E" },
{ 0x00016E7F, U"\U00016E5F\U00016E7F\U00016E5F" },
{ 0x0001E900, U"\U0001E900\U0001E922\U0001E900" },
{ 0x0001E901, U"\U0001E901\U0001E923\U0001E901" },
{ 0x0001E902, U"\U0001E902\U0001E924\U0001E902" },
{ 0x0001E903, U"\U0001E903\U0001E925\U0001E903" },
{ 0x0001E904, U"\U0001E904\U0001E926\U0001E904" },
{ 0x0001E905, U"\U0001E905\U0001E927\U0001E905" },
{ 0x0001E906, U"\U0001E906\U0001E928\U0001E906" },
{ 0x0001E907, U"\U0001E907\U0001E929\U0001E907" },
{ 0x0001E908, U"\U0001E908\U0001E92A\U0001E908" },
{ 0x0001E909, U"\U0001E909\U0001E92B\U0001E909" },
{ 0x0001E90A, U"\U0001E90A\U0001E92C\U0001E90A" },
{ 0x0001E90B, U"\U0001E90B\U0001E92D\U0001E90B" },
{ 0x0001E90C, U"\U0001E90C\U0001E92E\U0001E90C" },
{ 0x0001E90D, U"\U0001E90D\U0001E92F\U0001E90D" },
{ 0x0001E90E, U"\U0001E90E\U0001E930\U0001E90E" },
{ 0x0001E90F, U"\U0001E90F\U0001E931\U0001E90F" },
{ 0x0001E910, U"\U0001E910\U0001E932\U0001E910" },
{ 0x0001E911, U"\U0001E911\U0001E933\U0001E911" },
{ 0x0001E912, U"\U0001E912\U0001E934\U0001E912" },
{ 0x0001E913, U"\U0001E913\U0001E935\U0001E913" },
{ 0x0001E914, U"\U0001E914\U0001E936\U0001E914" },
{ 0x0001E915, U"\U0001E915\U0001E937\U0001E915" },
{ 0x0001E916, U"\U0001E916\U0001E938\U0001E916" },
{ 0x0001E917, U"\U0001E917\U0001E939\U0001E917" },
{ 0x0001E918, U"\U0001E918\U0001E93A\U0001E918" },
{ 0x0001E919, U"\U0001E919\U0001E93B\U0001E919" },
{ 0x0001E91A, U"\U0001E91A\U0001E93C\U0001E91A" },
{ 0x0001E91B, U"\U0001E91B\U0001E93D\U0001E91B" },
{ 0x0001E91C, U"\U0001E91C\U0001E93E\U0001E91C" },
{ 0x0001E91D, U"\U0001E91D\U0001E93F\U0001E91D" },
{ 0x0001E91E, U"\U0001E91E\U0001E940\U0001E91E" },
{ 0x0001E91F, U"\U0001E91F\U0001E941\U0001E91F" },
{ 0x0001E920, U"\U0001E920\U0001E942\U0001E920" },
{ 0x0001E921, U"\U0001E921\U0001E943\U0001E921" },
{ 0x0001E922, U"\U0001E900\U0001E922\U0001E900" },
{ 0x0001E923, U"\U0001E901\U0001E923\U0001E901" },
{ 0x0001E924, U"\U0001E902\U0001E924\U0001E902" },
{ 0x0001E925, U"\U0001E903\U0001E925\U0001E903" },
{ 0x0001E926, U"\U0001E904\U0001E926\U0001E904" },
{ 0x0001E927, U"\U0001E905\U0001E927\U0001E905" },
{ 0x0001E928, U"\U0001E906\U0001E928\U0001E906" },
{ 0x0001E929, U"\U0001E907\U0001E929\U0001E907" },
{ 0x0001E92A, U"\U0001E908\U0001E92A\U0001E908" },
{ 0x0001E92B, U"\U0001E909\U0001E92B\U0001E909" },
{ 0x0001E92C, U"\U0001E90A\U0001E92C\U0001E90A" },
{ 0x0001E92D, U"\U0001E90B\U0001E92D\U0001E90B" },
{ 0x0001E92E, U"\U0001E90C\U0001E92E\U0001E90C" },
{ 0x0001E92F, U"\U0001E90D\U0001E92F\U0001E90D" },
{ 0x0001E930, U"\U0001E90E\U0001E930\U0001E90E" },
{ 0x0001E931, U"\U0001E90F\U0001E931\U0001E90F" },
{ 0x0001E932, U"\U0001E910\U0001E932\U0001E910" },
{ 0x0001E933, U"\U0001E911\U0001E933\U0001E911" },
{ 0x0001E934, U"\U0001E912\U0001E934\U0001E912" },
{ 0x0001E935, U"\U0001E913\U0001E935\U0001E913" },
{ 0x0001E936, U"\U0001E914\U0001E936\U0001E914" },
{ 0x0001E937, U"\U0001E915\U0001E937\U0001E915" },
{ 0x0001E938, U"\U0001E916\U0001E938\U0001E916" },
{ 0x0001E939, U"\U0001E917\U0001E939\U0001E917" },
{ 0x0001E93A, U"\U0001E918\U0001E93A\U0001E918" },
{ 0x0001E93B, U"\U0001E919\U0001E93B\U0001E919" },
{ 0x0001E93C, U"\U0001E91A\U0001E93C\U0001E91A" },
{ 0x0001E93D, U"\U0001E91B\U0001E93D\U0001E91B" },
{ 0x0001E93E, U"\U0001E91C\U0001E93E\U0001E91C" },
{ 0x0001E93F, U"\U0001E91D\U0001E93F\U0001E91D" },
{ 0x0001E940, U"\U0001E91E\U0001E940\U0001E91E" },
{ 0x0001E941, U"\U0001E91F\U0001E941\U0001E91F" },
{ 0x0001E942, U"\U0001E920\U0001E942\U0001E920" },
{ 0x0001E943, U"\U0001E921\U0001E943\U0001E921" },
};
const std::unordered_multimap<char32_t, SpecialCasing> _special_case_mappings = {
{ 0x000003A3, { U"\U000003C2", U"\U000003A3", U"\U000003A3", 0, SpecialCasingContext::Final_Sigma } },
{ 0x00000307, { U"\U00000307", 0, 0, "lt", SpecialCasingContext::After_Soft_Dotted } },
{ 0x00000049, { U"\U00000069\U00000307", U"\U00000049", U"\U00000049", "lt", SpecialCasingContext::More_Above } },
{ 0x0000004A, { U"\U0000006A\U00000307", U"\U0000004A", U"\U0000004A", "lt", SpecialCasingContext::More_Above } },
{ 0x0000012E, { U"\U0000012F\U00000307", U"\U0000012E", U"\U0000012E", "lt", SpecialCasingContext::More_Above } },
{ 0x000000CC, { U"\U00000069\U00000307\U00000300", U"\U000000CC", U"\U000000CC", "lt", SpecialCasingContext::Unassigned } },
{ 0x000000CD, { U"\U00000069\U00000307\U00000301", U"\U000000CD", U"\U000000CD", "lt", SpecialCasingContext::Unassigned } },
{ 0x00000128, { U"\U00000069\U00000307\U00000303", U"\U00000128", U"\U00000128", "lt", SpecialCasingContext::Unassigned } },
{ 0x00000130, { U"\U00000069", U"\U00000130", U"\U00000130", "tr", SpecialCasingContext::Unassigned } },
{ 0x00000130, { U"\U00000069", U"\U00000130", U"\U00000130", "az", SpecialCasingContext::Unassigned } },
{ 0x00000307, { 0, U"\U00000307", U"\U00000307", "tr", SpecialCasingContext::After_I } },
{ 0x00000307, { 0, U"\U00000307", U"\U00000307", "az", SpecialCasingContext::After_I } },
{ 0x00000049, { U"\U00000131", U"\U00000049", U"\U00000049", "tr", SpecialCasingContext::Not_Before_Dot } },
{ 0x00000049, { U"\U00000131", U"\U00000049", U"\U00000049", "az", SpecialCasingContext::Not_Before_Dot } },
{ 0x00000069, { U"\U00000069", U"\U00000130", U"\U00000130", "tr", SpecialCasingContext::Unassigned } },
{ 0x00000069, { U"\U00000069", U"\U00000130", U"\U00000130", "az", SpecialCasingContext::Unassigned } },
};
const std::unordered_multimap<char32_t, SpecialCasing> _special_case_mappings_default = {
{ 0x000000DF, { U"\U000000DF", U"\U00000053\U00000073", U"\U00000053\U00000053", 0, SpecialCasingContext::Unassigned } },
{ 0x00000130, { U"\U00000069\U00000307", U"\U00000130", U"\U00000130", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB00, { U"\U0000FB00", U"\U00000046\U00000066", U"\U00000046\U00000046", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB01, { U"\U0000FB01", U"\U00000046\U00000069", U"\U00000046\U00000049", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB02, { U"\U0000FB02", U"\U00000046\U0000006C", U"\U00000046\U0000004C", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB03, { U"\U0000FB03", U"\U00000046\U00000066\U00000069", U"\U00000046\U00000046\U00000049", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB04, { U"\U0000FB04", U"\U00000046\U00000066\U0000006C", U"\U00000046\U00000046\U0000004C", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB05, { U"\U0000FB05", U"\U00000053\U00000074", U"\U00000053\U00000054", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB06, { U"\U0000FB06", U"\U00000053\U00000074", U"\U00000053\U00000054", 0, SpecialCasingContext::Unassigned } },
{ 0x00000587, { U"\U00000587", U"\U00000535\U00000582", U"\U00000535\U00000552", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB13, { U"\U0000FB13", U"\U00000544\U00000576", U"\U00000544\U00000546", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB14, { U"\U0000FB14", U"\U00000544\U00000565", U"\U00000544\U00000535", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB15, { U"\U0000FB15", U"\U00000544\U0000056B", U"\U00000544\U0000053B", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB16, { U"\U0000FB16", U"\U0000054E\U00000576", U"\U0000054E\U00000546", 0, SpecialCasingContext::Unassigned } },
{ 0x0000FB17, { U"\U0000FB17", U"\U00000544\U0000056D", U"\U00000544\U0000053D", 0, SpecialCasingContext::Unassigned } },
{ 0x00000149, { U"\U00000149", U"\U000002BC\U0000004E", U"\U000002BC\U0000004E", 0, SpecialCasingContext::Unassigned } },
{ 0x00000390, { U"\U00000390", U"\U00000399\U00000308\U00000301", U"\U00000399\U00000308\U00000301", 0, SpecialCasingContext::Unassigned } },
{ 0x000003B0, { U"\U000003B0", U"\U000003A5\U00000308\U00000301", U"\U000003A5\U00000308\U00000301", 0, SpecialCasingContext::Unassigned } },
{ 0x000001F0, { U"\U000001F0", U"\U0000004A\U0000030C", U"\U0000004A\U0000030C", 0, SpecialCasingContext::Unassigned } },
{ 0x00001E96, { U"\U00001E96", U"\U00000048\U00000331", U"\U00000048\U00000331", 0, SpecialCasingContext::Unassigned } },
{ 0x00001E97, { U"\U00001E97", U"\U00000054\U00000308", U"\U00000054\U00000308", 0, SpecialCasingContext::Unassigned } },
{ 0x00001E98, { U"\U00001E98", U"\U00000057\U0000030A", U"\U00000057\U0000030A", 0, SpecialCasingContext::Unassigned } },
{ 0x00001E99, { U"\U00001E99", U"\U00000059\U0000030A", U"\U00000059\U0000030A", 0, SpecialCasingContext::Unassigned } },
{ 0x00001E9A, { U"\U00001E9A", U"\U00000041\U000002BE", U"\U00000041\U000002BE", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F50, { U"\U00001F50", U"\U000003A5\U00000313", U"\U000003A5\U00000313", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F52, { U"\U00001F52", U"\U000003A5\U00000313\U00000300", U"\U000003A5\U00000313\U00000300", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F54, { U"\U00001F54", U"\U000003A5\U00000313\U00000301", U"\U000003A5\U00000313\U00000301", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F56, { U"\U00001F56", U"\U000003A5\U00000313\U00000342", U"\U000003A5\U00000313\U00000342", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FB6, { U"\U00001FB6", U"\U00000391\U00000342", U"\U00000391\U00000342", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FC6, { U"\U00001FC6", U"\U00000397\U00000342", U"\U00000397\U00000342", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FD2, { U"\U00001FD2", U"\U00000399\U00000308\U00000300", U"\U00000399\U00000308\U00000300", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FD3, { U"\U00001FD3", U"\U00000399\U00000308\U00000301", U"\U00000399\U00000308\U00000301", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FD6, { U"\U00001FD6", U"\U00000399\U00000342", U"\U00000399\U00000342", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FD7, { U"\U00001FD7", U"\U00000399\U00000308\U00000342", U"\U00000399\U00000308\U00000342", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FE2, { U"\U00001FE2", U"\U000003A5\U00000308\U00000300", U"\U000003A5\U00000308\U00000300", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FE3, { U"\U00001FE3", U"\U000003A5\U00000308\U00000301", U"\U000003A5\U00000308\U00000301", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FE4, { U"\U00001FE4", U"\U000003A1\U00000313", U"\U000003A1\U00000313", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FE6, { U"\U00001FE6", U"\U000003A5\U00000342", U"\U000003A5\U00000342", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FE7, { U"\U00001FE7", U"\U000003A5\U00000308\U00000342", U"\U000003A5\U00000308\U00000342", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FF6, { U"\U00001FF6", U"\U000003A9\U00000342", U"\U000003A9\U00000342", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F80, { U"\U00001F80", U"\U00001F88", U"\U00001F08\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F81, { U"\U00001F81", U"\U00001F89", U"\U00001F09\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F82, { U"\U00001F82", U"\U00001F8A", U"\U00001F0A\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F83, { U"\U00001F83", U"\U00001F8B", U"\U00001F0B\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F84, { U"\U00001F84", U"\U00001F8C", U"\U00001F0C\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F85, { U"\U00001F85", U"\U00001F8D", U"\U00001F0D\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F86, { U"\U00001F86", U"\U00001F8E", U"\U00001F0E\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F87, { U"\U00001F87", U"\U00001F8F", U"\U00001F0F\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F88, { U"\U00001F80", U"\U00001F88", U"\U00001F08\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F89, { U"\U00001F81", U"\U00001F89", U"\U00001F09\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F8A, { U"\U00001F82", U"\U00001F8A", U"\U00001F0A\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F8B, { U"\U00001F83", U"\U00001F8B", U"\U00001F0B\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F8C, { U"\U00001F84", U"\U00001F8C", U"\U00001F0C\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F8D, { U"\U00001F85", U"\U00001F8D", U"\U00001F0D\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F8E, { U"\U00001F86", U"\U00001F8E", U"\U00001F0E\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F8F, { U"\U00001F87", U"\U00001F8F", U"\U00001F0F\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F90, { U"\U00001F90", U"\U00001F98", U"\U00001F28\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F91, { U"\U00001F91", U"\U00001F99", U"\U00001F29\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F92, { U"\U00001F92", U"\U00001F9A", U"\U00001F2A\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F93, { U"\U00001F93", U"\U00001F9B", U"\U00001F2B\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F94, { U"\U00001F94", U"\U00001F9C", U"\U00001F2C\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F95, { U"\U00001F95", U"\U00001F9D", U"\U00001F2D\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F96, { U"\U00001F96", U"\U00001F9E", U"\U00001F2E\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F97, { U"\U00001F97", U"\U00001F9F", U"\U00001F2F\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F98, { U"\U00001F90", U"\U00001F98", U"\U00001F28\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F99, { U"\U00001F91", U"\U00001F99", U"\U00001F29\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F9A, { U"\U00001F92", U"\U00001F9A", U"\U00001F2A\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F9B, { U"\U00001F93", U"\U00001F9B", U"\U00001F2B\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F9C, { U"\U00001F94", U"\U00001F9C", U"\U00001F2C\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F9D, { U"\U00001F95", U"\U00001F9D", U"\U00001F2D\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F9E, { U"\U00001F96", U"\U00001F9E", U"\U00001F2E\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001F9F, { U"\U00001F97", U"\U00001F9F", U"\U00001F2F\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FA0, { U"\U00001FA0", U"\U00001FA8", U"\U00001F68\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FA1, { U"\U00001FA1", U"\U00001FA9", U"\U00001F69\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FA2, { U"\U00001FA2", U"\U00001FAA", U"\U00001F6A\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FA3, { U"\U00001FA3", U"\U00001FAB", U"\U00001F6B\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FA4, { U"\U00001FA4", U"\U00001FAC", U"\U00001F6C\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FA5, { U"\U00001FA5", U"\U00001FAD", U"\U00001F6D\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FA6, { U"\U00001FA6", U"\U00001FAE", U"\U00001F6E\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FA7, { U"\U00001FA7", U"\U00001FAF", U"\U00001F6F\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FA8, { U"\U00001FA0", U"\U00001FA8", U"\U00001F68\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FA9, { U"\U00001FA1", U"\U00001FA9", U"\U00001F69\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FAA, { U"\U00001FA2", U"\U00001FAA", U"\U00001F6A\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FAB, { U"\U00001FA3", U"\U00001FAB", U"\U00001F6B\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FAC, { U"\U00001FA4", U"\U00001FAC", U"\U00001F6C\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FAD, { U"\U00001FA5", U"\U00001FAD", U"\U00001F6D\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FAE, { U"\U00001FA6", U"\U00001FAE", U"\U00001F6E\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FAF, { U"\U00001FA7", U"\U00001FAF", U"\U00001F6F\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FB3, { U"\U00001FB3", U"\U00001FBC", U"\U00000391\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FBC, { U"\U00001FB3", U"\U00001FBC", U"\U00000391\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FC3, { U"\U00001FC3", U"\U00001FCC", U"\U00000397\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FCC, { U"\U00001FC3", U"\U00001FCC", U"\U00000397\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FF3, { U"\U00001FF3", U"\U00001FFC", U"\U000003A9\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FFC, { U"\U00001FF3", U"\U00001FFC", U"\U000003A9\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FB2, { U"\U00001FB2", U"\U00001FBA\U00000345", U"\U00001FBA\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FB4, { U"\U00001FB4", U"\U00000386\U00000345", U"\U00000386\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FC2, { U"\U00001FC2", U"\U00001FCA\U00000345", U"\U00001FCA\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FC4, { U"\U00001FC4", U"\U00000389\U00000345", U"\U00000389\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FF2, { U"\U00001FF2", U"\U00001FFA\U00000345", U"\U00001FFA\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FF4, { U"\U00001FF4", U"\U0000038F\U00000345", U"\U0000038F\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FB7, { U"\U00001FB7", U"\U00000391\U00000342\U00000345", U"\U00000391\U00000342\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FC7, { U"\U00001FC7", U"\U00000397\U00000342\U00000345", U"\U00000397\U00000342\U00000399", 0, SpecialCasingContext::Unassigned } },
{ 0x00001FF7, { U"\U00001FF7", U"\U000003A9\U00000342\U00000345", U"\U000003A9\U00000342\U00000399", 0, SpecialCasingContext::Unassigned } },
};
const std::unordered_map<char32_t, CaseFolding> _case_foldings = {
{ 0x00000041, { 0x00000061,  0x00000000, 0, 0x00000000 } },
{ 0x00000042, { 0x00000062,  0x00000000, 0, 0x00000000 } },
{ 0x00000043, { 0x00000063,  0x00000000, 0, 0x00000000 } },
{ 0x00000044, { 0x00000064,  0x00000000, 0, 0x00000000 } },
{ 0x00000045, { 0x00000065,  0x00000000, 0, 0x00000000 } },
{ 0x00000046, { 0x00000066,  0x00000000, 0, 0x00000000 } },
{ 0x00000047, { 0x00000067,  0x00000000, 0, 0x00000000 } },
{ 0x00000048, { 0x00000068,  0x00000000, 0, 0x00000000 } },
{ 0x00000049, { 0x00000069,  0x00000000, 0, 0x00000131 } },
{ 0x0000004A, { 0x0000006A,  0x00000000, 0, 0x00000000 } },
{ 0x0000004B, { 0x0000006B,  0x00000000, 0, 0x00000000 } },
{ 0x0000004C, { 0x0000006C,  0x00000000, 0, 0x00000000 } },
{ 0x0000004D, { 0x0000006D,  0x00000000, 0, 0x00000000 } },
{ 0x0000004E, { 0x0000006E,  0x00000000, 0, 0x00000000 } },
{ 0x0000004F, { 0x0000006F,  0x00000000, 0, 0x00000000 } },
{ 0x00000050, { 0x00000070,  0x00000000, 0, 0x00000000 } },
{ 0x00000051, { 0x00000071,  0x00000000, 0, 0x00000000 } },
{ 0x00000052, { 0x00000072,  0x00000000, 0, 0x00000000 } },
{ 0x00000053, { 0x00000073,  0x00000000, 0, 0x00000000 } },
{ 0x00000054, { 0x00000074,  0x00000000, 0, 0x00000000 } },
{ 0x00000055, { 0x00000075,  0x00000000, 0, 0x00000000 } },
{ 0x00000056, { 0x00000076,  0x00000000, 0, 0x00000000 } },
{ 0x00000057, { 0x00000077,  0x00000000, 0, 0x00000000 } },
{ 0x00000058, { 0x00000078,  0x00000000, 0, 0x00000000 } },
{ 0x00000059, { 0x00000079,  0x00000000, 0, 0x00000000 } },
{ 0x0000005A, { 0x0000007A,  0x00000000, 0, 0x00000000 } },
{ 0x000000B5, { 0x000003BC,  0x00000000, 0, 0x00000000 } },
{ 0x000000C0, { 0x000000E0,  0x00000000, 0, 0x00000000 } },
{ 0x000000C1, { 0x000000E1,  0x00000000, 0, 0x00000000 } },
{ 0x000000C2, { 0x000000E2,  0x00000000, 0, 0x00000000 } },
{ 0x000000C3, { 0x000000E3,  0x00000000, 0, 0x00000000 } },
{ 0x000000C4, { 0x000000E4,  0x00000000, 0, 0x00000000 } },
{ 0x000000C5, { 0x000000E5,  0x00000000, 0, 0x00000000 } },
{ 0x000000C6, { 0x000000E6,  0x00000000, 0, 0x00000000 } },
{ 0x000000C7, { 0x000000E7,  0x00000000, 0, 0x00000000 } },
{ 0x000000C8, { 0x000000E8,  0x00000000, 0, 0x00000000 } },
{ 0x000000C9, { 0x000000E9,  0x00000000, 0, 0x00000000 } },
{ 0x000000CA, { 0x000000EA,  0x00000000, 0, 0x00000000 } },
{ 0x000000CB, { 0x000000EB,  0x00000000, 0, 0x00000000 } },
{ 0x000000CC, { 0x000000EC,  0x00000000, 0, 0x00000000 } },
{ 0x000000CD, { 0x000000ED,  0x00000000, 0, 0x00000000 } },
{ 0x000000CE, { 0x000000EE,  0x00000000, 0, 0x00000000 } },
{ 0x000000CF, { 0x000000EF,  0x00000000, 0, 0x00000000 } },
{ 0x000000D0, { 0x000000F0,  0x00000000, 0, 0x00000000 } },
{ 0x000000D1, { 0x000000F1,  0x00000000, 0, 0x00000000 } },
{ 0x000000D2, { 0x000000F2,  0x00000000, 0, 0x00000000 } },
{ 0x000000D3, { 0x000000F3,  0x00000000, 0, 0x00000000 } },
{ 0x000000D4, { 0x000000F4,  0x00000000, 0, 0x00000000 } },
{ 0x000000D5, { 0x000000F5,  0x00000000, 0, 0x00000000 } },
{ 0x000000D6, { 0x000000F6,  0x00000000, 0, 0x00000000 } },
{ 0x000000D8, { 0x000000F8,  0x00000000, 0, 0x00000000 } },
{ 0x000000D9, { 0x000000F9,  0x00000000, 0, 0x00000000 } },
{ 0x000000DA, { 0x000000FA,  0x00000000, 0, 0x00000000 } },
{ 0x000000DB, { 0x000000FB,  0x00000000, 0, 0x00000000 } },
{ 0x000000DC, { 0x000000FC,  0x00000000, 0, 0x00000000 } },
{ 0x000000DD, { 0x000000FD,  0x00000000, 0, 0x00000000 } },
{ 0x000000DE, { 0x000000FE,  0x00000000, 0, 0x00000000 } },
{ 0x000000DF, { 0x00000000,  0x00000000, U"\U00000073\U00000073", 0x00000000 } },
{ 0x00000100, { 0x00000101,  0x00000000, 0, 0x00000000 } },
{ 0x00000102, { 0x00000103,  0x00000000, 0, 0x00000000 } },
{ 0x00000104, { 0x00000105,  0x00000000, 0, 0x00000000 } },
{ 0x00000106, { 0x00000107,  0x00000000, 0, 0x00000000 } },
{ 0x00000108, { 0x00000109,  0x00000000, 0, 0x00000000 } },
{ 0x0000010A, { 0x0000010B,  0x00000000, 0, 0x00000000 } },
{ 0x0000010C, { 0x0000010D,  0x00000000, 0, 0x00000000 } },
{ 0x0000010E, { 0x0000010F,  0x00000000, 0, 0x00000000 } },
{ 0x00000110, { 0x00000111,  0x00000000, 0, 0x00000000 } },
{ 0x00000112, { 0x00000113,  0x00000000, 0, 0x00000000 } },
{ 0x00000114, { 0x00000115,  0x00000000, 0, 0x00000000 } },
{ 0x00000116, { 0x00000117,  0x00000000, 0, 0x00000000 } },
{ 0x00000118, { 0x00000119,  0x00000000, 0, 0x00000000 } },
{ 0x0000011A, { 0x0000011B,  0x00000000, 0, 0x00000000 } },
{ 0x0000011C, { 0x0000011D,  0x00000000, 0, 0x00000000 } },
{ 0x0000011E, { 0x0000011F,  0x00000000, 0, 0x00000000 } },
{ 0x00000120, { 0x00000121,  0x00000000, 0, 0x00000000 } },
{ 0x00000122, { 0x00000123,  0x00000000, 0, 0x00000000 } },
{ 0x00000124, { 0x00000125,  0x00000000, 0, 0x00000000 } },
{ 0x00000126, { 0x00000127,  0x00000000, 0, 0x00000000 } },
{ 0x00000128, { 0x00000129,  0x00000000, 0, 0x00000000 } },
{ 0x0000012A, { 0x0000012B,  0x00000000, 0, 0x00000000 } },
{ 0x0000212B, { 0x000000E5,  0x00000000, 0, 0x00000000 } },
{ 0x0000012C, { 0x0000012D,  0x00000000, 0, 0x00000000 } },
{ 0x0000012E, { 0x0000012F,  0x00000000, 0, 0x00000000 } },
{ 0x00000130, { 0x00000000,  0x00000000, U"\U00000069\U00000307", 0x00000069 } },
{ 0x00000132, { 0x00000133,  0x00000000, 0, 0x00000000 } },
{ 0x00000134, { 0x00000135,  0x00000000, 0, 0x00000000 } },
{ 0x00000136, { 0x00000137,  0x00000000, 0, 0x00000000 } },
{ 0x00000139, { 0x0000013A,  0x00000000, 0, 0x00000000 } },
{ 0x0000013B, { 0x0000013C,  0x00000000, 0, 0x00000000 } },
{ 0x0000013D, { 0x0000013E,  0x00000000, 0, 0x00000000 } },
{ 0x0000013F, { 0x00000140,  0x00000000, 0, 0x00000000 } },
{ 0x00000141, { 0x00000142,  0x00000000, 0, 0x00000000 } },
{ 0x00000143, { 0x00000144,  0x00000000, 0, 0x00000000 } },
{ 0x00000145, { 0x00000146,  0x00000000, 0, 0x00000000 } },
{ 0x00000147, { 0x00000148,  0x00000000, 0, 0x00000000 } },
{ 0x00000149, { 0x00000000,  0x00000000, U"\U000002BC\U0000006E", 0x00000000 } },
{ 0x0000014A, { 0x0000014B,  0x00000000, 0, 0x00000000 } },
{ 0x0000014C, { 0x0000014D,  0x00000000, 0, 0x00000000 } },
{ 0x0000014E, { 0x0000014F,  0x00000000, 0, 0x00000000 } },
{ 0x00000150, { 0x00000151,  0x00000000, 0, 0x00000000 } },
{ 0x00000152, { 0x00000153,  0x00000000, 0, 0x00000000 } },
{ 0x00000154, { 0x00000155,  0x00000000, 0, 0x00000000 } },
{ 0x00000156, { 0x00000157,  0x00000000, 0, 0x00000000 } },
{ 0x00000158, { 0x00000159,  0x00000000, 0, 0x00000000 } },
{ 0x0000015A, { 0x0000015B,  0x00000000, 0, 0x00000000 } },
{ 0x0000015C, { 0x0000015D,  0x00000000, 0, 0x00000000 } },
{ 0x0000015E, { 0x0000015F,  0x00000000, 0, 0x00000000 } },
{ 0x00000160, { 0x00000161,  0x00000000, 0, 0x00000000 } },
{ 0x00002161, { 0x00002171,  0x00000000, 0, 0x00000000 } },
{ 0x00000162, { 0x00000163,  0x00000000, 0, 0x00000000 } },
{ 0x00002163, { 0x00002173,  0x00000000, 0, 0x00000000 } },
{ 0x00000164, { 0x00000165,  0x00000000, 0, 0x00000000 } },
{ 0x00002165, { 0x00002175,  0x00000000, 0, 0x00000000 } },
{ 0x00002166, { 0x00002176,  0x00000000, 0, 0x00000000 } },
{ 0x00002167, { 0x00002177,  0x00000000, 0, 0x00000000 } },
{ 0x00002168, { 0x00002178,  0x00000000, 0, 0x00000000 } },
{ 0x00002169, { 0x00002179,  0x00000000, 0, 0x00000000 } },
{ 0x0000216A, { 0x0000217A,  0x00000000, 0, 0x00000000 } },
{ 0x0000216B, { 0x0000217B,  0x00000000, 0, 0x00000000 } },
{ 0x0000216C, { 0x0000217C,  0x00000000, 0, 0x00000000 } },
{ 0x0000216D, { 0x0000217D,  0x00000000, 0, 0x00000000 } },
{ 0x0000216E, { 0x0000217E,  0x00000000, 0, 0x00000000 } },
{ 0x0000216F, { 0x0000217F,  0x00000000, 0, 0x00000000 } },
{ 0x00000170, { 0x00000171,  0x00000000, 0, 0x00000000 } },
{ 0x00000172, { 0x00000173,  0x00000000, 0, 0x00000000 } },
{ 0x00000174, { 0x00000175,  0x00000000, 0, 0x00000000 } },
{ 0x00000176, { 0x00000177,  0x00000000, 0, 0x00000000 } },
{ 0x00000178, { 0x000000FF,  0x00000000, 0, 0x00000000 } },
{ 0x00000179, { 0x0000017A,  0x00000000, 0, 0x00000000 } },
{ 0x0000017B, { 0x0000017C,  0x00000000, 0, 0x00000000 } },
{ 0x0000017D, { 0x0000017E,  0x00000000, 0, 0x00000000 } },
{ 0x0000017F, { 0x00000073,  0x00000000, 0, 0x00000000 } },
{ 0x00000181, { 0x00000253,  0x00000000, 0, 0x00000000 } },
{ 0x00000182, { 0x00000183,  0x00000000, 0, 0x00000000 } },
{ 0x00002183, { 0x00002184,  0x00000000, 0, 0x00000000 } },
{ 0x00000184, { 0x00000185,  0x00000000, 0, 0x00000000 } },
{ 0x00000186, { 0x00000254,  0x00000000, 0, 0x00000000 } },
{ 0x00000187, { 0x00000188,  0x00000000, 0, 0x00000000 } },
{ 0x00000189, { 0x00000256,  0x00000000, 0, 0x00000000 } },
{ 0x0000018A, { 0x00000257,  0x00000000, 0, 0x00000000 } },
{ 0x0000018B, { 0x0000018C,  0x00000000, 0, 0x00000000 } },
{ 0x0000018E, { 0x000001DD,  0x00000000, 0, 0x00000000 } },
{ 0x0000018F, { 0x00000259,  0x00000000, 0, 0x00000000 } },
{ 0x00000190, { 0x0000025B,  0x00000000, 0, 0x00000000 } },
{ 0x00000191, { 0x00000192,  0x00000000, 0, 0x00000000 } },
{ 0x00000193, { 0x00000260,  0x00000000, 0, 0x00000000 } },
{ 0x00000194, { 0x00000263,  0x00000000, 0, 0x00000000 } },
{ 0x00000196, { 0x00000269,  0x00000000, 0, 0x00000000 } },
{ 0x00000197, { 0x00000268,  0x00000000, 0, 0x00000000 } },
{ 0x00000198, { 0x00000199,  0x00000000, 0, 0x00000000 } },
{ 0x0000019C, { 0x0000026F,  0x00000000, 0, 0x00000000 } },
{ 0x0000019D, { 0x00000272,  0x00000000, 0, 0x00000000 } },
{ 0x0000019F, { 0x00000275,  0x00000000, 0, 0x00000000 } },
{ 0x000001A0, { 0x000001A1,  0x00000000, 0, 0x00000000 } },
{ 0x000001A2, { 0x000001A3,  0x00000000, 0, 0x00000000 } },
{ 0x000001A4, { 0x000001A5,  0x00000000, 0, 0x00000000 } },
{ 0x000001A6, { 0x00000280,  0x00000000, 0, 0x00000000 } },
{ 0x000001A7, { 0x000001A8,  0x00000000, 0, 0x00000000 } },
{ 0x000001A9, { 0x00000283,  0x00000000, 0, 0x00000000 } },
{ 0x000001AC, { 0x000001AD,  0x00000000, 0, 0x00000000 } },
{ 0x000001AE, { 0x00000288,  0x00000000, 0, 0x00000000 } },
{ 0x000001AF, { 0x000001B0,  0x00000000, 0, 0x00000000 } },
{ 0x000001B1, { 0x0000028A,  0x00000000, 0, 0x00000000 } },
{ 0x000001B2, { 0x0000028B,  0x00000000, 0, 0x00000000 } },
{ 0x000001B3, { 0x000001B4,  0x00000000, 0, 0x00000000 } },
{ 0x000001B5, { 0x000001B6,  0x00000000, 0, 0x00000000 } },
{ 0x000001B7, { 0x00000292,  0x00000000, 0, 0x00000000 } },
{ 0x000001B8, { 0x000001B9,  0x00000000, 0, 0x00000000 } },
{ 0x000001BC, { 0x000001BD,  0x00000000, 0, 0x00000000 } },
{ 0x000001C4, { 0x000001C6,  0x00000000, 0, 0x00000000 } },
{ 0x000001C5, { 0x000001C6,  0x00000000, 0, 0x00000000 } },
{ 0x000001C7, { 0x000001C9,  0x00000000, 0, 0x00000000 } },
{ 0x000001C8, { 0x000001C9,  0x00000000, 0, 0x00000000 } },
{ 0x000001CA, { 0x000001CC,  0x00000000, 0, 0x00000000 } },
{ 0x000001CB, { 0x000001CC,  0x00000000, 0, 0x00000000 } },
{ 0x000001CD, { 0x000001CE,  0x00000000, 0, 0x00000000 } },
{ 0x000001CF, { 0x000001D0,  0x00000000, 0, 0x00000000 } },
{ 0x000001D1, { 0x000001D2,  0x00000000, 0, 0x00000000 } },
{ 0x000001D3, { 0x000001D4,  0x00000000, 0, 0x00000000 } },
{ 0x000001D5, { 0x000001D6,  0x00000000, 0, 0x00000000 } },
{ 0x000001D7, { 0x000001D8,  0x00000000, 0, 0x00000000 } },
{ 0x000001D9, { 0x000001DA,  0x00000000, 0, 0x00000000 } },
{ 0x000001DB, { 0x000001DC,  0x00000000, 0, 0x00000000 } },
{ 0x000001DE, { 0x000001DF,  0x00000000, 0, 0x00000000 } },
{ 0x000001E0, { 0x000001E1,  0x00000000, 0, 0x00000000 } },
{ 0x000001E2, { 0x000001E3,  0x00000000, 0, 0x00000000 } },
{ 0x000001E4, { 0x000001E5,  0x00000000, 0, 0x00000000 } },
{ 0x000001E6, { 0x000001E7,  0x00000000, 0, 0x00000000 } },
{ 0x000001E8, { 0x000001E9,  0x00000000, 0, 0x00000000 } },
{ 0x000001EA, { 0x000001EB,  0x00000000, 0, 0x00000000 } },
{ 0x000001EC, { 0x000001ED,  0x00000000, 0, 0x00000000 } },
{ 0x000001EE, { 0x000001EF,  0x00000000, 0, 0x00000000 } },
{ 0x000001F0, { 0x00000000,  0x00000000, U"\U0000006A\U0000030C", 0x00000000 } },
{ 0x000001F1, { 0x000001F3,  0x00000000, 0, 0x00000000 } },
{ 0x000001F2, { 0x000001F3,  0x00000000, 0, 0x00000000 } },
{ 0x000001F4, { 0x000001F5,  0x00000000, 0, 0x00000000 } },
{ 0x000001F6, { 0x00000195,  0x00000000, 0, 0x00000000 } },
{ 0x000001F7, { 0x000001BF,  0x00000000, 0, 0x00000000 } },
{ 0x000001F8, { 0x000001F9,  0x00000000, 0, 0x00000000 } },
{ 0x000001FA, { 0x000001FB,  0x00000000, 0, 0x00000000 } },
{ 0x000001FC, { 0x000001FD,  0x00000000, 0, 0x00000000 } },
{ 0x000001FE, { 0x000001FF,  0x00000000, 0, 0x00000000 } },
{ 0x00000200, { 0x00000201,  0x00000000, 0, 0x00000000 } },
{ 0x00000202, { 0x00000203,  0x00000000, 0, 0x00000000 } },
{ 0x00000204, { 0x00000205,  0x00000000, 0, 0x00000000 } },
{ 0x00000206, { 0x00000207,  0x00000000, 0, 0x00000000 } },
{ 0x00000208, { 0x00000209,  0x00000000, 0, 0x00000000 } },
{ 0x0000020A, { 0x0000020B,  0x00000000, 0, 0x00000000 } },
{ 0x0000020C, { 0x0000020D,  0x00000000, 0, 0x00000000 } },
{ 0x0000020E, { 0x0000020F,  0x00000000, 0, 0x00000000 } },
{ 0x00000210, { 0x00000211,  0x00000000, 0, 0x00000000 } },
{ 0x00000212, { 0x00000213,  0x00000000, 0, 0x00000000 } },
{ 0x00000214, { 0x00000215,  0x00000000, 0, 0x00000000 } },
{ 0x00000216, { 0x00000217,  0x00000000, 0, 0x00000000 } },
{ 0x00000218, { 0x00000219,  0x00000000, 0, 0x00000000 } },
{ 0x0000021A, { 0x0000021B,  0x00000000, 0, 0x00000000 } },
{ 0x0000021C, { 0x0000021D,  0x00000000, 0, 0x00000000 } },
{ 0x0000021E, { 0x0000021F,  0x00000000, 0, 0x00000000 } },
{ 0x00000220, { 0x0000019E,  0x00000000, 0, 0x00000000 } },
{ 0x00000222, { 0x00000223,  0x00000000, 0, 0x00000000 } },
{ 0x00000224, { 0x00000225,  0x00000000, 0, 0x00000000 } },
{ 0x00000226, { 0x00000227,  0x00000000, 0, 0x00000000 } },
{ 0x00000228, { 0x00000229,  0x00000000, 0, 0x00000000 } },
{ 0x0000022A, { 0x0000022B,  0x00000000, 0, 0x00000000 } },
{ 0x0000022C, { 0x0000022D,  0x00000000, 0, 0x00000000 } },
{ 0x0000022E, { 0x0000022F,  0x00000000, 0, 0x00000000 } },
{ 0x00000230, { 0x00000231,  0x00000000, 0, 0x00000000 } },
{ 0x00000232, { 0x00000233,  0x00000000, 0, 0x00000000 } },
{ 0x0000023A, { 0x00002C65,  0x00000000, 0, 0x00000000 } },
{ 0x0000023B, { 0x0000023C,  0x00000000, 0, 0x00000000 } },
{ 0x0000023D, { 0x0000019A,  0x00000000, 0, 0x00000000 } },
{ 0x0000023E, { 0x00002C66,  0x00000000, 0, 0x00000000 } },
{ 0x00000241, { 0x00000242,  0x00000000, 0, 0x00000000 } },
{ 0x00000243, { 0x00000180,  0x00000000, 0, 0x00000000 } },
{ 0x00000244, { 0x00000289,  0x00000000, 0, 0x00000000 } },
{ 0x00000245, { 0x0000028C,  0x00000000, 0, 0x00000000 } },
{ 0x00000246, { 0x00000247,  0x00000000, 0, 0x00000000 } },
{ 0x00000248, { 0x00000249,  0x00000000, 0, 0x00000000 } },
{ 0x0000024A, { 0x0000024B,  0x00000000, 0, 0x00000000 } },
{ 0x0000024C, { 0x0000024D,  0x00000000, 0, 0x00000000 } },
{ 0x0000024E, { 0x0000024F,  0x00000000, 0, 0x00000000 } },
{ 0x00010C9A, { 0x00010CDA,  0x00000000, 0, 0x00000000 } },
{ 0x00010C9C, { 0x00010CDC,  0x00000000, 0, 0x00000000 } },
{ 0x00010C9E, { 0x00010CDE,  0x00000000, 0, 0x00000000 } },
{ 0x00000345, { 0x000003B9,  0x00000000, 0, 0x00000000 } },
{ 0x00000370, { 0x00000371,  0x00000000, 0, 0x00000000 } },
{ 0x00000372, { 0x00000373,  0x00000000, 0, 0x00000000 } },
{ 0x00000376, { 0x00000377,  0x00000000, 0, 0x00000000 } },
{ 0x0000037F, { 0x000003F3,  0x00000000, 0, 0x00000000 } },
{ 0x00000386, { 0x000003AC,  0x00000000, 0, 0x00000000 } },
{ 0x00000388, { 0x000003AD,  0x00000000, 0, 0x00000000 } },
{ 0x00000389, { 0x000003AE,  0x00000000, 0, 0x00000000 } },
{ 0x0000038A, { 0x000003AF,  0x00000000, 0, 0x00000000 } },
{ 0x0000038C, { 0x000003CC,  0x00000000, 0, 0x00000000 } },
{ 0x0000038E, { 0x000003CD,  0x00000000, 0, 0x00000000 } },
{ 0x0000038F, { 0x000003CE,  0x00000000, 0, 0x00000000 } },
{ 0x00000390, { 0x00000000,  0x00000000, U"\U000003B9\U00000308\U00000301", 0x00000000 } },
{ 0x00000391, { 0x000003B1,  0x00000000, 0, 0x00000000 } },
{ 0x00000392, { 0x000003B2,  0x00000000, 0, 0x00000000 } },
{ 0x00000393, { 0x000003B3,  0x00000000, 0, 0x00000000 } },
{ 0x00000394, { 0x000003B4,  0x00000000, 0, 0x00000000 } },
{ 0x00000395, { 0x000003B5,  0x00000000, 0, 0x00000000 } },
{ 0x00000396, { 0x000003B6,  0x00000000, 0, 0x00000000 } },
{ 0x00000397, { 0x000003B7,  0x00000000, 0, 0x00000000 } },
{ 0x00000398, { 0x000003B8,  0x00000000, 0, 0x00000000 } },
{ 0x00000399, { 0x000003B9,  0x00000000, 0, 0x00000000 } },
{ 0x0000039A, { 0x000003BA,  0x00000000, 0, 0x00000000 } },
{ 0x0000039B, { 0x000003BB,  0x00000000, 0, 0x00000000 } },
{ 0x0000039C, { 0x000003BC,  0x00000000, 0, 0x00000000 } },
{ 0x0000039D, { 0x000003BD,  0x00000000, 0, 0x00000000 } },
{ 0x0000039E, { 0x000003BE,  0x00000000, 0, 0x00000000 } },
{ 0x0000039F, { 0x000003BF,  0x00000000, 0, 0x00000000 } },
{ 0x000003A0, { 0x000003C0,  0x00000000, 0, 0x00000000 } },
{ 0x000003A1, { 0x000003C1,  0x00000000, 0, 0x00000000 } },
{ 0x000003A3, { 0x000003C3,  0x00000000, 0, 0x00000000 } },
{ 0x000003A4, { 0x000003C4,  0x00000000, 0, 0x00000000 } },
{ 0x000003A5, { 0x000003C5,  0x00000000, 0, 0x00000000 } },
{ 0x000003A6, { 0x000003C6,  0x00000000, 0, 0x00000000 } },
{ 0x000003A7, { 0x000003C7,  0x00000000, 0, 0x00000000 } },
{ 0x000003A8, { 0x000003C8,  0x00000000, 0, 0x00000000 } },
{ 0x000003A9, { 0x000003C9,  0x00000000, 0, 0x00000000 } },
{ 0x000003AA, { 0x000003CA,  0x00000000, 0, 0x00000000 } },
{ 0x000003AB, { 0x000003CB,  0x00000000, 0, 0x00000000 } },
{ 0x000003B0, { 0x00000000,  0x00000000, U"\U000003C5\U00000308\U00000301", 0x00000000 } },
{ 0x000003C2, { 0x000003C3,  0x00000000, 0, 0x00000000 } },
{ 0x000003CF, { 0x000003D7,  0x00000000, 0, 0x00000000 } },
{ 0x000003D0, { 0x000003B2,  0x00000000, 0, 0x00000000 } },
{ 0x000003D1, { 0x000003B8,  0x00000000, 0, 0x00000000 } },
{ 0x000003D5, { 0x000003C6,  0x00000000, 0, 0x00000000 } },
{ 0x000003D6, { 0x000003C0,  0x00000000, 0, 0x00000000 } },
{ 0x000003D8, { 0x000003D9,  0x00000000, 0, 0x00000000 } },
{ 0x000003DA, { 0x000003DB,  0x00000000, 0, 0x00000000 } },
{ 0x000003DC, { 0x000003DD,  0x00000000, 0, 0x00000000 } },
{ 0x000003DE, { 0x000003DF,  0x00000000, 0, 0x00000000 } },
{ 0x000003E0, { 0x000003E1,  0x00000000, 0, 0x00000000 } },
{ 0x000003E2, { 0x000003E3,  0x00000000, 0, 0x00000000 } },
{ 0x000003E4, { 0x000003E5,  0x00000000, 0, 0x00000000 } },
{ 0x000003E6, { 0x000003E7,  0x00000000, 0, 0x00000000 } },
{ 0x000003E8, { 0x000003E9,  0x00000000, 0, 0x00000000 } },
{ 0x000003EA, { 0x000003EB,  0x00000000, 0, 0x00000000 } },
{ 0x000003EC, { 0x000003ED,  0x00000000, 0, 0x00000000 } },
{ 0x000003EE, { 0x000003EF,  0x00000000, 0, 0x00000000 } },
{ 0x000003F0, { 0x000003BA,  0x00000000, 0, 0x00000000 } },
{ 0x000003F1, { 0x000003C1,  0x00000000, 0, 0x00000000 } },
{ 0x000003F4, { 0x000003B8,  0x00000000, 0, 0x00000000 } },
{ 0x000003F5, { 0x000003B5,  0x00000000, 0, 0x00000000 } },
{ 0x000003F7, { 0x000003F8,  0x00000000, 0, 0x00000000 } },
{ 0x000003F9, { 0x000003F2,  0x00000000, 0, 0x00000000 } },
{ 0x000003FA, { 0x000003FB,  0x00000000, 0, 0x00000000 } },
{ 0x000003FD, { 0x0000037B,  0x00000000, 0, 0x00000000 } },
{ 0x000003FE, { 0x0000037C,  0x00000000, 0, 0x00000000 } },
{ 0x000003FF, { 0x0000037D,  0x00000000, 0, 0x00000000 } },
{ 0x00010400, { 0x00010428,  0x00000000, 0, 0x00000000 } },
{ 0x00000401, { 0x00000451,  0x00000000, 0, 0x00000000 } },
{ 0x00010402, { 0x0001042A,  0x00000000, 0, 0x00000000 } },
{ 0x00010403, { 0x0001042B,  0x00000000, 0, 0x00000000 } },
{ 0x00010404, { 0x0001042C,  0x00000000, 0, 0x00000000 } },
{ 0x00010405, { 0x0001042D,  0x00000000, 0, 0x00000000 } },
{ 0x00010406, { 0x0001042E,  0x00000000, 0, 0x00000000 } },
{ 0x00010407, { 0x0001042F,  0x00000000, 0, 0x00000000 } },
{ 0x00000408, { 0x00000458,  0x00000000, 0, 0x00000000 } },
{ 0x00010409, { 0x00010431,  0x00000000, 0, 0x00000000 } },
{ 0x0001040A, { 0x00010432,  0x00000000, 0, 0x00000000 } },
{ 0x0001040B, { 0x00010433,  0x00000000, 0, 0x00000000 } },
{ 0x0001040C, { 0x00010434,  0x00000000, 0, 0x00000000 } },
{ 0x0000040D, { 0x0000045D,  0x00000000, 0, 0x00000000 } },
{ 0x0000040E, { 0x0000045E,  0x00000000, 0, 0x00000000 } },
{ 0x0001040F, { 0x00010437,  0x00000000, 0, 0x00000000 } },
{ 0x00010410, { 0x00010438,  0x00000000, 0, 0x00000000 } },
{ 0x00000411, { 0x00000431,  0x00000000, 0, 0x00000000 } },
{ 0x00010412, { 0x0001043A,  0x00000000, 0, 0x00000000 } },
{ 0x00010413, { 0x0001043B,  0x00000000, 0, 0x00000000 } },
{ 0x00010414, { 0x0001043C,  0x00000000, 0, 0x00000000 } },
{ 0x00010415, { 0x0001043D,  0x00000000, 0, 0x00000000 } },
{ 0x00000416, { 0x00000436,  0x00000000, 0, 0x00000000 } },
{ 0x00010417, { 0x0001043F,  0x00000000, 0, 0x00000000 } },
{ 0x00000418, { 0x00000438,  0x00000000, 0, 0x00000000 } },
{ 0x00000419, { 0x00000439,  0x00000000, 0, 0x00000000 } },
{ 0x0001041A, { 0x00010442,  0x00000000, 0, 0x00000000 } },
{ 0x0000041B, { 0x0000043B,  0x00000000, 0, 0x00000000 } },
{ 0x0001041C, { 0x00010444,  0x00000000, 0, 0x00000000 } },
{ 0x0001041D, { 0x00010445,  0x00000000, 0, 0x00000000 } },
{ 0x0001041E, { 0x00010446,  0x00000000, 0, 0x00000000 } },
{ 0x0001041F, { 0x00010447,  0x00000000, 0, 0x00000000 } },
{ 0x00010420, { 0x00010448,  0x00000000, 0, 0x00000000 } },
{ 0x00000421, { 0x00000441,  0x00000000, 0, 0x00000000 } },
{ 0x00000422, { 0x00000442,  0x00000000, 0, 0x00000000 } },
{ 0x00000423, { 0x00000443,  0x00000000, 0, 0x00000000 } },
{ 0x00000424, { 0x00000444,  0x00000000, 0, 0x00000000 } },
{ 0x00000425, { 0x00000445,  0x00000000, 0, 0x00000000 } },
{ 0x00000426, { 0x00000446,  0x00000000, 0, 0x00000000 } },
{ 0x00000427, { 0x00000447,  0x00000000, 0, 0x00000000 } },
{ 0x00000428, { 0x00000448,  0x00000000, 0, 0x00000000 } },
{ 0x00000429, { 0x00000449,  0x00000000, 0, 0x00000000 } },
{ 0x0000042A, { 0x0000044A,  0x00000000, 0, 0x00000000 } },
{ 0x0000042B, { 0x0000044B,  0x00000000, 0, 0x00000000 } },
{ 0x0000042C, { 0x0000044C,  0x00000000, 0, 0x00000000 } },
{ 0x0000042D, { 0x0000044D,  0x00000000, 0, 0x00000000 } },
{ 0x0000042E, { 0x0000044E,  0x00000000, 0, 0x00000000 } },
{ 0x0000042F, { 0x0000044F,  0x00000000, 0, 0x00000000 } },
{ 0x00000460, { 0x00000461,  0x00000000, 0, 0x00000000 } },
{ 0x00000462, { 0x00000463,  0x00000000, 0, 0x00000000 } },
{ 0x00000464, { 0x00000465,  0x00000000, 0, 0x00000000 } },
{ 0x00000466, { 0x00000467,  0x00000000, 0, 0x00000000 } },
{ 0x00000468, { 0x00000469,  0x00000000, 0, 0x00000000 } },
{ 0x0000046A, { 0x0000046B,  0x00000000, 0, 0x00000000 } },
{ 0x0000046C, { 0x0000046D,  0x00000000, 0, 0x00000000 } },
{ 0x0000046E, { 0x0000046F,  0x00000000, 0, 0x00000000 } },
{ 0x00000470, { 0x00000471,  0x00000000, 0, 0x00000000 } },
{ 0x00000472, { 0x00000473,  0x00000000, 0, 0x00000000 } },
{ 0x00000474, { 0x00000475,  0x00000000, 0, 0x00000000 } },
{ 0x00000476, { 0x00000477,  0x00000000, 0, 0x00000000 } },
{ 0x00000478, { 0x00000479,  0x00000000, 0, 0x00000000 } },
{ 0x0000047A, { 0x0000047B,  0x00000000, 0, 0x00000000 } },
{ 0x0000047C, { 0x0000047D,  0x00000000, 0, 0x00000000 } },
{ 0x0000047E, { 0x0000047F,  0x00000000, 0, 0x00000000 } },
{ 0x00000480, { 0x00000481,  0x00000000, 0, 0x00000000 } },
{ 0x0000048A, { 0x0000048B,  0x00000000, 0, 0x00000000 } },
{ 0x0000048C, { 0x0000048D,  0x00000000, 0, 0x00000000 } },
{ 0x0000048E, { 0x0000048F,  0x00000000, 0, 0x00000000 } },
{ 0x00000490, { 0x00000491,  0x00000000, 0, 0x00000000 } },
{ 0x00000492, { 0x00000493,  0x00000000, 0, 0x00000000 } },
{ 0x00010CAC, { 0x00010CEC,  0x00000000, 0, 0x00000000 } },
{ 0x00000494, { 0x00000495,  0x00000000, 0, 0x00000000 } },
{ 0x00000496, { 0x00000497,  0x00000000, 0, 0x00000000 } },
{ 0x00000498, { 0x00000499,  0x00000000, 0, 0x00000000 } },
{ 0x0000049A, { 0x0000049B,  0x00000000, 0, 0x00000000 } },
{ 0x0000049C, { 0x0000049D,  0x00000000, 0, 0x00000000 } },
{ 0x0000049E, { 0x0000049F,  0x00000000, 0, 0x00000000 } },
{ 0x000004A0, { 0x000004A1,  0x00000000, 0, 0x00000000 } },
{ 0x000004A2, { 0x000004A3,  0x00000000, 0, 0x00000000 } },
{ 0x000004A4, { 0x000004A5,  0x00000000, 0, 0x00000000 } },
{ 0x000004A6, { 0x000004A7,  0x00000000, 0, 0x00000000 } },
{ 0x000004A8, { 0x000004A9,  0x00000000, 0, 0x00000000 } },
{ 0x000004AA, { 0x000004AB,  0x00000000, 0, 0x00000000 } },
{ 0x000004AC, { 0x000004AD,  0x00000000, 0, 0x00000000 } },
{ 0x000004AE, { 0x000004AF,  0x00000000, 0, 0x00000000 } },
{ 0x000104B0, { 0x000104D8,  0x00000000, 0, 0x00000000 } },
{ 0x000104B1, { 0x000104D9,  0x00000000, 0, 0x00000000 } },
{ 0x000004B2, { 0x000004B3,  0x00000000, 0, 0x00000000 } },
{ 0x000104B3, { 0x000104DB,  0x00000000, 0, 0x00000000 } },
{ 0x000104B4, { 0x000104DC,  0x00000000, 0, 0x00000000 } },
{ 0x000104B5, { 0x000104DD,  0x00000000, 0, 0x00000000 } },
{ 0x000104B6, { 0x000104DE,  0x00000000, 0, 0x00000000 } },
{ 0x000024B7, { 0x000024D1,  0x00000000, 0, 0x00000000 } },
{ 0x000104B8, { 0x000104E0,  0x00000000, 0, 0x00000000 } },
{ 0x000104B9, { 0x000104E1,  0x00000000, 0, 0x00000000 } },
{ 0x000104BA, { 0x000104E2,  0x00000000, 0, 0x00000000 } },
{ 0x000104BB, { 0x000104E3,  0x00000000, 0, 0x00000000 } },
{ 0x000024BC, { 0x000024D6,  0x00000000, 0, 0x00000000 } },
{ 0x000104BD, { 0x000104E5,  0x00000000, 0, 0x00000000 } },
{ 0x000024BE, { 0x000024D8,  0x00000000, 0, 0x00000000 } },
{ 0x000104BF, { 0x000104E7,  0x00000000, 0, 0x00000000 } },
{ 0x000104C0, { 0x000104E8,  0x00000000, 0, 0x00000000 } },
{ 0x000104C1, { 0x000104E9,  0x00000000, 0, 0x00000000 } },
{ 0x000104C2, { 0x000104EA,  0x00000000, 0, 0x00000000 } },
{ 0x000024C3, { 0x000024DD,  0x00000000, 0, 0x00000000 } },
{ 0x000104C4, { 0x000104EC,  0x00000000, 0, 0x00000000 } },
{ 0x000104C5, { 0x000104ED,  0x00000000, 0, 0x00000000 } },
{ 0x000104C6, { 0x000104EE,  0x00000000, 0, 0x00000000 } },
{ 0x000024C7, { 0x000024E1,  0x00000000, 0, 0x00000000 } },
{ 0x000024C8, { 0x000024E2,  0x00000000, 0, 0x00000000 } },
{ 0x000024C9, { 0x000024E3,  0x00000000, 0, 0x00000000 } },
{ 0x000024CA, { 0x000024E4,  0x00000000, 0, 0x00000000 } },
{ 0x000104CB, { 0x000104F3,  0x00000000, 0, 0x00000000 } },
{ 0x000104CC, { 0x000104F4,  0x00000000, 0, 0x00000000 } },
{ 0x000104CD, { 0x000104F5,  0x00000000, 0, 0x00000000 } },
{ 0x000104CE, { 0x000104F6,  0x00000000, 0, 0x00000000 } },
{ 0x000104CF, { 0x000104F7,  0x00000000, 0, 0x00000000 } },
{ 0x000104D0, { 0x000104F8,  0x00000000, 0, 0x00000000 } },
{ 0x000104D1, { 0x000104F9,  0x00000000, 0, 0x00000000 } },
{ 0x000104D2, { 0x000104FA,  0x00000000, 0, 0x00000000 } },
{ 0x000104D3, { 0x000104FB,  0x00000000, 0, 0x00000000 } },
{ 0x000004D4, { 0x000004D5,  0x00000000, 0, 0x00000000 } },
{ 0x000004D6, { 0x000004D7,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB79, { 0x000013A9,  0x00000000, 0, 0x00000000 } },
{ 0x000004D8, { 0x000004D9,  0x00000000, 0, 0x00000000 } },
{ 0x000004DA, { 0x000004DB,  0x00000000, 0, 0x00000000 } },
{ 0x000004DC, { 0x000004DD,  0x00000000, 0, 0x00000000 } },
{ 0x000004DE, { 0x000004DF,  0x00000000, 0, 0x00000000 } },
{ 0x000004E0, { 0x000004E1,  0x00000000, 0, 0x00000000 } },
{ 0x000004E2, { 0x000004E3,  0x00000000, 0, 0x00000000 } },
{ 0x000004E4, { 0x000004E5,  0x00000000, 0, 0x00000000 } },
{ 0x000004E6, { 0x000004E7,  0x00000000, 0, 0x00000000 } },
{ 0x000004E8, { 0x000004E9,  0x00000000, 0, 0x00000000 } },
{ 0x000004EA, { 0x000004EB,  0x00000000, 0, 0x00000000 } },
{ 0x000004EC, { 0x000004ED,  0x00000000, 0, 0x00000000 } },
{ 0x000004EE, { 0x000004EF,  0x00000000, 0, 0x00000000 } },
{ 0x000004F0, { 0x000004F1,  0x00000000, 0, 0x00000000 } },
{ 0x000004F2, { 0x000004F3,  0x00000000, 0, 0x00000000 } },
{ 0x000004F4, { 0x000004F5,  0x00000000, 0, 0x00000000 } },
{ 0x000004F6, { 0x000004F7,  0x00000000, 0, 0x00000000 } },
{ 0x000004F8, { 0x000004F9,  0x00000000, 0, 0x00000000 } },
{ 0x000004FA, { 0x000004FB,  0x00000000, 0, 0x00000000 } },
{ 0x000004FC, { 0x000004FD,  0x00000000, 0, 0x00000000 } },
{ 0x000004FE, { 0x000004FF,  0x00000000, 0, 0x00000000 } },
{ 0x00000500, { 0x00000501,  0x00000000, 0, 0x00000000 } },
{ 0x00000502, { 0x00000503,  0x00000000, 0, 0x00000000 } },
{ 0x00000504, { 0x00000505,  0x00000000, 0, 0x00000000 } },
{ 0x00000506, { 0x00000507,  0x00000000, 0, 0x00000000 } },
{ 0x00000508, { 0x00000509,  0x00000000, 0, 0x00000000 } },
{ 0x0000050A, { 0x0000050B,  0x00000000, 0, 0x00000000 } },
{ 0x00010CB0, { 0x00010CF0,  0x00000000, 0, 0x00000000 } },
{ 0x0000050C, { 0x0000050D,  0x00000000, 0, 0x00000000 } },
{ 0x0000050E, { 0x0000050F,  0x00000000, 0, 0x00000000 } },
{ 0x00000510, { 0x00000511,  0x00000000, 0, 0x00000000 } },
{ 0x00000512, { 0x00000513,  0x00000000, 0, 0x00000000 } },
{ 0x00000514, { 0x00000515,  0x00000000, 0, 0x00000000 } },
{ 0x00000516, { 0x00000517,  0x00000000, 0, 0x00000000 } },
{ 0x00000518, { 0x00000519,  0x00000000, 0, 0x00000000 } },
{ 0x0000051A, { 0x0000051B,  0x00000000, 0, 0x00000000 } },
{ 0x0000051C, { 0x0000051D,  0x00000000, 0, 0x00000000 } },
{ 0x0000051E, { 0x0000051F,  0x00000000, 0, 0x00000000 } },
{ 0x00000520, { 0x00000521,  0x00000000, 0, 0x00000000 } },
{ 0x00000522, { 0x00000523,  0x00000000, 0, 0x00000000 } },
{ 0x00000524, { 0x00000525,  0x00000000, 0, 0x00000000 } },
{ 0x00000526, { 0x00000527,  0x00000000, 0, 0x00000000 } },
{ 0x00000528, { 0x00000529,  0x00000000, 0, 0x00000000 } },
{ 0x0000052A, { 0x0000052B,  0x00000000, 0, 0x00000000 } },
{ 0x0000052C, { 0x0000052D,  0x00000000, 0, 0x00000000 } },
{ 0x0000052E, { 0x0000052F,  0x00000000, 0, 0x00000000 } },
{ 0x00000531, { 0x00000561,  0x00000000, 0, 0x00000000 } },
{ 0x00000532, { 0x00000562,  0x00000000, 0, 0x00000000 } },
{ 0x00000533, { 0x00000563,  0x00000000, 0, 0x00000000 } },
{ 0x00000534, { 0x00000564,  0x00000000, 0, 0x00000000 } },
{ 0x00000535, { 0x00000565,  0x00000000, 0, 0x00000000 } },
{ 0x00000536, { 0x00000566,  0x00000000, 0, 0x00000000 } },
{ 0x00000537, { 0x00000567,  0x00000000, 0, 0x00000000 } },
{ 0x00000538, { 0x00000568,  0x00000000, 0, 0x00000000 } },
{ 0x00000539, { 0x00000569,  0x00000000, 0, 0x00000000 } },
{ 0x0000053A, { 0x0000056A,  0x00000000, 0, 0x00000000 } },
{ 0x0000053B, { 0x0000056B,  0x00000000, 0, 0x00000000 } },
{ 0x0000053C, { 0x0000056C,  0x00000000, 0, 0x00000000 } },
{ 0x0000053D, { 0x0000056D,  0x00000000, 0, 0x00000000 } },
{ 0x0000053E, { 0x0000056E,  0x00000000, 0, 0x00000000 } },
{ 0x0000053F, { 0x0000056F,  0x00000000, 0, 0x00000000 } },
{ 0x00000540, { 0x00000570,  0x00000000, 0, 0x00000000 } },
{ 0x00000541, { 0x00000571,  0x00000000, 0, 0x00000000 } },
{ 0x00000542, { 0x00000572,  0x00000000, 0, 0x00000000 } },
{ 0x00000543, { 0x00000573,  0x00000000, 0, 0x00000000 } },
{ 0x00000544, { 0x00000574,  0x00000000, 0, 0x00000000 } },
{ 0x00000545, { 0x00000575,  0x00000000, 0, 0x00000000 } },
{ 0x00000546, { 0x00000576,  0x00000000, 0, 0x00000000 } },
{ 0x00000547, { 0x00000577,  0x00000000, 0, 0x00000000 } },
{ 0x00000548, { 0x00000578,  0x00000000, 0, 0x00000000 } },
{ 0x00000549, { 0x00000579,  0x00000000, 0, 0x00000000 } },
{ 0x0000054A, { 0x0000057A,  0x00000000, 0, 0x00000000 } },
{ 0x0000054B, { 0x0000057B,  0x00000000, 0, 0x00000000 } },
{ 0x0000054C, { 0x0000057C,  0x00000000, 0, 0x00000000 } },
{ 0x0000054D, { 0x0000057D,  0x00000000, 0, 0x00000000 } },
{ 0x0000054E, { 0x0000057E,  0x00000000, 0, 0x00000000 } },
{ 0x0000054F, { 0x0000057F,  0x00000000, 0, 0x00000000 } },
{ 0x00000550, { 0x00000580,  0x00000000, 0, 0x00000000 } },
{ 0x00000551, { 0x00000581,  0x00000000, 0, 0x00000000 } },
{ 0x00000552, { 0x00000582,  0x00000000, 0, 0x00000000 } },
{ 0x00000553, { 0x00000583,  0x00000000, 0, 0x00000000 } },
{ 0x00000554, { 0x00000584,  0x00000000, 0, 0x00000000 } },
{ 0x00000555, { 0x00000585,  0x00000000, 0, 0x00000000 } },
{ 0x00000556, { 0x00000586,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB91, { 0x000013C1,  0x00000000, 0, 0x00000000 } },
{ 0x00000587, { 0x00000000,  0x00000000, U"\U00000565\U00000582", 0x00000000 } },
{ 0x0000A640, { 0x0000A641,  0x00000000, 0, 0x00000000 } },
{ 0x0000A642, { 0x0000A643,  0x00000000, 0, 0x00000000 } },
{ 0x0000A644, { 0x0000A645,  0x00000000, 0, 0x00000000 } },
{ 0x0000A646, { 0x0000A647,  0x00000000, 0, 0x00000000 } },
{ 0x0000A648, { 0x0000A649,  0x00000000, 0, 0x00000000 } },
{ 0x0000A64A, { 0x0000A64B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A64C, { 0x0000A64D,  0x00000000, 0, 0x00000000 } },
{ 0x0000A64E, { 0x0000A64F,  0x00000000, 0, 0x00000000 } },
{ 0x0000A650, { 0x0000A651,  0x00000000, 0, 0x00000000 } },
{ 0x0000A652, { 0x0000A653,  0x00000000, 0, 0x00000000 } },
{ 0x0000A654, { 0x0000A655,  0x00000000, 0, 0x00000000 } },
{ 0x0000A656, { 0x0000A657,  0x00000000, 0, 0x00000000 } },
{ 0x0000A658, { 0x0000A659,  0x00000000, 0, 0x00000000 } },
{ 0x0000A65A, { 0x0000A65B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A65C, { 0x0000A65D,  0x00000000, 0, 0x00000000 } },
{ 0x0000A65E, { 0x0000A65F,  0x00000000, 0, 0x00000000 } },
{ 0x0000A660, { 0x0000A661,  0x00000000, 0, 0x00000000 } },
{ 0x0000A662, { 0x0000A663,  0x00000000, 0, 0x00000000 } },
{ 0x0000A664, { 0x0000A665,  0x00000000, 0, 0x00000000 } },
{ 0x0000A666, { 0x0000A667,  0x00000000, 0, 0x00000000 } },
{ 0x0000A668, { 0x0000A669,  0x00000000, 0, 0x00000000 } },
{ 0x0000A66A, { 0x0000A66B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A66C, { 0x0000A66D,  0x00000000, 0, 0x00000000 } },
{ 0x0000A680, { 0x0000A681,  0x00000000, 0, 0x00000000 } },
{ 0x0000A682, { 0x0000A683,  0x00000000, 0, 0x00000000 } },
{ 0x0000A684, { 0x0000A685,  0x00000000, 0, 0x00000000 } },
{ 0x0000A686, { 0x0000A687,  0x00000000, 0, 0x00000000 } },
{ 0x0000A688, { 0x0000A689,  0x00000000, 0, 0x00000000 } },
{ 0x0000A68A, { 0x0000A68B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A68C, { 0x0000A68D,  0x00000000, 0, 0x00000000 } },
{ 0x0000A68E, { 0x0000A68F,  0x00000000, 0, 0x00000000 } },
{ 0x0000A690, { 0x0000A691,  0x00000000, 0, 0x00000000 } },
{ 0x0000A692, { 0x0000A693,  0x00000000, 0, 0x00000000 } },
{ 0x0000A694, { 0x0000A695,  0x00000000, 0, 0x00000000 } },
{ 0x0000A696, { 0x0000A697,  0x00000000, 0, 0x00000000 } },
{ 0x0000A698, { 0x0000A699,  0x00000000, 0, 0x00000000 } },
{ 0x0000A69A, { 0x0000A69B,  0x00000000, 0, 0x00000000 } },
{ 0x00002126, { 0x000003C9,  0x00000000, 0, 0x00000000 } },
{ 0x0000212A, { 0x0000006B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A722, { 0x0000A723,  0x00000000, 0, 0x00000000 } },
{ 0x0000A724, { 0x0000A725,  0x00000000, 0, 0x00000000 } },
{ 0x0000A726, { 0x0000A727,  0x00000000, 0, 0x00000000 } },
{ 0x0000A728, { 0x0000A729,  0x00000000, 0, 0x00000000 } },
{ 0x0000A72A, { 0x0000A72B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A72C, { 0x0000A72D,  0x00000000, 0, 0x00000000 } },
{ 0x00002132, { 0x0000214E,  0x00000000, 0, 0x00000000 } },
{ 0x0000A72E, { 0x0000A72F,  0x00000000, 0, 0x00000000 } },
{ 0x0000A732, { 0x0000A733,  0x00000000, 0, 0x00000000 } },
{ 0x0000A734, { 0x0000A735,  0x00000000, 0, 0x00000000 } },
{ 0x0000A736, { 0x0000A737,  0x00000000, 0, 0x00000000 } },
{ 0x0000A738, { 0x0000A739,  0x00000000, 0, 0x00000000 } },
{ 0x0000A73A, { 0x0000A73B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A73C, { 0x0000A73D,  0x00000000, 0, 0x00000000 } },
{ 0x0000A73E, { 0x0000A73F,  0x00000000, 0, 0x00000000 } },
{ 0x0000A740, { 0x0000A741,  0x00000000, 0, 0x00000000 } },
{ 0x0000A742, { 0x0000A743,  0x00000000, 0, 0x00000000 } },
{ 0x0000A744, { 0x0000A745,  0x00000000, 0, 0x00000000 } },
{ 0x0000A746, { 0x0000A747,  0x00000000, 0, 0x00000000 } },
{ 0x0000A748, { 0x0000A749,  0x00000000, 0, 0x00000000 } },
{ 0x0000A74A, { 0x0000A74B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A74C, { 0x0000A74D,  0x00000000, 0, 0x00000000 } },
{ 0x0000A74E, { 0x0000A74F,  0x00000000, 0, 0x00000000 } },
{ 0x0000A750, { 0x0000A751,  0x00000000, 0, 0x00000000 } },
{ 0x0000A752, { 0x0000A753,  0x00000000, 0, 0x00000000 } },
{ 0x0000A754, { 0x0000A755,  0x00000000, 0, 0x00000000 } },
{ 0x0000A756, { 0x0000A757,  0x00000000, 0, 0x00000000 } },
{ 0x0000A758, { 0x0000A759,  0x00000000, 0, 0x00000000 } },
{ 0x0000A75A, { 0x0000A75B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A75C, { 0x0000A75D,  0x00000000, 0, 0x00000000 } },
{ 0x0000A75E, { 0x0000A75F,  0x00000000, 0, 0x00000000 } },
{ 0x0000A760, { 0x0000A761,  0x00000000, 0, 0x00000000 } },
{ 0x0000A762, { 0x0000A763,  0x00000000, 0, 0x00000000 } },
{ 0x0000A764, { 0x0000A765,  0x00000000, 0, 0x00000000 } },
{ 0x0000A766, { 0x0000A767,  0x00000000, 0, 0x00000000 } },
{ 0x0000A768, { 0x0000A769,  0x00000000, 0, 0x00000000 } },
{ 0x0000A76A, { 0x0000A76B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A76C, { 0x0000A76D,  0x00000000, 0, 0x00000000 } },
{ 0x0000A76E, { 0x0000A76F,  0x00000000, 0, 0x00000000 } },
{ 0x0000A779, { 0x0000A77A,  0x00000000, 0, 0x00000000 } },
{ 0x0000A77B, { 0x0000A77C,  0x00000000, 0, 0x00000000 } },
{ 0x0000A77D, { 0x00001D79,  0x00000000, 0, 0x00000000 } },
{ 0x0000A77E, { 0x0000A77F,  0x00000000, 0, 0x00000000 } },
{ 0x0000A780, { 0x0000A781,  0x00000000, 0, 0x00000000 } },
{ 0x0000A782, { 0x0000A783,  0x00000000, 0, 0x00000000 } },
{ 0x0000A784, { 0x0000A785,  0x00000000, 0, 0x00000000 } },
{ 0x0000A786, { 0x0000A787,  0x00000000, 0, 0x00000000 } },
{ 0x0000A78B, { 0x0000A78C,  0x00000000, 0, 0x00000000 } },
{ 0x0000A78D, { 0x00000265,  0x00000000, 0, 0x00000000 } },
{ 0x0000A790, { 0x0000A791,  0x00000000, 0, 0x00000000 } },
{ 0x0000A792, { 0x0000A793,  0x00000000, 0, 0x00000000 } },
{ 0x0000A796, { 0x0000A797,  0x00000000, 0, 0x00000000 } },
{ 0x0000A798, { 0x0000A799,  0x00000000, 0, 0x00000000 } },
{ 0x0000A79A, { 0x0000A79B,  0x00000000, 0, 0x00000000 } },
{ 0x0000A79C, { 0x0000A79D,  0x00000000, 0, 0x00000000 } },
{ 0x0000A79E, { 0x0000A79F,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7A0, { 0x0000A7A1,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7A2, { 0x0000A7A3,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7A4, { 0x0000A7A5,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7A6, { 0x0000A7A7,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7A8, { 0x0000A7A9,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7AA, { 0x00000266,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7AB, { 0x0000025C,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7AC, { 0x00000261,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7AD, { 0x0000026C,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7AE, { 0x0000026A,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7B0, { 0x0000029E,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7B1, { 0x00000287,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7B2, { 0x0000029D,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7B3, { 0x0000AB53,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7B4, { 0x0000A7B5,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7B6, { 0x0000A7B7,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7B8, { 0x0000A7B9,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7BA, { 0x0000A7BB,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7BC, { 0x0000A7BD,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7BE, { 0x0000A7BF,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7C2, { 0x0000A7C3,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7C4, { 0x0000A794,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7C5, { 0x00000282,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7C6, { 0x00001D8E,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7C7, { 0x0000A7C8,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7C9, { 0x0000A7CA,  0x00000000, 0, 0x00000000 } },
{ 0x0000A7F5, { 0x0000A7F6,  0x00000000, 0, 0x00000000 } },
{ 0x00001F2B, { 0x00001F23,  0x00000000, 0, 0x00000000 } },
{ 0x00002160, { 0x00002170,  0x00000000, 0, 0x00000000 } },
{ 0x00002162, { 0x00002172,  0x00000000, 0, 0x00000000 } },
{ 0x00002164, { 0x00002174,  0x00000000, 0, 0x00000000 } },
{ 0x00000166, { 0x00000167,  0x00000000, 0, 0x00000000 } },
{ 0x00000168, { 0x00000169,  0x00000000, 0, 0x00000000 } },
{ 0x0000016A, { 0x0000016B,  0x00000000, 0, 0x00000000 } },
{ 0x0000016C, { 0x0000016D,  0x00000000, 0, 0x00000000 } },
{ 0x0000016E, { 0x0000016F,  0x00000000, 0, 0x00000000 } },
{ 0x0001E900, { 0x0001E922,  0x00000000, 0, 0x00000000 } },
{ 0x0001E901, { 0x0001E923,  0x00000000, 0, 0x00000000 } },
{ 0x0001E902, { 0x0001E924,  0x00000000, 0, 0x00000000 } },
{ 0x0001E903, { 0x0001E925,  0x00000000, 0, 0x00000000 } },
{ 0x0001E904, { 0x0001E926,  0x00000000, 0, 0x00000000 } },
{ 0x0001E905, { 0x0001E927,  0x00000000, 0, 0x00000000 } },
{ 0x0001E906, { 0x0001E928,  0x00000000, 0, 0x00000000 } },
{ 0x0001E907, { 0x0001E929,  0x00000000, 0, 0x00000000 } },
{ 0x0001E908, { 0x0001E92A,  0x00000000, 0, 0x00000000 } },
{ 0x0001E909, { 0x0001E92B,  0x00000000, 0, 0x00000000 } },
{ 0x0001E90A, { 0x0001E92C,  0x00000000, 0, 0x00000000 } },
{ 0x0001E90B, { 0x0001E92D,  0x00000000, 0, 0x00000000 } },
{ 0x0001E90C, { 0x0001E92E,  0x00000000, 0, 0x00000000 } },
{ 0x0001E90D, { 0x0001E92F,  0x00000000, 0, 0x00000000 } },
{ 0x0001E90E, { 0x0001E930,  0x00000000, 0, 0x00000000 } },
{ 0x0001E90F, { 0x0001E931,  0x00000000, 0, 0x00000000 } },
{ 0x0001E910, { 0x0001E932,  0x00000000, 0, 0x00000000 } },
{ 0x0001E911, { 0x0001E933,  0x00000000, 0, 0x00000000 } },
{ 0x0001E912, { 0x0001E934,  0x00000000, 0, 0x00000000 } },
{ 0x0001E913, { 0x0001E935,  0x00000000, 0, 0x00000000 } },
{ 0x0001E914, { 0x0001E936,  0x00000000, 0, 0x00000000 } },
{ 0x0001E915, { 0x0001E937,  0x00000000, 0, 0x00000000 } },
{ 0x0001E916, { 0x0001E938,  0x00000000, 0, 0x00000000 } },
{ 0x0001E917, { 0x0001E939,  0x00000000, 0, 0x00000000 } },
{ 0x0001E918, { 0x0001E93A,  0x00000000, 0, 0x00000000 } },
{ 0x0001E919, { 0x0001E93B,  0x00000000, 0, 0x00000000 } },
{ 0x0001E91A, { 0x0001E93C,  0x00000000, 0, 0x00000000 } },
{ 0x0001E91B, { 0x0001E93D,  0x00000000, 0, 0x00000000 } },
{ 0x0001E91C, { 0x0001E93E,  0x00000000, 0, 0x00000000 } },
{ 0x0001E91D, { 0x0001E93F,  0x00000000, 0, 0x00000000 } },
{ 0x0001E91E, { 0x0001E940,  0x00000000, 0, 0x00000000 } },
{ 0x0001E91F, { 0x0001E941,  0x00000000, 0, 0x00000000 } },
{ 0x0001E920, { 0x0001E942,  0x00000000, 0, 0x00000000 } },
{ 0x0001E921, { 0x0001E943,  0x00000000, 0, 0x00000000 } },
{ 0x00010C96, { 0x00010CD6,  0x00000000, 0, 0x00000000 } },
{ 0x00010C80, { 0x00010CC0,  0x00000000, 0, 0x00000000 } },
{ 0x00002C82, { 0x00002C83,  0x00000000, 0, 0x00000000 } },
{ 0x00002C84, { 0x00002C85,  0x00000000, 0, 0x00000000 } },
{ 0x00002C86, { 0x00002C87,  0x00000000, 0, 0x00000000 } },
{ 0x00001C87, { 0x00000463,  0x00000000, 0, 0x00000000 } },
{ 0x00002C88, { 0x00002C89,  0x00000000, 0, 0x00000000 } },
{ 0x00002C8A, { 0x00002C8B,  0x00000000, 0, 0x00000000 } },
{ 0x00010C8C, { 0x00010CCC,  0x00000000, 0, 0x00000000 } },
{ 0x00010C8E, { 0x00010CCE,  0x00000000, 0, 0x00000000 } },
{ 0x00002C90, { 0x00002C91,  0x00000000, 0, 0x00000000 } },
{ 0x00002C92, { 0x00002C93,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB70, { 0x000013A0,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB71, { 0x000013A1,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB72, { 0x000013A2,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB73, { 0x000013A3,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB74, { 0x000013A4,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB75, { 0x000013A5,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB76, { 0x000013A6,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB77, { 0x000013A7,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB78, { 0x000013A8,  0x00000000, 0, 0x00000000 } },
{ 0x00002C94, { 0x00002C95,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB7A, { 0x000013AA,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB7B, { 0x000013AB,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB7C, { 0x000013AC,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB7D, { 0x000013AD,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB7E, { 0x000013AE,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB7F, { 0x000013AF,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB80, { 0x000013B0,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB81, { 0x000013B1,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB82, { 0x000013B2,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB83, { 0x000013B3,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB84, { 0x000013B4,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB85, { 0x000013B5,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB86, { 0x000013B6,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB87, { 0x000013B7,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB88, { 0x000013B8,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB89, { 0x000013B9,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB8A, { 0x000013BA,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB8B, { 0x000013BB,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB8C, { 0x000013BC,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB8D, { 0x000013BD,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB8E, { 0x000013BE,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB8F, { 0x000013BF,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB90, { 0x000013C0,  0x00000000, 0, 0x00000000 } },
{ 0x00010C98, { 0x00010CD8,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB92, { 0x000013C2,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB93, { 0x000013C3,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB94, { 0x000013C4,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB95, { 0x000013C5,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB96, { 0x000013C6,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB97, { 0x000013C7,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB98, { 0x000013C8,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB99, { 0x000013C9,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB9A, { 0x000013CA,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB9B, { 0x000013CB,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB9C, { 0x000013CC,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB9D, { 0x000013CD,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB9E, { 0x000013CE,  0x00000000, 0, 0x00000000 } },
{ 0x0000AB9F, { 0x000013CF,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABA0, { 0x000013D0,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABA1, { 0x000013D1,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABA2, { 0x000013D2,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABA3, { 0x000013D3,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABA4, { 0x000013D4,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABA5, { 0x000013D5,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABA6, { 0x000013D6,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABA7, { 0x000013D7,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABA8, { 0x000013D8,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABA9, { 0x000013D9,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABAA, { 0x000013DA,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABAB, { 0x000013DB,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABAC, { 0x000013DC,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABAD, { 0x000013DD,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABAE, { 0x000013DE,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABAF, { 0x000013DF,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABB0, { 0x000013E0,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABB1, { 0x000013E1,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABB2, { 0x000013E2,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABB3, { 0x000013E3,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABB4, { 0x000013E4,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABB5, { 0x000013E5,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABB6, { 0x000013E6,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABB7, { 0x000013E7,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABB8, { 0x000013E8,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABB9, { 0x000013E9,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABBA, { 0x000013EA,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABBB, { 0x000013EB,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABBC, { 0x000013EC,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABBD, { 0x000013ED,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABBE, { 0x000013EE,  0x00000000, 0, 0x00000000 } },
{ 0x0000ABBF, { 0x000013EF,  0x00000000, 0, 0x00000000 } },
{ 0x00010CA0, { 0x00010CE0,  0x00000000, 0, 0x00000000 } },
{ 0x00010CA2, { 0x00010CE2,  0x00000000, 0, 0x00000000 } },
{ 0x00010CA4, { 0x00010CE4,  0x00000000, 0, 0x00000000 } },
{ 0x00010CA6, { 0x00010CE6,  0x00000000, 0, 0x00000000 } },
{ 0x00010CA8, { 0x00010CE8,  0x00000000, 0, 0x00000000 } },
{ 0x00010CAA, { 0x00010CEA,  0x00000000, 0, 0x00000000 } },
{ 0x00002C00, { 0x00002C30,  0x00000000, 0, 0x00000000 } },
{ 0x00002C01, { 0x00002C31,  0x00000000, 0, 0x00000000 } },
{ 0x00002C02, { 0x00002C32,  0x00000000, 0, 0x00000000 } },
{ 0x00002C03, { 0x00002C33,  0x00000000, 0, 0x00000000 } },
{ 0x00002C04, { 0x00002C34,  0x00000000, 0, 0x00000000 } },
{ 0x00002C05, { 0x00002C35,  0x00000000, 0, 0x00000000 } },
{ 0x00002C06, { 0x00002C36,  0x00000000, 0, 0x00000000 } },
{ 0x00002C07, { 0x00002C37,  0x00000000, 0, 0x00000000 } },
{ 0x00002C08, { 0x00002C38,  0x00000000, 0, 0x00000000 } },
{ 0x00002C09, { 0x00002C39,  0x00000000, 0, 0x00000000 } },
{ 0x00002C0A, { 0x00002C3A,  0x00000000, 0, 0x00000000 } },
{ 0x00002C0B, { 0x00002C3B,  0x00000000, 0, 0x00000000 } },
{ 0x00002C0C, { 0x00002C3C,  0x00000000, 0, 0x00000000 } },
{ 0x00002C0D, { 0x00002C3D,  0x00000000, 0, 0x00000000 } },
{ 0x00002C0E, { 0x00002C3E,  0x00000000, 0, 0x00000000 } },
{ 0x00002C0F, { 0x00002C3F,  0x00000000, 0, 0x00000000 } },
{ 0x00002C10, { 0x00002C40,  0x00000000, 0, 0x00000000 } },
{ 0x00002C11, { 0x00002C41,  0x00000000, 0, 0x00000000 } },
{ 0x00002C12, { 0x00002C42,  0x00000000, 0, 0x00000000 } },
{ 0x00002C13, { 0x00002C43,  0x00000000, 0, 0x00000000 } },
{ 0x00002C14, { 0x00002C44,  0x00000000, 0, 0x00000000 } },
{ 0x00002C15, { 0x00002C45,  0x00000000, 0, 0x00000000 } },
{ 0x00002C16, { 0x00002C46,  0x00000000, 0, 0x00000000 } },
{ 0x00002C17, { 0x00002C47,  0x00000000, 0, 0x00000000 } },
{ 0x00002C18, { 0x00002C48,  0x00000000, 0, 0x00000000 } },
{ 0x00002C19, { 0x00002C49,  0x00000000, 0, 0x00000000 } },
{ 0x00002C1A, { 0x00002C4A,  0x00000000, 0, 0x00000000 } },
{ 0x00002C1B, { 0x00002C4B,  0x00000000, 0, 0x00000000 } },
{ 0x00002C1C, { 0x00002C4C,  0x00000000, 0, 0x00000000 } },
{ 0x00002C1D, { 0x00002C4D,  0x00000000, 0, 0x00000000 } },
{ 0x00002C1E, { 0x00002C4E,  0x00000000, 0, 0x00000000 } },
{ 0x00002C1F, { 0x00002C4F,  0x00000000, 0, 0x00000000 } },
{ 0x00002C20, { 0x00002C50,  0x00000000, 0, 0x00000000 } },
{ 0x00002C21, { 0x00002C51,  0x00000000, 0, 0x00000000 } },
{ 0x00002C22, { 0x00002C52,  0x00000000, 0, 0x00000000 } },
{ 0x00002C23, { 0x00002C53,  0x00000000, 0, 0x00000000 } },
{ 0x00002C24, { 0x00002C54,  0x00000000, 0, 0x00000000 } },
{ 0x00002C25, { 0x00002C55,  0x00000000, 0, 0x00000000 } },
{ 0x00002C26, { 0x00002C56,  0x00000000, 0, 0x00000000 } },
{ 0x00002C27, { 0x00002C57,  0x00000000, 0, 0x00000000 } },
{ 0x00002C28, { 0x00002C58,  0x00000000, 0, 0x00000000 } },
{ 0x00002C29, { 0x00002C59,  0x00000000, 0, 0x00000000 } },
{ 0x00002C2A, { 0x00002C5A,  0x00000000, 0, 0x00000000 } },
{ 0x00002C2B, { 0x00002C5B,  0x00000000, 0, 0x00000000 } },
{ 0x00002C2C, { 0x00002C5C,  0x00000000, 0, 0x00000000 } },
{ 0x00002C2D, { 0x00002C5D,  0x00000000, 0, 0x00000000 } },
{ 0x00002C2E, { 0x00002C5E,  0x00000000, 0, 0x00000000 } },
{ 0x00002C60, { 0x00002C61,  0x00000000, 0, 0x00000000 } },
{ 0x00002C62, { 0x0000026B,  0x00000000, 0, 0x00000000 } },
{ 0x00002C63, { 0x00001D7D,  0x00000000, 0, 0x00000000 } },
{ 0x00002C64, { 0x0000027D,  0x00000000, 0, 0x00000000 } },
{ 0x00002C67, { 0x00002C68,  0x00000000, 0, 0x00000000 } },
{ 0x00002C69, { 0x00002C6A,  0x00000000, 0, 0x00000000 } },
{ 0x00002C6B, { 0x00002C6C,  0x00000000, 0, 0x00000000 } },
{ 0x00002C6D, { 0x00000251,  0x00000000, 0, 0x00000000 } },
{ 0x00002C6E, { 0x00000271,  0x00000000, 0, 0x00000000 } },
{ 0x00002C6F, { 0x00000250,  0x00000000, 0, 0x00000000 } },
{ 0x00002C70, { 0x00000252,  0x00000000, 0, 0x00000000 } },
{ 0x00002C72, { 0x00002C73,  0x00000000, 0, 0x00000000 } },
{ 0x00002C75, { 0x00002C76,  0x00000000, 0, 0x00000000 } },
{ 0x00002C7E, { 0x0000023F,  0x00000000, 0, 0x00000000 } },
{ 0x00002C7F, { 0x00000240,  0x00000000, 0, 0x00000000 } },
{ 0x00002C80, { 0x00002C81,  0x00000000, 0, 0x00000000 } },
{ 0x00010C81, { 0x00010CC1,  0x00000000, 0, 0x00000000 } },
{ 0x00010C82, { 0x00010CC2,  0x00000000, 0, 0x00000000 } },
{ 0x00010C83, { 0x00010CC3,  0x00000000, 0, 0x00000000 } },
{ 0x00010C84, { 0x00010CC4,  0x00000000, 0, 0x00000000 } },
{ 0x00010C85, { 0x00010CC5,  0x00000000, 0, 0x00000000 } },
{ 0x00010C86, { 0x00010CC6,  0x00000000, 0, 0x00000000 } },
{ 0x00010C87, { 0x00010CC7,  0x00000000, 0, 0x00000000 } },
{ 0x00010C88, { 0x00010CC8,  0x00000000, 0, 0x00000000 } },
{ 0x00010C89, { 0x00010CC9,  0x00000000, 0, 0x00000000 } },
{ 0x00010C8A, { 0x00010CCA,  0x00000000, 0, 0x00000000 } },
{ 0x00010C8B, { 0x00010CCB,  0x00000000, 0, 0x00000000 } },
{ 0x00002C8C, { 0x00002C8D,  0x00000000, 0, 0x00000000 } },
{ 0x00010C8D, { 0x00010CCD,  0x00000000, 0, 0x00000000 } },
{ 0x00002C8E, { 0x00002C8F,  0x00000000, 0, 0x00000000 } },
{ 0x00010C8F, { 0x00010CCF,  0x00000000, 0, 0x00000000 } },
{ 0x00010C90, { 0x00010CD0,  0x00000000, 0, 0x00000000 } },
{ 0x00010C91, { 0x00010CD1,  0x00000000, 0, 0x00000000 } },
{ 0x00010C92, { 0x00010CD2,  0x00000000, 0, 0x00000000 } },
{ 0x00010C93, { 0x00010CD3,  0x00000000, 0, 0x00000000 } },
{ 0x00010C94, { 0x00010CD4,  0x00000000, 0, 0x00000000 } },
{ 0x00010C95, { 0x00010CD5,  0x00000000, 0, 0x00000000 } },
{ 0x00002C96, { 0x00002C97,  0x00000000, 0, 0x00000000 } },
{ 0x00010C97, { 0x00010CD7,  0x00000000, 0, 0x00000000 } },
{ 0x00002C98, { 0x00002C99,  0x00000000, 0, 0x00000000 } },
{ 0x00010C99, { 0x00010CD9,  0x00000000, 0, 0x00000000 } },
{ 0x00002C9A, { 0x00002C9B,  0x00000000, 0, 0x00000000 } },
{ 0x00010C9B, { 0x00010CDB,  0x00000000, 0, 0x00000000 } },
{ 0x00002C9C, { 0x00002C9D,  0x00000000, 0, 0x00000000 } },
{ 0x00010C9D, { 0x00010CDD,  0x00000000, 0, 0x00000000 } },
{ 0x00002C9E, { 0x00002C9F,  0x00000000, 0, 0x00000000 } },
{ 0x00010C9F, { 0x00010CDF,  0x00000000, 0, 0x00000000 } },
{ 0x00002CA0, { 0x00002CA1,  0x00000000, 0, 0x00000000 } },
{ 0x00010CA1, { 0x00010CE1,  0x00000000, 0, 0x00000000 } },
{ 0x00002CA2, { 0x00002CA3,  0x00000000, 0, 0x00000000 } },
{ 0x00010CA3, { 0x00010CE3,  0x00000000, 0, 0x00000000 } },
{ 0x00002CA4, { 0x00002CA5,  0x00000000, 0, 0x00000000 } },
{ 0x00010CA5, { 0x00010CE5,  0x00000000, 0, 0x00000000 } },
{ 0x00002CA6, { 0x00002CA7,  0x00000000, 0, 0x00000000 } },
{ 0x00010CA7, { 0x00010CE7,  0x00000000, 0, 0x00000000 } },
{ 0x00002CA8, { 0x00002CA9,  0x00000000, 0, 0x00000000 } },
{ 0x00010CA9, { 0x00010CE9,  0x00000000, 0, 0x00000000 } },
{ 0x00002CAA, { 0x00002CAB,  0x00000000, 0, 0x00000000 } },
{ 0x00010CAB, { 0x00010CEB,  0x00000000, 0, 0x00000000 } },
{ 0x00002CAC, { 0x00002CAD,  0x00000000, 0, 0x00000000 } },
{ 0x00010CAD, { 0x00010CED,  0x00000000, 0, 0x00000000 } },
{ 0x00002CAE, { 0x00002CAF,  0x00000000, 0, 0x00000000 } },
{ 0x00010CAF, { 0x00010CEF,  0x00000000, 0, 0x00000000 } },
{ 0x00002CB0, { 0x00002CB1,  0x00000000, 0, 0x00000000 } },
{ 0x00010CB1, { 0x00010CF1,  0x00000000, 0, 0x00000000 } },
{ 0x00002CB2, { 0x00002CB3,  0x00000000, 0, 0x00000000 } },
{ 0x00002CB4, { 0x00002CB5,  0x00000000, 0, 0x00000000 } },
{ 0x00002CB6, { 0x00002CB7,  0x00000000, 0, 0x00000000 } },
{ 0x00002CB8, { 0x00002CB9,  0x00000000, 0, 0x00000000 } },
{ 0x00002CBA, { 0x00002CBB,  0x00000000, 0, 0x00000000 } },
{ 0x00002CBC, { 0x00002CBD,  0x00000000, 0, 0x00000000 } },
{ 0x00002CBE, { 0x00002CBF,  0x00000000, 0, 0x00000000 } },
{ 0x00002CC0, { 0x00002CC1,  0x00000000, 0, 0x00000000 } },
{ 0x00002CC2, { 0x00002CC3,  0x00000000, 0, 0x00000000 } },
{ 0x00002CC4, { 0x00002CC5,  0x00000000, 0, 0x00000000 } },
{ 0x00002CC6, { 0x00002CC7,  0x00000000, 0, 0x00000000 } },
{ 0x00002CC8, { 0x00002CC9,  0x00000000, 0, 0x00000000 } },
{ 0x00002CCA, { 0x00002CCB,  0x00000000, 0, 0x00000000 } },
{ 0x00002CCC, { 0x00002CCD,  0x00000000, 0, 0x00000000 } },
{ 0x00002CCE, { 0x00002CCF,  0x00000000, 0, 0x00000000 } },
{ 0x00002CD0, { 0x00002CD1,  0x00000000, 0, 0x00000000 } },
{ 0x00002CD2, { 0x00002CD3,  0x00000000, 0, 0x00000000 } },
{ 0x00002CD4, { 0x00002CD5,  0x00000000, 0, 0x00000000 } },
{ 0x00002CD6, { 0x00002CD7,  0x00000000, 0, 0x00000000 } },
{ 0x00002CD8, { 0x00002CD9,  0x00000000, 0, 0x00000000 } },
{ 0x00002CDA, { 0x00002CDB,  0x00000000, 0, 0x00000000 } },
{ 0x00002CDC, { 0x00002CDD,  0x00000000, 0, 0x00000000 } },
{ 0x00002CDE, { 0x00002CDF,  0x00000000, 0, 0x00000000 } },
{ 0x00002CE0, { 0x00002CE1,  0x00000000, 0, 0x00000000 } },
{ 0x00002CE2, { 0x00002CE3,  0x00000000, 0, 0x00000000 } },
{ 0x00002CEB, { 0x00002CEC,  0x00000000, 0, 0x00000000 } },
{ 0x00002CED, { 0x00002CEE,  0x00000000, 0, 0x00000000 } },
{ 0x00002CF2, { 0x00002CF3,  0x00000000, 0, 0x00000000 } },
{ 0x00016E40, { 0x00016E60,  0x00000000, 0, 0x00000000 } },
{ 0x00016E41, { 0x00016E61,  0x00000000, 0, 0x00000000 } },
{ 0x00016E42, { 0x00016E62,  0x00000000, 0, 0x00000000 } },
{ 0x00016E43, { 0x00016E63,  0x00000000, 0, 0x00000000 } },
{ 0x00016E44, { 0x00016E64,  0x00000000, 0, 0x00000000 } },
{ 0x00016E45, { 0x00016E65,  0x00000000, 0, 0x00000000 } },
{ 0x00016E46, { 0x00016E66,  0x00000000, 0, 0x00000000 } },
{ 0x00016E47, { 0x00016E67,  0x00000000, 0, 0x00000000 } },
{ 0x00016E48, { 0x00016E68,  0x00000000, 0, 0x00000000 } },
{ 0x00016E49, { 0x00016E69,  0x00000000, 0, 0x00000000 } },
{ 0x00016E4A, { 0x00016E6A,  0x00000000, 0, 0x00000000 } },
{ 0x00016E4B, { 0x00016E6B,  0x00000000, 0, 0x00000000 } },
{ 0x00016E4C, { 0x00016E6C,  0x00000000, 0, 0x00000000 } },
{ 0x00016E4D, { 0x00016E6D,  0x00000000, 0, 0x00000000 } },
{ 0x00016E4E, { 0x00016E6E,  0x00000000, 0, 0x00000000 } },
{ 0x00016E4F, { 0x00016E6F,  0x00000000, 0, 0x00000000 } },
{ 0x00016E50, { 0x00016E70,  0x00000000, 0, 0x00000000 } },
{ 0x00016E51, { 0x00016E71,  0x00000000, 0, 0x00000000 } },
{ 0x00016E52, { 0x00016E72,  0x00000000, 0, 0x00000000 } },
{ 0x00016E53, { 0x00016E73,  0x00000000, 0, 0x00000000 } },
{ 0x00016E54, { 0x00016E74,  0x00000000, 0, 0x00000000 } },
{ 0x00016E55, { 0x00016E75,  0x00000000, 0, 0x00000000 } },
{ 0x00016E56, { 0x00016E76,  0x00000000, 0, 0x00000000 } },
{ 0x00016E57, { 0x00016E77,  0x00000000, 0, 0x00000000 } },
{ 0x00016E58, { 0x00016E78,  0x00000000, 0, 0x00000000 } },
{ 0x00016E59, { 0x00016E79,  0x00000000, 0, 0x00000000 } },
{ 0x00016E5A, { 0x00016E7A,  0x00000000, 0, 0x00000000 } },
{ 0x00016E5B, { 0x00016E7B,  0x00000000, 0, 0x00000000 } },
{ 0x00016E5C, { 0x00016E7C,  0x00000000, 0, 0x00000000 } },
{ 0x00016E5D, { 0x00016E7D,  0x00000000, 0, 0x00000000 } },
{ 0x00016E5E, { 0x00016E7E,  0x00000000, 0, 0x00000000 } },
{ 0x00016E5F, { 0x00016E7F,  0x00000000, 0, 0x00000000 } },
{ 0x000004BE, { 0x000004BF,  0x00000000, 0, 0x00000000 } },
{ 0x000024B6, { 0x000024D0,  0x00000000, 0, 0x00000000 } },
{ 0x000004C0, { 0x000004CF,  0x00000000, 0, 0x00000000 } },
{ 0x000024B8, { 0x000024D2,  0x00000000, 0, 0x00000000 } },
{ 0x000004C1, { 0x000004C2,  0x00000000, 0, 0x00000000 } },
{ 0x000024BA, { 0x000024D4,  0x00000000, 0, 0x00000000 } },
{ 0x000004C3, { 0x000004C4,  0x00000000, 0, 0x00000000 } },
{ 0x000004C5, { 0x000004C6,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF2E, { 0x0000FF4E,  0x00000000, 0, 0x00000000 } },
{ 0x000024C0, { 0x000024DA,  0x00000000, 0, 0x00000000 } },
{ 0x000004C9, { 0x000004CA,  0x00000000, 0, 0x00000000 } },
{ 0x000004CB, { 0x000004CC,  0x00000000, 0, 0x00000000 } },
{ 0x000024C4, { 0x000024DE,  0x00000000, 0, 0x00000000 } },
{ 0x000024C5, { 0x000024DF,  0x00000000, 0, 0x00000000 } },
{ 0x000024C6, { 0x000024E0,  0x00000000, 0, 0x00000000 } },
{ 0x000010A0, { 0x00002D00,  0x00000000, 0, 0x00000000 } },
{ 0x000010A1, { 0x00002D01,  0x00000000, 0, 0x00000000 } },
{ 0x000010A2, { 0x00002D02,  0x00000000, 0, 0x00000000 } },
{ 0x000010A3, { 0x00002D03,  0x00000000, 0, 0x00000000 } },
{ 0x000010A4, { 0x00002D04,  0x00000000, 0, 0x00000000 } },
{ 0x000010A5, { 0x00002D05,  0x00000000, 0, 0x00000000 } },
{ 0x000010A6, { 0x00002D06,  0x00000000, 0, 0x00000000 } },
{ 0x000010A7, { 0x00002D07,  0x00000000, 0, 0x00000000 } },
{ 0x000010A8, { 0x00002D08,  0x00000000, 0, 0x00000000 } },
{ 0x000010A9, { 0x00002D09,  0x00000000, 0, 0x00000000 } },
{ 0x000010AA, { 0x00002D0A,  0x00000000, 0, 0x00000000 } },
{ 0x000010AB, { 0x00002D0B,  0x00000000, 0, 0x00000000 } },
{ 0x000010AC, { 0x00002D0C,  0x00000000, 0, 0x00000000 } },
{ 0x000010AD, { 0x00002D0D,  0x00000000, 0, 0x00000000 } },
{ 0x000010AE, { 0x00002D0E,  0x00000000, 0, 0x00000000 } },
{ 0x000010AF, { 0x00002D0F,  0x00000000, 0, 0x00000000 } },
{ 0x000010B0, { 0x00002D10,  0x00000000, 0, 0x00000000 } },
{ 0x000010B1, { 0x00002D11,  0x00000000, 0, 0x00000000 } },
{ 0x000010B2, { 0x00002D12,  0x00000000, 0, 0x00000000 } },
{ 0x000010B3, { 0x00002D13,  0x00000000, 0, 0x00000000 } },
{ 0x000010B4, { 0x00002D14,  0x00000000, 0, 0x00000000 } },
{ 0x000010B5, { 0x00002D15,  0x00000000, 0, 0x00000000 } },
{ 0x000010B6, { 0x00002D16,  0x00000000, 0, 0x00000000 } },
{ 0x000010B7, { 0x00002D17,  0x00000000, 0, 0x00000000 } },
{ 0x000010B8, { 0x00002D18,  0x00000000, 0, 0x00000000 } },
{ 0x000010B9, { 0x00002D19,  0x00000000, 0, 0x00000000 } },
{ 0x000010BA, { 0x00002D1A,  0x00000000, 0, 0x00000000 } },
{ 0x000010BB, { 0x00002D1B,  0x00000000, 0, 0x00000000 } },
{ 0x000010BC, { 0x00002D1C,  0x00000000, 0, 0x00000000 } },
{ 0x000010BD, { 0x00002D1D,  0x00000000, 0, 0x00000000 } },
{ 0x000010BE, { 0x00002D1E,  0x00000000, 0, 0x00000000 } },
{ 0x000010BF, { 0x00002D1F,  0x00000000, 0, 0x00000000 } },
{ 0x000010C0, { 0x00002D20,  0x00000000, 0, 0x00000000 } },
{ 0x000010C1, { 0x00002D21,  0x00000000, 0, 0x00000000 } },
{ 0x000010C2, { 0x00002D22,  0x00000000, 0, 0x00000000 } },
{ 0x000010C3, { 0x00002D23,  0x00000000, 0, 0x00000000 } },
{ 0x000010C4, { 0x00002D24,  0x00000000, 0, 0x00000000 } },
{ 0x000010C5, { 0x00002D25,  0x00000000, 0, 0x00000000 } },
{ 0x000010C7, { 0x00002D27,  0x00000000, 0, 0x00000000 } },
{ 0x000010CD, { 0x00002D2D,  0x00000000, 0, 0x00000000 } },
{ 0x000024CD, { 0x000024E7,  0x00000000, 0, 0x00000000 } },
{ 0x000118A3, { 0x000118C3,  0x00000000, 0, 0x00000000 } },
{ 0x000118A9, { 0x000118C9,  0x00000000, 0, 0x00000000 } },
{ 0x000013F8, { 0x000013F0,  0x00000000, 0, 0x00000000 } },
{ 0x000013F9, { 0x000013F1,  0x00000000, 0, 0x00000000 } },
{ 0x000013FA, { 0x000013F2,  0x00000000, 0, 0x00000000 } },
{ 0x000013FB, { 0x000013F3,  0x00000000, 0, 0x00000000 } },
{ 0x000013FC, { 0x000013F4,  0x00000000, 0, 0x00000000 } },
{ 0x000013FD, { 0x000013F5,  0x00000000, 0, 0x00000000 } },
{ 0x000118AF, { 0x000118CF,  0x00000000, 0, 0x00000000 } },
{ 0x000118B5, { 0x000118D5,  0x00000000, 0, 0x00000000 } },
{ 0x000118BB, { 0x000118DB,  0x00000000, 0, 0x00000000 } },
{ 0x000104BC, { 0x000104E4,  0x00000000, 0, 0x00000000 } },
{ 0x000104C3, { 0x000104EB,  0x00000000, 0, 0x00000000 } },
{ 0x000104C7, { 0x000104EF,  0x00000000, 0, 0x00000000 } },
{ 0x000104C8, { 0x000104F0,  0x00000000, 0, 0x00000000 } },
{ 0x000104C9, { 0x000104F1,  0x00000000, 0, 0x00000000 } },
{ 0x000104CA, { 0x000104F2,  0x00000000, 0, 0x00000000 } },
{ 0x00000400, { 0x00000450,  0x00000000, 0, 0x00000000 } },
{ 0x00010401, { 0x00010429,  0x00000000, 0, 0x00000000 } },
{ 0x00000402, { 0x00000452,  0x00000000, 0, 0x00000000 } },
{ 0x00000403, { 0x00000453,  0x00000000, 0, 0x00000000 } },
{ 0x00000404, { 0x00000454,  0x00000000, 0, 0x00000000 } },
{ 0x00000405, { 0x00000455,  0x00000000, 0, 0x00000000 } },
{ 0x00000406, { 0x00000456,  0x00000000, 0, 0x00000000 } },
{ 0x00000407, { 0x00000457,  0x00000000, 0, 0x00000000 } },
{ 0x00010408, { 0x00010430,  0x00000000, 0, 0x00000000 } },
{ 0x00000409, { 0x00000459,  0x00000000, 0, 0x00000000 } },
{ 0x0000040A, { 0x0000045A,  0x00000000, 0, 0x00000000 } },
{ 0x0000040B, { 0x0000045B,  0x00000000, 0, 0x00000000 } },
{ 0x0000040C, { 0x0000045C,  0x00000000, 0, 0x00000000 } },
{ 0x0001040D, { 0x00010435,  0x00000000, 0, 0x00000000 } },
{ 0x0001040E, { 0x00010436,  0x00000000, 0, 0x00000000 } },
{ 0x0000040F, { 0x0000045F,  0x00000000, 0, 0x00000000 } },
{ 0x00000410, { 0x00000430,  0x00000000, 0, 0x00000000 } },
{ 0x00010411, { 0x00010439,  0x00000000, 0, 0x00000000 } },
{ 0x00000412, { 0x00000432,  0x00000000, 0, 0x00000000 } },
{ 0x00000413, { 0x00000433,  0x00000000, 0, 0x00000000 } },
{ 0x00000414, { 0x00000434,  0x00000000, 0, 0x00000000 } },
{ 0x00000415, { 0x00000435,  0x00000000, 0, 0x00000000 } },
{ 0x00010416, { 0x0001043E,  0x00000000, 0, 0x00000000 } },
{ 0x00000417, { 0x00000437,  0x00000000, 0, 0x00000000 } },
{ 0x00010418, { 0x00010440,  0x00000000, 0, 0x00000000 } },
{ 0x00010419, { 0x00010441,  0x00000000, 0, 0x00000000 } },
{ 0x0000041A, { 0x0000043A,  0x00000000, 0, 0x00000000 } },
{ 0x000118A0, { 0x000118C0,  0x00000000, 0, 0x00000000 } },
{ 0x000118A1, { 0x000118C1,  0x00000000, 0, 0x00000000 } },
{ 0x000118A2, { 0x000118C2,  0x00000000, 0, 0x00000000 } },
{ 0x0001041B, { 0x00010443,  0x00000000, 0, 0x00000000 } },
{ 0x000118A4, { 0x000118C4,  0x00000000, 0, 0x00000000 } },
{ 0x000118A5, { 0x000118C5,  0x00000000, 0, 0x00000000 } },
{ 0x000118A6, { 0x000118C6,  0x00000000, 0, 0x00000000 } },
{ 0x000118A7, { 0x000118C7,  0x00000000, 0, 0x00000000 } },
{ 0x000118A8, { 0x000118C8,  0x00000000, 0, 0x00000000 } },
{ 0x0000041C, { 0x0000043C,  0x00000000, 0, 0x00000000 } },
{ 0x000118AA, { 0x000118CA,  0x00000000, 0, 0x00000000 } },
{ 0x000118AB, { 0x000118CB,  0x00000000, 0, 0x00000000 } },
{ 0x000118AC, { 0x000118CC,  0x00000000, 0, 0x00000000 } },
{ 0x000118AD, { 0x000118CD,  0x00000000, 0, 0x00000000 } },
{ 0x000118AE, { 0x000118CE,  0x00000000, 0, 0x00000000 } },
{ 0x0000041D, { 0x0000043D,  0x00000000, 0, 0x00000000 } },
{ 0x000118B0, { 0x000118D0,  0x00000000, 0, 0x00000000 } },
{ 0x000118B1, { 0x000118D1,  0x00000000, 0, 0x00000000 } },
{ 0x000118B2, { 0x000118D2,  0x00000000, 0, 0x00000000 } },
{ 0x000118B3, { 0x000118D3,  0x00000000, 0, 0x00000000 } },
{ 0x000118B4, { 0x000118D4,  0x00000000, 0, 0x00000000 } },
{ 0x0000041E, { 0x0000043E,  0x00000000, 0, 0x00000000 } },
{ 0x000118B6, { 0x000118D6,  0x00000000, 0, 0x00000000 } },
{ 0x000118B7, { 0x000118D7,  0x00000000, 0, 0x00000000 } },
{ 0x000118B8, { 0x000118D8,  0x00000000, 0, 0x00000000 } },
{ 0x000118B9, { 0x000118D9,  0x00000000, 0, 0x00000000 } },
{ 0x000118BA, { 0x000118DA,  0x00000000, 0, 0x00000000 } },
{ 0x0000041F, { 0x0000043F,  0x00000000, 0, 0x00000000 } },
{ 0x000118BC, { 0x000118DC,  0x00000000, 0, 0x00000000 } },
{ 0x000118BD, { 0x000118DD,  0x00000000, 0, 0x00000000 } },
{ 0x000118BE, { 0x000118DE,  0x00000000, 0, 0x00000000 } },
{ 0x000118BF, { 0x000118DF,  0x00000000, 0, 0x00000000 } },
{ 0x00000420, { 0x00000440,  0x00000000, 0, 0x00000000 } },
{ 0x00010421, { 0x00010449,  0x00000000, 0, 0x00000000 } },
{ 0x00010422, { 0x0001044A,  0x00000000, 0, 0x00000000 } },
{ 0x00010423, { 0x0001044B,  0x00000000, 0, 0x00000000 } },
{ 0x00010424, { 0x0001044C,  0x00000000, 0, 0x00000000 } },
{ 0x00010425, { 0x0001044D,  0x00000000, 0, 0x00000000 } },
{ 0x00010426, { 0x0001044E,  0x00000000, 0, 0x00000000 } },
{ 0x00010427, { 0x0001044F,  0x00000000, 0, 0x00000000 } },
{ 0x00010CAE, { 0x00010CEE,  0x00000000, 0, 0x00000000 } },
{ 0x00010CB2, { 0x00010CF2,  0x00000000, 0, 0x00000000 } },
{ 0x00001F28, { 0x00001F20,  0x00000000, 0, 0x00000000 } },
{ 0x00001F29, { 0x00001F21,  0x00000000, 0, 0x00000000 } },
{ 0x00001F2A, { 0x00001F22,  0x00000000, 0, 0x00000000 } },
{ 0x0000FB00, { 0x00000000,  0x00000000, U"\U00000066\U00000066", 0x00000000 } },
{ 0x0000FB01, { 0x00000000,  0x00000000, U"\U00000066\U00000069", 0x00000000 } },
{ 0x0000FB02, { 0x00000000,  0x00000000, U"\U00000066\U0000006C", 0x00000000 } },
{ 0x0000FB03, { 0x00000000,  0x00000000, U"\U00000066\U00000066\U00000069", 0x00000000 } },
{ 0x0000FB04, { 0x00000000,  0x00000000, U"\U00000066\U00000066\U0000006C", 0x00000000 } },
{ 0x0000FB05, { 0x00000000,  0x00000000, U"\U00000073\U00000074", 0x00000000 } },
{ 0x0000FB06, { 0x00000000,  0x00000000, U"\U00000073\U00000074", 0x00000000 } },
{ 0x00001F2C, { 0x00001F24,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF2D, { 0x0000FF4D,  0x00000000, 0, 0x00000000 } },
{ 0x0000FB13, { 0x00000000,  0x00000000, U"\U00000574\U00000576", 0x00000000 } },
{ 0x0000FB14, { 0x00000000,  0x00000000, U"\U00000574\U00000565", 0x00000000 } },
{ 0x0000FB15, { 0x00000000,  0x00000000, U"\U00000574\U0000056B", 0x00000000 } },
{ 0x0000FB16, { 0x00000000,  0x00000000, U"\U0000057E\U00000576", 0x00000000 } },
{ 0x0000FB17, { 0x00000000,  0x00000000, U"\U00000574\U0000056D", 0x00000000 } },
{ 0x00001F2F, { 0x00001F27,  0x00000000, 0, 0x00000000 } },
{ 0x00001F38, { 0x00001F30,  0x00000000, 0, 0x00000000 } },
{ 0x00001F39, { 0x00001F31,  0x00000000, 0, 0x00000000 } },
{ 0x00001F3A, { 0x00001F32,  0x00000000, 0, 0x00000000 } },
{ 0x000004B0, { 0x000004B1,  0x00000000, 0, 0x00000000 } },
{ 0x000104B2, { 0x000104DA,  0x00000000, 0, 0x00000000 } },
{ 0x000004B4, { 0x000004B5,  0x00000000, 0, 0x00000000 } },
{ 0x000004B6, { 0x000004B7,  0x00000000, 0, 0x00000000 } },
{ 0x000104B7, { 0x000104DF,  0x00000000, 0, 0x00000000 } },
{ 0x000004B8, { 0x000004B9,  0x00000000, 0, 0x00000000 } },
{ 0x000024B9, { 0x000024D3,  0x00000000, 0, 0x00000000 } },
{ 0x000004BA, { 0x000004BB,  0x00000000, 0, 0x00000000 } },
{ 0x000024BB, { 0x000024D5,  0x00000000, 0, 0x00000000 } },
{ 0x000004BC, { 0x000004BD,  0x00000000, 0, 0x00000000 } },
{ 0x000024BD, { 0x000024D7,  0x00000000, 0, 0x00000000 } },
{ 0x000104BE, { 0x000104E6,  0x00000000, 0, 0x00000000 } },
{ 0x000024BF, { 0x000024D9,  0x00000000, 0, 0x00000000 } },
{ 0x00001C80, { 0x00000432,  0x00000000, 0, 0x00000000 } },
{ 0x00001C81, { 0x00000434,  0x00000000, 0, 0x00000000 } },
{ 0x00001C82, { 0x0000043E,  0x00000000, 0, 0x00000000 } },
{ 0x00001C83, { 0x00000441,  0x00000000, 0, 0x00000000 } },
{ 0x00001C84, { 0x00000442,  0x00000000, 0, 0x00000000 } },
{ 0x00001C85, { 0x00000442,  0x00000000, 0, 0x00000000 } },
{ 0x00001C86, { 0x0000044A,  0x00000000, 0, 0x00000000 } },
{ 0x000024C1, { 0x000024DB,  0x00000000, 0, 0x00000000 } },
{ 0x00001C88, { 0x0000A64B,  0x00000000, 0, 0x00000000 } },
{ 0x000024C2, { 0x000024DC,  0x00000000, 0, 0x00000000 } },
{ 0x00001C90, { 0x000010D0,  0x00000000, 0, 0x00000000 } },
{ 0x00001C91, { 0x000010D1,  0x00000000, 0, 0x00000000 } },
{ 0x00001C92, { 0x000010D2,  0x00000000, 0, 0x00000000 } },
{ 0x00001C93, { 0x000010D3,  0x00000000, 0, 0x00000000 } },
{ 0x00001C94, { 0x000010D4,  0x00000000, 0, 0x00000000 } },
{ 0x00001C95, { 0x000010D5,  0x00000000, 0, 0x00000000 } },
{ 0x00001C96, { 0x000010D6,  0x00000000, 0, 0x00000000 } },
{ 0x00001C97, { 0x000010D7,  0x00000000, 0, 0x00000000 } },
{ 0x00001C98, { 0x000010D8,  0x00000000, 0, 0x00000000 } },
{ 0x00001C99, { 0x000010D9,  0x00000000, 0, 0x00000000 } },
{ 0x00001C9A, { 0x000010DA,  0x00000000, 0, 0x00000000 } },
{ 0x00001C9B, { 0x000010DB,  0x00000000, 0, 0x00000000 } },
{ 0x00001C9C, { 0x000010DC,  0x00000000, 0, 0x00000000 } },
{ 0x00001C9D, { 0x000010DD,  0x00000000, 0, 0x00000000 } },
{ 0x00001C9E, { 0x000010DE,  0x00000000, 0, 0x00000000 } },
{ 0x00001C9F, { 0x000010DF,  0x00000000, 0, 0x00000000 } },
{ 0x00001CA0, { 0x000010E0,  0x00000000, 0, 0x00000000 } },
{ 0x00001CA1, { 0x000010E1,  0x00000000, 0, 0x00000000 } },
{ 0x00001CA2, { 0x000010E2,  0x00000000, 0, 0x00000000 } },
{ 0x00001CA3, { 0x000010E3,  0x00000000, 0, 0x00000000 } },
{ 0x00001CA4, { 0x000010E4,  0x00000000, 0, 0x00000000 } },
{ 0x00001CA5, { 0x000010E5,  0x00000000, 0, 0x00000000 } },
{ 0x00001CA6, { 0x000010E6,  0x00000000, 0, 0x00000000 } },
{ 0x00001CA7, { 0x000010E7,  0x00000000, 0, 0x00000000 } },
{ 0x00001CA8, { 0x000010E8,  0x00000000, 0, 0x00000000 } },
{ 0x00001CA9, { 0x000010E9,  0x00000000, 0, 0x00000000 } },
{ 0x00001CAA, { 0x000010EA,  0x00000000, 0, 0x00000000 } },
{ 0x000004C7, { 0x000004C8,  0x00000000, 0, 0x00000000 } },
{ 0x00001CAC, { 0x000010EC,  0x00000000, 0, 0x00000000 } },
{ 0x00001CAD, { 0x000010ED,  0x00000000, 0, 0x00000000 } },
{ 0x00001CAE, { 0x000010EE,  0x00000000, 0, 0x00000000 } },
{ 0x00001CAF, { 0x000010EF,  0x00000000, 0, 0x00000000 } },
{ 0x00001CB0, { 0x000010F0,  0x00000000, 0, 0x00000000 } },
{ 0x00001CB1, { 0x000010F1,  0x00000000, 0, 0x00000000 } },
{ 0x00001CB2, { 0x000010F2,  0x00000000, 0, 0x00000000 } },
{ 0x00001CB3, { 0x000010F3,  0x00000000, 0, 0x00000000 } },
{ 0x00001CB4, { 0x000010F4,  0x00000000, 0, 0x00000000 } },
{ 0x00001CB5, { 0x000010F5,  0x00000000, 0, 0x00000000 } },
{ 0x00001CB6, { 0x000010F6,  0x00000000, 0, 0x00000000 } },
{ 0x00001CB7, { 0x000010F7,  0x00000000, 0, 0x00000000 } },
{ 0x00001CB8, { 0x000010F8,  0x00000000, 0, 0x00000000 } },
{ 0x00001CB9, { 0x000010F9,  0x00000000, 0, 0x00000000 } },
{ 0x00001CBA, { 0x000010FA,  0x00000000, 0, 0x00000000 } },
{ 0x00001CBD, { 0x000010FD,  0x00000000, 0, 0x00000000 } },
{ 0x00001CBE, { 0x000010FE,  0x00000000, 0, 0x00000000 } },
{ 0x00001CBF, { 0x000010FF,  0x00000000, 0, 0x00000000 } },
{ 0x000024CB, { 0x000024E5,  0x00000000, 0, 0x00000000 } },
{ 0x000024CC, { 0x000024E6,  0x00000000, 0, 0x00000000 } },
{ 0x000004CD, { 0x000004CE,  0x00000000, 0, 0x00000000 } },
{ 0x000024CE, { 0x000024E8,  0x00000000, 0, 0x00000000 } },
{ 0x000024CF, { 0x000024E9,  0x00000000, 0, 0x00000000 } },
{ 0x000004D0, { 0x000004D1,  0x00000000, 0, 0x00000000 } },
{ 0x000004D2, { 0x000004D3,  0x00000000, 0, 0x00000000 } },
{ 0x00001CAB, { 0x000010EB,  0x00000000, 0, 0x00000000 } },
{ 0x00001E00, { 0x00001E01,  0x00000000, 0, 0x00000000 } },
{ 0x00001E02, { 0x00001E03,  0x00000000, 0, 0x00000000 } },
{ 0x00001E04, { 0x00001E05,  0x00000000, 0, 0x00000000 } },
{ 0x00001E06, { 0x00001E07,  0x00000000, 0, 0x00000000 } },
{ 0x00001E08, { 0x00001E09,  0x00000000, 0, 0x00000000 } },
{ 0x00001E0A, { 0x00001E0B,  0x00000000, 0, 0x00000000 } },
{ 0x00001E0C, { 0x00001E0D,  0x00000000, 0, 0x00000000 } },
{ 0x00001E0E, { 0x00001E0F,  0x00000000, 0, 0x00000000 } },
{ 0x00001E10, { 0x00001E11,  0x00000000, 0, 0x00000000 } },
{ 0x00001E12, { 0x00001E13,  0x00000000, 0, 0x00000000 } },
{ 0x00001E14, { 0x00001E15,  0x00000000, 0, 0x00000000 } },
{ 0x00001E16, { 0x00001E17,  0x00000000, 0, 0x00000000 } },
{ 0x00001E18, { 0x00001E19,  0x00000000, 0, 0x00000000 } },
{ 0x00001E1A, { 0x00001E1B,  0x00000000, 0, 0x00000000 } },
{ 0x00001E1C, { 0x00001E1D,  0x00000000, 0, 0x00000000 } },
{ 0x00001E1E, { 0x00001E1F,  0x00000000, 0, 0x00000000 } },
{ 0x00001E20, { 0x00001E21,  0x00000000, 0, 0x00000000 } },
{ 0x00001E22, { 0x00001E23,  0x00000000, 0, 0x00000000 } },
{ 0x00001E24, { 0x00001E25,  0x00000000, 0, 0x00000000 } },
{ 0x00001E26, { 0x00001E27,  0x00000000, 0, 0x00000000 } },
{ 0x00001E28, { 0x00001E29,  0x00000000, 0, 0x00000000 } },
{ 0x00001E2A, { 0x00001E2B,  0x00000000, 0, 0x00000000 } },
{ 0x00001E2C, { 0x00001E2D,  0x00000000, 0, 0x00000000 } },
{ 0x00001E2E, { 0x00001E2F,  0x00000000, 0, 0x00000000 } },
{ 0x00001E30, { 0x00001E31,  0x00000000, 0, 0x00000000 } },
{ 0x00001E32, { 0x00001E33,  0x00000000, 0, 0x00000000 } },
{ 0x00001E34, { 0x00001E35,  0x00000000, 0, 0x00000000 } },
{ 0x00001E36, { 0x00001E37,  0x00000000, 0, 0x00000000 } },
{ 0x00001E38, { 0x00001E39,  0x00000000, 0, 0x00000000 } },
{ 0x00001E3A, { 0x00001E3B,  0x00000000, 0, 0x00000000 } },
{ 0x00001E3C, { 0x00001E3D,  0x00000000, 0, 0x00000000 } },
{ 0x00001E3E, { 0x00001E3F,  0x00000000, 0, 0x00000000 } },
{ 0x00001E40, { 0x00001E41,  0x00000000, 0, 0x00000000 } },
{ 0x00001E42, { 0x00001E43,  0x00000000, 0, 0x00000000 } },
{ 0x00001E44, { 0x00001E45,  0x00000000, 0, 0x00000000 } },
{ 0x00001E46, { 0x00001E47,  0x00000000, 0, 0x00000000 } },
{ 0x00001E48, { 0x00001E49,  0x00000000, 0, 0x00000000 } },
{ 0x00001E4A, { 0x00001E4B,  0x00000000, 0, 0x00000000 } },
{ 0x00001E4C, { 0x00001E4D,  0x00000000, 0, 0x00000000 } },
{ 0x00001E4E, { 0x00001E4F,  0x00000000, 0, 0x00000000 } },
{ 0x00001E50, { 0x00001E51,  0x00000000, 0, 0x00000000 } },
{ 0x00001E52, { 0x00001E53,  0x00000000, 0, 0x00000000 } },
{ 0x00001E54, { 0x00001E55,  0x00000000, 0, 0x00000000 } },
{ 0x00001E56, { 0x00001E57,  0x00000000, 0, 0x00000000 } },
{ 0x00001E58, { 0x00001E59,  0x00000000, 0, 0x00000000 } },
{ 0x00001E5A, { 0x00001E5B,  0x00000000, 0, 0x00000000 } },
{ 0x00001E5C, { 0x00001E5D,  0x00000000, 0, 0x00000000 } },
{ 0x00001E5E, { 0x00001E5F,  0x00000000, 0, 0x00000000 } },
{ 0x00001E60, { 0x00001E61,  0x00000000, 0, 0x00000000 } },
{ 0x00001E62, { 0x00001E63,  0x00000000, 0, 0x00000000 } },
{ 0x00001E64, { 0x00001E65,  0x00000000, 0, 0x00000000 } },
{ 0x00001E66, { 0x00001E67,  0x00000000, 0, 0x00000000 } },
{ 0x00001E68, { 0x00001E69,  0x00000000, 0, 0x00000000 } },
{ 0x00001E6A, { 0x00001E6B,  0x00000000, 0, 0x00000000 } },
{ 0x00001E6C, { 0x00001E6D,  0x00000000, 0, 0x00000000 } },
{ 0x00001E6E, { 0x00001E6F,  0x00000000, 0, 0x00000000 } },
{ 0x00001E70, { 0x00001E71,  0x00000000, 0, 0x00000000 } },
{ 0x00001E72, { 0x00001E73,  0x00000000, 0, 0x00000000 } },
{ 0x00001E74, { 0x00001E75,  0x00000000, 0, 0x00000000 } },
{ 0x00001E76, { 0x00001E77,  0x00000000, 0, 0x00000000 } },
{ 0x00001E78, { 0x00001E79,  0x00000000, 0, 0x00000000 } },
{ 0x00001E7A, { 0x00001E7B,  0x00000000, 0, 0x00000000 } },
{ 0x00001E7C, { 0x00001E7D,  0x00000000, 0, 0x00000000 } },
{ 0x00001E7E, { 0x00001E7F,  0x00000000, 0, 0x00000000 } },
{ 0x00001E80, { 0x00001E81,  0x00000000, 0, 0x00000000 } },
{ 0x00001E82, { 0x00001E83,  0x00000000, 0, 0x00000000 } },
{ 0x00001E84, { 0x00001E85,  0x00000000, 0, 0x00000000 } },
{ 0x00001E86, { 0x00001E87,  0x00000000, 0, 0x00000000 } },
{ 0x00001E88, { 0x00001E89,  0x00000000, 0, 0x00000000 } },
{ 0x00001E8A, { 0x00001E8B,  0x00000000, 0, 0x00000000 } },
{ 0x00001E8C, { 0x00001E8D,  0x00000000, 0, 0x00000000 } },
{ 0x00001E8E, { 0x00001E8F,  0x00000000, 0, 0x00000000 } },
{ 0x00001E90, { 0x00001E91,  0x00000000, 0, 0x00000000 } },
{ 0x00001E92, { 0x00001E93,  0x00000000, 0, 0x00000000 } },
{ 0x00001E94, { 0x00001E95,  0x00000000, 0, 0x00000000 } },
{ 0x00001E96, { 0x00000000,  0x00000000, U"\U00000068\U00000331", 0x00000000 } },
{ 0x00001E97, { 0x00000000,  0x00000000, U"\U00000074\U00000308", 0x00000000 } },
{ 0x00001E98, { 0x00000000,  0x00000000, U"\U00000077\U0000030A", 0x00000000 } },
{ 0x00001E99, { 0x00000000,  0x00000000, U"\U00000079\U0000030A", 0x00000000 } },
{ 0x00001E9A, { 0x00000000,  0x00000000, U"\U00000061\U000002BE", 0x00000000 } },
{ 0x00001E9B, { 0x00001E61,  0x00000000, 0, 0x00000000 } },
{ 0x00001E9E, { 0x00000000,  0x000000DF, U"\U00000073\U00000073", 0x00000000 } },
{ 0x00001EA0, { 0x00001EA1,  0x00000000, 0, 0x00000000 } },
{ 0x00001EA2, { 0x00001EA3,  0x00000000, 0, 0x00000000 } },
{ 0x00001EA4, { 0x00001EA5,  0x00000000, 0, 0x00000000 } },
{ 0x00001EA6, { 0x00001EA7,  0x00000000, 0, 0x00000000 } },
{ 0x00001EA8, { 0x00001EA9,  0x00000000, 0, 0x00000000 } },
{ 0x00001EAA, { 0x00001EAB,  0x00000000, 0, 0x00000000 } },
{ 0x00001EAC, { 0x00001EAD,  0x00000000, 0, 0x00000000 } },
{ 0x00001EAE, { 0x00001EAF,  0x00000000, 0, 0x00000000 } },
{ 0x00001EB0, { 0x00001EB1,  0x00000000, 0, 0x00000000 } },
{ 0x00001EB2, { 0x00001EB3,  0x00000000, 0, 0x00000000 } },
{ 0x00001EB4, { 0x00001EB5,  0x00000000, 0, 0x00000000 } },
{ 0x00001EB6, { 0x00001EB7,  0x00000000, 0, 0x00000000 } },
{ 0x00001EB8, { 0x00001EB9,  0x00000000, 0, 0x00000000 } },
{ 0x00001EBA, { 0x00001EBB,  0x00000000, 0, 0x00000000 } },
{ 0x00001EBC, { 0x00001EBD,  0x00000000, 0, 0x00000000 } },
{ 0x00001EBE, { 0x00001EBF,  0x00000000, 0, 0x00000000 } },
{ 0x00001EC0, { 0x00001EC1,  0x00000000, 0, 0x00000000 } },
{ 0x00001EC2, { 0x00001EC3,  0x00000000, 0, 0x00000000 } },
{ 0x00001EC4, { 0x00001EC5,  0x00000000, 0, 0x00000000 } },
{ 0x00001EC6, { 0x00001EC7,  0x00000000, 0, 0x00000000 } },
{ 0x00001EC8, { 0x00001EC9,  0x00000000, 0, 0x00000000 } },
{ 0x00001ECA, { 0x00001ECB,  0x00000000, 0, 0x00000000 } },
{ 0x00001ECC, { 0x00001ECD,  0x00000000, 0, 0x00000000 } },
{ 0x00001ECE, { 0x00001ECF,  0x00000000, 0, 0x00000000 } },
{ 0x00001ED0, { 0x00001ED1,  0x00000000, 0, 0x00000000 } },
{ 0x00001ED2, { 0x00001ED3,  0x00000000, 0, 0x00000000 } },
{ 0x00001ED4, { 0x00001ED5,  0x00000000, 0, 0x00000000 } },
{ 0x00001ED6, { 0x00001ED7,  0x00000000, 0, 0x00000000 } },
{ 0x00001ED8, { 0x00001ED9,  0x00000000, 0, 0x00000000 } },
{ 0x00001EDA, { 0x00001EDB,  0x00000000, 0, 0x00000000 } },
{ 0x00001EDC, { 0x00001EDD,  0x00000000, 0, 0x00000000 } },
{ 0x00001EDE, { 0x00001EDF,  0x00000000, 0, 0x00000000 } },
{ 0x00001EE0, { 0x00001EE1,  0x00000000, 0, 0x00000000 } },
{ 0x00001EE2, { 0x00001EE3,  0x00000000, 0, 0x00000000 } },
{ 0x00001EE4, { 0x00001EE5,  0x00000000, 0, 0x00000000 } },
{ 0x00001EE6, { 0x00001EE7,  0x00000000, 0, 0x00000000 } },
{ 0x00001EE8, { 0x00001EE9,  0x00000000, 0, 0x00000000 } },
{ 0x00001EEA, { 0x00001EEB,  0x00000000, 0, 0x00000000 } },
{ 0x00001EEC, { 0x00001EED,  0x00000000, 0, 0x00000000 } },
{ 0x00001EEE, { 0x00001EEF,  0x00000000, 0, 0x00000000 } },
{ 0x00001EF0, { 0x00001EF1,  0x00000000, 0, 0x00000000 } },
{ 0x00001EF2, { 0x00001EF3,  0x00000000, 0, 0x00000000 } },
{ 0x00001EF4, { 0x00001EF5,  0x00000000, 0, 0x00000000 } },
{ 0x00001EF6, { 0x00001EF7,  0x00000000, 0, 0x00000000 } },
{ 0x00001EF8, { 0x00001EF9,  0x00000000, 0, 0x00000000 } },
{ 0x00001EFA, { 0x00001EFB,  0x00000000, 0, 0x00000000 } },
{ 0x00001EFC, { 0x00001EFD,  0x00000000, 0, 0x00000000 } },
{ 0x00001EFE, { 0x00001EFF,  0x00000000, 0, 0x00000000 } },
{ 0x00001F08, { 0x00001F00,  0x00000000, 0, 0x00000000 } },
{ 0x00001F09, { 0x00001F01,  0x00000000, 0, 0x00000000 } },
{ 0x00001F0A, { 0x00001F02,  0x00000000, 0, 0x00000000 } },
{ 0x00001F0B, { 0x00001F03,  0x00000000, 0, 0x00000000 } },
{ 0x00001F0C, { 0x00001F04,  0x00000000, 0, 0x00000000 } },
{ 0x00001F0D, { 0x00001F05,  0x00000000, 0, 0x00000000 } },
{ 0x00001F0E, { 0x00001F06,  0x00000000, 0, 0x00000000 } },
{ 0x00001F0F, { 0x00001F07,  0x00000000, 0, 0x00000000 } },
{ 0x00001F18, { 0x00001F10,  0x00000000, 0, 0x00000000 } },
{ 0x00001F19, { 0x00001F11,  0x00000000, 0, 0x00000000 } },
{ 0x00001F1A, { 0x00001F12,  0x00000000, 0, 0x00000000 } },
{ 0x00001F1B, { 0x00001F13,  0x00000000, 0, 0x00000000 } },
{ 0x00001F1C, { 0x00001F14,  0x00000000, 0, 0x00000000 } },
{ 0x00001F1D, { 0x00001F15,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF21, { 0x0000FF41,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF22, { 0x0000FF42,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF23, { 0x0000FF43,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF24, { 0x0000FF44,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF25, { 0x0000FF45,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF26, { 0x0000FF46,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF27, { 0x0000FF47,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF28, { 0x0000FF48,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF29, { 0x0000FF49,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF2A, { 0x0000FF4A,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF2B, { 0x0000FF4B,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF2C, { 0x0000FF4C,  0x00000000, 0, 0x00000000 } },
{ 0x00001F2D, { 0x00001F25,  0x00000000, 0, 0x00000000 } },
{ 0x00001F2E, { 0x00001F26,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF2F, { 0x0000FF4F,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF30, { 0x0000FF50,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF31, { 0x0000FF51,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF32, { 0x0000FF52,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF33, { 0x0000FF53,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF34, { 0x0000FF54,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF35, { 0x0000FF55,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF36, { 0x0000FF56,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF37, { 0x0000FF57,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF38, { 0x0000FF58,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF39, { 0x0000FF59,  0x00000000, 0, 0x00000000 } },
{ 0x0000FF3A, { 0x0000FF5A,  0x00000000, 0, 0x00000000 } },
{ 0x00001F3B, { 0x00001F33,  0x00000000, 0, 0x00000000 } },
{ 0x00001F3C, { 0x00001F34,  0x00000000, 0, 0x00000000 } },
{ 0x00001F3D, { 0x00001F35,  0x00000000, 0, 0x00000000 } },
{ 0x00001F3E, { 0x00001F36,  0x00000000, 0, 0x00000000 } },
{ 0x00001F3F, { 0x00001F37,  0x00000000, 0, 0x00000000 } },
{ 0x00001F48, { 0x00001F40,  0x00000000, 0, 0x00000000 } },
{ 0x00001F49, { 0x00001F41,  0x00000000, 0, 0x00000000 } },
{ 0x00001F4A, { 0x00001F42,  0x00000000, 0, 0x00000000 } },
{ 0x00001F4B, { 0x00001F43,  0x00000000, 0, 0x00000000 } },
{ 0x00001F4C, { 0x00001F44,  0x00000000, 0, 0x00000000 } },
{ 0x00001F4D, { 0x00001F45,  0x00000000, 0, 0x00000000 } },
{ 0x00001F50, { 0x00000000,  0x00000000, U"\U000003C5\U00000313", 0x00000000 } },
{ 0x00001F52, { 0x00000000,  0x00000000, U"\U000003C5\U00000313\U00000300", 0x00000000 } },
{ 0x00001F54, { 0x00000000,  0x00000000, U"\U000003C5\U00000313\U00000301", 0x00000000 } },
{ 0x00001F56, { 0x00000000,  0x00000000, U"\U000003C5\U00000313\U00000342", 0x00000000 } },
{ 0x00001F59, { 0x00001F51,  0x00000000, 0, 0x00000000 } },
{ 0x00001F5B, { 0x00001F53,  0x00000000, 0, 0x00000000 } },
{ 0x00001F5D, { 0x00001F55,  0x00000000, 0, 0x00000000 } },
{ 0x00001F5F, { 0x00001F57,  0x00000000, 0, 0x00000000 } },
{ 0x00001F68, { 0x00001F60,  0x00000000, 0, 0x00000000 } },
{ 0x00001F69, { 0x00001F61,  0x00000000, 0, 0x00000000 } },
{ 0x00001F6A, { 0x00001F62,  0x00000000, 0, 0x00000000 } },
{ 0x00001F6B, { 0x00001F63,  0x00000000, 0, 0x00000000 } },
{ 0x00001F6C, { 0x00001F64,  0x00000000, 0, 0x00000000 } },
{ 0x00001F6D, { 0x00001F65,  0x00000000, 0, 0x00000000 } },
{ 0x00001F6E, { 0x00001F66,  0x00000000, 0, 0x00000000 } },
{ 0x00001F6F, { 0x00001F67,  0x00000000, 0, 0x00000000 } },
{ 0x00001F80, { 0x00000000,  0x00000000, U"\U00001F00\U000003B9", 0x00000000 } },
{ 0x00001F81, { 0x00000000,  0x00000000, U"\U00001F01\U000003B9", 0x00000000 } },
{ 0x00001F82, { 0x00000000,  0x00000000, U"\U00001F02\U000003B9", 0x00000000 } },
{ 0x00001F83, { 0x00000000,  0x00000000, U"\U00001F03\U000003B9", 0x00000000 } },
{ 0x00001F84, { 0x00000000,  0x00000000, U"\U00001F04\U000003B9", 0x00000000 } },
{ 0x00001F85, { 0x00000000,  0x00000000, U"\U00001F05\U000003B9", 0x00000000 } },
{ 0x00001F86, { 0x00000000,  0x00000000, U"\U00001F06\U000003B9", 0x00000000 } },
{ 0x00001F87, { 0x00000000,  0x00000000, U"\U00001F07\U000003B9", 0x00000000 } },
{ 0x00001F88, { 0x00000000,  0x00001F80, U"\U00001F00\U000003B9", 0x00000000 } },
{ 0x00001F89, { 0x00000000,  0x00001F81, U"\U00001F01\U000003B9", 0x00000000 } },
{ 0x00001F8A, { 0x00000000,  0x00001F82, U"\U00001F02\U000003B9", 0x00000000 } },
{ 0x00001F8B, { 0x00000000,  0x00001F83, U"\U00001F03\U000003B9", 0x00000000 } },
{ 0x00001F8C, { 0x00000000,  0x00001F84, U"\U00001F04\U000003B9", 0x00000000 } },
{ 0x00001F8D, { 0x00000000,  0x00001F85, U"\U00001F05\U000003B9", 0x00000000 } },
{ 0x00001F8E, { 0x00000000,  0x00001F86, U"\U00001F06\U000003B9", 0x00000000 } },
{ 0x00001F8F, { 0x00000000,  0x00001F87, U"\U00001F07\U000003B9", 0x00000000 } },
{ 0x00001F90, { 0x00000000,  0x00000000, U"\U00001F20\U000003B9", 0x00000000 } },
{ 0x00001F91, { 0x00000000,  0x00000000, U"\U00001F21\U000003B9", 0x00000000 } },
{ 0x00001F92, { 0x00000000,  0x00000000, U"\U00001F22\U000003B9", 0x00000000 } },
{ 0x00001F93, { 0x00000000,  0x00000000, U"\U00001F23\U000003B9", 0x00000000 } },
{ 0x00001F94, { 0x00000000,  0x00000000, U"\U00001F24\U000003B9", 0x00000000 } },
{ 0x00001F95, { 0x00000000,  0x00000000, U"\U00001F25\U000003B9", 0x00000000 } },
{ 0x00001F96, { 0x00000000,  0x00000000, U"\U00001F26\U000003B9", 0x00000000 } },
{ 0x00001F97, { 0x00000000,  0x00000000, U"\U00001F27\U000003B9", 0x00000000 } },
{ 0x00001F98, { 0x00000000,  0x00001F90, U"\U00001F20\U000003B9", 0x00000000 } },
{ 0x00001F99, { 0x00000000,  0x00001F91, U"\U00001F21\U000003B9", 0x00000000 } },
{ 0x00001F9A, { 0x00000000,  0x00001F92, U"\U00001F22\U000003B9", 0x00000000 } },
{ 0x00001F9B, { 0x00000000,  0x00001F93, U"\U00001F23\U000003B9", 0x00000000 } },
{ 0x00001F9C, { 0x00000000,  0x00001F94, U"\U00001F24\U000003B9", 0x00000000 } },
{ 0x00001F9D, { 0x00000000,  0x00001F95, U"\U00001F25\U000003B9", 0x00000000 } },
{ 0x00001F9E, { 0x00000000,  0x00001F96, U"\U00001F26\U000003B9", 0x00000000 } },
{ 0x00001F9F, { 0x00000000,  0x00001F97, U"\U00001F27\U000003B9", 0x00000000 } },
{ 0x00001FA0, { 0x00000000,  0x00000000, U"\U00001F60\U000003B9", 0x00000000 } },
{ 0x00001FA1, { 0x00000000,  0x00000000, U"\U00001F61\U000003B9", 0x00000000 } },
{ 0x00001FA2, { 0x00000000,  0x00000000, U"\U00001F62\U000003B9", 0x00000000 } },
{ 0x00001FA3, { 0x00000000,  0x00000000, U"\U00001F63\U000003B9", 0x00000000 } },
{ 0x00001FA4, { 0x00000000,  0x00000000, U"\U00001F64\U000003B9", 0x00000000 } },
{ 0x00001FA5, { 0x00000000,  0x00000000, U"\U00001F65\U000003B9", 0x00000000 } },
{ 0x00001FA6, { 0x00000000,  0x00000000, U"\U00001F66\U000003B9", 0x00000000 } },
{ 0x00001FA7, { 0x00000000,  0x00000000, U"\U00001F67\U000003B9", 0x00000000 } },
{ 0x00001FA8, { 0x00000000,  0x00001FA0, U"\U00001F60\U000003B9", 0x00000000 } },
{ 0x00001FA9, { 0x00000000,  0x00001FA1, U"\U00001F61\U000003B9", 0x00000000 } },
{ 0x00001FAA, { 0x00000000,  0x00001FA2, U"\U00001F62\U000003B9", 0x00000000 } },
{ 0x00001FAB, { 0x00000000,  0x00001FA3, U"\U00001F63\U000003B9", 0x00000000 } },
{ 0x00001FAC, { 0x00000000,  0x00001FA4, U"\U00001F64\U000003B9", 0x00000000 } },
{ 0x00001FAD, { 0x00000000,  0x00001FA5, U"\U00001F65\U000003B9", 0x00000000 } },
{ 0x00001FAE, { 0x00000000,  0x00001FA6, U"\U00001F66\U000003B9", 0x00000000 } },
{ 0x00001FAF, { 0x00000000,  0x00001FA7, U"\U00001F67\U000003B9", 0x00000000 } },
{ 0x00001FB2, { 0x00000000,  0x00000000, U"\U00001F70\U000003B9", 0x00000000 } },
{ 0x00001FB3, { 0x00000000,  0x00000000, U"\U000003B1\U000003B9", 0x00000000 } },
{ 0x00001FB4, { 0x00000000,  0x00000000, U"\U000003AC\U000003B9", 0x00000000 } },
{ 0x00001FB6, { 0x00000000,  0x00000000, U"\U000003B1\U00000342", 0x00000000 } },
{ 0x00001FB7, { 0x00000000,  0x00000000, U"\U000003B1\U00000342\U000003B9", 0x00000000 } },
{ 0x00001FB8, { 0x00001FB0,  0x00000000, 0, 0x00000000 } },
{ 0x00001FB9, { 0x00001FB1,  0x00000000, 0, 0x00000000 } },
{ 0x00001FBA, { 0x00001F70,  0x00000000, 0, 0x00000000 } },
{ 0x00001FBB, { 0x00001F71,  0x00000000, 0, 0x00000000 } },
{ 0x00001FBC, { 0x00000000,  0x00001FB3, U"\U000003B1\U000003B9", 0x00000000 } },
{ 0x00001FBE, { 0x000003B9,  0x00000000, 0, 0x00000000 } },
{ 0x00001FC2, { 0x00000000,  0x00000000, U"\U00001F74\U000003B9", 0x00000000 } },
{ 0x00001FC3, { 0x00000000,  0x00000000, U"\U000003B7\U000003B9", 0x00000000 } },
{ 0x00001FC4, { 0x00000000,  0x00000000, U"\U000003AE\U000003B9", 0x00000000 } },
{ 0x00001FC6, { 0x00000000,  0x00000000, U"\U000003B7\U00000342", 0x00000000 } },
{ 0x00001FC7, { 0x00000000,  0x00000000, U"\U000003B7\U00000342\U000003B9", 0x00000000 } },
{ 0x00001FC8, { 0x00001F72,  0x00000000, 0, 0x00000000 } },
{ 0x00001FC9, { 0x00001F73,  0x00000000, 0, 0x00000000 } },
{ 0x00001FCA, { 0x00001F74,  0x00000000, 0, 0x00000000 } },
{ 0x00001FCB, { 0x00001F75,  0x00000000, 0, 0x00000000 } },
{ 0x00001FCC, { 0x00000000,  0x00001FC3, U"\U000003B7\U000003B9", 0x00000000 } },
{ 0x00001FD2, { 0x00000000,  0x00000000, U"\U000003B9\U00000308\U00000300", 0x00000000 } },
{ 0x00001FD3, { 0x00000000,  0x00000000, U"\U000003B9\U00000308\U00000301", 0x00000000 } },
{ 0x00001FD6, { 0x00000000,  0x00000000, U"\U000003B9\U00000342", 0x00000000 } },
{ 0x00001FD7, { 0x00000000,  0x00000000, U"\U000003B9\U00000308\U00000342", 0x00000000 } },
{ 0x00001FD8, { 0x00001FD0,  0x00000000, 0, 0x00000000 } },
{ 0x00001FD9, { 0x00001FD1,  0x00000000, 0, 0x00000000 } },
{ 0x00001FDA, { 0x00001F76,  0x00000000, 0, 0x00000000 } },
{ 0x00001FDB, { 0x00001F77,  0x00000000, 0, 0x00000000 } },
{ 0x00001FE2, { 0x00000000,  0x00000000, U"\U000003C5\U00000308\U00000300", 0x00000000 } },
{ 0x00001FE3, { 0x00000000,  0x00000000, U"\U000003C5\U00000308\U00000301", 0x00000000 } },
{ 0x00001FE4, { 0x00000000,  0x00000000, U"\U000003C1\U00000313", 0x00000000 } },
{ 0x00001FE6, { 0x00000000,  0x00000000, U"\U000003C5\U00000342", 0x00000000 } },
{ 0x00001FE7, { 0x00000000,  0x00000000, U"\U000003C5\U00000308\U00000342", 0x00000000 } },
{ 0x00001FE8, { 0x00001FE0,  0x00000000, 0, 0x00000000 } },
{ 0x00001FE9, { 0x00001FE1,  0x00000000, 0, 0x00000000 } },
{ 0x00001FEA, { 0x00001F7A,  0x00000000, 0, 0x00000000 } },
{ 0x00001FEB, { 0x00001F7B,  0x00000000, 0, 0x00000000 } },
{ 0x00001FEC, { 0x00001FE5,  0x00000000, 0, 0x00000000 } },
{ 0x00001FF2, { 0x00000000,  0x00000000, U"\U00001F7C\U000003B9", 0x00000000 } },
{ 0x00001FF3, { 0x00000000,  0x00000000, U"\U000003C9\U000003B9", 0x00000000 } },
{ 0x00001FF4, { 0x00000000,  0x00000000, U"\U000003CE\U000003B9", 0x00000000 } },
{ 0x00001FF6, { 0x00000000,  0x00000000, U"\U000003C9\U00000342", 0x00000000 } },
{ 0x00001FF7, { 0x00000000,  0x00000000, U"\U000003C9\U00000342\U000003B9", 0x00000000 } },
{ 0x00001FF8, { 0x00001F78,  0x00000000, 0, 0x00000000 } },
{ 0x00001FF9, { 0x00001F79,  0x00000000, 0, 0x00000000 } },
{ 0x00001FFA, { 0x00001F7C,  0x00000000, 0, 0x00000000 } },
{ 0x00001FFB, { 0x00001F7D,  0x00000000, 0, 0x00000000 } },
{ 0x00001FFC, { 0x00000000,  0x00001FF3, U"\U000003C9\U000003B9", 0x00000000 } },
};
namespace _block_properties {
using T = Block;
const auto D = Block::Unassigned;
static const size_t _block_size = 128;
static const Block _4[] = { T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::LatinExtendedB,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions, };
static const Block _5[] = { T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::IPAExtensions,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters,T::SpacingModifierLetters, };
static const Block _6[] = { T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::CombiningDiacriticalMarks,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic,T::GreekAndCoptic, };
static const Block _10[] = { T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::CyrillicSupplement,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian, };
static const Block _11[] = { T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew, };
static const Block _14[] = { T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement,T::ArabicSupplement, };
static const Block _15[] = { T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo,T::NKo, };
static const Block _16[] = { T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,T::SyriacSupplement,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _17[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA,T::ArabicExtendedA, };
static const Block _33[] = { T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian, };
static const Block _39[] = { T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::EthiopicSupplement,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee, };
static const Block _45[] = { T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic, };
static const Block _46[] = { T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa, };
static const Block _49[] = { T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended,T::UnifiedCanadianAboriginalSyllabicsExtended, };
static const Block _50[] = { T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe,T::TaiLe, };
static const Block _51[] = { T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::NewTaiLue,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols,T::KhmerSymbols, };
static const Block _52[] = { T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham, };
static const Block _53[] = { T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::TaiTham,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended,T::CombiningDiacriticalMarksExtended, };
static const Block _55[] = { T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak, };
static const Block _56[] = { T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki,T::OlChiki, };
static const Block _57[] = { T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::CyrillicExtendedC,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::GeorgianExtended,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::SundaneseSupplement,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions,T::VedicExtensions, };
static const Block _59[] = { T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::PhoneticExtensionsSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement,T::CombiningDiacriticalMarksSupplement, };
static const Block _64[] = { T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::GeneralPunctuation,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts, };
static const Block _65[] = { T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::SuperscriptsAndSubscripts,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CurrencySymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols,T::CombiningDiacriticalMarksForSymbols, };
static const Block _66[] = { T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::LetterlikeSymbols,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms, };
static const Block _67[] = { T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::NumberForms,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows,T::Arrows, };
static const Block _72[] = { T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::ControlPictures,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::OpticalCharacterRecognition,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics,T::EnclosedAlphanumerics, };
static const Block _75[] = { T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::BlockElements,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes,T::GeometricShapes, };
static const Block _79[] = { T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::Dingbats,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::MiscellaneousMathematicalSymbolsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA,T::SupplementalArrowsA, };
static const Block _88[] = { T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC,T::LatinExtendedC, };
static const Block _90[] = { T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::GeorgianSupplement,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh, };
static const Block _91[] = { T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::EthiopicExtended,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA,T::CyrillicExtendedA, };
static const Block _95[] = { T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,T::KangxiRadicals,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters,T::IdeographicDescriptionCharacters, };
static const Block _96[] = { T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::CJKSymbolsAndPunctuation,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana, };
static const Block _97[] = { T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana, };
static const Block _98[] = { T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo, };
static const Block _99[] = { T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::HangulCompatibilityJamo,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::Kanbun,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::BopomofoExtended,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::CJKStrokes,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions,T::KatakanaPhoneticExtensions, };
static const Block _155[] = { T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols,T::YijingHexagramSymbols, };
static const Block _329[] = { T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::YiRadicals,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu, };
static const Block _332[] = { T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB, };
static const Block _333[] = { T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::CyrillicExtendedB,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum, };
static const Block _334[] = { T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::ModifierToneLetters,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD,T::LatinExtendedD, };
static const Block _336[] = { T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::SylotiNagri,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::CommonIndicNumberForms,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa,T::PhagsPa, };
static const Block _337[] = { T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended,T::DevanagariExtended, };
static const Block _338[] = { T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::KayahLi,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA,T::HangulJamoExtendedA, };
static const Block _339[] = { T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB,T::MyanmarExtendedB, };
static const Block _340[] = { T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA,T::MyanmarExtendedA, };
static const Block _341[] = { T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::TaiViet,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions,T::MeeteiMayekExtensions, };
static const Block _342[] = { T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::EthiopicExtendedA,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::LatinExtendedE,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement, };
static const Block _343[] = { T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::CherokeeSupplement,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek,T::MeeteiMayek, };
static const Block _431[] = { T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB,T::HangulJamoExtendedB, };
static const Block _502[] = { T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::AlphabeticPresentationForms,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA, };
static const Block _508[] = { T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VariationSelectors,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::VerticalForms,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CombiningHalfMarks,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::CJKCompatibilityForms,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::SmallFormVariants,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB,T::ArabicPresentationFormsB, };
static const Block _511[] = { T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::HalfwidthAndFullwidthForms,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials,T::Specials, };
static const Block _514[] = { T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AegeanNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers, };
static const Block _515[] = { T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientGreekNumbers,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::AncientSymbols,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc,T::PhaistosDisc, };
static const Block _517[] = { T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers,T::CopticEpactNumbers, };
static const Block _518[] = { T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::OldItalic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic,T::OldPermic, };
static const Block _519[] = { T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,T::OldPersian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _520[] = { T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian, };
static const Block _521[] = { T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage, };
static const Block _522[] = { T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,T::CaucasianAlbanian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _528[] = { T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::CypriotSyllabary,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::ImperialAramaic,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene, };
static const Block _529[] = { T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran, };
static const Block _530[] = { T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _531[] = { T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticHieroglyphs,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive,T::MeroiticCursive, };
static const Block _532[] = { T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian,T::OldSouthArabian, };
static const Block _533[] = { T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,T::OldNorthArabian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean, };
static const Block _534[] = { T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalParthian,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi,T::InscriptionalPahlavi, };
static const Block _535[] = { T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,T::PsalterPahlavi,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _536[] = { T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,T::OldTurkic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _538[] = { T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,T::HanifiRohingya,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _540[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols,T::RumiNumeralSymbols, };
static const Block _541[] = { T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _542[] = { T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::OldSogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _543[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic, };
static const Block _545[] = { T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng,T::SoraSompeng, };
static const Block _546[] = { T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani, };
static const Block _547[] = { T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers,T::SinhalaArchaicNumbers, };
static const Block _548[] = { T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _549[] = { T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi, };
static const Block _553[] = { T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _556[] = { T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement,T::MongolianSupplement, };
static const Block _557[] = { T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _558[] = { T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _560[] = { T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _561[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti,T::WarangCiti, };
static const Block _562[] = { T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,T::DivesAkuru,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _563[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari, };
static const Block _564[] = { T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::ZanabazarSquare,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo, };
static const Block _565[] = { T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau,T::PauCinHau, };
static const Block _568[] = { T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen, };
static const Block _569[] = { T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _570[] = { T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::MasaramGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi, };
static const Block _571[] = { T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,T::GunjalaGondi,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _573[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar, };
static const Block _575[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::LisuSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement,T::TamilSupplement, };
static const Block _586[] = { T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,T::EarlyDynasticCuneiform,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _616[] = { T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,T::EgyptianHieroglyphFormatControls,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _724[] = { T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _725[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah,T::BassaVah, };
static const Block _727[] = { T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,T::PahawhHmong,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _732[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin, };
static const Block _733[] = { T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _735[] = { T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation,T::IdeographicSymbolsAndPunctuation, };
static const Block _795[] = { T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,T::TangutSupplement,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _866[] = { T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::KanaExtendedA,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::SmallKanaExtension,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu, };
static const Block _889[] = { T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,T::ShorthandFormatControls,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _932[] = { T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,T::AncientGreekMusicalNotation,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _933[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals,T::MayanNumerals, };
static const Block _934[] = { T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::TaiXuanJingSymbols,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals,T::CountingRodNumerals, };
static const Block _949[] = { T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _960[] = { T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,T::GlagoliticSupplement,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _962[] = { T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,T::NyiakengPuachueHmong,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _965[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho, };
static const Block _977[] = { T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,T::MendeKikakui,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _978[] = { T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _984[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers, };
static const Block _985[] = { T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,T::IndicSiyaqNumbers,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _986[] = { T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,T::OttomanSiyaqNumbers,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _992[] = { T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::MahjongTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles, };
static const Block _993[] = { T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::DominoTiles,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards,T::PlayingCards, };
static const Block _1004[] = { T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::Emoticons,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats,T::OrnamentalDingbats, };
static const Block _1012[] = { T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::ChessSymbols,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA,T::SymbolsAndPictographsExtendedA, };
static const Block _1357[] = { T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _1390[] = { T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD, };
static const Block _1392[] = { T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionD,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE, };
static const Block _1437[] = { T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF, };
static const Block _1495[] = { T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _1524[] = { T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _1574[] = { T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block _7171[] = { T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,T::VariationSelectorsSupplement,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Block *_blocks[] = {
 0,0,0,0,_4,_5,_6,0,
 0,0,_10,_11,0,0,_14,_15,
 _16,_17,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_33,0,0,0,0,0,_39,
 0,0,0,0,0,_45,_46,0,
 0,_49,_50,_51,_52,_53,0,_55,
 _56,_57,0,_59,0,0,0,0,
 _64,_65,_66,_67,0,0,0,0,
 _72,0,0,_75,0,0,0,_79,
 0,0,0,0,0,0,0,0,
 _88,0,_90,_91,0,0,0,_95,
 _96,_97,_98,_99,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_155,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_329,0,0,_332,_333,_334,0,
 _336,_337,_338,_339,_340,_341,_342,_343,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_431,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_502,0,
 0,0,0,0,_508,0,0,_511,
 0,0,_514,_515,0,_517,_518,_519,
 _520,_521,_522,0,0,0,0,0,
 _528,_529,_530,_531,_532,_533,_534,_535,
 _536,0,_538,0,_540,_541,_542,_543,
 0,_545,_546,_547,_548,_549,0,0,
 0,_553,0,0,_556,_557,_558,0,
 _560,_561,_562,_563,_564,_565,0,0,
 _568,_569,_570,_571,0,_573,0,_575,
 0,0,0,0,0,0,0,0,
 0,0,_586,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _616,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_724,_725,0,_727,
 0,0,0,0,_732,_733,0,_735,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_795,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_866,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_889,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_932,_933,_934,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_949,0,0,
 0,0,0,0,0,0,0,0,
 _960,0,_962,0,0,_965,0,0,
 0,0,0,0,0,0,0,0,
 0,_977,_978,0,0,0,0,0,
 _984,_985,_986,0,0,0,0,0,
 _992,_993,0,0,0,0,0,0,
 0,0,0,0,_1004,0,0,0,
 0,0,0,0,_1012,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1357,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1390,0,
 _1392,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1437,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_1495,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_1524,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1574,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_7171,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
static const Block _block_values[] = {
 T::BasicLatin,T::Latin1Supplement,T::LatinExtendedA,T::LatinExtendedB,D,D,D,T::GreekAndCoptic,
 T::Cyrillic,T::Cyrillic,D,D,T::Arabic,T::Arabic,D,D,
 D,D,T::Devanagari,T::Bengali,T::Gurmukhi,T::Gujarati,T::Oriya,T::Tamil,
 T::Telugu,T::Kannada,T::Malayalam,T::Sinhala,T::Thai,T::Lao,T::Tibetan,T::Tibetan,
 T::Myanmar,D,T::HangulJamo,T::HangulJamo,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,
 T::UnifiedCanadianAboriginalSyllabics,T::UnifiedCanadianAboriginalSyllabics,T::UnifiedCanadianAboriginalSyllabics,T::UnifiedCanadianAboriginalSyllabics,T::UnifiedCanadianAboriginalSyllabics,D,D,T::Khmer,
 T::Mongolian,D,D,D,D,D,T::Balinese,D,
 D,D,T::PhoneticExtensions,D,T::LatinExtendedAdditional,T::LatinExtendedAdditional,T::GreekExtended,T::GreekExtended,
 D,D,D,D,T::MathematicalOperators,T::MathematicalOperators,T::MiscellaneousTechnical,T::MiscellaneousTechnical,
 D,T::EnclosedAlphanumerics,T::BoxDrawing,D,T::MiscellaneousSymbols,T::MiscellaneousSymbols,T::Dingbats,D,
 T::BraillePatterns,T::BraillePatterns,T::SupplementalArrowsB,T::MiscellaneousMathematicalSymbolsB,T::SupplementalMathematicalOperators,T::SupplementalMathematicalOperators,T::MiscellaneousSymbolsAndArrows,T::MiscellaneousSymbolsAndArrows,
 D,T::Coptic,D,D,T::SupplementalPunctuation,T::CJKRadicalsSupplement,T::KangxiRadicals,D,
 D,D,D,D,T::EnclosedCJKLettersAndMonths,T::EnclosedCJKLettersAndMonths,T::CJKCompatibility,T::CJKCompatibility,
 T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,
 T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,
 T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,
 T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,
 T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,
 T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,
 T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,T::CJKUnifiedIdeographsExtensionA,D,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,T::CJKUnifiedIdeographs,
 T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,T::YiSyllables,
 T::YiSyllables,D,T::Vai,T::Vai,D,D,D,T::LatinExtendedD,
 D,D,D,D,D,D,D,D,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,
 T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,T::HangulSyllables,D,
 T::HighSurrogates,T::HighSurrogates,T::HighSurrogates,T::HighSurrogates,T::HighSurrogates,T::HighSurrogates,T::HighSurrogates,T::HighPrivateUseSurrogates,
 T::LowSurrogates,T::LowSurrogates,T::LowSurrogates,T::LowSurrogates,T::LowSurrogates,T::LowSurrogates,T::LowSurrogates,T::LowSurrogates,
 T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,
 T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,
 T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,
 T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,
 T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,
 T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,T::PrivateUseArea,
 T::PrivateUseArea,T::PrivateUseArea,T::CJKCompatibilityIdeographs,T::CJKCompatibilityIdeographs,T::CJKCompatibilityIdeographs,T::CJKCompatibilityIdeographs,D,T::ArabicPresentationFormsA,
 T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,T::ArabicPresentationFormsA,D,T::ArabicPresentationFormsB,T::HalfwidthAndFullwidthForms,D,
 T::LinearBSyllabary,T::LinearBIdeograms,D,D,D,D,D,D,
 D,D,D,D,T::LinearA,T::LinearA,T::LinearA,D,
 D,D,D,D,D,D,D,D,
 D,T::OldHungarian,D,D,D,D,D,D,
 T::Brahmi,D,D,D,D,D,T::Grantha,D,
 T::Newa,D,D,T::Siddham,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,
 T::CuneiformNumbersAndPunctuation,T::EarlyDynasticCuneiform,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,T::EgyptianHieroglyphs,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::AnatolianHieroglyphs,T::AnatolianHieroglyphs,T::AnatolianHieroglyphs,T::AnatolianHieroglyphs,T::AnatolianHieroglyphs,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,T::BamumSupplement,D,D,T::PahawhHmong,D,
 D,D,D,D,D,D,T::Miao,D,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::TangutComponents,T::TangutComponents,T::TangutComponents,T::TangutComponents,T::TangutComponents,T::TangutComponents,T::KhitanSmallScript,T::KhitanSmallScript,
 T::KhitanSmallScript,T::KhitanSmallScript,T::TangutSupplement,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::KanaSupplement,T::KanaSupplement,D,T::Nushu,T::Nushu,T::Nushu,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Duployan,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::ByzantineMusicalSymbols,T::ByzantineMusicalSymbols,T::MusicalSymbols,T::MusicalSymbols,D,D,D,D,
 T::MathematicalAlphanumericSymbols,T::MathematicalAlphanumericSymbols,T::MathematicalAlphanumericSymbols,T::MathematicalAlphanumericSymbols,T::MathematicalAlphanumericSymbols,T::MathematicalAlphanumericSymbols,T::MathematicalAlphanumericSymbols,T::MathematicalAlphanumericSymbols,
 T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,T::SuttonSignWriting,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::MendeKikakui,D,D,D,D,D,D,D,
 D,D,D,D,T::ArabicMathematicalAlphabeticSymbols,T::ArabicMathematicalAlphabeticSymbols,D,D,
 D,D,T::EnclosedAlphanumericSupplement,T::EnclosedAlphanumericSupplement,T::EnclosedIdeographicSupplement,T::EnclosedIdeographicSupplement,T::MiscellaneousSymbolsAndPictographs,T::MiscellaneousSymbolsAndPictographs,
 T::MiscellaneousSymbolsAndPictographs,T::MiscellaneousSymbolsAndPictographs,T::MiscellaneousSymbolsAndPictographs,T::MiscellaneousSymbolsAndPictographs,D,T::TransportAndMapSymbols,T::AlchemicalSymbols,T::GeometricShapesExtended,
 T::SupplementalArrowsC,T::SupplementalArrowsC,T::SupplementalSymbolsAndPictographs,T::SupplementalSymbolsAndPictographs,D,T::SymbolsAndPictographsExtendedA,T::SymbolsForLegacyComputing,T::SymbolsForLegacyComputing,
 D,D,D,D,D,D,D,D,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,
 T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,T::CJKUnifiedIdeographsExtensionB,D,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,
 T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,
 T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,
 T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,
 T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,T::CJKUnifiedIdeographsExtensionC,D,T::CJKUnifiedIdeographsExtensionD,
 D,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,
 T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,
 T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,
 T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,
 T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,
 T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,T::CJKUnifiedIdeographsExtensionE,D,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,
 T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,
 T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,
 T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,
 T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,
 T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,
 T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,
 T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,T::CJKUnifiedIdeographsExtensionF,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,T::CJKCompatibilityIdeographsSupplement,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,
 T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,
 T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,
 T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,
 T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,T::CJKUnifiedIdeographsExtensionG,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Tags,D,T::VariationSelectorsSupplement,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,T::SupplementaryPrivateUseAreaA,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
 T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,T::SupplementaryPrivateUseAreaB,
};
inline Block get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}
namespace _script_properties {
using T = Script;
const auto D = Script::Unassigned;
static const size_t _block_size = 128;
static const Script _0[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _1[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin, };
static const Script _5[] = { T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common,T::Bopomofo,T::Bopomofo,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _6[] = { T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Greek,T::Greek,T::Greek,T::Greek,T::Common,T::Greek,T::Greek,T::Greek,D,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Common,T::Greek, };
static const Script _7[] = { D,D,D,D,T::Greek,T::Common,T::Greek,T::Common,T::Greek,T::Greek,T::Greek,D,T::Greek,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek, };
static const Script _9[] = { T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Inherited,T::Inherited,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic, };
static const Script _10[] = { T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,D,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,D,D,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian, };
static const Script _11[] = { T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,D,D,T::Armenian,T::Armenian,T::Armenian,D,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,D,D,D,D,D,D,D,D,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,D,D,D,D,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _12[] = { T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Common,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Common,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Common,T::Arabic,D,T::Arabic,T::Common,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Common,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Inherited,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic, };
static const Script _13[] = { T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Common,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic, };
static const Script _14[] = { T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,D,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,D,D,T::Syriac,T::Syriac,T::Syriac,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic, };
static const Script _15[] = { T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,T::Thaana,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,T::Nko,D,D,T::Nko,T::Nko,T::Nko, };
static const Script _16[] = { T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,D,D,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,T::Samaritan,D,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,T::Mandaic,D,D,T::Mandaic,D,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,T::Syriac,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _17[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,D,D,D,D,D,D,D,D,D,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Common,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic, };
static const Script _18[] = { T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Common,T::Common,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari, };
static const Script _19[] = { T::Bengali,T::Bengali,T::Bengali,T::Bengali,D,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,D,D,T::Bengali,T::Bengali,D,D,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,D,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,D,T::Bengali,D,D,D,T::Bengali,T::Bengali,T::Bengali,T::Bengali,D,D,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,D,D,T::Bengali,T::Bengali,D,D,T::Bengali,T::Bengali,T::Bengali,T::Bengali,D,D,D,D,D,D,D,D,T::Bengali,D,D,D,D,T::Bengali,T::Bengali,D,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,D,D,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,T::Bengali,D, };
static const Script _20[] = { D,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,D,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,D,D,D,D,T::Gurmukhi,T::Gurmukhi,D,D,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,D,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,D,T::Gurmukhi,T::Gurmukhi,D,T::Gurmukhi,T::Gurmukhi,D,T::Gurmukhi,T::Gurmukhi,D,D,T::Gurmukhi,D,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,D,D,D,D,T::Gurmukhi,T::Gurmukhi,D,D,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,D,D,D,T::Gurmukhi,D,D,D,D,D,D,D,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,D,T::Gurmukhi,D,D,D,D,D,D,D,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,T::Gurmukhi,D,D,D,D,D,D,D,D,D, };
static const Script _21[] = { D,T::Gujarati,T::Gujarati,T::Gujarati,D,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,D,T::Gujarati,T::Gujarati,T::Gujarati,D,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,D,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,D,T::Gujarati,T::Gujarati,D,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,D,D,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,D,T::Gujarati,T::Gujarati,T::Gujarati,D,T::Gujarati,T::Gujarati,T::Gujarati,D,D,T::Gujarati,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,D,D,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,D,D,D,D,D,D,D,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati,T::Gujarati, };
static const Script _22[] = { D,T::Oriya,T::Oriya,T::Oriya,D,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,D,D,T::Oriya,T::Oriya,D,D,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,D,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,D,T::Oriya,T::Oriya,D,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,D,D,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,D,D,T::Oriya,T::Oriya,D,D,T::Oriya,T::Oriya,T::Oriya,D,D,D,D,D,D,D,T::Oriya,T::Oriya,T::Oriya,D,D,D,D,T::Oriya,T::Oriya,D,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,D,D,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,T::Oriya,D,D,D,D,D,D,D,D, };
static const Script _23[] = { D,D,T::Tamil,T::Tamil,D,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,D,D,D,T::Tamil,T::Tamil,T::Tamil,D,T::Tamil,T::Tamil,T::Tamil,T::Tamil,D,D,D,T::Tamil,T::Tamil,D,T::Tamil,D,T::Tamil,T::Tamil,D,D,D,T::Tamil,T::Tamil,D,D,D,T::Tamil,T::Tamil,T::Tamil,D,D,D,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,D,D,D,D,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,D,D,D,T::Tamil,T::Tamil,T::Tamil,D,T::Tamil,T::Tamil,T::Tamil,T::Tamil,D,D,T::Tamil,D,D,D,D,D,D,T::Tamil,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,D,D,D,D,D, };
static const Script _24[] = { T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,D,T::Telugu,T::Telugu,T::Telugu,D,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,D,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,D,D,D,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,D,T::Telugu,T::Telugu,T::Telugu,D,T::Telugu,T::Telugu,T::Telugu,T::Telugu,D,D,D,D,D,D,D,T::Telugu,T::Telugu,D,T::Telugu,T::Telugu,T::Telugu,D,D,D,D,D,T::Telugu,T::Telugu,T::Telugu,T::Telugu,D,D,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,D,D,D,D,D,D,D,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu,T::Telugu, };
static const Script _25[] = { T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,D,T::Kannada,T::Kannada,T::Kannada,D,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,D,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,D,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,D,D,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,D,T::Kannada,T::Kannada,T::Kannada,D,T::Kannada,T::Kannada,T::Kannada,T::Kannada,D,D,D,D,D,D,D,T::Kannada,T::Kannada,D,D,D,D,D,D,D,T::Kannada,D,T::Kannada,T::Kannada,T::Kannada,T::Kannada,D,D,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,T::Kannada,D,T::Kannada,T::Kannada,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _26[] = { T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,D,T::Malayalam,T::Malayalam,T::Malayalam,D,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,D,T::Malayalam,T::Malayalam,T::Malayalam,D,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,D,D,D,D,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,D,D,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam,T::Malayalam, };
static const Script _27[] = { D,T::Sinhala,T::Sinhala,T::Sinhala,D,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,D,D,D,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,D,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,D,T::Sinhala,D,D,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,D,D,D,T::Sinhala,D,D,D,D,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,D,T::Sinhala,D,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,D,D,D,D,D,D,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,D,D,T::Sinhala,T::Sinhala,T::Sinhala,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _28[] = { D,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,D,D,D,D,T::Common,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,T::Thai,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _29[] = { D,T::Lao,T::Lao,D,T::Lao,D,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,D,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,D,T::Lao,D,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,D,D,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,D,T::Lao,D,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,D,D,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,T::Lao,D,D,T::Lao,T::Lao,T::Lao,T::Lao,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _30[] = { T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,D,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,D,D,D,D,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan, };
static const Script _31[] = { T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,D,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,D,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,D,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Tibetan,T::Common,T::Common,T::Common,T::Common,T::Tibetan,T::Tibetan,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _33[] = { T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,D,T::Georgian,D,D,D,D,D,T::Georgian,D,D,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Common,T::Georgian,T::Georgian,T::Georgian,T::Georgian, };
static const Script _36[] = { T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic, };
static const Script _37[] = { T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic, };
static const Script _38[] = { T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,D, };
static const Script _39[] = { T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,D,D,D,D,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,D,D,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,D,D, };
static const Script _45[] = { T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,T::Ogham,D,D,D,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Common,T::Common,T::Common,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,T::Runic,D,D,D,D,D,D,D, };
static const Script _46[] = { T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,D,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,T::Tagalog,D,D,D,D,D,D,D,D,D,D,D,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Hanunoo,T::Common,T::Common,D,D,D,D,D,D,D,D,D,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,T::Buhid,D,D,D,D,D,D,D,D,D,D,D,D,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,D,T::Tagbanwa,T::Tagbanwa,T::Tagbanwa,D,T::Tagbanwa,T::Tagbanwa,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _47[] = { T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,D,D,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,D,D,D,D,D,D,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,D,D,D,D,D,D, };
static const Script _48[] = { T::Mongolian,T::Mongolian,T::Common,T::Common,T::Mongolian,T::Common,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,D,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,D,D,D,D,D,D,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,D,D,D,D,D,D,D, };
static const Script _49[] = { T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,D,D,D,D,D,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,D,D,D,D,D,D,D,D,D,D, };
static const Script _50[] = { T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,D,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,D,D,D,D,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,D,D,D,D,T::Limbu,D,D,D,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Limbu,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,D,D,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,T::Tai_Le,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _51[] = { T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,D,D,D,D,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,D,D,D,D,D,D,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,T::New_Tai_Lue,D,D,D,T::New_Tai_Lue,T::New_Tai_Lue,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer,T::Khmer, };
static const Script _52[] = { T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,T::Buginese,D,D,T::Buginese,T::Buginese,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,D,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,D,D,T::Tai_Tham, };
static const Script _53[] = { T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,D,D,D,D,D,D,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,D,D,D,D,D,D,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,T::Tai_Tham,D,D,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _54[] = { T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,D,D,D,D,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,T::Balinese,D,D,D, };
static const Script _55[] = { T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,T::Batak,D,D,D,D,D,D,D,D,T::Batak,T::Batak,T::Batak,T::Batak, };
static const Script _56[] = { T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,D,D,D,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,T::Lepcha,D,D,D,T::Lepcha,T::Lepcha,T::Lepcha,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki,T::Ol_Chiki, };
static const Script _57[] = { T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,D,D,D,D,D,D,D,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,D,D,T::Georgian,T::Georgian,T::Georgian,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,T::Sundanese,D,D,D,D,D,D,D,D,T::Inherited,T::Inherited,T::Inherited,T::Common,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Common,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Common,T::Common,T::Common,T::Common,T::Inherited,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Inherited,T::Common,T::Common,T::Common,T::Inherited,T::Inherited,T::Common,D,D,D,D,D, };
static const Script _58[] = { T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Cyrillic,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Latin,T::Latin,T::Latin,T::Latin,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Cyrillic,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin, };
static const Script _59[] = { T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Greek,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,D,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited, };
static const Script _62[] = { T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,T::Greek,D,T::Greek,D,T::Greek,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,D, };
static const Script _63[] = { T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,D,T::Greek,T::Greek,T::Greek,D,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D, };
static const Script _64[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Inherited,T::Inherited,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin, };
static const Script _65[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _66[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Greek,T::Common,T::Common,T::Common,T::Latin,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin, };
static const Script _67[] = { T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Common,T::Common,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _72[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _86[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _87[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _88[] = { T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,D,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,D,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin, };
static const Script _89[] = { T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,D,D,D,D,D,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic,T::Coptic, };
static const Script _90[] = { T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,T::Georgian,D,T::Georgian,D,D,D,D,D,T::Georgian,D,D,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,T::Tifinagh,D,D,D,D,D,D,D,T::Tifinagh,T::Tifinagh,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Tifinagh, };
static const Script _91[] = { T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,D,D,D,D,D,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic, };
static const Script _92[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _93[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _95[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D, };
static const Script _96[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Han,T::Common,T::Han,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Hangul,T::Hangul,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Han,T::Han,T::Han,T::Han,T::Common,T::Common,T::Common,T::Common,D,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana, };
static const Script _97[] = { T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,D,D,T::Inherited,T::Inherited,T::Common,T::Common,T::Hiragana,T::Hiragana,T::Hiragana,T::Common,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Common,T::Common,T::Katakana,T::Katakana,T::Katakana, };
static const Script _98[] = { D,D,D,D,D,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,D,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul, };
static const Script _99[] = { T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Bopomofo,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana, };
static const Script _100[] = { T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Common, };
static const Script _101[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Common, };
static const Script _102[] = { T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _155[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _319[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,D, };
static const Script _329[] = { T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,D,D,D,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,D,D,D,D,D,D,D,D,D,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu,T::Lisu, };
static const Script _332[] = { T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,T::Vai,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic, };
static const Script _333[] = { T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Cyrillic,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,D,D,D,D,D,D,D,D, };
static const Script _334[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin, };
static const Script _335[] = { T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Common,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,D,D,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin, };
static const Script _336[] = { T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,T::Syloti_Nagri,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,T::Phags_Pa,D,D,D,D,D,D,D,D, };
static const Script _337[] = { T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,D,D,D,D,D,D,D,D,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,T::Saurashtra,D,D,D,D,D,D,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari,T::Devanagari, };
static const Script _338[] = { T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Kayah_Li,T::Common,T::Kayah_Li,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,T::Rejang,D,D,D,D,D,D,D,D,D,D,D,T::Rejang,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,D,D, };
static const Script _339[] = { T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,D,T::Common,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,T::Javanese,D,D,D,D,T::Javanese,T::Javanese,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,D, };
static const Script _340[] = { T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,D,D,D,D,D,D,D,D,D,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,D,D,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,T::Cham,D,D,T::Cham,T::Cham,T::Cham,T::Cham,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar,T::Myanmar, };
static const Script _341[] = { T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Tai_Viet,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,D,D,D,D,D,D,D,D,D, };
static const Script _342[] = { D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,D,D,D,D,D,D,D,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,T::Ethiopic,D,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Greek,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Common,D,D,D,D,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee, };
static const Script _343[] = { T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Cherokee,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,D,D,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,T::Meetei_Mayek,D,D,D,D,D,D, };
static const Script _431[] = { T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,D,D,D,D,D,D,D,D,D,D,D,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,D,D,D,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,D,D,D, };
static const Script _500[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han, };
static const Script _501[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _502[] = { T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,D,D,D,D,D,D,D,D,D,D,D,D,T::Armenian,T::Armenian,T::Armenian,T::Armenian,T::Armenian,D,D,D,D,D,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,D,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,D,T::Hebrew,D,T::Hebrew,T::Hebrew,D,T::Hebrew,T::Hebrew,D,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Hebrew,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic, };
static const Script _503[] = { T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic, };
static const Script _506[] = { T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic, };
static const Script _507[] = { T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,D, };
static const Script _508[] = { T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Cyrillic,T::Cyrillic,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,D,D,D,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic, };
static const Script _509[] = { T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,D,T::Common, };
static const Script _510[] = { D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Latin,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Common,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana, };
static const Script _511[] = { T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Common,T::Common,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,D,D,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,D,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,D,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,D,T::Hangul,T::Hangul,T::Hangul,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,D,D, };
static const Script _512[] = { T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,D,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,D,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,D,T::Linear_B,T::Linear_B,D,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,D,D,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _513[] = { T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,T::Linear_B,D,D,D,D,D, };
static const Script _514[] = { T::Common,T::Common,T::Common,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek, };
static const Script _515[] = { T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,T::Greek,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Inherited,D,D, };
static const Script _517[] = { T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,T::Lycian,D,D,D,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,T::Carian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Inherited,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D, };
static const Script _518[] = { T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Old_Italic,D,D,D,D,D,D,D,D,D,T::Old_Italic,T::Old_Italic,T::Old_Italic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,T::Gothic,D,D,D,D,D,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,T::Old_Permic,D,D,D,D,D, };
static const Script _519[] = { T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,T::Ugaritic,D,T::Ugaritic,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,D,D,D,D,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,T::Old_Persian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _520[] = { T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Deseret,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian,T::Shavian, };
static const Script _521[] = { T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,D,D,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,T::Osmanya,D,D,D,D,D,D,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,D,D,D,D,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,T::Osage,D,D,D,D, };
static const Script _522[] = { T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,T::Elbasan,D,D,D,D,D,D,D,D,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,T::Caucasian_Albanian,D,D,D,D,D,D,D,D,D,D,D,T::Caucasian_Albanian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _526[] = { T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,D,D,D,D,D,D,D,D,D,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,D,D,D,D,D,D,D,D,D,D,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,T::Linear_A,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _528[] = { T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,D,D,T::Cypriot,D,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,T::Cypriot,D,T::Cypriot,T::Cypriot,D,D,D,T::Cypriot,D,D,T::Cypriot,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,D,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Imperial_Aramaic,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene,T::Palmyrene, };
static const Script _529[] = { T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,D,D,D,D,D,D,D,D,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,T::Nabataean,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran,D,T::Hatran,T::Hatran,D,D,D,D,D,T::Hatran,T::Hatran,T::Hatran,T::Hatran,T::Hatran, };
static const Script _530[] = { T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,T::Phoenician,D,D,D,T::Phoenician,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,T::Lydian,D,D,D,D,D,T::Lydian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _531[] = { T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Hieroglyphs,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,D,D,D,D,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,D,D,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive,T::Meroitic_Cursive, };
static const Script _532[] = { T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,D,T::Kharoshthi,T::Kharoshthi,D,D,D,D,D,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,D,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,D,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,D,D,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,D,D,D,D,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,D,D,D,D,D,D,D,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,T::Kharoshthi,D,D,D,D,D,D,D,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian,T::Old_South_Arabian, };
static const Script _533[] = { T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,T::Old_North_Arabian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,D,D,D,D,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,T::Manichaean,D,D,D,D,D,D,D,D,D, };
static const Script _534[] = { T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,D,D,D,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Avestan,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,D,D,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Parthian,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,D,D,D,D,D,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi,T::Inscriptional_Pahlavi, };
static const Script _535[] = { T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,D,D,D,D,D,D,D,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,D,D,D,D,D,D,D,D,D,D,D,D,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,T::Psalter_Pahlavi,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _536[] = { T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,T::Old_Turkic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _537[] = { T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,D,D,D,D,D,D,D,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian,T::Old_Hungarian, };
static const Script _538[] = { T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,D,D,D,D,D,D,D,D,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,T::Hanifi_Rohingya,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _540[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D, };
static const Script _541[] = { T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,T::Yezidi,D,T::Yezidi,T::Yezidi,T::Yezidi,D,D,T::Yezidi,T::Yezidi,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _542[] = { T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,T::Old_Sogdian,D,D,D,D,D,D,D,D,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,T::Sogdian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _543[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,T::Chorasmian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,T::Elymaic,D,D,D,D,D,D,D,D,D, };
static const Script _544[] = { T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,D,D,D,D,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,T::Brahmi,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Brahmi, };
static const Script _545[] = { T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,T::Kaithi,D,D,D,D,D,D,D,D,D,D,D,T::Kaithi,D,D,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,D,D,D,D,D,D,D,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,T::Sora_Sompeng,D,D,D,D,D,D, };
static const Script _546[] = { T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,D,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,T::Chakma,D,D,D,D,D,D,D,D,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,T::Mahajani,D,D,D,D,D,D,D,D,D, };
static const Script _547[] = { T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,T::Sharada,D,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,T::Sinhala,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _548[] = { T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,D,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,T::Khojki,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _549[] = { T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,D,T::Multani,D,T::Multani,T::Multani,T::Multani,T::Multani,D,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,D,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,T::Multani,D,D,D,D,D,D,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,D,D,D,D,D,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,T::Khudawadi,D,D,D,D,D,D, };
static const Script _550[] = { T::Grantha,T::Grantha,T::Grantha,T::Grantha,D,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,D,D,T::Grantha,T::Grantha,D,D,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,D,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,D,T::Grantha,T::Grantha,D,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,D,T::Inherited,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,D,D,T::Grantha,T::Grantha,D,D,T::Grantha,T::Grantha,T::Grantha,D,D,T::Grantha,D,D,D,D,D,D,T::Grantha,D,D,D,D,D,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,D,D,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,D,D,D,T::Grantha,T::Grantha,T::Grantha,T::Grantha,T::Grantha,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _552[] = { T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,D,T::Newa,T::Newa,T::Newa,T::Newa,T::Newa,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _553[] = { T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,D,D,D,D,D,D,D,D,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,T::Tirhuta,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _555[] = { T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,D,D,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,T::Siddham,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _556[] = { T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,D,D,D,D,D,D,D,D,D,D,D,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,T::Modi,D,D,D,D,D,D,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,T::Mongolian,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _557[] = { T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,D,D,D,D,D,D,D,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,T::Takri,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _558[] = { T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,D,D,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,D,D,D,D,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,T::Ahom,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _560[] = { T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,T::Dogra,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _561[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,T::Warang_Citi,D,D,D,D,D,D,D,D,D,D,D,D,T::Warang_Citi, };
static const Script _562[] = { T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,D,D,T::Dives_Akuru,D,D,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,D,T::Dives_Akuru,T::Dives_Akuru,D,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,D,T::Dives_Akuru,T::Dives_Akuru,D,D,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,D,D,D,D,D,D,D,D,D,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,T::Dives_Akuru,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _563[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,D,D,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,D,D,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,T::Nandinagari,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _564[] = { T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,T::Zanabazar_Square,D,D,D,D,D,D,D,D,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo, };
static const Script _565[] = { T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,T::Soyombo,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,T::Pau_Cin_Hau,D,D,D,D,D,D,D, };
static const Script _568[] = { T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,D,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,D,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,D,D,D,D,D,D,D,D,D,D,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,T::Bhaiksuki,D,D,D,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen, };
static const Script _569[] = { T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,D,D,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,D,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,T::Marchen,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _570[] = { T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,D,T::Masaram_Gondi,T::Masaram_Gondi,D,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,D,D,D,T::Masaram_Gondi,D,T::Masaram_Gondi,T::Masaram_Gondi,D,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,D,D,D,D,D,D,D,D,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,T::Masaram_Gondi,D,D,D,D,D,D,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,D,T::Gunjala_Gondi,T::Gunjala_Gondi,D,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi, };
static const Script _571[] = { T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,D,T::Gunjala_Gondi,T::Gunjala_Gondi,D,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,D,D,D,D,D,D,D,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,T::Gunjala_Gondi,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _573[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,T::Makasar,D,D,D,D,D,D,D, };
static const Script _575[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lisu,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,T::Tamil,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Tamil, };
static const Script _583[] = { T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _584[] = { T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,D,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _586[] = { T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _616[] = { T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,D,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _652[] = { T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _724[] = { T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,T::Bamum,D,D,D,D,D,D,D,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,D,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,T::Mro,D,D,D,D,T::Mro,T::Mro,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _725[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,D,D,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,T::Bassa_Vah,D,D,D,D,D,D,D,D,D,D, };
static const Script _726[] = { T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,D,D,D,D,D,D,D,D,D,D,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,D,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,D,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,D,D,D,D,D,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong, };
static const Script _727[] = { T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,T::Pahawh_Hmong,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _732[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin, };
static const Script _733[] = { T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,T::Medefaidrin,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _734[] = { T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,D,D,D,D,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao, };
static const Script _735[] = { T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,D,D,D,D,D,D,D,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,T::Miao,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Tangut,T::Nushu,T::Common,T::Common,T::Khitan_Small_Script,D,D,D,D,D,D,D,D,D,D,D,T::Han,T::Han,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _783[] = { T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,D,D,D,D,D,D,D,D, };
static const Script _793[] = { T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,T::Khitan_Small_Script,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _794[] = { T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _864[] = { T::Katakana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana, };
static const Script _866[] = { T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,T::Hiragana,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Hiragana,T::Hiragana,T::Hiragana,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Katakana,T::Katakana,T::Katakana,T::Katakana,D,D,D,D,D,D,D,D,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu, };
static const Script _869[] = { T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,T::Nushu,D,D,D,D, };
static const Script _888[] = { T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,D,D,D,D,D,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,D,D,D, };
static const Script _889[] = { T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,D,D,D,D,D,D,D,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Duployan,D,D,T::Duployan,T::Duployan,T::Duployan,T::Duployan,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _929[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D, };
static const Script _930[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Inherited,T::Inherited,T::Inherited,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited, };
static const Script _931[] = { T::Inherited,T::Inherited,T::Inherited,T::Common,T::Common,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _932[] = { T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,T::Greek,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _933[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _934[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D, };
static const Script _936[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _937[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,D,D,T::Common,D,D,T::Common,T::Common,D,D,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _938[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _941[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _943[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _949[] = { T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,D,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _960[] = { T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,D,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,D,D,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,D,T::Glagolitic,T::Glagolitic,D,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,T::Glagolitic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _962[] = { T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,D,D,D,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,D,D,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,D,D,D,D,T::Nyiakeng_Puachue_Hmong,T::Nyiakeng_Puachue_Hmong,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _965[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,T::Wancho,D,D,D,D,D,T::Wancho, };
static const Script _977[] = { T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,D,D,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,T::Mende_Kikakui,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _978[] = { T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,D,D,D,D,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,T::Adlam,D,D,D,D,T::Adlam,T::Adlam,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _984[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _985[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _986[] = { D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _988[] = { T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,D,T::Arabic,D,D,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,D,T::Arabic,D,D,D,D,D,D,T::Arabic,D,D,D,D,T::Arabic,D,T::Arabic,D,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,D,T::Arabic,D,D,T::Arabic,D,T::Arabic,D,T::Arabic,D,T::Arabic,D,T::Arabic,D,T::Arabic,T::Arabic,D,T::Arabic,D,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,D, };
static const Script _989[] = { T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,D,D,D,D,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,T::Arabic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Arabic,T::Arabic,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _992[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _993[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D, };
static const Script _995[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _996[] = { T::Hiragana,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _1005[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D, };
static const Script _1006[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _1007[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _1008[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _1009[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _1010[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _1011[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _1012[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,T::Common,T::Common,T::Common,D,D,D,D,D, };
static const Script _1013[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _1015[] = { T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,D,D,D, };
static const Script _1357[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _1390[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,D,D,D,D,D,D,D,D,D,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han, };
static const Script _1392[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han, };
static const Script _1437[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han, };
static const Script _1495[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _1524[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _1574[] = { T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script _7168[] = { D,T::Common,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common, };
static const Script _7171[] = { T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,T::Inherited,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Script *_blocks[] = {
 _0,_1,0,0,0,_5,_6,_7,
 0,_9,_10,_11,_12,_13,_14,_15,
 _16,_17,_18,_19,_20,_21,_22,_23,
 _24,_25,_26,_27,_28,_29,_30,_31,
 0,_33,0,0,_36,_37,_38,_39,
 0,0,0,0,0,_45,_46,_47,
 _48,_49,_50,_51,_52,_53,_54,_55,
 _56,_57,_58,_59,0,0,_62,_63,
 _64,_65,_66,_67,0,0,0,0,
 _72,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_86,_87,
 _88,_89,_90,_91,_92,_93,0,_95,
 _96,_97,_98,_99,_100,_101,_102,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_155,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_319,
 0,0,0,0,0,0,0,0,
 0,_329,0,0,_332,_333,_334,_335,
 _336,_337,_338,_339,_340,_341,_342,_343,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_431,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_500,_501,_502,_503,
 0,0,_506,_507,_508,_509,_510,_511,
 _512,_513,_514,_515,0,_517,_518,_519,
 _520,_521,_522,0,0,0,_526,0,
 _528,_529,_530,_531,_532,_533,_534,_535,
 _536,_537,_538,0,_540,_541,_542,_543,
 _544,_545,_546,_547,_548,_549,_550,0,
 _552,_553,0,_555,_556,_557,_558,0,
 _560,_561,_562,_563,_564,_565,0,0,
 _568,_569,_570,_571,0,_573,0,_575,
 0,0,0,0,0,0,0,_583,
 _584,0,_586,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _616,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_652,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_724,_725,_726,_727,
 0,0,0,0,_732,_733,_734,_735,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_783,
 0,0,0,0,0,0,0,0,
 0,_793,_794,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _864,0,_866,0,0,_869,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _888,_889,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_929,_930,_931,_932,_933,_934,0,
 _936,_937,_938,0,0,_941,0,_943,
 0,0,0,0,0,_949,0,0,
 0,0,0,0,0,0,0,0,
 _960,0,_962,0,0,_965,0,0,
 0,0,0,0,0,0,0,0,
 0,_977,_978,0,0,0,0,0,
 _984,_985,_986,0,_988,_989,0,0,
 _992,_993,0,_995,_996,0,0,0,
 0,0,0,0,0,_1005,_1006,_1007,
 _1008,_1009,_1010,_1011,_1012,_1013,0,_1015,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1357,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1390,0,
 _1392,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1437,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_1495,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_1524,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1574,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _7168,0,0,_7171,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
static const Script _block_values[] = {
 D,D,T::Latin,T::Latin,T::Latin,D,D,D,
 T::Cyrillic,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Myanmar,D,T::Hangul,T::Hangul,D,D,D,D,
 T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,T::Canadian_Aboriginal,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,T::Latin,T::Latin,D,D,
 D,D,D,D,T::Common,T::Common,T::Common,T::Common,
 D,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,T::Common,
 T::Braille,T::Braille,T::Common,T::Common,T::Common,T::Common,D,D,
 D,D,D,D,D,D,T::Han,D,
 D,D,D,D,D,D,D,T::Common,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,D,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,
 T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,T::Yi,
 T::Yi,D,T::Vai,T::Vai,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,
 T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,T::Hangul,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::Han,T::Han,D,D,D,D,
 T::Arabic,T::Arabic,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,T::Linear_A,T::Linear_A,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,T::Cuneiform,D,
 D,T::Cuneiform,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,T::Egyptian_Hieroglyphs,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,T::Anatolian_Hieroglyphs,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Bamum,T::Bamum,T::Bamum,T::Bamum,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,D,
 T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Tangut,T::Khitan_Small_Script,T::Khitan_Small_Script,
 T::Khitan_Small_Script,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,T::Hiragana,D,T::Nushu,T::Nushu,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Common,D,D,D,D,D,D,D,
 D,D,D,T::Common,T::Common,D,T::Common,D,
 T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,T::SignWriting,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Mende_Kikakui,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::Common,D,D,D,T::Common,T::Common,
 T::Common,T::Common,T::Common,T::Common,T::Common,D,D,D,
 D,D,D,D,D,D,T::Common,D,
 D,D,D,D,D,D,D,D,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,D,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,T::Han,
 D,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,D,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Han,T::Han,T::Han,T::Han,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,
 T::Han,T::Han,T::Han,T::Han,T::Han,T::Han,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::Inherited,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
};
inline Script get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}
namespace _script_extension_ids {
static const size_t _block_size = 256;
static const int _3[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,3,-1,-1,3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,5,5,5,5,5,5,5,5,5,5,5,5,5,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _4[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,17,15,16,16,15,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _6[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,44,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,44,32,-1,-1,44,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,48,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,9,9,9,9,9,9,9,9,9,9,9,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,33,33,33,33,33,33,33,33,33,33,-1,-1,-1,-1,-1,-1,9,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,8,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _9[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,52,51,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,55,56,43,43,43,43,43,43,43,43,43,43,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,34,34,34,34,34,34,34,34,34,34,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _10[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,26,26,26,26,26,26,26,26,26,26,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,25,25,25,25,25,25,25,25,25,25,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _11[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,24,24,24,24,24,24,24,24,24,24,24,24,24,24,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _12[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,29,29,29,29,29,29,29,29,29,29,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _16[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,35,35,35,35,35,35,35,35,35,35,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,23,-1,-1,-1,-1, };
static const int _23[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,42,42,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _24[] = { -1,-1,31,31,-1,31,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _28[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,41,1,41,19,1,11,11,21,11,21,47,1,21,21,1,1,21,11,1,1,1,1,1,1,1,20,11,1,1,11,1,1,1,1,49,19,37,11,11,0,19,19,6,-1,-1,-1,-1,-1, };
static const int _29[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,3,3,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,18,-1,-1,-1,-1,-1,-1,-1, };
static const int _32[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,30,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,38,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _46[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,15,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _48[] = { -1,46,46,45,-1,-1,4,-1,46,46,46,46,46,46,46,46,46,46,-1,45,46,46,46,46,46,46,46,46,45,45,45,45,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,12,12,12,12,-1,-1,45,28,28,28,28,28,-1,45,-1,-1,-1,-1,39,39,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,28,28,28,28,-1,-1,-1,28,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,46,28,-1,-1,-1, };
static const int _49[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _50[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,4,4,4,4,4,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4, };
static const int _51[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,-1, };
static const int _166[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,15,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _167[] = { 27,27,27,27,27,27,27,27,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _168[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,54,54,54,53,53,53,50,50,50,50,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,11,-1,22,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _169[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,40,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,13,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _253[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,10,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,10,-1,-1, };
static const int _254[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,45,45,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _255[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,46,46,46,46,46,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,28,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,28,28,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _257[] = { 14,14,14,-1,-1,-1,-1,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,36,-1,-1,-1,14,14,14,14,14,14,14,14,14,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _258[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,-1,-1,-1,-1, };
static const int _275[] = { -1,24,-1,24,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,24,24,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _287[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,24,24,-1,24,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _444[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,2,2,2,2,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _467[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int _498[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,4,4,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1, };
static const int *_blocks[] = {
 0,0,0,_3,_4,0,_6,0,
 0,_9,_10,_11,_12,0,0,0,
 _16,0,0,0,0,0,0,_23,
 _24,0,0,0,_28,_29,0,0,
 _32,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_46,0,
 _48,_49,_50,_51,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_166,_167,
 _168,_169,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_253,_254,_255,
 0,_257,_258,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_275,0,0,0,0,
 0,0,0,0,0,0,0,_287,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_444,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_467,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_498,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
static const int _block_values[] = {
 -1,-1,-1,0,0,-1,0,-1,
 -1,0,0,0,0,-1,-1,-1,
 0,-1,-1,-1,-1,-1,-1,0,
 0,-1,-1,-1,0,0,-1,-1,
 0,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,0,-1,
 0,0,0,0,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,0,0,
 0,0,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,0,0,0,
 -1,0,0,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,0,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,0,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,0,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,0,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,0,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
 -1,-1,-1,-1,-1,-1,-1,-1,
};
inline int get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}
const std::vector<std::vector<Script>> _script_extension_properties_for_id = {
{
    Script::Bengali, 
},
{
    Script::Devanagari, 
},
{
    Script::Duployan, 
},
{
    Script::Greek, 
},
{
    Script::Han, 
},
{
    Script::Latin, 
},
{
    Script::Nandinagari, 
},
{
    Script::Arabic, 
    Script::Coptic, 
},
{
    Script::Arabic, 
    Script::Hanifi_Rohingya, 
},
{
    Script::Arabic, 
    Script::Syriac, 
},
{
    Script::Arabic, 
    Script::Thaana, 
},
{
    Script::Bengali, 
    Script::Devanagari, 
},
{
    Script::Bopomofo, 
    Script::Han, 
},
{
    Script::Buginese, 
    Script::Javanese, 
},
{
    Script::Cypriot, 
    Script::Linear_B, 
},
{
    Script::Cyrillic, 
    Script::Glagolitic, 
},
{
    Script::Cyrillic, 
    Script::Latin, 
},
{
    Script::Cyrillic, 
    Script::Old_Permic, 
},
{
    Script::Cyrillic, 
    Script::Syriac, 
},
{
    Script::Devanagari, 
    Script::Grantha, 
},
{
    Script::Devanagari, 
    Script::Nandinagari, 
},
{
    Script::Devanagari, 
    Script::Sharada, 
},
{
    Script::Devanagari, 
    Script::Tamil, 
},
{
    Script::Georgian, 
    Script::Latin, 
},
{
    Script::Grantha, 
    Script::Tamil, 
},
{
    Script::Gujarati, 
    Script::Khojki, 
},
{
    Script::Gurmukhi, 
    Script::Multani, 
},
{
    Script::Han, 
    Script::Latin, 
},
{
    Script::Hiragana, 
    Script::Katakana, 
},
{
    Script::Kannada, 
    Script::Nandinagari, 
},
{
    Script::Latin, 
    Script::Mongolian, 
},
{
    Script::Mongolian, 
    Script::Phags_Pa, 
},
{
    Script::Arabic, 
    Script::Syriac, 
    Script::Thaana, 
},
{
    Script::Arabic, 
    Script::Thaana, 
    Script::Yezidi, 
},
{
    Script::Bengali, 
    Script::Chakma, 
    Script::Syloti_Nagri, 
},
{
    Script::Chakma, 
    Script::Myanmar, 
    Script::Tai_Le, 
},
{
    Script::Cypriot, 
    Script::Linear_A, 
    Script::Linear_B, 
},
{
    Script::Devanagari, 
    Script::Grantha, 
    Script::Kannada, 
},
{
    Script::Devanagari, 
    Script::Grantha, 
    Script::Latin, 
},
{
    Script::Han, 
    Script::Hiragana, 
    Script::Katakana, 
},
{
    Script::Kayah_Li, 
    Script::Latin, 
    Script::Myanmar, 
},
{
    Script::Bengali, 
    Script::Devanagari, 
    Script::Grantha, 
    Script::Kannada, 
},
{
    Script::Buhid, 
    Script::Hanunoo, 
    Script::Tagbanwa, 
    Script::Tagalog, 
},
{
    Script::Devanagari, 
    Script::Dogra, 
    Script::Kaithi, 
    Script::Mahajani, 
},
{
    Script::Arabic, 
    Script::Hanifi_Rohingya, 
    Script::Syriac, 
    Script::Thaana, 
    Script::Yezidi, 
},
{
    Script::Bopomofo, 
    Script::Hangul, 
    Script::Han, 
    Script::Hiragana, 
    Script::Katakana, 
},
{
    Script::Bopomofo, 
    Script::Hangul, 
    Script::Han, 
    Script::Hiragana, 
    Script::Katakana, 
    Script::Yi, 
},
{
    Script::Devanagari, 
    Script::Kannada, 
    Script::Malayalam, 
    Script::Oriya, 
    Script::Tamil, 
    Script::Telugu, 
},
{
    Script::Adlam, 
    Script::Arabic, 
    Script::Mandaic, 
    Script::Manichaean, 
    Script::Psalter_Pahlavi, 
    Script::Hanifi_Rohingya, 
    Script::Sogdian, 
    Script::Syriac, 
},
{
    Script::Bengali, 
    Script::Devanagari, 
    Script::Grantha, 
    Script::Kannada, 
    Script::Nandinagari, 
    Script::Oriya, 
    Script::Telugu, 
    Script::Tirhuta, 
},
{
    Script::Devanagari, 
    Script::Dogra, 
    Script::Gujarati, 
    Script::Gurmukhi, 
    Script::Khojki, 
    Script::Kaithi, 
    Script::Mahajani, 
    Script::Modi, 
    Script::Khudawadi, 
    Script::Takri, 
    Script::Tirhuta, 
},
{
    Script::Bengali, 
    Script::Devanagari, 
    Script::Grantha, 
    Script::Gujarati, 
    Script::Gurmukhi, 
    Script::Kannada, 
    Script::Latin, 
    Script::Malayalam, 
    Script::Oriya, 
    Script::Tamil, 
    Script::Telugu, 
    Script::Tirhuta, 
},
{
    Script::Bengali, 
    Script::Devanagari, 
    Script::Grantha, 
    Script::Gujarati, 
    Script::Gurmukhi, 
    Script::Kannada, 
    Script::Latin, 
    Script::Malayalam, 
    Script::Oriya, 
    Script::Sharada, 
    Script::Tamil, 
    Script::Telugu, 
    Script::Tirhuta, 
},
{
    Script::Devanagari, 
    Script::Dogra, 
    Script::Gujarati, 
    Script::Gurmukhi, 
    Script::Khojki, 
    Script::Kannada, 
    Script::Kaithi, 
    Script::Mahajani, 
    Script::Modi, 
    Script::Nandinagari, 
    Script::Khudawadi, 
    Script::Takri, 
    Script::Tirhuta, 
},
{
    Script::Devanagari, 
    Script::Dogra, 
    Script::Gujarati, 
    Script::Gurmukhi, 
    Script::Khojki, 
    Script::Kannada, 
    Script::Kaithi, 
    Script::Mahajani, 
    Script::Malayalam, 
    Script::Modi, 
    Script::Nandinagari, 
    Script::Khudawadi, 
    Script::Takri, 
    Script::Tirhuta, 
},
{
    Script::Bengali, 
    Script::Devanagari, 
    Script::Dogra, 
    Script::Gunjala_Gondi, 
    Script::Masaram_Gondi, 
    Script::Grantha, 
    Script::Gujarati, 
    Script::Gurmukhi, 
    Script::Kannada, 
    Script::Mahajani, 
    Script::Malayalam, 
    Script::Nandinagari, 
    Script::Oriya, 
    Script::Khudawadi, 
    Script::Sinhala, 
    Script::Syloti_Nagri, 
    Script::Takri, 
    Script::Tamil, 
    Script::Telugu, 
    Script::Tirhuta, 
},
{
    Script::Bengali, 
    Script::Devanagari, 
    Script::Dogra, 
    Script::Gunjala_Gondi, 
    Script::Masaram_Gondi, 
    Script::Grantha, 
    Script::Gujarati, 
    Script::Gurmukhi, 
    Script::Kannada, 
    Script::Limbu, 
    Script::Mahajani, 
    Script::Malayalam, 
    Script::Nandinagari, 
    Script::Oriya, 
    Script::Khudawadi, 
    Script::Sinhala, 
    Script::Syloti_Nagri, 
    Script::Takri, 
    Script::Tamil, 
    Script::Telugu, 
    Script::Tirhuta, 
},
};
namespace _normalization_properties {
static const size_t _block_size = 128;
static const NormalizationProperties _1[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"noBreak",U"\U00000020"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000020\U00000308"},{0,0,0},{0,"super",U"\U00000061"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000020\U00000304"},{0,0,0},{0,0,0},{0,"super",U"\U00000032"},{0,"super",U"\U00000033"},{0,"compat",U"\U00000020\U00000301"},{0,"compat",U"\U000003BC"},{0,0,0},{0,0,0},{0,"compat",U"\U00000020\U00000327"},{0,"super",U"\U00000031"},{0,"super",U"\U0000006F"},{0,0,0},{0,"fraction",U"\U00000031\U00002044\U00000034"},{0,"fraction",U"\U00000031\U00002044\U00000032"},{0,"fraction",U"\U00000033\U00002044\U00000034"},{0,0,0},{0,0,U"\U00000041\U00000300"},{0,0,U"\U00000041\U00000301"},{0,0,U"\U00000041\U00000302"},{0,0,U"\U00000041\U00000303"},{0,0,U"\U00000041\U00000308"},{0,0,U"\U00000041\U0000030A"},{0,0,0},{0,0,U"\U00000043\U00000327"},{0,0,U"\U00000045\U00000300"},{0,0,U"\U00000045\U00000301"},{0,0,U"\U00000045\U00000302"},{0,0,U"\U00000045\U00000308"},{0,0,U"\U00000049\U00000300"},{0,0,U"\U00000049\U00000301"},{0,0,U"\U00000049\U00000302"},{0,0,U"\U00000049\U00000308"},{0,0,0},{0,0,U"\U0000004E\U00000303"},{0,0,U"\U0000004F\U00000300"},{0,0,U"\U0000004F\U00000301"},{0,0,U"\U0000004F\U00000302"},{0,0,U"\U0000004F\U00000303"},{0,0,U"\U0000004F\U00000308"},{0,0,0},{0,0,0},{0,0,U"\U00000055\U00000300"},{0,0,U"\U00000055\U00000301"},{0,0,U"\U00000055\U00000302"},{0,0,U"\U00000055\U00000308"},{0,0,U"\U00000059\U00000301"},{0,0,0},{0,0,0},{0,0,U"\U00000061\U00000300"},{0,0,U"\U00000061\U00000301"},{0,0,U"\U00000061\U00000302"},{0,0,U"\U00000061\U00000303"},{0,0,U"\U00000061\U00000308"},{0,0,U"\U00000061\U0000030A"},{0,0,0},{0,0,U"\U00000063\U00000327"},{0,0,U"\U00000065\U00000300"},{0,0,U"\U00000065\U00000301"},{0,0,U"\U00000065\U00000302"},{0,0,U"\U00000065\U00000308"},{0,0,U"\U00000069\U00000300"},{0,0,U"\U00000069\U00000301"},{0,0,U"\U00000069\U00000302"},{0,0,U"\U00000069\U00000308"},{0,0,0},{0,0,U"\U0000006E\U00000303"},{0,0,U"\U0000006F\U00000300"},{0,0,U"\U0000006F\U00000301"},{0,0,U"\U0000006F\U00000302"},{0,0,U"\U0000006F\U00000303"},{0,0,U"\U0000006F\U00000308"},{0,0,0},{0,0,0},{0,0,U"\U00000075\U00000300"},{0,0,U"\U00000075\U00000301"},{0,0,U"\U00000075\U00000302"},{0,0,U"\U00000075\U00000308"},{0,0,U"\U00000079\U00000301"},{0,0,0},{0,0,U"\U00000079\U00000308"}, };
static const NormalizationProperties _2[] = { {0,0,U"\U00000041\U00000304"},{0,0,U"\U00000061\U00000304"},{0,0,U"\U00000041\U00000306"},{0,0,U"\U00000061\U00000306"},{0,0,U"\U00000041\U00000328"},{0,0,U"\U00000061\U00000328"},{0,0,U"\U00000043\U00000301"},{0,0,U"\U00000063\U00000301"},{0,0,U"\U00000043\U00000302"},{0,0,U"\U00000063\U00000302"},{0,0,U"\U00000043\U00000307"},{0,0,U"\U00000063\U00000307"},{0,0,U"\U00000043\U0000030C"},{0,0,U"\U00000063\U0000030C"},{0,0,U"\U00000044\U0000030C"},{0,0,U"\U00000064\U0000030C"},{0,0,0},{0,0,0},{0,0,U"\U00000045\U00000304"},{0,0,U"\U00000065\U00000304"},{0,0,U"\U00000045\U00000306"},{0,0,U"\U00000065\U00000306"},{0,0,U"\U00000045\U00000307"},{0,0,U"\U00000065\U00000307"},{0,0,U"\U00000045\U00000328"},{0,0,U"\U00000065\U00000328"},{0,0,U"\U00000045\U0000030C"},{0,0,U"\U00000065\U0000030C"},{0,0,U"\U00000047\U00000302"},{0,0,U"\U00000067\U00000302"},{0,0,U"\U00000047\U00000306"},{0,0,U"\U00000067\U00000306"},{0,0,U"\U00000047\U00000307"},{0,0,U"\U00000067\U00000307"},{0,0,U"\U00000047\U00000327"},{0,0,U"\U00000067\U00000327"},{0,0,U"\U00000048\U00000302"},{0,0,U"\U00000068\U00000302"},{0,0,0},{0,0,0},{0,0,U"\U00000049\U00000303"},{0,0,U"\U00000069\U00000303"},{0,0,U"\U00000049\U00000304"},{0,0,U"\U00000069\U00000304"},{0,0,U"\U00000049\U00000306"},{0,0,U"\U00000069\U00000306"},{0,0,U"\U00000049\U00000328"},{0,0,U"\U00000069\U00000328"},{0,0,U"\U00000049\U00000307"},{0,0,0},{0,"compat",U"\U00000049\U0000004A"},{0,"compat",U"\U00000069\U0000006A"},{0,0,U"\U0000004A\U00000302"},{0,0,U"\U0000006A\U00000302"},{0,0,U"\U0000004B\U00000327"},{0,0,U"\U0000006B\U00000327"},{0,0,0},{0,0,U"\U0000004C\U00000301"},{0,0,U"\U0000006C\U00000301"},{0,0,U"\U0000004C\U00000327"},{0,0,U"\U0000006C\U00000327"},{0,0,U"\U0000004C\U0000030C"},{0,0,U"\U0000006C\U0000030C"},{0,"compat",U"\U0000004C\U000000B7"},{0,"compat",U"\U0000006C\U000000B7"},{0,0,0},{0,0,0},{0,0,U"\U0000004E\U00000301"},{0,0,U"\U0000006E\U00000301"},{0,0,U"\U0000004E\U00000327"},{0,0,U"\U0000006E\U00000327"},{0,0,U"\U0000004E\U0000030C"},{0,0,U"\U0000006E\U0000030C"},{0,"compat",U"\U000002BC\U0000006E"},{0,0,0},{0,0,0},{0,0,U"\U0000004F\U00000304"},{0,0,U"\U0000006F\U00000304"},{0,0,U"\U0000004F\U00000306"},{0,0,U"\U0000006F\U00000306"},{0,0,U"\U0000004F\U0000030B"},{0,0,U"\U0000006F\U0000030B"},{0,0,0},{0,0,0},{0,0,U"\U00000052\U00000301"},{0,0,U"\U00000072\U00000301"},{0,0,U"\U00000052\U00000327"},{0,0,U"\U00000072\U00000327"},{0,0,U"\U00000052\U0000030C"},{0,0,U"\U00000072\U0000030C"},{0,0,U"\U00000053\U00000301"},{0,0,U"\U00000073\U00000301"},{0,0,U"\U00000053\U00000302"},{0,0,U"\U00000073\U00000302"},{0,0,U"\U00000053\U00000327"},{0,0,U"\U00000073\U00000327"},{0,0,U"\U00000053\U0000030C"},{0,0,U"\U00000073\U0000030C"},{0,0,U"\U00000054\U00000327"},{0,0,U"\U00000074\U00000327"},{0,0,U"\U00000054\U0000030C"},{0,0,U"\U00000074\U0000030C"},{0,0,0},{0,0,0},{0,0,U"\U00000055\U00000303"},{0,0,U"\U00000075\U00000303"},{0,0,U"\U00000055\U00000304"},{0,0,U"\U00000075\U00000304"},{0,0,U"\U00000055\U00000306"},{0,0,U"\U00000075\U00000306"},{0,0,U"\U00000055\U0000030A"},{0,0,U"\U00000075\U0000030A"},{0,0,U"\U00000055\U0000030B"},{0,0,U"\U00000075\U0000030B"},{0,0,U"\U00000055\U00000328"},{0,0,U"\U00000075\U00000328"},{0,0,U"\U00000057\U00000302"},{0,0,U"\U00000077\U00000302"},{0,0,U"\U00000059\U00000302"},{0,0,U"\U00000079\U00000302"},{0,0,U"\U00000059\U00000308"},{0,0,U"\U0000005A\U00000301"},{0,0,U"\U0000007A\U00000301"},{0,0,U"\U0000005A\U00000307"},{0,0,U"\U0000007A\U00000307"},{0,0,U"\U0000005A\U0000030C"},{0,0,U"\U0000007A\U0000030C"},{0,"compat",U"\U00000073"}, };
static const NormalizationProperties _3[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0000004F\U0000031B"},{0,0,U"\U0000006F\U0000031B"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000055\U0000031B"},{0,0,U"\U00000075\U0000031B"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000044\U0000017D"},{0,"compat",U"\U00000044\U0000017E"},{0,"compat",U"\U00000064\U0000017E"},{0,"compat",U"\U0000004C\U0000004A"},{0,"compat",U"\U0000004C\U0000006A"},{0,"compat",U"\U0000006C\U0000006A"},{0,"compat",U"\U0000004E\U0000004A"},{0,"compat",U"\U0000004E\U0000006A"},{0,"compat",U"\U0000006E\U0000006A"},{0,0,U"\U00000041\U0000030C"},{0,0,U"\U00000061\U0000030C"},{0,0,U"\U00000049\U0000030C"},{0,0,U"\U00000069\U0000030C"},{0,0,U"\U0000004F\U0000030C"},{0,0,U"\U0000006F\U0000030C"},{0,0,U"\U00000055\U0000030C"},{0,0,U"\U00000075\U0000030C"},{0,0,U"\U000000DC\U00000304"},{0,0,U"\U000000FC\U00000304"},{0,0,U"\U000000DC\U00000301"},{0,0,U"\U000000FC\U00000301"},{0,0,U"\U000000DC\U0000030C"},{0,0,U"\U000000FC\U0000030C"},{0,0,U"\U000000DC\U00000300"},{0,0,U"\U000000FC\U00000300"},{0,0,0},{0,0,U"\U000000C4\U00000304"},{0,0,U"\U000000E4\U00000304"},{0,0,U"\U00000226\U00000304"},{0,0,U"\U00000227\U00000304"},{0,0,U"\U000000C6\U00000304"},{0,0,U"\U000000E6\U00000304"},{0,0,0},{0,0,0},{0,0,U"\U00000047\U0000030C"},{0,0,U"\U00000067\U0000030C"},{0,0,U"\U0000004B\U0000030C"},{0,0,U"\U0000006B\U0000030C"},{0,0,U"\U0000004F\U00000328"},{0,0,U"\U0000006F\U00000328"},{0,0,U"\U000001EA\U00000304"},{0,0,U"\U000001EB\U00000304"},{0,0,U"\U000001B7\U0000030C"},{0,0,U"\U00000292\U0000030C"},{0,0,U"\U0000006A\U0000030C"},{0,"compat",U"\U00000044\U0000005A"},{0,"compat",U"\U00000044\U0000007A"},{0,"compat",U"\U00000064\U0000007A"},{0,0,U"\U00000047\U00000301"},{0,0,U"\U00000067\U00000301"},{0,0,0},{0,0,0},{0,0,U"\U0000004E\U00000300"},{0,0,U"\U0000006E\U00000300"},{0,0,U"\U000000C5\U00000301"},{0,0,U"\U000000E5\U00000301"},{0,0,U"\U000000C6\U00000301"},{0,0,U"\U000000E6\U00000301"},{0,0,U"\U000000D8\U00000301"},{0,0,U"\U000000F8\U00000301"}, };
static const NormalizationProperties _4[] = { {0,0,U"\U00000041\U0000030F"},{0,0,U"\U00000061\U0000030F"},{0,0,U"\U00000041\U00000311"},{0,0,U"\U00000061\U00000311"},{0,0,U"\U00000045\U0000030F"},{0,0,U"\U00000065\U0000030F"},{0,0,U"\U00000045\U00000311"},{0,0,U"\U00000065\U00000311"},{0,0,U"\U00000049\U0000030F"},{0,0,U"\U00000069\U0000030F"},{0,0,U"\U00000049\U00000311"},{0,0,U"\U00000069\U00000311"},{0,0,U"\U0000004F\U0000030F"},{0,0,U"\U0000006F\U0000030F"},{0,0,U"\U0000004F\U00000311"},{0,0,U"\U0000006F\U00000311"},{0,0,U"\U00000052\U0000030F"},{0,0,U"\U00000072\U0000030F"},{0,0,U"\U00000052\U00000311"},{0,0,U"\U00000072\U00000311"},{0,0,U"\U00000055\U0000030F"},{0,0,U"\U00000075\U0000030F"},{0,0,U"\U00000055\U00000311"},{0,0,U"\U00000075\U00000311"},{0,0,U"\U00000053\U00000326"},{0,0,U"\U00000073\U00000326"},{0,0,U"\U00000054\U00000326"},{0,0,U"\U00000074\U00000326"},{0,0,0},{0,0,0},{0,0,U"\U00000048\U0000030C"},{0,0,U"\U00000068\U0000030C"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000041\U00000307"},{0,0,U"\U00000061\U00000307"},{0,0,U"\U00000045\U00000327"},{0,0,U"\U00000065\U00000327"},{0,0,U"\U000000D6\U00000304"},{0,0,U"\U000000F6\U00000304"},{0,0,U"\U000000D5\U00000304"},{0,0,U"\U000000F5\U00000304"},{0,0,U"\U0000004F\U00000307"},{0,0,U"\U0000006F\U00000307"},{0,0,U"\U0000022E\U00000304"},{0,0,U"\U0000022F\U00000304"},{0,0,U"\U00000059\U00000304"},{0,0,U"\U00000079\U00000304"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _5[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U00000068"},{0,"super",U"\U00000266"},{0,"super",U"\U0000006A"},{0,"super",U"\U00000072"},{0,"super",U"\U00000279"},{0,"super",U"\U0000027B"},{0,"super",U"\U00000281"},{0,"super",U"\U00000077"},{0,"super",U"\U00000079"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000020\U00000306"},{0,"compat",U"\U00000020\U00000307"},{0,"compat",U"\U00000020\U0000030A"},{0,"compat",U"\U00000020\U00000328"},{0,"compat",U"\U00000020\U00000303"},{0,"compat",U"\U00000020\U0000030B"},{0,0,0},{0,0,0},{0,"super",U"\U00000263"},{0,"super",U"\U0000006C"},{0,"super",U"\U00000073"},{0,"super",U"\U00000078"},{0,"super",U"\U00000295"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _6[] = { {230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{232,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{232,0,0},{216,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{202,0,0},{202,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{202,0,0},{202,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{1,0,0},{1,0,0},{1,0,0},{1,0,0},{1,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,U"\U00000300"},{230,0,U"\U00000301"},{230,0,0},{230,0,U"\U00000313"},{230,0,U"\U00000308\U00000301"},{240,0,0},{230,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{232,0,0},{220,0,0},{220,0,0},{230,0,0},{233,0,0},{234,0,0},{234,0,0},{233,0,0},{234,0,0},{234,0,0},{233,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000002B9"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000020\U00000345"},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0000003B"},{0,0,0}, };
static const NormalizationProperties _7[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000020\U00000301"},{0,0,U"\U000000A8\U00000301"},{0,0,U"\U00000391\U00000301"},{0,0,U"\U000000B7"},{0,0,U"\U00000395\U00000301"},{0,0,U"\U00000397\U00000301"},{0,0,U"\U00000399\U00000301"},{0,0,0},{0,0,U"\U0000039F\U00000301"},{0,0,0},{0,0,U"\U000003A5\U00000301"},{0,0,U"\U000003A9\U00000301"},{0,0,U"\U000003CA\U00000301"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000399\U00000308"},{0,0,U"\U000003A5\U00000308"},{0,0,U"\U000003B1\U00000301"},{0,0,U"\U000003B5\U00000301"},{0,0,U"\U000003B7\U00000301"},{0,0,U"\U000003B9\U00000301"},{0,0,U"\U000003CB\U00000301"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000003B9\U00000308"},{0,0,U"\U000003C5\U00000308"},{0,0,U"\U000003BF\U00000301"},{0,0,U"\U000003C5\U00000301"},{0,0,U"\U000003C9\U00000301"},{0,0,0},{0,"compat",U"\U000003B2"},{0,"compat",U"\U000003B8"},{0,"compat",U"\U000003A5"},{0,0,U"\U000003D2\U00000301"},{0,0,U"\U000003D2\U00000308"},{0,"compat",U"\U000003C6"},{0,"compat",U"\U000003C0"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U000003BA"},{0,"compat",U"\U000003C1"},{0,"compat",U"\U000003C2"},{0,0,0},{0,"compat",U"\U00000398"},{0,"compat",U"\U000003B5"},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U000003A3"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _8[] = { {0,0,U"\U00000415\U00000300"},{0,0,U"\U00000415\U00000308"},{0,0,0},{0,0,U"\U00000413\U00000301"},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000406\U00000308"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0000041A\U00000301"},{0,0,U"\U00000418\U00000300"},{0,0,U"\U00000423\U00000306"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000418\U00000306"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000438\U00000306"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000435\U00000300"},{0,0,U"\U00000435\U00000308"},{0,0,0},{0,0,U"\U00000433\U00000301"},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000456\U00000308"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0000043A\U00000301"},{0,0,U"\U00000438\U00000300"},{0,0,U"\U00000443\U00000306"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000474\U0000030F"},{0,0,U"\U00000475\U0000030F"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _9[] = { {0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000416\U00000306"},{0,0,U"\U00000436\U00000306"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000410\U00000306"},{0,0,U"\U00000430\U00000306"},{0,0,U"\U00000410\U00000308"},{0,0,U"\U00000430\U00000308"},{0,0,0},{0,0,0},{0,0,U"\U00000415\U00000306"},{0,0,U"\U00000435\U00000306"},{0,0,0},{0,0,0},{0,0,U"\U000004D8\U00000308"},{0,0,U"\U000004D9\U00000308"},{0,0,U"\U00000416\U00000308"},{0,0,U"\U00000436\U00000308"},{0,0,U"\U00000417\U00000308"},{0,0,U"\U00000437\U00000308"},{0,0,0},{0,0,0},{0,0,U"\U00000418\U00000304"},{0,0,U"\U00000438\U00000304"},{0,0,U"\U00000418\U00000308"},{0,0,U"\U00000438\U00000308"},{0,0,U"\U0000041E\U00000308"},{0,0,U"\U0000043E\U00000308"},{0,0,0},{0,0,0},{0,0,U"\U000004E8\U00000308"},{0,0,U"\U000004E9\U00000308"},{0,0,U"\U0000042D\U00000308"},{0,0,U"\U0000044D\U00000308"},{0,0,U"\U00000423\U00000304"},{0,0,U"\U00000443\U00000304"},{0,0,U"\U00000423\U00000308"},{0,0,U"\U00000443\U00000308"},{0,0,U"\U00000423\U0000030B"},{0,0,U"\U00000443\U0000030B"},{0,0,U"\U00000427\U00000308"},{0,0,U"\U00000447\U00000308"},{0,0,0},{0,0,0},{0,0,U"\U0000042B\U00000308"},{0,0,U"\U0000044B\U00000308"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _11[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000565\U00000582"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{222,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{222,0,0},{228,0,0},{230,0,0},{10,0,0},{11,0,0},{12,0,0},{13,0,0},{14,0,0},{15,0,0},{16,0,0},{17,0,0},{18,0,0},{19,0,0},{19,0,0},{20,0,0},{21,0,0},{22,0,0},{0,0,0},{23,0,0},{0,0,0},{24,0,0},{25,0,0},{0,0,0},{230,0,0},{220,0,0},{0,0,0},{18,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _12[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{30,0,0},{31,0,0},{32,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000627\U00000653"},{0,0,U"\U00000627\U00000654"},{0,0,U"\U00000648\U00000654"},{0,0,U"\U00000627\U00000655"},{0,0,U"\U0000064A\U00000654"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{27,0,0},{28,0,0},{29,0,0},{30,0,0},{31,0,0},{32,0,0},{33,0,0},{34,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{35,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000627\U00000674"},{0,"compat",U"\U00000648\U00000674"},{0,"compat",U"\U000006C7\U00000674"},{0,"compat",U"\U0000064A\U00000674"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _13[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000006D5\U00000654"},{0,0,0},{0,0,U"\U000006C1\U00000654"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000006D2\U00000654"},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{0,0,0},{220,0,0},{230,0,0},{230,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _14[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{36,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{220,0,0},{220,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{220,0,0},{230,0,0},{220,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _15[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{220,0,0},{220,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _16[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{220,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _17[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{220,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{220,0,0},{27,0,0},{28,0,0},{29,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0}, };
static const NormalizationProperties _18[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000928\U0000093C"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000930\U0000093C"},{0,0,0},{0,0,0},{0,0,U"\U00000933\U0000093C"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000915\U0000093C"},{0,0,U"\U00000916\U0000093C"},{0,0,U"\U00000917\U0000093C"},{0,0,U"\U0000091C\U0000093C"},{0,0,U"\U00000921\U0000093C"},{0,0,U"\U00000922\U0000093C"},{0,0,U"\U0000092B\U0000093C"},{0,0,U"\U0000092F\U0000093C"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _19[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{7,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000009C7\U000009BE"},{0,0,U"\U000009C7\U000009D7"},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000009A1\U000009BC"},{0,0,U"\U000009A2\U000009BC"},{0,0,0},{0,0,U"\U000009AF\U000009BC"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{0,0,0}, };
static const NormalizationProperties _20[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000A32\U00000A3C"},{0,0,0},{0,0,0},{0,0,U"\U00000A38\U00000A3C"},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{7,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000A16\U00000A3C"},{0,0,U"\U00000A17\U00000A3C"},{0,0,U"\U00000A1C\U00000A3C"},{0,0,0},{0,0,0},{0,0,U"\U00000A2B\U00000A3C"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _21[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{7,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _22[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{7,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000B47\U00000B56"},{0,0,0},{0,0,0},{0,0,U"\U00000B47\U00000B3E"},{0,0,U"\U00000B47\U00000B57"},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000B21\U00000B3C"},{0,0,U"\U00000B22\U00000B3C"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _23[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000B92\U00000BD7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000BC6\U00000BBE"},{0,0,U"\U00000BC7\U00000BBE"},{0,0,U"\U00000BC6\U00000BD7"},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _24[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000C46\U00000C56"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{84,0,0},{84,0,0},{84,0,0},{84,0,0},{84,0,0},{84,0,0},{84,0,0},{84,0,0},{91,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _25[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{7,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000CBF\U00000CD5"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000CC6\U00000CD5"},{0,0,U"\U00000CC6\U00000CD6"},{0,0,0},{0,0,U"\U00000CC6\U00000CC2"},{0,0,U"\U00000CCA\U00000CD5"},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _26[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000D46\U00000D3E"},{0,0,U"\U00000D47\U00000D3E"},{0,0,U"\U00000D46\U00000D57"},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _27[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000DD9\U00000DCA"},{0,0,0},{0,0,U"\U00000DD9\U00000DCF"},{0,0,U"\U00000DDC\U00000DCA"},{0,0,U"\U00000DD9\U00000DDF"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _28[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000E4D\U00000E32"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{103,0,0},{103,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{107,0,0},{107,0,0},{107,0,0},{107,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _29[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000ECD\U00000EB2"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{118,0,0},{118,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{122,0,0},{122,0,0},{122,0,0},{122,0,0},{122,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000EAB\U00000E99"},{0,"compat",U"\U00000EAB\U00000EA1"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _30[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"noBreak",U"\U00000F0B"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{0,0,0},{220,0,0},{0,0,0},{216,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000F42\U00000FB7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000F4C\U00000FB7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000F51\U00000FB7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000F56\U00000FB7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000F5B\U00000FB7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000F40\U00000FB5"},{0,0,0},{0,0,0},{0,0,0},{129,0,0},{129,0,0},{129,0,0},{129,0,0},{129,0,0},{130,0,0},{0,0,U"\U00000F71\U00000F72"},{132,0,0},{0,0,U"\U00000F71\U00000F74"},{0,0,U"\U00000FB2\U00000F80"},{0,"compat",U"\U00000FB2\U00000F81"},{0,0,U"\U00000FB3\U00000F80"},{0,"compat",U"\U00000FB3\U00000F81"},{130,0,0},{130,0,0},{130,0,0},{130,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _31[] = { {130,0,0},{0,0,U"\U00000F71\U00000F80"},{230,0,0},{230,0,0},{9,0,0},{0,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000F92\U00000FB7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000F9C\U00000FB7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000FA1\U00000FB7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000FA6\U00000FB7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000FAB\U00000FB7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000F90\U00000FB5"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _32[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00001025\U0000102E"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{9,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _33[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U000010DC"},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _38[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _46[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _47[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _49[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{228,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _50[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{222,0,0},{230,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _52[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{220,0,0}, };
static const NormalizationProperties _53[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{220,0,0},{0,0,0},{220,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _54[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00001B05\U00001B35"},{0,0,0},{0,0,U"\U00001B07\U00001B35"},{0,0,0},{0,0,U"\U00001B09\U00001B35"},{0,0,0},{0,0,U"\U00001B0B\U00001B35"},{0,0,0},{0,0,U"\U00001B0D\U00001B35"},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00001B11\U00001B35"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00001B3A\U00001B35"},{0,0,0},{0,0,U"\U00001B3C\U00001B35"},{0,0,0},{0,0,0},{0,0,U"\U00001B3E\U00001B35"},{0,0,U"\U00001B3F\U00001B35"},{0,0,0},{0,0,U"\U00001B42\U00001B35"},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _55[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _56[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _57[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{1,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{0,0,0},{1,0,0},{1,0,0},{1,0,0},{1,0,0},{1,0,0},{1,0,0},{1,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _58[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U00000041"},{0,"super",U"\U000000C6"},{0,"super",U"\U00000042"},{0,0,0},{0,"super",U"\U00000044"},{0,"super",U"\U00000045"},{0,"super",U"\U0000018E"},{0,"super",U"\U00000047"},{0,"super",U"\U00000048"},{0,"super",U"\U00000049"},{0,"super",U"\U0000004A"},{0,"super",U"\U0000004B"},{0,"super",U"\U0000004C"},{0,"super",U"\U0000004D"},{0,"super",U"\U0000004E"},{0,0,0},{0,"super",U"\U0000004F"},{0,"super",U"\U00000222"},{0,"super",U"\U00000050"},{0,"super",U"\U00000052"},{0,"super",U"\U00000054"},{0,"super",U"\U00000055"},{0,"super",U"\U00000057"},{0,"super",U"\U00000061"},{0,"super",U"\U00000250"},{0,"super",U"\U00000251"},{0,"super",U"\U00001D02"},{0,"super",U"\U00000062"},{0,"super",U"\U00000064"},{0,"super",U"\U00000065"},{0,"super",U"\U00000259"},{0,"super",U"\U0000025B"},{0,"super",U"\U0000025C"},{0,"super",U"\U00000067"},{0,0,0},{0,"super",U"\U0000006B"},{0,"super",U"\U0000006D"},{0,"super",U"\U0000014B"},{0,"super",U"\U0000006F"},{0,"super",U"\U00000254"},{0,"super",U"\U00001D16"},{0,"super",U"\U00001D17"},{0,"super",U"\U00000070"},{0,"super",U"\U00000074"},{0,"super",U"\U00000075"},{0,"super",U"\U00001D1D"},{0,"super",U"\U0000026F"},{0,"super",U"\U00000076"},{0,"super",U"\U00001D25"},{0,"super",U"\U000003B2"},{0,"super",U"\U000003B3"},{0,"super",U"\U000003B4"},{0,"super",U"\U000003C6"},{0,"super",U"\U000003C7"},{0,"sub",U"\U00000069"},{0,"sub",U"\U00000072"},{0,"sub",U"\U00000075"},{0,"sub",U"\U00000076"},{0,"sub",U"\U000003B2"},{0,"sub",U"\U000003B3"},{0,"sub",U"\U000003C1"},{0,"sub",U"\U000003C6"},{0,"sub",U"\U000003C7"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U0000043D"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _59[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U00000252"},{0,"super",U"\U00000063"},{0,"super",U"\U00000255"},{0,"super",U"\U000000F0"},{0,"super",U"\U0000025C"},{0,"super",U"\U00000066"},{0,"super",U"\U0000025F"},{0,"super",U"\U00000261"},{0,"super",U"\U00000265"},{0,"super",U"\U00000268"},{0,"super",U"\U00000269"},{0,"super",U"\U0000026A"},{0,"super",U"\U00001D7B"},{0,"super",U"\U0000029D"},{0,"super",U"\U0000026D"},{0,"super",U"\U00001D85"},{0,"super",U"\U0000029F"},{0,"super",U"\U00000271"},{0,"super",U"\U00000270"},{0,"super",U"\U00000272"},{0,"super",U"\U00000273"},{0,"super",U"\U00000274"},{0,"super",U"\U00000275"},{0,"super",U"\U00000278"},{0,"super",U"\U00000282"},{0,"super",U"\U00000283"},{0,"super",U"\U000001AB"},{0,"super",U"\U00000289"},{0,"super",U"\U0000028A"},{0,"super",U"\U00001D1C"},{0,"super",U"\U0000028B"},{0,"super",U"\U0000028C"},{0,"super",U"\U0000007A"},{0,"super",U"\U00000290"},{0,"super",U"\U00000291"},{0,"super",U"\U00000292"},{0,"super",U"\U000003B8"},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{230,0,0},{234,0,0},{214,0,0},{220,0,0},{202,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{232,0,0},{228,0,0},{228,0,0},{220,0,0},{230,0,0},{230,0,0},{233,0,0},{220,0,0},{230,0,0},{220,0,0}, };
static const NormalizationProperties _60[] = { {0,0,U"\U00000041\U00000325"},{0,0,U"\U00000061\U00000325"},{0,0,U"\U00000042\U00000307"},{0,0,U"\U00000062\U00000307"},{0,0,U"\U00000042\U00000323"},{0,0,U"\U00000062\U00000323"},{0,0,U"\U00000042\U00000331"},{0,0,U"\U00000062\U00000331"},{0,0,U"\U000000C7\U00000301"},{0,0,U"\U000000E7\U00000301"},{0,0,U"\U00000044\U00000307"},{0,0,U"\U00000064\U00000307"},{0,0,U"\U00000044\U00000323"},{0,0,U"\U00000064\U00000323"},{0,0,U"\U00000044\U00000331"},{0,0,U"\U00000064\U00000331"},{0,0,U"\U00000044\U00000327"},{0,0,U"\U00000064\U00000327"},{0,0,U"\U00000044\U0000032D"},{0,0,U"\U00000064\U0000032D"},{0,0,U"\U00000112\U00000300"},{0,0,U"\U00000113\U00000300"},{0,0,U"\U00000112\U00000301"},{0,0,U"\U00000113\U00000301"},{0,0,U"\U00000045\U0000032D"},{0,0,U"\U00000065\U0000032D"},{0,0,U"\U00000045\U00000330"},{0,0,U"\U00000065\U00000330"},{0,0,U"\U00000228\U00000306"},{0,0,U"\U00000229\U00000306"},{0,0,U"\U00000046\U00000307"},{0,0,U"\U00000066\U00000307"},{0,0,U"\U00000047\U00000304"},{0,0,U"\U00000067\U00000304"},{0,0,U"\U00000048\U00000307"},{0,0,U"\U00000068\U00000307"},{0,0,U"\U00000048\U00000323"},{0,0,U"\U00000068\U00000323"},{0,0,U"\U00000048\U00000308"},{0,0,U"\U00000068\U00000308"},{0,0,U"\U00000048\U00000327"},{0,0,U"\U00000068\U00000327"},{0,0,U"\U00000048\U0000032E"},{0,0,U"\U00000068\U0000032E"},{0,0,U"\U00000049\U00000330"},{0,0,U"\U00000069\U00000330"},{0,0,U"\U000000CF\U00000301"},{0,0,U"\U000000EF\U00000301"},{0,0,U"\U0000004B\U00000301"},{0,0,U"\U0000006B\U00000301"},{0,0,U"\U0000004B\U00000323"},{0,0,U"\U0000006B\U00000323"},{0,0,U"\U0000004B\U00000331"},{0,0,U"\U0000006B\U00000331"},{0,0,U"\U0000004C\U00000323"},{0,0,U"\U0000006C\U00000323"},{0,0,U"\U00001E36\U00000304"},{0,0,U"\U00001E37\U00000304"},{0,0,U"\U0000004C\U00000331"},{0,0,U"\U0000006C\U00000331"},{0,0,U"\U0000004C\U0000032D"},{0,0,U"\U0000006C\U0000032D"},{0,0,U"\U0000004D\U00000301"},{0,0,U"\U0000006D\U00000301"},{0,0,U"\U0000004D\U00000307"},{0,0,U"\U0000006D\U00000307"},{0,0,U"\U0000004D\U00000323"},{0,0,U"\U0000006D\U00000323"},{0,0,U"\U0000004E\U00000307"},{0,0,U"\U0000006E\U00000307"},{0,0,U"\U0000004E\U00000323"},{0,0,U"\U0000006E\U00000323"},{0,0,U"\U0000004E\U00000331"},{0,0,U"\U0000006E\U00000331"},{0,0,U"\U0000004E\U0000032D"},{0,0,U"\U0000006E\U0000032D"},{0,0,U"\U000000D5\U00000301"},{0,0,U"\U000000F5\U00000301"},{0,0,U"\U000000D5\U00000308"},{0,0,U"\U000000F5\U00000308"},{0,0,U"\U0000014C\U00000300"},{0,0,U"\U0000014D\U00000300"},{0,0,U"\U0000014C\U00000301"},{0,0,U"\U0000014D\U00000301"},{0,0,U"\U00000050\U00000301"},{0,0,U"\U00000070\U00000301"},{0,0,U"\U00000050\U00000307"},{0,0,U"\U00000070\U00000307"},{0,0,U"\U00000052\U00000307"},{0,0,U"\U00000072\U00000307"},{0,0,U"\U00000052\U00000323"},{0,0,U"\U00000072\U00000323"},{0,0,U"\U00001E5A\U00000304"},{0,0,U"\U00001E5B\U00000304"},{0,0,U"\U00000052\U00000331"},{0,0,U"\U00000072\U00000331"},{0,0,U"\U00000053\U00000307"},{0,0,U"\U00000073\U00000307"},{0,0,U"\U00000053\U00000323"},{0,0,U"\U00000073\U00000323"},{0,0,U"\U0000015A\U00000307"},{0,0,U"\U0000015B\U00000307"},{0,0,U"\U00000160\U00000307"},{0,0,U"\U00000161\U00000307"},{0,0,U"\U00001E62\U00000307"},{0,0,U"\U00001E63\U00000307"},{0,0,U"\U00000054\U00000307"},{0,0,U"\U00000074\U00000307"},{0,0,U"\U00000054\U00000323"},{0,0,U"\U00000074\U00000323"},{0,0,U"\U00000054\U00000331"},{0,0,U"\U00000074\U00000331"},{0,0,U"\U00000054\U0000032D"},{0,0,U"\U00000074\U0000032D"},{0,0,U"\U00000055\U00000324"},{0,0,U"\U00000075\U00000324"},{0,0,U"\U00000055\U00000330"},{0,0,U"\U00000075\U00000330"},{0,0,U"\U00000055\U0000032D"},{0,0,U"\U00000075\U0000032D"},{0,0,U"\U00000168\U00000301"},{0,0,U"\U00000169\U00000301"},{0,0,U"\U0000016A\U00000308"},{0,0,U"\U0000016B\U00000308"},{0,0,U"\U00000056\U00000303"},{0,0,U"\U00000076\U00000303"},{0,0,U"\U00000056\U00000323"},{0,0,U"\U00000076\U00000323"}, };
static const NormalizationProperties _61[] = { {0,0,U"\U00000057\U00000300"},{0,0,U"\U00000077\U00000300"},{0,0,U"\U00000057\U00000301"},{0,0,U"\U00000077\U00000301"},{0,0,U"\U00000057\U00000308"},{0,0,U"\U00000077\U00000308"},{0,0,U"\U00000057\U00000307"},{0,0,U"\U00000077\U00000307"},{0,0,U"\U00000057\U00000323"},{0,0,U"\U00000077\U00000323"},{0,0,U"\U00000058\U00000307"},{0,0,U"\U00000078\U00000307"},{0,0,U"\U00000058\U00000308"},{0,0,U"\U00000078\U00000308"},{0,0,U"\U00000059\U00000307"},{0,0,U"\U00000079\U00000307"},{0,0,U"\U0000005A\U00000302"},{0,0,U"\U0000007A\U00000302"},{0,0,U"\U0000005A\U00000323"},{0,0,U"\U0000007A\U00000323"},{0,0,U"\U0000005A\U00000331"},{0,0,U"\U0000007A\U00000331"},{0,0,U"\U00000068\U00000331"},{0,0,U"\U00000074\U00000308"},{0,0,U"\U00000077\U0000030A"},{0,0,U"\U00000079\U0000030A"},{0,"compat",U"\U00000061\U000002BE"},{0,0,U"\U0000017F\U00000307"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00000041\U00000323"},{0,0,U"\U00000061\U00000323"},{0,0,U"\U00000041\U00000309"},{0,0,U"\U00000061\U00000309"},{0,0,U"\U000000C2\U00000301"},{0,0,U"\U000000E2\U00000301"},{0,0,U"\U000000C2\U00000300"},{0,0,U"\U000000E2\U00000300"},{0,0,U"\U000000C2\U00000309"},{0,0,U"\U000000E2\U00000309"},{0,0,U"\U000000C2\U00000303"},{0,0,U"\U000000E2\U00000303"},{0,0,U"\U00001EA0\U00000302"},{0,0,U"\U00001EA1\U00000302"},{0,0,U"\U00000102\U00000301"},{0,0,U"\U00000103\U00000301"},{0,0,U"\U00000102\U00000300"},{0,0,U"\U00000103\U00000300"},{0,0,U"\U00000102\U00000309"},{0,0,U"\U00000103\U00000309"},{0,0,U"\U00000102\U00000303"},{0,0,U"\U00000103\U00000303"},{0,0,U"\U00001EA0\U00000306"},{0,0,U"\U00001EA1\U00000306"},{0,0,U"\U00000045\U00000323"},{0,0,U"\U00000065\U00000323"},{0,0,U"\U00000045\U00000309"},{0,0,U"\U00000065\U00000309"},{0,0,U"\U00000045\U00000303"},{0,0,U"\U00000065\U00000303"},{0,0,U"\U000000CA\U00000301"},{0,0,U"\U000000EA\U00000301"},{0,0,U"\U000000CA\U00000300"},{0,0,U"\U000000EA\U00000300"},{0,0,U"\U000000CA\U00000309"},{0,0,U"\U000000EA\U00000309"},{0,0,U"\U000000CA\U00000303"},{0,0,U"\U000000EA\U00000303"},{0,0,U"\U00001EB8\U00000302"},{0,0,U"\U00001EB9\U00000302"},{0,0,U"\U00000049\U00000309"},{0,0,U"\U00000069\U00000309"},{0,0,U"\U00000049\U00000323"},{0,0,U"\U00000069\U00000323"},{0,0,U"\U0000004F\U00000323"},{0,0,U"\U0000006F\U00000323"},{0,0,U"\U0000004F\U00000309"},{0,0,U"\U0000006F\U00000309"},{0,0,U"\U000000D4\U00000301"},{0,0,U"\U000000F4\U00000301"},{0,0,U"\U000000D4\U00000300"},{0,0,U"\U000000F4\U00000300"},{0,0,U"\U000000D4\U00000309"},{0,0,U"\U000000F4\U00000309"},{0,0,U"\U000000D4\U00000303"},{0,0,U"\U000000F4\U00000303"},{0,0,U"\U00001ECC\U00000302"},{0,0,U"\U00001ECD\U00000302"},{0,0,U"\U000001A0\U00000301"},{0,0,U"\U000001A1\U00000301"},{0,0,U"\U000001A0\U00000300"},{0,0,U"\U000001A1\U00000300"},{0,0,U"\U000001A0\U00000309"},{0,0,U"\U000001A1\U00000309"},{0,0,U"\U000001A0\U00000303"},{0,0,U"\U000001A1\U00000303"},{0,0,U"\U000001A0\U00000323"},{0,0,U"\U000001A1\U00000323"},{0,0,U"\U00000055\U00000323"},{0,0,U"\U00000075\U00000323"},{0,0,U"\U00000055\U00000309"},{0,0,U"\U00000075\U00000309"},{0,0,U"\U000001AF\U00000301"},{0,0,U"\U000001B0\U00000301"},{0,0,U"\U000001AF\U00000300"},{0,0,U"\U000001B0\U00000300"},{0,0,U"\U000001AF\U00000309"},{0,0,U"\U000001B0\U00000309"},{0,0,U"\U000001AF\U00000303"},{0,0,U"\U000001B0\U00000303"},{0,0,U"\U000001AF\U00000323"},{0,0,U"\U000001B0\U00000323"},{0,0,U"\U00000059\U00000300"},{0,0,U"\U00000079\U00000300"},{0,0,U"\U00000059\U00000323"},{0,0,U"\U00000079\U00000323"},{0,0,U"\U00000059\U00000309"},{0,0,U"\U00000079\U00000309"},{0,0,U"\U00000059\U00000303"},{0,0,U"\U00000079\U00000303"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _62[] = { {0,0,U"\U000003B1\U00000313"},{0,0,U"\U000003B1\U00000314"},{0,0,U"\U00001F00\U00000300"},{0,0,U"\U00001F01\U00000300"},{0,0,U"\U00001F00\U00000301"},{0,0,U"\U00001F01\U00000301"},{0,0,U"\U00001F00\U00000342"},{0,0,U"\U00001F01\U00000342"},{0,0,U"\U00000391\U00000313"},{0,0,U"\U00000391\U00000314"},{0,0,U"\U00001F08\U00000300"},{0,0,U"\U00001F09\U00000300"},{0,0,U"\U00001F08\U00000301"},{0,0,U"\U00001F09\U00000301"},{0,0,U"\U00001F08\U00000342"},{0,0,U"\U00001F09\U00000342"},{0,0,U"\U000003B5\U00000313"},{0,0,U"\U000003B5\U00000314"},{0,0,U"\U00001F10\U00000300"},{0,0,U"\U00001F11\U00000300"},{0,0,U"\U00001F10\U00000301"},{0,0,U"\U00001F11\U00000301"},{0,0,0},{0,0,0},{0,0,U"\U00000395\U00000313"},{0,0,U"\U00000395\U00000314"},{0,0,U"\U00001F18\U00000300"},{0,0,U"\U00001F19\U00000300"},{0,0,U"\U00001F18\U00000301"},{0,0,U"\U00001F19\U00000301"},{0,0,0},{0,0,0},{0,0,U"\U000003B7\U00000313"},{0,0,U"\U000003B7\U00000314"},{0,0,U"\U00001F20\U00000300"},{0,0,U"\U00001F21\U00000300"},{0,0,U"\U00001F20\U00000301"},{0,0,U"\U00001F21\U00000301"},{0,0,U"\U00001F20\U00000342"},{0,0,U"\U00001F21\U00000342"},{0,0,U"\U00000397\U00000313"},{0,0,U"\U00000397\U00000314"},{0,0,U"\U00001F28\U00000300"},{0,0,U"\U00001F29\U00000300"},{0,0,U"\U00001F28\U00000301"},{0,0,U"\U00001F29\U00000301"},{0,0,U"\U00001F28\U00000342"},{0,0,U"\U00001F29\U00000342"},{0,0,U"\U000003B9\U00000313"},{0,0,U"\U000003B9\U00000314"},{0,0,U"\U00001F30\U00000300"},{0,0,U"\U00001F31\U00000300"},{0,0,U"\U00001F30\U00000301"},{0,0,U"\U00001F31\U00000301"},{0,0,U"\U00001F30\U00000342"},{0,0,U"\U00001F31\U00000342"},{0,0,U"\U00000399\U00000313"},{0,0,U"\U00000399\U00000314"},{0,0,U"\U00001F38\U00000300"},{0,0,U"\U00001F39\U00000300"},{0,0,U"\U00001F38\U00000301"},{0,0,U"\U00001F39\U00000301"},{0,0,U"\U00001F38\U00000342"},{0,0,U"\U00001F39\U00000342"},{0,0,U"\U000003BF\U00000313"},{0,0,U"\U000003BF\U00000314"},{0,0,U"\U00001F40\U00000300"},{0,0,U"\U00001F41\U00000300"},{0,0,U"\U00001F40\U00000301"},{0,0,U"\U00001F41\U00000301"},{0,0,0},{0,0,0},{0,0,U"\U0000039F\U00000313"},{0,0,U"\U0000039F\U00000314"},{0,0,U"\U00001F48\U00000300"},{0,0,U"\U00001F49\U00000300"},{0,0,U"\U00001F48\U00000301"},{0,0,U"\U00001F49\U00000301"},{0,0,0},{0,0,0},{0,0,U"\U000003C5\U00000313"},{0,0,U"\U000003C5\U00000314"},{0,0,U"\U00001F50\U00000300"},{0,0,U"\U00001F51\U00000300"},{0,0,U"\U00001F50\U00000301"},{0,0,U"\U00001F51\U00000301"},{0,0,U"\U00001F50\U00000342"},{0,0,U"\U00001F51\U00000342"},{0,0,0},{0,0,U"\U000003A5\U00000314"},{0,0,0},{0,0,U"\U00001F59\U00000300"},{0,0,0},{0,0,U"\U00001F59\U00000301"},{0,0,0},{0,0,U"\U00001F59\U00000342"},{0,0,U"\U000003C9\U00000313"},{0,0,U"\U000003C9\U00000314"},{0,0,U"\U00001F60\U00000300"},{0,0,U"\U00001F61\U00000300"},{0,0,U"\U00001F60\U00000301"},{0,0,U"\U00001F61\U00000301"},{0,0,U"\U00001F60\U00000342"},{0,0,U"\U00001F61\U00000342"},{0,0,U"\U000003A9\U00000313"},{0,0,U"\U000003A9\U00000314"},{0,0,U"\U00001F68\U00000300"},{0,0,U"\U00001F69\U00000300"},{0,0,U"\U00001F68\U00000301"},{0,0,U"\U00001F69\U00000301"},{0,0,U"\U00001F68\U00000342"},{0,0,U"\U00001F69\U00000342"},{0,0,U"\U000003B1\U00000300"},{0,0,U"\U000003AC"},{0,0,U"\U000003B5\U00000300"},{0,0,U"\U000003AD"},{0,0,U"\U000003B7\U00000300"},{0,0,U"\U000003AE"},{0,0,U"\U000003B9\U00000300"},{0,0,U"\U000003AF"},{0,0,U"\U000003BF\U00000300"},{0,0,U"\U000003CC"},{0,0,U"\U000003C5\U00000300"},{0,0,U"\U000003CD"},{0,0,U"\U000003C9\U00000300"},{0,0,U"\U000003CE"},{0,0,0},{0,0,0}, };
static const NormalizationProperties _63[] = { {0,0,U"\U00001F00\U00000345"},{0,0,U"\U00001F01\U00000345"},{0,0,U"\U00001F02\U00000345"},{0,0,U"\U00001F03\U00000345"},{0,0,U"\U00001F04\U00000345"},{0,0,U"\U00001F05\U00000345"},{0,0,U"\U00001F06\U00000345"},{0,0,U"\U00001F07\U00000345"},{0,0,U"\U00001F08\U00000345"},{0,0,U"\U00001F09\U00000345"},{0,0,U"\U00001F0A\U00000345"},{0,0,U"\U00001F0B\U00000345"},{0,0,U"\U00001F0C\U00000345"},{0,0,U"\U00001F0D\U00000345"},{0,0,U"\U00001F0E\U00000345"},{0,0,U"\U00001F0F\U00000345"},{0,0,U"\U00001F20\U00000345"},{0,0,U"\U00001F21\U00000345"},{0,0,U"\U00001F22\U00000345"},{0,0,U"\U00001F23\U00000345"},{0,0,U"\U00001F24\U00000345"},{0,0,U"\U00001F25\U00000345"},{0,0,U"\U00001F26\U00000345"},{0,0,U"\U00001F27\U00000345"},{0,0,U"\U00001F28\U00000345"},{0,0,U"\U00001F29\U00000345"},{0,0,U"\U00001F2A\U00000345"},{0,0,U"\U00001F2B\U00000345"},{0,0,U"\U00001F2C\U00000345"},{0,0,U"\U00001F2D\U00000345"},{0,0,U"\U00001F2E\U00000345"},{0,0,U"\U00001F2F\U00000345"},{0,0,U"\U00001F60\U00000345"},{0,0,U"\U00001F61\U00000345"},{0,0,U"\U00001F62\U00000345"},{0,0,U"\U00001F63\U00000345"},{0,0,U"\U00001F64\U00000345"},{0,0,U"\U00001F65\U00000345"},{0,0,U"\U00001F66\U00000345"},{0,0,U"\U00001F67\U00000345"},{0,0,U"\U00001F68\U00000345"},{0,0,U"\U00001F69\U00000345"},{0,0,U"\U00001F6A\U00000345"},{0,0,U"\U00001F6B\U00000345"},{0,0,U"\U00001F6C\U00000345"},{0,0,U"\U00001F6D\U00000345"},{0,0,U"\U00001F6E\U00000345"},{0,0,U"\U00001F6F\U00000345"},{0,0,U"\U000003B1\U00000306"},{0,0,U"\U000003B1\U00000304"},{0,0,U"\U00001F70\U00000345"},{0,0,U"\U000003B1\U00000345"},{0,0,U"\U000003AC\U00000345"},{0,0,0},{0,0,U"\U000003B1\U00000342"},{0,0,U"\U00001FB6\U00000345"},{0,0,U"\U00000391\U00000306"},{0,0,U"\U00000391\U00000304"},{0,0,U"\U00000391\U00000300"},{0,0,U"\U00000386"},{0,0,U"\U00000391\U00000345"},{0,"compat",U"\U00000020\U00000313"},{0,0,U"\U000003B9"},{0,"compat",U"\U00000020\U00000313"},{0,"compat",U"\U00000020\U00000342"},{0,0,U"\U000000A8\U00000342"},{0,0,U"\U00001F74\U00000345"},{0,0,U"\U000003B7\U00000345"},{0,0,U"\U000003AE\U00000345"},{0,0,0},{0,0,U"\U000003B7\U00000342"},{0,0,U"\U00001FC6\U00000345"},{0,0,U"\U00000395\U00000300"},{0,0,U"\U00000388"},{0,0,U"\U00000397\U00000300"},{0,0,U"\U00000389"},{0,0,U"\U00000397\U00000345"},{0,0,U"\U00001FBF\U00000300"},{0,0,U"\U00001FBF\U00000301"},{0,0,U"\U00001FBF\U00000342"},{0,0,U"\U000003B9\U00000306"},{0,0,U"\U000003B9\U00000304"},{0,0,U"\U000003CA\U00000300"},{0,0,U"\U00000390"},{0,0,0},{0,0,0},{0,0,U"\U000003B9\U00000342"},{0,0,U"\U000003CA\U00000342"},{0,0,U"\U00000399\U00000306"},{0,0,U"\U00000399\U00000304"},{0,0,U"\U00000399\U00000300"},{0,0,U"\U0000038A"},{0,0,0},{0,0,U"\U00001FFE\U00000300"},{0,0,U"\U00001FFE\U00000301"},{0,0,U"\U00001FFE\U00000342"},{0,0,U"\U000003C5\U00000306"},{0,0,U"\U000003C5\U00000304"},{0,0,U"\U000003CB\U00000300"},{0,0,U"\U000003B0"},{0,0,U"\U000003C1\U00000313"},{0,0,U"\U000003C1\U00000314"},{0,0,U"\U000003C5\U00000342"},{0,0,U"\U000003CB\U00000342"},{0,0,U"\U000003A5\U00000306"},{0,0,U"\U000003A5\U00000304"},{0,0,U"\U000003A5\U00000300"},{0,0,U"\U0000038E"},{0,0,U"\U000003A1\U00000314"},{0,0,U"\U000000A8\U00000300"},{0,0,U"\U00000385"},{0,0,U"\U00000060"},{0,0,0},{0,0,0},{0,0,U"\U00001F7C\U00000345"},{0,0,U"\U000003C9\U00000345"},{0,0,U"\U000003CE\U00000345"},{0,0,0},{0,0,U"\U000003C9\U00000342"},{0,0,U"\U00001FF6\U00000345"},{0,0,U"\U0000039F\U00000300"},{0,0,U"\U0000038C"},{0,0,U"\U000003A9\U00000300"},{0,0,U"\U0000038F"},{0,0,U"\U000003A9\U00000345"},{0,0,U"\U000000B4"},{0,"compat",U"\U00000020\U00000314"},{0,0,0}, };
static const NormalizationProperties _64[] = { {0,0,U"\U00002002"},{0,0,U"\U00002003"},{0,"compat",U"\U00000020"},{0,"compat",U"\U00000020"},{0,"compat",U"\U00000020"},{0,"compat",U"\U00000020"},{0,"compat",U"\U00000020"},{0,"noBreak",U"\U00000020"},{0,"compat",U"\U00000020"},{0,"compat",U"\U00000020"},{0,"compat",U"\U00000020"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"noBreak",U"\U00002010"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000020\U00000333"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U0000002E"},{0,"compat",U"\U0000002E\U0000002E"},{0,"compat",U"\U0000002E\U0000002E\U0000002E"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"noBreak",U"\U00000020"},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00002032\U00002032"},{0,"compat",U"\U00002032\U00002032\U00002032"},{0,0,0},{0,"compat",U"\U00002035\U00002035"},{0,"compat",U"\U00002035\U00002035\U00002035"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000021\U00000021"},{0,0,0},{0,"compat",U"\U00000020\U00000305"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U0000003F\U0000003F"},{0,"compat",U"\U0000003F\U00000021"},{0,"compat",U"\U00000021\U0000003F"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00002032\U00002032\U00002032\U00002032"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000020"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U00000030"},{0,"super",U"\U00000069"},{0,0,0},{0,0,0},{0,"super",U"\U00000034"},{0,"super",U"\U00000035"},{0,"super",U"\U00000036"},{0,"super",U"\U00000037"},{0,"super",U"\U00000038"},{0,"super",U"\U00000039"},{0,"super",U"\U0000002B"},{0,"super",U"\U00002212"},{0,"super",U"\U0000003D"},{0,"super",U"\U00000028"},{0,"super",U"\U00000029"},{0,"super",U"\U0000006E"}, };
static const NormalizationProperties _65[] = { {0,"sub",U"\U00000030"},{0,"sub",U"\U00000031"},{0,"sub",U"\U00000032"},{0,"sub",U"\U00000033"},{0,"sub",U"\U00000034"},{0,"sub",U"\U00000035"},{0,"sub",U"\U00000036"},{0,"sub",U"\U00000037"},{0,"sub",U"\U00000038"},{0,"sub",U"\U00000039"},{0,"sub",U"\U0000002B"},{0,"sub",U"\U00002212"},{0,"sub",U"\U0000003D"},{0,"sub",U"\U00000028"},{0,"sub",U"\U00000029"},{0,0,0},{0,"sub",U"\U00000061"},{0,"sub",U"\U00000065"},{0,"sub",U"\U0000006F"},{0,"sub",U"\U00000078"},{0,"sub",U"\U00000259"},{0,"sub",U"\U00000068"},{0,"sub",U"\U0000006B"},{0,"sub",U"\U0000006C"},{0,"sub",U"\U0000006D"},{0,"sub",U"\U0000006E"},{0,"sub",U"\U00000070"},{0,"sub",U"\U00000073"},{0,"sub",U"\U00000074"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000052\U00000073"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{1,0,0},{1,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{1,0,0},{1,0,0},{1,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{1,0,0},{1,0,0},{230,0,0},{220,0,0},{230,0,0},{1,0,0},{1,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _66[] = { {0,"compat",U"\U00000061\U0000002F\U00000063"},{0,"compat",U"\U00000061\U0000002F\U00000073"},{0,"font",U"\U00000043"},{0,"compat",U"\U000000B0\U00000043"},{0,0,0},{0,"compat",U"\U00000063\U0000002F\U0000006F"},{0,"compat",U"\U00000063\U0000002F\U00000075"},{0,"compat",U"\U00000190"},{0,0,0},{0,"compat",U"\U000000B0\U00000046"},{0,"font",U"\U00000067"},{0,"font",U"\U00000048"},{0,"font",U"\U00000048"},{0,"font",U"\U00000048"},{0,"font",U"\U00000068"},{0,"font",U"\U00000127"},{0,"font",U"\U00000049"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000006C"},{0,0,0},{0,"font",U"\U0000004E"},{0,"compat",U"\U0000004E\U0000006F"},{0,0,0},{0,0,0},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000052"},{0,"font",U"\U00000052"},{0,0,0},{0,0,0},{0,"super",U"\U00000053\U0000004D"},{0,"compat",U"\U00000054\U00000045\U0000004C"},{0,"super",U"\U00000054\U0000004D"},{0,0,0},{0,"font",U"\U0000005A"},{0,0,0},{0,0,U"\U000003A9"},{0,0,0},{0,"font",U"\U0000005A"},{0,0,0},{0,0,U"\U0000004B"},{0,0,U"\U000000C5"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,0,0},{0,"font",U"\U00000065"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,0,0},{0,"font",U"\U0000004D"},{0,"font",U"\U0000006F"},{0,"compat",U"\U000005D0"},{0,"compat",U"\U000005D1"},{0,"compat",U"\U000005D2"},{0,"compat",U"\U000005D3"},{0,"font",U"\U00000069"},{0,0,0},{0,"compat",U"\U00000046\U00000041\U00000058"},{0,"font",U"\U000003C0"},{0,"font",U"\U000003B3"},{0,"font",U"\U00000393"},{0,"font",U"\U000003A0"},{0,"font",U"\U00002211"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"font",U"\U00000044"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"fraction",U"\U00000031\U00002044\U00000037"},{0,"fraction",U"\U00000031\U00002044\U00000039"},{0,"fraction",U"\U00000031\U00002044\U00000031\U00000030"},{0,"fraction",U"\U00000031\U00002044\U00000033"},{0,"fraction",U"\U00000032\U00002044\U00000033"},{0,"fraction",U"\U00000031\U00002044\U00000035"},{0,"fraction",U"\U00000032\U00002044\U00000035"},{0,"fraction",U"\U00000033\U00002044\U00000035"},{0,"fraction",U"\U00000034\U00002044\U00000035"},{0,"fraction",U"\U00000031\U00002044\U00000036"},{0,"fraction",U"\U00000035\U00002044\U00000036"},{0,"fraction",U"\U00000031\U00002044\U00000038"},{0,"fraction",U"\U00000033\U00002044\U00000038"},{0,"fraction",U"\U00000035\U00002044\U00000038"},{0,"fraction",U"\U00000037\U00002044\U00000038"},{0,"fraction",U"\U00000031\U00002044"},{0,"compat",U"\U00000049"},{0,"compat",U"\U00000049\U00000049"},{0,"compat",U"\U00000049\U00000049\U00000049"},{0,"compat",U"\U00000049\U00000056"},{0,"compat",U"\U00000056"},{0,"compat",U"\U00000056\U00000049"},{0,"compat",U"\U00000056\U00000049\U00000049"},{0,"compat",U"\U00000056\U00000049\U00000049\U00000049"},{0,"compat",U"\U00000049\U00000058"},{0,"compat",U"\U00000058"},{0,"compat",U"\U00000058\U00000049"},{0,"compat",U"\U00000058\U00000049\U00000049"},{0,"compat",U"\U0000004C"},{0,"compat",U"\U00000043"},{0,"compat",U"\U00000044"},{0,"compat",U"\U0000004D"},{0,"compat",U"\U00000069"},{0,"compat",U"\U00000069\U00000069"},{0,"compat",U"\U00000069\U00000069\U00000069"},{0,"compat",U"\U00000069\U00000076"},{0,"compat",U"\U00000076"},{0,"compat",U"\U00000076\U00000069"},{0,"compat",U"\U00000076\U00000069\U00000069"},{0,"compat",U"\U00000076\U00000069\U00000069\U00000069"},{0,"compat",U"\U00000069\U00000078"},{0,"compat",U"\U00000078"},{0,"compat",U"\U00000078\U00000069"},{0,"compat",U"\U00000078\U00000069\U00000069"},{0,"compat",U"\U0000006C"},{0,"compat",U"\U00000063"},{0,"compat",U"\U00000064"},{0,"compat",U"\U0000006D"}, };
static const NormalizationProperties _67[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"fraction",U"\U00000030\U00002044\U00000033"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00002190\U00000338"},{0,0,U"\U00002192\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00002194\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000021D0\U00000338"},{0,0,U"\U000021D4\U00000338"},{0,0,U"\U000021D2\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _68[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00002203\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00002208\U00000338"},{0,0,0},{0,0,0},{0,0,U"\U0000220B\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00002223\U00000338"},{0,0,0},{0,0,U"\U00002225\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U0000222B\U0000222B"},{0,"compat",U"\U0000222B\U0000222B\U0000222B"},{0,0,0},{0,"compat",U"\U0000222E\U0000222E"},{0,"compat",U"\U0000222E\U0000222E\U0000222E"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0000223C\U00000338"},{0,0,0},{0,0,0},{0,0,U"\U00002243\U00000338"},{0,0,0},{0,0,0},{0,0,U"\U00002245\U00000338"},{0,0,0},{0,0,U"\U00002248\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0000003D\U00000338"},{0,0,0},{0,0,U"\U00002261\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0000224D\U00000338"},{0,0,U"\U0000003C\U00000338"},{0,0,U"\U0000003E\U00000338"},{0,0,U"\U00002264\U00000338"},{0,0,U"\U00002265\U00000338"},{0,0,0},{0,0,0},{0,0,U"\U00002272\U00000338"},{0,0,U"\U00002273\U00000338"},{0,0,0},{0,0,0},{0,0,U"\U00002276\U00000338"},{0,0,U"\U00002277\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _69[] = { {0,0,U"\U0000227A\U00000338"},{0,0,U"\U0000227B\U00000338"},{0,0,0},{0,0,0},{0,0,U"\U00002282\U00000338"},{0,0,U"\U00002283\U00000338"},{0,0,0},{0,0,0},{0,0,U"\U00002286\U00000338"},{0,0,U"\U00002287\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000022A2\U00000338"},{0,0,U"\U000022A8\U00000338"},{0,0,U"\U000022A9\U00000338"},{0,0,U"\U000022AB\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0000227C\U00000338"},{0,0,U"\U0000227D\U00000338"},{0,0,U"\U00002291\U00000338"},{0,0,U"\U00002292\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000022B2\U00000338"},{0,0,U"\U000022B3\U00000338"},{0,0,U"\U000022B4\U00000338"},{0,0,U"\U000022B5\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _70[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00003008"},{0,0,U"\U00003009"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _72[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"circle",U"\U00000031"},{0,"circle",U"\U00000032"},{0,"circle",U"\U00000033"},{0,"circle",U"\U00000034"},{0,"circle",U"\U00000035"},{0,"circle",U"\U00000036"},{0,"circle",U"\U00000037"},{0,"circle",U"\U00000038"},{0,"circle",U"\U00000039"},{0,"circle",U"\U00000031\U00000030"},{0,"circle",U"\U00000031\U00000031"},{0,"circle",U"\U00000031\U00000032"},{0,"circle",U"\U00000031\U00000033"},{0,"circle",U"\U00000031\U00000034"},{0,"circle",U"\U00000031\U00000035"},{0,"circle",U"\U00000031\U00000036"},{0,"circle",U"\U00000031\U00000037"},{0,"circle",U"\U00000031\U00000038"},{0,"circle",U"\U00000031\U00000039"},{0,"circle",U"\U00000032\U00000030"},{0,"compat",U"\U00000028\U00000031\U00000029"},{0,"compat",U"\U00000028\U00000032\U00000029"},{0,"compat",U"\U00000028\U00000033\U00000029"},{0,"compat",U"\U00000028\U00000034\U00000029"},{0,"compat",U"\U00000028\U00000035\U00000029"},{0,"compat",U"\U00000028\U00000036\U00000029"},{0,"compat",U"\U00000028\U00000037\U00000029"},{0,"compat",U"\U00000028\U00000038\U00000029"},{0,"compat",U"\U00000028\U00000039\U00000029"},{0,"compat",U"\U00000028\U00000031\U00000030\U00000029"},{0,"compat",U"\U00000028\U00000031\U00000031\U00000029"},{0,"compat",U"\U00000028\U00000031\U00000032\U00000029"}, };
static const NormalizationProperties _73[] = { {0,"compat",U"\U00000028\U00000031\U00000033\U00000029"},{0,"compat",U"\U00000028\U00000031\U00000034\U00000029"},{0,"compat",U"\U00000028\U00000031\U00000035\U00000029"},{0,"compat",U"\U00000028\U00000031\U00000036\U00000029"},{0,"compat",U"\U00000028\U00000031\U00000037\U00000029"},{0,"compat",U"\U00000028\U00000031\U00000038\U00000029"},{0,"compat",U"\U00000028\U00000031\U00000039\U00000029"},{0,"compat",U"\U00000028\U00000032\U00000030\U00000029"},{0,"compat",U"\U00000031\U0000002E"},{0,"compat",U"\U00000032\U0000002E"},{0,"compat",U"\U00000033\U0000002E"},{0,"compat",U"\U00000034\U0000002E"},{0,"compat",U"\U00000035\U0000002E"},{0,"compat",U"\U00000036\U0000002E"},{0,"compat",U"\U00000037\U0000002E"},{0,"compat",U"\U00000038\U0000002E"},{0,"compat",U"\U00000039\U0000002E"},{0,"compat",U"\U00000031\U00000030\U0000002E"},{0,"compat",U"\U00000031\U00000031\U0000002E"},{0,"compat",U"\U00000031\U00000032\U0000002E"},{0,"compat",U"\U00000031\U00000033\U0000002E"},{0,"compat",U"\U00000031\U00000034\U0000002E"},{0,"compat",U"\U00000031\U00000035\U0000002E"},{0,"compat",U"\U00000031\U00000036\U0000002E"},{0,"compat",U"\U00000031\U00000037\U0000002E"},{0,"compat",U"\U00000031\U00000038\U0000002E"},{0,"compat",U"\U00000031\U00000039\U0000002E"},{0,"compat",U"\U00000032\U00000030\U0000002E"},{0,"compat",U"\U00000028\U00000061\U00000029"},{0,"compat",U"\U00000028\U00000062\U00000029"},{0,"compat",U"\U00000028\U00000063\U00000029"},{0,"compat",U"\U00000028\U00000064\U00000029"},{0,"compat",U"\U00000028\U00000065\U00000029"},{0,"compat",U"\U00000028\U00000066\U00000029"},{0,"compat",U"\U00000028\U00000067\U00000029"},{0,"compat",U"\U00000028\U00000068\U00000029"},{0,"compat",U"\U00000028\U00000069\U00000029"},{0,"compat",U"\U00000028\U0000006A\U00000029"},{0,"compat",U"\U00000028\U0000006B\U00000029"},{0,"compat",U"\U00000028\U0000006C\U00000029"},{0,"compat",U"\U00000028\U0000006D\U00000029"},{0,"compat",U"\U00000028\U0000006E\U00000029"},{0,"compat",U"\U00000028\U0000006F\U00000029"},{0,"compat",U"\U00000028\U00000070\U00000029"},{0,"compat",U"\U00000028\U00000071\U00000029"},{0,"compat",U"\U00000028\U00000072\U00000029"},{0,"compat",U"\U00000028\U00000073\U00000029"},{0,"compat",U"\U00000028\U00000074\U00000029"},{0,"compat",U"\U00000028\U00000075\U00000029"},{0,"compat",U"\U00000028\U00000076\U00000029"},{0,"compat",U"\U00000028\U00000077\U00000029"},{0,"compat",U"\U00000028\U00000078\U00000029"},{0,"compat",U"\U00000028\U00000079\U00000029"},{0,"compat",U"\U00000028\U0000007A\U00000029"},{0,"circle",U"\U00000041"},{0,"circle",U"\U00000042"},{0,"circle",U"\U00000043"},{0,"circle",U"\U00000044"},{0,"circle",U"\U00000045"},{0,"circle",U"\U00000046"},{0,"circle",U"\U00000047"},{0,"circle",U"\U00000048"},{0,"circle",U"\U00000049"},{0,"circle",U"\U0000004A"},{0,"circle",U"\U0000004B"},{0,"circle",U"\U0000004C"},{0,"circle",U"\U0000004D"},{0,"circle",U"\U0000004E"},{0,"circle",U"\U0000004F"},{0,"circle",U"\U00000050"},{0,"circle",U"\U00000051"},{0,"circle",U"\U00000052"},{0,"circle",U"\U00000053"},{0,"circle",U"\U00000054"},{0,"circle",U"\U00000055"},{0,"circle",U"\U00000056"},{0,"circle",U"\U00000057"},{0,"circle",U"\U00000058"},{0,"circle",U"\U00000059"},{0,"circle",U"\U0000005A"},{0,"circle",U"\U00000061"},{0,"circle",U"\U00000062"},{0,"circle",U"\U00000063"},{0,"circle",U"\U00000064"},{0,"circle",U"\U00000065"},{0,"circle",U"\U00000066"},{0,"circle",U"\U00000067"},{0,"circle",U"\U00000068"},{0,"circle",U"\U00000069"},{0,"circle",U"\U0000006A"},{0,"circle",U"\U0000006B"},{0,"circle",U"\U0000006C"},{0,"circle",U"\U0000006D"},{0,"circle",U"\U0000006E"},{0,"circle",U"\U0000006F"},{0,"circle",U"\U00000070"},{0,"circle",U"\U00000071"},{0,"circle",U"\U00000072"},{0,"circle",U"\U00000073"},{0,"circle",U"\U00000074"},{0,"circle",U"\U00000075"},{0,"circle",U"\U00000076"},{0,"circle",U"\U00000077"},{0,"circle",U"\U00000078"},{0,"circle",U"\U00000079"},{0,"circle",U"\U0000007A"},{0,"circle",U"\U00000030"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _84[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U0000222B\U0000222B\U0000222B\U0000222B"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U0000003A\U0000003A\U0000003D"},{0,"compat",U"\U0000003D\U0000003D"},{0,"compat",U"\U0000003D\U0000003D\U0000003D"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _85[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00002ADD\U00000338"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _88[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"sub",U"\U0000006A"},{0,"super",U"\U00000056"},{0,0,0},{0,0,0}, };
static const NormalizationProperties _89[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _90[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U00002D61"},{0,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0}, };
static const NormalizationProperties _91[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0}, };
static const NormalizationProperties _93[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00006BCD"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00009F9F"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _94[] = { {0,"compat",U"\U00004E00"},{0,"compat",U"\U00004E28"},{0,"compat",U"\U00004E36"},{0,"compat",U"\U00004E3F"},{0,"compat",U"\U00004E59"},{0,"compat",U"\U00004E85"},{0,"compat",U"\U00004E8C"},{0,"compat",U"\U00004EA0"},{0,"compat",U"\U00004EBA"},{0,"compat",U"\U0000513F"},{0,"compat",U"\U00005165"},{0,"compat",U"\U0000516B"},{0,"compat",U"\U00005182"},{0,"compat",U"\U00005196"},{0,"compat",U"\U000051AB"},{0,"compat",U"\U000051E0"},{0,"compat",U"\U000051F5"},{0,"compat",U"\U00005200"},{0,"compat",U"\U0000529B"},{0,"compat",U"\U000052F9"},{0,"compat",U"\U00005315"},{0,"compat",U"\U0000531A"},{0,"compat",U"\U00005338"},{0,"compat",U"\U00005341"},{0,"compat",U"\U0000535C"},{0,"compat",U"\U00005369"},{0,"compat",U"\U00005382"},{0,"compat",U"\U000053B6"},{0,"compat",U"\U000053C8"},{0,"compat",U"\U000053E3"},{0,"compat",U"\U000056D7"},{0,"compat",U"\U0000571F"},{0,"compat",U"\U000058EB"},{0,"compat",U"\U00005902"},{0,"compat",U"\U0000590A"},{0,"compat",U"\U00005915"},{0,"compat",U"\U00005927"},{0,"compat",U"\U00005973"},{0,"compat",U"\U00005B50"},{0,"compat",U"\U00005B80"},{0,"compat",U"\U00005BF8"},{0,"compat",U"\U00005C0F"},{0,"compat",U"\U00005C22"},{0,"compat",U"\U00005C38"},{0,"compat",U"\U00005C6E"},{0,"compat",U"\U00005C71"},{0,"compat",U"\U00005DDB"},{0,"compat",U"\U00005DE5"},{0,"compat",U"\U00005DF1"},{0,"compat",U"\U00005DFE"},{0,"compat",U"\U00005E72"},{0,"compat",U"\U00005E7A"},{0,"compat",U"\U00005E7F"},{0,"compat",U"\U00005EF4"},{0,"compat",U"\U00005EFE"},{0,"compat",U"\U00005F0B"},{0,"compat",U"\U00005F13"},{0,"compat",U"\U00005F50"},{0,"compat",U"\U00005F61"},{0,"compat",U"\U00005F73"},{0,"compat",U"\U00005FC3"},{0,"compat",U"\U00006208"},{0,"compat",U"\U00006236"},{0,"compat",U"\U0000624B"},{0,"compat",U"\U0000652F"},{0,"compat",U"\U00006534"},{0,"compat",U"\U00006587"},{0,"compat",U"\U00006597"},{0,"compat",U"\U000065A4"},{0,"compat",U"\U000065B9"},{0,"compat",U"\U000065E0"},{0,"compat",U"\U000065E5"},{0,"compat",U"\U000066F0"},{0,"compat",U"\U00006708"},{0,"compat",U"\U00006728"},{0,"compat",U"\U00006B20"},{0,"compat",U"\U00006B62"},{0,"compat",U"\U00006B79"},{0,"compat",U"\U00006BB3"},{0,"compat",U"\U00006BCB"},{0,"compat",U"\U00006BD4"},{0,"compat",U"\U00006BDB"},{0,"compat",U"\U00006C0F"},{0,"compat",U"\U00006C14"},{0,"compat",U"\U00006C34"},{0,"compat",U"\U0000706B"},{0,"compat",U"\U0000722A"},{0,"compat",U"\U00007236"},{0,"compat",U"\U0000723B"},{0,"compat",U"\U0000723F"},{0,"compat",U"\U00007247"},{0,"compat",U"\U00007259"},{0,"compat",U"\U0000725B"},{0,"compat",U"\U000072AC"},{0,"compat",U"\U00007384"},{0,"compat",U"\U00007389"},{0,"compat",U"\U000074DC"},{0,"compat",U"\U000074E6"},{0,"compat",U"\U00007518"},{0,"compat",U"\U0000751F"},{0,"compat",U"\U00007528"},{0,"compat",U"\U00007530"},{0,"compat",U"\U0000758B"},{0,"compat",U"\U00007592"},{0,"compat",U"\U00007676"},{0,"compat",U"\U0000767D"},{0,"compat",U"\U000076AE"},{0,"compat",U"\U000076BF"},{0,"compat",U"\U000076EE"},{0,"compat",U"\U000077DB"},{0,"compat",U"\U000077E2"},{0,"compat",U"\U000077F3"},{0,"compat",U"\U0000793A"},{0,"compat",U"\U000079B8"},{0,"compat",U"\U000079BE"},{0,"compat",U"\U00007A74"},{0,"compat",U"\U00007ACB"},{0,"compat",U"\U00007AF9"},{0,"compat",U"\U00007C73"},{0,"compat",U"\U00007CF8"},{0,"compat",U"\U00007F36"},{0,"compat",U"\U00007F51"},{0,"compat",U"\U00007F8A"},{0,"compat",U"\U00007FBD"},{0,"compat",U"\U00008001"},{0,"compat",U"\U0000800C"},{0,"compat",U"\U00008012"},{0,"compat",U"\U00008033"}, };
static const NormalizationProperties _95[] = { {0,"compat",U"\U0000807F"},{0,"compat",U"\U00008089"},{0,"compat",U"\U000081E3"},{0,"compat",U"\U000081EA"},{0,"compat",U"\U000081F3"},{0,"compat",U"\U000081FC"},{0,"compat",U"\U0000820C"},{0,"compat",U"\U0000821B"},{0,"compat",U"\U0000821F"},{0,"compat",U"\U0000826E"},{0,"compat",U"\U00008272"},{0,"compat",U"\U00008278"},{0,"compat",U"\U0000864D"},{0,"compat",U"\U0000866B"},{0,"compat",U"\U00008840"},{0,"compat",U"\U0000884C"},{0,"compat",U"\U00008863"},{0,"compat",U"\U0000897E"},{0,"compat",U"\U0000898B"},{0,"compat",U"\U000089D2"},{0,"compat",U"\U00008A00"},{0,"compat",U"\U00008C37"},{0,"compat",U"\U00008C46"},{0,"compat",U"\U00008C55"},{0,"compat",U"\U00008C78"},{0,"compat",U"\U00008C9D"},{0,"compat",U"\U00008D64"},{0,"compat",U"\U00008D70"},{0,"compat",U"\U00008DB3"},{0,"compat",U"\U00008EAB"},{0,"compat",U"\U00008ECA"},{0,"compat",U"\U00008F9B"},{0,"compat",U"\U00008FB0"},{0,"compat",U"\U00008FB5"},{0,"compat",U"\U00009091"},{0,"compat",U"\U00009149"},{0,"compat",U"\U000091C6"},{0,"compat",U"\U000091CC"},{0,"compat",U"\U000091D1"},{0,"compat",U"\U00009577"},{0,"compat",U"\U00009580"},{0,"compat",U"\U0000961C"},{0,"compat",U"\U000096B6"},{0,"compat",U"\U000096B9"},{0,"compat",U"\U000096E8"},{0,"compat",U"\U00009751"},{0,"compat",U"\U0000975E"},{0,"compat",U"\U00009762"},{0,"compat",U"\U00009769"},{0,"compat",U"\U000097CB"},{0,"compat",U"\U000097ED"},{0,"compat",U"\U000097F3"},{0,"compat",U"\U00009801"},{0,"compat",U"\U000098A8"},{0,"compat",U"\U000098DB"},{0,"compat",U"\U000098DF"},{0,"compat",U"\U00009996"},{0,"compat",U"\U00009999"},{0,"compat",U"\U000099AC"},{0,"compat",U"\U00009AA8"},{0,"compat",U"\U00009AD8"},{0,"compat",U"\U00009ADF"},{0,"compat",U"\U00009B25"},{0,"compat",U"\U00009B2F"},{0,"compat",U"\U00009B32"},{0,"compat",U"\U00009B3C"},{0,"compat",U"\U00009B5A"},{0,"compat",U"\U00009CE5"},{0,"compat",U"\U00009E75"},{0,"compat",U"\U00009E7F"},{0,"compat",U"\U00009EA5"},{0,"compat",U"\U00009EBB"},{0,"compat",U"\U00009EC3"},{0,"compat",U"\U00009ECD"},{0,"compat",U"\U00009ED1"},{0,"compat",U"\U00009EF9"},{0,"compat",U"\U00009EFD"},{0,"compat",U"\U00009F0E"},{0,"compat",U"\U00009F13"},{0,"compat",U"\U00009F20"},{0,"compat",U"\U00009F3B"},{0,"compat",U"\U00009F4A"},{0,"compat",U"\U00009F52"},{0,"compat",U"\U00009F8D"},{0,"compat",U"\U00009F9C"},{0,"compat",U"\U00009FA0"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _96[] = { {0,"wide",U"\U00000020"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{218,0,0},{228,0,0},{232,0,0},{222,0,0},{224,0,0},{224,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00003012"},{0,0,0},{0,"compat",U"\U00005341"},{0,"compat",U"\U00005344"},{0,"compat",U"\U00005345"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0000304B\U00003099"},{0,0,0},{0,0,U"\U0000304D\U00003099"},{0,0,0},{0,0,U"\U0000304F\U00003099"},{0,0,0},{0,0,U"\U00003051\U00003099"},{0,0,0},{0,0,U"\U00003053\U00003099"},{0,0,0},{0,0,U"\U00003055\U00003099"},{0,0,0},{0,0,U"\U00003057\U00003099"},{0,0,0},{0,0,U"\U00003059\U00003099"},{0,0,0},{0,0,U"\U0000305B\U00003099"},{0,0,0},{0,0,U"\U0000305D\U00003099"},{0,0,0},{0,0,U"\U0000305F\U00003099"},{0,0,0},{0,0,U"\U00003061\U00003099"},{0,0,0},{0,0,0},{0,0,U"\U00003064\U00003099"},{0,0,0},{0,0,U"\U00003066\U00003099"},{0,0,0},{0,0,U"\U00003068\U00003099"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0000306F\U00003099"},{0,0,U"\U0000306F\U0000309A"},{0,0,0},{0,0,U"\U00003072\U00003099"},{0,0,U"\U00003072\U0000309A"},{0,0,0},{0,0,U"\U00003075\U00003099"},{0,0,U"\U00003075\U0000309A"},{0,0,0},{0,0,U"\U00003078\U00003099"},{0,0,U"\U00003078\U0000309A"},{0,0,0},{0,0,U"\U0000307B\U00003099"},{0,0,U"\U0000307B\U0000309A"},{0,0,0},{0,0,0}, };
static const NormalizationProperties _97[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00003046\U00003099"},{0,0,0},{0,0,0},{8,0,0},{8,0,0},{8,0,0},{8,0,0},{0,"compat",U"\U00000020\U00003099"},{0,"compat",U"\U00000020\U0000309A"},{0,0,0},{0,0,U"\U0000309D\U00003099"},{0,"vertical",U"\U00003088\U0000308A"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000030AB\U00003099"},{0,0,0},{0,0,U"\U000030AD\U00003099"},{0,0,0},{0,0,U"\U000030AF\U00003099"},{0,0,0},{0,0,U"\U000030B1\U00003099"},{0,0,0},{0,0,U"\U000030B3\U00003099"},{0,0,0},{0,0,U"\U000030B5\U00003099"},{0,0,0},{0,0,U"\U000030B7\U00003099"},{0,0,0},{0,0,U"\U000030B9\U00003099"},{0,0,0},{0,0,U"\U000030BB\U00003099"},{0,0,0},{0,0,U"\U000030BD\U00003099"},{0,0,0},{0,0,U"\U000030BF\U00003099"},{0,0,0},{0,0,U"\U000030C1\U00003099"},{0,0,0},{0,0,0},{0,0,U"\U000030C4\U00003099"},{0,0,0},{0,0,U"\U000030C6\U00003099"},{0,0,0},{0,0,U"\U000030C8\U00003099"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000030CF\U00003099"},{0,0,U"\U000030CF\U0000309A"},{0,0,0},{0,0,U"\U000030D2\U00003099"},{0,0,U"\U000030D2\U0000309A"},{0,0,0},{0,0,U"\U000030D5\U00003099"},{0,0,U"\U000030D5\U0000309A"},{0,0,0},{0,0,U"\U000030D8\U00003099"},{0,0,U"\U000030D8\U0000309A"},{0,0,0},{0,0,U"\U000030DB\U00003099"},{0,0,U"\U000030DB\U0000309A"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000030A6\U00003099"},{0,0,0},{0,0,0},{0,0,U"\U000030EF\U00003099"},{0,0,U"\U000030F0\U00003099"},{0,0,U"\U000030F1\U00003099"},{0,0,U"\U000030F2\U00003099"},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000030FD\U00003099"},{0,"vertical",U"\U000030B3\U000030C8"}, };
static const NormalizationProperties _98[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00001100"},{0,"compat",U"\U00001101"},{0,"compat",U"\U000011AA"},{0,"compat",U"\U00001102"},{0,"compat",U"\U000011AC"},{0,"compat",U"\U000011AD"},{0,"compat",U"\U00001103"},{0,"compat",U"\U00001104"},{0,"compat",U"\U00001105"},{0,"compat",U"\U000011B0"},{0,"compat",U"\U000011B1"},{0,"compat",U"\U000011B2"},{0,"compat",U"\U000011B3"},{0,"compat",U"\U000011B4"},{0,"compat",U"\U000011B5"},{0,"compat",U"\U0000111A"},{0,"compat",U"\U00001106"},{0,"compat",U"\U00001107"},{0,"compat",U"\U00001108"},{0,"compat",U"\U00001121"},{0,"compat",U"\U00001109"},{0,"compat",U"\U0000110A"},{0,"compat",U"\U0000110B"},{0,"compat",U"\U0000110C"},{0,"compat",U"\U0000110D"},{0,"compat",U"\U0000110E"},{0,"compat",U"\U0000110F"},{0,"compat",U"\U00001110"},{0,"compat",U"\U00001111"},{0,"compat",U"\U00001112"},{0,"compat",U"\U00001161"},{0,"compat",U"\U00001162"},{0,"compat",U"\U00001163"},{0,"compat",U"\U00001164"},{0,"compat",U"\U00001165"},{0,"compat",U"\U00001166"},{0,"compat",U"\U00001167"},{0,"compat",U"\U00001168"},{0,"compat",U"\U00001169"},{0,"compat",U"\U0000116A"},{0,"compat",U"\U0000116B"},{0,"compat",U"\U0000116C"},{0,"compat",U"\U0000116D"},{0,"compat",U"\U0000116E"},{0,"compat",U"\U0000116F"},{0,"compat",U"\U00001170"},{0,"compat",U"\U00001171"},{0,"compat",U"\U00001172"},{0,"compat",U"\U00001173"},{0,"compat",U"\U00001174"},{0,"compat",U"\U00001175"},{0,"compat",U"\U00001160"},{0,"compat",U"\U00001114"},{0,"compat",U"\U00001115"},{0,"compat",U"\U000011C7"},{0,"compat",U"\U000011C8"},{0,"compat",U"\U000011CC"},{0,"compat",U"\U000011CE"},{0,"compat",U"\U000011D3"},{0,"compat",U"\U000011D7"},{0,"compat",U"\U000011D9"},{0,"compat",U"\U0000111C"},{0,"compat",U"\U000011DD"},{0,"compat",U"\U000011DF"},{0,"compat",U"\U0000111D"},{0,"compat",U"\U0000111E"},{0,"compat",U"\U00001120"},{0,"compat",U"\U00001122"},{0,"compat",U"\U00001123"},{0,"compat",U"\U00001127"},{0,"compat",U"\U00001129"},{0,"compat",U"\U0000112B"},{0,"compat",U"\U0000112C"},{0,"compat",U"\U0000112D"},{0,"compat",U"\U0000112E"},{0,"compat",U"\U0000112F"},{0,"compat",U"\U00001132"},{0,"compat",U"\U00001136"},{0,"compat",U"\U00001140"}, };
static const NormalizationProperties _99[] = { {0,"compat",U"\U00001147"},{0,"compat",U"\U0000114C"},{0,"compat",U"\U000011F1"},{0,"compat",U"\U000011F2"},{0,"compat",U"\U00001157"},{0,"compat",U"\U00001158"},{0,"compat",U"\U00001159"},{0,"compat",U"\U00001184"},{0,"compat",U"\U00001185"},{0,"compat",U"\U00001188"},{0,"compat",U"\U00001191"},{0,"compat",U"\U00001192"},{0,"compat",U"\U00001194"},{0,"compat",U"\U0000119E"},{0,"compat",U"\U000011A1"},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U00004E00"},{0,"super",U"\U00004E8C"},{0,"super",U"\U00004E09"},{0,"super",U"\U000056DB"},{0,"super",U"\U00004E0A"},{0,"super",U"\U00004E2D"},{0,"super",U"\U00004E0B"},{0,"super",U"\U00007532"},{0,"super",U"\U00004E59"},{0,"super",U"\U00004E19"},{0,"super",U"\U00004E01"},{0,"super",U"\U00005929"},{0,"super",U"\U00005730"},{0,"super",U"\U00004EBA"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _100[] = { {0,"compat",U"\U00000028\U00001100\U00000029"},{0,"compat",U"\U00000028\U00001102\U00000029"},{0,"compat",U"\U00000028\U00001103\U00000029"},{0,"compat",U"\U00000028\U00001105\U00000029"},{0,"compat",U"\U00000028\U00001106\U00000029"},{0,"compat",U"\U00000028\U00001107\U00000029"},{0,"compat",U"\U00000028\U00001109\U00000029"},{0,"compat",U"\U00000028\U0000110B\U00000029"},{0,"compat",U"\U00000028\U0000110C\U00000029"},{0,"compat",U"\U00000028\U0000110E\U00000029"},{0,"compat",U"\U00000028\U0000110F\U00000029"},{0,"compat",U"\U00000028\U00001110\U00000029"},{0,"compat",U"\U00000028\U00001111\U00000029"},{0,"compat",U"\U00000028\U00001112\U00000029"},{0,"compat",U"\U00000028\U00001100\U00001161\U00000029"},{0,"compat",U"\U00000028\U00001102\U00001161\U00000029"},{0,"compat",U"\U00000028\U00001103\U00001161\U00000029"},{0,"compat",U"\U00000028\U00001105\U00001161\U00000029"},{0,"compat",U"\U00000028\U00001106\U00001161\U00000029"},{0,"compat",U"\U00000028\U00001107\U00001161\U00000029"},{0,"compat",U"\U00000028\U00001109\U00001161\U00000029"},{0,"compat",U"\U00000028\U0000110B\U00001161\U00000029"},{0,"compat",U"\U00000028\U0000110C\U00001161\U00000029"},{0,"compat",U"\U00000028\U0000110E\U00001161\U00000029"},{0,"compat",U"\U00000028\U0000110F\U00001161\U00000029"},{0,"compat",U"\U00000028\U00001110\U00001161\U00000029"},{0,"compat",U"\U00000028\U00001111\U00001161\U00000029"},{0,"compat",U"\U00000028\U00001112\U00001161\U00000029"},{0,"compat",U"\U00000028\U0000110C\U0000116E\U00000029"},{0,"compat",U"\U00000028\U0000110B\U00001169\U0000110C\U00001165\U000011AB\U00000029"},{0,"compat",U"\U00000028\U0000110B\U00001169\U00001112\U0000116E\U00000029"},{0,0,0},{0,"compat",U"\U00000028\U00004E00\U00000029"},{0,"compat",U"\U00000028\U00004E8C\U00000029"},{0,"compat",U"\U00000028\U00004E09\U00000029"},{0,"compat",U"\U00000028\U000056DB\U00000029"},{0,"compat",U"\U00000028\U00004E94\U00000029"},{0,"compat",U"\U00000028\U0000516D\U00000029"},{0,"compat",U"\U00000028\U00004E03\U00000029"},{0,"compat",U"\U00000028\U0000516B\U00000029"},{0,"compat",U"\U00000028\U00004E5D\U00000029"},{0,"compat",U"\U00000028\U00005341\U00000029"},{0,"compat",U"\U00000028\U00006708\U00000029"},{0,"compat",U"\U00000028\U0000706B\U00000029"},{0,"compat",U"\U00000028\U00006C34\U00000029"},{0,"compat",U"\U00000028\U00006728\U00000029"},{0,"compat",U"\U00000028\U000091D1\U00000029"},{0,"compat",U"\U00000028\U0000571F\U00000029"},{0,"compat",U"\U00000028\U000065E5\U00000029"},{0,"compat",U"\U00000028\U0000682A\U00000029"},{0,"compat",U"\U00000028\U00006709\U00000029"},{0,"compat",U"\U00000028\U0000793E\U00000029"},{0,"compat",U"\U00000028\U0000540D\U00000029"},{0,"compat",U"\U00000028\U00007279\U00000029"},{0,"compat",U"\U00000028\U00008CA1\U00000029"},{0,"compat",U"\U00000028\U0000795D\U00000029"},{0,"compat",U"\U00000028\U000052B4\U00000029"},{0,"compat",U"\U00000028\U00004EE3\U00000029"},{0,"compat",U"\U00000028\U0000547C\U00000029"},{0,"compat",U"\U00000028\U00005B66\U00000029"},{0,"compat",U"\U00000028\U000076E3\U00000029"},{0,"compat",U"\U00000028\U00004F01\U00000029"},{0,"compat",U"\U00000028\U00008CC7\U00000029"},{0,"compat",U"\U00000028\U00005354\U00000029"},{0,"compat",U"\U00000028\U0000796D\U00000029"},{0,"compat",U"\U00000028\U00004F11\U00000029"},{0,"compat",U"\U00000028\U000081EA\U00000029"},{0,"compat",U"\U00000028\U000081F3\U00000029"},{0,"circle",U"\U0000554F"},{0,"circle",U"\U00005E7C"},{0,"circle",U"\U00006587"},{0,"circle",U"\U00007B8F"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"square",U"\U00000050\U00000054\U00000045"},{0,"circle",U"\U00000032\U00000031"},{0,"circle",U"\U00000032\U00000032"},{0,"circle",U"\U00000032\U00000033"},{0,"circle",U"\U00000032\U00000034"},{0,"circle",U"\U00000032\U00000035"},{0,"circle",U"\U00000032\U00000036"},{0,"circle",U"\U00000032\U00000037"},{0,"circle",U"\U00000032\U00000038"},{0,"circle",U"\U00000032\U00000039"},{0,"circle",U"\U00000033\U00000030"},{0,"circle",U"\U00000033\U00000031"},{0,"circle",U"\U00000033\U00000032"},{0,"circle",U"\U00000033\U00000033"},{0,"circle",U"\U00000033\U00000034"},{0,"circle",U"\U00000033\U00000035"},{0,"circle",U"\U00001100"},{0,"circle",U"\U00001102"},{0,"circle",U"\U00001103"},{0,"circle",U"\U00001105"},{0,"circle",U"\U00001106"},{0,"circle",U"\U00001107"},{0,"circle",U"\U00001109"},{0,"circle",U"\U0000110B"},{0,"circle",U"\U0000110C"},{0,"circle",U"\U0000110E"},{0,"circle",U"\U0000110F"},{0,"circle",U"\U00001110"},{0,"circle",U"\U00001111"},{0,"circle",U"\U00001112"},{0,"circle",U"\U00001100\U00001161"},{0,"circle",U"\U00001102\U00001161"},{0,"circle",U"\U00001103\U00001161"},{0,"circle",U"\U00001105\U00001161"},{0,"circle",U"\U00001106\U00001161"},{0,"circle",U"\U00001107\U00001161"},{0,"circle",U"\U00001109\U00001161"},{0,"circle",U"\U0000110B\U00001161"},{0,"circle",U"\U0000110C\U00001161"},{0,"circle",U"\U0000110E\U00001161"},{0,"circle",U"\U0000110F\U00001161"},{0,"circle",U"\U00001110\U00001161"},{0,"circle",U"\U00001111\U00001161"},{0,"circle",U"\U00001112\U00001161"},{0,"circle",U"\U0000110E\U00001161\U000011B7\U00001100\U00001169"},{0,"circle",U"\U0000110C\U0000116E\U0000110B\U00001174"},{0,"circle",U"\U0000110B\U0000116E"},{0,0,0}, };
static const NormalizationProperties _101[] = { {0,"circle",U"\U00004E00"},{0,"circle",U"\U00004E8C"},{0,"circle",U"\U00004E09"},{0,"circle",U"\U000056DB"},{0,"circle",U"\U00004E94"},{0,"circle",U"\U0000516D"},{0,"circle",U"\U00004E03"},{0,"circle",U"\U0000516B"},{0,"circle",U"\U00004E5D"},{0,"circle",U"\U00005341"},{0,"circle",U"\U00006708"},{0,"circle",U"\U0000706B"},{0,"circle",U"\U00006C34"},{0,"circle",U"\U00006728"},{0,"circle",U"\U000091D1"},{0,"circle",U"\U0000571F"},{0,"circle",U"\U000065E5"},{0,"circle",U"\U0000682A"},{0,"circle",U"\U00006709"},{0,"circle",U"\U0000793E"},{0,"circle",U"\U0000540D"},{0,"circle",U"\U00007279"},{0,"circle",U"\U00008CA1"},{0,"circle",U"\U0000795D"},{0,"circle",U"\U000052B4"},{0,"circle",U"\U000079D8"},{0,"circle",U"\U00007537"},{0,"circle",U"\U00005973"},{0,"circle",U"\U00009069"},{0,"circle",U"\U0000512A"},{0,"circle",U"\U00005370"},{0,"circle",U"\U00006CE8"},{0,"circle",U"\U00009805"},{0,"circle",U"\U00004F11"},{0,"circle",U"\U00005199"},{0,"circle",U"\U00006B63"},{0,"circle",U"\U00004E0A"},{0,"circle",U"\U00004E2D"},{0,"circle",U"\U00004E0B"},{0,"circle",U"\U00005DE6"},{0,"circle",U"\U000053F3"},{0,"circle",U"\U0000533B"},{0,"circle",U"\U00005B97"},{0,"circle",U"\U00005B66"},{0,"circle",U"\U000076E3"},{0,"circle",U"\U00004F01"},{0,"circle",U"\U00008CC7"},{0,"circle",U"\U00005354"},{0,"circle",U"\U0000591C"},{0,"circle",U"\U00000033\U00000036"},{0,"circle",U"\U00000033\U00000037"},{0,"circle",U"\U00000033\U00000038"},{0,"circle",U"\U00000033\U00000039"},{0,"circle",U"\U00000034\U00000030"},{0,"circle",U"\U00000034\U00000031"},{0,"circle",U"\U00000034\U00000032"},{0,"circle",U"\U00000034\U00000033"},{0,"circle",U"\U00000034\U00000034"},{0,"circle",U"\U00000034\U00000035"},{0,"circle",U"\U00000034\U00000036"},{0,"circle",U"\U00000034\U00000037"},{0,"circle",U"\U00000034\U00000038"},{0,"circle",U"\U00000034\U00000039"},{0,"circle",U"\U00000035\U00000030"},{0,"compat",U"\U00000031\U00006708"},{0,"compat",U"\U00000032\U00006708"},{0,"compat",U"\U00000033\U00006708"},{0,"compat",U"\U00000034\U00006708"},{0,"compat",U"\U00000035\U00006708"},{0,"compat",U"\U00000036\U00006708"},{0,"compat",U"\U00000037\U00006708"},{0,"compat",U"\U00000038\U00006708"},{0,"compat",U"\U00000039\U00006708"},{0,"compat",U"\U00000031\U00000030\U00006708"},{0,"compat",U"\U00000031\U00000031\U00006708"},{0,"compat",U"\U00000031\U00000032\U00006708"},{0,"square",U"\U00000048\U00000067"},{0,"square",U"\U00000065\U00000072\U00000067"},{0,"square",U"\U00000065\U00000056"},{0,"square",U"\U0000004C\U00000054\U00000044"},{0,"circle",U"\U000030A2"},{0,"circle",U"\U000030A4"},{0,"circle",U"\U000030A6"},{0,"circle",U"\U000030A8"},{0,"circle",U"\U000030AA"},{0,"circle",U"\U000030AB"},{0,"circle",U"\U000030AD"},{0,"circle",U"\U000030AF"},{0,"circle",U"\U000030B1"},{0,"circle",U"\U000030B3"},{0,"circle",U"\U000030B5"},{0,"circle",U"\U000030B7"},{0,"circle",U"\U000030B9"},{0,"circle",U"\U000030BB"},{0,"circle",U"\U000030BD"},{0,"circle",U"\U000030BF"},{0,"circle",U"\U000030C1"},{0,"circle",U"\U000030C4"},{0,"circle",U"\U000030C6"},{0,"circle",U"\U000030C8"},{0,"circle",U"\U000030CA"},{0,"circle",U"\U000030CB"},{0,"circle",U"\U000030CC"},{0,"circle",U"\U000030CD"},{0,"circle",U"\U000030CE"},{0,"circle",U"\U000030CF"},{0,"circle",U"\U000030D2"},{0,"circle",U"\U000030D5"},{0,"circle",U"\U000030D8"},{0,"circle",U"\U000030DB"},{0,"circle",U"\U000030DE"},{0,"circle",U"\U000030DF"},{0,"circle",U"\U000030E0"},{0,"circle",U"\U000030E1"},{0,"circle",U"\U000030E2"},{0,"circle",U"\U000030E4"},{0,"circle",U"\U000030E6"},{0,"circle",U"\U000030E8"},{0,"circle",U"\U000030E9"},{0,"circle",U"\U000030EA"},{0,"circle",U"\U000030EB"},{0,"circle",U"\U000030EC"},{0,"circle",U"\U000030ED"},{0,"circle",U"\U000030EF"},{0,"circle",U"\U000030F0"},{0,"circle",U"\U000030F1"},{0,"circle",U"\U000030F2"},{0,"square",U"\U00004EE4\U0000548C"}, };
static const NormalizationProperties _102[] = { {0,"square",U"\U000030A2\U000030D1\U000030FC\U000030C8"},{0,"square",U"\U000030A2\U000030EB\U000030D5\U000030A1"},{0,"square",U"\U000030A2\U000030F3\U000030DA\U000030A2"},{0,"square",U"\U000030A2\U000030FC\U000030EB"},{0,"square",U"\U000030A4\U000030CB\U000030F3\U000030B0"},{0,"square",U"\U000030A4\U000030F3\U000030C1"},{0,"square",U"\U000030A6\U000030A9\U000030F3"},{0,"square",U"\U000030A8\U000030B9\U000030AF\U000030FC\U000030C9"},{0,"square",U"\U000030A8\U000030FC\U000030AB\U000030FC"},{0,"square",U"\U000030AA\U000030F3\U000030B9"},{0,"square",U"\U000030AA\U000030FC\U000030E0"},{0,"square",U"\U000030AB\U000030A4\U000030EA"},{0,"square",U"\U000030AB\U000030E9\U000030C3\U000030C8"},{0,"square",U"\U000030AB\U000030ED\U000030EA\U000030FC"},{0,"square",U"\U000030AC\U000030ED\U000030F3"},{0,"square",U"\U000030AC\U000030F3\U000030DE"},{0,"square",U"\U000030AE\U000030AC"},{0,"square",U"\U000030AE\U000030CB\U000030FC"},{0,"square",U"\U000030AD\U000030E5\U000030EA\U000030FC"},{0,"square",U"\U000030AE\U000030EB\U000030C0\U000030FC"},{0,"square",U"\U000030AD\U000030ED"},{0,"square",U"\U000030AD\U000030ED\U000030B0\U000030E9\U000030E0"},{0,"square",U"\U000030AD\U000030ED\U000030E1\U000030FC\U000030C8\U000030EB"},{0,"square",U"\U000030AD\U000030ED\U000030EF\U000030C3\U000030C8"},{0,"square",U"\U000030B0\U000030E9\U000030E0"},{0,"square",U"\U000030B0\U000030E9\U000030E0\U000030C8\U000030F3"},{0,"square",U"\U000030AF\U000030EB\U000030BC\U000030A4\U000030ED"},{0,"square",U"\U000030AF\U000030ED\U000030FC\U000030CD"},{0,"square",U"\U000030B1\U000030FC\U000030B9"},{0,"square",U"\U000030B3\U000030EB\U000030CA"},{0,"square",U"\U000030B3\U000030FC\U000030DD"},{0,"square",U"\U000030B5\U000030A4\U000030AF\U000030EB"},{0,"square",U"\U000030B5\U000030F3\U000030C1\U000030FC\U000030E0"},{0,"square",U"\U000030B7\U000030EA\U000030F3\U000030B0"},{0,"square",U"\U000030BB\U000030F3\U000030C1"},{0,"square",U"\U000030BB\U000030F3\U000030C8"},{0,"square",U"\U000030C0\U000030FC\U000030B9"},{0,"square",U"\U000030C7\U000030B7"},{0,"square",U"\U000030C9\U000030EB"},{0,"square",U"\U000030C8\U000030F3"},{0,"square",U"\U000030CA\U000030CE"},{0,"square",U"\U000030CE\U000030C3\U000030C8"},{0,"square",U"\U000030CF\U000030A4\U000030C4"},{0,"square",U"\U000030D1\U000030FC\U000030BB\U000030F3\U000030C8"},{0,"square",U"\U000030D1\U000030FC\U000030C4"},{0,"square",U"\U000030D0\U000030FC\U000030EC\U000030EB"},{0,"square",U"\U000030D4\U000030A2\U000030B9\U000030C8\U000030EB"},{0,"square",U"\U000030D4\U000030AF\U000030EB"},{0,"square",U"\U000030D4\U000030B3"},{0,"square",U"\U000030D3\U000030EB"},{0,"square",U"\U000030D5\U000030A1\U000030E9\U000030C3\U000030C9"},{0,"square",U"\U000030D5\U000030A3\U000030FC\U000030C8"},{0,"square",U"\U000030D6\U000030C3\U000030B7\U000030A7\U000030EB"},{0,"square",U"\U000030D5\U000030E9\U000030F3"},{0,"square",U"\U000030D8\U000030AF\U000030BF\U000030FC\U000030EB"},{0,"square",U"\U000030DA\U000030BD"},{0,"square",U"\U000030DA\U000030CB\U000030D2"},{0,"square",U"\U000030D8\U000030EB\U000030C4"},{0,"square",U"\U000030DA\U000030F3\U000030B9"},{0,"square",U"\U000030DA\U000030FC\U000030B8"},{0,"square",U"\U000030D9\U000030FC\U000030BF"},{0,"square",U"\U000030DD\U000030A4\U000030F3\U000030C8"},{0,"square",U"\U000030DC\U000030EB\U000030C8"},{0,"square",U"\U000030DB\U000030F3"},{0,"square",U"\U000030DD\U000030F3\U000030C9"},{0,"square",U"\U000030DB\U000030FC\U000030EB"},{0,"square",U"\U000030DB\U000030FC\U000030F3"},{0,"square",U"\U000030DE\U000030A4\U000030AF\U000030ED"},{0,"square",U"\U000030DE\U000030A4\U000030EB"},{0,"square",U"\U000030DE\U000030C3\U000030CF"},{0,"square",U"\U000030DE\U000030EB\U000030AF"},{0,"square",U"\U000030DE\U000030F3\U000030B7\U000030E7\U000030F3"},{0,"square",U"\U000030DF\U000030AF\U000030ED\U000030F3"},{0,"square",U"\U000030DF\U000030EA"},{0,"square",U"\U000030DF\U000030EA\U000030D0\U000030FC\U000030EB"},{0,"square",U"\U000030E1\U000030AC"},{0,"square",U"\U000030E1\U000030AC\U000030C8\U000030F3"},{0,"square",U"\U000030E1\U000030FC\U000030C8\U000030EB"},{0,"square",U"\U000030E4\U000030FC\U000030C9"},{0,"square",U"\U000030E4\U000030FC\U000030EB"},{0,"square",U"\U000030E6\U000030A2\U000030F3"},{0,"square",U"\U000030EA\U000030C3\U000030C8\U000030EB"},{0,"square",U"\U000030EA\U000030E9"},{0,"square",U"\U000030EB\U000030D4\U000030FC"},{0,"square",U"\U000030EB\U000030FC\U000030D6\U000030EB"},{0,"square",U"\U000030EC\U000030E0"},{0,"square",U"\U000030EC\U000030F3\U000030C8\U000030B2\U000030F3"},{0,"square",U"\U000030EF\U000030C3\U000030C8"},{0,"compat",U"\U00000030\U000070B9"},{0,"compat",U"\U00000031\U000070B9"},{0,"compat",U"\U00000032\U000070B9"},{0,"compat",U"\U00000033\U000070B9"},{0,"compat",U"\U00000034\U000070B9"},{0,"compat",U"\U00000035\U000070B9"},{0,"compat",U"\U00000036\U000070B9"},{0,"compat",U"\U00000037\U000070B9"},{0,"compat",U"\U00000038\U000070B9"},{0,"compat",U"\U00000039\U000070B9"},{0,"compat",U"\U00000031\U00000030\U000070B9"},{0,"compat",U"\U00000031\U00000031\U000070B9"},{0,"compat",U"\U00000031\U00000032\U000070B9"},{0,"compat",U"\U00000031\U00000033\U000070B9"},{0,"compat",U"\U00000031\U00000034\U000070B9"},{0,"compat",U"\U00000031\U00000035\U000070B9"},{0,"compat",U"\U00000031\U00000036\U000070B9"},{0,"compat",U"\U00000031\U00000037\U000070B9"},{0,"compat",U"\U00000031\U00000038\U000070B9"},{0,"compat",U"\U00000031\U00000039\U000070B9"},{0,"compat",U"\U00000032\U00000030\U000070B9"},{0,"compat",U"\U00000032\U00000031\U000070B9"},{0,"compat",U"\U00000032\U00000032\U000070B9"},{0,"compat",U"\U00000032\U00000033\U000070B9"},{0,"compat",U"\U00000032\U00000034\U000070B9"},{0,"square",U"\U00000068\U00000050\U00000061"},{0,"square",U"\U00000064\U00000061"},{0,"square",U"\U00000041\U00000055"},{0,"square",U"\U00000062\U00000061\U00000072"},{0,"square",U"\U0000006F\U00000056"},{0,"square",U"\U00000070\U00000063"},{0,"square",U"\U00000064\U0000006D"},{0,"square",U"\U00000064\U0000006D\U000000B2"},{0,"square",U"\U00000064\U0000006D\U000000B3"},{0,"square",U"\U00000049\U00000055"},{0,"square",U"\U00005E73\U00006210"},{0,"square",U"\U0000662D\U0000548C"},{0,"square",U"\U00005927\U00006B63"},{0,"square",U"\U0000660E\U00006CBB"},{0,"square",U"\U0000682A\U00005F0F\U00004F1A\U0000793E"}, };
static const NormalizationProperties _103[] = { {0,"square",U"\U00000070\U00000041"},{0,"square",U"\U0000006E\U00000041"},{0,"square",U"\U000003BC\U00000041"},{0,"square",U"\U0000006D\U00000041"},{0,"square",U"\U0000006B\U00000041"},{0,"square",U"\U0000004B\U00000042"},{0,"square",U"\U0000004D\U00000042"},{0,"square",U"\U00000047\U00000042"},{0,"square",U"\U00000063\U00000061\U0000006C"},{0,"square",U"\U0000006B\U00000063\U00000061\U0000006C"},{0,"square",U"\U00000070\U00000046"},{0,"square",U"\U0000006E\U00000046"},{0,"square",U"\U000003BC\U00000046"},{0,"square",U"\U000003BC\U00000067"},{0,"square",U"\U0000006D\U00000067"},{0,"square",U"\U0000006B\U00000067"},{0,"square",U"\U00000048\U0000007A"},{0,"square",U"\U0000006B\U00000048\U0000007A"},{0,"square",U"\U0000004D\U00000048\U0000007A"},{0,"square",U"\U00000047\U00000048\U0000007A"},{0,"square",U"\U00000054\U00000048\U0000007A"},{0,"square",U"\U000003BC\U00002113"},{0,"square",U"\U0000006D\U00002113"},{0,"square",U"\U00000064\U00002113"},{0,"square",U"\U0000006B\U00002113"},{0,"square",U"\U00000066\U0000006D"},{0,"square",U"\U0000006E\U0000006D"},{0,"square",U"\U000003BC\U0000006D"},{0,"square",U"\U0000006D\U0000006D"},{0,"square",U"\U00000063\U0000006D"},{0,"square",U"\U0000006B\U0000006D"},{0,"square",U"\U0000006D\U0000006D\U000000B2"},{0,"square",U"\U00000063\U0000006D\U000000B2"},{0,"square",U"\U0000006D\U000000B2"},{0,"square",U"\U0000006B\U0000006D\U000000B2"},{0,"square",U"\U0000006D\U0000006D\U000000B3"},{0,"square",U"\U00000063\U0000006D\U000000B3"},{0,"square",U"\U0000006D\U000000B3"},{0,"square",U"\U0000006B\U0000006D\U000000B3"},{0,"square",U"\U0000006D\U00002215\U00000073"},{0,"square",U"\U0000006D\U00002215\U00000073\U000000B2"},{0,"square",U"\U00000050\U00000061"},{0,"square",U"\U0000006B\U00000050\U00000061"},{0,"square",U"\U0000004D\U00000050\U00000061"},{0,"square",U"\U00000047\U00000050\U00000061"},{0,"square",U"\U00000072\U00000061\U00000064"},{0,"square",U"\U00000072\U00000061\U00000064\U00002215\U00000073"},{0,"square",U"\U00000072\U00000061\U00000064\U00002215\U00000073\U000000B2"},{0,"square",U"\U00000070\U00000073"},{0,"square",U"\U0000006E\U00000073"},{0,"square",U"\U000003BC\U00000073"},{0,"square",U"\U0000006D\U00000073"},{0,"square",U"\U00000070\U00000056"},{0,"square",U"\U0000006E\U00000056"},{0,"square",U"\U000003BC\U00000056"},{0,"square",U"\U0000006D\U00000056"},{0,"square",U"\U0000006B\U00000056"},{0,"square",U"\U0000004D\U00000056"},{0,"square",U"\U00000070\U00000057"},{0,"square",U"\U0000006E\U00000057"},{0,"square",U"\U000003BC\U00000057"},{0,"square",U"\U0000006D\U00000057"},{0,"square",U"\U0000006B\U00000057"},{0,"square",U"\U0000004D\U00000057"},{0,"square",U"\U0000006B\U000003A9"},{0,"square",U"\U0000004D\U000003A9"},{0,"square",U"\U00000061\U0000002E\U0000006D\U0000002E"},{0,"square",U"\U00000042\U00000071"},{0,"square",U"\U00000063\U00000063"},{0,"square",U"\U00000063\U00000064"},{0,"square",U"\U00000043\U00002215\U0000006B\U00000067"},{0,"square",U"\U00000043\U0000006F\U0000002E"},{0,"square",U"\U00000064\U00000042"},{0,"square",U"\U00000047\U00000079"},{0,"square",U"\U00000068\U00000061"},{0,"square",U"\U00000048\U00000050"},{0,"square",U"\U00000069\U0000006E"},{0,"square",U"\U0000004B\U0000004B"},{0,"square",U"\U0000004B\U0000004D"},{0,"square",U"\U0000006B\U00000074"},{0,"square",U"\U0000006C\U0000006D"},{0,"square",U"\U0000006C\U0000006E"},{0,"square",U"\U0000006C\U0000006F\U00000067"},{0,"square",U"\U0000006C\U00000078"},{0,"square",U"\U0000006D\U00000062"},{0,"square",U"\U0000006D\U00000069\U0000006C"},{0,"square",U"\U0000006D\U0000006F\U0000006C"},{0,"square",U"\U00000050\U00000048"},{0,"square",U"\U00000070\U0000002E\U0000006D\U0000002E"},{0,"square",U"\U00000050\U00000050\U0000004D"},{0,"square",U"\U00000050\U00000052"},{0,"square",U"\U00000073\U00000072"},{0,"square",U"\U00000053\U00000076"},{0,"square",U"\U00000057\U00000062"},{0,"square",U"\U00000056\U00002215\U0000006D"},{0,"square",U"\U00000041\U00002215\U0000006D"},{0,"compat",U"\U00000031\U000065E5"},{0,"compat",U"\U00000032\U000065E5"},{0,"compat",U"\U00000033\U000065E5"},{0,"compat",U"\U00000034\U000065E5"},{0,"compat",U"\U00000035\U000065E5"},{0,"compat",U"\U00000036\U000065E5"},{0,"compat",U"\U00000037\U000065E5"},{0,"compat",U"\U00000038\U000065E5"},{0,"compat",U"\U00000039\U000065E5"},{0,"compat",U"\U00000031\U00000030\U000065E5"},{0,"compat",U"\U00000031\U00000031\U000065E5"},{0,"compat",U"\U00000031\U00000032\U000065E5"},{0,"compat",U"\U00000031\U00000033\U000065E5"},{0,"compat",U"\U00000031\U00000034\U000065E5"},{0,"compat",U"\U00000031\U00000035\U000065E5"},{0,"compat",U"\U00000031\U00000036\U000065E5"},{0,"compat",U"\U00000031\U00000037\U000065E5"},{0,"compat",U"\U00000031\U00000038\U000065E5"},{0,"compat",U"\U00000031\U00000039\U000065E5"},{0,"compat",U"\U00000032\U00000030\U000065E5"},{0,"compat",U"\U00000032\U00000031\U000065E5"},{0,"compat",U"\U00000032\U00000032\U000065E5"},{0,"compat",U"\U00000032\U00000033\U000065E5"},{0,"compat",U"\U00000032\U00000034\U000065E5"},{0,"compat",U"\U00000032\U00000035\U000065E5"},{0,"compat",U"\U00000032\U00000036\U000065E5"},{0,"compat",U"\U00000032\U00000037\U000065E5"},{0,"compat",U"\U00000032\U00000038\U000065E5"},{0,"compat",U"\U00000032\U00000039\U000065E5"},{0,"compat",U"\U00000033\U00000030\U000065E5"},{0,"compat",U"\U00000033\U00000031\U000065E5"},{0,"square",U"\U00000067\U00000061\U0000006C"}, };
static const NormalizationProperties _332[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _333[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U0000044A"},{0,"super",U"\U0000044C"},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _334[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U0000A76F"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _335[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U00000126"},{0,"super",U"\U00000153"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _336[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _337[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _338[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{220,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _339[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _341[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{0,0,0},{230,0,0},{230,0,0},{220,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{0,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _342[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U0000A727"},{0,"super",U"\U0000AB37"},{0,"super",U"\U0000026B"},{0,"super",U"\U0000AB52"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U0000028D"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _343[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _498[] = { {0,0,U"\U00008C48"},{0,0,U"\U000066F4"},{0,0,U"\U00008ECA"},{0,0,U"\U00008CC8"},{0,0,U"\U00006ED1"},{0,0,U"\U00004E32"},{0,0,U"\U000053E5"},{0,0,U"\U00009F9C"},{0,0,U"\U00009F9C"},{0,0,U"\U00005951"},{0,0,U"\U000091D1"},{0,0,U"\U00005587"},{0,0,U"\U00005948"},{0,0,U"\U000061F6"},{0,0,U"\U00007669"},{0,0,U"\U00007F85"},{0,0,U"\U0000863F"},{0,0,U"\U000087BA"},{0,0,U"\U000088F8"},{0,0,U"\U0000908F"},{0,0,U"\U00006A02"},{0,0,U"\U00006D1B"},{0,0,U"\U000070D9"},{0,0,U"\U000073DE"},{0,0,U"\U0000843D"},{0,0,U"\U0000916A"},{0,0,U"\U000099F1"},{0,0,U"\U00004E82"},{0,0,U"\U00005375"},{0,0,U"\U00006B04"},{0,0,U"\U0000721B"},{0,0,U"\U0000862D"},{0,0,U"\U00009E1E"},{0,0,U"\U00005D50"},{0,0,U"\U00006FEB"},{0,0,U"\U000085CD"},{0,0,U"\U00008964"},{0,0,U"\U000062C9"},{0,0,U"\U000081D8"},{0,0,U"\U0000881F"},{0,0,U"\U00005ECA"},{0,0,U"\U00006717"},{0,0,U"\U00006D6A"},{0,0,U"\U000072FC"},{0,0,U"\U000090CE"},{0,0,U"\U00004F86"},{0,0,U"\U000051B7"},{0,0,U"\U000052DE"},{0,0,U"\U000064C4"},{0,0,U"\U00006AD3"},{0,0,U"\U00007210"},{0,0,U"\U000076E7"},{0,0,U"\U00008001"},{0,0,U"\U00008606"},{0,0,U"\U0000865C"},{0,0,U"\U00008DEF"},{0,0,U"\U00009732"},{0,0,U"\U00009B6F"},{0,0,U"\U00009DFA"},{0,0,U"\U0000788C"},{0,0,U"\U0000797F"},{0,0,U"\U00007DA0"},{0,0,U"\U000083C9"},{0,0,U"\U00009304"},{0,0,U"\U00009E7F"},{0,0,U"\U00008AD6"},{0,0,U"\U000058DF"},{0,0,U"\U00005F04"},{0,0,U"\U00007C60"},{0,0,U"\U0000807E"},{0,0,U"\U00007262"},{0,0,U"\U000078CA"},{0,0,U"\U00008CC2"},{0,0,U"\U000096F7"},{0,0,U"\U000058D8"},{0,0,U"\U00005C62"},{0,0,U"\U00006A13"},{0,0,U"\U00006DDA"},{0,0,U"\U00006F0F"},{0,0,U"\U00007D2F"},{0,0,U"\U00007E37"},{0,0,U"\U0000964B"},{0,0,U"\U000052D2"},{0,0,U"\U0000808B"},{0,0,U"\U000051DC"},{0,0,U"\U000051CC"},{0,0,U"\U00007A1C"},{0,0,U"\U00007DBE"},{0,0,U"\U000083F1"},{0,0,U"\U00009675"},{0,0,U"\U00008B80"},{0,0,U"\U000062CF"},{0,0,U"\U00006A02"},{0,0,U"\U00008AFE"},{0,0,U"\U00004E39"},{0,0,U"\U00005BE7"},{0,0,U"\U00006012"},{0,0,U"\U00007387"},{0,0,U"\U00007570"},{0,0,U"\U00005317"},{0,0,U"\U000078FB"},{0,0,U"\U00004FBF"},{0,0,U"\U00005FA9"},{0,0,U"\U00004E0D"},{0,0,U"\U00006CCC"},{0,0,U"\U00006578"},{0,0,U"\U00007D22"},{0,0,U"\U000053C3"},{0,0,U"\U0000585E"},{0,0,U"\U00007701"},{0,0,U"\U00008449"},{0,0,U"\U00008AAA"},{0,0,U"\U00006BBA"},{0,0,U"\U00008FB0"},{0,0,U"\U00006C88"},{0,0,U"\U000062FE"},{0,0,U"\U000082E5"},{0,0,U"\U000063A0"},{0,0,U"\U00007565"},{0,0,U"\U00004EAE"},{0,0,U"\U00005169"},{0,0,U"\U000051C9"},{0,0,U"\U00006881"},{0,0,U"\U00007CE7"},{0,0,U"\U0000826F"},{0,0,U"\U00008AD2"},{0,0,U"\U000091CF"},{0,0,U"\U000052F5"}, };
static const NormalizationProperties _499[] = { {0,0,U"\U00005442"},{0,0,U"\U00005973"},{0,0,U"\U00005EEC"},{0,0,U"\U000065C5"},{0,0,U"\U00006FFE"},{0,0,U"\U0000792A"},{0,0,U"\U000095AD"},{0,0,U"\U00009A6A"},{0,0,U"\U00009E97"},{0,0,U"\U00009ECE"},{0,0,U"\U0000529B"},{0,0,U"\U000066C6"},{0,0,U"\U00006B77"},{0,0,U"\U00008F62"},{0,0,U"\U00005E74"},{0,0,U"\U00006190"},{0,0,U"\U00006200"},{0,0,U"\U0000649A"},{0,0,U"\U00006F23"},{0,0,U"\U00007149"},{0,0,U"\U00007489"},{0,0,U"\U000079CA"},{0,0,U"\U00007DF4"},{0,0,U"\U0000806F"},{0,0,U"\U00008F26"},{0,0,U"\U000084EE"},{0,0,U"\U00009023"},{0,0,U"\U0000934A"},{0,0,U"\U00005217"},{0,0,U"\U000052A3"},{0,0,U"\U000054BD"},{0,0,U"\U000070C8"},{0,0,U"\U000088C2"},{0,0,U"\U00008AAA"},{0,0,U"\U00005EC9"},{0,0,U"\U00005FF5"},{0,0,U"\U0000637B"},{0,0,U"\U00006BAE"},{0,0,U"\U00007C3E"},{0,0,U"\U00007375"},{0,0,U"\U00004EE4"},{0,0,U"\U000056F9"},{0,0,U"\U00005BE7"},{0,0,U"\U00005DBA"},{0,0,U"\U0000601C"},{0,0,U"\U000073B2"},{0,0,U"\U00007469"},{0,0,U"\U00007F9A"},{0,0,U"\U00008046"},{0,0,U"\U00009234"},{0,0,U"\U000096F6"},{0,0,U"\U00009748"},{0,0,U"\U00009818"},{0,0,U"\U00004F8B"},{0,0,U"\U000079AE"},{0,0,U"\U000091B4"},{0,0,U"\U000096B8"},{0,0,U"\U000060E1"},{0,0,U"\U00004E86"},{0,0,U"\U000050DA"},{0,0,U"\U00005BEE"},{0,0,U"\U00005C3F"},{0,0,U"\U00006599"},{0,0,U"\U00006A02"},{0,0,U"\U000071CE"},{0,0,U"\U00007642"},{0,0,U"\U000084FC"},{0,0,U"\U0000907C"},{0,0,U"\U00009F8D"},{0,0,U"\U00006688"},{0,0,U"\U0000962E"},{0,0,U"\U00005289"},{0,0,U"\U0000677B"},{0,0,U"\U000067F3"},{0,0,U"\U00006D41"},{0,0,U"\U00006E9C"},{0,0,U"\U00007409"},{0,0,U"\U00007559"},{0,0,U"\U0000786B"},{0,0,U"\U00007D10"},{0,0,U"\U0000985E"},{0,0,U"\U0000516D"},{0,0,U"\U0000622E"},{0,0,U"\U00009678"},{0,0,U"\U0000502B"},{0,0,U"\U00005D19"},{0,0,U"\U00006DEA"},{0,0,U"\U00008F2A"},{0,0,U"\U00005F8B"},{0,0,U"\U00006144"},{0,0,U"\U00006817"},{0,0,U"\U00007387"},{0,0,U"\U00009686"},{0,0,U"\U00005229"},{0,0,U"\U0000540F"},{0,0,U"\U00005C65"},{0,0,U"\U00006613"},{0,0,U"\U0000674E"},{0,0,U"\U000068A8"},{0,0,U"\U00006CE5"},{0,0,U"\U00007406"},{0,0,U"\U000075E2"},{0,0,U"\U00007F79"},{0,0,U"\U000088CF"},{0,0,U"\U000088E1"},{0,0,U"\U000091CC"},{0,0,U"\U000096E2"},{0,0,U"\U0000533F"},{0,0,U"\U00006EBA"},{0,0,U"\U0000541D"},{0,0,U"\U000071D0"},{0,0,U"\U00007498"},{0,0,U"\U000085FA"},{0,0,U"\U000096A3"},{0,0,U"\U00009C57"},{0,0,U"\U00009E9F"},{0,0,U"\U00006797"},{0,0,U"\U00006DCB"},{0,0,U"\U000081E8"},{0,0,U"\U00007ACB"},{0,0,U"\U00007B20"},{0,0,U"\U00007C92"},{0,0,U"\U000072C0"},{0,0,U"\U00007099"},{0,0,U"\U00008B58"},{0,0,U"\U00004EC0"},{0,0,U"\U00008336"},{0,0,U"\U0000523A"}, };
static const NormalizationProperties _500[] = { {0,0,U"\U00005207"},{0,0,U"\U00005EA6"},{0,0,U"\U000062D3"},{0,0,U"\U00007CD6"},{0,0,U"\U00005B85"},{0,0,U"\U00006D1E"},{0,0,U"\U000066B4"},{0,0,U"\U00008F3B"},{0,0,U"\U0000884C"},{0,0,U"\U0000964D"},{0,0,U"\U0000898B"},{0,0,U"\U00005ED3"},{0,0,U"\U00005140"},{0,0,U"\U000055C0"},{0,0,0},{0,0,0},{0,0,U"\U0000585A"},{0,0,0},{0,0,U"\U00006674"},{0,0,0},{0,0,0},{0,0,U"\U000051DE"},{0,0,U"\U0000732A"},{0,0,U"\U000076CA"},{0,0,U"\U0000793C"},{0,0,U"\U0000795E"},{0,0,U"\U00007965"},{0,0,U"\U0000798F"},{0,0,U"\U00009756"},{0,0,U"\U00007CBE"},{0,0,U"\U00007FBD"},{0,0,0},{0,0,U"\U00008612"},{0,0,0},{0,0,U"\U00008AF8"},{0,0,0},{0,0,0},{0,0,U"\U00009038"},{0,0,U"\U000090FD"},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000098EF"},{0,0,U"\U000098FC"},{0,0,U"\U00009928"},{0,0,U"\U00009DB4"},{0,0,U"\U000090DE"},{0,0,U"\U000096B7"},{0,0,U"\U00004FAE"},{0,0,U"\U000050E7"},{0,0,U"\U0000514D"},{0,0,U"\U000052C9"},{0,0,U"\U000052E4"},{0,0,U"\U00005351"},{0,0,U"\U0000559D"},{0,0,U"\U00005606"},{0,0,U"\U00005668"},{0,0,U"\U00005840"},{0,0,U"\U000058A8"},{0,0,U"\U00005C64"},{0,0,U"\U00005C6E"},{0,0,U"\U00006094"},{0,0,U"\U00006168"},{0,0,U"\U0000618E"},{0,0,U"\U000061F2"},{0,0,U"\U0000654F"},{0,0,U"\U000065E2"},{0,0,U"\U00006691"},{0,0,U"\U00006885"},{0,0,U"\U00006D77"},{0,0,U"\U00006E1A"},{0,0,U"\U00006F22"},{0,0,U"\U0000716E"},{0,0,U"\U0000722B"},{0,0,U"\U00007422"},{0,0,U"\U00007891"},{0,0,U"\U0000793E"},{0,0,U"\U00007949"},{0,0,U"\U00007948"},{0,0,U"\U00007950"},{0,0,U"\U00007956"},{0,0,U"\U0000795D"},{0,0,U"\U0000798D"},{0,0,U"\U0000798E"},{0,0,U"\U00007A40"},{0,0,U"\U00007A81"},{0,0,U"\U00007BC0"},{0,0,U"\U00007DF4"},{0,0,U"\U00007E09"},{0,0,U"\U00007E41"},{0,0,U"\U00007F72"},{0,0,U"\U00008005"},{0,0,U"\U000081ED"},{0,0,U"\U00008279"},{0,0,U"\U00008279"},{0,0,U"\U00008457"},{0,0,U"\U00008910"},{0,0,U"\U00008996"},{0,0,U"\U00008B01"},{0,0,U"\U00008B39"},{0,0,U"\U00008CD3"},{0,0,U"\U00008D08"},{0,0,U"\U00008FB6"},{0,0,U"\U00009038"},{0,0,U"\U000096E3"},{0,0,U"\U000097FF"},{0,0,U"\U0000983B"},{0,0,U"\U00006075"},{0,0,U"\U000242EE"},{0,0,U"\U00008218"},{0,0,0},{0,0,0},{0,0,U"\U00004E26"},{0,0,U"\U000051B5"},{0,0,U"\U00005168"},{0,0,U"\U00004F80"},{0,0,U"\U00005145"},{0,0,U"\U00005180"},{0,0,U"\U000052C7"},{0,0,U"\U000052FA"},{0,0,U"\U0000559D"},{0,0,U"\U00005555"},{0,0,U"\U00005599"},{0,0,U"\U000055E2"},{0,0,U"\U0000585A"},{0,0,U"\U000058B3"},{0,0,U"\U00005944"},{0,0,U"\U00005954"}, };
static const NormalizationProperties _501[] = { {0,0,U"\U00005A62"},{0,0,U"\U00005B28"},{0,0,U"\U00005ED2"},{0,0,U"\U00005ED9"},{0,0,U"\U00005F69"},{0,0,U"\U00005FAD"},{0,0,U"\U000060D8"},{0,0,U"\U0000614E"},{0,0,U"\U00006108"},{0,0,U"\U0000618E"},{0,0,U"\U00006160"},{0,0,U"\U000061F2"},{0,0,U"\U00006234"},{0,0,U"\U000063C4"},{0,0,U"\U0000641C"},{0,0,U"\U00006452"},{0,0,U"\U00006556"},{0,0,U"\U00006674"},{0,0,U"\U00006717"},{0,0,U"\U0000671B"},{0,0,U"\U00006756"},{0,0,U"\U00006B79"},{0,0,U"\U00006BBA"},{0,0,U"\U00006D41"},{0,0,U"\U00006EDB"},{0,0,U"\U00006ECB"},{0,0,U"\U00006F22"},{0,0,U"\U0000701E"},{0,0,U"\U0000716E"},{0,0,U"\U000077A7"},{0,0,U"\U00007235"},{0,0,U"\U000072AF"},{0,0,U"\U0000732A"},{0,0,U"\U00007471"},{0,0,U"\U00007506"},{0,0,U"\U0000753B"},{0,0,U"\U0000761D"},{0,0,U"\U0000761F"},{0,0,U"\U000076CA"},{0,0,U"\U000076DB"},{0,0,U"\U000076F4"},{0,0,U"\U0000774A"},{0,0,U"\U00007740"},{0,0,U"\U000078CC"},{0,0,U"\U00007AB1"},{0,0,U"\U00007BC0"},{0,0,U"\U00007C7B"},{0,0,U"\U00007D5B"},{0,0,U"\U00007DF4"},{0,0,U"\U00007F3E"},{0,0,U"\U00008005"},{0,0,U"\U00008352"},{0,0,U"\U000083EF"},{0,0,U"\U00008779"},{0,0,U"\U00008941"},{0,0,U"\U00008986"},{0,0,U"\U00008996"},{0,0,U"\U00008ABF"},{0,0,U"\U00008AF8"},{0,0,U"\U00008ACB"},{0,0,U"\U00008B01"},{0,0,U"\U00008AFE"},{0,0,U"\U00008AED"},{0,0,U"\U00008B39"},{0,0,U"\U00008B8A"},{0,0,U"\U00008D08"},{0,0,U"\U00008F38"},{0,0,U"\U00009072"},{0,0,U"\U00009199"},{0,0,U"\U00009276"},{0,0,U"\U0000967C"},{0,0,U"\U000096E3"},{0,0,U"\U00009756"},{0,0,U"\U000097DB"},{0,0,U"\U000097FF"},{0,0,U"\U0000980B"},{0,0,U"\U0000983B"},{0,0,U"\U00009B12"},{0,0,U"\U00009F9C"},{0,0,U"\U0002284A"},{0,0,U"\U00022844"},{0,0,U"\U000233D5"},{0,0,U"\U00003B9D"},{0,0,U"\U00004018"},{0,0,U"\U00004039"},{0,0,U"\U00025249"},{0,0,U"\U00025CD0"},{0,0,U"\U00027ED3"},{0,0,U"\U00009F43"},{0,0,U"\U00009F8E"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _502[] = { {0,"compat",U"\U00000066\U00000066"},{0,"compat",U"\U00000066\U00000069"},{0,"compat",U"\U00000066\U0000006C"},{0,"compat",U"\U00000066\U00000066\U00000069"},{0,"compat",U"\U00000066\U00000066\U0000006C"},{0,"compat",U"\U0000017F\U00000074"},{0,"compat",U"\U00000073\U00000074"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000574\U00000576"},{0,"compat",U"\U00000574\U00000565"},{0,"compat",U"\U00000574\U0000056B"},{0,"compat",U"\U0000057E\U00000576"},{0,"compat",U"\U00000574\U0000056D"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000005D9\U000005B4"},{26,0,0},{0,0,U"\U000005F2\U000005B7"},{0,"font",U"\U000005E2"},{0,"font",U"\U000005D0"},{0,"font",U"\U000005D3"},{0,"font",U"\U000005D4"},{0,"font",U"\U000005DB"},{0,"font",U"\U000005DC"},{0,"font",U"\U000005DD"},{0,"font",U"\U000005E8"},{0,"font",U"\U000005EA"},{0,"font",U"\U0000002B"},{0,0,U"\U000005E9\U000005C1"},{0,0,U"\U000005E9\U000005C2"},{0,0,U"\U0000FB49\U000005C1"},{0,0,U"\U0000FB49\U000005C2"},{0,0,U"\U000005D0\U000005B7"},{0,0,U"\U000005D0\U000005B8"},{0,0,U"\U000005D0\U000005BC"},{0,0,U"\U000005D1\U000005BC"},{0,0,U"\U000005D2\U000005BC"},{0,0,U"\U000005D3\U000005BC"},{0,0,U"\U000005D4\U000005BC"},{0,0,U"\U000005D5\U000005BC"},{0,0,U"\U000005D6\U000005BC"},{0,0,0},{0,0,U"\U000005D8\U000005BC"},{0,0,U"\U000005D9\U000005BC"},{0,0,U"\U000005DA\U000005BC"},{0,0,U"\U000005DB\U000005BC"},{0,0,U"\U000005DC\U000005BC"},{0,0,0},{0,0,U"\U000005DE\U000005BC"},{0,0,0},{0,0,U"\U000005E0\U000005BC"},{0,0,U"\U000005E1\U000005BC"},{0,0,0},{0,0,U"\U000005E3\U000005BC"},{0,0,U"\U000005E4\U000005BC"},{0,0,0},{0,0,U"\U000005E6\U000005BC"},{0,0,U"\U000005E7\U000005BC"},{0,0,U"\U000005E8\U000005BC"},{0,0,U"\U000005E9\U000005BC"},{0,0,U"\U000005EA\U000005BC"},{0,0,U"\U000005D5\U000005B9"},{0,0,U"\U000005D1\U000005BF"},{0,0,U"\U000005DB\U000005BF"},{0,0,U"\U000005E4\U000005BF"},{0,"compat",U"\U000005D0\U000005DC"},{0,"isolated",U"\U00000671"},{0,"final",U"\U00000671"},{0,"isolated",U"\U0000067B"},{0,"final",U"\U0000067B"},{0,"initial",U"\U0000067B"},{0,"medial",U"\U0000067B"},{0,"isolated",U"\U0000067E"},{0,"final",U"\U0000067E"},{0,"initial",U"\U0000067E"},{0,"medial",U"\U0000067E"},{0,"isolated",U"\U00000680"},{0,"final",U"\U00000680"},{0,"initial",U"\U00000680"},{0,"medial",U"\U00000680"},{0,"isolated",U"\U0000067A"},{0,"final",U"\U0000067A"},{0,"initial",U"\U0000067A"},{0,"medial",U"\U0000067A"},{0,"isolated",U"\U0000067F"},{0,"final",U"\U0000067F"},{0,"initial",U"\U0000067F"},{0,"medial",U"\U0000067F"},{0,"isolated",U"\U00000679"},{0,"final",U"\U00000679"},{0,"initial",U"\U00000679"},{0,"medial",U"\U00000679"},{0,"isolated",U"\U000006A4"},{0,"final",U"\U000006A4"},{0,"initial",U"\U000006A4"},{0,"medial",U"\U000006A4"},{0,"isolated",U"\U000006A6"},{0,"final",U"\U000006A6"},{0,"initial",U"\U000006A6"},{0,"medial",U"\U000006A6"},{0,"isolated",U"\U00000684"},{0,"final",U"\U00000684"},{0,"initial",U"\U00000684"},{0,"medial",U"\U00000684"},{0,"isolated",U"\U00000683"},{0,"final",U"\U00000683"},{0,"initial",U"\U00000683"},{0,"medial",U"\U00000683"},{0,"isolated",U"\U00000686"},{0,"final",U"\U00000686"},{0,"initial",U"\U00000686"},{0,"medial",U"\U00000686"},{0,"isolated",U"\U00000687"},{0,"final",U"\U00000687"}, };
static const NormalizationProperties _503[] = { {0,"initial",U"\U00000687"},{0,"medial",U"\U00000687"},{0,"isolated",U"\U0000068D"},{0,"final",U"\U0000068D"},{0,"isolated",U"\U0000068C"},{0,"final",U"\U0000068C"},{0,"isolated",U"\U0000068E"},{0,"final",U"\U0000068E"},{0,"isolated",U"\U00000688"},{0,"final",U"\U00000688"},{0,"isolated",U"\U00000698"},{0,"final",U"\U00000698"},{0,"isolated",U"\U00000691"},{0,"final",U"\U00000691"},{0,"isolated",U"\U000006A9"},{0,"final",U"\U000006A9"},{0,"initial",U"\U000006A9"},{0,"medial",U"\U000006A9"},{0,"isolated",U"\U000006AF"},{0,"final",U"\U000006AF"},{0,"initial",U"\U000006AF"},{0,"medial",U"\U000006AF"},{0,"isolated",U"\U000006B3"},{0,"final",U"\U000006B3"},{0,"initial",U"\U000006B3"},{0,"medial",U"\U000006B3"},{0,"isolated",U"\U000006B1"},{0,"final",U"\U000006B1"},{0,"initial",U"\U000006B1"},{0,"medial",U"\U000006B1"},{0,"isolated",U"\U000006BA"},{0,"final",U"\U000006BA"},{0,"isolated",U"\U000006BB"},{0,"final",U"\U000006BB"},{0,"initial",U"\U000006BB"},{0,"medial",U"\U000006BB"},{0,"isolated",U"\U000006C0"},{0,"final",U"\U000006C0"},{0,"isolated",U"\U000006C1"},{0,"final",U"\U000006C1"},{0,"initial",U"\U000006C1"},{0,"medial",U"\U000006C1"},{0,"isolated",U"\U000006BE"},{0,"final",U"\U000006BE"},{0,"initial",U"\U000006BE"},{0,"medial",U"\U000006BE"},{0,"isolated",U"\U000006D2"},{0,"final",U"\U000006D2"},{0,"isolated",U"\U000006D3"},{0,"final",U"\U000006D3"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"isolated",U"\U000006AD"},{0,"final",U"\U000006AD"},{0,"initial",U"\U000006AD"},{0,"medial",U"\U000006AD"},{0,"isolated",U"\U000006C7"},{0,"final",U"\U000006C7"},{0,"isolated",U"\U000006C6"},{0,"final",U"\U000006C6"},{0,"isolated",U"\U000006C8"},{0,"final",U"\U000006C8"},{0,"isolated",U"\U00000677"},{0,"isolated",U"\U000006CB"},{0,"final",U"\U000006CB"},{0,"isolated",U"\U000006C5"},{0,"final",U"\U000006C5"},{0,"isolated",U"\U000006C9"},{0,"final",U"\U000006C9"},{0,"isolated",U"\U000006D0"},{0,"final",U"\U000006D0"},{0,"initial",U"\U000006D0"},{0,"medial",U"\U000006D0"},{0,"initial",U"\U00000649"},{0,"medial",U"\U00000649"},{0,"isolated",U"\U00000626\U00000627"},{0,"final",U"\U00000626\U00000627"},{0,"isolated",U"\U00000626\U000006D5"},{0,"final",U"\U00000626\U000006D5"},{0,"isolated",U"\U00000626\U00000648"},{0,"final",U"\U00000626\U00000648"},{0,"isolated",U"\U00000626\U000006C7"},{0,"final",U"\U00000626\U000006C7"},{0,"isolated",U"\U00000626\U000006C6"},{0,"final",U"\U00000626\U000006C6"},{0,"isolated",U"\U00000626\U000006C8"},{0,"final",U"\U00000626\U000006C8"},{0,"isolated",U"\U00000626\U000006D0"},{0,"final",U"\U00000626\U000006D0"},{0,"initial",U"\U00000626\U000006D0"},{0,"isolated",U"\U00000626\U00000649"},{0,"final",U"\U00000626\U00000649"},{0,"initial",U"\U00000626\U00000649"},{0,"isolated",U"\U000006CC"},{0,"final",U"\U000006CC"},{0,"initial",U"\U000006CC"},{0,"medial",U"\U000006CC"}, };
static const NormalizationProperties _504[] = { {0,"isolated",U"\U00000626\U0000062C"},{0,"isolated",U"\U00000626\U0000062D"},{0,"isolated",U"\U00000626\U00000645"},{0,"isolated",U"\U00000626\U00000649"},{0,"isolated",U"\U00000626\U0000064A"},{0,"isolated",U"\U00000628\U0000062C"},{0,"isolated",U"\U00000628\U0000062D"},{0,"isolated",U"\U00000628\U0000062E"},{0,"isolated",U"\U00000628\U00000645"},{0,"isolated",U"\U00000628\U00000649"},{0,"isolated",U"\U00000628\U0000064A"},{0,"isolated",U"\U0000062A\U0000062C"},{0,"isolated",U"\U0000062A\U0000062D"},{0,"isolated",U"\U0000062A\U0000062E"},{0,"isolated",U"\U0000062A\U00000645"},{0,"isolated",U"\U0000062A\U00000649"},{0,"isolated",U"\U0000062A\U0000064A"},{0,"isolated",U"\U0000062B\U0000062C"},{0,"isolated",U"\U0000062B\U00000645"},{0,"isolated",U"\U0000062B\U00000649"},{0,"isolated",U"\U0000062B\U0000064A"},{0,"isolated",U"\U0000062C\U0000062D"},{0,"isolated",U"\U0000062C\U00000645"},{0,"isolated",U"\U0000062D\U0000062C"},{0,"isolated",U"\U0000062D\U00000645"},{0,"isolated",U"\U0000062E\U0000062C"},{0,"isolated",U"\U0000062E\U0000062D"},{0,"isolated",U"\U0000062E\U00000645"},{0,"isolated",U"\U00000633\U0000062C"},{0,"isolated",U"\U00000633\U0000062D"},{0,"isolated",U"\U00000633\U0000062E"},{0,"isolated",U"\U00000633\U00000645"},{0,"isolated",U"\U00000635\U0000062D"},{0,"isolated",U"\U00000635\U00000645"},{0,"isolated",U"\U00000636\U0000062C"},{0,"isolated",U"\U00000636\U0000062D"},{0,"isolated",U"\U00000636\U0000062E"},{0,"isolated",U"\U00000636\U00000645"},{0,"isolated",U"\U00000637\U0000062D"},{0,"isolated",U"\U00000637\U00000645"},{0,"isolated",U"\U00000638\U00000645"},{0,"isolated",U"\U00000639\U0000062C"},{0,"isolated",U"\U00000639\U00000645"},{0,"isolated",U"\U0000063A\U0000062C"},{0,"isolated",U"\U0000063A\U00000645"},{0,"isolated",U"\U00000641\U0000062C"},{0,"isolated",U"\U00000641\U0000062D"},{0,"isolated",U"\U00000641\U0000062E"},{0,"isolated",U"\U00000641\U00000645"},{0,"isolated",U"\U00000641\U00000649"},{0,"isolated",U"\U00000641\U0000064A"},{0,"isolated",U"\U00000642\U0000062D"},{0,"isolated",U"\U00000642\U00000645"},{0,"isolated",U"\U00000642\U00000649"},{0,"isolated",U"\U00000642\U0000064A"},{0,"isolated",U"\U00000643\U00000627"},{0,"isolated",U"\U00000643\U0000062C"},{0,"isolated",U"\U00000643\U0000062D"},{0,"isolated",U"\U00000643\U0000062E"},{0,"isolated",U"\U00000643\U00000644"},{0,"isolated",U"\U00000643\U00000645"},{0,"isolated",U"\U00000643\U00000649"},{0,"isolated",U"\U00000643\U0000064A"},{0,"isolated",U"\U00000644\U0000062C"},{0,"isolated",U"\U00000644\U0000062D"},{0,"isolated",U"\U00000644\U0000062E"},{0,"isolated",U"\U00000644\U00000645"},{0,"isolated",U"\U00000644\U00000649"},{0,"isolated",U"\U00000644\U0000064A"},{0,"isolated",U"\U00000645\U0000062C"},{0,"isolated",U"\U00000645\U0000062D"},{0,"isolated",U"\U00000645\U0000062E"},{0,"isolated",U"\U00000645\U00000645"},{0,"isolated",U"\U00000645\U00000649"},{0,"isolated",U"\U00000645\U0000064A"},{0,"isolated",U"\U00000646\U0000062C"},{0,"isolated",U"\U00000646\U0000062D"},{0,"isolated",U"\U00000646\U0000062E"},{0,"isolated",U"\U00000646\U00000645"},{0,"isolated",U"\U00000646\U00000649"},{0,"isolated",U"\U00000646\U0000064A"},{0,"isolated",U"\U00000647\U0000062C"},{0,"isolated",U"\U00000647\U00000645"},{0,"isolated",U"\U00000647\U00000649"},{0,"isolated",U"\U00000647\U0000064A"},{0,"isolated",U"\U0000064A\U0000062C"},{0,"isolated",U"\U0000064A\U0000062D"},{0,"isolated",U"\U0000064A\U0000062E"},{0,"isolated",U"\U0000064A\U00000645"},{0,"isolated",U"\U0000064A\U00000649"},{0,"isolated",U"\U0000064A\U0000064A"},{0,"isolated",U"\U00000630\U00000670"},{0,"isolated",U"\U00000631\U00000670"},{0,"isolated",U"\U00000649\U00000670"},{0,"isolated",U"\U00000020\U0000064C\U00000651"},{0,"isolated",U"\U00000020\U0000064D\U00000651"},{0,"isolated",U"\U00000020\U0000064E\U00000651"},{0,"isolated",U"\U00000020\U0000064F\U00000651"},{0,"isolated",U"\U00000020\U00000650\U00000651"},{0,"isolated",U"\U00000020\U00000651\U00000670"},{0,"final",U"\U00000626\U00000631"},{0,"final",U"\U00000626\U00000632"},{0,"final",U"\U00000626\U00000645"},{0,"final",U"\U00000626\U00000646"},{0,"final",U"\U00000626\U00000649"},{0,"final",U"\U00000626\U0000064A"},{0,"final",U"\U00000628\U00000631"},{0,"final",U"\U00000628\U00000632"},{0,"final",U"\U00000628\U00000645"},{0,"final",U"\U00000628\U00000646"},{0,"final",U"\U00000628\U00000649"},{0,"final",U"\U00000628\U0000064A"},{0,"final",U"\U0000062A\U00000631"},{0,"final",U"\U0000062A\U00000632"},{0,"final",U"\U0000062A\U00000645"},{0,"final",U"\U0000062A\U00000646"},{0,"final",U"\U0000062A\U00000649"},{0,"final",U"\U0000062A\U0000064A"},{0,"final",U"\U0000062B\U00000631"},{0,"final",U"\U0000062B\U00000632"},{0,"final",U"\U0000062B\U00000645"},{0,"final",U"\U0000062B\U00000646"},{0,"final",U"\U0000062B\U00000649"},{0,"final",U"\U0000062B\U0000064A"},{0,"final",U"\U00000641\U00000649"},{0,"final",U"\U00000641\U0000064A"},{0,"final",U"\U00000642\U00000649"},{0,"final",U"\U00000642\U0000064A"}, };
static const NormalizationProperties _505[] = { {0,"final",U"\U00000643\U00000627"},{0,"final",U"\U00000643\U00000644"},{0,"final",U"\U00000643\U00000645"},{0,"final",U"\U00000643\U00000649"},{0,"final",U"\U00000643\U0000064A"},{0,"final",U"\U00000644\U00000645"},{0,"final",U"\U00000644\U00000649"},{0,"final",U"\U00000644\U0000064A"},{0,"final",U"\U00000645\U00000627"},{0,"final",U"\U00000645\U00000645"},{0,"final",U"\U00000646\U00000631"},{0,"final",U"\U00000646\U00000632"},{0,"final",U"\U00000646\U00000645"},{0,"final",U"\U00000646\U00000646"},{0,"final",U"\U00000646\U00000649"},{0,"final",U"\U00000646\U0000064A"},{0,"final",U"\U00000649\U00000670"},{0,"final",U"\U0000064A\U00000631"},{0,"final",U"\U0000064A\U00000632"},{0,"final",U"\U0000064A\U00000645"},{0,"final",U"\U0000064A\U00000646"},{0,"final",U"\U0000064A\U00000649"},{0,"final",U"\U0000064A\U0000064A"},{0,"initial",U"\U00000626\U0000062C"},{0,"initial",U"\U00000626\U0000062D"},{0,"initial",U"\U00000626\U0000062E"},{0,"initial",U"\U00000626\U00000645"},{0,"initial",U"\U00000626\U00000647"},{0,"initial",U"\U00000628\U0000062C"},{0,"initial",U"\U00000628\U0000062D"},{0,"initial",U"\U00000628\U0000062E"},{0,"initial",U"\U00000628\U00000645"},{0,"initial",U"\U00000628\U00000647"},{0,"initial",U"\U0000062A\U0000062C"},{0,"initial",U"\U0000062A\U0000062D"},{0,"initial",U"\U0000062A\U0000062E"},{0,"initial",U"\U0000062A\U00000645"},{0,"initial",U"\U0000062A\U00000647"},{0,"initial",U"\U0000062B\U00000645"},{0,"initial",U"\U0000062C\U0000062D"},{0,"initial",U"\U0000062C\U00000645"},{0,"initial",U"\U0000062D\U0000062C"},{0,"initial",U"\U0000062D\U00000645"},{0,"initial",U"\U0000062E\U0000062C"},{0,"initial",U"\U0000062E\U00000645"},{0,"initial",U"\U00000633\U0000062C"},{0,"initial",U"\U00000633\U0000062D"},{0,"initial",U"\U00000633\U0000062E"},{0,"initial",U"\U00000633\U00000645"},{0,"initial",U"\U00000635\U0000062D"},{0,"initial",U"\U00000635\U0000062E"},{0,"initial",U"\U00000635\U00000645"},{0,"initial",U"\U00000636\U0000062C"},{0,"initial",U"\U00000636\U0000062D"},{0,"initial",U"\U00000636\U0000062E"},{0,"initial",U"\U00000636\U00000645"},{0,"initial",U"\U00000637\U0000062D"},{0,"initial",U"\U00000638\U00000645"},{0,"initial",U"\U00000639\U0000062C"},{0,"initial",U"\U00000639\U00000645"},{0,"initial",U"\U0000063A\U0000062C"},{0,"initial",U"\U0000063A\U00000645"},{0,"initial",U"\U00000641\U0000062C"},{0,"initial",U"\U00000641\U0000062D"},{0,"initial",U"\U00000641\U0000062E"},{0,"initial",U"\U00000641\U00000645"},{0,"initial",U"\U00000642\U0000062D"},{0,"initial",U"\U00000642\U00000645"},{0,"initial",U"\U00000643\U0000062C"},{0,"initial",U"\U00000643\U0000062D"},{0,"initial",U"\U00000643\U0000062E"},{0,"initial",U"\U00000643\U00000644"},{0,"initial",U"\U00000643\U00000645"},{0,"initial",U"\U00000644\U0000062C"},{0,"initial",U"\U00000644\U0000062D"},{0,"initial",U"\U00000644\U0000062E"},{0,"initial",U"\U00000644\U00000645"},{0,"initial",U"\U00000644\U00000647"},{0,"initial",U"\U00000645\U0000062C"},{0,"initial",U"\U00000645\U0000062D"},{0,"initial",U"\U00000645\U0000062E"},{0,"initial",U"\U00000645\U00000645"},{0,"initial",U"\U00000646\U0000062C"},{0,"initial",U"\U00000646\U0000062D"},{0,"initial",U"\U00000646\U0000062E"},{0,"initial",U"\U00000646\U00000645"},{0,"initial",U"\U00000646\U00000647"},{0,"initial",U"\U00000647\U0000062C"},{0,"initial",U"\U00000647\U00000645"},{0,"initial",U"\U00000647\U00000670"},{0,"initial",U"\U0000064A\U0000062C"},{0,"initial",U"\U0000064A\U0000062D"},{0,"initial",U"\U0000064A\U0000062E"},{0,"initial",U"\U0000064A\U00000645"},{0,"initial",U"\U0000064A\U00000647"},{0,"medial",U"\U00000626\U00000645"},{0,"medial",U"\U00000626\U00000647"},{0,"medial",U"\U00000628\U00000645"},{0,"medial",U"\U00000628\U00000647"},{0,"medial",U"\U0000062A\U00000645"},{0,"medial",U"\U0000062A\U00000647"},{0,"medial",U"\U0000062B\U00000645"},{0,"medial",U"\U0000062B\U00000647"},{0,"medial",U"\U00000633\U00000645"},{0,"medial",U"\U00000633\U00000647"},{0,"medial",U"\U00000634\U00000645"},{0,"medial",U"\U00000634\U00000647"},{0,"medial",U"\U00000643\U00000644"},{0,"medial",U"\U00000643\U00000645"},{0,"medial",U"\U00000644\U00000645"},{0,"medial",U"\U00000646\U00000645"},{0,"medial",U"\U00000646\U00000647"},{0,"medial",U"\U0000064A\U00000645"},{0,"medial",U"\U0000064A\U00000647"},{0,"medial",U"\U00000640\U0000064E\U00000651"},{0,"medial",U"\U00000640\U0000064F\U00000651"},{0,"medial",U"\U00000640\U00000650\U00000651"},{0,"isolated",U"\U00000637\U00000649"},{0,"isolated",U"\U00000637\U0000064A"},{0,"isolated",U"\U00000639\U00000649"},{0,"isolated",U"\U00000639\U0000064A"},{0,"isolated",U"\U0000063A\U00000649"},{0,"isolated",U"\U0000063A\U0000064A"},{0,"isolated",U"\U00000633\U00000649"},{0,"isolated",U"\U00000633\U0000064A"},{0,"isolated",U"\U00000634\U00000649"},{0,"isolated",U"\U00000634\U0000064A"},{0,"isolated",U"\U0000062D\U00000649"}, };
static const NormalizationProperties _506[] = { {0,"isolated",U"\U0000062D\U0000064A"},{0,"isolated",U"\U0000062C\U00000649"},{0,"isolated",U"\U0000062C\U0000064A"},{0,"isolated",U"\U0000062E\U00000649"},{0,"isolated",U"\U0000062E\U0000064A"},{0,"isolated",U"\U00000635\U00000649"},{0,"isolated",U"\U00000635\U0000064A"},{0,"isolated",U"\U00000636\U00000649"},{0,"isolated",U"\U00000636\U0000064A"},{0,"isolated",U"\U00000634\U0000062C"},{0,"isolated",U"\U00000634\U0000062D"},{0,"isolated",U"\U00000634\U0000062E"},{0,"isolated",U"\U00000634\U00000645"},{0,"isolated",U"\U00000634\U00000631"},{0,"isolated",U"\U00000633\U00000631"},{0,"isolated",U"\U00000635\U00000631"},{0,"isolated",U"\U00000636\U00000631"},{0,"final",U"\U00000637\U00000649"},{0,"final",U"\U00000637\U0000064A"},{0,"final",U"\U00000639\U00000649"},{0,"final",U"\U00000639\U0000064A"},{0,"final",U"\U0000063A\U00000649"},{0,"final",U"\U0000063A\U0000064A"},{0,"final",U"\U00000633\U00000649"},{0,"final",U"\U00000633\U0000064A"},{0,"final",U"\U00000634\U00000649"},{0,"final",U"\U00000634\U0000064A"},{0,"final",U"\U0000062D\U00000649"},{0,"final",U"\U0000062D\U0000064A"},{0,"final",U"\U0000062C\U00000649"},{0,"final",U"\U0000062C\U0000064A"},{0,"final",U"\U0000062E\U00000649"},{0,"final",U"\U0000062E\U0000064A"},{0,"final",U"\U00000635\U00000649"},{0,"final",U"\U00000635\U0000064A"},{0,"final",U"\U00000636\U00000649"},{0,"final",U"\U00000636\U0000064A"},{0,"final",U"\U00000634\U0000062C"},{0,"final",U"\U00000634\U0000062D"},{0,"final",U"\U00000634\U0000062E"},{0,"final",U"\U00000634\U00000645"},{0,"final",U"\U00000634\U00000631"},{0,"final",U"\U00000633\U00000631"},{0,"final",U"\U00000635\U00000631"},{0,"final",U"\U00000636\U00000631"},{0,"initial",U"\U00000634\U0000062C"},{0,"initial",U"\U00000634\U0000062D"},{0,"initial",U"\U00000634\U0000062E"},{0,"initial",U"\U00000634\U00000645"},{0,"initial",U"\U00000633\U00000647"},{0,"initial",U"\U00000634\U00000647"},{0,"initial",U"\U00000637\U00000645"},{0,"medial",U"\U00000633\U0000062C"},{0,"medial",U"\U00000633\U0000062D"},{0,"medial",U"\U00000633\U0000062E"},{0,"medial",U"\U00000634\U0000062C"},{0,"medial",U"\U00000634\U0000062D"},{0,"medial",U"\U00000634\U0000062E"},{0,"medial",U"\U00000637\U00000645"},{0,"medial",U"\U00000638\U00000645"},{0,"final",U"\U00000627\U0000064B"},{0,"isolated",U"\U00000627\U0000064B"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"initial",U"\U0000062A\U0000062C\U00000645"},{0,"final",U"\U0000062A\U0000062D\U0000062C"},{0,"initial",U"\U0000062A\U0000062D\U0000062C"},{0,"initial",U"\U0000062A\U0000062D\U00000645"},{0,"initial",U"\U0000062A\U0000062E\U00000645"},{0,"initial",U"\U0000062A\U00000645\U0000062C"},{0,"initial",U"\U0000062A\U00000645\U0000062D"},{0,"initial",U"\U0000062A\U00000645\U0000062E"},{0,"final",U"\U0000062C\U00000645\U0000062D"},{0,"initial",U"\U0000062C\U00000645\U0000062D"},{0,"final",U"\U0000062D\U00000645\U0000064A"},{0,"final",U"\U0000062D\U00000645\U00000649"},{0,"initial",U"\U00000633\U0000062D\U0000062C"},{0,"initial",U"\U00000633\U0000062C\U0000062D"},{0,"final",U"\U00000633\U0000062C\U00000649"},{0,"final",U"\U00000633\U00000645\U0000062D"},{0,"initial",U"\U00000633\U00000645\U0000062D"},{0,"initial",U"\U00000633\U00000645\U0000062C"},{0,"final",U"\U00000633\U00000645\U00000645"},{0,"initial",U"\U00000633\U00000645\U00000645"},{0,"final",U"\U00000635\U0000062D\U0000062D"},{0,"initial",U"\U00000635\U0000062D\U0000062D"},{0,"final",U"\U00000635\U00000645\U00000645"},{0,"final",U"\U00000634\U0000062D\U00000645"},{0,"initial",U"\U00000634\U0000062D\U00000645"},{0,"final",U"\U00000634\U0000062C\U0000064A"},{0,"final",U"\U00000634\U00000645\U0000062E"},{0,"initial",U"\U00000634\U00000645\U0000062E"},{0,"final",U"\U00000634\U00000645\U00000645"},{0,"initial",U"\U00000634\U00000645\U00000645"},{0,"final",U"\U00000636\U0000062D\U00000649"},{0,"final",U"\U00000636\U0000062E\U00000645"},{0,"initial",U"\U00000636\U0000062E\U00000645"},{0,"final",U"\U00000637\U00000645\U0000062D"},{0,"initial",U"\U00000637\U00000645\U0000062D"},{0,"initial",U"\U00000637\U00000645\U00000645"},{0,"final",U"\U00000637\U00000645\U0000064A"},{0,"final",U"\U00000639\U0000062C\U00000645"},{0,"final",U"\U00000639\U00000645\U00000645"},{0,"initial",U"\U00000639\U00000645\U00000645"},{0,"final",U"\U00000639\U00000645\U00000649"},{0,"final",U"\U0000063A\U00000645\U00000645"},{0,"final",U"\U0000063A\U00000645\U0000064A"},{0,"final",U"\U0000063A\U00000645\U00000649"},{0,"final",U"\U00000641\U0000062E\U00000645"},{0,"initial",U"\U00000641\U0000062E\U00000645"},{0,"final",U"\U00000642\U00000645\U0000062D"},{0,"final",U"\U00000642\U00000645\U00000645"}, };
static const NormalizationProperties _507[] = { {0,"final",U"\U00000644\U0000062D\U00000645"},{0,"final",U"\U00000644\U0000062D\U0000064A"},{0,"final",U"\U00000644\U0000062D\U00000649"},{0,"initial",U"\U00000644\U0000062C\U0000062C"},{0,"final",U"\U00000644\U0000062C\U0000062C"},{0,"final",U"\U00000644\U0000062E\U00000645"},{0,"initial",U"\U00000644\U0000062E\U00000645"},{0,"final",U"\U00000644\U00000645\U0000062D"},{0,"initial",U"\U00000644\U00000645\U0000062D"},{0,"initial",U"\U00000645\U0000062D\U0000062C"},{0,"initial",U"\U00000645\U0000062D\U00000645"},{0,"final",U"\U00000645\U0000062D\U0000064A"},{0,"initial",U"\U00000645\U0000062C\U0000062D"},{0,"initial",U"\U00000645\U0000062C\U00000645"},{0,"initial",U"\U00000645\U0000062E\U0000062C"},{0,"initial",U"\U00000645\U0000062E\U00000645"},{0,0,0},{0,0,0},{0,"initial",U"\U00000645\U0000062C\U0000062E"},{0,"initial",U"\U00000647\U00000645\U0000062C"},{0,"initial",U"\U00000647\U00000645\U00000645"},{0,"initial",U"\U00000646\U0000062D\U00000645"},{0,"final",U"\U00000646\U0000062D\U00000649"},{0,"final",U"\U00000646\U0000062C\U00000645"},{0,"initial",U"\U00000646\U0000062C\U00000645"},{0,"final",U"\U00000646\U0000062C\U00000649"},{0,"final",U"\U00000646\U00000645\U0000064A"},{0,"final",U"\U00000646\U00000645\U00000649"},{0,"final",U"\U0000064A\U00000645\U00000645"},{0,"initial",U"\U0000064A\U00000645\U00000645"},{0,"final",U"\U00000628\U0000062E\U0000064A"},{0,"final",U"\U0000062A\U0000062C\U0000064A"},{0,"final",U"\U0000062A\U0000062C\U00000649"},{0,"final",U"\U0000062A\U0000062E\U0000064A"},{0,"final",U"\U0000062A\U0000062E\U00000649"},{0,"final",U"\U0000062A\U00000645\U0000064A"},{0,"final",U"\U0000062A\U00000645\U00000649"},{0,"final",U"\U0000062C\U00000645\U0000064A"},{0,"final",U"\U0000062C\U0000062D\U00000649"},{0,"final",U"\U0000062C\U00000645\U00000649"},{0,"final",U"\U00000633\U0000062E\U00000649"},{0,"final",U"\U00000635\U0000062D\U0000064A"},{0,"final",U"\U00000634\U0000062D\U0000064A"},{0,"final",U"\U00000636\U0000062D\U0000064A"},{0,"final",U"\U00000644\U0000062C\U0000064A"},{0,"final",U"\U00000644\U00000645\U0000064A"},{0,"final",U"\U0000064A\U0000062D\U0000064A"},{0,"final",U"\U0000064A\U0000062C\U0000064A"},{0,"final",U"\U0000064A\U00000645\U0000064A"},{0,"final",U"\U00000645\U00000645\U0000064A"},{0,"final",U"\U00000642\U00000645\U0000064A"},{0,"final",U"\U00000646\U0000062D\U0000064A"},{0,"initial",U"\U00000642\U00000645\U0000062D"},{0,"initial",U"\U00000644\U0000062D\U00000645"},{0,"final",U"\U00000639\U00000645\U0000064A"},{0,"final",U"\U00000643\U00000645\U0000064A"},{0,"initial",U"\U00000646\U0000062C\U0000062D"},{0,"final",U"\U00000645\U0000062E\U0000064A"},{0,"initial",U"\U00000644\U0000062C\U00000645"},{0,"final",U"\U00000643\U00000645\U00000645"},{0,"final",U"\U00000644\U0000062C\U00000645"},{0,"final",U"\U00000646\U0000062C\U0000062D"},{0,"final",U"\U0000062C\U0000062D\U0000064A"},{0,"final",U"\U0000062D\U0000062C\U0000064A"},{0,"final",U"\U00000645\U0000062C\U0000064A"},{0,"final",U"\U00000641\U00000645\U0000064A"},{0,"final",U"\U00000628\U0000062D\U0000064A"},{0,"initial",U"\U00000643\U00000645\U00000645"},{0,"initial",U"\U00000639\U0000062C\U00000645"},{0,"initial",U"\U00000635\U00000645\U00000645"},{0,"final",U"\U00000633\U0000062E\U0000064A"},{0,"final",U"\U00000646\U0000062C\U0000064A"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"isolated",U"\U00000635\U00000644\U000006D2"},{0,"isolated",U"\U00000642\U00000644\U000006D2"},{0,"isolated",U"\U00000627\U00000644\U00000644\U00000647"},{0,"isolated",U"\U00000627\U00000643\U00000628\U00000631"},{0,"isolated",U"\U00000645\U0000062D\U00000645\U0000062F"},{0,"isolated",U"\U00000635\U00000644\U00000639\U00000645"},{0,"isolated",U"\U00000631\U00000633\U00000648\U00000644"},{0,"isolated",U"\U00000639\U00000644\U0000064A\U00000647"},{0,"isolated",U"\U00000648\U00000633\U00000644\U00000645"},{0,"isolated",U"\U00000635\U00000644\U00000649"},{0,"isolated",U"\U00000635\U00000644\U00000649\U00000020\U00000627\U00000644\U00000644\U00000647\U00000020\U00000639\U00000644\U0000064A\U00000647\U00000020\U00000648\U00000633\U00000644\U00000645"},{0,"isolated",U"\U0000062C\U00000644\U00000020\U0000062C\U00000644\U00000627\U00000644\U00000647"},{0,"isolated",U"\U00000631\U000006CC\U00000627\U00000644"},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _508[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"vertical",U"\U0000002C"},{0,"vertical",U"\U00003001"},{0,"vertical",U"\U00003002"},{0,"vertical",U"\U0000003A"},{0,"vertical",U"\U0000003B"},{0,"vertical",U"\U00000021"},{0,"vertical",U"\U0000003F"},{0,"vertical",U"\U00003016"},{0,"vertical",U"\U00003017"},{0,"vertical",U"\U00002026"},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{0,"vertical",U"\U00002025"},{0,"vertical",U"\U00002014"},{0,"vertical",U"\U00002013"},{0,"vertical",U"\U0000005F"},{0,"vertical",U"\U0000005F"},{0,"vertical",U"\U00000028"},{0,"vertical",U"\U00000029"},{0,"vertical",U"\U0000007B"},{0,"vertical",U"\U0000007D"},{0,"vertical",U"\U00003014"},{0,"vertical",U"\U00003015"},{0,"vertical",U"\U00003010"},{0,"vertical",U"\U00003011"},{0,"vertical",U"\U0000300A"},{0,"vertical",U"\U0000300B"},{0,"vertical",U"\U00003008"},{0,"vertical",U"\U00003009"},{0,"vertical",U"\U0000300C"},{0,"vertical",U"\U0000300D"},{0,"vertical",U"\U0000300E"},{0,"vertical",U"\U0000300F"},{0,0,0},{0,0,0},{0,"vertical",U"\U0000005B"},{0,"vertical",U"\U0000005D"},{0,"compat",U"\U0000203E"},{0,"compat",U"\U0000203E"},{0,"compat",U"\U0000203E"},{0,"compat",U"\U0000203E"},{0,"compat",U"\U0000005F"},{0,"compat",U"\U0000005F"},{0,"compat",U"\U0000005F"},{0,"small",U"\U0000002C"},{0,"small",U"\U00003001"},{0,"small",U"\U0000002E"},{0,0,0},{0,"small",U"\U0000003B"},{0,"small",U"\U0000003A"},{0,"small",U"\U0000003F"},{0,"small",U"\U00000021"},{0,"small",U"\U00002014"},{0,"small",U"\U00000028"},{0,"small",U"\U00000029"},{0,"small",U"\U0000007B"},{0,"small",U"\U0000007D"},{0,"small",U"\U00003014"},{0,"small",U"\U00003015"},{0,"small",U"\U00000023"},{0,"small",U"\U00000026"},{0,"small",U"\U0000002A"},{0,"small",U"\U0000002B"},{0,"small",U"\U0000002D"},{0,"small",U"\U0000003C"},{0,"small",U"\U0000003E"},{0,"small",U"\U0000003D"},{0,0,0},{0,"small",U"\U0000005C"},{0,"small",U"\U00000024"},{0,"small",U"\U00000025"},{0,"small",U"\U00000040"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"isolated",U"\U00000020\U0000064B"},{0,"medial",U"\U00000640\U0000064B"},{0,"isolated",U"\U00000020\U0000064C"},{0,0,0},{0,"isolated",U"\U00000020\U0000064D"},{0,0,0},{0,"isolated",U"\U00000020\U0000064E"},{0,"medial",U"\U00000640\U0000064E"},{0,"isolated",U"\U00000020\U0000064F"},{0,"medial",U"\U00000640\U0000064F"},{0,"isolated",U"\U00000020\U00000650"},{0,"medial",U"\U00000640\U00000650"},{0,"isolated",U"\U00000020\U00000651"},{0,"medial",U"\U00000640\U00000651"},{0,"isolated",U"\U00000020\U00000652"},{0,"medial",U"\U00000640\U00000652"}, };
static const NormalizationProperties _509[] = { {0,"isolated",U"\U00000621"},{0,"isolated",U"\U00000622"},{0,"final",U"\U00000622"},{0,"isolated",U"\U00000623"},{0,"final",U"\U00000623"},{0,"isolated",U"\U00000624"},{0,"final",U"\U00000624"},{0,"isolated",U"\U00000625"},{0,"final",U"\U00000625"},{0,"isolated",U"\U00000626"},{0,"final",U"\U00000626"},{0,"initial",U"\U00000626"},{0,"medial",U"\U00000626"},{0,"isolated",U"\U00000627"},{0,"final",U"\U00000627"},{0,"isolated",U"\U00000628"},{0,"final",U"\U00000628"},{0,"initial",U"\U00000628"},{0,"medial",U"\U00000628"},{0,"isolated",U"\U00000629"},{0,"final",U"\U00000629"},{0,"isolated",U"\U0000062A"},{0,"final",U"\U0000062A"},{0,"initial",U"\U0000062A"},{0,"medial",U"\U0000062A"},{0,"isolated",U"\U0000062B"},{0,"final",U"\U0000062B"},{0,"initial",U"\U0000062B"},{0,"medial",U"\U0000062B"},{0,"isolated",U"\U0000062C"},{0,"final",U"\U0000062C"},{0,"initial",U"\U0000062C"},{0,"medial",U"\U0000062C"},{0,"isolated",U"\U0000062D"},{0,"final",U"\U0000062D"},{0,"initial",U"\U0000062D"},{0,"medial",U"\U0000062D"},{0,"isolated",U"\U0000062E"},{0,"final",U"\U0000062E"},{0,"initial",U"\U0000062E"},{0,"medial",U"\U0000062E"},{0,"isolated",U"\U0000062F"},{0,"final",U"\U0000062F"},{0,"isolated",U"\U00000630"},{0,"final",U"\U00000630"},{0,"isolated",U"\U00000631"},{0,"final",U"\U00000631"},{0,"isolated",U"\U00000632"},{0,"final",U"\U00000632"},{0,"isolated",U"\U00000633"},{0,"final",U"\U00000633"},{0,"initial",U"\U00000633"},{0,"medial",U"\U00000633"},{0,"isolated",U"\U00000634"},{0,"final",U"\U00000634"},{0,"initial",U"\U00000634"},{0,"medial",U"\U00000634"},{0,"isolated",U"\U00000635"},{0,"final",U"\U00000635"},{0,"initial",U"\U00000635"},{0,"medial",U"\U00000635"},{0,"isolated",U"\U00000636"},{0,"final",U"\U00000636"},{0,"initial",U"\U00000636"},{0,"medial",U"\U00000636"},{0,"isolated",U"\U00000637"},{0,"final",U"\U00000637"},{0,"initial",U"\U00000637"},{0,"medial",U"\U00000637"},{0,"isolated",U"\U00000638"},{0,"final",U"\U00000638"},{0,"initial",U"\U00000638"},{0,"medial",U"\U00000638"},{0,"isolated",U"\U00000639"},{0,"final",U"\U00000639"},{0,"initial",U"\U00000639"},{0,"medial",U"\U00000639"},{0,"isolated",U"\U0000063A"},{0,"final",U"\U0000063A"},{0,"initial",U"\U0000063A"},{0,"medial",U"\U0000063A"},{0,"isolated",U"\U00000641"},{0,"final",U"\U00000641"},{0,"initial",U"\U00000641"},{0,"medial",U"\U00000641"},{0,"isolated",U"\U00000642"},{0,"final",U"\U00000642"},{0,"initial",U"\U00000642"},{0,"medial",U"\U00000642"},{0,"isolated",U"\U00000643"},{0,"final",U"\U00000643"},{0,"initial",U"\U00000643"},{0,"medial",U"\U00000643"},{0,"isolated",U"\U00000644"},{0,"final",U"\U00000644"},{0,"initial",U"\U00000644"},{0,"medial",U"\U00000644"},{0,"isolated",U"\U00000645"},{0,"final",U"\U00000645"},{0,"initial",U"\U00000645"},{0,"medial",U"\U00000645"},{0,"isolated",U"\U00000646"},{0,"final",U"\U00000646"},{0,"initial",U"\U00000646"},{0,"medial",U"\U00000646"},{0,"isolated",U"\U00000647"},{0,"final",U"\U00000647"},{0,"initial",U"\U00000647"},{0,"medial",U"\U00000647"},{0,"isolated",U"\U00000648"},{0,"final",U"\U00000648"},{0,"isolated",U"\U00000649"},{0,"final",U"\U00000649"},{0,"isolated",U"\U0000064A"},{0,"final",U"\U0000064A"},{0,"initial",U"\U0000064A"},{0,"medial",U"\U0000064A"},{0,"isolated",U"\U00000644\U00000622"},{0,"final",U"\U00000644\U00000622"},{0,"isolated",U"\U00000644\U00000623"},{0,"final",U"\U00000644\U00000623"},{0,"isolated",U"\U00000644\U00000625"},{0,"final",U"\U00000644\U00000625"},{0,"isolated",U"\U00000644\U00000627"},{0,"final",U"\U00000644\U00000627"},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _510[] = { {0,0,0},{0,"wide",U"\U00000021"},{0,"wide",U"\U00000022"},{0,"wide",U"\U00000023"},{0,"wide",U"\U00000024"},{0,"wide",U"\U00000025"},{0,"wide",U"\U00000026"},{0,"wide",U"\U00000027"},{0,"wide",U"\U00000028"},{0,"wide",U"\U00000029"},{0,"wide",U"\U0000002A"},{0,"wide",U"\U0000002B"},{0,"wide",U"\U0000002C"},{0,"wide",U"\U0000002D"},{0,"wide",U"\U0000002E"},{0,"wide",U"\U0000002F"},{0,"wide",U"\U00000030"},{0,"wide",U"\U00000031"},{0,"wide",U"\U00000032"},{0,"wide",U"\U00000033"},{0,"wide",U"\U00000034"},{0,"wide",U"\U00000035"},{0,"wide",U"\U00000036"},{0,"wide",U"\U00000037"},{0,"wide",U"\U00000038"},{0,"wide",U"\U00000039"},{0,"wide",U"\U0000003A"},{0,"wide",U"\U0000003B"},{0,"wide",U"\U0000003C"},{0,"wide",U"\U0000003D"},{0,"wide",U"\U0000003E"},{0,"wide",U"\U0000003F"},{0,"wide",U"\U00000040"},{0,"wide",U"\U00000041"},{0,"wide",U"\U00000042"},{0,"wide",U"\U00000043"},{0,"wide",U"\U00000044"},{0,"wide",U"\U00000045"},{0,"wide",U"\U00000046"},{0,"wide",U"\U00000047"},{0,"wide",U"\U00000048"},{0,"wide",U"\U00000049"},{0,"wide",U"\U0000004A"},{0,"wide",U"\U0000004B"},{0,"wide",U"\U0000004C"},{0,"wide",U"\U0000004D"},{0,"wide",U"\U0000004E"},{0,"wide",U"\U0000004F"},{0,"wide",U"\U00000050"},{0,"wide",U"\U00000051"},{0,"wide",U"\U00000052"},{0,"wide",U"\U00000053"},{0,"wide",U"\U00000054"},{0,"wide",U"\U00000055"},{0,"wide",U"\U00000056"},{0,"wide",U"\U00000057"},{0,"wide",U"\U00000058"},{0,"wide",U"\U00000059"},{0,"wide",U"\U0000005A"},{0,"wide",U"\U0000005B"},{0,"wide",U"\U0000005C"},{0,"wide",U"\U0000005D"},{0,"wide",U"\U0000005E"},{0,"wide",U"\U0000005F"},{0,"wide",U"\U00000060"},{0,"wide",U"\U00000061"},{0,"wide",U"\U00000062"},{0,"wide",U"\U00000063"},{0,"wide",U"\U00000064"},{0,"wide",U"\U00000065"},{0,"wide",U"\U00000066"},{0,"wide",U"\U00000067"},{0,"wide",U"\U00000068"},{0,"wide",U"\U00000069"},{0,"wide",U"\U0000006A"},{0,"wide",U"\U0000006B"},{0,"wide",U"\U0000006C"},{0,"wide",U"\U0000006D"},{0,"wide",U"\U0000006E"},{0,"wide",U"\U0000006F"},{0,"wide",U"\U00000070"},{0,"wide",U"\U00000071"},{0,"wide",U"\U00000072"},{0,"wide",U"\U00000073"},{0,"wide",U"\U00000074"},{0,"wide",U"\U00000075"},{0,"wide",U"\U00000076"},{0,"wide",U"\U00000077"},{0,"wide",U"\U00000078"},{0,"wide",U"\U00000079"},{0,"wide",U"\U0000007A"},{0,"wide",U"\U0000007B"},{0,"wide",U"\U0000007C"},{0,"wide",U"\U0000007D"},{0,"wide",U"\U0000007E"},{0,"wide",U"\U00002985"},{0,"wide",U"\U00002986"},{0,"narrow",U"\U00003002"},{0,"narrow",U"\U0000300C"},{0,"narrow",U"\U0000300D"},{0,"narrow",U"\U00003001"},{0,"narrow",U"\U000030FB"},{0,"narrow",U"\U000030F2"},{0,"narrow",U"\U000030A1"},{0,"narrow",U"\U000030A3"},{0,"narrow",U"\U000030A5"},{0,"narrow",U"\U000030A7"},{0,"narrow",U"\U000030A9"},{0,"narrow",U"\U000030E3"},{0,"narrow",U"\U000030E5"},{0,"narrow",U"\U000030E7"},{0,"narrow",U"\U000030C3"},{0,"narrow",U"\U000030FC"},{0,"narrow",U"\U000030A2"},{0,"narrow",U"\U000030A4"},{0,"narrow",U"\U000030A6"},{0,"narrow",U"\U000030A8"},{0,"narrow",U"\U000030AA"},{0,"narrow",U"\U000030AB"},{0,"narrow",U"\U000030AD"},{0,"narrow",U"\U000030AF"},{0,"narrow",U"\U000030B1"},{0,"narrow",U"\U000030B3"},{0,"narrow",U"\U000030B5"},{0,"narrow",U"\U000030B7"},{0,"narrow",U"\U000030B9"},{0,"narrow",U"\U000030BB"},{0,"narrow",U"\U000030BD"}, };
static const NormalizationProperties _511[] = { {0,"narrow",U"\U000030BF"},{0,"narrow",U"\U000030C1"},{0,"narrow",U"\U000030C4"},{0,"narrow",U"\U000030C6"},{0,"narrow",U"\U000030C8"},{0,"narrow",U"\U000030CA"},{0,"narrow",U"\U000030CB"},{0,"narrow",U"\U000030CC"},{0,"narrow",U"\U000030CD"},{0,"narrow",U"\U000030CE"},{0,"narrow",U"\U000030CF"},{0,"narrow",U"\U000030D2"},{0,"narrow",U"\U000030D5"},{0,"narrow",U"\U000030D8"},{0,"narrow",U"\U000030DB"},{0,"narrow",U"\U000030DE"},{0,"narrow",U"\U000030DF"},{0,"narrow",U"\U000030E0"},{0,"narrow",U"\U000030E1"},{0,"narrow",U"\U000030E2"},{0,"narrow",U"\U000030E4"},{0,"narrow",U"\U000030E6"},{0,"narrow",U"\U000030E8"},{0,"narrow",U"\U000030E9"},{0,"narrow",U"\U000030EA"},{0,"narrow",U"\U000030EB"},{0,"narrow",U"\U000030EC"},{0,"narrow",U"\U000030ED"},{0,"narrow",U"\U000030EF"},{0,"narrow",U"\U000030F3"},{0,"narrow",U"\U00003099"},{0,"narrow",U"\U0000309A"},{0,"narrow",U"\U00003164"},{0,"narrow",U"\U00003131"},{0,"narrow",U"\U00003132"},{0,"narrow",U"\U00003133"},{0,"narrow",U"\U00003134"},{0,"narrow",U"\U00003135"},{0,"narrow",U"\U00003136"},{0,"narrow",U"\U00003137"},{0,"narrow",U"\U00003138"},{0,"narrow",U"\U00003139"},{0,"narrow",U"\U0000313A"},{0,"narrow",U"\U0000313B"},{0,"narrow",U"\U0000313C"},{0,"narrow",U"\U0000313D"},{0,"narrow",U"\U0000313E"},{0,"narrow",U"\U0000313F"},{0,"narrow",U"\U00003140"},{0,"narrow",U"\U00003141"},{0,"narrow",U"\U00003142"},{0,"narrow",U"\U00003143"},{0,"narrow",U"\U00003144"},{0,"narrow",U"\U00003145"},{0,"narrow",U"\U00003146"},{0,"narrow",U"\U00003147"},{0,"narrow",U"\U00003148"},{0,"narrow",U"\U00003149"},{0,"narrow",U"\U0000314A"},{0,"narrow",U"\U0000314B"},{0,"narrow",U"\U0000314C"},{0,"narrow",U"\U0000314D"},{0,"narrow",U"\U0000314E"},{0,0,0},{0,0,0},{0,0,0},{0,"narrow",U"\U0000314F"},{0,"narrow",U"\U00003150"},{0,"narrow",U"\U00003151"},{0,"narrow",U"\U00003152"},{0,"narrow",U"\U00003153"},{0,"narrow",U"\U00003154"},{0,0,0},{0,0,0},{0,"narrow",U"\U00003155"},{0,"narrow",U"\U00003156"},{0,"narrow",U"\U00003157"},{0,"narrow",U"\U00003158"},{0,"narrow",U"\U00003159"},{0,"narrow",U"\U0000315A"},{0,0,0},{0,0,0},{0,"narrow",U"\U0000315B"},{0,"narrow",U"\U0000315C"},{0,"narrow",U"\U0000315D"},{0,"narrow",U"\U0000315E"},{0,"narrow",U"\U0000315F"},{0,"narrow",U"\U00003160"},{0,0,0},{0,0,0},{0,"narrow",U"\U00003161"},{0,"narrow",U"\U00003162"},{0,"narrow",U"\U00003163"},{0,0,0},{0,0,0},{0,0,0},{0,"wide",U"\U000000A2"},{0,"wide",U"\U000000A3"},{0,"wide",U"\U000000AC"},{0,"wide",U"\U000000AF"},{0,"wide",U"\U000000A6"},{0,"wide",U"\U000000A5"},{0,"wide",U"\U000020A9"},{0,0,0},{0,"narrow",U"\U00002502"},{0,"narrow",U"\U00002190"},{0,"narrow",U"\U00002191"},{0,"narrow",U"\U00002192"},{0,"narrow",U"\U00002193"},{0,"narrow",U"\U000025A0"},{0,"narrow",U"\U000025CB"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _515[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _517[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _518[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _532[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{0,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{1,0,0},{220,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _533[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _538[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _541[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _542[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{220,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{230,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _544[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0},{9,0,0}, };
static const NormalizationProperties _545[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00011099\U000110BA"},{0,0,0},{0,0,U"\U0001109B\U000110BA"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000110A5\U000110BA"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0}, };
static const NormalizationProperties _546[] = { {230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00011131\U00011127"},{0,0,U"\U00011132\U00011127"},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _547[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _548[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _549[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _550[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{7,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00011347\U0001133E"},{0,0,U"\U00011347\U00011357"},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _552[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _553[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000114B9\U000114BA"},{0,0,U"\U000114B9\U000114B0"},{0,0,0},{0,0,U"\U000114B9\U000114BD"},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _555[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U000115B8\U000115AF"},{0,0,U"\U000115B9\U000115AF"},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _556[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _557[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _558[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _560[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _562[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U00011935\U00011930"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _563[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _564[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _565[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _568[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _570[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{7,0,0},{0,0,0},{9,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _571[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{9,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _725[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{1,0,0},{1,0,0},{1,0,0},{1,0,0},{1,0,0},{1,0,0},{1,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _726[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _735[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{6,0,0},{6,0,0},{6,0,0},{6,0,0},{6,0,0},{6,0,0},{6,0,0},{6,0,0},{6,0,0},{6,0,0},{6,0,0},{6,0,0},{6,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _889[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{1,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _930[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0001D157\U0001D165"},{0,0,U"\U0001D158\U0001D165"},{0,0,U"\U0001D15F\U0001D16E"},{0,0,U"\U0001D15F\U0001D16F"},{0,0,U"\U0001D15F\U0001D170"},{0,0,U"\U0001D15F\U0001D171"},{0,0,U"\U0001D15F\U0001D172"},{216,0,0},{216,0,0},{1,0,0},{1,0,0},{1,0,0},{0,0,0},{0,0,0},{0,0,0},{226,0,0},{216,0,0},{216,0,0},{216,0,0},{216,0,0},{216,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0}, };
static const NormalizationProperties _931[] = { {220,0,0},{220,0,0},{220,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{220,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,U"\U0001D1B9\U0001D165"},{0,0,U"\U0001D1BA\U0001D165"},{0,0,U"\U0001D1BB\U0001D16E"},{0,0,U"\U0001D1BC\U0001D16E"},{0,0,U"\U0001D1BB\U0001D16F"},{0,0,U"\U0001D1BC\U0001D16F"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _932[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _936[] = { {0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,"font",U"\U00000048"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,"font",U"\U00000048"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,0,0},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,"font",U"\U00000048"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"}, };
static const NormalizationProperties _937[] = { {0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,0,0},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,0,0},{0,0,0},{0,"font",U"\U00000047"},{0,0,0},{0,0,0},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,0,0},{0,0,0},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,0,0},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,0,0},{0,"font",U"\U00000066"},{0,0,0},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,0,0},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,"font",U"\U00000048"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"}, };
static const NormalizationProperties _938[] = { {0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,0,0},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,0,0},{0,0,0},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,0,0},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,0,0},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,0,0},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,0,0},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,0,0},{0,"font",U"\U0000004F"},{0,0,0},{0,0,0},{0,0,0},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,0,0},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,"font",U"\U00000048"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"}, };
static const NormalizationProperties _939[] = { {0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,"font",U"\U00000048"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,"font",U"\U00000048"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"}, };
static const NormalizationProperties _940[] = { {0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,"font",U"\U00000048"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,"font",U"\U00000048"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"},{0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000041"},{0,"font",U"\U00000042"},{0,"font",U"\U00000043"},{0,"font",U"\U00000044"},{0,"font",U"\U00000045"},{0,"font",U"\U00000046"},{0,"font",U"\U00000047"},{0,"font",U"\U00000048"},{0,"font",U"\U00000049"},{0,"font",U"\U0000004A"},{0,"font",U"\U0000004B"},{0,"font",U"\U0000004C"},{0,"font",U"\U0000004D"},{0,"font",U"\U0000004E"},{0,"font",U"\U0000004F"},{0,"font",U"\U00000050"}, };
static const NormalizationProperties _941[] = { {0,"font",U"\U00000051"},{0,"font",U"\U00000052"},{0,"font",U"\U00000053"},{0,"font",U"\U00000054"},{0,"font",U"\U00000055"},{0,"font",U"\U00000056"},{0,"font",U"\U00000057"},{0,"font",U"\U00000058"},{0,"font",U"\U00000059"},{0,"font",U"\U0000005A"},{0,"font",U"\U00000061"},{0,"font",U"\U00000062"},{0,"font",U"\U00000063"},{0,"font",U"\U00000064"},{0,"font",U"\U00000065"},{0,"font",U"\U00000066"},{0,"font",U"\U00000067"},{0,"font",U"\U00000068"},{0,"font",U"\U00000069"},{0,"font",U"\U0000006A"},{0,"font",U"\U0000006B"},{0,"font",U"\U0000006C"},{0,"font",U"\U0000006D"},{0,"font",U"\U0000006E"},{0,"font",U"\U0000006F"},{0,"font",U"\U00000070"},{0,"font",U"\U00000071"},{0,"font",U"\U00000072"},{0,"font",U"\U00000073"},{0,"font",U"\U00000074"},{0,"font",U"\U00000075"},{0,"font",U"\U00000076"},{0,"font",U"\U00000077"},{0,"font",U"\U00000078"},{0,"font",U"\U00000079"},{0,"font",U"\U0000007A"},{0,"font",U"\U00000131"},{0,"font",U"\U00000237"},{0,0,0},{0,0,0},{0,"font",U"\U00000391"},{0,"font",U"\U00000392"},{0,"font",U"\U00000393"},{0,"font",U"\U00000394"},{0,"font",U"\U00000395"},{0,"font",U"\U00000396"},{0,"font",U"\U00000397"},{0,"font",U"\U00000398"},{0,"font",U"\U00000399"},{0,"font",U"\U0000039A"},{0,"font",U"\U0000039B"},{0,"font",U"\U0000039C"},{0,"font",U"\U0000039D"},{0,"font",U"\U0000039E"},{0,"font",U"\U0000039F"},{0,"font",U"\U000003A0"},{0,"font",U"\U000003A1"},{0,"font",U"\U000003F4"},{0,"font",U"\U000003A3"},{0,"font",U"\U000003A4"},{0,"font",U"\U000003A5"},{0,"font",U"\U000003A6"},{0,"font",U"\U000003A7"},{0,"font",U"\U000003A8"},{0,"font",U"\U000003A9"},{0,"font",U"\U00002207"},{0,"font",U"\U000003B1"},{0,"font",U"\U000003B2"},{0,"font",U"\U000003B3"},{0,"font",U"\U000003B4"},{0,"font",U"\U000003B5"},{0,"font",U"\U000003B6"},{0,"font",U"\U000003B7"},{0,"font",U"\U000003B8"},{0,"font",U"\U000003B9"},{0,"font",U"\U000003BA"},{0,"font",U"\U000003BB"},{0,"font",U"\U000003BC"},{0,"font",U"\U000003BD"},{0,"font",U"\U000003BE"},{0,"font",U"\U000003BF"},{0,"font",U"\U000003C0"},{0,"font",U"\U000003C1"},{0,"font",U"\U000003C2"},{0,"font",U"\U000003C3"},{0,"font",U"\U000003C4"},{0,"font",U"\U000003C5"},{0,"font",U"\U000003C6"},{0,"font",U"\U000003C7"},{0,"font",U"\U000003C8"},{0,"font",U"\U000003C9"},{0,"font",U"\U00002202"},{0,"font",U"\U000003F5"},{0,"font",U"\U000003D1"},{0,"font",U"\U000003F0"},{0,"font",U"\U000003D5"},{0,"font",U"\U000003F1"},{0,"font",U"\U000003D6"},{0,"font",U"\U00000391"},{0,"font",U"\U00000392"},{0,"font",U"\U00000393"},{0,"font",U"\U00000394"},{0,"font",U"\U00000395"},{0,"font",U"\U00000396"},{0,"font",U"\U00000397"},{0,"font",U"\U00000398"},{0,"font",U"\U00000399"},{0,"font",U"\U0000039A"},{0,"font",U"\U0000039B"},{0,"font",U"\U0000039C"},{0,"font",U"\U0000039D"},{0,"font",U"\U0000039E"},{0,"font",U"\U0000039F"},{0,"font",U"\U000003A0"},{0,"font",U"\U000003A1"},{0,"font",U"\U000003F4"},{0,"font",U"\U000003A3"},{0,"font",U"\U000003A4"},{0,"font",U"\U000003A5"},{0,"font",U"\U000003A6"},{0,"font",U"\U000003A7"},{0,"font",U"\U000003A8"},{0,"font",U"\U000003A9"},{0,"font",U"\U00002207"},{0,"font",U"\U000003B1"},{0,"font",U"\U000003B2"},{0,"font",U"\U000003B3"},{0,"font",U"\U000003B4"}, };
static const NormalizationProperties _942[] = { {0,"font",U"\U000003B5"},{0,"font",U"\U000003B6"},{0,"font",U"\U000003B7"},{0,"font",U"\U000003B8"},{0,"font",U"\U000003B9"},{0,"font",U"\U000003BA"},{0,"font",U"\U000003BB"},{0,"font",U"\U000003BC"},{0,"font",U"\U000003BD"},{0,"font",U"\U000003BE"},{0,"font",U"\U000003BF"},{0,"font",U"\U000003C0"},{0,"font",U"\U000003C1"},{0,"font",U"\U000003C2"},{0,"font",U"\U000003C3"},{0,"font",U"\U000003C4"},{0,"font",U"\U000003C5"},{0,"font",U"\U000003C6"},{0,"font",U"\U000003C7"},{0,"font",U"\U000003C8"},{0,"font",U"\U000003C9"},{0,"font",U"\U00002202"},{0,"font",U"\U000003F5"},{0,"font",U"\U000003D1"},{0,"font",U"\U000003F0"},{0,"font",U"\U000003D5"},{0,"font",U"\U000003F1"},{0,"font",U"\U000003D6"},{0,"font",U"\U00000391"},{0,"font",U"\U00000392"},{0,"font",U"\U00000393"},{0,"font",U"\U00000394"},{0,"font",U"\U00000395"},{0,"font",U"\U00000396"},{0,"font",U"\U00000397"},{0,"font",U"\U00000398"},{0,"font",U"\U00000399"},{0,"font",U"\U0000039A"},{0,"font",U"\U0000039B"},{0,"font",U"\U0000039C"},{0,"font",U"\U0000039D"},{0,"font",U"\U0000039E"},{0,"font",U"\U0000039F"},{0,"font",U"\U000003A0"},{0,"font",U"\U000003A1"},{0,"font",U"\U000003F4"},{0,"font",U"\U000003A3"},{0,"font",U"\U000003A4"},{0,"font",U"\U000003A5"},{0,"font",U"\U000003A6"},{0,"font",U"\U000003A7"},{0,"font",U"\U000003A8"},{0,"font",U"\U000003A9"},{0,"font",U"\U00002207"},{0,"font",U"\U000003B1"},{0,"font",U"\U000003B2"},{0,"font",U"\U000003B3"},{0,"font",U"\U000003B4"},{0,"font",U"\U000003B5"},{0,"font",U"\U000003B6"},{0,"font",U"\U000003B7"},{0,"font",U"\U000003B8"},{0,"font",U"\U000003B9"},{0,"font",U"\U000003BA"},{0,"font",U"\U000003BB"},{0,"font",U"\U000003BC"},{0,"font",U"\U000003BD"},{0,"font",U"\U000003BE"},{0,"font",U"\U000003BF"},{0,"font",U"\U000003C0"},{0,"font",U"\U000003C1"},{0,"font",U"\U000003C2"},{0,"font",U"\U000003C3"},{0,"font",U"\U000003C4"},{0,"font",U"\U000003C5"},{0,"font",U"\U000003C6"},{0,"font",U"\U000003C7"},{0,"font",U"\U000003C8"},{0,"font",U"\U000003C9"},{0,"font",U"\U00002202"},{0,"font",U"\U000003F5"},{0,"font",U"\U000003D1"},{0,"font",U"\U000003F0"},{0,"font",U"\U000003D5"},{0,"font",U"\U000003F1"},{0,"font",U"\U000003D6"},{0,"font",U"\U00000391"},{0,"font",U"\U00000392"},{0,"font",U"\U00000393"},{0,"font",U"\U00000394"},{0,"font",U"\U00000395"},{0,"font",U"\U00000396"},{0,"font",U"\U00000397"},{0,"font",U"\U00000398"},{0,"font",U"\U00000399"},{0,"font",U"\U0000039A"},{0,"font",U"\U0000039B"},{0,"font",U"\U0000039C"},{0,"font",U"\U0000039D"},{0,"font",U"\U0000039E"},{0,"font",U"\U0000039F"},{0,"font",U"\U000003A0"},{0,"font",U"\U000003A1"},{0,"font",U"\U000003F4"},{0,"font",U"\U000003A3"},{0,"font",U"\U000003A4"},{0,"font",U"\U000003A5"},{0,"font",U"\U000003A6"},{0,"font",U"\U000003A7"},{0,"font",U"\U000003A8"},{0,"font",U"\U000003A9"},{0,"font",U"\U00002207"},{0,"font",U"\U000003B1"},{0,"font",U"\U000003B2"},{0,"font",U"\U000003B3"},{0,"font",U"\U000003B4"},{0,"font",U"\U000003B5"},{0,"font",U"\U000003B6"},{0,"font",U"\U000003B7"},{0,"font",U"\U000003B8"},{0,"font",U"\U000003B9"},{0,"font",U"\U000003BA"},{0,"font",U"\U000003BB"},{0,"font",U"\U000003BC"},{0,"font",U"\U000003BD"},{0,"font",U"\U000003BE"},{0,"font",U"\U000003BF"},{0,"font",U"\U000003C0"}, };
static const NormalizationProperties _943[] = { {0,"font",U"\U000003C1"},{0,"font",U"\U000003C2"},{0,"font",U"\U000003C3"},{0,"font",U"\U000003C4"},{0,"font",U"\U000003C5"},{0,"font",U"\U000003C6"},{0,"font",U"\U000003C7"},{0,"font",U"\U000003C8"},{0,"font",U"\U000003C9"},{0,"font",U"\U00002202"},{0,"font",U"\U000003F5"},{0,"font",U"\U000003D1"},{0,"font",U"\U000003F0"},{0,"font",U"\U000003D5"},{0,"font",U"\U000003F1"},{0,"font",U"\U000003D6"},{0,"font",U"\U00000391"},{0,"font",U"\U00000392"},{0,"font",U"\U00000393"},{0,"font",U"\U00000394"},{0,"font",U"\U00000395"},{0,"font",U"\U00000396"},{0,"font",U"\U00000397"},{0,"font",U"\U00000398"},{0,"font",U"\U00000399"},{0,"font",U"\U0000039A"},{0,"font",U"\U0000039B"},{0,"font",U"\U0000039C"},{0,"font",U"\U0000039D"},{0,"font",U"\U0000039E"},{0,"font",U"\U0000039F"},{0,"font",U"\U000003A0"},{0,"font",U"\U000003A1"},{0,"font",U"\U000003F4"},{0,"font",U"\U000003A3"},{0,"font",U"\U000003A4"},{0,"font",U"\U000003A5"},{0,"font",U"\U000003A6"},{0,"font",U"\U000003A7"},{0,"font",U"\U000003A8"},{0,"font",U"\U000003A9"},{0,"font",U"\U00002207"},{0,"font",U"\U000003B1"},{0,"font",U"\U000003B2"},{0,"font",U"\U000003B3"},{0,"font",U"\U000003B4"},{0,"font",U"\U000003B5"},{0,"font",U"\U000003B6"},{0,"font",U"\U000003B7"},{0,"font",U"\U000003B8"},{0,"font",U"\U000003B9"},{0,"font",U"\U000003BA"},{0,"font",U"\U000003BB"},{0,"font",U"\U000003BC"},{0,"font",U"\U000003BD"},{0,"font",U"\U000003BE"},{0,"font",U"\U000003BF"},{0,"font",U"\U000003C0"},{0,"font",U"\U000003C1"},{0,"font",U"\U000003C2"},{0,"font",U"\U000003C3"},{0,"font",U"\U000003C4"},{0,"font",U"\U000003C5"},{0,"font",U"\U000003C6"},{0,"font",U"\U000003C7"},{0,"font",U"\U000003C8"},{0,"font",U"\U000003C9"},{0,"font",U"\U00002202"},{0,"font",U"\U000003F5"},{0,"font",U"\U000003D1"},{0,"font",U"\U000003F0"},{0,"font",U"\U000003D5"},{0,"font",U"\U000003F1"},{0,"font",U"\U000003D6"},{0,"font",U"\U000003DC"},{0,"font",U"\U000003DD"},{0,0,0},{0,0,0},{0,"font",U"\U00000030"},{0,"font",U"\U00000031"},{0,"font",U"\U00000032"},{0,"font",U"\U00000033"},{0,"font",U"\U00000034"},{0,"font",U"\U00000035"},{0,"font",U"\U00000036"},{0,"font",U"\U00000037"},{0,"font",U"\U00000038"},{0,"font",U"\U00000039"},{0,"font",U"\U00000030"},{0,"font",U"\U00000031"},{0,"font",U"\U00000032"},{0,"font",U"\U00000033"},{0,"font",U"\U00000034"},{0,"font",U"\U00000035"},{0,"font",U"\U00000036"},{0,"font",U"\U00000037"},{0,"font",U"\U00000038"},{0,"font",U"\U00000039"},{0,"font",U"\U00000030"},{0,"font",U"\U00000031"},{0,"font",U"\U00000032"},{0,"font",U"\U00000033"},{0,"font",U"\U00000034"},{0,"font",U"\U00000035"},{0,"font",U"\U00000036"},{0,"font",U"\U00000037"},{0,"font",U"\U00000038"},{0,"font",U"\U00000039"},{0,"font",U"\U00000030"},{0,"font",U"\U00000031"},{0,"font",U"\U00000032"},{0,"font",U"\U00000033"},{0,"font",U"\U00000034"},{0,"font",U"\U00000035"},{0,"font",U"\U00000036"},{0,"font",U"\U00000037"},{0,"font",U"\U00000038"},{0,"font",U"\U00000039"},{0,"font",U"\U00000030"},{0,"font",U"\U00000031"},{0,"font",U"\U00000032"},{0,"font",U"\U00000033"},{0,"font",U"\U00000034"},{0,"font",U"\U00000035"},{0,"font",U"\U00000036"},{0,"font",U"\U00000037"},{0,"font",U"\U00000038"},{0,"font",U"\U00000039"}, };
static const NormalizationProperties _949[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0}, };
static const NormalizationProperties _960[] = { {230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _962[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _965[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _977[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{220,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _978[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{7,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _988[] = { {0,"font",U"\U00000627"},{0,"font",U"\U00000628"},{0,"font",U"\U0000062C"},{0,"font",U"\U0000062F"},{0,0,0},{0,"font",U"\U00000648"},{0,"font",U"\U00000632"},{0,"font",U"\U0000062D"},{0,"font",U"\U00000637"},{0,"font",U"\U0000064A"},{0,"font",U"\U00000643"},{0,"font",U"\U00000644"},{0,"font",U"\U00000645"},{0,"font",U"\U00000646"},{0,"font",U"\U00000633"},{0,"font",U"\U00000639"},{0,"font",U"\U00000641"},{0,"font",U"\U00000635"},{0,"font",U"\U00000642"},{0,"font",U"\U00000631"},{0,"font",U"\U00000634"},{0,"font",U"\U0000062A"},{0,"font",U"\U0000062B"},{0,"font",U"\U0000062E"},{0,"font",U"\U00000630"},{0,"font",U"\U00000636"},{0,"font",U"\U00000638"},{0,"font",U"\U0000063A"},{0,"font",U"\U0000066E"},{0,"font",U"\U000006BA"},{0,"font",U"\U000006A1"},{0,"font",U"\U0000066F"},{0,0,0},{0,"font",U"\U00000628"},{0,"font",U"\U0000062C"},{0,0,0},{0,"font",U"\U00000647"},{0,0,0},{0,0,0},{0,"font",U"\U0000062D"},{0,0,0},{0,"font",U"\U0000064A"},{0,"font",U"\U00000643"},{0,"font",U"\U00000644"},{0,"font",U"\U00000645"},{0,"font",U"\U00000646"},{0,"font",U"\U00000633"},{0,"font",U"\U00000639"},{0,"font",U"\U00000641"},{0,"font",U"\U00000635"},{0,"font",U"\U00000642"},{0,0,0},{0,"font",U"\U00000634"},{0,"font",U"\U0000062A"},{0,"font",U"\U0000062B"},{0,"font",U"\U0000062E"},{0,0,0},{0,"font",U"\U00000636"},{0,0,0},{0,"font",U"\U0000063A"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"font",U"\U0000062C"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"font",U"\U0000062D"},{0,0,0},{0,"font",U"\U0000064A"},{0,0,0},{0,"font",U"\U00000644"},{0,0,0},{0,"font",U"\U00000646"},{0,"font",U"\U00000633"},{0,"font",U"\U00000639"},{0,0,0},{0,"font",U"\U00000635"},{0,"font",U"\U00000642"},{0,0,0},{0,"font",U"\U00000634"},{0,0,0},{0,0,0},{0,"font",U"\U0000062E"},{0,0,0},{0,"font",U"\U00000636"},{0,0,0},{0,"font",U"\U0000063A"},{0,0,0},{0,"font",U"\U000006BA"},{0,0,0},{0,"font",U"\U0000066F"},{0,0,0},{0,"font",U"\U00000628"},{0,"font",U"\U0000062C"},{0,0,0},{0,"font",U"\U00000647"},{0,0,0},{0,0,0},{0,"font",U"\U0000062D"},{0,"font",U"\U00000637"},{0,"font",U"\U0000064A"},{0,"font",U"\U00000643"},{0,0,0},{0,"font",U"\U00000645"},{0,"font",U"\U00000646"},{0,"font",U"\U00000633"},{0,"font",U"\U00000639"},{0,"font",U"\U00000641"},{0,"font",U"\U00000635"},{0,"font",U"\U00000642"},{0,0,0},{0,"font",U"\U00000634"},{0,"font",U"\U0000062A"},{0,"font",U"\U0000062B"},{0,"font",U"\U0000062E"},{0,0,0},{0,"font",U"\U00000636"},{0,"font",U"\U00000638"},{0,"font",U"\U0000063A"},{0,"font",U"\U0000066E"},{0,0,0},{0,"font",U"\U000006A1"},{0,0,0}, };
static const NormalizationProperties _989[] = { {0,"font",U"\U00000627"},{0,"font",U"\U00000628"},{0,"font",U"\U0000062C"},{0,"font",U"\U0000062F"},{0,"font",U"\U00000647"},{0,"font",U"\U00000648"},{0,"font",U"\U00000632"},{0,"font",U"\U0000062D"},{0,"font",U"\U00000637"},{0,"font",U"\U0000064A"},{0,0,0},{0,"font",U"\U00000644"},{0,"font",U"\U00000645"},{0,"font",U"\U00000646"},{0,"font",U"\U00000633"},{0,"font",U"\U00000639"},{0,"font",U"\U00000641"},{0,"font",U"\U00000635"},{0,"font",U"\U00000642"},{0,"font",U"\U00000631"},{0,"font",U"\U00000634"},{0,"font",U"\U0000062A"},{0,"font",U"\U0000062B"},{0,"font",U"\U0000062E"},{0,"font",U"\U00000630"},{0,"font",U"\U00000636"},{0,"font",U"\U00000638"},{0,"font",U"\U0000063A"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"font",U"\U00000628"},{0,"font",U"\U0000062C"},{0,"font",U"\U0000062F"},{0,0,0},{0,"font",U"\U00000648"},{0,"font",U"\U00000632"},{0,"font",U"\U0000062D"},{0,"font",U"\U00000637"},{0,"font",U"\U0000064A"},{0,0,0},{0,"font",U"\U00000644"},{0,"font",U"\U00000645"},{0,"font",U"\U00000646"},{0,"font",U"\U00000633"},{0,"font",U"\U00000639"},{0,"font",U"\U00000641"},{0,"font",U"\U00000635"},{0,"font",U"\U00000642"},{0,"font",U"\U00000631"},{0,"font",U"\U00000634"},{0,"font",U"\U0000062A"},{0,"font",U"\U0000062B"},{0,"font",U"\U0000062E"},{0,"font",U"\U00000630"},{0,"font",U"\U00000636"},{0,"font",U"\U00000638"},{0,"font",U"\U0000063A"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _994[] = { {0,"compat",U"\U00000030\U0000002E"},{0,"compat",U"\U00000030\U0000002C"},{0,"compat",U"\U00000031\U0000002C"},{0,"compat",U"\U00000032\U0000002C"},{0,"compat",U"\U00000033\U0000002C"},{0,"compat",U"\U00000034\U0000002C"},{0,"compat",U"\U00000035\U0000002C"},{0,"compat",U"\U00000036\U0000002C"},{0,"compat",U"\U00000037\U0000002C"},{0,"compat",U"\U00000038\U0000002C"},{0,"compat",U"\U00000039\U0000002C"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00000028\U00000041\U00000029"},{0,"compat",U"\U00000028\U00000042\U00000029"},{0,"compat",U"\U00000028\U00000043\U00000029"},{0,"compat",U"\U00000028\U00000044\U00000029"},{0,"compat",U"\U00000028\U00000045\U00000029"},{0,"compat",U"\U00000028\U00000046\U00000029"},{0,"compat",U"\U00000028\U00000047\U00000029"},{0,"compat",U"\U00000028\U00000048\U00000029"},{0,"compat",U"\U00000028\U00000049\U00000029"},{0,"compat",U"\U00000028\U0000004A\U00000029"},{0,"compat",U"\U00000028\U0000004B\U00000029"},{0,"compat",U"\U00000028\U0000004C\U00000029"},{0,"compat",U"\U00000028\U0000004D\U00000029"},{0,"compat",U"\U00000028\U0000004E\U00000029"},{0,"compat",U"\U00000028\U0000004F\U00000029"},{0,"compat",U"\U00000028\U00000050\U00000029"},{0,"compat",U"\U00000028\U00000051\U00000029"},{0,"compat",U"\U00000028\U00000052\U00000029"},{0,"compat",U"\U00000028\U00000053\U00000029"},{0,"compat",U"\U00000028\U00000054\U00000029"},{0,"compat",U"\U00000028\U00000055\U00000029"},{0,"compat",U"\U00000028\U00000056\U00000029"},{0,"compat",U"\U00000028\U00000057\U00000029"},{0,"compat",U"\U00000028\U00000058\U00000029"},{0,"compat",U"\U00000028\U00000059\U00000029"},{0,"compat",U"\U00000028\U0000005A\U00000029"},{0,"compat",U"\U00003014\U00000053\U00003015"},{0,"circle",U"\U00000043"},{0,"circle",U"\U00000052"},{0,"circle",U"\U00000043\U00000044"},{0,"circle",U"\U00000057\U0000005A"},{0,0,0},{0,"square",U"\U00000041"},{0,"square",U"\U00000042"},{0,"square",U"\U00000043"},{0,"square",U"\U00000044"},{0,"square",U"\U00000045"},{0,"square",U"\U00000046"},{0,"square",U"\U00000047"},{0,"square",U"\U00000048"},{0,"square",U"\U00000049"},{0,"square",U"\U0000004A"},{0,"square",U"\U0000004B"},{0,"square",U"\U0000004C"},{0,"square",U"\U0000004D"},{0,"square",U"\U0000004E"},{0,"square",U"\U0000004F"},{0,"square",U"\U00000050"},{0,"square",U"\U00000051"},{0,"square",U"\U00000052"},{0,"square",U"\U00000053"},{0,"square",U"\U00000054"},{0,"square",U"\U00000055"},{0,"square",U"\U00000056"},{0,"square",U"\U00000057"},{0,"square",U"\U00000058"},{0,"square",U"\U00000059"},{0,"square",U"\U0000005A"},{0,"square",U"\U00000048\U00000056"},{0,"square",U"\U0000004D\U00000056"},{0,"square",U"\U00000053\U00000044"},{0,"square",U"\U00000053\U00000053"},{0,"square",U"\U00000050\U00000050\U00000056"},{0,"square",U"\U00000057\U00000043"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"super",U"\U0000004D\U00000043"},{0,"super",U"\U0000004D\U00000044"},{0,"super",U"\U0000004D\U00000052"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _995[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"square",U"\U00000044\U0000004A"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _996[] = { {0,"square",U"\U0000307B\U0000304B"},{0,"square",U"\U000030B3\U000030B3"},{0,"square",U"\U000030B5"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"square",U"\U0000624B"},{0,"square",U"\U00005B57"},{0,"square",U"\U000053CC"},{0,"square",U"\U000030C7"},{0,"square",U"\U00004E8C"},{0,"square",U"\U0000591A"},{0,"square",U"\U000089E3"},{0,"square",U"\U00005929"},{0,"square",U"\U00004EA4"},{0,"square",U"\U00006620"},{0,"square",U"\U00007121"},{0,"square",U"\U00006599"},{0,"square",U"\U0000524D"},{0,"square",U"\U00005F8C"},{0,"square",U"\U0000518D"},{0,"square",U"\U000065B0"},{0,"square",U"\U0000521D"},{0,"square",U"\U00007D42"},{0,"square",U"\U0000751F"},{0,"square",U"\U00008CA9"},{0,"square",U"\U000058F0"},{0,"square",U"\U00005439"},{0,"square",U"\U00006F14"},{0,"square",U"\U00006295"},{0,"square",U"\U00006355"},{0,"square",U"\U00004E00"},{0,"square",U"\U00004E09"},{0,"square",U"\U0000904A"},{0,"square",U"\U00005DE6"},{0,"square",U"\U00004E2D"},{0,"square",U"\U000053F3"},{0,"square",U"\U00006307"},{0,"square",U"\U00008D70"},{0,"square",U"\U00006253"},{0,"square",U"\U00007981"},{0,"square",U"\U00007A7A"},{0,"square",U"\U00005408"},{0,"square",U"\U00006E80"},{0,"square",U"\U00006709"},{0,"square",U"\U00006708"},{0,"square",U"\U00007533"},{0,"square",U"\U00005272"},{0,"square",U"\U000055B6"},{0,"square",U"\U0000914D"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"compat",U"\U00003014\U0000672C\U00003015"},{0,"compat",U"\U00003014\U00004E09\U00003015"},{0,"compat",U"\U00003014\U00004E8C\U00003015"},{0,"compat",U"\U00003014\U00005B89\U00003015"},{0,"compat",U"\U00003014\U000070B9\U00003015"},{0,"compat",U"\U00003014\U00006253\U00003015"},{0,"compat",U"\U00003014\U000076D7\U00003015"},{0,"compat",U"\U00003014\U000052DD\U00003015"},{0,"compat",U"\U00003014\U00006557\U00003015"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"circle",U"\U00005F97"},{0,"circle",U"\U000053EF"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _1015[] = { {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,"font",U"\U00000030"},{0,"font",U"\U00000031"},{0,"font",U"\U00000032"},{0,"font",U"\U00000033"},{0,"font",U"\U00000034"},{0,"font",U"\U00000035"},{0,"font",U"\U00000036"},{0,"font",U"\U00000037"},{0,"font",U"\U00000038"},{0,"font",U"\U00000039"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties _1520[] = { {0,0,U"\U00004E3D"},{0,0,U"\U00004E38"},{0,0,U"\U00004E41"},{0,0,U"\U00020122"},{0,0,U"\U00004F60"},{0,0,U"\U00004FAE"},{0,0,U"\U00004FBB"},{0,0,U"\U00005002"},{0,0,U"\U0000507A"},{0,0,U"\U00005099"},{0,0,U"\U000050E7"},{0,0,U"\U000050CF"},{0,0,U"\U0000349E"},{0,0,U"\U0002063A"},{0,0,U"\U0000514D"},{0,0,U"\U00005154"},{0,0,U"\U00005164"},{0,0,U"\U00005177"},{0,0,U"\U0002051C"},{0,0,U"\U000034B9"},{0,0,U"\U00005167"},{0,0,U"\U0000518D"},{0,0,U"\U0002054B"},{0,0,U"\U00005197"},{0,0,U"\U000051A4"},{0,0,U"\U00004ECC"},{0,0,U"\U000051AC"},{0,0,U"\U000051B5"},{0,0,U"\U000291DF"},{0,0,U"\U000051F5"},{0,0,U"\U00005203"},{0,0,U"\U000034DF"},{0,0,U"\U0000523B"},{0,0,U"\U00005246"},{0,0,U"\U00005272"},{0,0,U"\U00005277"},{0,0,U"\U00003515"},{0,0,U"\U000052C7"},{0,0,U"\U000052C9"},{0,0,U"\U000052E4"},{0,0,U"\U000052FA"},{0,0,U"\U00005305"},{0,0,U"\U00005306"},{0,0,U"\U00005317"},{0,0,U"\U00005349"},{0,0,U"\U00005351"},{0,0,U"\U0000535A"},{0,0,U"\U00005373"},{0,0,U"\U0000537D"},{0,0,U"\U0000537F"},{0,0,U"\U0000537F"},{0,0,U"\U0000537F"},{0,0,U"\U00020A2C"},{0,0,U"\U00007070"},{0,0,U"\U000053CA"},{0,0,U"\U000053DF"},{0,0,U"\U00020B63"},{0,0,U"\U000053EB"},{0,0,U"\U000053F1"},{0,0,U"\U00005406"},{0,0,U"\U0000549E"},{0,0,U"\U00005438"},{0,0,U"\U00005448"},{0,0,U"\U00005468"},{0,0,U"\U000054A2"},{0,0,U"\U000054F6"},{0,0,U"\U00005510"},{0,0,U"\U00005553"},{0,0,U"\U00005563"},{0,0,U"\U00005584"},{0,0,U"\U00005584"},{0,0,U"\U00005599"},{0,0,U"\U000055AB"},{0,0,U"\U000055B3"},{0,0,U"\U000055C2"},{0,0,U"\U00005716"},{0,0,U"\U00005606"},{0,0,U"\U00005717"},{0,0,U"\U00005651"},{0,0,U"\U00005674"},{0,0,U"\U00005207"},{0,0,U"\U000058EE"},{0,0,U"\U000057CE"},{0,0,U"\U000057F4"},{0,0,U"\U0000580D"},{0,0,U"\U0000578B"},{0,0,U"\U00005832"},{0,0,U"\U00005831"},{0,0,U"\U000058AC"},{0,0,U"\U000214E4"},{0,0,U"\U000058F2"},{0,0,U"\U000058F7"},{0,0,U"\U00005906"},{0,0,U"\U0000591A"},{0,0,U"\U00005922"},{0,0,U"\U00005962"},{0,0,U"\U000216A8"},{0,0,U"\U000216EA"},{0,0,U"\U000059EC"},{0,0,U"\U00005A1B"},{0,0,U"\U00005A27"},{0,0,U"\U000059D8"},{0,0,U"\U00005A66"},{0,0,U"\U000036EE"},{0,0,U"\U000036FC"},{0,0,U"\U00005B08"},{0,0,U"\U00005B3E"},{0,0,U"\U00005B3E"},{0,0,U"\U000219C8"},{0,0,U"\U00005BC3"},{0,0,U"\U00005BD8"},{0,0,U"\U00005BE7"},{0,0,U"\U00005BF3"},{0,0,U"\U00021B18"},{0,0,U"\U00005BFF"},{0,0,U"\U00005C06"},{0,0,U"\U00005F53"},{0,0,U"\U00005C22"},{0,0,U"\U00003781"},{0,0,U"\U00005C60"},{0,0,U"\U00005C6E"},{0,0,U"\U00005CC0"},{0,0,U"\U00005C8D"},{0,0,U"\U00021DE4"},{0,0,U"\U00005D43"},{0,0,U"\U00021DE6"},{0,0,U"\U00005D6E"},{0,0,U"\U00005D6B"}, };
static const NormalizationProperties _1521[] = { {0,0,U"\U00005D7C"},{0,0,U"\U00005DE1"},{0,0,U"\U00005DE2"},{0,0,U"\U0000382F"},{0,0,U"\U00005DFD"},{0,0,U"\U00005E28"},{0,0,U"\U00005E3D"},{0,0,U"\U00005E69"},{0,0,U"\U00003862"},{0,0,U"\U00022183"},{0,0,U"\U0000387C"},{0,0,U"\U00005EB0"},{0,0,U"\U00005EB3"},{0,0,U"\U00005EB6"},{0,0,U"\U00005ECA"},{0,0,U"\U0002A392"},{0,0,U"\U00005EFE"},{0,0,U"\U00022331"},{0,0,U"\U00022331"},{0,0,U"\U00008201"},{0,0,U"\U00005F22"},{0,0,U"\U00005F22"},{0,0,U"\U000038C7"},{0,0,U"\U000232B8"},{0,0,U"\U000261DA"},{0,0,U"\U00005F62"},{0,0,U"\U00005F6B"},{0,0,U"\U000038E3"},{0,0,U"\U00005F9A"},{0,0,U"\U00005FCD"},{0,0,U"\U00005FD7"},{0,0,U"\U00005FF9"},{0,0,U"\U00006081"},{0,0,U"\U0000393A"},{0,0,U"\U0000391C"},{0,0,U"\U00006094"},{0,0,U"\U000226D4"},{0,0,U"\U000060C7"},{0,0,U"\U00006148"},{0,0,U"\U0000614C"},{0,0,U"\U0000614E"},{0,0,U"\U0000614C"},{0,0,U"\U0000617A"},{0,0,U"\U0000618E"},{0,0,U"\U000061B2"},{0,0,U"\U000061A4"},{0,0,U"\U000061AF"},{0,0,U"\U000061DE"},{0,0,U"\U000061F2"},{0,0,U"\U000061F6"},{0,0,U"\U00006210"},{0,0,U"\U0000621B"},{0,0,U"\U0000625D"},{0,0,U"\U000062B1"},{0,0,U"\U000062D4"},{0,0,U"\U00006350"},{0,0,U"\U00022B0C"},{0,0,U"\U0000633D"},{0,0,U"\U000062FC"},{0,0,U"\U00006368"},{0,0,U"\U00006383"},{0,0,U"\U000063E4"},{0,0,U"\U00022BF1"},{0,0,U"\U00006422"},{0,0,U"\U000063C5"},{0,0,U"\U000063A9"},{0,0,U"\U00003A2E"},{0,0,U"\U00006469"},{0,0,U"\U0000647E"},{0,0,U"\U0000649D"},{0,0,U"\U00006477"},{0,0,U"\U00003A6C"},{0,0,U"\U0000654F"},{0,0,U"\U0000656C"},{0,0,U"\U0002300A"},{0,0,U"\U000065E3"},{0,0,U"\U000066F8"},{0,0,U"\U00006649"},{0,0,U"\U00003B19"},{0,0,U"\U00006691"},{0,0,U"\U00003B08"},{0,0,U"\U00003AE4"},{0,0,U"\U00005192"},{0,0,U"\U00005195"},{0,0,U"\U00006700"},{0,0,U"\U0000669C"},{0,0,U"\U000080AD"},{0,0,U"\U000043D9"},{0,0,U"\U00006717"},{0,0,U"\U0000671B"},{0,0,U"\U00006721"},{0,0,U"\U0000675E"},{0,0,U"\U00006753"},{0,0,U"\U000233C3"},{0,0,U"\U00003B49"},{0,0,U"\U000067FA"},{0,0,U"\U00006785"},{0,0,U"\U00006852"},{0,0,U"\U00006885"},{0,0,U"\U0002346D"},{0,0,U"\U0000688E"},{0,0,U"\U0000681F"},{0,0,U"\U00006914"},{0,0,U"\U00003B9D"},{0,0,U"\U00006942"},{0,0,U"\U000069A3"},{0,0,U"\U000069EA"},{0,0,U"\U00006AA8"},{0,0,U"\U000236A3"},{0,0,U"\U00006ADB"},{0,0,U"\U00003C18"},{0,0,U"\U00006B21"},{0,0,U"\U000238A7"},{0,0,U"\U00006B54"},{0,0,U"\U00003C4E"},{0,0,U"\U00006B72"},{0,0,U"\U00006B9F"},{0,0,U"\U00006BBA"},{0,0,U"\U00006BBB"},{0,0,U"\U00023A8D"},{0,0,U"\U00021D0B"},{0,0,U"\U00023AFA"},{0,0,U"\U00006C4E"},{0,0,U"\U00023CBC"},{0,0,U"\U00006CBF"},{0,0,U"\U00006CCD"},{0,0,U"\U00006C67"},{0,0,U"\U00006D16"}, };
static const NormalizationProperties _1522[] = { {0,0,U"\U00006D3E"},{0,0,U"\U00006D77"},{0,0,U"\U00006D41"},{0,0,U"\U00006D69"},{0,0,U"\U00006D78"},{0,0,U"\U00006D85"},{0,0,U"\U00023D1E"},{0,0,U"\U00006D34"},{0,0,U"\U00006E2F"},{0,0,U"\U00006E6E"},{0,0,U"\U00003D33"},{0,0,U"\U00006ECB"},{0,0,U"\U00006EC7"},{0,0,U"\U00023ED1"},{0,0,U"\U00006DF9"},{0,0,U"\U00006F6E"},{0,0,U"\U00023F5E"},{0,0,U"\U00023F8E"},{0,0,U"\U00006FC6"},{0,0,U"\U00007039"},{0,0,U"\U0000701E"},{0,0,U"\U0000701B"},{0,0,U"\U00003D96"},{0,0,U"\U0000704A"},{0,0,U"\U0000707D"},{0,0,U"\U00007077"},{0,0,U"\U000070AD"},{0,0,U"\U00020525"},{0,0,U"\U00007145"},{0,0,U"\U00024263"},{0,0,U"\U0000719C"},{0,0,U"\U000243AB"},{0,0,U"\U00007228"},{0,0,U"\U00007235"},{0,0,U"\U00007250"},{0,0,U"\U00024608"},{0,0,U"\U00007280"},{0,0,U"\U00007295"},{0,0,U"\U00024735"},{0,0,U"\U00024814"},{0,0,U"\U0000737A"},{0,0,U"\U0000738B"},{0,0,U"\U00003EAC"},{0,0,U"\U000073A5"},{0,0,U"\U00003EB8"},{0,0,U"\U00003EB8"},{0,0,U"\U00007447"},{0,0,U"\U0000745C"},{0,0,U"\U00007471"},{0,0,U"\U00007485"},{0,0,U"\U000074CA"},{0,0,U"\U00003F1B"},{0,0,U"\U00007524"},{0,0,U"\U00024C36"},{0,0,U"\U0000753E"},{0,0,U"\U00024C92"},{0,0,U"\U00007570"},{0,0,U"\U0002219F"},{0,0,U"\U00007610"},{0,0,U"\U00024FA1"},{0,0,U"\U00024FB8"},{0,0,U"\U00025044"},{0,0,U"\U00003FFC"},{0,0,U"\U00004008"},{0,0,U"\U000076F4"},{0,0,U"\U000250F3"},{0,0,U"\U000250F2"},{0,0,U"\U00025119"},{0,0,U"\U00025133"},{0,0,U"\U0000771E"},{0,0,U"\U0000771F"},{0,0,U"\U0000771F"},{0,0,U"\U0000774A"},{0,0,U"\U00004039"},{0,0,U"\U0000778B"},{0,0,U"\U00004046"},{0,0,U"\U00004096"},{0,0,U"\U0002541D"},{0,0,U"\U0000784E"},{0,0,U"\U0000788C"},{0,0,U"\U000078CC"},{0,0,U"\U000040E3"},{0,0,U"\U00025626"},{0,0,U"\U00007956"},{0,0,U"\U0002569A"},{0,0,U"\U000256C5"},{0,0,U"\U0000798F"},{0,0,U"\U000079EB"},{0,0,U"\U0000412F"},{0,0,U"\U00007A40"},{0,0,U"\U00007A4A"},{0,0,U"\U00007A4F"},{0,0,U"\U0002597C"},{0,0,U"\U00025AA7"},{0,0,U"\U00025AA7"},{0,0,U"\U00007AEE"},{0,0,U"\U00004202"},{0,0,U"\U00025BAB"},{0,0,U"\U00007BC6"},{0,0,U"\U00007BC9"},{0,0,U"\U00004227"},{0,0,U"\U00025C80"},{0,0,U"\U00007CD2"},{0,0,U"\U000042A0"},{0,0,U"\U00007CE8"},{0,0,U"\U00007CE3"},{0,0,U"\U00007D00"},{0,0,U"\U00025F86"},{0,0,U"\U00007D63"},{0,0,U"\U00004301"},{0,0,U"\U00007DC7"},{0,0,U"\U00007E02"},{0,0,U"\U00007E45"},{0,0,U"\U00004334"},{0,0,U"\U00026228"},{0,0,U"\U00026247"},{0,0,U"\U00004359"},{0,0,U"\U000262D9"},{0,0,U"\U00007F7A"},{0,0,U"\U0002633E"},{0,0,U"\U00007F95"},{0,0,U"\U00007FFA"},{0,0,U"\U00008005"},{0,0,U"\U000264DA"},{0,0,U"\U00026523"},{0,0,U"\U00008060"},{0,0,U"\U000265A8"},{0,0,U"\U00008070"}, };
static const NormalizationProperties _1523[] = { {0,0,U"\U0002335F"},{0,0,U"\U000043D5"},{0,0,U"\U000080B2"},{0,0,U"\U00008103"},{0,0,U"\U0000440B"},{0,0,U"\U0000813E"},{0,0,U"\U00005AB5"},{0,0,U"\U000267A7"},{0,0,U"\U000267B5"},{0,0,U"\U00023393"},{0,0,U"\U0002339C"},{0,0,U"\U00008201"},{0,0,U"\U00008204"},{0,0,U"\U00008F9E"},{0,0,U"\U0000446B"},{0,0,U"\U00008291"},{0,0,U"\U0000828B"},{0,0,U"\U0000829D"},{0,0,U"\U000052B3"},{0,0,U"\U000082B1"},{0,0,U"\U000082B3"},{0,0,U"\U000082BD"},{0,0,U"\U000082E6"},{0,0,U"\U00026B3C"},{0,0,U"\U000082E5"},{0,0,U"\U0000831D"},{0,0,U"\U00008363"},{0,0,U"\U000083AD"},{0,0,U"\U00008323"},{0,0,U"\U000083BD"},{0,0,U"\U000083E7"},{0,0,U"\U00008457"},{0,0,U"\U00008353"},{0,0,U"\U000083CA"},{0,0,U"\U000083CC"},{0,0,U"\U000083DC"},{0,0,U"\U00026C36"},{0,0,U"\U00026D6B"},{0,0,U"\U00026CD5"},{0,0,U"\U0000452B"},{0,0,U"\U000084F1"},{0,0,U"\U000084F3"},{0,0,U"\U00008516"},{0,0,U"\U000273CA"},{0,0,U"\U00008564"},{0,0,U"\U00026F2C"},{0,0,U"\U0000455D"},{0,0,U"\U00004561"},{0,0,U"\U00026FB1"},{0,0,U"\U000270D2"},{0,0,U"\U0000456B"},{0,0,U"\U00008650"},{0,0,U"\U0000865C"},{0,0,U"\U00008667"},{0,0,U"\U00008669"},{0,0,U"\U000086A9"},{0,0,U"\U00008688"},{0,0,U"\U0000870E"},{0,0,U"\U000086E2"},{0,0,U"\U00008779"},{0,0,U"\U00008728"},{0,0,U"\U0000876B"},{0,0,U"\U00008786"},{0,0,U"\U000045D7"},{0,0,U"\U000087E1"},{0,0,U"\U00008801"},{0,0,U"\U000045F9"},{0,0,U"\U00008860"},{0,0,U"\U00008863"},{0,0,U"\U00027667"},{0,0,U"\U000088D7"},{0,0,U"\U000088DE"},{0,0,U"\U00004635"},{0,0,U"\U000088FA"},{0,0,U"\U000034BB"},{0,0,U"\U000278AE"},{0,0,U"\U00027966"},{0,0,U"\U000046BE"},{0,0,U"\U000046C7"},{0,0,U"\U00008AA0"},{0,0,U"\U00008AED"},{0,0,U"\U00008B8A"},{0,0,U"\U00008C55"},{0,0,U"\U00027CA8"},{0,0,U"\U00008CAB"},{0,0,U"\U00008CC1"},{0,0,U"\U00008D1B"},{0,0,U"\U00008D77"},{0,0,U"\U00027F2F"},{0,0,U"\U00020804"},{0,0,U"\U00008DCB"},{0,0,U"\U00008DBC"},{0,0,U"\U00008DF0"},{0,0,U"\U000208DE"},{0,0,U"\U00008ED4"},{0,0,U"\U00008F38"},{0,0,U"\U000285D2"},{0,0,U"\U000285ED"},{0,0,U"\U00009094"},{0,0,U"\U000090F1"},{0,0,U"\U00009111"},{0,0,U"\U0002872E"},{0,0,U"\U0000911B"},{0,0,U"\U00009238"},{0,0,U"\U000092D7"},{0,0,U"\U000092D8"},{0,0,U"\U0000927C"},{0,0,U"\U000093F9"},{0,0,U"\U00009415"},{0,0,U"\U00028BFA"},{0,0,U"\U0000958B"},{0,0,U"\U00004995"},{0,0,U"\U000095B7"},{0,0,U"\U00028D77"},{0,0,U"\U000049E6"},{0,0,U"\U000096C3"},{0,0,U"\U00005DB2"},{0,0,U"\U00009723"},{0,0,U"\U00029145"},{0,0,U"\U0002921A"},{0,0,U"\U00004A6E"},{0,0,U"\U00004A76"},{0,0,U"\U000097E0"},{0,0,U"\U0002940A"},{0,0,U"\U00004AB2"},{0,0,U"\U00029496"},{0,0,U"\U0000980B"},{0,0,U"\U0000980B"}, };
static const NormalizationProperties _1524[] = { {0,0,U"\U00009829"},{0,0,U"\U000295B6"},{0,0,U"\U000098E2"},{0,0,U"\U00004B33"},{0,0,U"\U00009929"},{0,0,U"\U000099A7"},{0,0,U"\U000099C2"},{0,0,U"\U000099FE"},{0,0,U"\U00004BCE"},{0,0,U"\U00029B30"},{0,0,U"\U00009B12"},{0,0,U"\U00009C40"},{0,0,U"\U00009CFD"},{0,0,U"\U00004CCE"},{0,0,U"\U00004CED"},{0,0,U"\U00009D67"},{0,0,U"\U0002A0CE"},{0,0,U"\U00004CF8"},{0,0,U"\U0002A105"},{0,0,U"\U0002A20E"},{0,0,U"\U0002A291"},{0,0,U"\U00009EBB"},{0,0,U"\U00004D56"},{0,0,U"\U00009EF9"},{0,0,U"\U00009EFE"},{0,0,U"\U00009F05"},{0,0,U"\U00009F0F"},{0,0,U"\U00009F16"},{0,0,U"\U00009F3B"},{0,0,U"\U0002A600"},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}, };
static const NormalizationProperties *_blocks[] = {
 0,_1,_2,_3,_4,_5,_6,_7,
 _8,_9,0,_11,_12,_13,_14,_15,
 _16,_17,_18,_19,_20,_21,_22,_23,
 _24,_25,_26,_27,_28,_29,_30,_31,
 _32,_33,0,0,0,0,_38,0,
 0,0,0,0,0,0,_46,_47,
 0,_49,_50,0,_52,_53,_54,_55,
 _56,_57,_58,_59,_60,_61,_62,_63,
 _64,_65,_66,_67,_68,_69,_70,0,
 _72,_73,0,0,0,0,0,0,
 0,0,0,0,_84,_85,0,0,
 _88,_89,_90,_91,0,_93,_94,_95,
 _96,_97,_98,_99,_100,_101,_102,_103,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_332,_333,_334,_335,
 _336,_337,_338,_339,0,_341,_342,_343,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_498,_499,_500,_501,_502,_503,
 _504,_505,_506,_507,_508,_509,_510,_511,
 0,0,0,_515,0,_517,_518,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_532,_533,0,0,
 0,0,_538,0,0,_541,_542,0,
 _544,_545,_546,_547,_548,_549,_550,0,
 _552,_553,0,_555,_556,_557,_558,0,
 _560,0,_562,_563,_564,_565,0,0,
 _568,0,_570,_571,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_725,_726,0,
 0,0,0,0,0,0,0,_735,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_889,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_930,_931,_932,0,0,0,
 _936,_937,_938,_939,_940,_941,_942,_943,
 0,0,0,0,0,_949,0,0,
 0,0,0,0,0,0,0,0,
 _960,0,_962,0,0,_965,0,0,
 0,0,0,0,0,0,0,0,
 0,_977,_978,0,0,0,0,0,
 0,0,0,0,_988,_989,0,0,
 0,0,_994,_995,_996,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_1015,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _1520,_1521,_1522,_1523,_1524,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
static const NormalizationProperties _block_values[] = {
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{230,0,0},{230,0,0},
 {230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},{230,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
 {0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0},
};
inline NormalizationProperties get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}
const std::unordered_map<std::u32string, char32_t> _normalization_composition = {
{ U"\U00000041\U00000300", 0x000000C0 },
{ U"\U00000041\U00000301", 0x000000C1 },
{ U"\U00000041\U00000302", 0x000000C2 },
{ U"\U00000041\U00000303", 0x000000C3 },
{ U"\U00000041\U00000308", 0x000000C4 },
{ U"\U00000041\U0000030A", 0x000000C5 },
{ U"\U00000043\U00000327", 0x000000C7 },
{ U"\U00000045\U00000300", 0x000000C8 },
{ U"\U00000045\U00000301", 0x000000C9 },
{ U"\U00000045\U00000302", 0x000000CA },
{ U"\U00000045\U00000308", 0x000000CB },
{ U"\U00000049\U00000300", 0x000000CC },
{ U"\U00000049\U00000301", 0x000000CD },
{ U"\U00000049\U00000302", 0x000000CE },
{ U"\U00000049\U00000308", 0x000000CF },
{ U"\U0000004E\U00000303", 0x000000D1 },
{ U"\U0000004F\U00000300", 0x000000D2 },
{ U"\U0000004F\U00000301", 0x000000D3 },
{ U"\U0000004F\U00000302", 0x000000D4 },
{ U"\U0000004F\U00000303", 0x000000D5 },
{ U"\U0000004F\U00000308", 0x000000D6 },
{ U"\U00000055\U00000300", 0x000000D9 },
{ U"\U00000055\U00000301", 0x000000DA },
{ U"\U00000055\U00000302", 0x000000DB },
{ U"\U00000055\U00000308", 0x000000DC },
{ U"\U00000059\U00000301", 0x000000DD },
{ U"\U00000061\U00000300", 0x000000E0 },
{ U"\U00000061\U00000301", 0x000000E1 },
{ U"\U00000061\U00000302", 0x000000E2 },
{ U"\U00000061\U00000303", 0x000000E3 },
{ U"\U00000061\U00000308", 0x000000E4 },
{ U"\U00000061\U0000030A", 0x000000E5 },
{ U"\U00000063\U00000327", 0x000000E7 },
{ U"\U00000065\U00000300", 0x000000E8 },
{ U"\U00000065\U00000301", 0x000000E9 },
{ U"\U00000065\U00000302", 0x000000EA },
{ U"\U00000065\U00000308", 0x000000EB },
{ U"\U00000069\U00000300", 0x000000EC },
{ U"\U00000069\U00000301", 0x000000ED },
{ U"\U00000069\U00000302", 0x000000EE },
{ U"\U00000069\U00000308", 0x000000EF },
{ U"\U0000006E\U00000303", 0x000000F1 },
{ U"\U0000006F\U00000300", 0x000000F2 },
{ U"\U0000006F\U00000301", 0x000000F3 },
{ U"\U0000006F\U00000302", 0x000000F4 },
{ U"\U0000006F\U00000303", 0x000000F5 },
{ U"\U0000006F\U00000308", 0x000000F6 },
{ U"\U00000075\U00000300", 0x000000F9 },
{ U"\U00000075\U00000301", 0x000000FA },
{ U"\U00000075\U00000302", 0x000000FB },
{ U"\U00000075\U00000308", 0x000000FC },
{ U"\U00000079\U00000301", 0x000000FD },
{ U"\U00000079\U00000308", 0x000000FF },
{ U"\U00000041\U00000304", 0x00000100 },
{ U"\U00000061\U00000304", 0x00000101 },
{ U"\U00000041\U00000306", 0x00000102 },
{ U"\U00000061\U00000306", 0x00000103 },
{ U"\U00000041\U00000328", 0x00000104 },
{ U"\U00000061\U00000328", 0x00000105 },
{ U"\U00000043\U00000301", 0x00000106 },
{ U"\U00000063\U00000301", 0x00000107 },
{ U"\U00000043\U00000302", 0x00000108 },
{ U"\U00000063\U00000302", 0x00000109 },
{ U"\U00000043\U00000307", 0x0000010A },
{ U"\U00000063\U00000307", 0x0000010B },
{ U"\U00000043\U0000030C", 0x0000010C },
{ U"\U00000063\U0000030C", 0x0000010D },
{ U"\U00000044\U0000030C", 0x0000010E },
{ U"\U00000064\U0000030C", 0x0000010F },
{ U"\U00000045\U00000304", 0x00000112 },
{ U"\U00000065\U00000304", 0x00000113 },
{ U"\U00000045\U00000306", 0x00000114 },
{ U"\U00000065\U00000306", 0x00000115 },
{ U"\U00000045\U00000307", 0x00000116 },
{ U"\U00000065\U00000307", 0x00000117 },
{ U"\U00000045\U00000328", 0x00000118 },
{ U"\U00000065\U00000328", 0x00000119 },
{ U"\U00000045\U0000030C", 0x0000011A },
{ U"\U00000065\U0000030C", 0x0000011B },
{ U"\U00000047\U00000302", 0x0000011C },
{ U"\U00000067\U00000302", 0x0000011D },
{ U"\U00000047\U00000306", 0x0000011E },
{ U"\U00000067\U00000306", 0x0000011F },
{ U"\U00000047\U00000307", 0x00000120 },
{ U"\U00000067\U00000307", 0x00000121 },
{ U"\U00000047\U00000327", 0x00000122 },
{ U"\U00000067\U00000327", 0x00000123 },
{ U"\U00000048\U00000302", 0x00000124 },
{ U"\U00000068\U00000302", 0x00000125 },
{ U"\U00000049\U00000303", 0x00000128 },
{ U"\U00000069\U00000303", 0x00000129 },
{ U"\U00000049\U00000304", 0x0000012A },
{ U"\U00000069\U00000304", 0x0000012B },
{ U"\U00000049\U00000306", 0x0000012C },
{ U"\U00000069\U00000306", 0x0000012D },
{ U"\U00000049\U00000328", 0x0000012E },
{ U"\U00000069\U00000328", 0x0000012F },
{ U"\U00000049\U00000307", 0x00000130 },
{ U"\U0000004A\U00000302", 0x00000134 },
{ U"\U0000006A\U00000302", 0x00000135 },
{ U"\U0000004B\U00000327", 0x00000136 },
{ U"\U0000006B\U00000327", 0x00000137 },
{ U"\U0000004C\U00000301", 0x00000139 },
{ U"\U0000006C\U00000301", 0x0000013A },
{ U"\U0000004C\U00000327", 0x0000013B },
{ U"\U0000006C\U00000327", 0x0000013C },
{ U"\U0000004C\U0000030C", 0x0000013D },
{ U"\U0000006C\U0000030C", 0x0000013E },
{ U"\U0000004E\U00000301", 0x00000143 },
{ U"\U0000006E\U00000301", 0x00000144 },
{ U"\U0000004E\U00000327", 0x00000145 },
{ U"\U0000006E\U00000327", 0x00000146 },
{ U"\U0000004E\U0000030C", 0x00000147 },
{ U"\U0000006E\U0000030C", 0x00000148 },
{ U"\U0000004F\U00000304", 0x0000014C },
{ U"\U0000006F\U00000304", 0x0000014D },
{ U"\U0000004F\U00000306", 0x0000014E },
{ U"\U0000006F\U00000306", 0x0000014F },
{ U"\U0000004F\U0000030B", 0x00000150 },
{ U"\U0000006F\U0000030B", 0x00000151 },
{ U"\U00000052\U00000301", 0x00000154 },
{ U"\U00000072\U00000301", 0x00000155 },
{ U"\U00000052\U00000327", 0x00000156 },
{ U"\U00000072\U00000327", 0x00000157 },
{ U"\U00000052\U0000030C", 0x00000158 },
{ U"\U00000072\U0000030C", 0x00000159 },
{ U"\U00000053\U00000301", 0x0000015A },
{ U"\U00000073\U00000301", 0x0000015B },
{ U"\U00000053\U00000302", 0x0000015C },
{ U"\U00000073\U00000302", 0x0000015D },
{ U"\U00000053\U00000327", 0x0000015E },
{ U"\U00000073\U00000327", 0x0000015F },
{ U"\U00000053\U0000030C", 0x00000160 },
{ U"\U00000073\U0000030C", 0x00000161 },
{ U"\U00000054\U00000327", 0x00000162 },
{ U"\U00000074\U00000327", 0x00000163 },
{ U"\U00000054\U0000030C", 0x00000164 },
{ U"\U00000074\U0000030C", 0x00000165 },
{ U"\U00000055\U00000303", 0x00000168 },
{ U"\U00000075\U00000303", 0x00000169 },
{ U"\U00000055\U00000304", 0x0000016A },
{ U"\U00000075\U00000304", 0x0000016B },
{ U"\U00000055\U00000306", 0x0000016C },
{ U"\U00000075\U00000306", 0x0000016D },
{ U"\U00000055\U0000030A", 0x0000016E },
{ U"\U00000075\U0000030A", 0x0000016F },
{ U"\U00000055\U0000030B", 0x00000170 },
{ U"\U00000075\U0000030B", 0x00000171 },
{ U"\U00000055\U00000328", 0x00000172 },
{ U"\U00000075\U00000328", 0x00000173 },
{ U"\U00000057\U00000302", 0x00000174 },
{ U"\U00000077\U00000302", 0x00000175 },
{ U"\U00000059\U00000302", 0x00000176 },
{ U"\U00000079\U00000302", 0x00000177 },
{ U"\U00000059\U00000308", 0x00000178 },
{ U"\U0000005A\U00000301", 0x00000179 },
{ U"\U0000007A\U00000301", 0x0000017A },
{ U"\U0000005A\U00000307", 0x0000017B },
{ U"\U0000007A\U00000307", 0x0000017C },
{ U"\U0000005A\U0000030C", 0x0000017D },
{ U"\U0000007A\U0000030C", 0x0000017E },
{ U"\U0000004F\U0000031B", 0x000001A0 },
{ U"\U0000006F\U0000031B", 0x000001A1 },
{ U"\U00000055\U0000031B", 0x000001AF },
{ U"\U00000075\U0000031B", 0x000001B0 },
{ U"\U00000041\U0000030C", 0x000001CD },
{ U"\U00000061\U0000030C", 0x000001CE },
{ U"\U00000049\U0000030C", 0x000001CF },
{ U"\U00000069\U0000030C", 0x000001D0 },
{ U"\U0000004F\U0000030C", 0x000001D1 },
{ U"\U0000006F\U0000030C", 0x000001D2 },
{ U"\U00000055\U0000030C", 0x000001D3 },
{ U"\U00000075\U0000030C", 0x000001D4 },
{ U"\U000000DC\U00000304", 0x000001D5 },
{ U"\U000000FC\U00000304", 0x000001D6 },
{ U"\U000000DC\U00000301", 0x000001D7 },
{ U"\U000000FC\U00000301", 0x000001D8 },
{ U"\U000000DC\U0000030C", 0x000001D9 },
{ U"\U000000FC\U0000030C", 0x000001DA },
{ U"\U000000DC\U00000300", 0x000001DB },
{ U"\U000000FC\U00000300", 0x000001DC },
{ U"\U000000C4\U00000304", 0x000001DE },
{ U"\U000000E4\U00000304", 0x000001DF },
{ U"\U00000226\U00000304", 0x000001E0 },
{ U"\U00000227\U00000304", 0x000001E1 },
{ U"\U000000C6\U00000304", 0x000001E2 },
{ U"\U000000E6\U00000304", 0x000001E3 },
{ U"\U00000047\U0000030C", 0x000001E6 },
{ U"\U00000067\U0000030C", 0x000001E7 },
{ U"\U0000004B\U0000030C", 0x000001E8 },
{ U"\U0000006B\U0000030C", 0x000001E9 },
{ U"\U0000004F\U00000328", 0x000001EA },
{ U"\U0000006F\U00000328", 0x000001EB },
{ U"\U000001EA\U00000304", 0x000001EC },
{ U"\U000001EB\U00000304", 0x000001ED },
{ U"\U000001B7\U0000030C", 0x000001EE },
{ U"\U00000292\U0000030C", 0x000001EF },
{ U"\U0000006A\U0000030C", 0x000001F0 },
{ U"\U00000047\U00000301", 0x000001F4 },
{ U"\U00000067\U00000301", 0x000001F5 },
{ U"\U0000004E\U00000300", 0x000001F8 },
{ U"\U0000006E\U00000300", 0x000001F9 },
{ U"\U000000C5\U00000301", 0x000001FA },
{ U"\U000000E5\U00000301", 0x000001FB },
{ U"\U000000C6\U00000301", 0x000001FC },
{ U"\U000000E6\U00000301", 0x000001FD },
{ U"\U000000D8\U00000301", 0x000001FE },
{ U"\U000000F8\U00000301", 0x000001FF },
{ U"\U00000041\U0000030F", 0x00000200 },
{ U"\U00000061\U0000030F", 0x00000201 },
{ U"\U00000041\U00000311", 0x00000202 },
{ U"\U00000061\U00000311", 0x00000203 },
{ U"\U00000045\U0000030F", 0x00000204 },
{ U"\U00000065\U0000030F", 0x00000205 },
{ U"\U00000045\U00000311", 0x00000206 },
{ U"\U00000065\U00000311", 0x00000207 },
{ U"\U00000049\U0000030F", 0x00000208 },
{ U"\U00000069\U0000030F", 0x00000209 },
{ U"\U00000049\U00000311", 0x0000020A },
{ U"\U00000069\U00000311", 0x0000020B },
{ U"\U0000004F\U0000030F", 0x0000020C },
{ U"\U0000006F\U0000030F", 0x0000020D },
{ U"\U0000004F\U00000311", 0x0000020E },
{ U"\U0000006F\U00000311", 0x0000020F },
{ U"\U00000052\U0000030F", 0x00000210 },
{ U"\U00000072\U0000030F", 0x00000211 },
{ U"\U00000052\U00000311", 0x00000212 },
{ U"\U00000072\U00000311", 0x00000213 },
{ U"\U00000055\U0000030F", 0x00000214 },
{ U"\U00000075\U0000030F", 0x00000215 },
{ U"\U00000055\U00000311", 0x00000216 },
{ U"\U00000075\U00000311", 0x00000217 },
{ U"\U00000053\U00000326", 0x00000218 },
{ U"\U00000073\U00000326", 0x00000219 },
{ U"\U00000054\U00000326", 0x0000021A },
{ U"\U00000074\U00000326", 0x0000021B },
{ U"\U00000048\U0000030C", 0x0000021E },
{ U"\U00000068\U0000030C", 0x0000021F },
{ U"\U00000041\U00000307", 0x00000226 },
{ U"\U00000061\U00000307", 0x00000227 },
{ U"\U00000045\U00000327", 0x00000228 },
{ U"\U00000065\U00000327", 0x00000229 },
{ U"\U000000D6\U00000304", 0x0000022A },
{ U"\U000000F6\U00000304", 0x0000022B },
{ U"\U000000D5\U00000304", 0x0000022C },
{ U"\U000000F5\U00000304", 0x0000022D },
{ U"\U0000004F\U00000307", 0x0000022E },
{ U"\U0000006F\U00000307", 0x0000022F },
{ U"\U0000022E\U00000304", 0x00000230 },
{ U"\U0000022F\U00000304", 0x00000231 },
{ U"\U00000059\U00000304", 0x00000232 },
{ U"\U00000079\U00000304", 0x00000233 },
{ U"\U000000A8\U00000301", 0x00000385 },
{ U"\U00000391\U00000301", 0x00000386 },
{ U"\U00000395\U00000301", 0x00000388 },
{ U"\U00000397\U00000301", 0x00000389 },
{ U"\U00000399\U00000301", 0x0000038A },
{ U"\U0000039F\U00000301", 0x0000038C },
{ U"\U000003A5\U00000301", 0x0000038E },
{ U"\U000003A9\U00000301", 0x0000038F },
{ U"\U000003CA\U00000301", 0x00000390 },
{ U"\U00000399\U00000308", 0x000003AA },
{ U"\U000003A5\U00000308", 0x000003AB },
{ U"\U000003B1\U00000301", 0x000003AC },
{ U"\U000003B5\U00000301", 0x000003AD },
{ U"\U000003B7\U00000301", 0x000003AE },
{ U"\U000003B9\U00000301", 0x000003AF },
{ U"\U000003CB\U00000301", 0x000003B0 },
{ U"\U000003B9\U00000308", 0x000003CA },
{ U"\U000003C5\U00000308", 0x000003CB },
{ U"\U000003BF\U00000301", 0x000003CC },
{ U"\U000003C5\U00000301", 0x000003CD },
{ U"\U000003C9\U00000301", 0x000003CE },
{ U"\U000003D2\U00000301", 0x000003D3 },
{ U"\U000003D2\U00000308", 0x000003D4 },
{ U"\U00000415\U00000300", 0x00000400 },
{ U"\U00000415\U00000308", 0x00000401 },
{ U"\U00000413\U00000301", 0x00000403 },
{ U"\U00000406\U00000308", 0x00000407 },
{ U"\U0000041A\U00000301", 0x0000040C },
{ U"\U00000418\U00000300", 0x0000040D },
{ U"\U00000423\U00000306", 0x0000040E },
{ U"\U00000418\U00000306", 0x00000419 },
{ U"\U00000438\U00000306", 0x00000439 },
{ U"\U00000435\U00000300", 0x00000450 },
{ U"\U00000435\U00000308", 0x00000451 },
{ U"\U00000433\U00000301", 0x00000453 },
{ U"\U00000456\U00000308", 0x00000457 },
{ U"\U0000043A\U00000301", 0x0000045C },
{ U"\U00000438\U00000300", 0x0000045D },
{ U"\U00000443\U00000306", 0x0000045E },
{ U"\U00000474\U0000030F", 0x00000476 },
{ U"\U00000475\U0000030F", 0x00000477 },
{ U"\U00000416\U00000306", 0x000004C1 },
{ U"\U00000436\U00000306", 0x000004C2 },
{ U"\U00000410\U00000306", 0x000004D0 },
{ U"\U00000430\U00000306", 0x000004D1 },
{ U"\U00000410\U00000308", 0x000004D2 },
{ U"\U00000430\U00000308", 0x000004D3 },
{ U"\U00000415\U00000306", 0x000004D6 },
{ U"\U00000435\U00000306", 0x000004D7 },
{ U"\U000004D8\U00000308", 0x000004DA },
{ U"\U000004D9\U00000308", 0x000004DB },
{ U"\U00000416\U00000308", 0x000004DC },
{ U"\U00000436\U00000308", 0x000004DD },
{ U"\U00000417\U00000308", 0x000004DE },
{ U"\U00000437\U00000308", 0x000004DF },
{ U"\U00000418\U00000304", 0x000004E2 },
{ U"\U00000438\U00000304", 0x000004E3 },
{ U"\U00000418\U00000308", 0x000004E4 },
{ U"\U00000438\U00000308", 0x000004E5 },
{ U"\U0000041E\U00000308", 0x000004E6 },
{ U"\U0000043E\U00000308", 0x000004E7 },
{ U"\U000004E8\U00000308", 0x000004EA },
{ U"\U000004E9\U00000308", 0x000004EB },
{ U"\U0000042D\U00000308", 0x000004EC },
{ U"\U0000044D\U00000308", 0x000004ED },
{ U"\U00000423\U00000304", 0x000004EE },
{ U"\U00000443\U00000304", 0x000004EF },
{ U"\U00000423\U00000308", 0x000004F0 },
{ U"\U00000443\U00000308", 0x000004F1 },
{ U"\U00000423\U0000030B", 0x000004F2 },
{ U"\U00000443\U0000030B", 0x000004F3 },
{ U"\U00000427\U00000308", 0x000004F4 },
{ U"\U00000447\U00000308", 0x000004F5 },
{ U"\U0000042B\U00000308", 0x000004F8 },
{ U"\U0000044B\U00000308", 0x000004F9 },
{ U"\U00000627\U00000653", 0x00000622 },
{ U"\U00000627\U00000654", 0x00000623 },
{ U"\U00000648\U00000654", 0x00000624 },
{ U"\U00000627\U00000655", 0x00000625 },
{ U"\U0000064A\U00000654", 0x00000626 },
{ U"\U000006D5\U00000654", 0x000006C0 },
{ U"\U000006C1\U00000654", 0x000006C2 },
{ U"\U000006D2\U00000654", 0x000006D3 },
{ U"\U00000928\U0000093C", 0x00000929 },
{ U"\U00000930\U0000093C", 0x00000931 },
{ U"\U00000933\U0000093C", 0x00000934 },
{ U"\U000009C7\U000009BE", 0x000009CB },
{ U"\U000009C7\U000009D7", 0x000009CC },
{ U"\U00000B47\U00000B56", 0x00000B48 },
{ U"\U00000B47\U00000B3E", 0x00000B4B },
{ U"\U00000B47\U00000B57", 0x00000B4C },
{ U"\U00000B92\U00000BD7", 0x00000B94 },
{ U"\U00000BC6\U00000BBE", 0x00000BCA },
{ U"\U00000BC7\U00000BBE", 0x00000BCB },
{ U"\U00000BC6\U00000BD7", 0x00000BCC },
{ U"\U00000C46\U00000C56", 0x00000C48 },
{ U"\U00000CBF\U00000CD5", 0x00000CC0 },
{ U"\U00000CC6\U00000CD5", 0x00000CC7 },
{ U"\U00000CC6\U00000CD6", 0x00000CC8 },
{ U"\U00000CC6\U00000CC2", 0x00000CCA },
{ U"\U00000CCA\U00000CD5", 0x00000CCB },
{ U"\U00000D46\U00000D3E", 0x00000D4A },
{ U"\U00000D47\U00000D3E", 0x00000D4B },
{ U"\U00000D46\U00000D57", 0x00000D4C },
{ U"\U00000DD9\U00000DCA", 0x00000DDA },
{ U"\U00000DD9\U00000DCF", 0x00000DDC },
{ U"\U00000DDC\U00000DCA", 0x00000DDD },
{ U"\U00000DD9\U00000DDF", 0x00000DDE },
{ U"\U00001025\U0000102E", 0x00001026 },
{ U"\U00001B05\U00001B35", 0x00001B06 },
{ U"\U00001B07\U00001B35", 0x00001B08 },
{ U"\U00001B09\U00001B35", 0x00001B0A },
{ U"\U00001B0B\U00001B35", 0x00001B0C },
{ U"\U00001B0D\U00001B35", 0x00001B0E },
{ U"\U00001B11\U00001B35", 0x00001B12 },
{ U"\U00001B3A\U00001B35", 0x00001B3B },
{ U"\U00001B3C\U00001B35", 0x00001B3D },
{ U"\U00001B3E\U00001B35", 0x00001B40 },
{ U"\U00001B3F\U00001B35", 0x00001B41 },
{ U"\U00001B42\U00001B35", 0x00001B43 },
{ U"\U00000041\U00000325", 0x00001E00 },
{ U"\U00000061\U00000325", 0x00001E01 },
{ U"\U00000042\U00000307", 0x00001E02 },
{ U"\U00000062\U00000307", 0x00001E03 },
{ U"\U00000042\U00000323", 0x00001E04 },
{ U"\U00000062\U00000323", 0x00001E05 },
{ U"\U00000042\U00000331", 0x00001E06 },
{ U"\U00000062\U00000331", 0x00001E07 },
{ U"\U000000C7\U00000301", 0x00001E08 },
{ U"\U000000E7\U00000301", 0x00001E09 },
{ U"\U00000044\U00000307", 0x00001E0A },
{ U"\U00000064\U00000307", 0x00001E0B },
{ U"\U00000044\U00000323", 0x00001E0C },
{ U"\U00000064\U00000323", 0x00001E0D },
{ U"\U00000044\U00000331", 0x00001E0E },
{ U"\U00000064\U00000331", 0x00001E0F },
{ U"\U00000044\U00000327", 0x00001E10 },
{ U"\U00000064\U00000327", 0x00001E11 },
{ U"\U00000044\U0000032D", 0x00001E12 },
{ U"\U00000064\U0000032D", 0x00001E13 },
{ U"\U00000112\U00000300", 0x00001E14 },
{ U"\U00000113\U00000300", 0x00001E15 },
{ U"\U00000112\U00000301", 0x00001E16 },
{ U"\U00000113\U00000301", 0x00001E17 },
{ U"\U00000045\U0000032D", 0x00001E18 },
{ U"\U00000065\U0000032D", 0x00001E19 },
{ U"\U00000045\U00000330", 0x00001E1A },
{ U"\U00000065\U00000330", 0x00001E1B },
{ U"\U00000228\U00000306", 0x00001E1C },
{ U"\U00000229\U00000306", 0x00001E1D },
{ U"\U00000046\U00000307", 0x00001E1E },
{ U"\U00000066\U00000307", 0x00001E1F },
{ U"\U00000047\U00000304", 0x00001E20 },
{ U"\U00000067\U00000304", 0x00001E21 },
{ U"\U00000048\U00000307", 0x00001E22 },
{ U"\U00000068\U00000307", 0x00001E23 },
{ U"\U00000048\U00000323", 0x00001E24 },
{ U"\U00000068\U00000323", 0x00001E25 },
{ U"\U00000048\U00000308", 0x00001E26 },
{ U"\U00000068\U00000308", 0x00001E27 },
{ U"\U00000048\U00000327", 0x00001E28 },
{ U"\U00000068\U00000327", 0x00001E29 },
{ U"\U00000048\U0000032E", 0x00001E2A },
{ U"\U00000068\U0000032E", 0x00001E2B },
{ U"\U00000049\U00000330", 0x00001E2C },
{ U"\U00000069\U00000330", 0x00001E2D },
{ U"\U000000CF\U00000301", 0x00001E2E },
{ U"\U000000EF\U00000301", 0x00001E2F },
{ U"\U0000004B\U00000301", 0x00001E30 },
{ U"\U0000006B\U00000301", 0x00001E31 },
{ U"\U0000004B\U00000323", 0x00001E32 },
{ U"\U0000006B\U00000323", 0x00001E33 },
{ U"\U0000004B\U00000331", 0x00001E34 },
{ U"\U0000006B\U00000331", 0x00001E35 },
{ U"\U0000004C\U00000323", 0x00001E36 },
{ U"\U0000006C\U00000323", 0x00001E37 },
{ U"\U00001E36\U00000304", 0x00001E38 },
{ U"\U00001E37\U00000304", 0x00001E39 },
{ U"\U0000004C\U00000331", 0x00001E3A },
{ U"\U0000006C\U00000331", 0x00001E3B },
{ U"\U0000004C\U0000032D", 0x00001E3C },
{ U"\U0000006C\U0000032D", 0x00001E3D },
{ U"\U0000004D\U00000301", 0x00001E3E },
{ U"\U0000006D\U00000301", 0x00001E3F },
{ U"\U0000004D\U00000307", 0x00001E40 },
{ U"\U0000006D\U00000307", 0x00001E41 },
{ U"\U0000004D\U00000323", 0x00001E42 },
{ U"\U0000006D\U00000323", 0x00001E43 },
{ U"\U0000004E\U00000307", 0x00001E44 },
{ U"\U0000006E\U00000307", 0x00001E45 },
{ U"\U0000004E\U00000323", 0x00001E46 },
{ U"\U0000006E\U00000323", 0x00001E47 },
{ U"\U0000004E\U00000331", 0x00001E48 },
{ U"\U0000006E\U00000331", 0x00001E49 },
{ U"\U0000004E\U0000032D", 0x00001E4A },
{ U"\U0000006E\U0000032D", 0x00001E4B },
{ U"\U000000D5\U00000301", 0x00001E4C },
{ U"\U000000F5\U00000301", 0x00001E4D },
{ U"\U000000D5\U00000308", 0x00001E4E },
{ U"\U000000F5\U00000308", 0x00001E4F },
{ U"\U0000014C\U00000300", 0x00001E50 },
{ U"\U0000014D\U00000300", 0x00001E51 },
{ U"\U0000014C\U00000301", 0x00001E52 },
{ U"\U0000014D\U00000301", 0x00001E53 },
{ U"\U00000050\U00000301", 0x00001E54 },
{ U"\U00000070\U00000301", 0x00001E55 },
{ U"\U00000050\U00000307", 0x00001E56 },
{ U"\U00000070\U00000307", 0x00001E57 },
{ U"\U00000052\U00000307", 0x00001E58 },
{ U"\U00000072\U00000307", 0x00001E59 },
{ U"\U00000052\U00000323", 0x00001E5A },
{ U"\U00000072\U00000323", 0x00001E5B },
{ U"\U00001E5A\U00000304", 0x00001E5C },
{ U"\U00001E5B\U00000304", 0x00001E5D },
{ U"\U00000052\U00000331", 0x00001E5E },
{ U"\U00000072\U00000331", 0x00001E5F },
{ U"\U00000053\U00000307", 0x00001E60 },
{ U"\U00000073\U00000307", 0x00001E61 },
{ U"\U00000053\U00000323", 0x00001E62 },
{ U"\U00000073\U00000323", 0x00001E63 },
{ U"\U0000015A\U00000307", 0x00001E64 },
{ U"\U0000015B\U00000307", 0x00001E65 },
{ U"\U00000160\U00000307", 0x00001E66 },
{ U"\U00000161\U00000307", 0x00001E67 },
{ U"\U00001E62\U00000307", 0x00001E68 },
{ U"\U00001E63\U00000307", 0x00001E69 },
{ U"\U00000054\U00000307", 0x00001E6A },
{ U"\U00000074\U00000307", 0x00001E6B },
{ U"\U00000054\U00000323", 0x00001E6C },
{ U"\U00000074\U00000323", 0x00001E6D },
{ U"\U00000054\U00000331", 0x00001E6E },
{ U"\U00000074\U00000331", 0x00001E6F },
{ U"\U00000054\U0000032D", 0x00001E70 },
{ U"\U00000074\U0000032D", 0x00001E71 },
{ U"\U00000055\U00000324", 0x00001E72 },
{ U"\U00000075\U00000324", 0x00001E73 },
{ U"\U00000055\U00000330", 0x00001E74 },
{ U"\U00000075\U00000330", 0x00001E75 },
{ U"\U00000055\U0000032D", 0x00001E76 },
{ U"\U00000075\U0000032D", 0x00001E77 },
{ U"\U00000168\U00000301", 0x00001E78 },
{ U"\U00000169\U00000301", 0x00001E79 },
{ U"\U0000016A\U00000308", 0x00001E7A },
{ U"\U0000016B\U00000308", 0x00001E7B },
{ U"\U00000056\U00000303", 0x00001E7C },
{ U"\U00000076\U00000303", 0x00001E7D },
{ U"\U00000056\U00000323", 0x00001E7E },
{ U"\U00000076\U00000323", 0x00001E7F },
{ U"\U00000057\U00000300", 0x00001E80 },
{ U"\U00000077\U00000300", 0x00001E81 },
{ U"\U00000057\U00000301", 0x00001E82 },
{ U"\U00000077\U00000301", 0x00001E83 },
{ U"\U00000057\U00000308", 0x00001E84 },
{ U"\U00000077\U00000308", 0x00001E85 },
{ U"\U00000057\U00000307", 0x00001E86 },
{ U"\U00000077\U00000307", 0x00001E87 },
{ U"\U00000057\U00000323", 0x00001E88 },
{ U"\U00000077\U00000323", 0x00001E89 },
{ U"\U00000058\U00000307", 0x00001E8A },
{ U"\U00000078\U00000307", 0x00001E8B },
{ U"\U00000058\U00000308", 0x00001E8C },
{ U"\U00000078\U00000308", 0x00001E8D },
{ U"\U00000059\U00000307", 0x00001E8E },
{ U"\U00000079\U00000307", 0x00001E8F },
{ U"\U0000005A\U00000302", 0x00001E90 },
{ U"\U0000007A\U00000302", 0x00001E91 },
{ U"\U0000005A\U00000323", 0x00001E92 },
{ U"\U0000007A\U00000323", 0x00001E93 },
{ U"\U0000005A\U00000331", 0x00001E94 },
{ U"\U0000007A\U00000331", 0x00001E95 },
{ U"\U00000068\U00000331", 0x00001E96 },
{ U"\U00000074\U00000308", 0x00001E97 },
{ U"\U00000077\U0000030A", 0x00001E98 },
{ U"\U00000079\U0000030A", 0x00001E99 },
{ U"\U0000017F\U00000307", 0x00001E9B },
{ U"\U00000041\U00000323", 0x00001EA0 },
{ U"\U00000061\U00000323", 0x00001EA1 },
{ U"\U00000041\U00000309", 0x00001EA2 },
{ U"\U00000061\U00000309", 0x00001EA3 },
{ U"\U000000C2\U00000301", 0x00001EA4 },
{ U"\U000000E2\U00000301", 0x00001EA5 },
{ U"\U000000C2\U00000300", 0x00001EA6 },
{ U"\U000000E2\U00000300", 0x00001EA7 },
{ U"\U000000C2\U00000309", 0x00001EA8 },
{ U"\U000000E2\U00000309", 0x00001EA9 },
{ U"\U000000C2\U00000303", 0x00001EAA },
{ U"\U000000E2\U00000303", 0x00001EAB },
{ U"\U00001EA0\U00000302", 0x00001EAC },
{ U"\U00001EA1\U00000302", 0x00001EAD },
{ U"\U00000102\U00000301", 0x00001EAE },
{ U"\U00000103\U00000301", 0x00001EAF },
{ U"\U00000102\U00000300", 0x00001EB0 },
{ U"\U00000103\U00000300", 0x00001EB1 },
{ U"\U00000102\U00000309", 0x00001EB2 },
{ U"\U00000103\U00000309", 0x00001EB3 },
{ U"\U00000102\U00000303", 0x00001EB4 },
{ U"\U00000103\U00000303", 0x00001EB5 },
{ U"\U00001EA0\U00000306", 0x00001EB6 },
{ U"\U00001EA1\U00000306", 0x00001EB7 },
{ U"\U00000045\U00000323", 0x00001EB8 },
{ U"\U00000065\U00000323", 0x00001EB9 },
{ U"\U00000045\U00000309", 0x00001EBA },
{ U"\U00000065\U00000309", 0x00001EBB },
{ U"\U00000045\U00000303", 0x00001EBC },
{ U"\U00000065\U00000303", 0x00001EBD },
{ U"\U000000CA\U00000301", 0x00001EBE },
{ U"\U000000EA\U00000301", 0x00001EBF },
{ U"\U000000CA\U00000300", 0x00001EC0 },
{ U"\U000000EA\U00000300", 0x00001EC1 },
{ U"\U000000CA\U00000309", 0x00001EC2 },
{ U"\U000000EA\U00000309", 0x00001EC3 },
{ U"\U000000CA\U00000303", 0x00001EC4 },
{ U"\U000000EA\U00000303", 0x00001EC5 },
{ U"\U00001EB8\U00000302", 0x00001EC6 },
{ U"\U00001EB9\U00000302", 0x00001EC7 },
{ U"\U00000049\U00000309", 0x00001EC8 },
{ U"\U00000069\U00000309", 0x00001EC9 },
{ U"\U00000049\U00000323", 0x00001ECA },
{ U"\U00000069\U00000323", 0x00001ECB },
{ U"\U0000004F\U00000323", 0x00001ECC },
{ U"\U0000006F\U00000323", 0x00001ECD },
{ U"\U0000004F\U00000309", 0x00001ECE },
{ U"\U0000006F\U00000309", 0x00001ECF },
{ U"\U000000D4\U00000301", 0x00001ED0 },
{ U"\U000000F4\U00000301", 0x00001ED1 },
{ U"\U000000D4\U00000300", 0x00001ED2 },
{ U"\U000000F4\U00000300", 0x00001ED3 },
{ U"\U000000D4\U00000309", 0x00001ED4 },
{ U"\U000000F4\U00000309", 0x00001ED5 },
{ U"\U000000D4\U00000303", 0x00001ED6 },
{ U"\U000000F4\U00000303", 0x00001ED7 },
{ U"\U00001ECC\U00000302", 0x00001ED8 },
{ U"\U00001ECD\U00000302", 0x00001ED9 },
{ U"\U000001A0\U00000301", 0x00001EDA },
{ U"\U000001A1\U00000301", 0x00001EDB },
{ U"\U000001A0\U00000300", 0x00001EDC },
{ U"\U000001A1\U00000300", 0x00001EDD },
{ U"\U000001A0\U00000309", 0x00001EDE },
{ U"\U000001A1\U00000309", 0x00001EDF },
{ U"\U000001A0\U00000303", 0x00001EE0 },
{ U"\U000001A1\U00000303", 0x00001EE1 },
{ U"\U000001A0\U00000323", 0x00001EE2 },
{ U"\U000001A1\U00000323", 0x00001EE3 },
{ U"\U00000055\U00000323", 0x00001EE4 },
{ U"\U00000075\U00000323", 0x00001EE5 },
{ U"\U00000055\U00000309", 0x00001EE6 },
{ U"\U00000075\U00000309", 0x00001EE7 },
{ U"\U000001AF\U00000301", 0x00001EE8 },
{ U"\U000001B0\U00000301", 0x00001EE9 },
{ U"\U000001AF\U00000300", 0x00001EEA },
{ U"\U000001B0\U00000300", 0x00001EEB },
{ U"\U000001AF\U00000309", 0x00001EEC },
{ U"\U000001B0\U00000309", 0x00001EED },
{ U"\U000001AF\U00000303", 0x00001EEE },
{ U"\U000001B0\U00000303", 0x00001EEF },
{ U"\U000001AF\U00000323", 0x00001EF0 },
{ U"\U000001B0\U00000323", 0x00001EF1 },
{ U"\U00000059\U00000300", 0x00001EF2 },
{ U"\U00000079\U00000300", 0x00001EF3 },
{ U"\U00000059\U00000323", 0x00001EF4 },
{ U"\U00000079\U00000323", 0x00001EF5 },
{ U"\U00000059\U00000309", 0x00001EF6 },
{ U"\U00000079\U00000309", 0x00001EF7 },
{ U"\U00000059\U00000303", 0x00001EF8 },
{ U"\U00000079\U00000303", 0x00001EF9 },
{ U"\U000003B1\U00000313", 0x00001F00 },
{ U"\U000003B1\U00000314", 0x00001F01 },
{ U"\U00001F00\U00000300", 0x00001F02 },
{ U"\U00001F01\U00000300", 0x00001F03 },
{ U"\U00001F00\U00000301", 0x00001F04 },
{ U"\U00001F01\U00000301", 0x00001F05 },
{ U"\U00001F00\U00000342", 0x00001F06 },
{ U"\U00001F01\U00000342", 0x00001F07 },
{ U"\U00000391\U00000313", 0x00001F08 },
{ U"\U00000391\U00000314", 0x00001F09 },
{ U"\U00001F08\U00000300", 0x00001F0A },
{ U"\U00001F09\U00000300", 0x00001F0B },
{ U"\U00001F08\U00000301", 0x00001F0C },
{ U"\U00001F09\U00000301", 0x00001F0D },
{ U"\U00001F08\U00000342", 0x00001F0E },
{ U"\U00001F09\U00000342", 0x00001F0F },
{ U"\U000003B5\U00000313", 0x00001F10 },
{ U"\U000003B5\U00000314", 0x00001F11 },
{ U"\U00001F10\U00000300", 0x00001F12 },
{ U"\U00001F11\U00000300", 0x00001F13 },
{ U"\U00001F10\U00000301", 0x00001F14 },
{ U"\U00001F11\U00000301", 0x00001F15 },
{ U"\U00000395\U00000313", 0x00001F18 },
{ U"\U00000395\U00000314", 0x00001F19 },
{ U"\U00001F18\U00000300", 0x00001F1A },
{ U"\U00001F19\U00000300", 0x00001F1B },
{ U"\U00001F18\U00000301", 0x00001F1C },
{ U"\U00001F19\U00000301", 0x00001F1D },
{ U"\U000003B7\U00000313", 0x00001F20 },
{ U"\U000003B7\U00000314", 0x00001F21 },
{ U"\U00001F20\U00000300", 0x00001F22 },
{ U"\U00001F21\U00000300", 0x00001F23 },
{ U"\U00001F20\U00000301", 0x00001F24 },
{ U"\U00001F21\U00000301", 0x00001F25 },
{ U"\U00001F20\U00000342", 0x00001F26 },
{ U"\U00001F21\U00000342", 0x00001F27 },
{ U"\U00000397\U00000313", 0x00001F28 },
{ U"\U00000397\U00000314", 0x00001F29 },
{ U"\U00001F28\U00000300", 0x00001F2A },
{ U"\U00001F29\U00000300", 0x00001F2B },
{ U"\U00001F28\U00000301", 0x00001F2C },
{ U"\U00001F29\U00000301", 0x00001F2D },
{ U"\U00001F28\U00000342", 0x00001F2E },
{ U"\U00001F29\U00000342", 0x00001F2F },
{ U"\U000003B9\U00000313", 0x00001F30 },
{ U"\U000003B9\U00000314", 0x00001F31 },
{ U"\U00001F30\U00000300", 0x00001F32 },
{ U"\U00001F31\U00000300", 0x00001F33 },
{ U"\U00001F30\U00000301", 0x00001F34 },
{ U"\U00001F31\U00000301", 0x00001F35 },
{ U"\U00001F30\U00000342", 0x00001F36 },
{ U"\U00001F31\U00000342", 0x00001F37 },
{ U"\U00000399\U00000313", 0x00001F38 },
{ U"\U00000399\U00000314", 0x00001F39 },
{ U"\U00001F38\U00000300", 0x00001F3A },
{ U"\U00001F39\U00000300", 0x00001F3B },
{ U"\U00001F38\U00000301", 0x00001F3C },
{ U"\U00001F39\U00000301", 0x00001F3D },
{ U"\U00001F38\U00000342", 0x00001F3E },
{ U"\U00001F39\U00000342", 0x00001F3F },
{ U"\U000003BF\U00000313", 0x00001F40 },
{ U"\U000003BF\U00000314", 0x00001F41 },
{ U"\U00001F40\U00000300", 0x00001F42 },
{ U"\U00001F41\U00000300", 0x00001F43 },
{ U"\U00001F40\U00000301", 0x00001F44 },
{ U"\U00001F41\U00000301", 0x00001F45 },
{ U"\U0000039F\U00000313", 0x00001F48 },
{ U"\U0000039F\U00000314", 0x00001F49 },
{ U"\U00001F48\U00000300", 0x00001F4A },
{ U"\U00001F49\U00000300", 0x00001F4B },
{ U"\U00001F48\U00000301", 0x00001F4C },
{ U"\U00001F49\U00000301", 0x00001F4D },
{ U"\U000003C5\U00000313", 0x00001F50 },
{ U"\U000003C5\U00000314", 0x00001F51 },
{ U"\U00001F50\U00000300", 0x00001F52 },
{ U"\U00001F51\U00000300", 0x00001F53 },
{ U"\U00001F50\U00000301", 0x00001F54 },
{ U"\U00001F51\U00000301", 0x00001F55 },
{ U"\U00001F50\U00000342", 0x00001F56 },
{ U"\U00001F51\U00000342", 0x00001F57 },
{ U"\U000003A5\U00000314", 0x00001F59 },
{ U"\U00001F59\U00000300", 0x00001F5B },
{ U"\U00001F59\U00000301", 0x00001F5D },
{ U"\U00001F59\U00000342", 0x00001F5F },
{ U"\U000003C9\U00000313", 0x00001F60 },
{ U"\U000003C9\U00000314", 0x00001F61 },
{ U"\U00001F60\U00000300", 0x00001F62 },
{ U"\U00001F61\U00000300", 0x00001F63 },
{ U"\U00001F60\U00000301", 0x00001F64 },
{ U"\U00001F61\U00000301", 0x00001F65 },
{ U"\U00001F60\U00000342", 0x00001F66 },
{ U"\U00001F61\U00000342", 0x00001F67 },
{ U"\U000003A9\U00000313", 0x00001F68 },
{ U"\U000003A9\U00000314", 0x00001F69 },
{ U"\U00001F68\U00000300", 0x00001F6A },
{ U"\U00001F69\U00000300", 0x00001F6B },
{ U"\U00001F68\U00000301", 0x00001F6C },
{ U"\U00001F69\U00000301", 0x00001F6D },
{ U"\U00001F68\U00000342", 0x00001F6E },
{ U"\U00001F69\U00000342", 0x00001F6F },
{ U"\U000003B1\U00000300", 0x00001F70 },
{ U"\U000003B5\U00000300", 0x00001F72 },
{ U"\U000003B7\U00000300", 0x00001F74 },
{ U"\U000003B9\U00000300", 0x00001F76 },
{ U"\U000003BF\U00000300", 0x00001F78 },
{ U"\U000003C5\U00000300", 0x00001F7A },
{ U"\U000003C9\U00000300", 0x00001F7C },
{ U"\U00001F00\U00000345", 0x00001F80 },
{ U"\U00001F01\U00000345", 0x00001F81 },
{ U"\U00001F02\U00000345", 0x00001F82 },
{ U"\U00001F03\U00000345", 0x00001F83 },
{ U"\U00001F04\U00000345", 0x00001F84 },
{ U"\U00001F05\U00000345", 0x00001F85 },
{ U"\U00001F06\U00000345", 0x00001F86 },
{ U"\U00001F07\U00000345", 0x00001F87 },
{ U"\U00001F08\U00000345", 0x00001F88 },
{ U"\U00001F09\U00000345", 0x00001F89 },
{ U"\U00001F0A\U00000345", 0x00001F8A },
{ U"\U00001F0B\U00000345", 0x00001F8B },
{ U"\U00001F0C\U00000345", 0x00001F8C },
{ U"\U00001F0D\U00000345", 0x00001F8D },
{ U"\U00001F0E\U00000345", 0x00001F8E },
{ U"\U00001F0F\U00000345", 0x00001F8F },
{ U"\U00001F20\U00000345", 0x00001F90 },
{ U"\U00001F21\U00000345", 0x00001F91 },
{ U"\U00001F22\U00000345", 0x00001F92 },
{ U"\U00001F23\U00000345", 0x00001F93 },
{ U"\U00001F24\U00000345", 0x00001F94 },
{ U"\U00001F25\U00000345", 0x00001F95 },
{ U"\U00001F26\U00000345", 0x00001F96 },
{ U"\U00001F27\U00000345", 0x00001F97 },
{ U"\U00001F28\U00000345", 0x00001F98 },
{ U"\U00001F29\U00000345", 0x00001F99 },
{ U"\U00001F2A\U00000345", 0x00001F9A },
{ U"\U00001F2B\U00000345", 0x00001F9B },
{ U"\U00001F2C\U00000345", 0x00001F9C },
{ U"\U00001F2D\U00000345", 0x00001F9D },
{ U"\U00001F2E\U00000345", 0x00001F9E },
{ U"\U00001F2F\U00000345", 0x00001F9F },
{ U"\U00001F60\U00000345", 0x00001FA0 },
{ U"\U00001F61\U00000345", 0x00001FA1 },
{ U"\U00001F62\U00000345", 0x00001FA2 },
{ U"\U00001F63\U00000345", 0x00001FA3 },
{ U"\U00001F64\U00000345", 0x00001FA4 },
{ U"\U00001F65\U00000345", 0x00001FA5 },
{ U"\U00001F66\U00000345", 0x00001FA6 },
{ U"\U00001F67\U00000345", 0x00001FA7 },
{ U"\U00001F68\U00000345", 0x00001FA8 },
{ U"\U00001F69\U00000345", 0x00001FA9 },
{ U"\U00001F6A\U00000345", 0x00001FAA },
{ U"\U00001F6B\U00000345", 0x00001FAB },
{ U"\U00001F6C\U00000345", 0x00001FAC },
{ U"\U00001F6D\U00000345", 0x00001FAD },
{ U"\U00001F6E\U00000345", 0x00001FAE },
{ U"\U00001F6F\U00000345", 0x00001FAF },
{ U"\U000003B1\U00000306", 0x00001FB0 },
{ U"\U000003B1\U00000304", 0x00001FB1 },
{ U"\U00001F70\U00000345", 0x00001FB2 },
{ U"\U000003B1\U00000345", 0x00001FB3 },
{ U"\U000003AC\U00000345", 0x00001FB4 },
{ U"\U000003B1\U00000342", 0x00001FB6 },
{ U"\U00001FB6\U00000345", 0x00001FB7 },
{ U"\U00000391\U00000306", 0x00001FB8 },
{ U"\U00000391\U00000304", 0x00001FB9 },
{ U"\U00000391\U00000300", 0x00001FBA },
{ U"\U00000391\U00000345", 0x00001FBC },
{ U"\U000000A8\U00000342", 0x00001FC1 },
{ U"\U00001F74\U00000345", 0x00001FC2 },
{ U"\U000003B7\U00000345", 0x00001FC3 },
{ U"\U000003AE\U00000345", 0x00001FC4 },
{ U"\U000003B7\U00000342", 0x00001FC6 },
{ U"\U00001FC6\U00000345", 0x00001FC7 },
{ U"\U00000395\U00000300", 0x00001FC8 },
{ U"\U00000397\U00000300", 0x00001FCA },
{ U"\U00000397\U00000345", 0x00001FCC },
{ U"\U00001FBF\U00000300", 0x00001FCD },
{ U"\U00001FBF\U00000301", 0x00001FCE },
{ U"\U00001FBF\U00000342", 0x00001FCF },
{ U"\U000003B9\U00000306", 0x00001FD0 },
{ U"\U000003B9\U00000304", 0x00001FD1 },
{ U"\U000003CA\U00000300", 0x00001FD2 },
{ U"\U000003B9\U00000342", 0x00001FD6 },
{ U"\U000003CA\U00000342", 0x00001FD7 },
{ U"\U00000399\U00000306", 0x00001FD8 },
{ U"\U00000399\U00000304", 0x00001FD9 },
{ U"\U00000399\U00000300", 0x00001FDA },
{ U"\U00001FFE\U00000300", 0x00001FDD },
{ U"\U00001FFE\U00000301", 0x00001FDE },
{ U"\U00001FFE\U00000342", 0x00001FDF },
{ U"\U000003C5\U00000306", 0x00001FE0 },
{ U"\U000003C5\U00000304", 0x00001FE1 },
{ U"\U000003CB\U00000300", 0x00001FE2 },
{ U"\U000003C1\U00000313", 0x00001FE4 },
{ U"\U000003C1\U00000314", 0x00001FE5 },
{ U"\U000003C5\U00000342", 0x00001FE6 },
{ U"\U000003CB\U00000342", 0x00001FE7 },
{ U"\U000003A5\U00000306", 0x00001FE8 },
{ U"\U000003A5\U00000304", 0x00001FE9 },
{ U"\U000003A5\U00000300", 0x00001FEA },
{ U"\U000003A1\U00000314", 0x00001FEC },
{ U"\U000000A8\U00000300", 0x00001FED },
{ U"\U00001F7C\U00000345", 0x00001FF2 },
{ U"\U000003C9\U00000345", 0x00001FF3 },
{ U"\U000003CE\U00000345", 0x00001FF4 },
{ U"\U000003C9\U00000342", 0x00001FF6 },
{ U"\U00001FF6\U00000345", 0x00001FF7 },
{ U"\U0000039F\U00000300", 0x00001FF8 },
{ U"\U000003A9\U00000300", 0x00001FFA },
{ U"\U000003A9\U00000345", 0x00001FFC },
{ U"\U00002190\U00000338", 0x0000219A },
{ U"\U00002192\U00000338", 0x0000219B },
{ U"\U00002194\U00000338", 0x000021AE },
{ U"\U000021D0\U00000338", 0x000021CD },
{ U"\U000021D4\U00000338", 0x000021CE },
{ U"\U000021D2\U00000338", 0x000021CF },
{ U"\U00002203\U00000338", 0x00002204 },
{ U"\U00002208\U00000338", 0x00002209 },
{ U"\U0000220B\U00000338", 0x0000220C },
{ U"\U00002223\U00000338", 0x00002224 },
{ U"\U00002225\U00000338", 0x00002226 },
{ U"\U0000223C\U00000338", 0x00002241 },
{ U"\U00002243\U00000338", 0x00002244 },
{ U"\U00002245\U00000338", 0x00002247 },
{ U"\U00002248\U00000338", 0x00002249 },
{ U"\U0000003D\U00000338", 0x00002260 },
{ U"\U00002261\U00000338", 0x00002262 },
{ U"\U0000224D\U00000338", 0x0000226D },
{ U"\U0000003C\U00000338", 0x0000226E },
{ U"\U0000003E\U00000338", 0x0000226F },
{ U"\U00002264\U00000338", 0x00002270 },
{ U"\U00002265\U00000338", 0x00002271 },
{ U"\U00002272\U00000338", 0x00002274 },
{ U"\U00002273\U00000338", 0x00002275 },
{ U"\U00002276\U00000338", 0x00002278 },
{ U"\U00002277\U00000338", 0x00002279 },
{ U"\U0000227A\U00000338", 0x00002280 },
{ U"\U0000227B\U00000338", 0x00002281 },
{ U"\U00002282\U00000338", 0x00002284 },
{ U"\U00002283\U00000338", 0x00002285 },
{ U"\U00002286\U00000338", 0x00002288 },
{ U"\U00002287\U00000338", 0x00002289 },
{ U"\U000022A2\U00000338", 0x000022AC },
{ U"\U000022A8\U00000338", 0x000022AD },
{ U"\U000022A9\U00000338", 0x000022AE },
{ U"\U000022AB\U00000338", 0x000022AF },
{ U"\U0000227C\U00000338", 0x000022E0 },
{ U"\U0000227D\U00000338", 0x000022E1 },
{ U"\U00002291\U00000338", 0x000022E2 },
{ U"\U00002292\U00000338", 0x000022E3 },
{ U"\U000022B2\U00000338", 0x000022EA },
{ U"\U000022B3\U00000338", 0x000022EB },
{ U"\U000022B4\U00000338", 0x000022EC },
{ U"\U000022B5\U00000338", 0x000022ED },
{ U"\U0000304B\U00003099", 0x0000304C },
{ U"\U0000304D\U00003099", 0x0000304E },
{ U"\U0000304F\U00003099", 0x00003050 },
{ U"\U00003051\U00003099", 0x00003052 },
{ U"\U00003053\U00003099", 0x00003054 },
{ U"\U00003055\U00003099", 0x00003056 },
{ U"\U00003057\U00003099", 0x00003058 },
{ U"\U00003059\U00003099", 0x0000305A },
{ U"\U0000305B\U00003099", 0x0000305C },
{ U"\U0000305D\U00003099", 0x0000305E },
{ U"\U0000305F\U00003099", 0x00003060 },
{ U"\U00003061\U00003099", 0x00003062 },
{ U"\U00003064\U00003099", 0x00003065 },
{ U"\U00003066\U00003099", 0x00003067 },
{ U"\U00003068\U00003099", 0x00003069 },
{ U"\U0000306F\U00003099", 0x00003070 },
{ U"\U0000306F\U0000309A", 0x00003071 },
{ U"\U00003072\U00003099", 0x00003073 },
{ U"\U00003072\U0000309A", 0x00003074 },
{ U"\U00003075\U00003099", 0x00003076 },
{ U"\U00003075\U0000309A", 0x00003077 },
{ U"\U00003078\U00003099", 0x00003079 },
{ U"\U00003078\U0000309A", 0x0000307A },
{ U"\U0000307B\U00003099", 0x0000307C },
{ U"\U0000307B\U0000309A", 0x0000307D },
{ U"\U00003046\U00003099", 0x00003094 },
{ U"\U0000309D\U00003099", 0x0000309E },
{ U"\U000030AB\U00003099", 0x000030AC },
{ U"\U000030AD\U00003099", 0x000030AE },
{ U"\U000030AF\U00003099", 0x000030B0 },
{ U"\U000030B1\U00003099", 0x000030B2 },
{ U"\U000030B3\U00003099", 0x000030B4 },
{ U"\U000030B5\U00003099", 0x000030B6 },
{ U"\U000030B7\U00003099", 0x000030B8 },
{ U"\U000030B9\U00003099", 0x000030BA },
{ U"\U000030BB\U00003099", 0x000030BC },
{ U"\U000030BD\U00003099", 0x000030BE },
{ U"\U000030BF\U00003099", 0x000030C0 },
{ U"\U000030C1\U00003099", 0x000030C2 },
{ U"\U000030C4\U00003099", 0x000030C5 },
{ U"\U000030C6\U00003099", 0x000030C7 },
{ U"\U000030C8\U00003099", 0x000030C9 },
{ U"\U000030CF\U00003099", 0x000030D0 },
{ U"\U000030CF\U0000309A", 0x000030D1 },
{ U"\U000030D2\U00003099", 0x000030D3 },
{ U"\U000030D2\U0000309A", 0x000030D4 },
{ U"\U000030D5\U00003099", 0x000030D6 },
{ U"\U000030D5\U0000309A", 0x000030D7 },
{ U"\U000030D8\U00003099", 0x000030D9 },
{ U"\U000030D8\U0000309A", 0x000030DA },
{ U"\U000030DB\U00003099", 0x000030DC },
{ U"\U000030DB\U0000309A", 0x000030DD },
{ U"\U000030A6\U00003099", 0x000030F4 },
{ U"\U000030EF\U00003099", 0x000030F7 },
{ U"\U000030F0\U00003099", 0x000030F8 },
{ U"\U000030F1\U00003099", 0x000030F9 },
{ U"\U000030F2\U00003099", 0x000030FA },
{ U"\U000030FD\U00003099", 0x000030FE },
{ U"\U00011099\U000110BA", 0x0001109A },
{ U"\U0001109B\U000110BA", 0x0001109C },
{ U"\U000110A5\U000110BA", 0x000110AB },
{ U"\U00011131\U00011127", 0x0001112E },
{ U"\U00011132\U00011127", 0x0001112F },
{ U"\U00011347\U0001133E", 0x0001134B },
{ U"\U00011347\U00011357", 0x0001134C },
{ U"\U000114B9\U000114BA", 0x000114BB },
{ U"\U000114B9\U000114B0", 0x000114BC },
{ U"\U000114B9\U000114BD", 0x000114BE },
{ U"\U000115B8\U000115AF", 0x000115BA },
{ U"\U000115B9\U000115AF", 0x000115BB },
{ U"\U00011935\U00011930", 0x00011938 },
};
namespace _grapheme_break_properties {
using T = GraphemeBreak;
const auto D = GraphemeBreak::Unassigned;
static const size_t _block_size = 128;
static const GraphemeBreak _0[] = { T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::LF,T::Control,T::Control,T::CR,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Control, };
static const GraphemeBreak _1[] = { T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Control,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _6[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _9[] = { D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _11[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _12[] = { T::Prepend,T::Prepend,T::Prepend,T::Prepend,T::Prepend,T::Prepend,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Control,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _13[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Prepend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _14[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Prepend,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _15[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,T::Extend,D,D, };
static const GraphemeBreak _16[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _17[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Prepend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const GraphemeBreak _18[] = { T::Extend,T::Extend,T::Extend,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,T::Extend,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _19[] = { D,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::SpacingMark,T::SpacingMark,D,D,T::SpacingMark,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D, };
static const GraphemeBreak _20[] = { D,T::Extend,T::Extend,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _21[] = { D,T::Extend,T::Extend,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::SpacingMark,D,T::SpacingMark,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const GraphemeBreak _22[] = { D,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::SpacingMark,T::SpacingMark,D,D,T::SpacingMark,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _23[] = { D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _24[] = { T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _25[] = { D,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,D,T::Extend,T::SpacingMark,T::SpacingMark,D,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _26[] = { T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Prepend,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _27[] = { D,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,D,T::Extend,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _28[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _29[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _30[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::Extend,D,T::Extend,D,D,D,D,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark, };
static const GraphemeBreak _31[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _32[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _33[] = { D,D,T::Extend,D,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _34[] = { T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V, };
static const GraphemeBreak _35[] = { T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T, };
static const GraphemeBreak _38[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _46[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _47[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _48[] = { D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Control,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _49[] = { D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _50[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,D,D,D,D,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _52[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,D,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend, };
static const GraphemeBreak _53[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _54[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _55[] = { T::Extend,T::Extend,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _56[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _57[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,D,D,D,D,D,D,T::Extend,D,D,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D, };
static const GraphemeBreak _59[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const GraphemeBreak _64[] = { D,D,D,D,D,D,D,D,D,D,D,T::Control,T::Extend,T::ZWJ,T::Control,T::Control,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _65[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _89[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _90[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend, };
static const GraphemeBreak _91[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const GraphemeBreak _96[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _97[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _332[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D, };
static const GraphemeBreak _333[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _336[] = { D,D,T::Extend,D,D,D,T::Extend,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::SpacingMark,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _337[] = { T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend, };
static const GraphemeBreak _338[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,T::L,D,D,D, };
static const GraphemeBreak _339[] = { T::Extend,T::Extend,T::Extend,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _340[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D, };
static const GraphemeBreak _341[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,D,D,D,D,D,T::Extend,T::Extend,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _343[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,D,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _344[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _345[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _346[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _347[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _348[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _349[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _350[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _351[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _352[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _353[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _354[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _355[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _356[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _357[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _358[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _359[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _360[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _361[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _362[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _363[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _364[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _365[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _366[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _367[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _368[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _369[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _370[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _371[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _372[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _373[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _374[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _375[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _376[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _377[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _378[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _379[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _380[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _381[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _382[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _383[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _384[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _385[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _386[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _387[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _388[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _389[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _390[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _391[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _392[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _393[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _394[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _395[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _396[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _397[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _398[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _399[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _400[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _401[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _402[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _403[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _404[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _405[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _406[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _407[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _408[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _409[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _410[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _411[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _412[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _413[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _414[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _415[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _416[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _417[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _418[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _419[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _420[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _421[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _422[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _423[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _424[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _425[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _426[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _427[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _428[] = { T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _429[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _430[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT, };
static const GraphemeBreak _431[] = { T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LV,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,T::LVT,D,D,D,D,D,D,D,D,D,D,D,D,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,T::V,D,D,D,D,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,T::T,D,D,D,D, };
static const GraphemeBreak _502[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _508[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _509[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Control, };
static const GraphemeBreak _511[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,D,D,D,D, };
static const GraphemeBreak _515[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D, };
static const GraphemeBreak _517[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _518[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D, };
static const GraphemeBreak _532[] = { D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _533[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _538[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _541[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _542[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _544[] = { T::SpacingMark,T::Extend,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend, };
static const GraphemeBreak _545[] = { T::Extend,T::Extend,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,D,D,T::Prepend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Prepend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _546[] = { T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _547[] = { T::Extend,T::Extend,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,D,T::Prepend,T::Prepend,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _548[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _549[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _550[] = { T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,T::Extend,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,D,D,T::SpacingMark,T::SpacingMark,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _552[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _553[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _555[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _556[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _557[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _558[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _560[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _562[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,D,T::SpacingMark,T::SpacingMark,D,D,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Prepend,T::SpacingMark,T::Prepend,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _563[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::Extend,D,D,D,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _564[] = { D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Prepend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _565[] = { D,D,D,D,T::Prepend,T::Prepend,T::Prepend,T::Prepend,T::Prepend,T::Prepend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _568[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _569[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::SpacingMark,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _570[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Prepend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _571[] = { D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,D,T::Extend,T::Extend,D,T::SpacingMark,T::SpacingMark,T::Extend,T::SpacingMark,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _573[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _616[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _725[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _726[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _734[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark, };
static const GraphemeBreak _735[] = { T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,T::SpacingMark,T::SpacingMark,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _889[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,T::Control,T::Control,T::Control,T::Control,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _930[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::SpacingMark,T::Extend,T::Extend,T::Extend,D,D,D,T::SpacingMark,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const GraphemeBreak _931[] = { T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _932[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _948[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _949[] = { D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _960[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _962[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _965[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _977[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _978[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const GraphemeBreak _995[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator, };
static const GraphemeBreak _999[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const GraphemeBreak _7168[] = { T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const GraphemeBreak _7171[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control, };
static const GraphemeBreak *_blocks[] = {
 _0,_1,0,0,0,0,_6,0,
 0,_9,0,_11,_12,_13,_14,_15,
 _16,_17,_18,_19,_20,_21,_22,_23,
 _24,_25,_26,_27,_28,_29,_30,_31,
 _32,_33,_34,_35,0,0,_38,0,
 0,0,0,0,0,0,_46,_47,
 _48,_49,_50,0,_52,_53,_54,_55,
 _56,_57,0,_59,0,0,0,0,
 _64,_65,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_89,_90,_91,0,0,0,0,
 _96,_97,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_332,_333,0,0,
 _336,_337,_338,_339,_340,_341,0,_343,
 _344,_345,_346,_347,_348,_349,_350,_351,
 _352,_353,_354,_355,_356,_357,_358,_359,
 _360,_361,_362,_363,_364,_365,_366,_367,
 _368,_369,_370,_371,_372,_373,_374,_375,
 _376,_377,_378,_379,_380,_381,_382,_383,
 _384,_385,_386,_387,_388,_389,_390,_391,
 _392,_393,_394,_395,_396,_397,_398,_399,
 _400,_401,_402,_403,_404,_405,_406,_407,
 _408,_409,_410,_411,_412,_413,_414,_415,
 _416,_417,_418,_419,_420,_421,_422,_423,
 _424,_425,_426,_427,_428,_429,_430,_431,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_502,0,
 0,0,0,0,_508,_509,0,_511,
 0,0,0,_515,0,_517,_518,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_532,_533,0,0,
 0,0,_538,0,0,_541,_542,0,
 _544,_545,_546,_547,_548,_549,_550,0,
 _552,_553,0,_555,_556,_557,_558,0,
 _560,0,_562,_563,_564,_565,0,0,
 _568,_569,_570,_571,0,_573,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _616,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_725,_726,0,
 0,0,0,0,0,0,_734,_735,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_889,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_930,_931,_932,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_948,_949,0,0,
 0,0,0,0,0,0,0,0,
 _960,0,_962,0,0,_965,0,0,
 0,0,0,0,0,0,0,0,
 0,_977,_978,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_995,0,0,0,_999,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _7168,0,0,_7171,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
static const GraphemeBreak _block_values[] = {
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,T::Control,T::Extend,D,T::Control,T::Control,T::Control,T::Control,
 T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,
 T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,
 T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,T::Control,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
};
inline GraphemeBreak get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}
namespace _word_break_properties {
using T = WordBreak;
const auto D = WordBreak::Unassigned;
static const size_t _block_size = 128;
static const WordBreak _0[] = { D,D,D,D,D,D,D,D,D,D,T::LF,T::Newline,T::Newline,T::CR,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::WSegSpace,D,T::Double_Quote,D,D,D,D,T::Single_Quote,D,D,D,D,T::MidNum,D,T::MidNumLet,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::MidLetter,T::MidNum,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::ExtendNumLet,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D, };
static const WordBreak _1[] = { D,D,D,D,D,T::Newline,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,D,D,T::Format,D,D,D,D,D,D,D,T::ALetter,D,T::MidLetter,D,D,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _5[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _6[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::MidNum,T::ALetter, };
static const WordBreak _7[] = { D,D,D,D,D,D,T::ALetter,T::MidLetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _9[] = { T::ALetter,T::ALetter,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _10[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::MidLetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _11[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::MidNum,T::ALetter,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,D,T::Extend,D,D,D,D,D,D,D,D,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,D,D,D,D,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::ALetter,T::MidLetter,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _12[] = { T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,D,D,D,D,D,D,T::MidNum,T::MidNum,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Format,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,T::Numeric,T::MidNum,D,T::ALetter,T::ALetter,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _13[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Format,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter, };
static const WordBreak _14[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Format,T::ALetter,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _15[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,D,D,T::MidNum,D,T::ALetter,D,D,T::Extend,D,D, };
static const WordBreak _16[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _17[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Format,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const WordBreak _18[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _19[] = { T::ALetter,T::Extend,T::Extend,T::Extend,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::ALetter,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,T::ALetter,D,T::Extend,D, };
static const WordBreak _20[] = { D,T::Extend,T::Extend,T::Extend,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,D,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::Extend,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _21[] = { D,T::Extend,T::Extend,T::Extend,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const WordBreak _22[] = { D,T::Extend,T::Extend,T::Extend,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _23[] = { D,D,T::Extend,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,D,T::ALetter,D,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::ALetter,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _24[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,D,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _25[] = { T::ALetter,T::Extend,T::Extend,T::Extend,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,T::ALetter,D,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _26[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::Extend,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _27[] = { D,T::Extend,T::Extend,T::Extend,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::Extend,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _28[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _29[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _30[] = { T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::Extend,D,T::Extend,D,D,D,D,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const WordBreak _31[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _32[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _33[] = { D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,D,D,D,D,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _36[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _37[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _38[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _39[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D, };
static const WordBreak _40[] = { D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _44[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _45[] = { T::WSegSpace,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D, };
static const WordBreak _46[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _47[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _48[] = { D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Format,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D, };
static const WordBreak _49[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _50[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _51[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _52[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend, };
static const WordBreak _53[] = { T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _54[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _55[] = { T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _56[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D, };
static const WordBreak _57[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::ALetter,D,D,D,D,D, };
static const WordBreak _59[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const WordBreak _62[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,T::ALetter,D,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D, };
static const WordBreak _63[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D, };
static const WordBreak _64[] = { T::WSegSpace,T::WSegSpace,T::WSegSpace,T::WSegSpace,T::WSegSpace,T::WSegSpace,T::WSegSpace,D,T::WSegSpace,T::WSegSpace,T::WSegSpace,D,T::Extend,T::ZWJ,T::Format,T::Format,D,D,D,D,D,D,D,D,T::MidNumLet,T::MidNumLet,D,D,D,D,D,D,D,D,D,D,T::MidNumLet,D,D,T::MidLetter,T::Newline,T::Newline,T::Format,T::Format,T::Format,T::Format,T::Format,T::ExtendNumLet,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ExtendNumLet,T::ExtendNumLet,D,D,D,T::MidNum,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ExtendNumLet,D,D,D,D,D,D,D,D,D,D,T::WSegSpace,T::Format,T::Format,T::Format,T::Format,T::Format,D,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter, };
static const WordBreak _65[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _66[] = { D,D,T::ALetter,D,D,D,D,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,T::ALetter,D,T::ALetter,D,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _67[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _73[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _88[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _89[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _90[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,D,D,D,D,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend, };
static const WordBreak _91[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const WordBreak _92[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _96[] = { T::WSegSpace,D,D,D,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,D,D,D,D,D,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _97[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Katakana,T::Katakana,D,D,D,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,D,T::Katakana,T::Katakana,T::Katakana,T::Katakana, };
static const WordBreak _98[] = { D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _99[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana, };
static const WordBreak _101[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,D, };
static const WordBreak _102[] = { T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _329[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D, };
static const WordBreak _332[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::ALetter, };
static const WordBreak _333[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _334[] = { D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _335[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _336[] = { T::ALetter,T::ALetter,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _337[] = { T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,D,T::ALetter,T::ALetter,T::Extend, };
static const WordBreak _338[] = { T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D, };
static const WordBreak _339[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const WordBreak _340[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D, };
static const WordBreak _341[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,D,D,D,D,D,T::Extend,T::Extend,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D, };
static const WordBreak _342[] = { D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _343[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const WordBreak _431[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D, };
static const WordBreak _502[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,T::Hebrew_Letter,T::Extend,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,D,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,D,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,D,T::Hebrew_Letter,D,T::Hebrew_Letter,T::Hebrew_Letter,D,T::Hebrew_Letter,T::Hebrew_Letter,D,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::Hebrew_Letter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _503[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _506[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _507[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D, };
static const WordBreak _508[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::MidNum,D,D,T::MidLetter,T::MidNum,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::ExtendNumLet,T::ExtendNumLet,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ExtendNumLet,T::ExtendNumLet,T::ExtendNumLet,T::MidNum,D,T::MidNumLet,D,T::MidNum,T::MidLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _509[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Format, };
static const WordBreak _510[] = { D,D,D,D,D,D,D,T::MidNumLet,D,D,D,D,T::MidNum,D,T::MidNumLet,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::MidLetter,T::MidNum,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::ExtendNumLet,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana, };
static const WordBreak _511[] = { T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Katakana,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Format,T::Format,T::Format,D,D,D,D, };
static const WordBreak _512[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _513[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D, };
static const WordBreak _514[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _515[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D, };
static const WordBreak _517[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _518[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D, };
static const WordBreak _519[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _521[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D, };
static const WordBreak _522[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _526[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _528[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,D,D,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D, };
static const WordBreak _529[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _530[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _531[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _532[] = { T::ALetter,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D, };
static const WordBreak _533[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _534[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _535[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _536[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _537[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _538[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _541[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::Extend,T::Extend,D,D,D,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _542[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,T::ALetter,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _543[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D, };
static const WordBreak _544[] = { T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend, };
static const WordBreak _545[] = { T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Format,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Format,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const WordBreak _546[] = { T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,T::ALetter,T::Extend,T::Extend,T::ALetter,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,D,D,T::ALetter,D,D,D,D,D,D,D,D,D, };
static const WordBreak _547[] = { T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::ALetter,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _548[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _549[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const WordBreak _550[] = { T::Extend,T::Extend,T::Extend,T::Extend,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::Extend,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,D,D,T::ALetter,D,D,D,D,D,D,T::Extend,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _552[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,T::Extend,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _553[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,D,T::ALetter,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _555[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _556[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _557[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _558[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _560[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _561[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter, };
static const WordBreak _562[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::Extend,T::ALetter,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _563[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,D,T::ALetter,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _564[] = { T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _565[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D, };
static const WordBreak _568[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _569[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _570[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::Extend,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _571[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _573[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D, };
static const WordBreak _575[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _583[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _584[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _586[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _616[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _652[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _724[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _725[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _726[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _727[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _732[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _734[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,T::Extend,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const WordBreak _735[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,D,T::ALetter,T::Extend,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _864[] = { T::Katakana,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _866[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Katakana,T::Katakana,T::Katakana,T::Katakana,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _888[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D, };
static const WordBreak _889[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::Extend,T::Extend,D,T::Format,T::Format,T::Format,T::Format,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _930[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const WordBreak _931[] = { T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _932[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _936[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _937[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,D,T::ALetter,D,D,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _938[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _941[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _942[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _943[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric, };
static const WordBreak _948[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _949[] = { D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _960[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _962[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _965[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const WordBreak _977[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _978[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::ALetter,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _988[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,D,D,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D,T::ALetter,D,D,D,D,D,D,T::ALetter,D,D,D,D,T::ALetter,D,T::ALetter,D,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,D,D,T::ALetter,D,T::ALetter,D,T::ALetter,D,T::ALetter,D,T::ALetter,D,T::ALetter,T::ALetter,D,T::ALetter,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,D, };
static const WordBreak _989[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak _994[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter, };
static const WordBreak _995[] = { T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator,T::Regional_Indicator, };
static const WordBreak _999[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const WordBreak _1015[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const WordBreak _7168[] = { D,T::Format,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const WordBreak _7171[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const WordBreak *_blocks[] = {
 _0,_1,0,0,0,_5,_6,_7,
 0,_9,_10,_11,_12,_13,_14,_15,
 _16,_17,_18,_19,_20,_21,_22,_23,
 _24,_25,_26,_27,_28,_29,_30,_31,
 _32,_33,0,0,_36,_37,_38,_39,
 _40,0,0,0,_44,_45,_46,_47,
 _48,_49,_50,_51,_52,_53,_54,_55,
 _56,_57,0,_59,0,0,_62,_63,
 _64,_65,_66,_67,0,0,0,0,
 0,_73,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _88,_89,_90,_91,_92,0,0,0,
 _96,_97,_98,_99,0,_101,_102,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_329,0,0,_332,_333,_334,_335,
 _336,_337,_338,_339,_340,_341,_342,_343,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_431,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_502,_503,
 0,0,_506,_507,_508,_509,_510,_511,
 _512,_513,_514,_515,0,_517,_518,_519,
 0,_521,_522,0,0,0,_526,0,
 _528,_529,_530,_531,_532,_533,_534,_535,
 _536,_537,_538,0,0,_541,_542,_543,
 _544,_545,_546,_547,_548,_549,_550,0,
 _552,_553,0,_555,_556,_557,_558,0,
 _560,_561,_562,_563,_564,_565,0,0,
 _568,_569,_570,_571,0,_573,0,_575,
 0,0,0,0,0,0,0,_583,
 _584,0,_586,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _616,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_652,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_724,_725,_726,_727,
 0,0,0,0,_732,0,_734,_735,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _864,0,_866,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _888,_889,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_930,_931,_932,0,0,0,
 _936,_937,_938,0,0,_941,_942,_943,
 0,0,0,0,_948,_949,0,0,
 0,0,0,0,0,0,0,0,
 _960,0,_962,0,0,_965,0,0,
 0,0,0,0,0,0,0,0,
 0,_977,_978,0,0,0,0,0,
 0,0,0,0,_988,_989,0,0,
 0,0,_994,_995,0,0,0,_999,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_1015,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _7168,0,0,_7171,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
static const WordBreak _block_values[] = {
 D,D,T::ALetter,T::ALetter,T::ALetter,D,D,D,
 T::ALetter,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::ALetter,T::ALetter,D,D,D,D,
 D,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::ALetter,D,T::ALetter,T::ALetter,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,D,T::ALetter,T::ALetter,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::ALetter,T::ALetter,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::ALetter,D,D,D,T::ALetter,T::ALetter,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,
 D,T::ALetter,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,T::ALetter,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::ALetter,T::ALetter,T::ALetter,T::ALetter,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,T::ALetter,T::ALetter,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::ALetter,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::Extend,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
};
inline WordBreak get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}
namespace _sentence_break_properties {
using T = SentenceBreak;
const auto D = SentenceBreak::Unassigned;
static const size_t _block_size = 128;
static const SentenceBreak _0[] = { D,D,D,D,D,D,D,D,D,T::Sp,T::LF,T::Sp,T::Sp,T::CR,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Sp,T::STerm,T::Close,D,D,D,D,T::Close,T::Close,T::Close,D,D,T::SContinue,T::SContinue,T::ATerm,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::SContinue,D,D,D,D,T::STerm,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Close,D,T::Close,D,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Close,D,T::Close,D,D, };
static const SentenceBreak _1[] = { D,D,D,D,D,T::Sep,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Sp,D,D,D,D,D,D,D,D,D,T::Lower,T::Close,D,T::Format,D,D,D,D,D,D,D,T::Lower,D,D,D,D,T::Lower,T::Close,D,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower, };
static const SentenceBreak _2[] = { T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower, };
static const SentenceBreak _3[] = { T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Lower,T::OLetter,T::Upper,T::Lower,T::Lower,T::Lower,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Upper,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower, };
static const SentenceBreak _4[] = { T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower, };
static const SentenceBreak _5[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::OLetter,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Lower,T::Lower,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,D,D,D,D,D,T::OLetter,D,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _6[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Upper,T::Lower,T::Upper,T::Lower,T::OLetter,D,T::Upper,T::Lower,D,D,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Upper, };
static const SentenceBreak _7[] = { D,D,D,D,D,D,T::Upper,D,T::Upper,T::Upper,T::Upper,D,T::Upper,D,T::Upper,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Lower,D,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper, };
static const SentenceBreak _8[] = { T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower, };
static const SentenceBreak _9[] = { T::Upper,T::Lower,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower, };
static const SentenceBreak _10[] = { T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,T::OLetter,D,D,D,T::SContinue,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower, };
static const SentenceBreak _11[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::STerm,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,D,T::Extend,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _12[] = { T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,D,D,D,D,D,D,T::SContinue,T::SContinue,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Format,D,T::STerm,T::STerm,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,T::Numeric,T::Numeric,D,T::OLetter,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _13[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::STerm,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Format,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter, };
static const SentenceBreak _14[] = { T::STerm,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,T::Format,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _15[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,D,D,T::SContinue,T::STerm,T::OLetter,D,D,T::Extend,D,D, };
static const SentenceBreak _16[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,T::STerm,D,T::STerm,D,D,D,T::STerm,T::STerm,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _17[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Format,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const SentenceBreak _18[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::STerm,T::STerm,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _19[] = { T::OLetter,T::Extend,T::Extend,T::Extend,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::OLetter,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,T::OLetter,D,T::Extend,D, };
static const SentenceBreak _20[] = { D,T::Extend,T::Extend,T::Extend,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,D,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::Extend,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _21[] = { D,T::Extend,T::Extend,T::Extend,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,D,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const SentenceBreak _22[] = { D,T::Extend,T::Extend,T::Extend,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _23[] = { D,D,T::Extend,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::OLetter,D,T::OLetter,D,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::OLetter,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _24[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,D,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _25[] = { T::OLetter,T::Extend,T::Extend,T::Extend,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,T::OLetter,D,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _26[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::Extend,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _27[] = { D,T::Extend,T::Extend,T::Extend,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::Extend,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _28[] = { D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _29[] = { D,T::OLetter,T::OLetter,D,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _30[] = { T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,T::Extend,D,T::Extend,T::Close,T::Close,T::Close,T::Close,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const SentenceBreak _31[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _32[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::STerm,T::STerm,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _33[] = { T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::Extend,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Upper,D,D,D,D,D,T::Upper,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _36[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _37[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _38[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Extend,T::Extend,T::Extend,D,D,T::STerm,D,D,D,D,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _39[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D, };
static const SentenceBreak _40[] = { D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _44[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::STerm,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _45[] = { T::Sp,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Close,T::Close,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D, };
static const SentenceBreak _46[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _47[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::OLetter,D,D,D,D,T::OLetter,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _48[] = { D,D,T::SContinue,T::STerm,D,D,D,D,T::SContinue,T::STerm,D,T::Extend,T::Extend,T::Extend,T::Format,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D, };
static const SentenceBreak _49[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::OLetter,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _50[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,T::STerm,T::STerm,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _51[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _52[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend, };
static const SentenceBreak _53[] = { T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::STerm,T::STerm,T::STerm,T::STerm,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _54[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::STerm,T::STerm,D,D,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _55[] = { T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _56[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::STerm,T::STerm,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::STerm,T::STerm, };
static const SentenceBreak _57[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::OLetter,D,D,D,D,D, };
static const SentenceBreak _59[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const SentenceBreak _60[] = { T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower, };
static const SentenceBreak _61[] = { T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower, };
static const SentenceBreak _62[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Upper,D,T::Upper,D,T::Upper,D,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D, };
static const SentenceBreak _63[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Lower,D,D,D,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,D,T::Lower,T::Lower,T::Lower,T::Lower,D,D,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,D,D,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,D,D,D,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,D, };
static const SentenceBreak _64[] = { T::Sp,T::Sp,T::Sp,T::Sp,T::Sp,T::Sp,T::Sp,T::Sp,T::Sp,T::Sp,T::Sp,T::Format,T::Extend,T::Extend,T::Format,T::Format,D,D,D,T::SContinue,T::SContinue,D,D,D,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,D,D,D,T::ATerm,D,D,D,T::Sep,T::Sep,T::Format,T::Format,T::Format,T::Format,T::Format,T::Sp,D,D,D,D,D,D,D,D,D,T::Close,T::Close,D,T::STerm,T::STerm,D,D,D,D,D,D,D,T::Close,T::Close,T::STerm,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Sp,T::Format,T::Format,T::Format,T::Format,T::Format,D,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,D,T::Lower,D,D,D,D,D,D,D,D,D,D,D,T::Close,T::Close,T::Lower, };
static const SentenceBreak _65[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,T::Close,T::Close,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _66[] = { D,D,T::Upper,D,D,D,D,T::Upper,D,D,T::Lower,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Lower,D,T::Upper,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,D,D,D,D,T::Upper,D,T::Upper,D,T::Upper,D,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Lower,D,D,T::Lower,T::Lower,T::Upper,T::Upper,D,D,D,D,D,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,D,D,D,D,T::Lower,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower, };
static const SentenceBreak _67[] = { T::OLetter,T::OLetter,T::OLetter,T::Upper,T::Lower,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _70[] = { D,D,D,D,D,D,D,D,T::Close,T::Close,T::Close,T::Close,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Close,T::Close,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _73[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _78[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,D,D,D,D,D,D,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _79[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Close,T::Close,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _83[] = { D,D,D,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Close,T::Close,T::Close,T::Close,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Close,T::Close,D,D, };
static const SentenceBreak _88[] = { T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper, };
static const SentenceBreak _89[] = { T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,D,D,D,D,D,D,T::Upper,T::Lower,T::Upper,T::Lower,T::Extend,T::Extend,T::Extend,T::Upper,T::Lower,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _90[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,D,D,D,D,D,T::Lower,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend, };
static const SentenceBreak _91[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const SentenceBreak _92[] = { T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Close,T::Close,D,D,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,D,D,D,T::STerm,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,T::STerm,D,D,D,D,D,T::Close,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _96[] = { T::Sp,T::SContinue,T::STerm,D,D,T::OLetter,T::OLetter,T::OLetter,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,D,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,T::Close,T::Close,T::Close,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _97[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Extend,T::Extend,D,D,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _98[] = { D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _99[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _155[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _319[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D, };
static const SentenceBreak _329[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::STerm, };
static const SentenceBreak _332[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::STerm,T::STerm,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::OLetter, };
static const SentenceBreak _333[] = { T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Lower,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,D,T::STerm,D,D,D,T::STerm,D,D,D,D,D,D,D,D, };
static const SentenceBreak _334[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Lower, };
static const SentenceBreak _335[] = { T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::OLetter,D,D,T::Upper,T::Lower,T::Upper,T::Lower,T::OLetter,T::Upper,T::Lower,T::Upper,T::Lower,T::Lower,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,T::Upper,T::Lower,D,D,T::Upper,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Upper,T::Lower,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Upper,T::Lower,T::OLetter,T::Lower,T::Lower,T::Lower,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _336[] = { T::OLetter,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::STerm,T::STerm,D,D,D,D,D,D,D,D, };
static const SentenceBreak _337[] = { T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,T::STerm,T::STerm,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::OLetter,D,T::OLetter,T::OLetter,T::Extend, };
static const SentenceBreak _338[] = { T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::STerm,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D, };
static const SentenceBreak _339[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,T::STerm,T::STerm,D,D,D,D,D,T::OLetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D, };
static const SentenceBreak _340[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,T::STerm,T::STerm,T::STerm,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter, };
static const SentenceBreak _341[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::OLetter,T::Extend,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::STerm,T::STerm,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _342[] = { D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::OLetter,D,D,D,D,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower, };
static const SentenceBreak _343[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::STerm,T::Extend,T::Extend,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const SentenceBreak _431[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D, };
static const SentenceBreak _500[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _501[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _502[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,D,D,D,D,D,D,D,D,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,D,D,D,T::OLetter,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _503[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _506[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Close,T::Close,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _507[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D, };
static const SentenceBreak _508[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::SContinue,T::SContinue,D,T::SContinue,D,D,D,T::Close,T::Close,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::SContinue,T::SContinue,D,D,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,D,T::Close,T::Close,D,D,D,D,D,D,D,T::SContinue,T::SContinue,T::ATerm,D,D,T::SContinue,T::STerm,T::STerm,T::SContinue,T::Close,T::Close,T::Close,T::Close,T::Close,T::Close,D,D,D,D,T::SContinue,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _509[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Format, };
static const SentenceBreak _510[] = { D,T::STerm,D,D,D,D,D,D,T::Close,T::Close,D,D,T::SContinue,T::SContinue,T::ATerm,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::SContinue,D,D,D,D,T::STerm,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Close,D,T::Close,D,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Close,D,T::Close,D,T::Close,T::Close,T::STerm,T::Close,T::Close,T::SContinue,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _511[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Format,T::Format,T::Format,D,D,D,D, };
static const SentenceBreak _512[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _513[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D, };
static const SentenceBreak _514[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _515[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D, };
static const SentenceBreak _517[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _518[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D, };
static const SentenceBreak _519[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _520[] = { T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _521[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,D,D, };
static const SentenceBreak _522[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _526[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _528[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,D,D,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _529[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _530[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _531[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _532[] = { T::OLetter,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::STerm,T::STerm,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D, };
static const SentenceBreak _533[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _534[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _535[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _536[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _537[] = { T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _538[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _541[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::Extend,T::Extend,D,D,D,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _542[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,T::OLetter,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _543[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _544[] = { T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend, };
static const SentenceBreak _545[] = { T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Format,T::STerm,T::STerm,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,T::Format,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const SentenceBreak _546[] = { T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,T::STerm,T::STerm,T::STerm,T::OLetter,T::Extend,T::Extend,T::OLetter,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,D,D,T::OLetter,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _547[] = { T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::STerm,T::STerm,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::STerm,T::Extend,T::Extend,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::OLetter,D,T::OLetter,D,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _548[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::STerm,T::STerm,D,T::STerm,T::STerm,D,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _549[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::STerm,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const SentenceBreak _550[] = { T::Extend,T::Extend,T::Extend,T::Extend,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::Extend,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,D,D,T::OLetter,D,D,D,D,D,D,T::Extend,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _552[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::STerm,T::STerm,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,T::Extend,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _553[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,D,T::OLetter,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _555[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::STerm,T::STerm,D,D,D,D,D,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::STerm,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _556[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::STerm,T::STerm,D,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _557[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _558[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,T::STerm,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _560[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _561[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter, };
static const SentenceBreak _562[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::Extend,T::OLetter,T::Extend,T::Extend,T::STerm,D,T::STerm,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _563[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,D,T::OLetter,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _564[] = { T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::STerm,T::STerm,D,D,D,T::Extend,D,D,D,D,D,D,D,D,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _565[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::STerm,T::STerm,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D, };
static const SentenceBreak _568[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _569[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _570[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::Extend,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _571[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _573[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::STerm,T::STerm,D,D,D,D,D,D,D, };
static const SentenceBreak _575[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _583[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _584[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _586[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _616[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _652[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _724[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,T::STerm,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _725[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::STerm,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _726[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::STerm,T::STerm,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::STerm,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _727[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _732[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower, };
static const SentenceBreak _733[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _734[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,T::Extend,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const SentenceBreak _735[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,D,T::OLetter,T::Extend,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _783[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D, };
static const SentenceBreak _793[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _794[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _866[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _869[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D, };
static const SentenceBreak _888[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D, };
static const SentenceBreak _889[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::Extend,T::Extend,T::STerm,T::Format,T::Format,T::Format,T::Format,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _930[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Format,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const SentenceBreak _931[] = { T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _932[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _936[] = { T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper, };
static const SentenceBreak _937[] = { T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,D,T::Upper,T::Upper,D,D,T::Upper,D,D,T::Upper,T::Upper,D,D,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower, };
static const SentenceBreak _938[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,D,T::Upper,T::Upper,T::Upper,T::Upper,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,D,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Upper,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper, };
static const SentenceBreak _939[] = { T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower, };
static const SentenceBreak _940[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper, };
static const SentenceBreak _941[] = { T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Lower,T::Lower,T::Lower,T::Lower, };
static const SentenceBreak _942[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower, };
static const SentenceBreak _943[] = { T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,D,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Upper,T::Lower,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric, };
static const SentenceBreak _948[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,T::Extend,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _949[] = { D,D,D,D,T::Extend,D,D,D,T::STerm,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _960[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,T::Extend,T::Extend,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _962[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _965[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::Extend,T::Extend,T::Extend,T::Extend,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const SentenceBreak _977[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _978[] = { T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Lower,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::OLetter,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _988[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,D,D,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D,T::OLetter,D,D,D,D,D,D,T::OLetter,D,D,D,D,T::OLetter,D,T::OLetter,D,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,D,D,T::OLetter,D,T::OLetter,D,T::OLetter,D,T::OLetter,D,T::OLetter,D,T::OLetter,T::OLetter,D,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,D, };
static const SentenceBreak _989[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _994[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,D,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,D,D,D,D,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper, };
static const SentenceBreak _995[] = { T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,T::Upper,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _1004[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Close,T::Close,T::Close,D,D,D,D,D,D,D, };
static const SentenceBreak _1015[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,T::Numeric,D,D,D,D,D,D, };
static const SentenceBreak _1357[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _1390[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _1392[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _1437[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter, };
static const SentenceBreak _1495[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _1524[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _1574[] = { T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak _7168[] = { D,T::Format,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend, };
static const SentenceBreak _7171[] = { T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,T::Extend,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const SentenceBreak *_blocks[] = {
 _0,_1,_2,_3,_4,_5,_6,_7,
 _8,_9,_10,_11,_12,_13,_14,_15,
 _16,_17,_18,_19,_20,_21,_22,_23,
 _24,_25,_26,_27,_28,_29,_30,_31,
 _32,_33,0,0,_36,_37,_38,_39,
 _40,0,0,0,_44,_45,_46,_47,
 _48,_49,_50,_51,_52,_53,_54,_55,
 _56,_57,0,_59,_60,_61,_62,_63,
 _64,_65,_66,_67,0,0,_70,0,
 0,_73,0,0,0,0,_78,_79,
 0,0,0,_83,0,0,0,0,
 _88,_89,_90,_91,_92,0,0,0,
 _96,_97,_98,_99,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,_155,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_319,
 0,0,0,0,0,0,0,0,
 0,_329,0,0,_332,_333,_334,_335,
 _336,_337,_338,_339,_340,_341,_342,_343,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_431,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_500,_501,_502,_503,
 0,0,_506,_507,_508,_509,_510,_511,
 _512,_513,_514,_515,0,_517,_518,_519,
 _520,_521,_522,0,0,0,_526,0,
 _528,_529,_530,_531,_532,_533,_534,_535,
 _536,_537,_538,0,0,_541,_542,_543,
 _544,_545,_546,_547,_548,_549,_550,0,
 _552,_553,0,_555,_556,_557,_558,0,
 _560,_561,_562,_563,_564,_565,0,0,
 _568,_569,_570,_571,0,_573,0,_575,
 0,0,0,0,0,0,0,_583,
 _584,0,_586,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _616,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_652,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_724,_725,_726,_727,
 0,0,0,0,_732,_733,_734,_735,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_783,
 0,0,0,0,0,0,0,0,
 0,_793,_794,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_866,0,0,_869,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _888,_889,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,_930,_931,_932,0,0,0,
 _936,_937,_938,_939,_940,_941,_942,_943,
 0,0,0,0,_948,_949,0,0,
 0,0,0,0,0,0,0,0,
 _960,0,_962,0,0,_965,0,0,
 0,0,0,0,0,0,0,0,
 0,_977,_978,0,0,0,0,0,
 0,0,0,0,_988,_989,0,0,
 0,0,_994,_995,0,0,0,0,
 0,0,0,0,_1004,0,0,0,
 0,0,0,0,0,0,0,_1015,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1357,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1390,0,
 _1392,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,_1437,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,_1495,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,_1524,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_1574,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _7168,0,0,_7171,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
static const SentenceBreak _block_values[] = {
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::OLetter,T::OLetter,D,D,D,D,
 D,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::Lower,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,D,T::OLetter,T::OLetter,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::OLetter,T::OLetter,D,D,D,D,
 T::OLetter,T::OLetter,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,T::OLetter,T::OLetter,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,
 D,T::OLetter,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,
 D,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,
 T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,T::OLetter,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,T::Extend,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
};
inline SentenceBreak get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}
namespace _emoji_properties {
using T = Emoji;
const auto D = Emoji::Unassigned;
static const size_t _block_size = 256;
static const Emoji _0[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Emoji_Component,D,D,D,D,D,D,T::Emoji_Component,D,D,D,D,D,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji _32[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,T::Emoji_Component,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Emoji_Component,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji _33[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji _35[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D, };
static const Emoji _36[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji _37[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D, };
static const Emoji _38[] = { T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic, };
static const Emoji _39[] = { T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,T::Extended_Pictographic,D,T::Extended_Pictographic,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,T::Extended_Pictographic,D,D,D,D,T::Extended_Pictographic,D,T::Extended_Pictographic,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji _41[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji _43[] = { D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji _48[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji _50[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji _254[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Emoji_Component,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji _497[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component, };
static const Emoji _498[] = { D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic, };
static const Emoji _499[] = { T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component, };
static const Emoji _501[] = { T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic, };
static const Emoji _502[] = { T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic, };
static const Emoji _503[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic, };
static const Emoji _504[] = { D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic, };
static const Emoji _505[] = { D,D,D,D,D,D,D,D,D,D,D,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic, };
static const Emoji _511[] = { T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,D, };
static const Emoji _3584[] = { D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,T::Emoji_Component,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D,D, };
static const Emoji *_blocks[] = {
 _0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _32,_33,0,_35,_36,_37,_38,_39,
 0,_41,0,_43,0,0,0,0,
 _48,0,_50,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,_254,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,_497,_498,_499,0,_501,_502,_503,
 _504,_505,0,0,0,0,0,_511,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 _3584,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,
};
static const Emoji _block_values[] = {
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 T::Extended_Pictographic,D,D,D,T::Extended_Pictographic,D,D,D,
 D,D,T::Extended_Pictographic,D,T::Extended_Pictographic,T::Extended_Pictographic,T::Extended_Pictographic,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
 D,D,D,D,D,D,D,D,
};
inline Emoji get_value(char32_t cp) {
  auto i = cp / _block_size;
  auto bl = _blocks[i];
  if (bl) {
    auto off = cp % _block_size;
    return bl[off];
  }
  return _block_values[i];
}
}

//-----------------------------------------------------------------------------
// Hangul Algorithm
//-----------------------------------------------------------------------------

namespace hangul {

// Common Constants
const char32_t SBase = 0xAC00;
const char32_t LBase = 0x1100;
const char32_t VBase = 0x1161;
const char32_t TBase = 0x11A7;
const size_t LCount = 19;
const size_t VCount = 21;
const size_t TCount = 28;
const size_t NCount = VCount * TCount;  // 588
const size_t SCount = LCount * NCount;  // 11172

inline bool is_precomposed_syllable(char32_t cp) {
  return SBase <= cp && cp < SBase + SCount;
}

inline bool is_decomposed_syllable(const char32_t *source, size_t len) {
  if (len < 2) {
    return false;
  }

  auto first = source[0];
  auto second = source[1];

  if (LBase <= first && first < LBase + LCount) {
    if (VBase <= second && second < VBase + VCount) {
      return true;
    }
  }

  if (SBase <= first && first < SBase + SCount &&
      ((first - SBase) % TCount) == 0) {
    if (TBase <= second && second < TBase + TCount) {
      return true;
    }
  }

  return false;
}

inline void decompose_hangul(char32_t cp, std::u32string &out) {
  int SIndex = cp - SBase;
  char32_t L = LBase + SIndex / NCount;
  char32_t V = VBase + (SIndex % NCount) / TCount;
  char32_t T = TBase + SIndex % TCount;
  out += L;
  out += V;
  if (T != TBase) {
    out += T;
  }
}

inline size_t compose_hangul(const char32_t *source, size_t len,
                             std::u32string &out) {
  auto last = source[0];  // copy first char
  out += last;

  size_t i = 1;
  for (; i < len; i++) {
    auto ch = source[i];

    // 1. check to see if two current characters are L and V
    int LIndex = last - LBase;
    if (0 <= LIndex && LIndex < static_cast<int>(LCount)) {
      int VIndex = ch - VBase;
      if (0 <= VIndex && VIndex < static_cast<int>(VCount)) {
        // make syllable of form LV
        last =
            static_cast<char32_t>(SBase + (LIndex * VCount + VIndex) * TCount);
        out.back() = last;  // reset last
        continue;           // discard ch
      }
    }

    // 2. check to see if two current characters are LV and T
    int SIndex = last - SBase;
    if (0 <= SIndex && SIndex < static_cast<int>(SCount) &&
        (SIndex % TCount) == 0) {
      int TIndex = ch - TBase;
      if (0 <= TIndex && TIndex <= static_cast<int>(TCount)) {
        // make syllable of form LVT
        last += TIndex;
        out.back() = last;  // reset last
        continue;           // discard ch
      }
    }

    // if neither case was true, break
    break;
  }

  return i;
}

}  // namespace hangul

//-----------------------------------------------------------------------------
// General Category
//-----------------------------------------------------------------------------

inline GeneralCategory general_category(char32_t cp) {
  return _general_category_properties::get_value(cp);
}

inline bool is_cased_letter_category(GeneralCategory gc) {
  switch (gc) {
    case GeneralCategory::Lu:
    case GeneralCategory::Ll:
    case GeneralCategory::Lt:
      return true;
    default:
      return false;
  }
}

inline bool is_letter_category(GeneralCategory gc) {
  switch (gc) {
    case GeneralCategory::Lu:
    case GeneralCategory::Ll:
    case GeneralCategory::Lt:
    case GeneralCategory::Lm:
    case GeneralCategory::Lo:
      return true;
    default:
      return false;
  }
}

inline bool is_mark_category(GeneralCategory gc) {
  switch (gc) {
    case GeneralCategory::Mn:
    case GeneralCategory::Mc:
    case GeneralCategory::Me:
      return true;
    default:
      return false;
  }
}

inline bool is_number_category(GeneralCategory gc) {
  switch (gc) {
    case GeneralCategory::Nd:
    case GeneralCategory::Nl:
    case GeneralCategory::No:
      return true;
    default:
      return false;
  }
}

inline bool is_punctuation_category(GeneralCategory gc) {
  switch (gc) {
    case GeneralCategory::Pc:
    case GeneralCategory::Pd:
    case GeneralCategory::Ps:
    case GeneralCategory::Pe:
    case GeneralCategory::Pi:
    case GeneralCategory::Pf:
    case GeneralCategory::Po:
      return true;
    default:
      return false;
  }
}

inline bool is_symbol_category(GeneralCategory gc) {
  switch (gc) {
    case GeneralCategory::Sm:
    case GeneralCategory::Sc:
    case GeneralCategory::Sk:
    case GeneralCategory::So:
      return true;
    default:
      return false;
  }
}

inline bool is_separator_category(GeneralCategory gc) {
  switch (gc) {
    case GeneralCategory::Zs:
    case GeneralCategory::Zl:
    case GeneralCategory::Zp:
      return true;
    default:
      return false;
  }
}

inline bool is_other_category(GeneralCategory gc) {
  switch (gc) {
    case GeneralCategory::Cc:
    case GeneralCategory::Cf:
    case GeneralCategory::Cs:
    case GeneralCategory::Co:
    case GeneralCategory::Cn:
      return true;
    default:
      return false;
  }
}

inline bool is_cased_letter(char32_t cp) {
  return is_cased_letter_category(general_category(cp));
}

inline bool is_letter(char32_t cp) {
  return is_letter_category(general_category(cp));
}

inline bool is_mark(char32_t cp) {
  return is_mark_category(general_category(cp));
}

inline bool is_number(char32_t cp) {
  return is_number_category(general_category(cp));
}

inline bool is_punctuation(char32_t cp) {
  return is_punctuation_category(general_category(cp));
}

inline bool is_symbol(char32_t cp) {
  return is_symbol_category(general_category(cp));
}

inline bool is_separator(char32_t cp) {
  return is_separator_category(general_category(cp));
}

inline bool is_other(char32_t cp) {
  return is_other_category(general_category(cp));
}

//-----------------------------------------------------------------------------
// Property
//-----------------------------------------------------------------------------

const uint64_t Property_White_Space = 0b00000000000000000000000000000001;
const uint64_t Property_Bidi_Control = 0b00000000000000000000000000000010;
const uint64_t Property_Join_Control = 0b00000000000000000000000000000100;
const uint64_t Property_Dash = 0b00000000000000000000000000001000;
const uint64_t Property_Hyphen = 0b00000000000000000000000000010000;
const uint64_t Property_Quotation_Mark = 0b00000000000000000000000000100000;
const uint64_t Property_Terminal_Punctuation =
    0b00000000000000000000000001000000;
const uint64_t Property_Other_Math = 0b00000000000000000000000010000000;
const uint64_t Property_Hex_Digit = 0b00000000000000000000000100000000;
const uint64_t Property_ASCII_Hex_Digit = 0b00000000000000000000001000000000;
const uint64_t Property_Other_Alphabetic = 0b00000000000000000000010000000000;
const uint64_t Property_Ideographic = 0b00000000000000000000100000000000;
const uint64_t Property_Diacritic = 0b00000000000000000001000000000000;
const uint64_t Property_Extender = 0b00000000000000000010000000000000;
const uint64_t Property_Other_Lowercase = 0b00000000000000000100000000000000;
const uint64_t Property_Other_Uppercase = 0b00000000000000001000000000000000;
const uint64_t Property_Noncharacter_Code_Point =
    0b00000000000000010000000000000000;
const uint64_t Property_Other_Grapheme_Extend =
    0b00000000000000100000000000000000;
const uint64_t Property_IDS_Binary_Operator =
    0b00000000000001000000000000000000;
const uint64_t Property_IDS_Trinary_Operator =
    0b00000000000010000000000000000000;
const uint64_t Property_Radical = 0b00000000000100000000000000000000;
const uint64_t Property_Unified_Ideograph = 0b00000000001000000000000000000000;
const uint64_t Property_Other_Default_Ignorable_Code_Point =
    0b00000000010000000000000000000000;
const uint64_t Property_Deprecated = 0b00000000100000000000000000000000;
const uint64_t Property_Soft_Dotted = 0b00000001000000000000000000000000;
const uint64_t Property_Logical_Order_Exception =
    0b00000010000000000000000000000000;
const uint64_t Property_Other_ID_Start = 0b00000100000000000000000000000000;
const uint64_t Property_Other_ID_Continue = 0b00001000000000000000000000000000;
const uint64_t Property_Sentence_Terminal = 0b00010000000000000000000000000000;
const uint64_t Property_Variation_Selector = 0b00100000000000000000000000000000;
const uint64_t Property_Pattern_White_Space =
    0b01000000000000000000000000000000;
const uint64_t Property_Pattern_Syntax = 0b10000000000000000000000000000000;
const uint64_t Property_Prepended_Concatenation_Mark =
    0b100000000000000000000000000000000;

inline bool is_white_space(char32_t cp) {
  return (_properties::get_value(cp) & Property_White_Space) != 0;
}

inline bool is_bidi_control(char32_t cp) {
  return (_properties::get_value(cp) & Property_Bidi_Control) != 0;
}

inline bool is_join_control(char32_t cp) {
  return (_properties::get_value(cp) & Property_Join_Control) != 0;
}

inline bool is_dash(char32_t cp) {
  return (_properties::get_value(cp) & Property_Dash) != 0;
}

inline bool is_hyphen(char32_t cp) {
  return (_properties::get_value(cp) & Property_Hyphen) != 0;
}

inline bool is_quotation_mark(char32_t cp) {
  return (_properties::get_value(cp) & Property_Quotation_Mark) != 0;
}

inline bool is_terminal_punctuation(char32_t cp) {
  return (_properties::get_value(cp) & Property_Terminal_Punctuation) != 0;
}

inline bool is_other_math(char32_t cp) {
  return (_properties::get_value(cp) & Property_Other_Math) != 0;
}

inline bool is_hex_digit(char32_t cp) {
  return (_properties::get_value(cp) & Property_Hex_Digit) != 0;
}

inline bool is_ascii_hex_digit(char32_t cp) {
  return (_properties::get_value(cp) & Property_ASCII_Hex_Digit) != 0;
}

inline bool is_other_alphabetic(char32_t cp) {
  return (_properties::get_value(cp) & Property_Other_Alphabetic) != 0;
}

inline bool is_ideographic(char32_t cp) {
  return (_properties::get_value(cp) & Property_Ideographic) != 0;
}

inline bool is_diacritic(char32_t cp) {
  return (_properties::get_value(cp) & Property_Diacritic) != 0;
}

inline bool is_extender(char32_t cp) {
  return (_properties::get_value(cp) & Property_Extender) != 0;
}

inline bool is_other_lowercase(char32_t cp) {
  return (_properties::get_value(cp) & Property_Other_Lowercase) != 0;
}

inline bool is_other_uppercase(char32_t cp) {
  return (_properties::get_value(cp) & Property_Other_Uppercase) != 0;
}

inline bool is_noncharacter_code_point(char32_t cp) {
  return (_properties::get_value(cp) & Property_Noncharacter_Code_Point) != 0;
}

inline bool is_other_grapheme_extend(char32_t cp) {
  return (_properties::get_value(cp) & Property_Other_Grapheme_Extend) != 0;
}

inline bool is_ids_binary_operator(char32_t cp) {
  return (_properties::get_value(cp) & Property_IDS_Binary_Operator) != 0;
}

inline bool is_ids_trinary_operator(char32_t cp) {
  return (_properties::get_value(cp) & Property_IDS_Trinary_Operator) != 0;
}

inline bool is_radical(char32_t cp) {
  return (_properties::get_value(cp) & Property_Radical) != 0;
}

inline bool is_unified_ideograph(char32_t cp) {
  return (_properties::get_value(cp) & Property_Unified_Ideograph) != 0;
}

inline bool is_other_default_ignorable_code_point(char32_t cp) {
  return (_properties::get_value(cp) &
          Property_Other_Default_Ignorable_Code_Point) != 0;
}

inline bool is_deprecated(char32_t cp) {
  return (_properties::get_value(cp) & Property_Deprecated) != 0;
}

inline bool is_soft_dotted(char32_t cp) {
  return (_properties::get_value(cp) & Property_Soft_Dotted) != 0;
}

inline bool is_logical_order_exception(char32_t cp) {
  return (_properties::get_value(cp) & Property_Logical_Order_Exception) != 0;
}

inline bool is_other_id_start(char32_t cp) {
  return (_properties::get_value(cp) & Property_Other_ID_Start) != 0;
}

inline bool is_other_id_continue(char32_t cp) {
  return (_properties::get_value(cp) & Property_Other_ID_Continue) != 0;
}

inline bool is_sentence_terminal(char32_t cp) {
  return (_properties::get_value(cp) & Property_Sentence_Terminal) != 0;
}

inline bool is_variation_selector(char32_t cp) {
  return (_properties::get_value(cp) & Property_Variation_Selector) != 0;
}

inline bool is_pattern_white_space(char32_t cp) {
  return (_properties::get_value(cp) & Property_Pattern_White_Space) != 0;
}

inline bool is_pattern_syntax(char32_t cp) {
  return (_properties::get_value(cp) & Property_Pattern_Syntax) != 0;
}

inline bool is_prepended_concatenation_mark(char32_t cp) {
  return (_properties::get_value(cp) & Property_Prepended_Concatenation_Mark) !=
         0;
}

//-----------------------------------------------------------------------------
// Derived Property
//-----------------------------------------------------------------------------

const uint32_t DerivedProperty_Math = 0b00000000000000000000000000000001;
const uint32_t DerivedProperty_Alphabetic = 0b00000000000000000000000000000010;
const uint32_t DerivedProperty_Lowercase = 0b00000000000000000000000000000100;
const uint32_t DerivedProperty_Uppercase = 0b00000000000000000000000000001000;
const uint32_t DerivedProperty_Cased = 0b00000000000000000000000000010000;
const uint32_t DerivedProperty_Case_Ignorable =
    0b00000000000000000000000000100000;
const uint32_t DerivedProperty_Changes_When_Lowercased =
    0b00000000000000000000000001000000;
const uint32_t DerivedProperty_Changes_When_Uppercased =
    0b00000000000000000000000010000000;
const uint32_t DerivedProperty_Changes_When_Titlecased =
    0b00000000000000000000000100000000;
const uint32_t DerivedProperty_Changes_When_Casefolded =
    0b00000000000000000000001000000000;
const uint32_t DerivedProperty_Changes_When_Casemapped =
    0b00000000000000000000010000000000;
const uint32_t DerivedProperty_ID_Start = 0b00000000000000000000100000000000;
const uint32_t DerivedProperty_ID_Continue = 0b00000000000000000001000000000000;
const uint32_t DerivedProperty_XID_Start = 0b00000000000000000010000000000000;
const uint32_t DerivedProperty_XID_Continue =
    0b00000000000000000100000000000000;
const uint32_t DerivedProperty_Default_Ignorable_Code_Point =
    0b00000000000000001000000000000000;
const uint32_t DerivedProperty_Grapheme_Extend =
    0b00000000000000010000000000000000;
const uint32_t DerivedProperty_Grapheme_Base =
    0b00000000000000100000000000000000;
const uint32_t DerivedProperty_Grapheme_Link =
    0b00000000000001000000000000000000;

inline bool is_math(char32_t cp) {
  return (_derived_core_properties::get_value(cp) & DerivedProperty_Math) != 0;
}

inline bool is_alphabetic(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Alphabetic) != 0;
}

inline bool is_lowercase(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Lowercase) != 0;
}

inline bool is_uppercase(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Uppercase) != 0;
}

inline bool is_cased(char32_t cp) {
  return (_derived_core_properties::get_value(cp) & DerivedProperty_Cased) != 0;
}

inline bool is_case_ignorable(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Case_Ignorable) != 0;
}

inline bool is_changes_when_lowercased(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Changes_When_Lowercased) != 0;
}

inline bool is_changes_when_uppercased(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Changes_When_Uppercased) != 0;
}

inline bool is_changes_when_titlecased(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Changes_When_Titlecased) != 0;
}

inline bool is_changes_when_casefolded(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Changes_When_Casefolded) != 0;
}

inline bool is_changes_when_casemapped(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Changes_When_Casemapped) != 0;
}

inline bool is_id_start(char32_t cp) {
  return (_derived_core_properties::get_value(cp) & DerivedProperty_ID_Start) !=
         0;
}

inline bool is_id_continue(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_ID_Continue) != 0;
}

inline bool is_xid_start(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_XID_Start) != 0;
}

inline bool is_xid_continue(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_XID_Continue) != 0;
}

inline bool is_default_ignorable_code_point(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Default_Ignorable_Code_Point) != 0;
}

inline bool is_grapheme_extend(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Grapheme_Extend) != 0;
}

inline bool is_grapheme_base(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Grapheme_Base) != 0;
}

inline bool is_grapheme_link(char32_t cp) {
  return (_derived_core_properties::get_value(cp) &
          DerivedProperty_Grapheme_Link) != 0;
}

//-----------------------------------------------------------------------------
// Other Property
//-----------------------------------------------------------------------------

inline int combining_class(char32_t cp) {
  return _normalization_properties::get_value(cp).combining_class;
}

//-----------------------------------------------------------------------------
// Case
//-----------------------------------------------------------------------------

inline char32_t simple_case_mapping(char32_t cp, CaseMappingType type) {
  using namespace std;
  auto it = _simple_case_mappings.find(cp);
  if (it != _simple_case_mappings.end()) {
    return it->second[static_cast<size_t>(type)];
  }
  return cp;
}

inline char32_t simple_uppercase_mapping(char32_t cp) {
  return simple_case_mapping(cp, CaseMappingType::Upper);
}

inline char32_t simple_lowercase_mapping(char32_t cp) {
  return simple_case_mapping(cp, CaseMappingType::Lower);
}

inline char32_t simple_titlecase_mapping(char32_t cp) {
  return simple_case_mapping(cp, CaseMappingType::Title);
}

inline char32_t simple_case_folding(char32_t cp) {
  auto it = _case_foldings.find(cp);
  if (it != _case_foldings.end()) {
    const auto &cf = it->second;
    if (cf.S) {
      return cf.S;
    }
    return cf.C;
  }
  return cp;
}

inline bool is_language_qualified(const char *user_lang,
                                  const char *spec_lang) {
  return !spec_lang || (user_lang && !strcmp(user_lang, spec_lang));
}

inline bool is_final_sigma(const char32_t *s32, size_t l, size_t i) {
  // C is preceded by a sequence consisting of a cased letter and
  // then zero or more case-ignorable characters, and C is not
  // followed by a sequence consisting of zero or more case-ignorable
  // characters and then a cased letter

  // Before C: \p{cased} (\p{case-ignorable})*
  auto il = static_cast<int>(l);
  auto ii = static_cast<int>(i);
  auto pos = ii - 1;
  while (pos >= 0 && is_case_ignorable(s32[pos])) {
    pos--;
  }
  if (pos < 0 || !is_cased(s32[pos])) {
    return false;
  }

  // After C: !((\p{case-ignorable})* \p{cased})
  pos = ii + 1;
  while (pos < il && is_case_ignorable(s32[pos])) {
    pos++;
  }
  if (pos < il && is_cased(s32[pos])) {
    return false;
  }

  return true;
}

inline bool has_class_230_or_0(char32_t cp) {
  auto cls = combining_class(cp);
  return cls == 230 || cls == 0;
}

inline bool is_after_soft_dotted(const char32_t *s32, size_t l, size_t i) {
  // There is a Soft_Dotted character before C, with no intervening character of
  // combining class 0 or 230 (Above).

  // Before C: [\p{Soft_Dotted}] ([^\p{ccc=230} \p{ccc=0}])*
  auto pos = static_cast<int>(i) - 1;
  if (pos >= static_cast<int>(l)) {
    return false;
  }
  while (pos >= 0 && !has_class_230_or_0(s32[pos])) {
    pos--;
  }
  if (pos < 0 || !is_soft_dotted(s32[pos])) {
    return false;
  }

  return true;
}

inline bool is_more_above(const char32_t *s32, size_t l, size_t i) {
  // C is followed by a character of combining class 230 (Above) with no
  // intervening character of combining class 0 or 230 (Above).

  // After C: [^\p{ccc=230}\p{ccc=0}]* [\p{ccc=230}]
  auto pos = static_cast<int>(i) - 1;
  while (pos < static_cast<int>(l) && !has_class_230_or_0(s32[pos])) {
    pos++;
  }
  if (pos < static_cast<int>(l) && combining_class(s32[pos]) != 230) {
    return false;
  }

  return true;
}

inline bool is_before_dot(const char32_t *s32, size_t l, size_t i) {
  // C is followed by combining dot above (U+0307). Any sequence of characters
  // with a combining class that is neither 0 nor 230 may intervene between the
  // current character and the combining dot above.

  // After C: ([^\p{ccc=230} \p{ccc=0}])* [\u0307]
  auto pos = static_cast<int>(i) + 1;
  while (pos < static_cast<int>(l) && !has_class_230_or_0(s32[pos])) {
    pos++;
  }
  if (pos < static_cast<int>(l) && s32[pos] != 0x0307) {
    return false;
  }

  return true;
}

inline bool is_after_i(const char32_t *s32, size_t l, size_t i) {
  // There is an uppercase I before C, and there is no intervening combining
  // character class 230 (Above) or 0.

  // Before C: [I] ([^\p{ccc=230} \p{ccc=0}])*
  auto pos = static_cast<int>(i) - 1;
  if (pos >= static_cast<int>(l)) {
    return false;
  }
  while (pos >= 0 && !has_class_230_or_0(s32[pos])) {
    pos--;
  }
  if (pos < 0 || s32[pos] != U'I') {
    return false;
  }

  return true;
}

inline void full_case_mapping(const char32_t *s32, size_t l, size_t i,
                              const char *lang, CaseMappingType type,
                              std::u32string &out) {
  // D135 A character C is defined to be cased if and only if C has the
  // Lowercase or Uppercase property or has a General_Category value of
  // Titlecase_Letter. • The Uppercase and Lowercase property values are
  // specified in the data file DerivedCoreProperties.txt in the Unicode
  // Character Database. The derived property Cased is also listed in
  // DerivedCoreProperties.txt.
  //
  // D136 A character C is defined to be case-ignorable if C has the value
  // MidLetter (ML), MidNumLet (MB), or Single_Quote (SQ) for the Word_Break
  // property or its General_Category is one of Nonspacing_Mark (Mn),
  // Enclosing_Mark (Me), Format (Cf), Modifier_Letter (Lm), or Modifier_Symbol
  // (Sk). • The Word_Break property is defined in the data file
  // WordBreakProperty.txt in the Unicode Character Database. • The derived
  // property Case_Ignorable is listed in the data file
  // DerivedCoreProperties.txt in the Unicode Character Database. • The
  // Case_Ignorable property is defined for use in the context specifications of
  // Table 3-17. It is a narrow-use property, and is not intended for use in
  // other contexts. The more broadly applicable string casing function,
  // isCased(X), is defined in D143.
  //
  // D137 Case-ignorable sequence: A sequence of zero or more case-ignorable
  // characters.
  //
  // D138 A character C is in a particular casing context for context-dependent
  // matching if and only if it matches the corresponding specification in Table
  // 3-17.
  auto cp = s32[i];
  auto count = _special_case_mappings.count(cp);
  if (count != 0) {
    auto r = _special_case_mappings.equal_range(cp);
    for (auto it = r.first; it != r.second; ++it) {
      const auto &sc = it->second;
      if (is_language_qualified(lang, sc.language)) {
        bool handle = false;
        switch (sc.context) {
          case SpecialCasingContext::Final_Sigma:
            handle = is_final_sigma(s32, l, i);
            break;
          case SpecialCasingContext::Not_Final_Sigma:
            handle = !is_final_sigma(s32, l, i);
            break;
          case SpecialCasingContext::After_Soft_Dotted:
            handle = is_after_soft_dotted(s32, l, i);
            break;
          case SpecialCasingContext::More_Above:
            handle = is_more_above(s32, l, i);
            break;
          case SpecialCasingContext::Before_Dot:
            handle = is_before_dot(s32, l, i);
            break;
          case SpecialCasingContext::Not_Before_Dot:
            handle = !is_before_dot(s32, l, i);
            break;
          case SpecialCasingContext::After_I:
            handle = is_after_i(s32, l, i);
            break;
          default:
            // NOTREACHED
            break;
        }
        if (handle) {
          out += sc.case_mapping_codes(type);
          return;
        }
      }
    }
  }

  auto it = _special_case_mappings_default.find(cp);
  if (it != _special_case_mappings_default.end()) {
    out += it->second.case_mapping_codes(type);
  } else {
    out += simple_case_mapping(s32[i], type);
  }
}

inline void uppercase_mapping(const char32_t *s32, size_t l, size_t i,
                              const char *lang, std::u32string &out) {
  full_case_mapping(s32, l, i, lang, CaseMappingType::Upper, out);
}

inline void lowercase_mapping(const char32_t *s32, size_t l, size_t i,
                              const char *lang, std::u32string &out) {
  full_case_mapping(s32, l, i, lang, CaseMappingType::Lower, out);
}

inline void titlecase_mapping(const char32_t *s32, size_t l, size_t i,
                              const char *lang, std::u32string &out) {
  full_case_mapping(s32, l, i, lang, CaseMappingType::Title, out);
}

inline std::u32string to_uppercase(const char32_t *s32, size_t l,
                                   const char *lang) {
  // R1 toUppercase(X): Map each character C in X to Uppercase_Mapping(C)
  std::u32string out;
  for (size_t i = 0; i < l; i++) {
    uppercase_mapping(s32, l, i, lang, out);
  }
  return out;
}

inline std::u32string to_lowercase(const char32_t *s32, size_t l,
                                   const char *lang) {
  // R2 toLowercase(X): Map each character C in X to Lowercase_Mapping(C)
  std::u32string out;
  for (size_t i = 0; i < l; i++) {
    lowercase_mapping(s32, l, i, lang, out);
  }
  return out;
}

inline std::u32string to_titlecase(const char32_t *s32, size_t l,
                                   const char *lang) {
  // R3 toTitlecase(X): Find the word boundaries in X according to Unicode
  // Standard Annex #29, “Unicode Text Segmentation.” For each word boundary,
  // find the first cased character F following the word boundary. If F exists,
  // map F to Titlecase_Mapping(F); then map all characters C between F and the
  // following word boundary to Lowercase_Mapping(C)
  std::u32string out;
  size_t i = 0;
  while (i < l) {
    while (i < l && !is_cased(s32[i])) {
      out += s32[i];
      i++;
    }

    if (i == l) {
      break;
    }

    titlecase_mapping(s32, l, i, lang, out);
    i++;

    if (i == l) {
      break;
    }

    while (i < l && !is_word_boundary(s32, l, i)) {
      lowercase_mapping(s32, l, i, lang, out);
      i++;
    }
  }
  return out;
}

inline void case_folding(
    char32_t cp, bool special_case_for_uppercase_I_and_dotted_uppercase_I,
    std::u32string &out) {
  auto it = _case_foldings.find(cp);
  if (it != _case_foldings.end()) {
    const auto &cf = it->second;
    if (special_case_for_uppercase_I_and_dotted_uppercase_I && cf.T) {
      out += cf.T;
      return;
    } else if (cf.F) {
      out += cf.F;
      return;
    } else if (cf.S) {
      out += cf.S;
      return;
    } else if (cf.C) {
      out += cf.C;
      return;
    }
  }
  out += cp;
}

inline std::u32string to_case_fold(
    const char32_t *s32, size_t l,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I) {
  // R4 toCasefold(X): Map each character C in X to Case_Folding(C)
  std::u32string out;
  for (size_t i = 0; i < l; i++) {
    case_folding(s32[i], special_case_for_uppercase_I_and_dotted_uppercase_I,
                 out);
  }
  return out;
}

inline bool is_uppercase(const char32_t *s32, size_t l) {
  // D140 isUppercase(X): isUppercase(X) is true when toUppercase(Y) = Y
  for (size_t i = 0; i < l; i++) {
    if (is_changes_when_uppercased(s32[i])) {
      return false;
    }
  }
  return true;
}

inline bool is_lowercase(const char32_t *s32, size_t l) {
  // D139 isLowercase(X): isLowercase(X) is true when toLowercase(Y) = Y
  for (size_t i = 0; i < l; i++) {
    if (is_changes_when_lowercased(s32[i])) {
      return false;
    }
  }
  return true;
}

inline bool is_titlecase(const char32_t *s32, size_t l) {
  // D141 isTitlecase(X): isTitlecase(X) is true when toTitlecase(Y) = Y
  size_t i = 0;
  while (i < l) {
    while (i < l && !is_cased(s32[i])) {
      if (is_changes_when_lowercased(s32[i])) {
        return false;
      }
      i++;
    }

    if (i == l) {
      break;
    }

    if (is_changes_when_titlecased(s32[i])) {
      return false;
    }
    i++;

    if (i == l) {
      break;
    }

    while (i < l && !is_word_boundary(s32, l, i)) {
      if (is_changes_when_lowercased(s32[i])) {
        return false;
      }
      i++;
    }
  }

  return true;
}

inline bool is_case_fold(const char32_t *s32, size_t l) {
  // D142 isCasefolded(X): isCasefolded(X) is true when toCasefold(Y) = Y
  for (size_t i = 0; i < l; i++) {
    if (is_changes_when_casefolded(s32[i])) {
      return false;
    }
  }
  return true;
}

inline bool caseless_match(
    const char32_t *s1, size_t l1, const char32_t *s2, size_t l2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I) {
  // D144 A string X is a caseless match for a string Y if and only if
  // toCasefold(X) = toCasefold(Y)
  return to_case_fold(s1, l1,
                      special_case_for_uppercase_I_and_dotted_uppercase_I) ==
         to_case_fold(s2, l2,
                      special_case_for_uppercase_I_and_dotted_uppercase_I);
}

inline bool canonical_caseless_match(
    const char32_t *s1, size_t l1, const char32_t *s2, size_t l2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I) {
  // D145 A string X is a canonical caseless match for a string Y if and only if
  // NFD(toCasefold(NFD(X))) = NFD(toCasefold(NFD(Y)))
  return to_nfd(to_case_fold(
             to_nfd(s1, l1),
             special_case_for_uppercase_I_and_dotted_uppercase_I)) ==
         to_nfd(
             to_case_fold(to_nfd(s2, l2),
                          special_case_for_uppercase_I_and_dotted_uppercase_I));
}

inline bool compatibility_caseless_match(
    const char32_t *s1, size_t l1, const char32_t *s2, size_t l2,
    bool special_case_for_uppercase_I_and_dotted_uppercase_I) {
  // D146 A string X is a compatibility caseless match for a string Y if and
  // only if NFKD(toCasefold(NFKD(toCasefold(NFD(X))))) =
  // NFKD(toCasefold(NFKD(toCasefold(NFD(Y)))))
  return to_nfkd(to_case_fold(
             to_nfkd(to_case_fold(
                 to_nfd(s1, l1),
                 special_case_for_uppercase_I_and_dotted_uppercase_I)),
             special_case_for_uppercase_I_and_dotted_uppercase_I)) ==
         to_nfkd(to_case_fold(
             to_nfkd(to_case_fold(
                 to_nfd(s2, l2),
                 special_case_for_uppercase_I_and_dotted_uppercase_I)),
             special_case_for_uppercase_I_and_dotted_uppercase_I));
}

//-----------------------------------------------------------------------------
// Combining Character Sequence
//-----------------------------------------------------------------------------

inline bool is_graphic_character(char32_t cp) {
  // D50 Graphic character: A character with the General Category of Letter (L),
  // Combining Mark (M), Number (N), Punctuation (P), Symbol (S), or Space
  // Separator (Zs).
  return is_base_character(cp) || is_mark(cp);
}

inline bool is_base_character(char32_t cp) {
  // D51 Base character: Any graphic character except for those with the General
  // Category of Combining Mark (M).
  auto gc = general_category(cp);
  switch (gc) {
    case GeneralCategory::Zs:
      return true;
    default:
      return is_letter_category(gc) || is_number_category(gc) ||
             is_punctuation_category(gc) || is_symbol_category(gc);
  }
}

inline bool is_standard_korean_syllable_block(const char32_t *s32, size_t l,
                                              size_t &length) {
  // D134 Standard Korean syllable block: A sequence of one or more L followed
  // by a sequence of one or more V and a sequence of zero or more T, or any
  // other sequence that is canonically equivalent.
  size_t i = 0;
  if (i == l ||
      _grapheme_break_properties::get_value(s32[i]) != GraphemeBreak::L) {
    return false;
  }
  i++;
  while (i < l &&
         _grapheme_break_properties::get_value(s32[i]) == GraphemeBreak::L) {
    i++;
  }
  if (i == l ||
      _grapheme_break_properties::get_value(s32[i]) != GraphemeBreak::V) {
    return false;
  }
  i++;
  while (i < l &&
         _grapheme_break_properties::get_value(s32[i]) == GraphemeBreak::V) {
    i++;
  }
  while (i < l ||
         _grapheme_break_properties::get_value(s32[i]) == GraphemeBreak::T) {
    i++;
  }
  length = i;
  return true;
}

inline bool is_extended_base(const char32_t *s32, size_t l, size_t &length) {
  // D51a Extended base: Any base character, or any standard Korean syllable
  // block.
  if (l > 0) {
    if (is_standard_korean_syllable_block(s32, l, length)) {
      return true;
    } else if (is_base_character(s32[0])) {
      length = 1;
      return true;
    }
  }
  return false;
}

inline bool is_combining_character(char32_t cp) {
  // D52 Combining character: A character with the General Category of Combining
  // Mark (M)
  return is_mark(cp);
}

const char32_t ZERO_WIDTH_JOINER = 0x200D;
const char32_t ZERO_WIDTH_NON_JOINER = 0x200C;

inline size_t combining_character_sequence_length(const char32_t *s32,
                                                  size_t l) {
  // D56 Combining character sequence: A maximal character sequence consisting
  // of either a base character followed by a sequence of one or more characters
  // where each is a combining character, zero width joiner, or zero width
  // non-joiner; or a sequence of one or more characters where each is a
  // combining character, zero width joiner, or zero width non-joiner.
  size_t i = 0;
  if (l) {
    if (is_base_character(s32[i])) {
      i++;
    }
    while (i < l) {
      auto cp = s32[i];
      if (is_combining_character(cp) || cp == ZERO_WIDTH_JOINER ||
          cp == ZERO_WIDTH_NON_JOINER) {
        i++;
      } else {
        break;
      }
    }
  }
  return i;
}

inline size_t extended_combining_character_sequence_length(const char32_t *s32,
                                                           size_t l) {
  // D56a Extended combining character sequence: A maximal character sequence
  // consisting of either an extended base followed by a sequence of one or more
  // characters where each is a combining character, zero width joiner, or zero
  // width non-joiner; or a sequence of one or more characters where each is a
  // combining character, zero width joiner, or zero width non-joiner.
  size_t i = 0;
  if (l) {
    size_t length;
    if (is_extended_base(s32, l, length)) {
      i += length;
    }
    while (i < l) {
      auto cp = s32[i];
      if (is_combining_character(cp) || cp == ZERO_WIDTH_JOINER ||
          cp == ZERO_WIDTH_NON_JOINER) {
        i++;
      } else {
        break;
      }
    }
  }
  return i;
}

inline size_t combining_character_sequence_count(const char32_t *s32,
                                                 size_t l) {
  size_t count = 0;
  size_t i = 0;
  while (i < l) {
    count++;
    i += combining_character_sequence_length(s32 + i, l - i);
  }
  return count;
}

inline size_t extended_combining_character_sequence_count(const char32_t *s32,
                                                          size_t l) {
  size_t count = 0;
  size_t i = 0;
  while (i < l) {
    count++;
    i += extended_combining_character_sequence_length(s32 + i, l - i);
  }
  return count;
}

//-----------------------------------------------------------------------------
// Grapheme Cluster Segmentation
//-----------------------------------------------------------------------------

inline bool is_grapheme_boundary(const char32_t *s32, size_t l, size_t i) {
  //---------------------------------------------------------------------------
  // Break at the start and end of text, unless the text empty.
  //---------------------------------------------------------------------------

  // GB1: sot ÷
  if (i == 0) {
    return true;
  }

  // GB2: ÷ eot
  if (i == l) {
    return true;
  }

  auto lp = _grapheme_break_properties::get_value(s32[i - 1]);
  auto rp = _grapheme_break_properties::get_value(s32[i]);

  //---------------------------------------------------------------------------
  // Do not break between a CR and LF. Otherwise, break before and after
  // controls.
  //---------------------------------------------------------------------------

  // GB3: CR × LF
  if ((lp == GraphemeBreak::CR) && (rp == GraphemeBreak::LF)) {
    return false;
  }

  // GB4: (Control|CR|LF) ÷
  if ((lp == GraphemeBreak::Control || lp == GraphemeBreak::CR ||
       lp == GraphemeBreak::LF)) {
    return true;
  }

  // GB5: ÷ (Control|CR|LF)
  if ((rp == GraphemeBreak::Control || rp == GraphemeBreak::CR ||
       rp == GraphemeBreak::LF)) {
    return true;
  }

  //---------------------------------------------------------------------------
  // Do not break Hangul syllable sequences.
  //---------------------------------------------------------------------------

  // GB6: L × (L|V|LV|LVT)
  if ((lp == GraphemeBreak::L) &&
      (rp == GraphemeBreak::L || rp == GraphemeBreak::V ||
       rp == GraphemeBreak::LV || rp == GraphemeBreak::LVT)) {
    return false;
  }

  // GB7: (LV|V) × (V|T)
  if ((lp == GraphemeBreak::LV || lp == GraphemeBreak::V) &&
      (rp == GraphemeBreak::V || rp == GraphemeBreak::T)) {
    return false;
  }

  // GB8: (LVT|T) × T
  if ((lp == GraphemeBreak::LVT || lp == GraphemeBreak::T) &&
      (rp == GraphemeBreak::T)) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Do not break before extending characters or ZWJ.
  //---------------------------------------------------------------------------

  // GB9: × Extend
  if (rp == GraphemeBreak::Extend || rp == GraphemeBreak::ZWJ) {
    return false;
  }

  //---------------------------------------------------------------------------
  // The GB9a and GB9b rules only apply to extended grapheme clusters:
  // Do not break before SpacingMakrs, or after Prepend characters.
  //---------------------------------------------------------------------------

  // GB9a: × SpacingMark
  if (rp == GraphemeBreak::SpacingMark) {
    return false;
  }

  // GB9b: Prepend ×
  if (lp == GraphemeBreak::Prepend) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Do not break within emoji modifier sequences or emoji zwj sewuences.
  //---------------------------------------------------------------------------

  // GB11: \p{Extended_Pictographic} Extend* ZWJ x \p{Extended_Pictographic}
  {
    auto rpEmoji = _emoji_properties::get_value(s32[i]);

    if (lp == GraphemeBreak::ZWJ && rpEmoji == Emoji::Extended_Pictographic) {
      auto pos = static_cast<int>(i) - 2;
      while (pos >= 0 && _grapheme_break_properties::get_value(s32[pos]) ==
                             GraphemeBreak::Extend) {
        pos--;
      }
      if (pos >= 0) {
        auto lpEmoji = _emoji_properties::get_value(s32[pos]);
        if (lpEmoji == Emoji::Extended_Pictographic) {
          return false;
        }
      }
    }
  }

  //---------------------------------------------------------------------------
  // Do not break within emoji flag sequences. That is, do not break between
  // regional indicator (RI) symbols if there is an odd number of RI
  // characters before the break point.
  //---------------------------------------------------------------------------

  // GB12: ^ (RI RI)* RI x RI
  // GB13: [^RI] (RI RI)* RI x RI
  if (lp == GraphemeBreak::Regional_Indicator &&
      rp == GraphemeBreak::Regional_Indicator) {
    auto pos = static_cast<int>(i) - 2;
    while (pos >= 1 &&
           _grapheme_break_properties::get_value(s32[pos]) ==
               GraphemeBreak::Regional_Indicator &&
           _grapheme_break_properties::get_value(s32[pos - 1]) ==
               GraphemeBreak::Regional_Indicator) {
      pos -= 2;
    }
    if (pos < 0) {
      return false;
    }
    if (_grapheme_break_properties::get_value(s32[pos]) !=
        GraphemeBreak::Regional_Indicator) {
      return false;
    }
  }

  //---------------------------------------------------------------------------
  // Othewise, break everywhere.
  //---------------------------------------------------------------------------

  // GB999: Any ÷ Any
  return true;
}

inline size_t grapheme_length(const char32_t *s32, size_t l) {
  size_t i = 1;
  for (; i < l; i++) {
    if (is_grapheme_boundary(s32, l, i)) {
      return i;
    }
  }
  return i;
}

inline size_t grapheme_count(const char32_t *s32, size_t l) {
  size_t count = 0;
  size_t i = 0;
  while (i < l) {
    count++;
    i += grapheme_length(s32 + i, l - i);
  }
  return count;
}

//-----------------------------------------------------------------------------
// Word Segmentation
//-----------------------------------------------------------------------------

inline bool AHLetter(WordBreak p) {
  return p == WordBreak::ALetter || p == WordBreak::Hebrew_Letter;
}

inline bool MidNumLetQ(WordBreak p) {
  return p == WordBreak::MidNumLet || p == WordBreak::Single_Quote;
}

inline int previous_word_break_property_position(const char32_t *s32,
                                                 size_t i) {
  auto prop = WordBreak::Unassigned;
  auto pos = static_cast<int>(i) - 1;
  while (pos >= 0) {
    prop = _word_break_properties::get_value(s32[pos]);
    if (prop != WordBreak::Extend && prop != WordBreak::Format &&
        prop != WordBreak::ZWJ) {
      break;
    }
    pos--;
  }
  return pos;
}

inline size_t next_word_break_property_position(const char32_t *s32, size_t l,
                                                size_t i) {
  auto prop = WordBreak::Unassigned;
  auto pos = i + 1;
  while (pos < l) {
    prop = _word_break_properties::get_value(s32[pos]);
    if (prop != WordBreak::Extend && prop != WordBreak::Format &&
        prop != WordBreak::ZWJ) {
      break;
    }
    pos++;
  }
  return pos;
}

inline bool is_word_boundary(const char32_t *s32, size_t l, size_t i) {
  //---------------------------------------------------------------------------
  // Break at the start and end of text, unless the text is empty
  //---------------------------------------------------------------------------

  // WB1: sot ÷
  if (i == 0) {
    return true;
  }

  // WB2: ÷ eot
  if (i == l) {
    return true;
  }

  auto lp = _word_break_properties::get_value(s32[i - 1]);
  auto rp = _word_break_properties::get_value(s32[i]);

  //---------------------------------------------------------------------------
  // Do not break within CRLF
  //---------------------------------------------------------------------------

  // WB3: CR × LF
  if ((lp == WordBreak::CR) && (rp == WordBreak::LF)) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Otherwise break before and after Newlines (including CR and LF)
  //---------------------------------------------------------------------------

  // WB3a: (Newline|CR|LF) ÷
  if ((lp == WordBreak::Newline || lp == WordBreak::CR ||
       lp == WordBreak::LF)) {
    return true;
  }

  // WB3b: ÷ (Newline|CR|LF)
  if ((rp == WordBreak::Newline || rp == WordBreak::CR ||
       rp == WordBreak::LF)) {
    return true;
  }

  //---------------------------------------------------------------------------
  // Do not break within emoji zwj sequences.
  //---------------------------------------------------------------------------

  // WB3c: ZWJ x \p{Extended_Pictographic}
  {
    auto rpEmoji = _emoji_properties::get_value(s32[i]);

    if (lp == WordBreak::ZWJ && rpEmoji == Emoji::Extended_Pictographic) {
      return false;
    }
  }

  //---------------------------------------------------------------------------
  // Keep horizontal whitespace together.
  //---------------------------------------------------------------------------

  // WB3d: WSegSpace x WSegSpace
  if (lp == WordBreak::WSegSpace && rp == WordBreak::WSegSpace) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Ignore Format and Extend characters, except after sot, CR, LF, and
  // Newline. (See Section 6.2, Replacing Ignore Rules.) This also has the
  // effect of: Any × (Format | Extend | ZWJ)
  //---------------------------------------------------------------------------

  // WB4: X (Extend|Format|ZWJ)* → X
  if ((rp == WordBreak::Extend || rp == WordBreak::Format ||
       rp == WordBreak::ZWJ)) {
    return false;
  }

  // Find left property
  lp = WordBreak::Unassigned;
  auto lpos = previous_word_break_property_position(s32, i);
  if (lpos >= 0) {
    lp = _word_break_properties::get_value(s32[lpos]);
  }

  //---------------------------------------------------------------------------
  // Do not break between most letters.
  //---------------------------------------------------------------------------

  // WB5: AHLetter × AHLetter
  if (AHLetter(lp) && AHLetter(rp)) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Do not break across certain punctuation.
  //---------------------------------------------------------------------------

  auto rp1 = WordBreak::Unassigned;
  auto rpos = next_word_break_property_position(s32, l, i);
  if (rpos < l) {
    rp1 = _word_break_properties::get_value(s32[rpos]);
  }

  // WB6: AHLetter × (MidLetter | MidNumLetQ) AHLetter
  if ((AHLetter(lp)) &&
      ((rp == WordBreak::MidLetter || MidNumLetQ(rp)) && AHLetter(rp1))) {
    return false;
  }

  auto lp1 = WordBreak::Unassigned;
  lpos = previous_word_break_property_position(s32, lpos);
  if (lpos >= 0) {
    lp1 = _word_break_properties::get_value(s32[lpos]);
  }

  // WB7: AHLetter (MidLetter | MidNumLetQ) × AHLetter
  if ((AHLetter(lp1) && (lp == WordBreak::MidLetter || MidNumLetQ(lp))) &&
      (AHLetter(rp))) {
    return false;
  }

  // WB7a: Hebrew_Letter × Single_Quote
  if ((lp == WordBreak::Hebrew_Letter) && (rp == WordBreak::Single_Quote)) {
    return false;
  }

  // WB7b: Hebrew_Letter × Double_Quote Hebrew_Letter
  if ((lp == WordBreak::Hebrew_Letter) &&
      (rp == WordBreak::Double_Quote && rp1 == WordBreak::Hebrew_Letter)) {
    return false;
  }

  // WB7c: Hebrew_Letter Double_Quote × Hebrew_Letter
  if ((lp1 == WordBreak::Hebrew_Letter && lp == WordBreak::Double_Quote) &&
      (rp == WordBreak::Hebrew_Letter)) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Do not break within sequences of digits, or digits adjacent to letters
  // ("3a", or "A3").
  //---------------------------------------------------------------------------

  // WB8: Numeric × Numeric
  if ((lp == WordBreak::Numeric) && (rp == WordBreak::Numeric)) {
    return false;
  }

  // WB9: AHLetter × Numeric
  if ((AHLetter(lp)) && (rp == WordBreak::Numeric)) {
    return false;
  }

  // WB10: Numeric × AHLetter
  if ((lp == WordBreak::Numeric) && (AHLetter(rp))) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Do not break within sequences, such as "3.2" or "3,456.789"
  //---------------------------------------------------------------------------

  // WB11: Numeric (MidNum | MidNumLetQ) × Numeric
  if ((lp1 == WordBreak::Numeric &&
       (lp == WordBreak::MidNum || MidNumLetQ(lp))) &&
      (rp == WordBreak::Numeric)) {
    return false;
  }

  // WB12: Numeric × (MidNum | MidNumLetQ) Numeric
  if ((lp == WordBreak::Numeric) &&
      ((rp == WordBreak::MidNum || MidNumLetQ(rp)) &&
       rp1 == WordBreak::Numeric)) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Do not break between Katakana.
  //---------------------------------------------------------------------------

  // WB13: Katakana × Katakana
  if ((lp == WordBreak::Katakana) && (rp == WordBreak::Katakana)) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Do not break from extenders.
  //---------------------------------------------------------------------------

  // WB13a: (AHLetter | Numeric | Katakana | ExtendNumLet) × ExtendNumLet
  if ((AHLetter(lp) || lp == WordBreak::Katakana || lp == WordBreak::Numeric ||
       lp == WordBreak::Katakana || lp == WordBreak::ExtendNumLet) &&
      (rp == WordBreak::ExtendNumLet)) {
    return false;
  }

  // WB13b: ExtendNumLet × (AHLetter | Numeric | Katakana)
  if ((lp == WordBreak::ExtendNumLet) &&
      (AHLetter(rp) || rp == WordBreak::Numeric || rp == WordBreak::Katakana)) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Do not break within emoji flag sequences. That is, do not break between
  // regional indicator (RI) symbols if there is an odd number of RI
  // characters before the break point.
  //---------------------------------------------------------------------------

  // WB15: ^ (RI RI)* RI x RI
  // WB16: [^RI] (RI RI)* RI x RI
  {
    if (lp == WordBreak::Regional_Indicator &&
        rp == WordBreak::Regional_Indicator) {
      auto lpos = previous_word_break_property_position(s32, i);

      while (true) {
        lpos = previous_word_break_property_position(s32, lpos);
        if (lpos < 0 || _word_break_properties::get_value(s32[lpos]) !=
                            WordBreak::Regional_Indicator) {
          return false;
        }

        lpos = previous_word_break_property_position(s32, lpos);
        if (lpos < 0 || _word_break_properties::get_value(s32[lpos]) !=
                            WordBreak::Regional_Indicator) {
          break;
        }
      }
    }
  }

  //---------------------------------------------------------------------------
  // Othewise, break everywhere.
  //---------------------------------------------------------------------------

  // WB14: Any ÷ Any
  return true;
}

//-----------------------------------------------------------------------------
// Sentence Segmentation
//-----------------------------------------------------------------------------

inline bool ParaSep(SentenceBreak p) {
  return p == SentenceBreak::Sep || p == SentenceBreak::CR ||
         p == SentenceBreak::LF;
}

inline bool SATerm(SentenceBreak p) {
  return p == SentenceBreak::STerm || p == SentenceBreak::ATerm;
}

inline int previous_sentence_break_property_position(const char32_t *s32,
                                                     size_t i) {
  auto prop = SentenceBreak::Unassigned;
  auto pos = static_cast<int>(i) - 1;
  while (pos >= 0) {
    prop = _sentence_break_properties::get_value(s32[pos]);
    if (prop != SentenceBreak::Extend && prop != SentenceBreak::Format) {
      break;
    }
    pos--;
  }
  return pos;
}

inline size_t next_sentence_break_property_position(const char32_t *s32,
                                                    size_t l, size_t i) {
  auto pos = i + 1;
  while (pos < l) {
    auto prop = _sentence_break_properties::get_value(s32[pos]);
    if (prop != SentenceBreak::Extend && prop != SentenceBreak::Format) {
      break;
    }
    pos++;
  }
  return pos;
}

inline bool is_sentence_boundary(const char32_t *s32, size_t l, size_t i) {
  //---------------------------------------------------------------------------
  // Break at the start and end of text, unless the text is empty.
  //---------------------------------------------------------------------------

  // SB1: sot ÷
  if (i == 0) {
    return true;
  }

  // SB2: ÷ eot
  if (i == l) {
    return true;
  }

  //---------------------------------------------------------------------------
  // Do not break within CRLF.
  //---------------------------------------------------------------------------

  auto lp = _sentence_break_properties::get_value(s32[i - 1]);
  auto rp = _sentence_break_properties::get_value(s32[i]);

  // SB3: CR × LF
  if ((lp == SentenceBreak::CR) && (rp == SentenceBreak::LF)) {
    return false;
  }

  //---------------------------------------------------------------------------
  // Break after paragraph separators.
  //---------------------------------------------------------------------------

  // SB4: ParaSep ÷
  if (ParaSep(lp)) {
    return true;
  }

  //---------------------------------------------------------------------------
  // Ignore Format and Extend characters, except after sot, ParaSep, and
  // within CRLF. (See Section 6.2, Replacing Ignore Rules.) This also has the
  // effect of: Any × (Format | Extend)
  //---------------------------------------------------------------------------

  // SB5: X (Extend | Format)* → X
  if ((rp == SentenceBreak::Extend || rp == SentenceBreak::Format)) {
    return false;
  }

  // Find left property
  lp = SentenceBreak::Unassigned;
  auto lpos = previous_sentence_break_property_position(s32, i);
  if (lpos >= 0) {
    lp = _sentence_break_properties::get_value(s32[lpos]);
  }

  //---------------------------------------------------------------------------
  // Do not break after full stop in certain contexts.
  //---------------------------------------------------------------------------

  // SB6: ATerm × Numeric
  if ((lp == SentenceBreak::ATerm) && (rp == SentenceBreak::Numeric)) {
    return false;
  }

  auto lp1 = SentenceBreak::Unassigned;
  if (lpos > 0) {
    lpos = previous_sentence_break_property_position(s32, lpos);
    if (lpos >= 0) {
      lp1 = _sentence_break_properties::get_value(s32[lpos]);
    }
  }

  // SB7: (Upper | Lower) ATerm × Upper
  if (((lp1 == SentenceBreak::Upper || lp1 == SentenceBreak::Lower) &&
       (lp == SentenceBreak::ATerm)) &&
      (rp == SentenceBreak::Upper)) {
    return false;
  }

  auto lp2 = SentenceBreak::Unassigned;
  {
    lp2 = SentenceBreak::Unassigned;
    auto pos = previous_sentence_break_property_position(s32, i);
    while (pos >= 0) {
      lp2 = _sentence_break_properties::get_value(s32[pos]);
      if (lp2 != SentenceBreak::Sp) {
        break;
      }
      pos = previous_sentence_break_property_position(s32, pos);
    }
    while (pos >= 0) {
      lp2 = _sentence_break_properties::get_value(s32[pos]);
      if (lp2 != SentenceBreak::Close) {
        break;
      }
      pos = previous_sentence_break_property_position(s32, pos);
    }
  }

  auto rp2 = SentenceBreak::Unassigned;
  {
    size_t pos = i;
    while (pos < l) {
      rp2 = _sentence_break_properties::get_value(s32[pos]);
      if (ParaSep(rp2) || SATerm(rp2) || rp2 == SentenceBreak::OLetter ||
          rp2 == SentenceBreak::Upper || rp2 == SentenceBreak::Lower) {
        break;
      }
      pos = next_sentence_break_property_position(s32, l, pos);
    }
  }

  // SB8: ATerm Close* Sp* × (¬(OLetter | Upper | Lower | ParaSep | SATerm))*
  // Lower
  if ((lp2 == SentenceBreak::ATerm) && (rp2 == SentenceBreak::Lower)) {
    return false;
  }

  // SB8a: SATerm Close* Sp* × (SContinue | SATerm)
  if ((SATerm(lp2)) && (rp == SentenceBreak::SContinue || SATerm(rp))) {
    return false;
  }

  auto lp3 = SentenceBreak::Unassigned;
  {
    auto pos = previous_sentence_break_property_position(s32, i);
    while (pos >= 0) {
      lp3 = _sentence_break_properties::get_value(s32[pos]);
      if (lp3 != SentenceBreak::Close) {
        break;
      }
      pos = previous_sentence_break_property_position(s32, pos);
    }
  }

  //---------------------------------------------------------------------------
  // Break after sentence terminators, but include closing punctuation,
  // trailing spaces, and any paragraph separator.
  //---------------------------------------------------------------------------

  // SB9: SATerm Close* × (Close | Sp | ParaSep)
  if ((SATerm(lp3)) &&
      (rp == SentenceBreak::Close || rp == SentenceBreak::Sp || ParaSep(rp))) {
    return false;
  }

  // SB10: SATerm Close* Sp* × (Sp | ParaSep)
  if ((SATerm(lp2)) && (rp == SentenceBreak::Sp || ParaSep(rp))) {
    return false;
  }

  // SB11: SATerm Close* Sp* ParaSep? ÷
  // NOTE: 'ParaSep?' has been handled in SB4.
  if (SATerm(lp2)) {
    return true;
  }

  //---------------------------------------------------------------------------
  // Otherwise, do not break.
  //---------------------------------------------------------------------------

  // SB998: Any × Any
  return false;
}

//-----------------------------------------------------------------------------
// Block
//-----------------------------------------------------------------------------

inline Block block(char32_t cp) { return _block_properties::get_value(cp); }

//-----------------------------------------------------------------------------
// Script
//-----------------------------------------------------------------------------

inline Script script(char32_t cp) { return _script_properties::get_value(cp); }

inline bool is_script(Script sc, char32_t cp) {
  auto val = script(cp);
  switch (val) {
    case Script::Common:
    case Script::Inherited: {
      auto id = _script_extension_ids::get_value(cp);
      const auto &props = _script_extension_properties_for_id[id];
      return std::find(props.begin(), props.end(), sc) != props.end();
    }
    default:
      return sc == val;
  }
}

//-----------------------------------------------------------------------------
// Normalization
//-----------------------------------------------------------------------------

enum class Normalization {
  NFC,
  NFD,
  NFKC,
  NFKD,
};

inline void decompose_code(const char32_t cp, std::u32string &out,
                           Normalization norm) {
  if (hangul::is_precomposed_syllable(cp)) {
    hangul::decompose_hangul(cp, out);
  } else {
    const auto &prop = _normalization_properties::get_value(cp);
    if (prop.codes && (!prop.compat_format || norm == Normalization::NFKC ||
                       norm == Normalization::NFKD)) {
      size_t i = 0;
      while (prop.codes[i]) {
        decompose_code(prop.codes[i], out, norm);
        i++;
      }
    } else {
      out += cp;
    }
  }
}

inline std::u32string decompose(const char32_t *s32, size_t l,
                                Normalization norm) {
  std::u32string out;

  // Decompose
  for (size_t i = 0; i < l; i++) {
    decompose_code(s32[i], out, norm);
  }

  // Reorder combining marks with 'Canonical Ordering Algorithm'.
  for (size_t i = 0; i < out.length(); i++) {
    const auto &prop = _normalization_properties::get_value(out[i]);
    if (prop.combining_class > 0) {
      for (size_t j = i; j > 0; j--) {
        auto prev = out[j - 1];
        auto curr = out[j];
        if (combining_class(prev) <= combining_class(curr)) {
          break;
        }
        std::swap(out[j - 1], out[j]);
      }
    }
  }

  return out;
}

inline bool compose_pair(char32_t cp0, char32_t cp1, char32_t &cp) {
  std::u32string key = {cp0, cp1};
  auto it = _normalization_composition.find(key);
  if (it != _normalization_composition.end()) {
    cp = it->second;
    return true;
  }
  return false;
}

inline size_t compose_codes(const char32_t *s32, size_t l,
                            std::u32string &out) {
  auto starter = s32[0];
  std::vector<bool> cheched(l, false);

  // Compose successively
  bool handled = true;
  while (handled) {
    handled = false;
    int max_class = -1;
    size_t i = 1;
    for (; i < l; i++) {
      if (!cheched[i]) {
        auto klass =
            _normalization_properties::get_value(s32[i]).combining_class;
        if (max_class < klass) {
          if (compose_pair(starter, s32[i], starter)) {
            handled = true;
            cheched[i] = true;
            break;
          }
        }
        if (klass == 0) {
          break;
        }
        max_class = std::max(max_class, klass);
      }
    }
  }

  // Output
  out += starter;
  size_t i = 1;
  for (; i < l; i++) {
    if (!cheched[i]) {
      if (_normalization_properties::get_value(s32[i]).combining_class == 0) {
        break;
      }
      out += s32[i];
    }
  }

  return i;
}

inline std::u32string compose(const std::u32string_view s32) {
  std::u32string out;
  size_t i = 0;
  while (i < s32.length()) {
    if (hangul::is_decomposed_syllable(s32.data() + i, s32.length() - i)) {
      i += hangul::compose_hangul(s32.data() + i, s32.length() - i, out);
    } else {
      i += compose_codes(s32.data() + i, s32.length() - i, out);
    }
  }
  return out;
}

inline std::u32string to_nfc(const char32_t *s32, size_t l) {
  return compose(decompose(s32, l, Normalization::NFC));
}

inline std::u32string to_nfd(const char32_t *s32, size_t l) {
  return decompose(s32, l, Normalization::NFD);
}

inline std::u32string to_nfkc(const char32_t *s32, size_t l) {
  return compose(decompose(s32, l, Normalization::NFKC));
}

inline std::u32string to_nfkd(const char32_t *s32, size_t l) {
  return decompose(s32, l, Normalization::NFKD);
}

// ----------------------------------------------------------------------------

}  // namespace unicode

// vim: et ts=2 sw=2 cin cino=\:0 ff=unix
